/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package layouts 
{
	import com.adobe.portfolio.animations.AnimateFactory;
	import com.adobe.portfolio.cards.ICardProvider;
	import com.adobe.portfolio.cards.ISimpleCard;
	import com.adobe.portfolio.components.supportClasses.AutoListen;
	import com.adobe.portfolio.components.supportClasses.IListenAndWatch;
	import com.adobe.portfolio.events.LayoutEvent;
	import com.adobe.portfolio.events.ThumbnailManagerEvent;
	import com.adobe.portfolio.layouts.supportClasses.CardLayoutHelper;
	import com.adobe.portfolio.layouts.supportClasses.ICardLayout;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.PropertyManager;
	import com.adobe.portfolio.managers.ThumbnailManager;
	import com.adobe.portfolio.properties.PropertyContext;
	import com.adobe.portfolio.utils.MathUtil;
	
	import components.Panel;
	
	import events.PanelEvent;
	
	import flash.events.Event;
	import flash.events.IEventDispatcher;
	import flash.geom.Matrix3D;
	import flash.geom.PerspectiveProjection;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.geom.Vector3D;
	import flash.utils.Dictionary;
	
	import managers.PanelManager;
	
	import mx.binding.utils.BindingUtils;
	import mx.collections.IList;
	import mx.core.ILayoutElement;
	import mx.core.IUIComponent;
	import mx.core.IVisualElement;
	import mx.core.UIComponent;
	import mx.events.CollectionEvent;
	import mx.events.EffectEvent;
	import mx.events.ResizeEvent;
	import mx.logging.ILogger;
	import mx.logging.Log;
	
	import properties.CustomProperties;
	
	import spark.components.List;
	import spark.components.supportClasses.GroupBase;
	import spark.effects.Animate;
	import spark.effects.easing.EaseInOutBase;
	import spark.filters.BlurFilter;
	import spark.filters.ColorMatrixFilter;
	import spark.layouts.supportClasses.LayoutBase;
	
	import utils.StringUtil;
	
	/**
	 *  Dispatched after the layout has set its focus to a valid attachment.
	 *  @eventType com.adobe.portfolio.events.NavLayoutEvent.HANDLED_ATTACHMENT_FOCUS
	 */
	[Event(name="HANDLED_ATTACHMENT_FOCUS", type="com.adobe.portfolio.events.NavLayoutEvent")]
	
	/**
	 * The WaveLayout class extends LayoutBase to implement the layout of the Wave navigator.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class BorgesLayout extends LayoutBase implements IListenAndWatch, ICardLayout
	{
		[Bindable]
		public var customProperties:CustomProperties;

		private const logger:ILogger = Log.getLogger("BorgesLayout");
		
		protected const DEG2RAD:Number = 2 * Math.PI / 360;

		protected var geometryDict:Dictionary;
		protected var panels:Dictionary;
		protected var cardScrollInfo:CardScrollInfo;
		
		private var _cardLayoutHelper:CardLayoutHelper;
		private var _galleryGeometry:String;

		protected var autoListen:AutoListen;
		protected var fadeDelay:Number;
		protected var perspectiveProjection:PerspectiveProjection;
		protected var animate:Animate;
		protected var totalWidth:Number;
		protected var layoutWidth:Number;
		protected var focusPanelInvalid:Boolean;
		protected var sizeInvalid:Boolean;
		protected var panelSizesInvalid:Boolean;
		
		private var _panelWidth:Number;
		private var _panelHeight:Number;
		private var _fieldOfView:Number; // degrees
		private var _focusIndex:int;
		private var _zDistance:Number;
		private var _xAngle:Number;
		private var _spotlight:Boolean;

		private var _easer:EaseInOutBase;
		private var _thumbFitSize:Point;
		private var _cardPosition:Number;
		private var _maxPanelPosition:Number;
		private var _pendingCardPosition:Number;
		private var _cardProvider:ICardProvider;
		private var _cardsPerPanel:int;

		private static var blurFilters:Array;
		private static var colorFilters:Array;
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function BorgesLayout() 
		{
			super();
			initColorFilters();
			autoListen = new AutoListen();
			_fieldOfView = 25; 
			_focusIndex = 0;
			_cardPosition = 0;
			cardLayoutHelper = new CardLayoutHelper(this);
			cardScrollInfo = new CardScrollInfo();
			addListenAndWatch();
		}
		
		/**
		 *  Add event listeners and change watchers to the navigator.
		 * 
		 * @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function addListenAndWatch():void
		{
			autoListen.addListener(ThumbnailManager.instance, ThumbnailManagerEvent.ASPECT_RATIO_CHANGED, onAspectRatioChanged);
			autoListen.addListener(PanelManager.instance, PanelEvent.ORIENTATION_CHANGED, onPanelOrientationChange);
			BindingUtils.bindProperty(this, "customProperties", PropertyManager.instance, "commonProperties");
			BindingUtils.bindSetter(onCardPositionChanged, this, ["cardProvider", "cardPosition"]);
		}
		
		/**
		 *  Remove all event listeners and change watchers from the navigator
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function removeListenAndWatch():void
		{
			autoListen.removeAllListeners();
		}

		/**
		 *  Set the target host for this navigator
		 * 
		 *  @param value The GroupBase instance of the new target
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function set target(value:GroupBase):void 
		{
			autoListen.removeListener(target);
			super.target = value;
			autoListen.addListener(target, ResizeEvent.RESIZE, onResize);
		}
		
		protected function onAspectRatioChanged(event:Event):void
		{
			panelSizesInvalid = true;
		}
		
		protected function onCardPositionChanged(cardPosition:Number):void
		{
			scrollTo(cardPosition / cardsPerPanel);
		}
		
		protected function onResize(event:ResizeEvent):void
		{
			invalidateSize();
		}
		
		/**
		 *  Handles the bookkeeping when a ResizeEvent invalidates
		 *  the current panel sizes.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function invalidateSize():void
		{
			sizeInvalid = true;
			panelSizesInvalid = true;
			if (target)
				target.invalidateSize();
			invalidateGallery();
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function measure():void
		{
			super.measure();
			commitProperties();
		}
		
		protected function commitProperties():void
		{
			if (panelSizesInvalid) {
				resetPanelSizes();
			}
			if (sizeInvalid) {
				resetSize();
			}
			if (focusPanelInvalid) {
				scroll();
				if (target.numElements > 0)
					cardProvider.pendingCardPosition = focusIndex * cardsPerPanel;
				focusPanelInvalid = false;
			}
		}
		
		protected function resetSize():void {
			
			if (!target)
				return;
			
			var numElements:int = target.numElements;
			layoutWidth = target.width;
			totalWidth = layoutWidth + (numElements - 1) * panelDistance;
			updateMaxPanelPosition();
			target.setContentSize(totalWidth, target.height);
			resetPerspective();
			scroll(0);
			sizeInvalid = false;
		}

		protected function resetPerspective():void 
		{
			if (!target)
				return;
			
			if (!perspectiveProjection) {
				perspectiveProjection = new PerspectiveProjection();
			}
			perspectiveProjection.fieldOfView = fieldOfView;
			perspectiveProjection.projectionCenter = getProjectionCenter();
			target.parent.transform.perspectiveProjection = perspectiveProjection;
			invalidateTarget();
		}
		
		protected function getProjectionCenter():Point 
		{
			return target.localToGlobal(
				new Point (target.x + target.width/2, 
						   target.y + target.height/2));
		}

		protected function initColorFilters():void 
		{
			if (!colorFilters) {
				colorFilters = [];
				for (var i:int=0; i<256; i++) {
					var v:Number = 1 - i/255;
					// i.e. 1, 254/255, 253/255, etc.
					colorFilters[i] = new ColorMatrixFilter(
						[v,0,0,0,0,
							0,v,0,0,0,
							0,0,v,0,0,
							0,0,0,1,0]);
				}
			}
		}
		
		/**
		 *  Scroll to the new focus panel based on the focusIndex
		 * 
		 *  @param duration The time duration, in miliseconds, for the scrolling.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function scroll(duration:Number = NaN):void
		{
			if (focusIndex == -1)
				return;
			
			
			var newPanelPosition:Number = panelPositionFromIndex(focusIndex);
			if (newPanelPosition == panelPosition)
				return;
			
			if (isNaN(duration))
				duration = 750 * Math.abs(newPanelPosition - panelPosition) / panelDistance;
			if (duration > 10000)
				duration = 10000;
			else if (duration < 0 || isNaN(duration))
				duration = 0;
			
			enablePanels(false);

			if (animate) {
				animate.removeEventListener(EffectEvent.EFFECT_END, onEndScroll);
				animate.end();
			}
			
			ThumbnailManager.instance.pauseQueue = true;
			animate = AnimateFactory.createAnimate(this, duration, "panelPosition", newPanelPosition);
			if (easer)
				animate.easer = easer;
			animate.addEventListener(EffectEvent.EFFECT_END, onEndScroll);
			animate.play();
		}
		
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function updateDisplayList (width:Number, height:Number):void
		{
			super.updateDisplayList(width, height);
			
			updateCardScrollInfo();
			
			var numElements:int = target.numElements;
			for (var i:int = 0; i < numElements; i++) {
				
				var element:IVisualElement = target.getElementAt(i);
				if (element == null) {
					continue;
				}
				
				var matrix:Matrix3D = new Matrix3D();
				var position:Number = panelDistance * i - panelPosition;
				var indexedPosition:Number = position / panelDistance;
				var roundedIndexedPosition:Number = Math.round(indexedPosition);
				
				if (Math.abs(indexedPosition - roundedIndexedPosition) < 0.01)
					indexedPosition = roundedIndexedPosition;
				
				
				setAlpha(element, indexedPosition);
				setFilters(element, indexedPosition);
				
				rotateYPanel(element, matrix, indexedPosition, i);
				element.setLayoutMatrix3D(matrix, false);
			}
			
			cardLayoutHelper.enableThumbnailGeneration();
		}
		
		protected function enablePanels(enabled:Boolean):void
		{
			if (!target)
				return;
			
			var numElements:int = target.numElements;
			for (var i:int = 0; i < numElements; i++) {
				
				var panel:IUIComponent = target.getElementAt(i) as IUIComponent;
				if (panel == null) {
					continue;
				}
				panel.enabled = enabled;
			}
		}
		
		/**
		 *  Scroll the layout display to the panel with the specified index.
		 * 
		 *  @param cardIndex The index of the panel that is to be made the focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function scrollTo(panelIndex:int):void {
			if (_focusIndex == panelIndex)
				return;
			focusIndex = panelIndex;
			scroll();
		}
		
		/**
		 *  ctochatreplay is called when the user drags a panel either left or right.  In response 
		 *  the function scrolls one panel to the left or right.
		 * 
		 *  @param delta The amount of the drag move.  This must be at least 20 pixels.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function nudgePanel(delta:Number):void 
		{
			var absDelta:Number = Math.abs(delta);
			
			if (absDelta > 10 && absDelta < panelDistance) {
				var newIndex:int = focusIndex + (delta > 0 ? 1 : -1);
				newIndex = MathUtil.clipToRange(newIndex, 0, target.numElements-1);
				scrollTo(newIndex);
			} else {
				scrollTo(focusIndex);
			}
		}
		
		/**
		 *  Handles an end scroll event
		 * 
		 *  @param e The onEndScroll effect to be handled
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onEndScroll(e:EffectEvent):void 
		{
			var numElements:int = target.numElements;
			animate.removeEventListener(EffectEvent.EFFECT_END, onEndScroll);
			animate = null;
			dispatchEvent(new Event("panelPositionChanged"));
			enablePanels(true);
			ThumbnailManager.instance.pauseQueue = false;
		}

		/**
		 *  Handles the bookeeping when an element is added
		 * 
		 *  @param index The index value of the element that has been added
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function elementAdded(index:int):void {
			super.elementAdded(index);
			panelSizesInvalid = true;
			sizeInvalid = true;
			focusPanelInvalid = true;
		}
		
		protected function updateMaxPanelPosition():void
		{
			maxPanelPosition = panelPositionFromIndex(target.numElements - 1);  
		}
		
		/**
		 *  Set the target's focus to be the specified attachment
		 * 
		 *  @param attachment The IAttachement element that is to be the new focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function setFocusAttachment(attachment:Object):void
		{
			dispatchEvent(new LayoutEvent(LayoutEvent.HANDLED_ATTACHMENT_FOCUS));
		}
		
		/**
		 *  Reset the size of the panels displayed.  
		 *  This is called when the target has been resized
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function resetPanelSizes():void 
		{
			if (!panelSizesInvalid)
				return;
			
			if (!target)
				return;
			
			var width:Number = target.width;
			var height:Number = target.height;
		
			var fitSize:Number = 0.8 * Math.min(width,height);
			fitSize = Math.min(fitSize, 800);
			var numElements:int = target.numElements;
			if (numElements <= 0)
				return;
			
			panelSizesInvalid = false;
			for (var i:int = 0; i < numElements; i++) {
				
				var panel:IVisualElement = target.getElementAt(i);
				if (panel == null) {
					continue;
				}
				panel.setLayoutBoundsSize(panelWidth, panelHeight, false);
			}
		}

		/**
		 *  Sets the alpha transparency value of the specified card.
		 * 
		 *  @see mx.core.IFlexDisplayObject
		 *  @param card The card for which you want to set the transparency
		 *  @param indexedPosition
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function setAlpha(panel:IVisualElement, indexedPosition:Number):void 
		{
			var absIndexedPosition:Number = Math.abs(indexedPosition);

			panel.alpha = (absIndexedPosition < 0.25) ? 1 : 0.75;
		}
		
		/**
		 *  Set the color and blur filters for the specified visual layout element
		 * 
		 *  @param layoutElement The visual element for which you want to set the filters
		 *  @param indexedPosition
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function setFilters(layoutElement:ILayoutElement, indexedPosition:Number):void
		{
			var uiElement:UIComponent = layoutElement as UIComponent;
			var absIndexedPosition:Number = Math.abs(indexedPosition);
			
			var blurIndex:uint = Math.min(3, absIndexedPosition > 0.25 ? 1 : 0) as uint;
//			var newBlurFilter:BlurFilter = blurFilters[blurIndex];
//			var curBlurFilter:BlurFilter = uiElement.filters ? uiElement.filters[0] : null;
			
			var colorIndex:uint = Math.min(128, absIndexedPosition*32);
			var newColorFilter:ColorMatrixFilter = colorFilters[colorIndex];
			var curColorFilter:ColorMatrixFilter = uiElement.filters ? uiElement.filters[1] : null;
			
			if (/*!curBlurFilter || curBlurFilter.blurX != newBlurFilter.blurX
					|| */!curColorFilter || curColorFilter.matrix[0] != newColorFilter.matrix[0]) {
				uiElement.filters = [/*newBlurFilter, */newColorFilter];
			}
		}
		
		/**
		 *  Sets the horizontal position of a visual element
		 * 
		 *  @param matrix The 3D transformation matrix that determines the position 
		 *  and orientation of the cards.
		 *  @param position The horizontal position to set
		 *
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function setXPosition(matrix:Matrix3D, position:Number):void 
		{
			matrix.appendTranslation(position, 0, 0);
		}
		
		/**
		 *  Clips the given position value so that it is within the min-to-max range.
		 *  
		 * 
		 *  @param val the original value to clip
		 *  @param min The minimum value of the range
		 *  @param max The maximum value of the range
		 *  @return The new value, which is between min and max, inclusive
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private static function clipToRange(val:Number, min:Number, max:Number): Number 
		{
			if (val < min)
				val = min;
			else if (val > max)
				val = max;
			
			return val;
		}

		protected function setZDistance(element:IVisualElement, matrix:Matrix3D, indexedPosition:Number):void 
		{
			var distance:Number = zDistance;
			element.depth = -distance;
			matrix.appendTranslation(0, 0, distance);
		}

		/**
		 *
		 */
		protected static function centerElement(element:IVisualElement, matrix:Matrix3D, width:Number, height:Number):void 
		{
			matrix.appendTranslation(width / 2 - (element.width / 2),  
									  height / 2 - (element.height / 2), 0); 
		}
		
		/**
		 *  Returns the horizontal card position, in pixels, given
		 *  the cards index value.
		 * 
		 *  @param index The index value for the card
		 *  @return the horizontal position of the card, given in pixels
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function panelPositionFromIndex(index:int):Number
		{
			return index * ((totalWidth - layoutWidth) / (target.numElements - 1));
		}
		
		[Bindable (name="panelPositionChanged")]
		public function get panelPosition():Number
		{
			return _cardPosition;
		}
		
		public function set panelPosition(value:Number):void
		{
			if (_cardPosition == value)
				return;
			
			value = MathUtil.clipToRange(value, 0, maxPanelPosition);
			
			_cardPosition = value;
			invalidateTarget();
			
			if (!animate) {
				dispatchEvent(new Event("panelPositionChanged"));
			}
		}
		
		/**
		 *  The maximum horizontal position that a panel can have.
		 *  This is dependent on the target width.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function get maxPanelPosition():Number
		{
			return _maxPanelPosition;
		}
		
		/**
		 *  @private
		 *
		 */
		protected function set maxPanelPosition(value:Number):void
		{
			value = Math.floor(value);
			if (_maxPanelPosition == value)
				return;
			
			_maxPanelPosition = value;
		}
		
		/**
		 *  Set the target instance as invalid.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function invalidateTarget():void
		{
			if (!target)
				return;
			
			target.invalidateDisplayList();
		}

		[Bindable]
		/**
		 *  The index of the centered item
		 * 
		 *  @param value The index value of the centered item
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get focusIndex():int
		{
			return _focusIndex;
		}
		/**
		 *  @private
		 * 
		 */
		public function set focusIndex(value:int):void
		{
			_focusIndex = value;
			invalidateTarget();
			focusPanelInvalid = true;
		}
		
		[Bindable]
		/**
		 *  The apparent depth of the visual element
		 * 
		 *  @see flash.geom.PerspectiveProjection
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get zDistance():Number
		{
			return _zDistance;
		}
		/**
		 *  @private
		 */
		public function set zDistance(value:Number):void
		{
			_zDistance = value;
			invalidateGallery();
		}
		
		[Bindable]
		/**
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get xAngle():Number
		{
			return _xAngle;
		}
		/**
		 *  @private
		 */
		public function set xAngle(value:Number):void
		{
			_xAngle = value;
			invalidateGallery();
		}

		[Bindable]
		public function get spotlight():Boolean
		{
			return _spotlight;
		}
		
		public function set spotlight(value:Boolean):void
		{
			_spotlight = value;
			invalidateGallery();
		}

		[Bindable]
		/**
		 *  The card width for background cards
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get panelWidth():Number
		{
			return _panelWidth;
		}
		
		/**
		 *  @private
		 */
		public function set panelWidth(value:Number):void
		{
			_panelWidth = value;
			invalidateSize();
		}
		
		protected function dispatchCollectionChangeEvent():void
		{
			if (!target)
				return;
			
			var eventDispatcher:IEventDispatcher = target["dataProvider"] as IEventDispatcher;
			if (!eventDispatcher)
				return;
			
			eventDispatcher.dispatchEvent(new CollectionEvent(CollectionEvent.COLLECTION_CHANGE));
		}

		[Bindable]
		/**
		 *  The card height for background cards
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get panelHeight():Number
		{
			return _panelHeight;
		}
		/**
		 *  @private
		 */
		public function set panelHeight(value:Number):void
		{
			_panelHeight = value;
			invalidateSize();
		}

		[Bindable]
		/**
		 *  The instance of the EaseInOutBase class that provide easing capability.
		 * 
		 *  @see spark.effects.easing
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get easer():EaseInOutBase
		{
			return _easer;
		}
		/**
		 *  @private
		 */
		public function set easer(value:EaseInOutBase):void
		{
			_easer = value;
		}

		[Bindable]
		/**
		 *  An angle, specified in degrees between 0 and 180, for the field of view in three dimensions.
		 * 
		 *  @see flash.geom.PerspectiveProjection
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get fieldOfView():Number
		{
			return _fieldOfView;
		}
		/**
		 *  @private
		 */
		public function set fieldOfView(value:Number):void
		{
			_fieldOfView = value;
			resetPerspective();
		}

		/**
		 *  Find the list of cards that overlap the specified marquee rectangle. 
		 * 
		 *  @param marqueeRect A bounding rectangle
		 *  @return A vector of integers that represent the indexes
		 *  of the cards in the layout that intersect the specified marquee rectangle.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function findCardsThatIntersect(marqueeRect:Rectangle):Vector.<int>
		{
			return null; // only necessary for marquee support
		}

		/**
		 *  The size, represented as a width and height ordered pair, to use for thumbnails
		 * 
		 *  @param value The image size to use for thumbnails
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get thumbFitSize():Point
		{
			return _thumbFitSize;
		}
		/**
		 *  @private
		 */
		public function set thumbFitSize(value:Point):void
		{
			_thumbFitSize = value;
		}
		
		protected function getPanelInfoFocusCard():PanelInfo 
		{
			return getPanelInfo(focusIndex);
		}
		
		protected function getFloorPanelIndexFromPosition(panelPosition:Number):int
		{
			var index:int = Math.floor(panelPosition / panelDistance);
			if (index < 0)
				index = 0;
			if (index >= target.numElements)
				index = target.numElements - 1;
			return index;
		}
		
		/** returns a number in the range [0,1] */
		protected function getIndexedDelta():Number
		{
			var pos:Number = panelPosition / panelDistance;
			
			return pos - Math.floor(pos);
		}
		
		protected function appendViewMatrices():void
		{
			cardScrollInfo.viewMatrix.identity();
			
			for (var pos:Number = 0; (pos + panelDistance) <= panelPosition; pos += panelDistance) {
				var cardIndex:int = getFloorPanelIndexFromPosition(pos);
				if (cardIndex == -1)
					return;
				var element:IVisualElement = target.getElementAt(cardIndex+1);
				if (!element)
					return;
				
				var panelInfo:PanelInfo = panels[element];
				const t:Number = 1;
				appendViewMatrix(element, panelInfo, t);
			}
		}
		
		protected function validatePanels():void
		{
			if (!panels)
				panels = new Dictionary();
			
			validateGeometry();
			var geometry:String = galleryGeometry;
			var numElements:int = target.numElements;
			for (var i:int = 0; i < numElements; i++) {
				getPanelInfo(i);
				var panel:Panel = target.getElementAt(i) as Panel;
				if (!panel)
					continue;
				panel.panelGeometry = geometry.charAt(i);
			}
		}
		
		protected function validateGeometry():void
		{
			var numElements:int = target.numElements;
			if (numElements == 0)
				return;
			
			if (galleryGeometry.length >= numElements)
				return;
			
			var temp:String = "";
			var geometry:String = galleryGeometry;
			
			for (var i:int = 0; i < numElements; i++) {
				var panelGeometry:String = geometry.charAt(i);
				if (panelGeometry == "") {
					panelGeometry = (i % 2) ? "O" : "I";
				}
				temp += panelGeometry;
			}
			
			galleryGeometry = temp;
		}
		
		protected function updateCardScrollInfo():void
		{
			validatePanels();
			appendViewMatrices();
			
			var index:int = getFloorPanelIndexFromPosition(panelPosition + panelDistance - 0.00001);
			var element:IVisualElement = target.getElementAt(index);
			
			if (!panels)
				return;
			
			var panelInfo:PanelInfo = panels[element];
			if (!panelInfo)
				return;
			
			var t:Number = getIndexedDelta();
			if (t == 0) {
				cardScrollInfo.globalAngle = panelInfo.angle;
				//				logger.debug("t=0: cardScrollInfo: globalAngle= " + cardScrollInfo.globalAngle);
			} else {
				appendViewMatrix(element, panelInfo, t);
			}
			rotateAngle(Vector3D.X_AXIS, new Vector3D(0,-panelHeight/2,0), xAngle, cardScrollInfo.viewMatrix, false);
		}
		
		protected function appendViewMatrix(element:IVisualElement, panelInfo:PanelInfo, t:Number):void 
		{
			if (!panelInfo)
				return;
			
			var rotateIn:Boolean = panelInfo.angleDelta == 90;
			var fullAngle:Number = -panelInfo.angleDelta;
			
			if (fullAngle == 0) {
				cardScrollInfo.globalAngle = panelInfo.angle;
				var delta:Number = interpolate(0, -element.width, t);
				cardScrollInfo.viewMatrix.appendTranslation(delta, 0, 0);
			} else {
				var halfWidth:Number = element.width / 2;
				var deltaAngle:Number = interpolate(0, fullAngle, 1-t);
				cardScrollInfo.globalAngle = panelInfo.angle + deltaAngle;
				var angle:Number = interpolate(0, fullAngle, t);
				var origin:Vector3D = new Vector3D(-halfWidth, 0, rotateIn ? halfWidth : -halfWidth);
				rotateYAngle(origin, angle, cardScrollInfo.viewMatrix, false);
			}
		}
		
		public function get panelDistance():Number 
		{
			return panelWidth;
		}
		
		protected function rotateYPanel(element:IVisualElement, matrix:Matrix3D, indexedPosition:Number, elementIndex:int):void 
		{
			var panelInfo:PanelInfo = getPanelInfo(elementIndex);
			if (!panelInfo)
				return;
			
			matrix.append(panelInfo.matrix);
			
			if (cardScrollInfo.viewMatrix)
				matrix.append(cardScrollInfo.viewMatrix);
			setZDistance(element, matrix, indexedPosition);
			centerElement(element, matrix, target.width, target.height);
			
			var v:Vector3D = matrix.transformVector(new Vector3D(element.width/2, element.height/2, 0));
			element.depth = -v.z;
			
			setPanelVisibility(element, matrix, indexedPosition, elementIndex, panelInfo);
		}
		
		protected function setPanelVisibility(element:IVisualElement, matrix:Matrix3D, 
											  indexedPosition:Number, elementIndex:int, 
											  panelInfo:PanelInfo):void 
		{
			if (!spotlight) {
				element.visible = true;
				return;
			}
			
			var v0:Vector3D = matrix.transformVector(new Vector3D(0, 0, 0));
			var v1:Vector3D = matrix.transformVector(new Vector3D(0, element.height, 0));
			var v2:Vector3D = matrix.transformVector(new Vector3D(element.width, element.height, 0));
			var v3:Vector3D = matrix.transformVector(new Vector3D(element.width, 0, 0));
			
			var minElementZ:Number = Math.min(v0.z, v1.z, v2.z, v3.z);
			var maxElementZ:Number = Math.max(v0.z, v1.z, v2.z, v3.z);
			var minZ:Number = zDistance;
			var visible:Boolean = true;
			if (minElementZ < minZ && maxElementZ < minZ)
				visible = false;

			element.visible =  visible;
		}
		
		protected function getPanelInfo(elementIndex:int):PanelInfo 
		{
			if (!panels)
				return null;
			
			var element:IVisualElement = target.getElementAt(elementIndex);
			var panelInfo:PanelInfo = panels[element];
			if (panelInfo)
				return panelInfo;
			
			var prevElement:IVisualElement = target.getElementAt(elementIndex-1);
			var prevPanelInfo:PanelInfo = panels[prevElement];
			if (prevPanelInfo) {
				var matrix:Matrix3D = prevPanelInfo.matrix.clone();
				var angleDelta:Number = getYAngleDelta(elementIndex);
				var prevAngle:Number = prevPanelInfo.angle;
				if (angleDelta == 0) {
					matrix.prependTranslation(prevElement.width, 0, 0);
				} else {
					var halfWidth:Number = element.width/2;
					var origin:Vector3D = new Vector3D(halfWidth, 0, angleDelta == 90 ? -halfWidth : halfWidth);
					rotateYAngle(origin, angleDelta, matrix);
				}
				panelInfo = new PanelInfo(matrix, angleDelta + prevAngle, angleDelta);
				logger.debug("getPanelInfo: panelInfo: matrix= " + panelInfo.matrix + " angle= " + panelInfo.angle + " angleDelta= " + panelInfo.angleDelta);
			} else {
				panelInfo = new PanelInfo(new Matrix3D(), 0, 90);
			}
			panels[element] = panelInfo;
			
			return panelInfo;
		}
		
		protected function interpolate(a:Number, b:Number, t:Number):Number 
		{
			return a + (b-a) * t;
		}
		
		protected function rotateAngle(axis:Vector3D, origin:Vector3D, angle:Number, matrix:Matrix3D, prepend:Boolean = true):void 
		{
			if (angle == 0)
				return;
			
			if (prepend) {
				matrix.prependTranslation(origin.x, origin.y, origin.z);
				matrix.prependRotation(angle, axis);
				matrix.prependTranslation(-origin.x, -origin.y, -origin.z);
			} else {
				matrix.appendTranslation(origin.x, origin.y, origin.z);
				matrix.appendRotation(angle, axis);
				matrix.appendTranslation(-origin.x, -origin.y, -origin.z);
				
			}
		}
		
		protected function rotateYAngle(origin:Vector3D, angle:Number, matrix:Matrix3D, prepend:Boolean = true):void 
		{
			if (angle == 0)
				return;
			
			rotateAngle(Vector3D.Y_AXIS, origin, angle, matrix, prepend);
		}
		
		protected function getYAngleDelta(elementIndex:int):Number
		{
			if (!geometryDict) {
				geometryDict = new Dictionary();
				geometryDict["-"] = 0;
				geometryDict["I"] = 90;
				geometryDict["O"] = -90;
			}
			
			if (elementIndex == 0)
				return 0;
			
			var cardGeometry:String = galleryGeometry.charAt(elementIndex-1);
			
			return geometryDict[cardGeometry];
		}
		
		protected function onPanelOrientationChange(event:PanelEvent):void
		{
			var index:int = target.getElementIndex(event.panel);
			if (index > -1 && index < target.numElements) {
				var panelGeometry:String = galleryGeometry.charAt(index);
				switch (panelGeometry) {
					case "I":
						panelGeometry = "O";
						break;
					
					case "O":
						panelGeometry = "-";
						break;
					
					case "-":
						panelGeometry = "I";
						break;
				}
				galleryGeometry = StringUtil.replaceChar(galleryGeometry, index, panelGeometry, "I");
			}
		}
		
		[Bindable]
		public function get galleryGeometry():String
		{
			return _galleryGeometry;
		}
		
		public function set galleryGeometry(value:String):void
		{
			_galleryGeometry = value;
			customProperties.galleryGeometry.value = value;
			invalidateGallery();
		}
		
		protected function invalidateGallery():void 
		{
			invalidateTarget();
			panels = null;
		}
		
		public function get cardLayoutHelper():CardLayoutHelper
		{
			return _cardLayoutHelper;
		}
		
		public function set cardLayoutHelper(value:CardLayoutHelper):void
		{
			_cardLayoutHelper = value;
		}

		[Bindable]
		public function get cardProvider():ICardProvider
		{
			return _cardProvider;
		}

		public function set cardProvider(value:ICardProvider):void
		{
			_cardProvider = value;
		}

		[Bindable]
		public function get cardsPerPanel():int
		{
			return _cardsPerPanel;
		}

		public function set cardsPerPanel(value:int):void
		{
			_cardsPerPanel = value;
		}


	}
}


import flash.geom.Matrix3D;

class PanelInfo {
	
	public var matrix:Matrix3D;
	public var angle:Number;
	public var angleDelta:Number;
	
	public function PanelInfo(matrix:Matrix3D, angle:Number, angleDelta:Number)
	{
		this.matrix = matrix;
		this.angle = angle;
		this.angleDelta = angleDelta;
	}
}

class CardScrollInfo {
	
	public var globalAngle:Number;
	public var viewMatrix:Matrix3D;
	
	public function CardScrollInfo()
	{
		globalAngle = 0;
		viewMatrix = new Matrix3D();
	}
}

