/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.animations
{
	import com.adobe.portfolio.utils.LogHelper;
	import com.adobe.portfolio.utils.OrderedDictionary;
	
	import flash.events.Event;
	import flash.events.IEventDispatcher;
	import flash.events.TimerEvent;
	import flash.utils.Timer;
	
	import mx.events.EffectEvent;
	import mx.logging.ILogger;
	import mx.utils.ObjectUtil;
	import mx.utils.UIDUtil;
	
	import spark.effects.Animate;

	/**
	 *  The AnimationSequence class animates a property value of each object in a list of objects.  The animation
	 *  proceeds sequentially through the list of objects, animating the value of the specified property to a new value.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class AnimationSequence
	{
		
		private var animations:OrderedDictionary;
		private var playing:Boolean = false;
		private var waitingToPlay:Boolean = false;
		
		private var _onDone:Function;
		private var _onElementDone:Function;
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function AnimationSequence()
		{
		}
		
		/**
		 * Creates an animation of the specified property and adds it to a queue, and starts processing the 
		 * queue if it's not running already subject to the specified delay.
		 * 
		 *  @param target The target object
		 *  @param duration The time duration of the animation effect in milliseconds
		 *  @param propertyName The name of the property (of the target object) to animate.
		 *  @param propertyValue The desired final value of the property.
		 *  @param delay The delay (in msec) before starting the animation.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function addAnimation(target:Object, duration:Number, propertyName:String, propertyValue:*, delay:Number):void
		{
			if (animations && animations.getValue(target))
				return;
			
			animations = playSequence(animations, target, 
				AnimateFactory.createAnimate(target, duration, propertyName, propertyValue), delay);
		}
		
		/**
		 *  Returns the number of animations to run.  If there are no animations
		 *  to run, this will return 0.
		 * 
		 *  @return The number of animations to run.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get length():uint 
		{
			return animations ? animations.length : 0;
		}
		
		/**
		 *  @private
		 */
		protected function playSequence(animations:OrderedDictionary, key:Object, animate:Animate, delay:Number):OrderedDictionary 
		{
			CONFIG::logging{
				LogHelper.logger.debug("playSequence: " + animate.target.toString());
			}
			if (!animations)
				animations = new OrderedDictionary();
			
			if (playing || waitingToPlay)
				animations.push(key, animate);
			else {
				var t:Timer = new Timer(delay);
				t.addEventListener(TimerEvent.TIMER,
					function (e:Event):void {
						t.stop();
						t.removeEventListener(e.type, arguments.callee);
						waitingToPlay = false;
						startPlay(animate);
					}
				);
				t.start();
				waitingToPlay = true;
			}
			return animations;
		}
		
		/**
		 *  @private
		 */
		protected function startPlay(animate:Animate):void 
		{
			if (!animate) {
				CONFIG::logging{
					LogHelper.logger.debug("startPlay(): stopped: animate == null");
				}
				if (onDone != null) {
					onDone();
				}
				return;
			}
			if (playing)
				return;
			
			CONFIG::logging{
				LogHelper.logger.debug("startPlay(): start: " + animate.target.toString());
			}
			animate.addEventListener(EffectEvent.EFFECT_END, effectEnd, false, 0, true );
			animate.play();
			playing = true;
		}
		
		/**
		 *  @private
		 */
		protected function effectEnd(e:EffectEvent):void 
		{
			(e.target as IEventDispatcher).removeEventListener(e.type, arguments.callee);
			playing = false;
			if (onElementDone != null) 
				onElementDone(e.target.target); // i.e. onElementDone gets called w/ the animate target that just completed
			
			startPlay(animations.shift());
		}
		
		/**
		 *  The callback function to run when the animation sequence has completed.
		 * 
		 *  @param value A callback function of the form: callback(target:Object):void, 
		 *  where target is the object that dispatched the animation sequence. The callback function is called
		 *  when the entire animation sequence is completed.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function set onDone(value:Function):void
		{
			_onDone = value;
		}
		/**
		 *  @private
		 */
		public function get onDone():Function
		{
			return _onDone;
		}

		/**
		 *  The callback function to run when a animation in the sequence has completed.
		 * 
		 *  @param value Q callback function of the form: callback(target:Object):void, 
		 *  where target is the animate target. The callback function is called
		 *  when this element's animation is completed.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function set onElementDone(value:Function):void
		{
			_onElementDone = value;
		}
		/**
		 *  @private
		 */
		public function get onElementDone():Function
		{
			return _onElementDone;
		}
	}
}


