/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.cards
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.managers.PropertyManager;
	import com.adobe.portfolio.utils.LogHelper;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.core.ClassFactory;
	import mx.core.IFactory;
	
	/**
	 *  The CardFactory class is a singleton that can be called to construct Cards and associated ItemRenderer functions.
	 * 
	 *  <p>You specify a generator class when you construct the cardFactory object. 
	 *  Then you set the properties property on the cardFactory object. 
	 *  Flex uses the factory object to generate instances by calling the 
	 *  cardFactory object's newInstance() method.</p>
	 * 
	 *  <p>The newInstance() method creates a new instance of the generator class,
	 *  and sets the properties specified by properties in the new instance. 
	 *  If you need to further customize the generated instances, you can 
	 *  override the newInstance() method.</p>
	 * 
	 *  <p>The CardFactory class implements the IFactory interface. Therefore, 
	 *  it lets you create objects that can be assigned to properties of type IFactory, 
	 *  such as the itemRenderer property.</p>
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */	
	public class CardFactory extends ClassFactory
	{
		/**
		 *  String literal used to represent a SimpleCard value for the cardType
		 *  when constructing a CardFactory.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const SIMPLE_CARD:String = "SimpleCard";
		
		/**
		 *  String literal used to represent a BasicCard value for the cardType
		 *  when constructing a CardFactory.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const BASIC_CARD:String = "BasicCard";
		
		/**
		 *  @review
		 * 
		 *  String literal used to represent a PreviewCard value for the cardType
		 *  when constructing a CardFactory.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PREVIEW_CARD:String = "PreviewCard";
		
		/**
		 *  String literal used to represent a FolderCard value for the cardType
		 *  when constructing a CardFactory.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const FOLDER_CARD:String = "FolderCard";

		/**
		 *  String literal used to represent a simple state value for the cardType
		 *  when constructing a CardFactory.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const STATE_SIMPLE:String = "simple";
		
		/**
		 *  @review
		 * 
		 *  String literal used to represent a small state value for the cardType
		 *  when constructing a CardFactory.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const STATE_SMALL:String = "small";
		
		/**
		 *  @review
		 * 
		 *  String literal used to represent a medium state value for the cardType
		 *  when constructing a CardFactory.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const STATE_MEDIUM:String = "medium";
		
		/**
		 *  @review
		 * 
		 *  String literal used to represent a large state value for the cardType
		 *  when constructing a CardFactory.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const STATE_LARGE:String = "large";
		

		/**
		 * A string literal used to represent the preferred location, "none", of a dropshadow effect.
		 * 
		 * @see SimpleCard#dropShadowPreference
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const DROP_SHADOW_NONE:String = "none";
		
		/**
		 * A string literal used to represent the preferred location, "bottom" of a dropshadow effect.
		 * 
		 * @see SimpleCard#dropShadowPreference
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const DROP_SHADOW_BOTTOM:String = "bottom";
		
		/**
		 * A string literal used to represent the preferred location, "rear" of a dropshadow effect.
		 * 
		 * @see SimpleCard#dropShadowPreference
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const DROP_SHADOW_REAR:String = "rear";
		
		/**
         *  String literal used to represent "none" as the preferred type of rollover effect.
		 *  This is used as a hint to skins when specifying filters.
		 * 
		 *  @see SimpleCard#rollOverPreference
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public static const ROLL_OVER_NONE:String = "none";
		
		/**
		 *  String literal used to represent "glow" as the preferred type of rollover effect.
		 *  This is used as a hint to skins when specifying filters.
		 * 
		 *  @see SimpleCard#rollOverPreference
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public static const ROLL_OVER_GLOW:String = "glow";
		
		/**
		 *  String literal used to represent "scale" as the preferred type of rollover effect.
		 *  This is used as a hint to skins when specifying filters.
		 * 
		 *  @see SimpleCard#rollOverPreference
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public static const ROLL_OVER_SCALE:String = "scale";
		
		private static var rendererFunctions:Object = new Object();
		private static var factories:Object = new Object();
		private static var folderFactories:Object = new Object();

		private var cardSizeWatcher:ChangeWatcher;
		
		/**
		 *  Returns a function that will provide an appropriate CardFactory based on a particular IAttachment.
		 * 	This returned function can be used as a dataGroup's itemRendererFunction.
		 * 
		 *  @param defaultType The type of card to be used for standard attachment types.
		 *  @param defaultState The initial state to be used for standard attachment types. 
		 *  @param dropShadowPreference The preferred drop shadow effect. 
		 *  This is a string specifying the preferred location for a drop shadow effect on this component.
		 *  This is used as a hint to skins when specifying filters.
		 *  @param rollOverPreference The preferred rollover effect: none, glow, or scale.
		 *  @param alpha The alpha value (transparency) of the item; the default is 1 (completely opaque)
		 *  @param enableThumbnailGeneration If <code>true</code> (the default) then generate thumbnail images for the card
		 *  @param maintainAspectRatio If <code>true</code> then adjust the aspect ratio when rendering; default is <code>false</code>
		 * 	@param supportFolder If <code>true</code> then folders are supported.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public static function getItemRendererFunction(defaultType:String = "BasicCard",
													   defaultState:String = null,
													   dropShadowPreference:String = null,
													   rollOverPreference:String = null,
													   alpha:Number = 1,
													   enableThumbnailGeneration:Boolean = true,
													   maintainAspectRatio:Boolean = false,
													   supportFolder:Boolean = true):Function
		{
			var factoryKey:String = defaultType +
									alpha.toString() +
									enableThumbnailGeneration.toString() +
									maintainAspectRatio.toString() + 
									((defaultState == null) ? "null" : defaultState) +
									((dropShadowPreference == null) ? "null" : dropShadowPreference) +
									((rollOverPreference == null) ? "null" : rollOverPreference);
			var functionKey:String = factoryKey + supportFolder.toString();
			
			var returnFunction:Function = rendererFunctions[functionKey] as Function;
			if (returnFunction != null) {
				return returnFunction;
			}
			
			var cardFactory:CardFactory = factories[factoryKey] as CardFactory;
			if (cardFactory == null) {
				cardFactory = new CardFactory(
					defaultType, defaultState, dropShadowPreference, rollOverPreference, alpha, enableThumbnailGeneration, maintainAspectRatio);
				factories[factoryKey] = cardFactory;
			}
			var folderFactory:CardFactory = supportFolder ? folderFactories[factoryKey] as CardFactory : cardFactory;
			if (folderFactory == null) {
				folderFactory = new CardFactory(
					FOLDER_CARD, defaultState, dropShadowPreference, rollOverPreference, alpha, enableThumbnailGeneration, maintainAspectRatio);
				folderFactories[factoryKey] = folderFactory;
			}
			
			returnFunction = function(data:Object):IFactory
			{
				var attachment:IAttachment = data as IAttachment;
				if (attachment == null) {
					if (data != null) {
						;// fixes warning
						CONFIG::logging{
							LogHelper.logger.warn("Unexpected data type in CardFactory");
						}
					}
					return cardFactory;
				}
				if (attachment.isFolder) {
					return folderFactory;
				} else {
					return cardFactory;
				}
			};
			
			rendererFunctions[functionKey] = returnFunction;
			return returnFunction;
		}
		
		/**
		 *  Returns new instances of the specified card type.
		 * 
		 *  @param cardType The type of card to create; default is a BasicCard
		 *  @param defaultState The initial state of the card
		 *  @param dropShadowPreference The preferred drop shadown effect. 
		 *  This is a string specifying the preferred location for a drop shadow effect on this component.
		 *  This is used as a hint to skins when specifying filters.
		 *  @param rollOverPreference The preferred rollover effect: none, glow, or scale.
		 *  @param alpha The alpha value (transparency) of the card; the default is 1 (completely opaque)
		 *  @param enableThumbnailGeneration If <code>true</code> (the default) then generate thumbnail images for the card.
		 *  @param maintainAspectRatio If <code>true</code> then adjust the aspect ratio when rendering; default is <code>false</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function CardFactory(cardType:String = BASIC_CARD,
									defaultState:String = null,
									dropShadowPreference:String = null,
									rollOverPreference:String = null,
									alpha:Number = 1,
									enableThumbnailGeneration:Boolean = true,
									maintainAspectRatio:Boolean = false)
		{
			var cardClass:Class = getClassForType(cardType);
			super(cardClass);
			
			var defaultProperties:Object = new Object();
			var currentState:String = getStateForType(cardType, defaultState);
			if (currentState != null) {
				defaultProperties.currentState = currentState;
			} else if (supportsBasicStates(cardType)) {
				// Bind to the card size property
				cardSizeWatcher = BindingUtils.bindSetter(setCardSize, PropertyManager.instance, ["commonProperties", "cardSize", "name"]);
			}
			if (dropShadowPreference != null) {
				defaultProperties.dropShadowPreference = dropShadowPreference;
			}
			if (rollOverPreference != null) {
				defaultProperties.rollOverPreference = rollOverPreference;
			}
			defaultProperties.alpha = alpha;
			defaultProperties.enableThumbnailGeneration = enableThumbnailGeneration;
			defaultProperties.maintainAspectRatio = maintainAspectRatio;
			
			properties = defaultProperties;
		}

		/**
		 *  Returns the Class corresponding to a particular card type.
		 * 	Subclasses can override this function to provide support for
		 * 	alternative or additional card classes.
		 * 
		 */ 
		protected function getClassForType(cardType:String):Class
		{
			switch (cardType) {
				case SIMPLE_CARD:
					return SimpleCard;
				case FOLDER_CARD:
					return FolderCard;
				case BASIC_CARD:
					return BasicCard;
				case PREVIEW_CARD:
					return PreviewCard;
				default:
					CONFIG::logging{
						LogHelper.logger.warn("Unexpected cardType constructing CardFactory: "+cardType);
					}
					break;
			}
			return BasicCard;
		}
		
		/**
		 *  Returns the state string corresponding to a particular card type and state.
		 * 	
		 * 	@return	The state name string representing the given card state for
		 * 			the class indicated by the card type. The returned value may
		 * 			not match the given card state, or may be null if the state is
		 * 			unsupported by the given card type.
		 * 
		 * 	@see #getClassForType
		 */ 
		protected function getStateForType(cardType:String, cardState:String):String
		{
			if (cardState == null) {
				return null;
			}
			if ((cardType == SIMPLE_CARD) || (cardType == PREVIEW_CARD)) {
				return null;
			}
			if ((cardType == FOLDER_CARD) && (cardState == STATE_SIMPLE)) {
				return FolderCard.STATE_SIMPLE;
			}
			if (supportsBasicStates(cardType)) {
				switch (cardState) {
					case STATE_SMALL:
						return BasicCard.STATE_SMALL;
					case STATE_MEDIUM:
						return BasicCard.STATE_MEDIUM;
					case STATE_LARGE:
						return BasicCard.STATE_LARGE;
				}
			}
			return null;
		}
		
		/**
		 *  Returns true if the card class returned by this factory supports the basic card states.
		 * 
		 * 	@see BasicCard
		 */ 
		protected function supportsBasicStates(cardType:String):Boolean
		{
			return ((cardType == BASIC_CARD) || (cardType == FOLDER_CARD));
		}
		
		/**
		 *  @private
		 */	
		private function setCardSize(value:String):void
		{
			if (value == null) {
				return;
			}
			if (!validCardSize(value)) {
				return;
			}
			
			if (properties == null) {
				properties = { currentState: value };
			} else {
				properties.currentState = value;
			}
		}
		
		/**
		 *  @private
		 */	
		private function validCardSize(cardSize:String):Boolean 
		{
			return cardSize == BasicCard.STATE_SMALL || cardSize == BasicCard.STATE_MEDIUM || cardSize == BasicCard.STATE_LARGE;
		}
	}
}
