/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.cards
{
	import com.adobe.portfolio.events.PreviewEvent;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.CommandManager;
	import com.adobe.portfolio.managers.ThumbnailManager;
	import com.adobe.portfolio.previews.Preview;
	import com.adobe.portfolio.previews.PreviewFactory;
	import com.adobe.portfolio.utils.LogHelper;
	import com.adobe.portfolio.utils.MathUtil;
	
	import flash.events.Event;
	import flash.events.IEventDispatcher;
	import flash.events.MouseEvent;
	import flash.geom.Point;
	
	import mx.states.State;
	
	import spark.components.Button;

	/**
	 *  Dispatched when the user clicks the close button for this component.
	 */
	[Event(name="closeButton", type="flash.events.Event")]
	
	[SkinState("previewNormal")]
	[SkinState("previewInfo")]
	/**
	 *  The PreviewCard class extends SimpleCard to show large thumbnails corresponding to an IAttachment. 
	 * 
	 *  It implements IListItemRenderer,
	 *  so it can be used as an item renderer for relevant containers. The skin is not required
	 *  to provide all the relevant controls, which allows this object to act as the core for
	 *  both simple (mini-nav) and complex (basic click-through card) rendering.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class PreviewCard extends SimpleCard implements IPreviewCard
	{
		/**
		 *  A string literal used to represent the preview state of a previewCard.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const STATE_PREVIEW:String = "preview";
		
		/**
		 *  A string literal used to represent the info state of a previewCard.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const STATE_INFO:String = "info";

		/**
		 * String literals used to represent the states of a PreviewCard skin.
		 * TODO: Clean up mismatch
		 */
		
		/**
		 *  A string literal used to represent the preview state of a PreviewCard skin.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const SKIN_STATE_PREVIEW_NORMAL:String = "previewNormal";
		
		/**
		 *  A string literal used to represent the info state of a PreviewCard skin.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const SKIN_STATE_PREVIEW_INFO:String = "previewInfo";

		
		[SkinPart(required="false")] 
		/**
		 *  closeButton is the button control to close the preview card.
		 *  It is an optional skin part, so it is not necessarily implemented
		 *  for all skins.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var closeButton:Button;

		[SkinPart(required="false")] 
		/**
		 *  infoCloseButton is the button control to close the info card.
		 *  It is an optional skin part, so not necessarily implemented
		 *  for all skins.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var infoCloseButton:Button;
		
		private var _pageNavigationEnabled:Boolean;

		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function PreviewCard()
		{
			super();
			// Setup the states for this instance. They should match the skin states.
			states = [	new State({name: STATE_PREVIEW}),
				new State({name: STATE_INFO})];
			currentState = STATE_PREVIEW;
			enableThumbnailGeneration = false;
			downSampleOK = true;
			_pageNavigationEnabled = false;
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function setCurrentState(stateName:String, playTransition:Boolean = true):void
		{
			if (stateName != currentState) {
				super.setCurrentState(stateName, playTransition);
				
				// We used to call skin.currentState = stateName; here, but simply invalidating
				// plays nicer with the Flex model.
				invalidateSkinState();
			}
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function getCurrentSkinState():String
		{
			switch (currentState) {
				case STATE_PREVIEW:
					return SKIN_STATE_PREVIEW_NORMAL;
				case STATE_INFO:
					return SKIN_STATE_PREVIEW_INFO;
				default:
					CONFIG::logging{
						LogHelper.logger.warn("Unexpected component state: "+currentState);
					}
					return SKIN_STATE_PREVIEW_NORMAL;
			}
		}
				
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function partAdded(partName:String, instance:Object):void
		{
			//handlePartChange will call super.partAdded
			handlePartChange(partName, instance, true);
		}

		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function partRemoved(partName:String, instance:Object):void
		{
			//handlePartChange will call super.partRemoved
			handlePartChange(partName, instance, false);
		}
		
		/** 
		 *  @private
		 * 
		 *  Internal function for adding and removing event listeners for skin parts, depending on
		 *  whether the part is being added or removed for this instance.
		 */
		private function handlePartChange(partName:String, instance:Object, adding:Boolean):void
		{
			var listeners:Array = null;
			var types:Array = null;
			if ((instance != null) && (instance is IEventDispatcher)) {
				var activePart:IEventDispatcher = instance as IEventDispatcher;
				
				// Build up an array of event types and listeners that apply to this object
				switch (instance) {
					case closeButton:
						types = [ MouseEvent.CLICK ];
						listeners = [ onCloseClick ];
						break;
					case infoCloseButton:
						types = [ MouseEvent.CLICK ];
						listeners = [ onInfoCloseClick ];
						break;
					case preview:
						types = [ PreviewEvent.HANDLE_FILE_INFO, PreviewEvent.HANDLE_FILE_EXTRACT];
						listeners = [ handleFileInfo, handleFileExtract];
						break;
				}
				
				// Add or remove all the relevant event listeners
				var count:uint = (listeners == null) ? 0 : listeners.length;
				if (count > 0) {
					for (var i:uint = 0; i < count; ++i) {
						adding ? 	activePart.addEventListener(types[i], listeners[i]) :
							activePart.removeEventListener(types[i], listeners[i]);
					}
				}
			}
			adding ? super.partAdded(partName, instance) : super.partRemoved(partName, instance);
		}
		
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function get allowMultiPage():Boolean
		{
			// Calculating the page count is expensive, so only allow multi-page at all if our attachment supports it
			var previewState:String = collectionManager.apiIsSupported(CollectionManager.API_VERSION_10) ?
				PreviewFactory.instance.previewStateFromMimeType(mimeTypeFromAttachment(), extensionFromAttachment()) :
				Preview.THUMBNAIL;
			return (previewState == Preview.PAGING_THUMBNAIL);
		}
		
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function get prefersMediaPreview():Boolean
		{
			var preferredState:String = preferredPreviewState;
			return ((preferredState == Preview.AUDIO) || (preferredState == Preview.VIDEO));
		}
		
		
		/**
		 *  Returns the preferred PreviewState. This is based on the MIME type
		 *  of the associated attachement if supported. 
		 * 
		 * 	@inheritDoc
		 * 
		 * 	@see Preview#AUDIO
		 * 	@see Preview#PAGING_THUMBNAIL
		 * 	@see Preview#THUMBNAIL
		 * 	@see Preview#VIDEO
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function get preferredPreviewState():String
		{
			var previewState:String = collectionManager.apiIsSupported(CollectionManager.API_VERSION_10) ?
				PreviewFactory.instance.previewStateFromMimeType(mimeTypeFromAttachment(), extensionFromAttachment()) :
				Preview.THUMBNAIL;
			if (previewState == Preview.PAGING_THUMBNAIL && (!allowMultiPage || !ThumbnailManager.instance.enableDynamicThumbnails || (attachment.pageCount <= 1)))
				previewState = Preview.THUMBNAIL;
			
			return previewState;
		}
		
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function invalidatePreviewFromSetData(invalidateModel:Boolean = false, generateThumb:Boolean = true):void {
			invalidatePreview(false, false);
		}
		
		
		/**
		 *  Dispatches a closeButton event
		 * 
		 *  @param event The triggering event
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onCloseClick(evt:Event):void
		{
			dispatchEvent(new Event("closeButton"));
		}
		
		
		/**
		 *  The event handler for closing the previewCard's info state.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onInfoCloseClick(evt:Event):void
		{
			currentState = STATE_PREVIEW;
		}
		
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		private function handleFileInfo(e:PreviewEvent):void
		{
			if (currentState == STATE_INFO) {
				currentState = STATE_PREVIEW;
			} else {
				currentState = STATE_INFO;
			}
		}
		
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		private function handleFileExtract(e:PreviewEvent):void
		{
			if (attachment == null)
				return;
			
			CommandManager.instance.extractAttachment(attachment);
		}
		
//		override public function get downSampleOK():Boolean
//		{
//			return true; // always ok
//		}

		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function getPreviewSize():Point
		{
			const FIT_MIN:Number = 128;
			const FIT_MAX:Number = 2048;
			var size:Point = new Point(preview.width, preview.height);
			
			var width:Number = MathUtil.clipToRange(size.x, FIT_MIN, FIT_MAX);
			var height:Number = MathUtil.clipToRange(size.y, FIT_MIN, FIT_MAX);
			
			return new Point(width,height);
		}
		
		/**
		 * 	@inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function generateThumbnail():void 
		{
			preview.thumbnailPriority = ThumbnailManager.PREVIEW_THUMB_PRIORITY;
			super.generateThumbnail();
		}
		
		[Bindable]
		/**
		 *  Whether page navigation is enabled (via keyboard, mousewheel)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.1
		 */	
		public function get pageNavigationEnabled():Boolean
		{
			return _pageNavigationEnabled;
		}

		public function set pageNavigationEnabled(value:Boolean):void
		{
			_pageNavigationEnabled = value;
		}

	}
}
