/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.cards
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.api.IField;
	import com.adobe.portfolio.components.NavSkinnableComponent;
	import com.adobe.portfolio.components.supportClasses.ColorPalette;
	import com.adobe.portfolio.events.PreviewEvent;
	import com.adobe.portfolio.events.ThumbnailManagerEvent;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.ThumbnailManager;
	import com.adobe.portfolio.previews.IPreview;
	import com.adobe.portfolio.previews.IPreviewMediaModel;
	import com.adobe.portfolio.previews.IPreviewModel;
	import com.adobe.portfolio.previews.IPreviewMultiPageModel;
	import com.adobe.portfolio.previews.Preview;
	import com.adobe.portfolio.previews.PreviewFactory;
	import com.adobe.portfolio.previews.PreviewMediaModel;
	import com.adobe.portfolio.utils.FieldUtils;
	import com.adobe.portfolio.utils.FileUtil;
	import com.adobe.portfolio.utils.MathUtil;
	
	import flash.display.Bitmap;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.events.TimerEvent;
	import flash.geom.Point;
	import flash.utils.Timer;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.controls.Image;
	import mx.controls.listClasses.IListItemRenderer;
	import mx.core.UIComponent;
	import mx.events.CollectionEvent;
	import mx.events.CollectionEventKind;
	import mx.events.FlexEvent;
	import mx.events.PropertyChangeEvent;
	import mx.events.ResizeEvent;
	import mx.logging.ILogger;
	import mx.logging.Log;
	import mx.managers.IFocusManagerComponent;
	
	import spark.components.Button;
	import spark.components.Group;
	import spark.components.IItemRenderer;
	import spark.components.supportClasses.Skin;
	import spark.primitives.BitmapImage;
	
	/**
	 *  Name of the skin class to use for this component when its data is
	 * 	set to a folder attachment.
	 *  
	 *  @langversion 3.0
	 *  @playerversion Flash 10
	 *  @playerversion AIR 1.5
	 *  @productversion Flex 4
	 */
	[Style(name="folderSkinClass", type="Class")]
	
	[SkinState("normal")]
	[SkinState("over")]
	[SkinState("selected")]
	[SkinState("overAndSelected")]
	
	/**
	 *  The SimpleCard class extends NavSkinnableComponent to show a simple card that displays a thumbnail corresponding to an IAttachment. 
	 * 
	 *  SimpleCard is a skinnable renderer for an IAttachment. It implements IListItemRenderer,
	 *  so it can be used as an item renderer for relevant containers. The skin is not required
	 *  to provide all the relevant controls, which allows this object to act as the core for
	 *  both simple (mini-nav) and complex (basic click-through card) rendering.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */	
	public class SimpleCard extends NavSkinnableComponent implements IFocusManagerComponent, IListItemRenderer, IItemRenderer, ISimpleCard
	{
		// String literals used to represent the skin parts exposed by SimpleCard to its skins.
		
		/**
		 *  A string literal used to represent the preview skin part exposed by SimpleCard
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PART_PREVIEW:String = "preview";
		
		/**
		 *  A string literal used to represent the icon skin part exposed by SimpleCard
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PART_ICON:String = "icon";
		
		/**
		 *  A string literal used to represent the ribbon skin part exposed by SimpleCard
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PART_RIBBON:String = "ribbon";
		
		// String literals used to represent the states of a SimpleCard skin.
		
		/**
		 *  A string literal used to represent the normal skin state of the SimpleCard
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const SKIN_STATE_NORMAL:String = "normal";
		
		/**
		 *  A string literal used to represent the "over" skin state of the SimpleCard
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const SKIN_STATE_OVER:String = "over";
		
		/**
		 *  A string literal used to represent the selected skin state of the SimpleCard
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const SKIN_STATE_SELECTED:String = "selected";
		
		/**
		 *  A string literal used to represent the over and selected skin state of the SimpleCard
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const SKIN_STATE_OVER_AND_SELECTED:String = "overAndSelected";
		
		
		/**
		 *  A string literal representing the name for the aspect ratio field
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const ASPECT_RATIO_FIELD:String = "adobe:AspectRatio";
		
		/**
		 *  The navigator's CollectionManager instance. All host interaction should go through this
		 *  instance, rather than the INavigatorHost directly.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected const collectionManager:CollectionManager = CollectionManager.instance;
		
		
		/**
		 *  A logger that can be used by this instance to log warnings, errors, or additional information.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected const logger:ILogger = Log.getLogger("SimpleCard");
		
		
		/**
		 *  The navigator's thumbnailManager instance. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected static const thumbnailMgr:ThumbnailManager = ThumbnailManager.instance;
		
		/**
		 *  If <code>true</code> the preview for this card is invalid.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var invalidPreview:Boolean;
		
		private var _delayThumbnailGeneration:Boolean;
		private var _enableThumbnailGeneration:Boolean;
		private var _downSampleOK:Boolean;

		private var previewResizeTimer:Timer;
		private var _fitWidth:Number;
		private var _fitHeight:Number;
		private var _generateThumb:Boolean;
		
		private var hostPageCountChangeWatcher:ChangeWatcher;
		private var hostPageNumberChangeWatcher:ChangeWatcher;
		private var enableDynamicThumbnailsWatcher:ChangeWatcher;
		
		
		/**
		 *  Default constructor for SimpleCard.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function SimpleCard()
		{
			super();
			maxWidth = 2048;
			maxHeight = 2048;
			_delayThumbnailGeneration = false;
			_enableThumbnailGeneration = false;
			_downSampleOK = false;
			invalidPreview = false;
			
			addEventListener(Event.ADDED_TO_STAGE, onAddedToStage, false, 0, true);
			addEventListener(FlexEvent.CREATION_COMPLETE,
				function (e:Event):void {
					e.target.removeEventListener(FlexEvent.CREATION_COMPLETE, arguments.callee);
					invalidatePreview();
				}
			);
//			logger.debug("SimpleCardCtor: " + this);
	}
		
		
		/**
		 *  @private
		 *
		 */	
		private function onAddedToStage(e:Event):void 
		{
			e.target.removeEventListener(e.type, arguments.callee);
			
			addEventListener(Event.REMOVED_FROM_STAGE, onRemovedFromStage, false, 0, true);
			addEventListener(MouseEvent.ROLL_OVER, onRoll, false, 0, true);
			addEventListener(MouseEvent.ROLL_OUT, onRoll, false, 0, true);
//			addEventListener(MouseEvent.CLICK, onMouseClick, false, 0, true);
			collectionManager.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onHostPropertyChange, false, 0, true);
			thumbnailMgr.addEventListener(ThumbnailManagerEvent.ASPECT_RATIO_CHANGED, onAspectRatioChanged, false, 0, true);
			enableDynamicThumbnailsWatcher = BindingUtils.bindSetter(onEnableDynamicThumbnails, thumbnailMgr, "enableDynamicThumbnails");
		}
		
		/**
		 *  @private
		 */	
		private function onRemovedFromStage(e:Event):void
		{
			e.target.removeEventListener(e.type, arguments.callee);
			
			removeEventListener(MouseEvent.ROLL_OVER, onRoll);
			removeEventListener(MouseEvent.ROLL_OUT, onRoll);
//			removeEventListener(MouseEvent.CLICK, onMouseClick);
			if (collectionManager)
				collectionManager.removeEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onHostPropertyChange);
			thumbnailMgr.removeEventListener(ThumbnailManagerEvent.ASPECT_RATIO_CHANGED, onAspectRatioChanged, false);
			if (enableDynamicThumbnailsWatcher != null) {
				enableDynamicThumbnailsWatcher.unwatch();
				enableDynamicThumbnailsWatcher = null;
			}
		}
		
		
		/**
		 *  Handle the enabling of dynamic thumbnails.
		 *  This function makes sure that cards supporting multi-page
		 *  start listening for page events.
		 * 
		 *  @param value If <code>true</code> cards supporting multi-page
		 *  will start listening for page events.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onEnableDynamicThumbnails(value:Boolean):void
		{
			if (value && preview) {
				// Make sure cards supporting multi-page start listening for page events
				updatePreviewModelListeners(false);
				updatePreviewModelListeners(true);
			}
		}
		
		private var _mouseIsOver:Boolean = false;
		
		/**
		 *  Returns <code>true</code> if the mouse is over the card;
		 *  returns <code>false</code> otherwise.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function get mouseIsOver():Boolean
		{
			return _mouseIsOver;
		}
		
		/**
		 *  @private
		 *
		 */	
		protected function set mouseIsOver(value:Boolean):void
		{
			if (value != _mouseIsOver) {
				_mouseIsOver = value;
				invalidateSkinState();
			}
		}
		/**
		 *  @private
		 *
		 */	
		private function onRoll(event:MouseEvent):void
		{
			var mouseIsOver:Boolean = (event.type == MouseEvent.ROLL_OVER);
			if (!mouseIsOver) {
//				logger.debug("event.stageX, Y = " + event.stageX + " , " + event.stageY);
				if (event.stageX < 1 && event.stageY < 1) // weird values for some rollout events which shouldn't be rollout
					mouseIsOver = true;
			}
			this.mouseIsOver = mouseIsOver;
		}
		
		/**
		 *  A listener for PropertyChangeEvents on the collectionManager.
		 * 
		 *  @param event The host's PropertyChangeEvent to be handled.
		 * 
		 * 	@see CollectionManager
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onHostPropertyChange(event:PropertyChangeEvent):void
		{
			switch (event.property) {
				case "authoringMode":
					// The changed property affects our skin state, so invalidate it
					invalidateSkinState();
					break;
				default:
					// nothing to do
					break;
			}
		}
		
		private var _dropShadowPreference:String = CardFactory.DROP_SHADOW_NONE;
		

		[Bindable]
		/**
		 *  @inheritDoc
		 * 
		 *  @see CardFactory
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get dropShadowPreference():String
		{
			return _dropShadowPreference;
		}
		
		/**
		 * @private
		 */
		public function set dropShadowPreference(value:String):void
		{
			_dropShadowPreference = value;
		}
		
		private var _rollOverPreference:String = CardFactory.ROLL_OVER_NONE;
		
		[Bindable]
		/**
		 *  @inheritDoc
		 * 
		 * 	@see CardFactory
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get rollOverPreference():String
		{
			return _rollOverPreference;
		}
		
		/**
		 * @private
		 */
		public function set rollOverPreference(value:String):void
		{
			_rollOverPreference = value;
		}
		
		private function previewAspectRatioChanged(event:PreviewEvent):void
		{
			setAspectRatio(preview.aspectRatio, preview.model is PreviewMediaModel);
		}
		
		[SkinPart(required="false")]
		/**
		 *  A UI Component whose left/right/top/bottom values can be adjusted to fit card
		 *  UI to a particular aspect ratio without affecting the bounds of the card.
		 * 
		 * 	@see #setLayoutBoundsSize
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var paddingGroup:UIComponent;

		[SkinPart(required="false")][Bindable]
		/**
		 *  The preview component for this card. This skin part can also be accessed
		 * 	more generically via previewComponent in the ISimpleCard interface.
		 * 
		 * 	@see #previewComponent
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var preview:Preview;
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get previewComponent():IPreview
		{
			return preview;
		}
		
		[SkinPart(required="false")]
		/**
		 *  A skin part representing the ribbon component that should be visible when the user
		 *  is editing the attament in an external application.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var ribbon:Group;
		
		/** 
		 *  @private
		 *  Internal representation of the String referenced by the fileName getter and setter.
		 */
		private var _fileName:String = "";
		[Bindable]
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get fileName():String
		{
			return _fileName;
		}
		/** 
		 *  @private
		 *  This property is considered read-only; as a result, its setter is not public,
		 *  and is only used to update the internal values based on attachment changes.
		 *
		 */	
		protected function set fileName(val:String):void
		{
			if ((attachment != null) && (attachment.fileName != val)) {
				logger.error("fileName should only be set according to attachment.fileName");
			}
			_fileName = val;
		}
		
		/** Internal representation of the String referenced by the description getter and setter.
		 */
		private var _description:String = "";
		[Bindable]
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get description():String
		{
			return _description;
		}
		/**
		 *  @private
		 */	
		public function set description(val:String):void
		{
			if (_description != val) {
				if (val != null) {
					_description = val;
				} else {
					_description = "";
				}
				// Change the attachment's description, if necessary
				if ((attachment != null) && (attachment.description != _description)) {
					attachment.description = _description;
				}
			}
		}
		
		/** 
		 *  @private
		 *  Internal representation of the Date referenced by the modDate getter and setter.
		 */
		private var _modDate:Date = new Date();
		[Bindable]
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get modDate():Date
		{
			return _modDate;
		}
		/** 
		 *  @private
		 *  This property is considered read-only; as a result, its setter is not public,
		 *  and is only used to update the internal values based on attachment changes.
		 */
		protected function set modDate(val:Date):void
		{
			if ((attachment != null) && (attachment.modDate != val)) {
				logger.error("modDate should only be set according to attachment.modDate");
			}
			_modDate = val;
		}
		
		/** 
		 *  @private
		 *  Internal representation of the Date referenced by the creationDate getter and setter.
		 */
		private var _creationDate:Date = new Date();
		[Bindable]
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get creationDate():Date
		{
			return _creationDate;
		}
		/** 
		 *  @private
		 *  This property is considered read-only; as a result, its setter is not public,
		 *  and is only used to update the internal values based on attachment changes.
		 */
		protected function set creationDate(val:Date):void
		{
			if ((attachment != null) && (attachment.creationDate != val)) {
				logger.error("creationDate should only be set according to attachment.creationDate");
			}
			_creationDate = val;
		}
		
		/** Internal representation of the uint referenced by the size getter and setter.
		 */
		private var _size:uint = 0;
		[Bindable]
		/**
		 * 	@inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get size():uint
		{
			return _size;
		}
		/** 
		 *  @private
		 *  This property is considered read-only; as a result, its setter is not public,
		 *  and is only used to update the internal values based on attachment changes.
		 */
		protected function set size(val:uint):void
		{
			if ((attachment != null) && (attachment.size != val)) {
				logger.error("size should only be set according to attachment.size");
			}
			_size = val;
		}
		
		/** Internal representation of the uint referenced by the compressedSize getter and setter.
		 */
		private var _compressedSize:uint = 0;
		[Bindable]
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get compressedSize():uint
		{
			return _compressedSize;
		}
		/** 
		 *  @private
		 *  This property is considered read-only; as a result, its setter is not public,
		 *  and is only used to update the internal values based on attachment changes.
		 */
		protected function set compressedSize(val:uint):void
		{
			if ((attachment != null) && (attachment.compressedSize != val)) {
				logger.error("compressedSize should only be set according to attachment.compressedSize");
			}
			_compressedSize = val;
		}
		
		/** Internal representation of the ColorPalette referenced by the colorPalette getter and setter.
		 */
		private var _localColorPalette:ColorPalette = null;

		[Bindable]
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get colorPalette():ColorPalette
		{
			return (_localColorPalette != null) ? _localColorPalette : collectionManager.colorPalette;
		}
		/**
		 *  @private
		 */	
		protected function set colorPalette(value:ColorPalette):void
		{
			// TODO This will most likely require writing some sort of properties into the collection
			_localColorPalette = value;
		}
		
		/**
		 *  The IAttachment data object being rendered by this component. It is set through the public
		 *  data setter.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var attachment:IAttachment = null;
		
		private var _numAttachments:int = 0;
		[Bindable]
		/**
		 *  The number of attachments in a folder
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get numAttachments():int
		{
			return _numAttachments;
		}
		
		/**
		 *  @private
		 */	
		private function set numAttachments(value:int):void
		{
			_numAttachments = value;
		}
		
		/**
		 *  @private
		 */	
		private function updateNumAttachments():void
		{
			if ((attachment == null) || (attachment.children == 0)) {
				numAttachments = 0;
			} else {
				numAttachments = attachment.children.length;
			}
		}
		
		/**
		 * An event listener that's called when this attachments children dispatches
		 * a collectionChange event.
		 * 
		 * @see IAttachment#children
		 */
		protected function onChildrenChanged(event:CollectionEvent):void
		{
			if (event.kind == CollectionEventKind.UPDATE) {
				// don't bother doing anything on an update
			} else {
				updateNumAttachments();
			}
		}

		[Bindable]
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get data():Object
		{
			return attachment;
		}
		
		/**
		 *  @private
		 */	
		public function set data(val:Object):void
		{
			if (val == attachment)
				return;
			
			if (attachment != null) {
				// Removing the old event listener is important to avoid leaks and garbage collection dependencies
				attachment.removeEventListener("fileNameChanged", onAttachmentChange);
				attachment.removeEventListener("descriptionChanged", onAttachmentChange);
				attachment.removeEventListener("modDateChanged", onAttachmentChange);
				attachment.removeEventListener("creationDateChanged", onAttachmentChange);
				attachment.removeEventListener("sizeChanged", onAttachmentChange);
				attachment.removeEventListener("compressedSizeChanged", onAttachmentChange);
				attachment.removeEventListener("isOpenChanged", isOpenChanged);
				if (icon != null) {
					attachment.removeEventListener("iconChanged", onIconChange);
				}
				if (attachment.isFolder) {
					attachment.children.removeEventListener(CollectionEvent.COLLECTION_CHANGE, onChildrenChanged);
				}
				if (preview != null) {
					invalidatePreview(true);
				}
			}
			
			var prevVal:IAttachment = attachment;
			attachment = val as IAttachment;
			
			if (attachment != null) {
				/*
				// This forces us to get an aspectRatio early, but can hurt intro animations due
				// to upfront activity.
				if (maintainAspectRatio && isNaN(thumbnailMgr.getAspectRatio(attachment, false, true))) {
					// Request the cached thumb right away, so we'll have its aspect ratio
					attachment.requestCachedThumbnail(this, 
						function (id:*, bitmap:Bitmap):void {
							if (bitmap &&
								bitmap.bitmapData &&
								(bitmap.bitmapData.width > 0) &&
								(bitmap.bitmapData.height > 0))
							{
								thumbnailMgr.setAspectRatio(attachment, bitmap.bitmapData.width / bitmap.bitmapData.height, false);
							}
						}
					);
				}
				*/

				attachment.addEventListener("fileNameChanged", onAttachmentChange);
				attachment.addEventListener("descriptionChanged", onAttachmentChange);
				attachment.addEventListener("modDateChanged", onAttachmentChange);
				attachment.addEventListener("creationDateChanged", onAttachmentChange);
				attachment.addEventListener("sizeChanged", onAttachmentChange);
				attachment.addEventListener("compressedSizeChanged", onAttachmentChange);
				attachment.addEventListener("isOpenChanged", isOpenChanged);
				if (icon != null) {
					attachment.addEventListener("iconChanged", onIconChange);
				}
				
				fileName = attachment.fileName;
				description = attachment.description;
				modDate = attachment.modDate;
				creationDate = attachment.creationDate;
				compressedSize = attachment.compressedSize;
				size = attachment.size;
				isOpen = attachment.isOpen;
				invalidatePreviewFromSetData(false);
				
				var existingSkin:Class = getStyle("skinClass");
				var folderSkin:Class = getStyle("folderSkinClass");
				if (attachment.isFolder) {
					attachment.children.addEventListener(CollectionEvent.COLLECTION_CHANGE, onChildrenChanged);
					numAttachments = attachment.children.length;
					
					if ((existingSkin != folderSkin) && (folderSkin != null)) {
						setStyle("skinClass", folderSkin);
					}
				} else {
					if (existingSkin == folderSkin) {
						// TODO: return skinClass to its inherited value
						clearStyle("skinClass");
					}
					numAttachments = -1;
				}
			} else {
				// There's no attachment, but internal values should still be reset to a known state.
				fileName = "";
				description = "";
				modDate = new Date();
				creationDate = new Date();
				compressedSize = 0;
				size = 0;
				isOpen = false;
				numAttachments = -1;
			}
			setIconBitmap();
			
			// Just in case any client wants to use the item renderer's label, fileName is an appropriate value
			label = fileName;
			if (maintainAspectRatio) {
				invalidateSize();
			}
			
			//			validateNow();   // probably unnecessary, but leaving it as a comment for the time being
		}
		
		/**
		 *  Invalidate the preview because a new data object to be rendered has been set
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function invalidatePreviewFromSetData(invalidateModel:Boolean = false, generateThumb:Boolean = true):void 
		{
			// override as necessary
			invalidatePreview(invalidateModel, generateThumb);
		}
		
		/**
		 *  Handles changes in the underlying attachment events
		 * 
		 *  @param evt The change in the underlying attachement event
		 * 
		 * 	@see IAttachment
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onAttachmentChange(evt:Event):void
		{
			switch (evt.type) {
				case "fileNameChanged":
					fileName = attachment.fileName;
					if (preview) {
						updatePreviewModelListeners(false);
						invalidatePreview(true);
					}
					break;
				case "descriptionChanged":
					description = attachment.description;
					break;
				case "modDateChanged":
					modDate = attachment.modDate;
					break;
				case "creationDateChanged":
					creationDate = attachment.creationDate;
					break;
				case "sizeChanged":
					size = attachment.size;
					break;
				case "compressedSizeChanged":
					compressedSize = attachment.compressedSize;
					break;
				default:
					// TODO We didn't handle an event; is this worth logging as a warning?
					break;
			}
		}
		
		/**
		 *  Specifies whether or not this card should use a multi-page preview state.
		 * 	Calculating the page count is expensive, so only allow multi-page at all
		 * 	if the current attachment supports it.
		 *  
		 *  @Return <code>true</code> if this card should attempt to use a multi-page preview state,
		 *  or <code>false</code> otherwise.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function get allowMultiPage():Boolean
		{
			// Calculating the pageCount of an attachment is expensive, so always return false by default
			return false;
		}
		
		/**
		 *  Returns <code>true</code> if this card supports in-place multimedia playback
		 * 	for it's current attachment; <code>false</code> otherwise.
		 * 
		 *  @return <code>true</code> if this card supports in-place multimedia playback
		 * 	for it's current attachment; <code>false</code> otherwise.
		 * 
		 * 	@see #preferredPreviewState
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function get prefersMediaPreview():Boolean
		{
			return false;
		}
		
		/**
		 *  The preferred preview state for this card
		 * 
		 *  @return the preferred preview state
		 * 
		 * 	@see Preview#AUDIO
		 * 	@see Preview#PAGING_THUMBNAIL
		 *	@see Preview#THUMBNAIL
		 * 	@see Preview#VIDEO
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function get preferredPreviewState():String
		{
			return Preview.THUMBNAIL;
		}
		
		/**
		 *  Invalidate the current preview state for the card
		 * 
		 *  @param invalidateModel If <code>true</code> then nulify the preview model.
		 *  @param generateThumb If <code>true</code> then generate a thumbnail image.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function invalidatePreview(invalidateModel:Boolean = false, generateThumb:Boolean = true):void 
		{
			invalidateProperties();
			invalidPreview = true;
			if (invalidateModel && preview)
				preview.model = null;
			this.generateThumb = generateThumb;
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function commitProperties():void {
			super.commitProperties();
			if (invalidPreview)
				initPreview();
		}

		/**
		 *  @private
		 */
		private function initPreview():void 
		{
			if (!attachment)
				return;
			
			PreviewFactory.instance.initPreview(preview, preferredPreviewState, attachment);
			initPageCount();
			updatePreviewModelListeners(true);
			
			if (generateThumb)
				generateThumbnail();
			
			invalidPreview = false;
		}
		
		/**
		 *  Initialize the page count for the preview multipage model 
		 *  to the page count of the underlying attachement.
		 * 	Calculating the page count is expensive, so avoid calling this function
		 * 	when allowMultiPage is false.
		 * 
		 * 	@see #allowMultiPage
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function initPageCount():void
		{
			if (previewMultiPageModel && attachment.pageCount > 1) {
				previewMultiPageModel.pageCount = attachment.pageCount;
			}
		}
		
		/**
		 *  Have the collectionManager handle modal previe experience
		 * 	for the underlying attachment. This is usually handled within
		 * 	the portfolio by the NavPreview component, or by the host
		 * 	application through the preview command.
		 * 
		 * 	@see NavPreview
		 * 	@see CommandType#PREVIEW
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function handlePreview():void 
		{
			collectionManager.handlePreview(attachment);
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function getCurrentSkinState():String
		{
			if (selected) {
				return mouseIsOver ? SKIN_STATE_OVER_AND_SELECTED : SKIN_STATE_SELECTED;
			} else {
				return mouseIsOver ? SKIN_STATE_OVER : SKIN_STATE_NORMAL;
			}
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function partAdded(partName:String, instance:Object):void
		{
			//handlePartChange will call super.partAdded
			handlePartChange(partName, instance, true);
		}

		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function partRemoved(partName:String, instance:Object):void
		{
			//handlePartChange will call super.partRemoved
			handlePartChange(partName, instance, false);
		}
		
		/** 
		 *  @private
		 *  Internal function for adding and removing event listeners for skin parts, depending on
		 *  whether the part is being added or removed for this instance.
		 */
		private function handlePartChange(partName:String, instance:Object, adding:Boolean):void
		{
			var listeners:Array = null;
			var types:Array = null;
			if ((instance != null) && (instance is IEventDispatcher)) {
				var activePart:IEventDispatcher = instance as IEventDispatcher;
				
				// Build up an array of event types and listeners that apply to this object
				switch (partName) {
					case PART_PREVIEW:
						// TODO Listen for ThumbnailManager.aspectRatio event instead
						types = [ ResizeEvent.RESIZE, MouseEvent.DOUBLE_CLICK, PreviewEvent.ASPECT_RATIO_CHANGED ];
						listeners = [ onPreviewResize, onPreviewMouseDoubleClick, previewAspectRatioChanged ];
						if (adding)
							invalidatePreview();
						else {
							updatePreviewModelListeners(adding);
							invalidatePreview(true, false);
						}
						break;
					case PART_ICON:
						if (attachment != null) {
							if (adding) {
								attachment.addEventListener("iconChanged", onIconChange);
							} else {
								attachment.removeEventListener("iconChanged", onIconChange);
							}
						}
						setIconBitmap();
						break;
					case PART_RIBBON:
						if (adding) {
							ribbon.visible = isOpen;
						}
						break;
				}
				
				// Add or remove all the relevant event listeners
				var count:uint = (listeners == null) ? 0 : listeners.length;
				if (count > 0) {
					for (var i:uint = 0; i < count; ++i) {
						adding ? 	activePart.addEventListener(types[i], listeners[i]) :
							activePart.removeEventListener(types[i], listeners[i]);
					}
				}
			}
			adding ? super.partAdded(partName, instance) : super.partRemoved(partName, instance);
		}
		
		private var addedPageListeners:Boolean;
		
		/**
		 *  Update the listeners for preview model events.
		 * 
		 *  @param adding Set to <code>true</code> if you are adding attachments
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function updatePreviewModelListeners(adding:Boolean):void 
		{
			var model:IPreviewModel = preview ? preview.model : null;
			if (!model)
				return;
			
			if (collectionManager.apiIsSupported(CollectionManager.API_VERSION_10)) {
				// page APIs were added in A10, so don't bother listening otherwise
				if (adding) {
					addedPageListeners = allowMultiPage && ThumbnailManager.instance.enableDynamicThumbnails;
					if (addedPageListeners) {
						hostPageCountChangeWatcher = BindingUtils.bindSetter(hostPageCountChanged, attachment, "pageCount");
						hostPageNumberChangeWatcher = BindingUtils.bindSetter(hostPageNumberChanged, attachment, "pageNumber");
						model.addEventListener("pageCountChanged", pageCountChanged); 
						model.addEventListener("pageNumberChanged", pageNumberChanged);
					}
				} else {
					if (hostPageCountChangeWatcher != null) {
						hostPageCountChangeWatcher.unwatch();
						hostPageCountChangeWatcher = null;
					}
					if (hostPageNumberChangeWatcher != null) {
						hostPageNumberChangeWatcher.unwatch();
						hostPageNumberChangeWatcher = null;
					}
					if (addedPageListeners) {
						model.removeEventListener("pageCountChanged", pageCountChanged);
						model.removeEventListener("pageNumberChanged", pageNumberChanged);
					}
					addedPageListeners = false;
				}
			}
		}
		
		/**
		 *  Get a preview model that supports multipage preview of attachments, or
		 * 	null if multipage preview is not supported by the current preview model.
		 * 
		 *  @return a multipage preview model
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function get previewMultiPageModel():IPreviewMultiPageModel
		{
			if (!preview)
				return null;
			
			return preview.model as IPreviewMultiPageModel;
		}
		
		/**
		 *  This function is called if the page count of the multipage model
		 *  has changed
		 * 
		 *  @param val If 0 then do nothing; otherwise invalidate the preview.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function hostPageCountChanged(val:uint):void 
		{
			if (val == 0)
				return;
			
			var model:IPreviewMultiPageModel = previewMultiPageModel;
			if (!model) {
				if (val > 1 && preferredPreviewState != Preview.THUMBNAIL) {
					invalidatePreview(true);
				}
			} else if (val == 1) { // back to a single-page model
				invalidatePreview(true);
			}
		}
		
		
		/**
		 *  @private
		 */		
		private function hostPageNumberChanged(val:uint):void 
		{
			var model:IPreviewMultiPageModel = previewMultiPageModel;
			if (!model)
				return;
			
			if (model.pageNumber == val)
				return;
			
			model.pageNumber = val;   
			CONFIG::logging {
				logger.debug("hostPageNumberChanged: val= " + val);
			}
		}
		
		/**
		 *  @private
		 */		
		private function pageCountChanged(e:Event):void 
		{
//			generateThumbnail();
			invalidatePreview(true, true);
		}
		
		/**
		 *  @private
		 */		
		private function pageNumberChanged(e:Event):void 
		{
			var model:IPreviewMultiPageModel = previewMultiPageModel;
			if (!model)
				return;
			
			invalidatePreview(false, true);
			attachment.pageNumber = model.pageNumber;
		}
		
		/**
		 *  @private
		 */		
		private function onPreviewMouseDoubleClick(e:MouseEvent):void {
			if (e.target is Button) {
				// Don't initiate a preview if the user has clicked a button twice quickly
				// unless we're using a media model, in which case the whole thing is a button
				if ((preview == null) || !(preview.model is IPreviewMediaModel)) {
					return;
				}
			}
			handlePreview();	
		}
		
		/**
		 *  @private
		 */		
		private function onPreviewResize(e:ResizeEvent):void 
		{
			if (!preview)
				return;
			
			if (!attachment)
				return;
			
			if (preview.width == 0 || preview.height == 0)
				return;
			
			if (!delayThumbnailGeneration) {
				invalidatePreview(false, true);
				return;
			}
			
			if (previewResizeTimer)
				return;
			
//			log("onPreviewResize: preview w,h= " + preview.width + "," + preview.height, "onPreviewResize");
			var startWidth:Number = preview.width;
			var startHeight:Number = preview.height;
			previewResizeTimer = new Timer(50, 0);
			previewResizeTimer.addEventListener(TimerEvent.TIMER,
				function (e:Event):void {
					previewResizeTimer.stop();
					if (startWidth == preview.width && startHeight == preview.height) {
						previewResizeTimer.removeEventListener(TimerEvent.TIMER, arguments.callee);
						invalidatePreview(false, true);
						previewResizeTimer = null;
					} else {
						startWidth = preview.width;
						startHeight = preview.height;
						previewResizeTimer.start();
					}
				}
			);
			previewResizeTimer.start();
		}
		
		/**
		 *  Generate a thumbnail image for the underlying attachment.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function generateThumbnail():void 
		{
			if (!preview || !preview.model)
				return;
			
			if (width == 0 || height == 0)
				return;
			
			if (preview.width == 0 || preview.height == 0)
				return;
			
			if (!enableThumbnailGeneration)
				return;
			
			if (delayThumbnailGeneration)
				return;
			
			preview.generateThumbnail(this, attachment, getPreviewSize(), downSampleOK, onThumbComplete);
		}
		
		/**
		 *  @private
		 * 
		 */
		private function onThumbComplete(retPageNumber:uint, retWidth:Number, retHeight:Number):void 
		{
			if (retPageNumber == 1 && retWidth != 0 && retHeight != 0) {
				setAspectRatio(retWidth / retHeight, false);
			}
			invalidateDisplayList();
		}

		/**
		 *  Get the size of the preview image
		 * 
		 *  @return the width and height of the preview image as an (x,y) pair.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function getPreviewSize():Point
		{
			const FIT_MIN:Number = 128;
			const FIT_MAX:Number = 2048;
			
			var width:Number = MathUtil.clipToRange(preview.width, FIT_MIN, FIT_MAX);
			var height:Number = MathUtil.clipToRange(preview.height, FIT_MIN, FIT_MAX);
			
			return new Point(width,height);
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function set visible(value:Boolean):void {
			var newlyVisible:Boolean = !visible && value;
			super.visible = value;
			if (newlyVisible)
				onPreviewResize(null);
		}
		
		[SkinPart(required="false")]
		/**
		 *  The icon image for the card
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var icon:BitmapImage;
		
		/**
		 *  @private
		 */			
		private function setIconBitmap():void
		{
			if (icon != null) {
				if ((attachment != null) &&
					(attachment.icon != null) &&
					(attachment.icon.bitmapData != null))
				{
					icon.source = new Bitmap(attachment.icon.bitmapData);
				} else {
					icon.source = null;
				}
			}
		}
		
		/**
		 *  Handle icon change events from the underlying attachment
		 * 
		 *  @param The icon change event
		 * 
		 * 	@see IAttachment#icon 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onIconChange(event:Event):void
		{
			setIconBitmap();
		}
		
		/** 
		 * @private
		 * Internal representation of the Boolean referenced by the isOpen getter and setter. 
		 */
		private var _isOpen:Boolean = false;
		
		[Bindable]
		/**
		 *  Is <code>true</code> if the attachment is open in a native editor;
		 *  otherwise <code>false</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get isOpen():Boolean
		{
			return _isOpen;
		}
		/**
		 *  @private
		 */
		private function set isOpen(value:Boolean):void
		{
			if (value != _isOpen) {
				_isOpen = value;
				if (ribbon != null) {
					// Show/hide the ribbon skin part based on the the isOpen value.
					// We could have invalidated the skin state instead, and had ribbon variants of our
					// skin states, but that would have doubled our skin states for very simple functionality.
					ribbon.visible = value;
				}
			}
		}
		
		/**
		 * @private
		 * A listener that sets the card's <code>isOpen</code> property based on the current attachment.
		 */
		private function isOpenChanged(event:Event):void
		{
			isOpen = (attachment != null) ? attachment.isOpen : false;
		}
		
		// IItemRenderer implementation
		/** 
		 *  @private
		 *  Internal representation of the int referenced by the itemIndex getter and setter.
		 */
		private var _itemIndex:int = -1;
		
		/** 
		 *  @private
		 *  Internal representation of the string referenced by the label getter and setter. 
		 */
		private var _label:String = ""; 
		
		/**
		 *  The index of the item in the data set 
		 *  that this renderer is visualizing.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get itemIndex():int
		{
			return _itemIndex;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set itemIndex(val:int):void
		{
			if(_itemIndex != val)
				_itemIndex = val;
		}
		
		/** 
		 *  @private
		 *  Internal representation of the Boolean referenced by the selected getter and setter. 
		 */
		private var _selected:Boolean = false;
		
		[Bindable]
		/**  
		 * Is <code>true</code> if the item renderer 
		 * can show itself as selected; otherwise it is <code>false</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get selected():Boolean
		{
			return _selected;
		}
		
		/**
		 *  @private
		 */	
		public function set selected(value:Boolean):void
		{
			if (value != _selected) {
				_selected = value;
				invalidateSkinState();
			}
		}
		
		
		/** 
		 *  @private
		 *  Internal representation of the Boolean referenced by the showsCaret getter and setter. 
		 */
		private var _showsCaret:Boolean = false;
		
		/**
		 *  Is <code>true</code> if the item renderer can show itself as focused;
		 *  otherwise it is <code>false</code>. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get showsCaret():Boolean
		{
			return _showsCaret;
		}
		/**
		 *  @private
		 */	
		public function set showsCaret(value:Boolean):void
		{
			if(value != _showsCaret)
				_showsCaret = value;
		}
		
		/**
		 *  Is <code>true</code> if the item renderer is being dragged;
		 *  otherwise <code>false</code>. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get dragging():Boolean
		{
			return false;
		}
		/**
		 *  @private
		 */	
		public function set dragging(value:Boolean):void
		{
			
		}
		
		/**
		 *  The String to display in the item renderer. 
		 *
		 *  <p>The host component of the item renderer can use the 
		 *  <code>itemToLabel()</code> method to convert the data item 
		 *  to a String for display by the item renderer. </p>
		 * 
		 *  <p>For controls like List and ButtonBar, you can use the 
		 *  <code>labelField</code> or <code>labelFunction</code> properties 
		 *  to specify the field of the data item that contains the String.
		 *  Otherwise the host component uses the <code>toString()</code> method 
		 *  to convert the data item to a String. </p>
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get label():String
		{
			return _label;
		}
		/**
		 *  @private
		 *
		 */	
		public function set label(val:String):void
		{
			if(_label != val)
				_label = val;
		}
		
		/**
		 *  The String to display in the item renderer. 
		 *
		 *  <p>For controls that do not have <code>labelField</code>
		 *  or <code>labelFunction</code> properties to specify the field
		 *  of the data item that contains the string to display,
		 *  use <code>toString()</code> to convert the data item to a string.</p>
		 * 
		 *  @return The String to display in the item renderer. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function toString():String 
		{
			var s:String = super.toString();
			s = s.substr(s.lastIndexOf("."));
			s += ": " + label;
			return s;
		}
		
		private var _maintainAspectRatio:Boolean = false;
		/**
		 *  Is <code>true</code> if the aspect ratio of the card should be
		 * 	maintained by setLayoutBoundsSize; otherwise it is <code>false</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get maintainAspectRatio():Boolean
		{
			return _maintainAspectRatio;
		}
		/**
		 *  @private
		 */	
		public function set maintainAspectRatio(value:Boolean):void
		{
			if (value == _maintainAspectRatio) {
				return;
			}
			_maintainAspectRatio = value;
		}
		
		/**
		 *  Get the aspect ratio of the thumbnail for the attachment.
		 * 
		 *  @return The aspect ratio as a single number representing width/height
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function getAspectRatio():Number
		{
			if (attachment == null) {
				return Number.NaN;
			}
			var aspectRatio:Number = attachment.getFieldValue(ASPECT_RATIO_FIELD);
			if (isNaN(aspectRatio)) {
				// If there's no aspect ratio field, try our cached value instead
				aspectRatio = thumbnailMgr.getAspectRatio(attachment, prefersMediaPreview, true);
			}
			return aspectRatio;
		}
		
		/**
		 *  Set the aspect ratio of the thumbnail for the attachment.
		 * 
		 *  @param value The aspect ratio as a single number representing width/height
		 *  @param mediaValue Set to <code>true</code> if the card for an audio or
		 *  video attachment; otherwise set to <code>false</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function setAspectRatio(value:Number, mediaValue:Boolean = false):void
		{
			thumbnailMgr.setAspectRatio(attachment, value, mediaValue);
			if (mediaValue) {
				// We only cache thumbnail values in the aspectRatio field
				return;
			}
			
			collectionManager.getDirty(
				function (isDirty:Boolean):void {
					if (isDirty)
						setAspectRatioField(value);
				}
			);
		}
		
		/**
		 *  The event handler for changes in aspect ratio of the thumbnail.
		 * 
		 *  @param event the change in aspect ratio event
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onAspectRatioChanged(event:PropertyChangeEvent):void
		{
			if ((event == null) || (event.source != attachment)) {
				return;
			}
			if ((event.property == "mediaAspectRatio") != prefersMediaPreview) {
				// Media cards care about mediaAspectRatio; thumbnail cards care about aspectRatio
				return;
			}
			
			explicitWidth = explicitHeight = NaN;  // must set these to NaN for validateSize() to work later on
			invalidateSize();
		}
		
		/**
		 *  Set the aspect ratio of the appropriate field for the attachment
		 * 
		 *  @param value The aspect ratio as a single number representing width/height
		 * 
		 * 	@see #ASPECT_RATIO_FIELD
		 * 	@see FieldUtils
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function setAspectRatioField(value:Number):void
		{
			var aspectRatioField:IField = FieldUtils.getFieldByFieldName(collectionManager.fields, ASPECT_RATIO_FIELD);
			if (aspectRatioField) {
				var aspectRatio:Number = FieldUtils.getFieldValue(attachment, aspectRatioField);
				if (isNaN(aspectRatio) || (aspectRatio == 1 && value != aspectRatio)) {
					FieldUtils.setFieldValue(attachment, aspectRatioField, value);
				}
			}
		}
		
		/**
		 *  Sets the layout size of the card. This is the size that 
		 *  the card element uses to draw on screen. 
		 *  <p>If the width and/or height parameters are left unspecified,
		 *  Flex sets the element's layout size to its preferred width and/or
		 *  preferred height.</p>
		 *  <p> Note that calls to the setLayoutBoundSize() method can affect
		 *  the layout position, so call setLayoutBoundPosition() after
		 *  calling setLayoutBoundSize().</p>
		 * 	<p> If maintainAspectRatio is true, then this call will attempt to
		 * 	size the card around its preview component by reducing the width or height,
		 * 	or by adjusting the layout values of paddingGroup.</p>
		 * 
		 *  @param width The desired width for the card
		 *  @param height The desired height for the card
		 *  @param postLayoutTransform  When postLayoutTransform is <code>true</code>,
		 *  the specified sizes are those of the card's bounding box.  The bounding box
		 *  is in card's parent coordinate space and is calculated from the card's layout size,
		 *  layout position and layout transform matrix.
		 * 
		 * 	@see #maintainAspectRatio
		 * 	@see #paddingGroup
		 * 	@see #preview
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function setLayoutBoundsSize(width:Number, height:Number, postLayoutTransform:Boolean=true):void
		{
			if (attachment == null) {
				super.setLayoutBoundsSize(width, height, postLayoutTransform);
				return;
			}
			if (preview == null) {
				if (maintainAspectRatio &&
					attachment.isFolder &&
					((getStyle("folderSkinClass") as Class) != null))
				{
					// TODO: 	We don't want to scale our folder cards beyond their measured width/height,
					//			but we should have a better way of indicating this behavior
					super.setLayoutBoundsSize(Math.min(measuredWidth, width), Math.min(measuredHeight, height));
				} else {
					super.setLayoutBoundsSize(width, height, postLayoutTransform);
				}
				return;
			}
			
			var previewPadding:Point = getPadding(preview);
			if (previewPadding == null) {
				previewPadding = new Point();
			}
			
			var adjustedWidth:Number = width;
			var adjustedHeight:Number = height;
			var shouldAdjustToAspectRatio:Boolean = false;
			if (isNaN(adjustedWidth)) {
				if (isNaN(adjustedHeight)) {
					// The preview should default to a good size, so just use the measured width/height
					adjustedWidth = measuredWidth;
					adjustedHeight = measuredHeight;
					shouldAdjustToAspectRatio = true;
				} else {
					// only width is NaN, so assume we can make it as wide as we want
					//adjustedWidth = maxWidth;
				}
			} else if (isNaN(adjustedHeight)) {
				// only height is NaN, so assume we can make it as high as we want
				//adjustedHeight = maxHeight;
			} else {
				shouldAdjustToAspectRatio = true;
			}
			
			// TODO If we wanted to explicitly set the desired preview width/height
			//var previewWidth:Number = adjustedWidth - previewPadding.x;
			//var previewHeight:Number = adjustedHeight - previewPadding.y;

			var aspectRatio:Number = getAspectRatio();
			if (isNaN(aspectRatio) || (aspectRatio <= 0)) {
				shouldAdjustToAspectRatio = false;
			}
			if (maintainAspectRatio && shouldAdjustToAspectRatio) {
				// Both width and height are real values, so we'll need to match our aspect ratio
				if (paddingGroup != null) {
					// Ignore this padding when calculating the relative size
					previewPadding.x -= (paddingGroup.left + paddingGroup.right);
					previewPadding.y -= (paddingGroup.top + paddingGroup.bottom);
				}
				var relativeWidth:uint = adjustedWidth;
				var relativeHeight:uint = ((adjustedWidth - previewPadding.x) / aspectRatio) + previewPadding.y;
				if (relativeHeight > adjustedHeight) {
					relativeWidth = ((adjustedHeight - previewPadding.y) * aspectRatio) + previewPadding.x;
					relativeHeight = adjustedHeight;
				}
				if (paddingGroup != null) {
					// If we have a padding group, account for the aspect ratio adjustments there
					var existingPadding:Number = (paddingGroup.left as Number) + (paddingGroup.right as Number);
					if (Math.abs(existingPadding - (width - relativeWidth)) > 2) {
						paddingGroup.left = paddingGroup.right = (width - relativeWidth) / 2;
					}
					existingPadding = (paddingGroup.top as Number) + (paddingGroup.bottom as Number);
					if (Math.abs(existingPadding - (height - relativeHeight)) > 2) {
						paddingGroup.top = paddingGroup.bottom = (height - relativeHeight) / 2;
					}
				} else {
					// If we don't have a padding group, override the requested width and height
					width = relativeWidth;
					height = relativeHeight;
				}
			} else {
				width = adjustedWidth;
				height = adjustedHeight;
			}
			
			super.setLayoutBoundsSize(width, height, postLayoutTransform);
		}

		/**
		 *  Get the horizontal and vertical padding for this UI component.
		 * 
		 *  @return The horizontal and vertical padding as an ordered pair:
		 *  (horizontal padding, vertical padding). The padding is the sum of
		 *  the padding on both sides (left + right for horizontal padding;
		 *  top + bottom for vertical padding).
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get padding():Point
		{
			if (!preview)
				return new Point();
			
			return getPadding(preview);
		}
		
		/**
		 *  @private
		 * 
		 */
		private function getPadding(component:UIComponent, shallowOnly:Boolean = false):Point 
		{
			if (!component)
				return null;
			
			var hPad:Number = 0;
			var vPad:Number = 0;
			
			do {
				if (component.left != null) {
					hPad += component.left;
				}
				if (component.right != null)
					hPad += component.right;
				
				if (component.top != null) {
					vPad += component.top;
				}
				if (component.bottom != null)
					vPad += component.bottom;
				
				if (shallowOnly) {
					break;
				}
				
//				logger.debug(this + " component.left= " + component.left);
//				logger.debug(this + " component.right= " + component.right);
//				logger.debug(this + " component.top= " + component.top);
//				logger.debug(this + " component.bottom= " + component.bottom);
				component = component.parent as UIComponent;
			} while ((component != null) && !(component is Skin));
			
			return new Point (hPad, vPad);
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function updateRenderer(oldRenderer:ISimpleCard):void
		{
			alpha = oldRenderer.alpha;
			updatePreview(oldRenderer);
			oldRenderer.invalidatePreview();
			
			// Drag-and-drop proxy doesn't behave properly without setting these measured values
			measuredWidth = oldRenderer.width;
			measuredHeight = oldRenderer.height;
//			logger.debug("SimpleCard.updateRenderer: this= " + this + " oldRenderer= " + oldRenderer);
		}
		
		/**
		 *  Update the preview for this card.
		 * 
		 *  @param oldRenderer The previous simpleCard renderer to use
		 *  as the basis for the updated preview
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function updatePreview(oldRenderer:ISimpleCard):void
		{
			preview = oldRenderer.previewComponent as Preview;
			if (!preview)
				return;
			enableThumbnailGeneration = true;
			invalidatePreview(true);
		}
		
		/**
		 *  Indicates whether generation of the thumbnail image should be
		 *  delayed or not.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get delayThumbnailGeneration():Boolean
		{
			return _delayThumbnailGeneration;
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set delayThumbnailGeneration(value:Boolean):void
		{
			_delayThumbnailGeneration = value;
		}
		
		/**
		 *  Indicates whether generation of the thumbnail image is enabled or not.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get enableThumbnailGeneration():Boolean
		{
			return _enableThumbnailGeneration;
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set enableThumbnailGeneration(value:Boolean):void
		{
			if (_enableThumbnailGeneration == value)
				return;
			
			_enableThumbnailGeneration = value;
			if (value) {
				generateThumbnail();  // using this (as opposed to invalidatePreview) gives better control over the order of generation
//				invalidatePreview(false, true);
			}
		}
		
		/**
		 *  Returns the file extension for the attachment.
		 *  If the attachment is a folder, this returns null.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function extensionFromAttachment():String {
			if (attachment.isFolder)
				return null;
			
			return FileUtil.getExtension(attachment.fileName);
		}
		
		/**
		 *  Returns the MIME type for the attachment.
		 *  If the attachment is a folder, this returns null.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function mimeTypeFromAttachment():String
		{
			if (!attachment)
				return null;
			
			return attachment.isFolder ? null : attachment.mimeType;
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function set doubleClickEnabled(value:Boolean):void
		{
			if (preview)
				preview.doubleClickEnabled = value;
			super.doubleClickEnabled = value;	
		}
		
//		leave for debug purposes
//		override public function set alpha(value:Number):void
//		{
//			super.alpha = value;
//			logger.debug(this + " alpha: " + alpha);
//		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get downSampleOK():Boolean
		{
			return previewMultiPageModel ? false : _downSampleOK;
		}

		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set downSampleOK(value:Boolean):void
		{
			if (_downSampleOK == value)
				return;
			
			_downSampleOK = value;
		}

		/**
		 *  Indicates whether or not to generate a thumbnail image for the card 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get generateThumb():Boolean
		{
			return _generateThumb;
		}

		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set generateThumb(value:Boolean):void
		{
			if (_generateThumb == value)
				return;
			
			_generateThumb = value;
			if (value)
				invalidatePreview();
		}
		
//		leave for debug purposes
//		override public function setLayoutBoundsPosition(x:Number, y:Number, postLayoutTransform:Boolean=true):void
//		{
//			logger.debug("setLayoutBoundsPosition: " + this + " x,y= " + x + " , " + y);
//			super.setLayoutBoundsPosition(x, y, postLayoutTransform);
//		}
	}
}
