/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components
{
	import com.adobe.portfolio.api.ICollection;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.PropertyManager;
	import com.adobe.portfolio.managers.ResourceManager;
	import com.adobe.portfolio.properties.CommonProperties;
	import com.adobe.portfolio.properties.PropertyBinaryValue;
	import com.adobe.portfolio.properties.PropertyListValue;
	import com.adobe.portfolio.properties.PropertyValue;
	import com.adobe.portfolio.utils.LogHelper;
	import com.adobe.portfolio.utils.NavURLLoader;
	
	import flash.display.DisplayObject;
	import flash.display.GradientType;
	import flash.display.Loader;
	import flash.events.Event;
	import flash.events.FocusEvent;
	import flash.events.IOErrorEvent;
	import flash.events.MouseEvent;
	import flash.events.SecurityErrorEvent;
	import flash.net.URLRequest;
	
	import mx.binding.utils.BindingUtils;
	import mx.events.FlexEvent;
	import mx.events.ResizeEvent;
	import mx.graphics.BitmapFillMode;
	import mx.managers.IFocusManagerComponent;
	import mx.states.State;
	import mx.utils.ColorUtil;
	
	import spark.components.supportClasses.SkinnableComponent;
	import spark.filters.BlurFilter;

	/**
	 *  Choosing this state enables the solid skin part of the background skin
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	[SkinState("solid")]
	
	/**
	 *  Choosing this state enables the linear gradient skin part of the background skin
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	[SkinState("linear")]
	
	/**
	 *  Choosing this state enables the radial gradient skin part of the background skin
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	[SkinState("radial")]
	
	/**
	 *  The Background class extends SkinnableComponent and draws the background of a navigator display window.  
	 *  It provides the capability to draw a mix of solid color, gradients, and images.
	 *  The background properties are specified in the navigator's properties.xml file.
	 * 
  	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class Background extends SkinnableComponent implements IFocusManagerComponent
	{
		/**
		 *  The mime type of a jpeg file.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const MIME_TYPE_JPG:String = "image/jpeg";

		/**
		 *  The mime type of a gif file.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const MIME_TYPE_GIF:String = "image/gif";

		/**
		 *  The mime type of a png file.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const MIME_TYPE_PNG:String = "image/png";
		
		public static const IMPORTED_IMAGE_NAME:String = "user_image";
		
		private static const MIME_TYPE_SWF:String = "application/x-shockwave-flash";

		private static const MAXIMUM_VALID_COLOR:uint = 0xFFFFFF;
		private static const INVALID_COLOR:uint = MAXIMUM_VALID_COLOR + 1;
		
		private static const DEFAULT_GRADIENT_INTENSITY:Number = 70;
		
		private static const ALIGN_LEFT:String = "alignLeft";
		private static const ALIGN_RIGHT:String = "alignRight";
		private static const ALIGN_TOP:String = "alignTop";
		private static const ALIGN_BOTTOM:String = "alignBottom";
		private static const ALIGN_CENTER:String = "alignCenter";

		private static const STATE_SOLID:String = "solid";
		private static const STATE_LINEAR:String = "linear";
		private static const STATE_RADIAL:String = "radial";
		
		private static const MAX_BLUR_VALUE:Number = 75;
		
		private var imageChanged:Boolean = false;

		// navLoader gets the actual bytes from the host, but we still need imageLoader
		// for the contentType, unless there's a better way.
		private var navLoader:NavURLLoader;
		
		private var imageLoader:Loader;
		private var imageLoadComplete:Boolean = false;
		private var loadedImageWidth:Number = 0.0;
		private var loadedImageHeight:Number = 0.0;

		private var gradientIntensity:Number = DEFAULT_GRADIENT_INTENSITY;
		private var imageScale:String = CommonProperties.PROPERTY_IMAGE_SCALE_ACTUAL_SIZE;
		private var imagePosition:String = CommonProperties.PROPERTY_IMAGE_POSITION_MIDDLE_CENTER;

		private const collectionMgr:CollectionManager = CollectionManager.instance;
		
		// Variables used by the gradient parts of the background skin
		/**
		 *  Used by the background skin to determine the opacity of the gradient.
		 * 
		 *  @default 1.0 (fully opaque)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var gradientAlpha:Number = 1.0;
		
		/**
		 *  Used by the background skin to determine the primary color of the gradient.
		 * 
		 *  @default 0 (black)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var gradientBaseColor:uint = 0;
		
		/**
		 *  Used by the background skin to determine the secondary color of the gradient.
		 * 
		 *  @default 0 (black)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var gradientDeltaColor:uint = 0;
		
		// Variables used by the image part of the background skin
		/**
		 *  Used by the background skin to determine if the image should be displayed.
		 * 
		 *  @default false (image is not displayed)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var imageVisible:Boolean = false;
		
		/**
		 *  Used by the background skin to determine the bluriness of the image (simulates depth of field).
		 *  If non-null, this is an array of one containing a blur filter.
		 *  @see spark.filters.BlurFilter
		 * 
		 *  @default no blur effect
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var imageFilters:Array = null;
		
		/**
		 *  Used by the background skin to display the loaded image.
		 * 
		 *  @default no image source
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var imageSource:Object = null;
		
		/**
		 *  Used by the background skin to determine the opacity of the image.
		 * 
		 *  @default 1.0 (fully opaque)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var imageAlpha:Number = 1.0;
		
		/**
		 *  Used by the background skin to determine the x position of the image.
		 * 
		 *  @default 0.0
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var imageX:Number = 0.0;
		
		/**
		 *  Used by the background skin to determine the y position of the image.
		 * 
		 *  @default 0.0
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var imageY:Number = 0.0;
		
		/**
		 *  Used by the background skin to determine the x-direction scale of the image.
		 * 
		 *  @default 1 (actual size)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var imageScaleX:Number = 1.0;
		
		/**
		 *  Used by the background skin to determine the y-direction scale of the image.
		 * 
		 *  @default 1 (actual size)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var imageScaleY:Number = 1.0;
		
		/**
		 *  Used by the background skin to determine how the image is displayed.
		 *  @see mx.graphics.BitmapFillMode
		 * 
		 *  @default BitmapFillMode.CLIP
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var imageFillMode:String = BitmapFillMode.CLIP;

		
		/**
		 *  Constructor.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function Background()
		{
			super();

			// Create the Image Loader
			imageLoader = new Loader();
			navLoader = new NavURLLoader(null, true);
			
			// Setup the states for this instance. They should match the skin states.
			states = [new State({name: STATE_SOLID}), new State({name: STATE_LINEAR}), new State({name: STATE_RADIAL})];

			// Initialize it's state
			currentState = STATE_SOLID;

			// Wait until we're initialized before setting up some bindings/listening
			addEventListener(FlexEvent.INITIALIZE, onInitialize);
		}
		
		/**
		 *  Sets the state of the skin as well as the state of this instance.
		 *
		 *  @param stateName The name of the new view state.
		 *  @param playTransition If <code>true</code>, play
		 *  the appropriate transition when the view state changes.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function setCurrentState(stateName:String, playTransition:Boolean = true):void
		{
			super.setCurrentState(stateName, playTransition);
			
			// We used to call skin.currentState = stateName; here, but simply invalidating
			// plays nicer with the Flex model.
			invalidateSkinState();
		}
		
		/** Returns the current state of this instance, since its skins should have a matching
		 *  set of states.
		 */
		override protected function getCurrentSkinState():String
		{
			return currentState;
		}

		private function onInitialize(evt:Event):void
		{
			// Listen for the resize event
			addEventListener(ResizeEvent.RESIZE, resizeEventHandler);
			
			// Listen for the background color change event
			BindingUtils.bindSetter(paletteBackgroundColorChanged, collectionMgr, ["colorPalette", "background"]);
			
			// Listen for the background property change events
			BindingUtils.bindSetter(gradientTypeChanged, PropertyManager.instance, ["commonProperties", "backgroundGradientType", "name"]);
			BindingUtils.bindSetter(gradientIntensityChanged, PropertyManager.instance, ["commonProperties", "backgroundGradientIntensity", "numValue"]);
			BindingUtils.bindSetter(imageSourceItemChanged, PropertyManager.instance, ["commonProperties", "backgroundImageSourceList", "activeValue"]);
			BindingUtils.bindSetter(imageScaleChanged, PropertyManager.instance, ["commonProperties", "backgroundImageScale", "name"]);
			BindingUtils.bindSetter(imagePositionChanged, PropertyManager.instance, ["commonProperties", "backgroundImagePosition", "name"]);
			BindingUtils.bindSetter(imageOpacityChanged, PropertyManager.instance, ["commonProperties", "backgroundImageOpacity", "numValue"]);
			BindingUtils.bindSetter(imageBlurChanged, PropertyManager.instance, ["commonProperties", "backgroundImageBlur", "numValue"]);

			// Set the current property context to Portfolio Properties when we receive focus
			addEventListener(FocusEvent.FOCUS_IN, portfolioContextHandler, false, 0, true);
			addEventListener(MouseEvent.CLICK, mouseClickHandler, false, 0, true);
		}

		private function paletteBackgroundColorChanged(value:uint):void
		{
			updateGradientColors();
		}
		
		private function gradientTypeChanged(value:String):void
		{
			// Decide which background is shown
			switch (value)
			{
				case GradientType.LINEAR:
					currentState = STATE_LINEAR;
					break;
				
				case GradientType.RADIAL:
					currentState = STATE_RADIAL;
					break;
				
				default:
					currentState = STATE_SOLID;
					break;
			}
		}
		
		private function gradientIntensityChanged(value:Number):void
		{
			if (isNaN(value))
				return;
			
			gradientIntensity = value;
			updateGradientColors();
		}
		
		private var lastImageValue:PropertyValue;
		private function imageSourceItemChanged(value:PropertyValue):void {
			if (value == null) {
				return;
			}
			switch (value.name) {
				case CommonProperties.PROPERTY_FILE_SELECT:
					// selectAndLoadImage will likely change the activeValue of the source list,
					// but we're probably in the middle of responding to such a change from the host application,
					// so let that clear out before we go ahead.
					// Without this, the host application may not be ready to update its UI accordingly.
					callLater(selectAndLoadImage);
					break;
				
				case CommonProperties.PROPERTY_FILE_REMOVE:
					lastImageValue = value;
					imageSourceChanged(null);
					break;
				
				case CommonProperties.PROPERTY_FILE_PLACEHOLDER:
					// Don't do anything here and don't execute
					// the code below or else you will get an
					// infinite loop since setting an active
					// value triggers this event.
					return;
					
				default:
					if (value is PropertyBinaryValue) {
						lastImageValue = value;
						imageSourceChanged((value as PropertyBinaryValue).url);
					}
					return;
			}
		}

		private function selectAndLoadImage():void
		{
			var addImageCallback:Function = function (response:Object):void
			{
				var imageList:PropertyListValue;
				if (PropertyManager.instance.commonProperties != null) {
					imageList = PropertyManager.instance.commonProperties.backgroundImageSourceList;
				}
				if (imageList == null) {
					return;
				}
				
				if (!response.success) {
					// If we weren't successful, go back to the previously selected entry
					if (lastImageValue != null) {
						imageList.activeValue = lastImageValue;
					}
					return;
				}
				
				var propertyLabel:String = response.relativeURL;
				var lastSlashIndex:int = propertyLabel.lastIndexOf("/");
				if ((lastSlashIndex >= 0) && (lastSlashIndex < (propertyLabel.length))) {
					propertyLabel = propertyLabel.substr(lastSlashIndex+1);
				}
				var newActiveValue:PropertyValue = PropertyValue.createPropertyValue(
					<PropertyBinary name={IMPORTED_IMAGE_NAME} label={propertyLabel} url={response.relativeURL}/>,
					imageList);
				if (newActiveValue == null) {
					// If we weren't successful, go back to the previously selected entry
					LogHelper.logger.error("Couldn't create newActiveValue from "+response.relativeURL);
					if (lastImageValue != null) {
						imageList.activeValue = lastImageValue;
					}
					return;
				}
				
				// add image item
				var didAddNewValue:Boolean = false;
				var newSubvalues:Array = [];
				var oldImportedURL:String;
				for each (var subvalue:PropertyValue in imageList.subvalues) {
					// We only allow a single custom image to be listed, so strip out any existing
					if (subvalue.name == IMPORTED_IMAGE_NAME) {
						if (subvalue is PropertyBinaryValue) {
							// Dereference the old custom image so that it can be purged
							ResourceManager.instance.dereferenceResource((subvalue as PropertyBinaryValue).url);
							oldImportedURL = (subvalue as PropertyBinaryValue).url;
						}
						continue;
					}
					newSubvalues.push(subvalue);
					if (!didAddNewValue && (subvalue.name == "remove")) {
						// We want to add our new value after 'remove'
						didAddNewValue = true;
						newSubvalues.push(newActiveValue);
					}
				}
				if (!didAddNewValue) {
					// We must not have had a 'remove' item, so just add our new item to the end
					newSubvalues.push(newActiveValue);
				}
			
				// Reference the new URL so that it doesn't get purged
				ResourceManager.instance.referenceResource(response.relativeURL);
				
				// Reinitialize other image settings when the user sets a custom background
				var commonProperties:CommonProperties = PropertyManager.instance.commonProperties;
				if (commonProperties != null) {
					if (commonProperties.backgroundImageOpacity != null) {
						commonProperties.backgroundImageOpacity.numValue = 100;
					}
					if (commonProperties.backgroundImageBlur != null) {
						commonProperties.backgroundImageBlur.numValue = 0;
					}
					if (commonProperties.backgroundImageScaleList != null) {
						commonProperties.backgroundImageScaleList.setValueFromString(CommonProperties.PROPERTY_IMAGE_SCALE_ACTUAL_SIZE);
					}
					if (commonProperties.backgroundImagePositionList != null) {
						commonProperties.backgroundImagePositionList.setValueFromString(CommonProperties.PROPERTY_IMAGE_POSITION_MIDDLE_CENTER);
					}
				}
				
				// This will trigger another change to backgroundImageSourceList.activeValue,
				// which will then be handled by our binding to set the image bits.
				imageList.updateSubvalues(newSubvalues, newActiveValue);
				
				if (oldImportedURL != null) {
					// Delete the old custom image
					ResourceManager.instance.deleteResource(oldImportedURL);
				}
			};
			
			var hostCollection:ICollection = (CollectionManager.instance.host != null) ? CollectionManager.instance.host.collection : null;
			var titleString:String = (hostCollection == null) ? "BACKGROUND_SELECT_IMAGE_TITLE" : hostCollection.getLocalizedString("backgroundSelectImageTitle", "BACKGROUND_SELECT_IMAGE_TITLE");
			ResourceManager.instance.addResource(titleString, [ ResourceManager.instance.imageFilter ], "navigator/user_background/", addImageCallback);
		}

		private function imageSourceChanged(value:String):void
		{
			if (value == navLoader.url) {
				// The image hasn't actually changed, so just return
				return;
			}
			
			loadedImageWidth = 0;
			loadedImageHeight = 0;
			
			// De-reference the old image so that it can be purged
			ResourceManager.instance.dereferenceResource(navLoader.url);
			
			if (value == null || value == "")
			{
				imageLoader.unload();
				imageVisible = false;
				navLoader.url = value;
				return;
			}
			
			navLoader.addEventListener(Event.COMPLETE, navLoaderHandler);
			navLoader.addEventListener(IOErrorEvent.IO_ERROR, navLoaderHandler);
			navLoader.addEventListener(SecurityErrorEvent.SECURITY_ERROR, navLoaderHandler);
			
			// Reference the new image so that it doesn't get purged
			ResourceManager.instance.referenceResource(value);
			
			navLoader.url = value;
		}
		
		private function imageOpacityChanged(value:Number):void
		{
			if (isNaN(value))
				return;
				
			// Update the opacity of the image
			imageAlpha = value / 100;
		}
		
		private function imageBlurChanged(value:Number):void
		{
			if (isNaN(value))
				return;
				
			imageFilters = null;
			if (value > 0)
			{
				// The property value is actually a percentage, so adjust it accordingly
				var blurValue:Number = MAX_BLUR_VALUE * value / 100;
				var blurFilter:BlurFilter = new BlurFilter(blurValue, blurValue);
				imageFilters = [blurFilter];
			}
		}
		
		private function imageScaleChanged(value:String):void
		{
			if (value == null)
				return;
			
			imageScale = value;
			updateImage();
		}

		private function imagePositionChanged(value:String):void
		{
			if (value == null)
				return;
			
			imagePosition = value;
			updateImage();
		}

		private function navLoaderHandler(event:Event):void
		{
			if (event.type == Event.COMPLETE) {
				imageLoader.contentLoaderInfo.addEventListener(Event.COMPLETE, loadImageCompleteHandler);
				imageLoader.contentLoaderInfo.addEventListener(IOErrorEvent.IO_ERROR, loadImageIOErrorHandler);
				imageLoader.contentLoaderInfo.addEventListener(SecurityErrorEvent.SECURITY_ERROR, loadImageSecurityErrorHandler);
				
				imageLoader.loadBytes(navLoader.dataBytes);
			} else {
				LogHelper.logger.error("navLoaderHandler. " + event.toString());
				ResourceManager.instance.showMissingResourcePopup(navLoader.url, parentApplication as DisplayObject);
				imageLoader.unload();
				imageVisible = false;
			}
			
			navLoader.removeEventListener(Event.COMPLETE, navLoaderHandler);
			navLoader.removeEventListener(IOErrorEvent.IO_ERROR, navLoaderHandler);
			navLoader.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, navLoaderHandler);
		}
				
		private function loadImageCompleteHandler(event:Event):void
		{
			removeImageLoaderListeners();
			
			// This prevents dealing with corrupted images
			if (imageLoader.width == 0 || imageLoader.height == 0)
			{
				imageLoader.unloadAndStop();
				return;
			}
			
			// Only allow images (swfs masquerading as images are a security risk)
			switch (imageLoader.contentLoaderInfo.contentType)
			{
				case MIME_TYPE_JPG:
				case MIME_TYPE_GIF:
				case MIME_TYPE_PNG:
					break;
				
				default:
					imageLoader.unloadAndStop();
					return;
			}
			
			loadedImageWidth = imageLoader.contentLoaderInfo.width;
			loadedImageHeight = imageLoader.contentLoaderInfo.height;
			imageLoadComplete = true;
			imageChanged = true;
			
			updateImage();
		}
		
		private function loadImageIOErrorHandler(event:IOErrorEvent):void
		{
			removeImageLoaderListeners();
			LogHelper.logger.error("loadImageIOErrorHandler. " + event.text);
		}
		
		private function loadImageSecurityErrorHandler(event:SecurityErrorEvent):void
		{
			removeImageLoaderListeners();
			LogHelper.logger.error("loadImageSecurityErrorHandler. " + event.text);
		}
		
		private function removeImageLoaderListeners():void
		{
			imageLoader.contentLoaderInfo.removeEventListener(Event.COMPLETE, loadImageCompleteHandler);
			imageLoader.contentLoaderInfo.removeEventListener(IOErrorEvent.IO_ERROR, loadImageIOErrorHandler);
			imageLoader.contentLoaderInfo.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, loadImageSecurityErrorHandler);
		}
		
		
		private function resizeEventHandler(event:ResizeEvent):void
		{
			updateImage();
		}
		
		private function portfolioContextHandler(event:Event):void
		{
			if ((PropertyManager.instance.commonProperties != null) &&
				collectionMgr.apiIsSupported(CollectionManager.API_VERSION_10))
			{
				PropertyManager.instance.commonProperties.setCurrentContextByName(CommonProperties.CONTEXT_PORTFOLIO, this);
			}
		}
		
		private function mouseClickHandler(event:MouseEvent):void
		{
			// Clear the selection if the user explicitly clicks on the background
			collectionMgr.setSelectedItems([]);
		}
		
		private function updateGradientColors():void
		{
			// All backgrounds (solid, linear/radial gradients) are based on the background theme color
			// If there is no valid background theme color, no changes are made
			var baseColor:uint = collectionMgr.colorPalette.background;
			if (baseColor > MAXIMUM_VALID_COLOR)
				return;
			
			// Gradients require a second color, which is a delta intensity of the base color
			var brightnessChange:Number = -gradientIntensity;
			brightnessChange = Math.min(brightnessChange, 100);
			brightnessChange = Math.max(brightnessChange, -100);
			var brighterColor:uint = ColorUtil.adjustBrightness2(baseColor, brightnessChange);
			
			// Update the colors of the skin elements which have colors
			gradientBaseColor = baseColor;
			gradientDeltaColor = brighterColor;
		}
		
		private function updateImage():void
		{
			// Assume that image should not be displayed until we confirm
			// that it is valid and should be (at end of this function).
			imageVisible = false;

			// If there is no image, nothing to do
			if (!imageLoadComplete)
				return;
			
			// If the image has no dimensions, nothing to do
			if (loadedImageWidth == 0 || loadedImageHeight == 0)
				return;
			
			var imageWidth:Number = width;
			var imageHeight:Number = height;
			var bmRatio:Number = loadedImageWidth / loadedImageHeight;
			var doRatio:Number = unscaledWidth / unscaledHeight;
			
			imageScaleX = 1.0;
			imageScaleY = 1.0;
			imageFillMode = BitmapFillMode.CLIP;
			
			// Figure out how the image should be scaled (and if it should be tiled)
			switch (imageScale)
			{
				case CommonProperties.PROPERTY_IMAGE_SCALE_ACTUAL_SIZE:
					imageWidth = loadedImageWidth;
					imageHeight = loadedImageHeight;
					// No scaling (default set above)
					// No repeating (default set above)
					break;
				
				case CommonProperties.PROPERTY_IMAGE_SCALE_NO_BORDER:
					imageWidth = bmRatio < doRatio ? unscaledWidth : unscaledHeight * (loadedImageWidth / loadedImageHeight);
					imageHeight = bmRatio >= doRatio ? unscaledHeight : unscaledWidth * (loadedImageHeight / loadedImageWidth);
					imageScaleX = imageWidth / loadedImageWidth;
					imageScaleY = imageHeight / loadedImageHeight;
					// No repeating (default set above)
					break;
				
				case CommonProperties.PROPERTY_IMAGE_SCALE_SHOW_ALL:
					imageWidth = bmRatio >= doRatio ? unscaledWidth : unscaledHeight * (loadedImageWidth / loadedImageHeight);
					imageHeight = bmRatio < doRatio ? unscaledHeight : unscaledWidth * (loadedImageHeight / loadedImageWidth);
					imageScaleX = imageWidth / loadedImageWidth;
					imageScaleY = imageHeight / loadedImageHeight;
					// No repeating (default set above)
					break;
				
				case CommonProperties.PROPERTY_IMAGE_SCALE_ACTUAL_SIZE_TILED:
					imageWidth = roundToOddNumberOfTilesLength(unscaledWidth, loadedImageWidth);
					imageHeight = roundToOddNumberOfTilesLength(unscaledHeight, loadedImageHeight);
					// No scaling (default set above)
					imageFillMode = BitmapFillMode.REPEAT;
					break;
				
				default:
					// Don't display the image if the scaling is invalid
					return;
			}
			
			// Figure out where the image should be positioned
			var xAlign:String;
			var yAlign:String;
			switch (imagePosition)
			{
				case CommonProperties.PROPERTY_IMAGE_POSITION_TOP_LEFT:
					xAlign = ALIGN_LEFT;
					yAlign = ALIGN_TOP;
					break;

				case CommonProperties.PROPERTY_IMAGE_POSITION_TOP_CENTER:
					xAlign = ALIGN_CENTER;
					yAlign = ALIGN_TOP;
					break;
				
				case CommonProperties.PROPERTY_IMAGE_POSITION_TOP_RIGHT:
					xAlign = ALIGN_RIGHT;
					yAlign = ALIGN_TOP;
					break;
				
				case CommonProperties.PROPERTY_IMAGE_POSITION_MIDDLE_LEFT:
					xAlign = ALIGN_LEFT;
					yAlign = ALIGN_CENTER;
					break;
				
				case CommonProperties.PROPERTY_IMAGE_POSITION_MIDDLE_CENTER:
					xAlign = ALIGN_CENTER;
					yAlign = ALIGN_CENTER;
					break;

				case CommonProperties.PROPERTY_IMAGE_POSITION_MIDDLE_RIGHT:
					xAlign = ALIGN_RIGHT;
					yAlign = ALIGN_CENTER;
					break;
				
				case CommonProperties.PROPERTY_IMAGE_POSITION_BOTTOM_LEFT:
					xAlign = ALIGN_LEFT;
					yAlign = ALIGN_BOTTOM;
					break;
				
				case CommonProperties.PROPERTY_IMAGE_POSITION_BOTTOM_CENTER:
					xAlign = ALIGN_CENTER;
					yAlign = ALIGN_BOTTOM;
					break;
				
				case CommonProperties.PROPERTY_IMAGE_POSITION_BOTTOM_RIGHT:
					xAlign = ALIGN_RIGHT;
					yAlign = ALIGN_BOTTOM;
					break;
				
				default:
					// Don't display the image if the positioning is invalid
					return;
			}
			
			switch(xAlign)
			{
				case ALIGN_LEFT:
					imageX = 0;
					break;
				
				case ALIGN_RIGHT:
					imageX = unscaledWidth - imageWidth;
					break;
				
				case ALIGN_CENTER:
				default:
					imageX = (unscaledWidth - imageWidth) / 2;
					break;
			}
			
			switch(yAlign)
			{
				case ALIGN_TOP:
					imageY = 0;
					break;
				
				case ALIGN_BOTTOM:
					imageY = unscaledHeight - imageHeight;
					break;
				
				case ALIGN_CENTER:
				default:
					imageY = (unscaledHeight - imageHeight) / 2;
					break;
			}
			
			if (imageChanged)
			{
				imageChanged = false;
				imageSource = imageLoader.contentLoaderInfo.content;
			}
			
			// Now that everything has been set, make it visible
			imageVisible = true;
			CONFIG::performance {
				collectionMgr.logPerformance("Background visible");
			}
		}
		
		// This is useful for determining how many tiled images fit across and down a window for 
		// the purpose of offsetting them so the central image is in the center of the window.
		private function roundToOddNumberOfTilesLength(displayObjectSize:Number, imageSize:Number):Number
		{
			var tilesPer:uint = displayObjectSize / imageSize;
			var oddNumTiles:Boolean = (tilesPer & 1) == 1;
			// If exact fit and odd number then return it as is
			if ((tilesPer * imageSize == displayObjectSize) && oddNumTiles)
				return displayObjectSize;
			tilesPer += oddNumTiles ? 2 : 1;
			return tilesPer * imageSize;
		}
	}
}
