/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components
{
	import com.adobe.portfolio.cards.ISimpleCard;
	import com.adobe.portfolio.events.MarqueeSelectionEvent;
	
	import flash.display.DisplayObject;
	import flash.display.DisplayObjectContainer;
	import flash.events.MouseEvent;
	import flash.geom.Point;
	
	import mx.states.State;
	
	import spark.components.Button;
	
	/**
	 *  Dispatched when a marquee selection is initiated.
	 *  
	 *  @eventType com.adobe.portfolio.events.MarqueeSelectionEvent.BEGIN
	 */
	[Event(name="marqueeSelectionBegin", type="com.adobe.portfolio.events.MarqueeSelectionEvent")]

	/**
	 *  Dispatched when a marquee selection in progress is moved.
	 *  
	 *  @eventType com.adobe.portfolio.events.MarqueeSelectionEvent.MOVE
	 */
	[Event(name="marqueeSelectionMove", type="com.adobe.portfolio.events.MarqueeSelectionEvent")]
	
	/**
	 *  Dispatched when a marquee selection is completed.
	 *  
	 *  @eventType com.adobe.portfolio.events.MarqueeSelectionEvent.END
	 */
	[Event(name="marqueeSelectionEnd", type="com.adobe.portfolio.events.MarqueeSelectionEvent")]
	
	/**
	 *  Dispatched when a marquee selection is initiated.
	 *  
	 *  @eventType com.adobe.portfolio.events.MarqueeSelectionEvent.CANCELLED
	 */
	[Event(name="marqueeSelectionCancelled", type="com.adobe.portfolio.events.MarqueeSelectionEvent")]
	
	/**
 	 * The MarqueeSelection class extends NavSkinnableComponent to provide marquee selection support.
	 * 
  	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class MarqueeSelection extends NavSkinnableComponent
	{
		private var mouseDownInBackground:Boolean = false;
		private var mouseDownActionPending:Boolean = false;
		private var xMouseDown:Number;
		private var yMouseDown:Number;
		
		/**
		 * Constructor.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function MarqueeSelection()
		{
			super();
			visible = false;
		}
		
		override protected function initStates():void
		{
			states = [
				new State({name: STATE_NORMAL})
			];
			currentState = STATE_NORMAL;
		}
		
		override protected function attachSkin():void
		{
			super.attachSkin();
			var root:DisplayObject = systemManager.getSandboxRoot();
			if (!root)
				return;
			addListener(root, MouseEvent.MOUSE_DOWN, onMouseDown);
			addListener(root, MouseEvent.MOUSE_UP, onMouseUp);
		}
		
		protected function onMouseDown(event:MouseEvent):void
		{
			if (!enabled)
				return;
			
			if (targetIsBackground(event)) {
				mouseDownInBackground = true;
				mouseDownActionPending = true;
				
				xMouseDown = event.stageX;
				yMouseDown = event.stageY;
				var localPos:Point = globalToParentLocal(event.stageX, event.stageY);
				x = localPos.x;
				y = localPos.y;
				width = 0;
				height = 0;
				var root:DisplayObject = systemManager.getSandboxRoot();
				if (root)
					root.addEventListener(MouseEvent.MOUSE_MOVE, onMouseMove, false, 0, true);
				sendMarqueeSelectionEvent(MarqueeSelectionEvent.BEGIN, event);
			}
		}
		
		protected function onMouseMove(event:MouseEvent):void
		{
			// Make sure the marquee is visible (and update its size)
			visible = true;
			
			var localPos:Point = globalToParentLocal(event.stageX, event.stageY);
			var w:Number = event.stageX - xMouseDown;
			if (w < 0) {
				w = -w;
				x = localPos.x;
			}
			var h:Number = event.stageY - yMouseDown;
			if (h < 0) {
				h = -h;
				y = localPos.y;
			}
			width = w;
			height = h;
			mouseDownActionPending = false;
			
			// Send a marquee move event
			sendMarqueeSelectionEvent(MarqueeSelectionEvent.MOVE, event);
		}
		
		protected function onMouseUp(event:MouseEvent):void
		{
			if (!enabled)
				return;
			
			if (!mouseDownInBackground)
				return;
			
			var root:DisplayObject = systemManager.getSandboxRoot();
			if (root)
				root.removeEventListener(MouseEvent.MOUSE_MOVE, onMouseMove);

			// Set the marquee back to invisible 
			visible = false;
			
			// Clear mouse down state
			mouseDownInBackground = false;

			// If mouse up immediately follows mouse down (no mouse moving),
			// then the marquee selection is considered to be cancelled
			// otherwise it is considered to be complete. A cancelled
			// marquee selection is just a single click in the background.
			sendMarqueeSelectionEvent(mouseDownActionPending ? MarqueeSelectionEvent.CANCELLED : MarqueeSelectionEvent.END, event);
			
			// Clear mouse down action pending
			mouseDownActionPending = false;
		}
		
		protected function targetIsBackground(event:MouseEvent):Boolean
		{
			if (!event)
				return false;
			
			var eventTarget:DisplayObject = event.target as DisplayObject;
			if (!eventTarget)
				return false;
			
			if (eventTarget is Button)
				return false;
			
			var parent:DisplayObjectContainer = eventTarget.parent;
			while (parent) {
				if (parent is ISimpleCard || parent is Button)
					return false;
				parent = parent.parent;
			}
				
			return true;
		}
		
		protected function sendMarqueeSelectionEvent(type:String, mouseEvent:MouseEvent):void
		{
			var marqueeEvent:MarqueeSelectionEvent = new MarqueeSelectionEvent(type, false, false);
			
			var pos:Point = parentLocalToGlobal();
			marqueeEvent.x = pos.x;
			marqueeEvent.y = pos.y;
			marqueeEvent.width = width;
			marqueeEvent.height = height;
			
			// Pass on the mouse keyboard modifiers (to support marquee selection add and subtract)
			marqueeEvent.ctrlKey = mouseEvent.ctrlKey;
			marqueeEvent.shiftKey = mouseEvent.shiftKey;
			
			dispatchEvent(marqueeEvent);
		}
		
		private function globalToParentLocal(x:Number, y:Number):Point {
			return parent.globalToLocal(new Point(x,y));
		}
		
		private function parentLocalToGlobal():Point {
			return parent.localToGlobal(new Point(x,y));
		}
	}
}