/*************************************************************************
 *
 * ADOBE SYSTEMS INCORPORATED
 *  Copyright 2010 Adobe Systems Incorporated
 *  All Rights Reserved.
 *
 * NOTICE:  Adobe permits you to use, modify, and distribute this 
 * file in accordance with the terms of the Adobe license agreement 
 * accompanying it.  If you have received this file from a source 
 * other than Adobe, then your use, modification, or distribution 
 * of it requires the prior written permission of Adobe.
 **************************************************************************/

package com.adobe.portfolio.components
{
	import com.adobe.portfolio.api.ICollection;
	import com.adobe.portfolio.api.INavigatorHost;
	import com.adobe.portfolio.components.supportClasses.CanvasItem;
	import com.adobe.portfolio.components.supportClasses.CanvasItemData;
	import com.adobe.portfolio.components.supportClasses.CanvasList;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.properties.CommonProperties;
	import com.adobe.portfolio.properties.PropertyContext;
	
	import flash.events.Event;
	import flash.events.MouseEvent;
	
	import mx.core.IButton;
	import mx.utils.UIDUtil;
	
	import spark.components.CheckBox;
	import spark.events.RendererExistenceEvent;
	
	[SkinState(name="closed")]
	
	/**
	 * The Welcome class extends CanvasList to add support for user-customizable swf items, a close state,
	 * and Don't-show-again behavior.
	 */ 
	public class Welcome extends CanvasList
	{
		/**
		 * A string literal representing the closed state of this component's skin.
		 */
		public static const SKIN_STATE_CLOSED:String = "closed";
		
		/**
		 * Default constructor
		 */
		public function Welcome()
		{
			super();
			
			// Initialize dataURL to the standard welcome location
			dataURL = "navigator/welcome.xml";
			swfClass = "WelcomeBackground";
			autoGenerateProperties = true;
			autoGeneratedLabel = "WELCOME_PROPERTIES";
			autoGeneratedLabelKey = "welcomeProperties";
			backgroundColor = collectionManager.colorPalette.background;
			backgroundOpacity = 90;
		}
		
		[SkinPart(required="false")]
		/**
		 * A button used to dismiss the Welcome component.
		 */
		public var closeButton:IButton;
		
		[SkinPart(required="false")]
		/**
		 * A button used to display the Welcome component.
		 */
		public var openButton:IButton;
		
		[SkinPart(required="false")]
		/**
		 * A checkbox used to indicate that the Welcome component should not be
		 * displayed by default.
		 */
		public var hideCheckbox:CheckBox;
		
		private var _closed:Boolean = false;
		[Bindable]
		/**
		 * True if this component is closed; false otherwise.
		 */
		public function get closed():Boolean
		{
			return _closed;
		}
		public function set closed(value:Boolean):void
		{
			_closed = value;
			invalidateSkinState();
			if ((dataGroup != null) && (dataProvider != null)) {
				// Stop all swfs on a close, or play on an open
				var dataLength:int = dataProvider.length;
				for (var i:int = 0; i < dataLength; ++i) {
					var canvasData:CanvasItemData = dataProvider.getItemAt(i) as CanvasItemData;
					if ((canvasData == null) || (canvasData.type != CanvasItemData.SWF_TYPE)) {
						continue;
					}
					var canvasItem:CanvasItem = dataGroup.getElementAt(i) as CanvasItem;
					if (canvasItem == null) {
						continue;
					}
					canvasItem.playSWF = !value;
				}
			}
		}
		
		/**
		 * Event listener for mouse clicks on the close button.
		 */
		protected function onCloseClick(event:Event):void
		{
			closed = true;
		}
		
		/**
		 * Event listener for mouse clicks on the open button.
		 */
		protected function onOpenClick(event:Event):void
		{
			closed = false;
		}
		
		/**
		 * Event listener for changes to the hide checkbox.
		 */
		protected function onHideChanged(event:Event):void
		{
			var hostCollection:ICollection = (collectionManager.host == null) ? null : collectionManager.host.collection;
			if (hostCollection != null) {
				hostCollection.setSessionData("hideWelcome", hideCheckbox.selected); 
			}
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function onRendererChange(event:RendererExistenceEvent):void
		{
			super.onRendererChange(event);
			if (event.type != RendererExistenceEvent.RENDERER_ADD) {
				return;
			}
			var item:CanvasItem = event.renderer as CanvasItem;
			if (item == null) {
				return;
			}
			item.playSWF = !closed;
		}

		/**
		 * @inheritDoc
		 */
		override protected function getCurrentSkinState():String
		{
			var returnState:String = super.getCurrentSkinState();
			if (closed && (returnState != CanvasList.SKIN_STATE_DISABLED)) {
				returnState = SKIN_STATE_CLOSED;
			}
			return returnState;
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function partAdded(partName:String, instance:Object):void
		{
			switch (instance) {
				case closeButton:
					closeButton.addEventListener(MouseEvent.CLICK, onCloseClick, false, 0, true);
					break;
				case openButton:
					openButton.addEventListener(MouseEvent.CLICK, onOpenClick, false, 0, true);
					break;
				case hideCheckbox:
					hideCheckbox.selected = hideWelcome;
					hideCheckbox.addEventListener(Event.CHANGE, onHideChanged, false, 0, true);
					break;
			}
			super.partAdded(partName, instance);
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function partRemoved(partName:String, instance:Object):void
		{
			switch (instance) {
				case closeButton:
					closeButton.removeEventListener(MouseEvent.CLICK, onCloseClick, false);
					break;
				case openButton:
					openButton.removeEventListener(MouseEvent.CLICK, onOpenClick, false);
					break;
				case hideCheckbox:
					hideCheckbox.removeEventListener(Event.CHANGE, onHideChanged, false);
					break;
			}
			super.partRemoved(partName, instance);
		}

		/**
		 * @inheritDoc
		 */
		override protected function generatePropertyContext():PropertyContext
		{
			var context:PropertyContext = new PropertyContext();
			context.loadFromXML(
				<PropertyContext labelKey={autoGeneratedLabelKey} label={autoGeneratedLabel} uid="com.adobe.navigator.welcome" global="false">
					<PropertyColor name="backgroundColor" labelKey="canvasBackgroundColor" label="CANVAS_BACKGROUND_COLOR" value={backgroundColor}/>
					<PropertyNumber name="backgroundOpacity" labelKey="canvasBackgroundOpacity" label="CANVAS_BACKGROUND_OPACITY" minValue="0" value={backgroundOpacity} maxValue="100"/>
					<PropertyList name="gradientType" labelKey="backgroundGradientType" label="BACKGROUND_GRADIENT_TYPE" active={backgroundGradientType}>
						<PropertyValue name="none" labelKey="backgroundGradientNone" label="BACKGROUND_GRADIENT_NONE"/>
						<PropertyValue name="linear" labelKey="backgroundGradientLinear" label="BACKGROUND_GRADIENT_LINEAR"/>
						<PropertyValue name="radial" labelKey="backgroundGradientRadial" label="BACKGROUND_GRADIENT_RADIAL"/>
					</PropertyList>
					<PropertyNumber name="gradientIntensity" labelKey="backgroundGradientIntensity" label="BACKGROUND_GRADIENT_INTENSITY" minValue="-100" value={backgroundGradientIntensity} maxValue="100"/>
					<PropertyAction name={CommonProperties.PROPERTY_ADD_TEXT} labelKey="addText" label="ADD_TEXT_BUTTON"/>
					<PropertyAction name={CommonProperties.PROPERTY_ADD_IMAGE} labelKey="addImage" label="ADD_IMAGE_BUTTON"/>
					<PropertyAction name={CommonProperties.PROPERTY_ADD_SWF} labelKey="addSwf" label="ADD_SWF_BUTTON"/>
				</PropertyContext>
			);
			return context;
		}

		/**
		 * @inheritDoc
		 */
		override public function addListenAndWatch():void
		{
			super.addListenAndWatch();
			refreshHideWelcome();
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function onNavHostEvent(event:CollectionManagerEvent):void
		{
			super.onNavHostEvent(event);
			if (event.kind == CollectionManagerEvent.HOST_CHANGED_KIND) {
				refreshHideWelcome();
			}
		}
		
		private var hideWelcome:Boolean = false;
		
		/**
		 * Updates the hideCheckbox and closed property based on existing
		 * cached session data.
		 */
		protected function refreshHideWelcome():void
		{
			if ((collectionManager.host == null) ||
				(collectionManager.host.collection == null))
			{
				return;
			}
			collectionManager.host.collection.getSessionData("hideWelcome",
				function (key:String, sessionData:*):void
				{
					if (sessionData != null) {
						closed = sessionData;
						hideWelcome = closed;
						if (hideCheckbox != null) {
							hideCheckbox.selected = closed;
						}
					} else {
						closed = false;
					}
				});
		}
	}
}