/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.details
{
	import com.adobe.portfolio.api.FieldType;
	import com.adobe.portfolio.api.IField;
	import com.adobe.portfolio.components.NavSkinnableComponent;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.utils.AttachmentCollection;
	import com.adobe.portfolio.utils.LogHelper;
	import com.adobe.portfolio.utils.SortUtil;
	
	import flash.events.MouseEvent;
	import flash.events.TimerEvent;
	import flash.geom.Point;
	import flash.system.Capabilities;
	import flash.utils.Dictionary;
	import flash.utils.Timer;
	
	import mx.binding.utils.BindingUtils;
	import mx.collections.ArrayList;
	import mx.collections.Sort;
	import mx.collections.SortField;
	import mx.containers.Canvas;
	import mx.core.UIFTETextField;
	import mx.events.CollectionEvent;
	import mx.events.CollectionEventKind;
	import mx.events.DataGridEvent;
	import mx.events.PropertyChangeEvent;
	import mx.events.ResizeEvent;
	
	/**
	 *  The Details class extends NavSkinnableComponent to implement the Details navigator. 
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */	
	public class Details extends NavSkinnableComponent
	{
		/**
		 *  The maximum number of data items that are checked to determine the optimum (default) width of a field/column.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const MAXIMUM_DATA_ITEMS_TO_CHECK:Number = 10;
		
		/**
		 *  The width, in pixels, of the icon which is displayed in each row of the details layout.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const PREVIEW_ICON_WIDTH:Number = 75;
		
		/**
		 *  The height, in pixels, of the icon which is displayed in each row of the details layout.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const PREVIEW_ICON_HEIGHT:Number = 50;
		
		/**
		 *  The navigator's CollectionManager instance.
		 *  All host interaction should go through this instance, rather than the INavigatorHost directly.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public const collectionMgr:CollectionManager = CollectionManager.instance;
		
		private static const MAXIMUM_COLUMN_WIDTH:Number = 300;
		private static const MINIMUM_COLUMN_WIDTH:Number = 70;
		
		private static const DETAILS_BACKGROUND_COLOR:uint = 0x323232;
		private static const DETAILS_CARD_BACKGROUND_COLOR:uint = 0xF7F7F7;
		private static const DETAILS_CARD_BORDER_COLOR:uint = 0xE4E0CE;
		private static const DETAILS_PRIMARY_TEXT_COLOR:uint = 0x333333;
		private static const DETAILS_SECONDARY_TEXT_COLOR:uint = 0x000000;
		
		/**
		 *  The header portion of the details card list.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="true")]
		public var detailsHeader:DetailsHeader;
		
		/**
		 *  The body portion of the details card list.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="true")]
		public var detailsCards:DetailsCardList;
		
		/**
		 *  A container needed to add an IUITextField for measuring text. When provided, measureField
		 *  will be created and added to its rawChildren.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="true")]
		public var measureCanvas:Canvas;
		
		/**
		 * A IUITextField used to measure text. It must be added to the stage to work properly,
		 * so it's always added to measureCanvas.
		 */
		private var measureField:UIFTETextField;
		
		/**
		 *  A common horizontalScrollPosition used by both the header and card list.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public var horizontalScrollPosition:Number = 0;

		private var resizeFieldLeft:Number = 0;
		private var resizeDetailsField:DetailsField = null;
		
		private var iFields:Array = null;
		
		private var _detailsFields:ArrayList = new ArrayList();
		/**
		 *  An array of DetailFields that corresponds to each of the fields in the IAttachment.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public function get detailsFields():ArrayList
		{
			return _detailsFields;
		}
		private function set detailsFields(value:ArrayList):void
		{
			_detailsFields = value;
		}
		
		private var detailsFieldsMap:Dictionary = new Dictionary();
		private var updateColumnSizes:Boolean = false;
		private var updateColumnSizesRetryTimer:Timer = null;
		private var updateLayoutsOfHeaderAndCardsNeeded:Boolean = false;
		
		private var pendingSortOrderChange:Boolean = false;
		private var pendingFieldOrderChange:Boolean = false;

		/**
		 *  Constructor.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function Details()
		{
			super();
			initDetailsColorPalette();
		}
		
		private function initDetailsColorPalette():void
		{
			var detailsColorPalette:Object = {
				isHighContrast: false,
				background:		DETAILS_BACKGROUND_COLOR,
				cardBackground:	DETAILS_CARD_BACKGROUND_COLOR,
				cardBorder:		DETAILS_CARD_BORDER_COLOR,
				primaryText:	DETAILS_PRIMARY_TEXT_COLOR,
				secondardText:	DETAILS_SECONDARY_TEXT_COLOR
			};
			collectionMgr.fixedNavigatorColorPalette = detailsColorPalette;
		}
		
		/**
		 *  Adds listeners and watchers that are automatically unwatched and unlistened.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		override public function addListenAndWatch():void
		{
			super.addListenAndWatch();
			addListener(collectionMgr, CollectionManagerEvent.COLLECTION_CHANGE, onHostEvent);
			addWatcher(BindingUtils.bindSetter(fieldsChanged, collectionMgr, "fields"));
			addWatcher(BindingUtils.bindSetter(sortChanged, collectionMgr, "sort"));
			addWatcher(BindingUtils.bindSetter(authoringModeChanged, collectionMgr, "authoringMode"));
		}
		
		override protected function measure():void
		{
			super.measure();
			if (updateColumnSizes)
				updateDefaultFieldColumnWidths();
		}
		
		override protected function updateDisplayList(unscaledWidth:Number, unscaledHeight:Number):void
		{
			if (updateLayoutsOfHeaderAndCardsNeeded) {
				updateLayoutsOfHeaderAndCards();
				updateLayoutsOfHeaderAndCardsNeeded = false;
			}
			super.updateDisplayList(unscaledWidth, unscaledHeight);
		}
		
		/** Called when a skin has added a skin part for this instance.
		 */
		override protected function partAdded(partName:String, instance:Object):void
		{
			super.partAdded(partName, instance);
			switch (instance) {
				case detailsHeader:
					detailsHeader.details = this;
					detailsHeader.fields = detailsFields;
					addListener(detailsHeader, DataGridEvent.HEADER_RELEASE, onHeaderClicked);
					addListener(detailsHeader, DataGridEvent.COLUMN_STRETCH, onColumnResizeStart);
					addListener(detailsHeader, ResizeEvent.RESIZE, onHeaderResize);
					break;
				
				case detailsCards:
					detailsCards.details = this;
					break;
				
				case measureCanvas:
					measureField = new UIFTETextField();
					measureCanvas.rawChildren.addChild(measureField);
					break;
				
				default:
					break;
			}
		}
		
		/** An event listener to respond to events from the collectionManager.
		 */
		protected function onHostEvent(event:CollectionManagerEvent):void
		{
			var collection:AttachmentCollection;
			
			switch (event.kind) {
				case CollectionManagerEvent.ITEMS_WILL_CHANGE_KIND:
					collection = event.oldVal as AttachmentCollection;
					if (collection)
						collection.removeEventListener(CollectionEvent.COLLECTION_CHANGE, onCollectionChange);
					break;
				
				case CollectionManagerEvent.ITEMS_DID_CHANGE_KIND:
					collection = event.newVal as AttachmentCollection;
					if (collection)
						collection.addEventListener(CollectionEvent.COLLECTION_CHANGE, onCollectionChange);
					break;
				
				default:
					break;
			}
		}
		
		/** An event listener to respond to events from the current collection.
		 */
		protected function onCollectionChange(event:CollectionEvent):void
		{
//			logger.info("Collection changed, reason: " + event.kind + ", num items: " + (collectionMgr.currentItems ? collectionMgr.currentItems.length : 0));
			
			switch (event.kind) {
				case CollectionEventKind.UPDATE:
//					updateColumnSizes = true;
//					invalidateSize();
					break;
				
				default:
					break;
			}
		}
		
		/**
		 *  Called when user changes field order by dragging/dropping (field) column headers.
		 * 
		 *  @param movedField: The field that is being moved.
		 *  @param afterFieldName: Determines the location of the moved field.
		 *  If not an empty string, the moved field is moved before this field.
		 *  If the empty string, the moved field is moved to the end of all fields.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function updateFieldOrder(movedField:IField, afterFieldName:String):void
		{
			var oldFields:Array = collectionMgr.fields;
			if (!oldFields || !movedField || afterFieldName == null)
				return;
			var newFields:Array = new Array();
			for each (var field:IField in oldFields) {
				if (!field)
					continue;
				if (field.name == movedField.name)
					continue;
				if (field.name == afterFieldName)
					newFields.push(movedField);
				newFields.push(field);
			}
			if (afterFieldName == "")
				newFields.push(movedField);
			collectionMgr.fields = newFields;
			
			// If this is a pending field order change, remember so it can
			// be made permanent if the user switches to authoring mode.
			if (!collectionMgr.authoringMode || !collectionMgr.permissionCheck("Document", "Modify"))
				pendingFieldOrderChange = true;
			// Else make the 'permanent' change now
			else if (collectionMgr.host && collectionMgr.host.collection)
				collectionMgr.host.collection.fields = collectionMgr.fields;
		}
		
		protected function fieldsChanged(value:Array):void
		{
			if (value == iFields)
				return;
			
			iFields = value;
			
			// The intent of this line is to "encourage" garbage collection on the old details fields
			// I don't know if it makes any difference, but that's why I am doing it
			detailsFields.removeAll();
			// Need a new arrayList, otherwise the setters in detailsHeader and detailsCards
			// do nothing since the value has not changed
			detailsFields = new ArrayList();
			var newDetailsFieldsMap:Dictionary = new Dictionary();
			var detailsField:DetailsField = null;
			for each (var iField:IField in iFields) {
				if (!iField || !iField.visible)
					continue;
				detailsField = new DetailsField(iField);
				// If this field existed before, update some info
				var prevDetailsField:DetailsField = detailsFieldsMap[detailsField.name] as DetailsField;
				if (prevDetailsField)
					detailsField.explicitWidth = prevDetailsField.explicitWidth;
				detailsFields.addItem(detailsField);
				newDetailsFieldsMap[detailsField.name] = detailsField;
			}
			detailsFieldsMap = newDetailsFieldsMap;
			
			if (detailsField)
				detailsField.resizable = false;
			
			if (detailsHeader)
				detailsHeader.fields = detailsFields;
			
			updateColumnSizes = true;
			invalidateSize();
		}
		
		/**
		 *  Called when user changes sort order by clicking on a (field) column header.
		 * 
		 *  @param fieldName: The field that the sort order is based on.
		 *  @param descending: Determines if the sort is descending (<code>true</code>) or ascending (<code>false</code>).
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function updateSortOrder(fieldName:String, descending:Boolean):void
		{
			var sortFields:Array = new Array;
			sortFields.push(new SortField(fieldName, true, descending));
			var sort:Sort = new Sort();
			sort.fields = sortFields;
			collectionMgr.sort = SortUtil.createSortArray(sort);

			// If this is a pending sort order change, remember so it can
			// be made permanent if the user switches to authoring mode.
			if (!collectionMgr.authoringMode || !collectionMgr.permissionCheck("Document", "Modify"))
				pendingSortOrderChange = true;
		}
		
		protected function sortChanged(value:Array):void
		{
			// Update the sort arrow and shading to match the new sort order
			if (detailsHeader)
				detailsHeader.updateSortIndicators();
			// Update the sort shading to match the new sort order
			if (detailsCards)
				detailsCards.updateSortIndicators();
		}
		
		protected function authoringModeChanged(value:Boolean):void
		{
			// Only need to check if switching to authoring mode (not when switching to preview mode)
			if (!value)
				return;
			// Nothing to do if no pending changes (sort order / field order)
			if (!pendingSortOrderChange && !pendingFieldOrderChange)
				return;
			// Can't revert if host or collection are invalid
			if (!collectionMgr.host || !collectionMgr.host.collection)
				return;
			// If a pending sort order change was made, set collection's sort order
			if (pendingSortOrderChange) {
				pendingSortOrderChange = false;
				collectionMgr.host.collection.sort = collectionMgr.sort;
			}
			// If a pending field order change was made, set collection's field order
			if (pendingFieldOrderChange) {
				pendingFieldOrderChange = false;
				collectionMgr.host.collection.fields = collectionMgr.fields;
			}
		}
		
		protected function onHeaderResize(event:ResizeEvent):void
		{
			// If the window resizes, the header gets the event, but the cards do not
			// Invalidating them will cause them to update due to the resize
			updateLayoutsOfHeaderAndCards();
		}
		
		protected function onHeaderClicked(event:DataGridEvent):void
		{
			// Clicking on the header (to sort by field/column)
			// took the focus away from the cards. This puts it back.
			if (detailsCards)
				detailsCards.returnFocusToDetailsCardList();
		}

		protected function onColumnResizeStart(event:DataGridEvent):void
		{
			var detailsHeaderItem:DetailsHeaderItem = event.itemRenderer as DetailsHeaderItem;
			if (!detailsHeaderItem || !detailsHeaderItem.detailsField) {
				// Let the header know that resizing is finished (aborted actually)
				if (detailsHeader)
					detailsHeader.columnResizingFinished();
				return;
			}
			// Find the left position of the field/column that is being resized
			var resizeFieldPos:Point = detailsHeaderItem.contentToGlobal(new Point());
			resizeFieldLeft = resizeFieldPos.x;
			resizeDetailsField = detailsHeaderItem.detailsField;
			// Listen for mouse move (resize) and up (resize done) events
			addEventListener(MouseEvent.MOUSE_MOVE, onMouseMoveInResizeArea);
			addEventListener(MouseEvent.MOUSE_UP, onMouseUpInResizeArea);
		}
	
		private function onMouseMoveInResizeArea(event:MouseEvent):void
		{
			// Sometimes we don't receive the mouse up event.
			// This is an alternate way to detect the mouse up.
			if (!event.buttonDown) {
				onMouseUpInResizeArea(event);
				return;
			}
			var rfWidth:Number = event.stageX - resizeFieldLeft;
			if (rfWidth >= MINIMUM_COLUMN_WIDTH) {
				resizeDetailsField.explicitWidth = rfWidth;
				updateColumnSizes = true;
				invalidateSize();
				invalidateDisplayList();
			}
		}
	
		private function onMouseUpInResizeArea(event:MouseEvent):void
		{
			removeEventListener(MouseEvent.MOUSE_MOVE, onMouseMoveInResizeArea);
			removeEventListener(MouseEvent.MOUSE_UP, onMouseUpInResizeArea);
			// Let the header know that resizing is finished
			if (detailsHeader)
				detailsHeader.columnResizingFinished();
			// Clicking in the header and resizing a field/column
			// took the focus away from the cards. This puts it back.
			if (detailsCards)
				detailsCards.returnFocusToDetailsCardList();
		}
		
		private function updateDefaultFieldColumnWidths():void
		{
			var error:String = tryToUpdateDefaultFieldColumnWidths();
			if (!error)
				return;
			
			// Log the reason why the update didn't complete
			CONFIG::logging {
				LogHelper.logger.info(error);
			}

			// Schedule a retry 200 ms later
			if (updateColumnSizesRetryTimer == null) {
				updateColumnSizesRetryTimer = new Timer(200, 1);
				var timerComplete:Function = function(event:TimerEvent):void
				{
					invalidateSize();
				};
				updateColumnSizesRetryTimer.addEventListener(TimerEvent.TIMER_COMPLETE, timerComplete);
			} else {
				updateColumnSizesRetryTimer.reset();
			}
			updateColumnSizesRetryTimer.start();
		}
		
		private function tryToUpdateDefaultFieldColumnWidths():String
		{
			CONFIG::logging {
				LogHelper.logger.info("updateDefaultFieldColumnWidths called, updateColumnSizes=" + updateColumnSizes);
			}
			var detailsField:DetailsField;
			
			if (!detailsFields || !detailsFields.source || !detailsFields.source.length) {
				return "UCW: No details fields";
			}
			
			// Initially set each column's calculated width based on the width of the header (text, sort indicator, etc)
			if (!detailsHeader || !detailsHeader.dataGroup || detailsHeader.dataGroup.numElements < 1) {
				return "UCW: No details header";
			}
			var headerItemRenderer:DetailsHeaderItem = detailsHeader.dataGroup.getElementAt(0) as DetailsHeaderItem;
			if (!headerItemRenderer) {
				return "UCW: No header item renderer";
			}
			var headerItemWidth:Number = 0;
			for each (detailsField in detailsFields.source) {
				if (!detailsField)
					continue;
				headerItemWidth = headerItemRenderer.getCalculatedWidth(detailsField, measureField);
				if (isNaN(headerItemWidth)) {
					return "UCW: No header item width info";
				}
				detailsField.calculatedWidth = Math.max(headerItemWidth, MINIMUM_COLUMN_WIDTH);
			}

			// Then look at a few of the files and increase the column widths for any columns with data wider than the header
			if (!detailsCards || !detailsCards.dataGroup) {
				return "UCW: No details cards";
			}
			var cardItemWidth:Number = 0;
			var maxCount:int = Math.min(detailsCards.dataGroup.numElements, MAXIMUM_DATA_ITEMS_TO_CHECK);
			for (var cardIndex:int = 0; cardIndex < maxCount; ++cardIndex) {
				var detailsCard:DetailsCard = detailsCards.dataGroup.getElementAt(cardIndex) as DetailsCard;
				if (!detailsCard || !detailsCard.dataGroup) {
					CONFIG::logging {
						LogHelper.logger.info("UCW: No details card, index is " + cardIndex);
					}
					continue;
				}
				for (var fieldIndex:int = 0; fieldIndex < detailsFields.source.length; ++fieldIndex) {
					detailsField = detailsFields.source[fieldIndex] as DetailsField;
					if (!detailsField)
						continue;
					var cardItemRenderer:DetailsCardItem = detailsCard.dataGroup.getElementAt(fieldIndex) as DetailsCardItem;
					if (!cardItemRenderer) {
						return "UCW: No card item renderer";
					}
					cardItemWidth = cardItemRenderer.getCalculatedWidth(detailsField, measureField);
					if (isNaN(cardItemWidth)) {
						return "UCW: No card item width info";
					}
					cardItemWidth = Math.min(Math.max(cardItemWidth, MINIMUM_COLUMN_WIDTH), MAXIMUM_COLUMN_WIDTH);
					if (cardItemWidth > detailsField.calculatedWidth)
						detailsField.calculatedWidth = cardItemWidth;
				}
			}

			// We updated column sizes based on header and data, so don't need to do this again until fields change
			CONFIG::logging {
				LogHelper.logger.info("Column sizes updated");
			}
			updateColumnSizes = false;
			
			// Since we were successful, cancel the retry timer if it is pending
			if (updateColumnSizesRetryTimer && updateColumnSizesRetryTimer.running)
				updateColumnSizesRetryTimer.stop();
			
			// Update the layouts after changing the column widths
			// This one is still needed because updateDisplayList will not
			// be called in the rare case where there are no cards (newly created portfolio).
			updateLayoutsOfHeaderAndCards();
			
			// Update the layouts after an updateDisplayList is received.
			// This is needed to update it after a column has been added.
			updateLayoutsOfHeaderAndCardsNeeded = true;
			
			return null;
		}
		
		private function updateLayoutsOfHeaderAndCards():void
		{
			// Update the layout of the header items
			if (detailsHeader && detailsHeader.dataGroup)
				detailsHeader.dataGroup.invalidateDisplayList();
			
			// Update the layout of each of the card items
			if (detailsCards && detailsCards.dataGroup) {
				for (var index:int = 0; index < detailsCards.dataGroup.numElements; ++index) {
					var detailsCard:DetailsCard = detailsCards.dataGroup.getElementAt(index) as DetailsCard;
					if (!detailsCard || !detailsCard.dataGroup)
						continue;
					// Update the size of the card (which seems to affect redrawing the horizontal scrollbar)
					detailsCard.invalidateSize();
					detailsCard.dataGroup.invalidateDisplayList();
				}
			}
		}
		
		private var _fieldColumnMinWidth:Number;
		/**
		 *  The sum of actual widths of all fields.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		[Bindable (event="propertyChange")]
		public function get fieldColumnMinWidth():Number
		{
			return _fieldColumnMinWidth;
		}
		
		private var _fieldColumnTotalWidth:Number;
		/**
		 *  The actual width of all fields, plus any additional space to equal the width of Details.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		[Bindable (event="propertyChange")]
		public function get fieldColumnTotalWidth():Number
		{
			return _fieldColumnTotalWidth;
		}
		
		/**
		 *  Called by DetailsHeaderLayout and DetailsCardLayout classes to decide the layout of each header and card item.
		 * 
		 *  @return an array of width values (Number) corresponding to the width of each field.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function getFieldColumnWidths():Array
		{
			var widths:Array = new Array();
			if (!detailsFields || !detailsFields.source)
				return widths;
			var fitToWidth:Number = width - Details.PREVIEW_ICON_WIDTH - 2;
			var totalWidths:Number = 0;
			var autoGrowFieldIndex:int = -1;
			var lastVisibleFieldIndex:int = -1;
			for each (var detailsField:DetailsField in detailsFields.source) {
				if (!detailsField)
					continue;
				if (detailsField.visible) {
					// Track the last visible field (which will get the extra space if description doesn't)
					lastVisibleFieldIndex = widths.length;
					// If this is the description field and the user hasn't explicitly resized it,
					// give it any extra space (delta between fitToWidth and the minScrollWidth)
					if (detailsField.type == FieldType.DESCRIPTION && isNaN(detailsField.explicitWidth))
						autoGrowFieldIndex = widths.length;
				}
				totalWidths += detailsField.width;
				widths.push(detailsField.width);
			}
			
			if (_fieldColumnMinWidth != totalWidths) {
				var originalMinValue:Number = _fieldColumnMinWidth;
				_fieldColumnMinWidth = totalWidths;
				dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "fieldColumnMinWidth", originalMinValue, _fieldColumnMinWidth));
			}
			
			// If no field wanted the extra space, give it to the last field
			if (autoGrowFieldIndex == -1)
				autoGrowFieldIndex = lastVisibleFieldIndex;
			// If a field can take the extra space and fields had widths and extra space exists
			// then give the extra space to the field that can use it (description field)
			if ((autoGrowFieldIndex >= 0) && totalWidths && (fitToWidth > totalWidths)) {
				widths[autoGrowFieldIndex] += (fitToWidth - totalWidths);
				totalWidths = fitToWidth;
			}
			if (_fieldColumnTotalWidth != totalWidths) {
				var originalTotalValue:Number = _fieldColumnTotalWidth;
				_fieldColumnTotalWidth = totalWidths;
				dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "fieldColumnTotalWidth", originalTotalValue, _fieldColumnTotalWidth));
				if ((detailsCards != null) && (detailsCards.dataGroup != null)) {
					// This should allow the list to adjust its scrollbars
					detailsCards.dataGroup.invalidateSize();
					detailsCards.dataGroup.invalidateDisplayList();
				}
			}
			
			return widths;
		}
		
		override protected function initializeAccessibility():void
		{
			var runningOnWindows:Boolean = Capabilities.version.indexOf('WIN') != -1;
			
			if (runningOnWindows)
				accessibilityImplementation = new DetailsAccImpl(this);
		}
	}
}
