/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.details
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.utils.AttachmentCollection;
	import com.adobe.portfolio.utils.FieldUtils;
	
	import flash.accessibility.Accessibility;
	import flash.events.Event;
	import flash.events.FocusEvent;
	import flash.ui.Keyboard;
	
	import mx.accessibility.AccConst;
	import mx.accessibility.AccImpl;
	import mx.binding.utils.*;
	import mx.core.UIComponent;
	import mx.events.CollectionEvent;
	import mx.events.CollectionEventKind;
	
	import spark.events.IndexChangeEvent;
	import spark.events.RendererExistenceEvent;
	
	/**
	 *  The DetailsAccImpl class extends AccImpl to implement accessability functionality for the Details navigator. 
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */	
	public class DetailsAccImpl extends AccImpl
	{
		/**
		 *  The navigator's CollectionManager instance.
		 *  All host interaction should go through this instance, rather than the INavigatorHost directly.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected const collectionMgr:CollectionManager = CollectionManager.instance;
		
		private var focusRowIndex:int  = 0;
		private var focusHeaderIndex:int = 0;
		private var firstObjectInFocusRowID:int;
		private var headerColumnFocusedID:int;
		private var numColumns:int = 0;
		private var numRows:int = 0;
		private var selectedRows:Vector.<int>;
		private var focusAtt:IAttachment = null;
		
		/**
		 *  Constructor.
		 * 
		 *  @param master The UIComponent instance that this AccImpl instance
		 *  is making accessible. In this case, the Details navigator.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function DetailsAccImpl(master:UIComponent)
		{
			super(master);
			
			headerColumnFocusedID = -1;
			firstObjectInFocusRowID = -1;
			selectedRows = new Vector.<int>;
			
			var details:Details = master as Details;
			
			var detailsCardList:DetailsCardList = details.detailsCards;
			detailsCardList.addEventListener( IndexChangeEvent.CARET_CHANGE, onCardIndexChangeEvent, false, 0, true);
			detailsCardList.addEventListener( FocusEvent.FOCUS_IN, cardFocusEvent, false, 0, true);
			detailsCardList.addEventListener( FocusEvent.FOCUS_OUT, cardFocusEvent, false, 0, true);
			BindingUtils.bindSetter( selectionHandler, detailsCardList, "selectedIndices");
			
			var detailsHeader:DetailsHeader = details.detailsHeader;
			detailsHeader.addEventListener( FocusEvent.FOCUS_IN, headerFocusEvent, false, 0, true);
			detailsHeader.addEventListener( FocusEvent.FOCUS_OUT, headerFocusEvent, false, 0, true);
			detailsHeader.dataGroup.addEventListener( RendererExistenceEvent.RENDERER_ADD, headerAddDel, false, 0, true);
			
			collectionMgr.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent);
		}
		
		// change of the current items i.e. folder change - alert MSAA
		protected function onNavHostEvent(event:CollectionManagerEvent):void
		{
			var aColl:AttachmentCollection;
			
			switch (event.kind)
			{
				case CollectionManagerEvent.ITEMS_WILL_CHANGE_KIND:
					aColl = event.oldVal as AttachmentCollection;
					if (aColl)
						aColl.removeEventListener( CollectionEvent.COLLECTION_CHANGE, collectionChanged);
					break;
				
				case CollectionManagerEvent.ITEMS_DID_CHANGE_KIND:
					aColl = event.newVal as AttachmentCollection;
					if (aColl)
					{
						numRows = aColl.length;   // starting point for any more additions by collectionChanged
						Accessibility.sendEvent(master, 0, AccConst.EVENT_OBJECT_REORDER);
						
						aColl.addEventListener(CollectionEvent.COLLECTION_CHANGE, collectionChanged);
					}
					break;
			}
		}
		
		// the attachments may not all have neen known when the collection changed above; track additions (and deletions) here
		protected function collectionChanged(evt:CollectionEvent):void
		{
			var madeChange:Boolean = false;

			switch(evt.kind)
			{
				case CollectionEventKind.ADD:
					numRows = numRows + evt.items.length;
					madeChange = true;
					break;
					
				case CollectionEventKind.REMOVE:
					numRows = numRows - evt.items.length;
					madeChange = true;
					break;
			}
			
			if (madeChange)
				Accessibility.sendEvent(master, 0, AccConst.EVENT_OBJECT_REORDER);  // keep letting MSAA know
		}

		/**
		 *  Method to return an array of childIDs.
		 *  The array consists of first the headers and then all of the cells of the table.
		 * 
		 *  @return Array.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function getChildIDArray():Array
		{
			var i:int;
			var childIndex:int;
						
			var childIDs:Array = [];
			
			// child ID 0 is reserved for the table as a whole - hence the "+1"
			childIndex = 0;
			for (i = 0; i<numColumns; i++)
				childIDs[childIndex++] = i + 1;
	
			for (i = 0; i< numRows * numColumns; i++)
			{
				childIDs[childIndex] = childIndex + 1;
				childIndex++;
			}
			
			return childIDs;
		}
		
		private function getTableCellN( numColumns:int, childID:uint):*
		{
			if ( childID <= numColumns)
				return "";
			
			var tableIndex:uint = childID - numColumns - 1;
			
			if (tableIndex >= numRows * numColumns)
				return "";
			
			var row:uint = tableIndex / numColumns;
			var column:uint = tableIndex % numColumns;
			
			var details:Details = master as Details;
			var detailsCardList:DetailsCardList = details.detailsCards;
			if ( detailsCardList == null)
				return "";
			
			var detailsCard:DetailsCard = detailsCardList.dataGroup.getElementAt(row) as DetailsCard;
			if ( detailsCard == null)
				return "";
			
			var detailsCardItem:DetailsCardItem = detailsCard.dataGroup.getElementAt(column) as DetailsCardItem;
			return detailsCardItem;
		}
		
		/**
		 *  Returns the value of the component.
		 * 
		 *  @param childID identifies internal part that this method applies to.
		 *  @return Value of the component.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function get_accValue(childID:uint):String
		{
			var details:Details = master as Details;
			var detailsHeader:DetailsHeader = details.detailsHeader;
					
			if (childID == 0)
				return "";
			
			// Is the ID for one of the column header children?
			if ( childID <= numColumns)
			{
				var detailHead:DetailsHeaderItem = detailsHeader.dataGroup.getChildAt( childID-1) as DetailsHeaderItem;
				if ( detailHead && detailHead.headerName)
					return detailHead.headerName.text;
				else
					return "";
			}

			var detailsCardItem:DetailsCardItem = getTableCellN( numColumns, childID) as DetailsCardItem;
			if ( detailsCardItem && detailsCardItem.detailsField)
				return FieldUtils.getTextFromField(detailsCardItem.attachment, detailsCardItem.detailsField.field);
			else
				return "";
		}
		
		/**
		 *  Returns the system role for the component.
		 * 
		 *  @param childID identifies internal part that this method applies to.
		 *  @return Role associated with the component.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function get_accRole(childID:uint):uint
		{
			if (childID == 0)
				return AccConst.ROLE_SYSTEM_TABLE;
		
			if (childID <= numColumns)
				return AccConst.ROLE_SYSTEM_COLUMNHEADER;
			
			return AccConst.ROLE_SYSTEM_CELL;
		}
		
		override protected function getName(childID:uint):String
		{
			 return "";
		/*
			// The spec says the names are to all be null/empty. Keeping this original naming code for debugging purposes.
			if (childID == 0)
				return "";

			if (childID <= numColumns)
				return "0," + (childID-1);

			var tableIndex:uint = childID - numColumns - 1;
			var row:uint = tableIndex / numColumns;
			return  row + "," + (tableIndex % numColumns);
		*/	
		}
		
		/**
		 *  Returns the location for the component.
		 * 
		 *  @param childID identifies internal part that this method applies to.
		 *  @return Location for the component.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function accLocation(childID:uint):*
		{
			var details:Details = master as Details;
			var detailsHeader:DetailsHeader = details.detailsHeader;
			
			if (childID == 0)
				return details;
			
			if (childID <= numColumns)
			{
				if ( detailsHeader && detailsHeader.dataGroup)
					return detailsHeader.dataGroup.getElementAt( childID - 1);
				else
					return "";
			}

			return getTableCellN( numColumns, childID);
		}
		
		override protected function get eventsToHandle():Array
		{
			return super.eventsToHandle.concat([ "select", "focus"]);
		}
		
		protected function selectionHandler(value:Vector.<int>):void
		{
			if ( value)
			{
				// what rows are selected now that weren't before - fire selection events for them
				for (var i:int=0; i<value.length; i++)
					if ( selectedRows.indexOf( value[i]) == -1)
						fireSelectionOnRow( value[i]);
			
				selectedRows = value;
			}
		}
		
		/**
		 *  Returns the state for the component.
		 * 
		 *  @param childID identifies internal part that this method applies to.
		 *  @return State for the component.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function get_accState(childID:uint):uint
		{	
			if ( childID == 0)  
				return 0;   // table as a whole
			
			if ( childID == headerColumnFocusedID)
				return AccConst.STATE_SYSTEM_FOCUSED | AccConst.STATE_SYSTEM_SELECTED | AccConst.STATE_SYSTEM_FOCUSABLE;		
			
			if ( childID <= numColumns)
			{
				if ( childID == ( focusHeaderIndex + 1))
					return AccConst.STATE_SYSTEM_SELECTED | AccConst.STATE_SYSTEM_FOCUSABLE;
				else
					return AccConst.STATE_SYSTEM_FOCUSABLE;
			}
			
			var focussed:uint = 0;
			if ( childID == firstObjectInFocusRowID)
				focussed = AccConst.STATE_SYSTEM_FOCUSED;
			
			var rowIndx:int = (childID - numColumns - 1) / numColumns;
			if ( selectedRows.indexOf( rowIndx) == -1)    // is this cell in one of the selected rows?
				return focussed | AccConst.STATE_SYSTEM_FOCUSABLE;   // no
			else
				return focussed | AccConst.STATE_SYSTEM_SELECTED;    // yes
		}
		
		/**
		 *  Performs actions based on the selection flag for the component.
		 * 
		 *  @param selFlag set of flags identifying action to respond to.
		 *  @param childID identifies internal part that this method applies to.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function accSelect( selFlag:uint, childID:uint):void
		{
			// This is the only case we're being asked to handle right now.
			if (selFlag == (AccConst.SELFLAG_TAKEFOCUS | AccConst.SELFLAG_TAKESELECTION))
				accDoDefaultAction( childID);
		}
		
		/*
			Note: One of the things that this code has to worry about is the tab focus. When we "click" 
			on a header or row via accDoDefaultAction, the tab focus must be moved to the header or body respectively.
			If we need to move the tab focus we use a FocusEvent.KEY_FOCUS_CHANGE. We can tell where the current tab focus
			is from the values of headerColumnFocusedID and firstObjectInFocusRowID and, we know that the tab cycle is
			header -> body -> go back button (if present) -> header. 
		
		*/
		/**
		 *  Performs default action for the component.
		 * 
		 *  @param childID identifies internal part that this method applies to.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function accDoDefaultAction( childID:uint):void
		{
			if ( childID == 0)
				return;
			
			var details:Details = master as Details;
			var focusEvent:FocusEvent
			
			if ( childID <= numColumns)
			{
					// click header item				
				var detailsHeader:DetailsHeader = details.detailsHeader;

				if ( !detailsHeader.dataGroup)
					return;
				
				var headerItem:DetailsHeaderItem = detailsHeader.dataGroup.getElementAt(childID-1) as DetailsHeaderItem;
				if ( headerItem)
				{	
					details.updateSortOrder( headerItem.detailsField.name, !headerItem.sortAscending);

					focusHeaderIndex = childID - 1;
								
					if ( headerColumnFocusedID == -1)
					{   // tab focus is not on header - move it; will trigger MSAA event too
						focusEvent = new FocusEvent(FocusEvent.KEY_FOCUS_CHANGE);
						focusEvent.keyCode = Keyboard.TAB;
						if ( firstObjectInFocusRowID != -1)
							focusEvent.shiftKey = true;    // body has tab focus so go backwards in cycle to get to header
						focusEvent.relatedObject = headerItem;
						
						headerItem.dispatchEvent( focusEvent);
					}
					else
					{	// tab focus is on header; just send event to MSAA clients
						headerColumnFocusedID = childID;
						Accessibility.sendEvent(master, focusHeaderIndex+1, AccConst.EVENT_OBJECT_FOCUS);
					}
				}
							
				return;
			}
			
			// select a row
			var rowNum:int = (childID-numColumns-1) / numColumns;
			if ( (rowNum >= 0) && (rowNum < numRows))
			{
				var detailsCardList:DetailsCardList = details.detailsCards;
				if ( detailsCardList)
				{
					var newList:Vector.<int> = new Vector.<int>();			
					newList.push( rowNum);
				
					detailsCardList.setSelectedIndicesWithChangeDispatch( newList);
				}
						
				if ( firstObjectInFocusRowID == -1)
				{   // tab focus is not on body - move it
					focusEvent = new FocusEvent(FocusEvent.KEY_FOCUS_CHANGE);
					focusEvent.keyCode = Keyboard.TAB;
					if ( headerColumnFocusedID == -1)
						focusEvent.shiftKey = true;    // tab focus not on header either must be on go back button so go backwards in cycle
					focusEvent.relatedObject = detailsCardList;
					
					detailsCardList.dispatchEvent( focusEvent);
				}
			}
		}
		
		/**
		 *  Returns the default action for the component.
		 * 
		 *  @param childID identifies internal part that this method applies to.
		 *  @return Default action.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function get_accDefaultAction(childID:uint):String
		{
			if ( childID == 0)
				return null;
			
			return "Click";
		}
		
		/*  This routine is used to fire off a blizzard of MSAA events when a row in the table is selected:
			A selection event is sent for each cell in the row.
		*/
		private function fireSelectionOnRow(index:int):void
		{
			var childID:uint = index * numColumns + numColumns + 1;

			for (var i:int=0; i<numColumns; i++)
				Accessibility.sendEvent(master, childID + i, AccConst.EVENT_OBJECT_SELECTION);	
		}
		
		private function fireFocusOnRow(index:int):void
		{
			var childID:uint = index * numColumns + numColumns + 1;
			
			focusRowIndex = index;
			firstObjectInFocusRowID = childID;

			Accessibility.sendEvent(master, childID, AccConst.EVENT_OBJECT_FOCUS);
		}
		
		protected function onCardIndexChangeEvent(event:IndexChangeEvent):void
		{
			if ( (event.newIndex >= 0) && collectionMgr.currentItems)
			{
				focusRowIndex = event.newIndex;
				
				var att:IAttachment = collectionMgr.currentItems.getItemAt(event.newIndex) as IAttachment;
				if ( att == null)
					return;
				
				if ( att == focusAtt)
					return;					// Don't send a focus event if index changes because the card in focus moved as a result
											// of new sorting. It's the same underlying card.
				
				focusAtt = att;
				
				fireFocusOnRow( event.newIndex);
			}
		}
		
		// we get here when tabbing and focus goes to/from the table body
		private function cardFocusEvent(event:FocusEvent):void
		{
			if ( event.type == FocusEvent.FOCUS_IN)
			{
				fireFocusOnRow( focusRowIndex);
			}
			else
			if ( event.type == FocusEvent.FOCUS_OUT)
			{
				firstObjectInFocusRowID = -1;
			}
		}
		
		// we get here when tabbing and focus goes to/from the header
		private function headerFocusEvent(event:FocusEvent):void
		{
			if ( event.type == FocusEvent.FOCUS_IN)
			{
				Accessibility.sendEvent(master, focusHeaderIndex+1, AccConst.EVENT_OBJECT_FOCUS);
				headerColumnFocusedID = focusHeaderIndex + 1;
			}
			else
			if ( event.type == FocusEvent.FOCUS_OUT)
			{
				headerColumnFocusedID = -1;
			}
		}
		
		private function headerItemFocusEvent(event:FocusEvent):void
		{
			var detailsHeaderItem:DetailsHeaderItem = event.currentTarget as DetailsHeaderItem;
			
			var details:Details = master as Details;
			var detailsHeader:DetailsHeader = details.detailsHeader;
			
			if ( (!detailsHeader.dataGroup) || (event.type != FocusEvent.FOCUS_IN))
				return;
			
			for (var index:int = 0; index < detailsHeader.dataGroup.numElements; ++index)
				if (detailsHeaderItem == (detailsHeader.dataGroup.getElementAt(index) as DetailsHeaderItem))
				{	
					var childID:uint = index + 1;
					headerColumnFocusedID = childID;
					focusHeaderIndex = index;
						
					Accessibility.sendEvent(master, childID, AccConst.EVENT_OBJECT_FOCUS);			
					return;
				}
		}
		
		protected function headerAddDel(event:RendererExistenceEvent):void
		{
			if ( event.type == RendererExistenceEvent.RENDERER_ADD)
			{
				numColumns++;
				
				var detailHeader:DetailsHeaderItem = event.renderer as DetailsHeaderItem;
				if (detailHeader)
					detailHeader.addEventListener( FocusEvent.FOCUS_IN, headerItemFocusEvent, false, 0, true);
			}
			// else?  Well, with no edit mode support for accessibility no need to deal with deletes
		}

		override protected function eventHandler(event:Event):void
		{   
			// Let AccImpl class handle the events
			// that all accessible UIComponents understand.
			$eventHandler(event);
		}
	}
}
