/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.details
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.cards.SimpleCard;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.collections.ArrayList;
	
	import spark.components.DataGroup;
	import spark.events.RendererExistenceEvent;
	
	/**
	 *  The DetailsCard class extends SimpleCard to implement a row (corresponding to an IAttachment) in the Details navigator. 
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */	
	public class DetailsCard extends SimpleCard
	{
		/**
		 *  The data group that holds the card items. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		[SkinPart(required="true")]
		public var dataGroup:DataGroup;

		/**
		 *  Constructor. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function DetailsCard()
		{
			super();
			
			tabFocusEnabled = false;       // no focus on the row as a whole (before stepping through editable fields - in 10.0+)
			hasFocusableChildren = false;  // step through fields with tab (until in edit mode which is not being done for 10.0)
		}
		
		private var _oddCard:Boolean;
		
		/**
		 *  Returns whether a card is in an odd numbered row or not. Used by the card skin to draw alternating colors for each row of cards.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		[Bindable]
		public function get oddCard():Boolean
		{
			return _oddCard;
		}
		
		/**
		 *  @private
		 */	
		public function set oddCard(value:Boolean):void
		{
			_oddCard = value;
		}
		
		private var _details:Details;

		/**
		 *  The main Details class.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function get details():Details
		{
			return _details;
		}

		/**
		 *  @private
		 */	
		public function set details(value:Details):void
		{
			if (value == _details)
				return;
			if (fieldsWatcher != null) {
				fieldsWatcher.unwatch();
			}
			_details = value;
			if (_details != null) {
				fieldsWatcher = BindingUtils.bindSetter(setFields, _details, "detailsFields");
			}
			if (dataGroup) {
				var dcLayout:DetailsCardLayout = dataGroup.layout as DetailsCardLayout;
				if (dcLayout)
					dcLayout.details = details;
			}
		}
		
		private var fieldsWatcher:ChangeWatcher;
		private function setFields(value:ArrayList):void
		{
			fields = value;
		}
		
		private var _fields:ArrayList;
		
		protected function get fields():ArrayList
		{
			return _fields;
		}
		
		/**
		 *  @private
		 */	
		protected function set fields(value:ArrayList):void
		{
			if (value == _fields)
				return;
			_fields = value;
			if (dataGroup)
				dataGroup.dataProvider = _fields;
		}
		
		/**
		 *  Subclass SimpleCard's adjustAspectRatio getter and return <code>false</code> as Details does not need to consider the aspect ratio.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		override public function get maintainAspectRatio():Boolean
		{
			// DetailsCard should never consider the aspectRatio of the preview for layout
			return false;
		}

		override protected function measure():void
		{
			super.measure();
			if ((details != null) && !isNaN(details.fieldColumnTotalWidth) && (details.fieldColumnTotalWidth > 0)) {
				measuredWidth = details.fieldColumnTotalWidth;
			}
		}
		
		/**
		 *  Subclass SimpleCard's itemIndex setter to determine if this card is in an even or odd row.
		 *  @see #oddCard
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		override public function set itemIndex(value:int):void
		{
			super.itemIndex = value;
			oddCard = value & 1 ? true : false;
		}
		
		/**
		 *  Subclass SimpleCard's selected setter so the color of the card can be set based on its selection state.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		override public function set selected(value:Boolean):void
		{
			super.selected = value;
			if (!colorPalette)
				return;
			// Set the text color for all card items on the card that use text
			// If in high contrast mode and this is the selected text, use the secondary text color [highlighted text]
			// otherwise use the primary text color.
			setStyle("color", colorPalette.isHighContrast && selected ? colorPalette.secondaryText : colorPalette.primaryText);
		}
		
		/** Called when a skin has added a skin part for this instance.
		 */
		override protected function partAdded(partName:String, instance:Object):void
		{
			super.partAdded(partName, instance);
			switch (instance) {
				case dataGroup:
					var dcLayout:DetailsCardLayout = dataGroup.layout as DetailsCardLayout;
					if (dcLayout)
						dcLayout.details = details;
					addListener(dataGroup, RendererExistenceEvent.RENDERER_ADD, rendererAddHandler);
					dataGroup.dataProvider = fields;
					break;
				
				default:
					break;
			}
		}
		
		/**
		 *  @private
		 *  Called when an item has been added to this component.
		 */
		private function rendererAddHandler(event:RendererExistenceEvent):void
		{
			var cardItem:DetailsCardItem = event.renderer as DetailsCardItem;
			if (!cardItem)
				return;
			cardItem.attachment = attachment;
		}
		
		/**
		 *  Called when the sort order is changed by clicking in the header of a column.
		 *  This method calls updateSortState in each of the DetailsCardItems for this card.
		 *  @see DetailsCardItem
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function updateSortIndicators():void
		{
			if (!dataGroup)
				return;
			for (var index:int = 0; index < dataGroup.numElements; ++index) {
				var detailsCardItem:DetailsCardItem = dataGroup.getElementAt(index) as DetailsCardItem;
				if (!detailsCardItem)
					continue;
				detailsCardItem.updateSortState();
			}
		}
		
		/**
		 *  Subclass SimpleCard's data provider so we can pass the attachment on to each of the card items.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		override public function set data(value:Object):void
		{
			var oldAttachment:IAttachment = attachment;
			// Make a first access to the folder so that its list of children is populated. 
			// This is required by things like reordering, or adding files to a closed folder. 
			var newAttachment:IAttachment = value as IAttachment;
			if (newAttachment && newAttachment.isFolder)
				newAttachment.children;
			super.data = value;
			if (attachment != oldAttachment && dataGroup) {
				// Tell each card item about the new attachment (whether it is null or not)
				for (var index:int = 0; index < dataGroup.numElements; ++index) {
					var detailsCardItem:DetailsCardItem = dataGroup.getElementAt(index) as DetailsCardItem;
					if (!detailsCardItem)
						continue;
					detailsCardItem.attachment = attachment;
				}
			}
		}
				
	}
}
