/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.details
{
	import com.adobe.portfolio.api.FieldType;
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.api.IField;
	import com.adobe.portfolio.api.ISort;
	import com.adobe.portfolio.components.NavSkinnableComponent;
	import com.adobe.portfolio.components.supportClasses.DateTimeChooser;
	import com.adobe.portfolio.components.supportClasses.FieldTextArea;
	import com.adobe.portfolio.components.supportClasses.FieldTextInput;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.utils.FieldUtils;
	
	import flash.events.FocusEvent;
	import flash.events.MouseEvent;
	
	import mx.controls.listClasses.IListItemRenderer;
	import mx.core.UIFTETextField;
	import mx.events.FlexEvent;
	
	/**
	 *  The DetailsCardItem class extends NavSkinnableComponent to implement a cell in a DetailsCard in the Details navigator. 
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */	
	public class DetailsCardItem extends NavSkinnableComponent implements IListItemRenderer
	{
		/**
		 *  The navigator's CollectionManager instance.
		 *  All host interaction should go through this instance, rather than the INavigatorHost directly.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected const collectionMgr:CollectionManager = CollectionManager.instance;
		
		/**
		 * String literals used to represent the states of the details card item.
		 */
		private static const STATE_DISABLED:String = "disabled";
		private static const STATE_ONE_LINE_TEXT:String = "oneLineText";
		private static const STATE_MANY_LINES_TEXT:String = "manyLinesText";
		private static const EDITING_DATE_TIME:String = "editingDateTime";
		
		// ToDo: Seems like it would be better if this came from the skin (or total width of skin parts)
		private static const DETAILS_CARD_ITEM_PADDING:Number = 20;
		private static const DETAILS_CARD_ITEM_FUDGE_FACTOR:Number = 1.25;
		
		/**
		 * Use this to measure a field which has no data
		 */
		private static const MINIMUM_FIELD_DATA:String = "X";
		
		private var editingDateField:Boolean = false;

		/**
		 *  The one line text field for the card item.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="true")]
		public var oneLineTextEntry:FieldTextInput;
		
		/**
		 *  The multi line text field for the card item.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="true")]
		public var manyLinesTextEntry:FieldTextArea;
		
		/**
		 *  The date/time chooser for user defined date fields.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="false")]
		public var dateTimeChooser:DateTimeChooser;

		/**
		 *  Constructor.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function DetailsCardItem()
		{
			super();
		}
		
		private var _fieldName:String;

		/**
		 *  The field name of the IField/DetailsField for this card item.
		 *  The DetailsCardItemSkin binds to this value and uses it to display the FieldTextArea field to render the field information.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public function get fieldName():String
		{
			return _fieldName;
		}

		/**
		 *  @private
		 *
		 */	
		public function set fieldName(value:String):void
		{
			if (value == _fieldName)
				return;
			_fieldName = value;
		}
		
		private var _attachment:IAttachment;

		/**
		 *  The IAttachment for this card item.
		 *  The DetailsCardItemSkin binds to this value and uses it to display the FieldTextArea field to render the field information.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public function get attachment():IAttachment
		{
			return _attachment;
		}

		/**
		 *  @private
		 */	
		public function set attachment(value:IAttachment):void
		{
			if (value == _attachment)
				return;
			_attachment = value;
		}
		
		private var _sortedField:Boolean = false;
		
		/**
		 *  The sorted state of this field.
		 *  The DetailsCardItemSkin binds to this value and uses it to draw sorted fields differently from other fields.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[Bindable]
		public function get sortedField():Boolean
		{
			return _sortedField;
		}
		
		/**
		 *  @private
		 */	
		public function set sortedField(value:Boolean):void
		{
			_sortedField = value;
		}
		
		/**
		 * Returns the current state of this instance, which is a variant of the component state depending on
		 * whether or not the component is selected.
		 */
		override protected function getCurrentSkinState():String
		{
			if (!detailsField || !detailsField.visible)
				return STATE_DISABLED;
			switch (detailsField.type) {
				case FieldType.DESCRIPTION:
				case FieldType.STRING:
					return STATE_MANY_LINES_TEXT;
					
				// Show date/time editor if editing user defined DATE field
				case FieldType.DATE:
					if (editingDateField) {
						initDateInDateTimeChooser();
						return EDITING_DATE_TIME;
					}
					// Note that if we are not editing a user defined date field
					// that we purposely fall thru to the default and use a
					// one line text field to display the date.
					
				default:
					return STATE_ONE_LINE_TEXT;
			}
		}
		
		private var detailsFieldChanged:Boolean = false;
		private var sortStateChanged:Boolean = false;
		override protected function commitProperties():void
		{
			// Call super's commitProperties first so the skins are updated before the fieldName is changed
			super.commitProperties();
			if (detailsFieldChanged) {
				// Initialize some things if the field is not null
				if (detailsField) {
					enabled = detailsField.visible;
					includeInLayout = detailsField.visible;
					fieldName = detailsField.field ? detailsField.field.name : "";
				}
					
				// Anything to be reset if the new field value is null?
				else {
					enabled = false;
					includeInLayout = false;
					fieldName = "";
				}
				detailsFieldChanged = false;
				sortStateChanged = true;
			}
			if (sortStateChanged) {
				if (collectionMgr.sort && collectionMgr.sort.length) {
					var field:IField = detailsField ? detailsField.field : null;
					var iSort:ISort = collectionMgr.sort[0] as ISort;
					// If sort exists, field exists and this is the primary sort field,
					// show the icon and as ascending/descending, whichever it is now
					if (iSort && field && field.visible && field.name == iSort.name) {
						sortedField = true;
					}
					else {
						sortedField = false;
					}
				}
				sortStateChanged = false;
			}
		}

		/** Called when a skin has added a skin part for this instance.
		 */
		override protected function partAdded(partName:String, instance:Object):void
		{
			super.partAdded(partName, instance);
			switch (instance) {
				case oneLineTextEntry:
					oneLineTextEntry.doubleClickEnabled = true;
					addListener(oneLineTextEntry, MouseEvent.CLICK, onItemClicked);
					addListener(oneLineTextEntry, MouseEvent.DOUBLE_CLICK, onItemDoubleClicked);
					break;
				
				case manyLinesTextEntry:
					manyLinesTextEntry.doubleClickEnabled = true;
					addListener(manyLinesTextEntry, MouseEvent.DOUBLE_CLICK, onItemDoubleClicked);
					break;
				
				case dateTimeChooser:
					initDateInDateTimeChooser();
					addListener(dateTimeChooser, FlexEvent.DATA_CHANGE, onDataChange);
					addListener(dateTimeChooser, FocusEvent.FOCUS_OUT, onFocusOut);
					break;
				
				default:
					break;
			}
		}
		
		private function onItemClicked(event:MouseEvent):void
		{
			// Ignore for anything other than user defined date fields
			if (!detailsField || detailsField.type != FieldType.DATE)
				return;
			// Ignore click (to edit date field) if not in authoring mode...
			if (!collectionMgr.authoringMode)
				return;
			editingDateField = true;
			invalidateSkinState();
			invalidateProperties();
		}
		
		private function onItemDoubleClicked(event:MouseEvent):void
		{
			// If we're in authoring mode and this is an editable field
			// then ignore the double click (since it should be handled by the field)
			if (collectionMgr.authoringMode && detailsField && detailsField.editable)
				return;
			// For any other case, preview the file or folder
			if (attachment && attachment.isFolder)
				collectionMgr.currentFolder = attachment;
			else
				collectionMgr.dispatchNavigatorPreview(attachment);
		}
		
		private function initDateInDateTimeChooser():void
		{
			if (!dateTimeChooser || !attachment || !detailsField)
				return;
			dateTimeChooser.selectedDate = FieldUtils.getFieldValue(attachment, detailsField.field) as Date;
		}
		
		private function onFocusOut(event:FocusEvent):void
		{
			editingDateField = false;
			invalidateSkinState();
			invalidateProperties();
		}
		
		private function onDataChange(event:FlexEvent):void
		{
			if (!detailsField || !dateTimeChooser || !dateTimeChooser.selectedDate)
				return;
			FieldUtils.setFieldValue(attachment, detailsField.field, dateTimeChooser.selectedDate);
		}
		
		/**
		 *  Get the ideal width for this card item. Usually this is the measured width of the text plus some padding.
		 * 
		 *  @param field The DetailsField for this card item. This method can be called before the DetailsField is set,
		 *  which is why this is passed as an argument.
		 *  @param measureField A field on the stage appropriate for measuring text.
		 *  @return Width of the text plus some padding.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getCalculatedWidth(field:DetailsField, measureField:UIFTETextField = null):Number
		{
			if (!field)
				return 0;
			var calcWidth:Number = 0;
			var text:String = FieldUtils.getTextFromField(attachment, field.field);
			if (text == null || !text.length)
				text = MINIMUM_FIELD_DATA;
			if (oneLineTextEntry) {
				if (measureField != null) {
					measureField.multiline = false;
					calcWidth = measureField.getUITextFormat().measureText(text).width;
				} else {
					calcWidth = oneLineTextEntry.measureText(text).width;
				}
			} else if (manyLinesTextEntry) {
				if (measureField != null) {
					measureField.multiline = true;
					calcWidth = measureField.getUITextFormat().measureText(text).width;
				} else {
					calcWidth = manyLinesTextEntry.measureText(text).width;
				}
				// measureText erroneously returns the width of only the first line
				// and if the first line is a CR, the width is returned as zero.
				// This checks for that condition and corrects it by returning
				// a width of 1 which will be converted to the minimum column
				// width by the caller (in Details.as).
				if (calcWidth == 0) {
					var firstChar:String = text.charAt();
					if (firstChar == "\n" || firstChar == "\r") {
						calcWidth = 1;
					}
				}
			}
//			trace("Calculated item width for '" + text + "' is: " + calcWidth);
			if (calcWidth == 0)
				return Number.NaN;
			calcWidth = calcWidth * DETAILS_CARD_ITEM_FUDGE_FACTOR + DETAILS_CARD_ITEM_PADDING;
			return calcWidth;
		}
		
		/**
		 *  Called when the sort order of the IAttachment fields change.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function updateSortState():void
		{
			sortStateChanged = true;
			invalidateProperties();
		}
		
				
		// IListItemRenderer(IDataRenderer) implementation
		
		/**
		 *  The DetailsField data object being rendered by this component, set through the public
		 *  data setter.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public var detailsField:DetailsField = null;
		
		/**
		 *  The data object to be rendered. Only DetailsField objects are supported.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get data():Object
		{
			return detailsField;
		}
		
		/**
		 *  @private
		 */	
		public function set data(val:Object):void
		{
			if (val == detailsField)
				return;
			
			// Anything to clean up for the old field (before it changes?)
			if (detailsField) {
			}
			
			// New field value
			detailsField = val as DetailsField;
			
			// A different item may be shown now that the field changed
			invalidateSkinState();
			// Update the skin and fields based on the detailsField change
			detailsFieldChanged = true;
			invalidateProperties();
		}
		
	}
}