/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.details
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.lists.CardList;
	
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.geom.Rectangle;
	import flash.ui.Keyboard;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.core.ClassFactory;
	import mx.events.ResizeEvent;
	
	import spark.events.RendererExistenceEvent;
	import spark.layouts.supportClasses.DropLocation;
	
	/**
	 *  The DetailsCardList class extends CardList to implement the 
	 *  contents of the Details navigator. 
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */	
	public class DetailsCardList extends CardList
	{
		/**
		 *  Constructor.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function DetailsCardList()
		{
			super();
			itemRenderer = new ClassFactory(DetailsCard);
			hasFocusableChildren = true;
			tabFocusEnabled = true;
	
			useDragDropReordering = true;
			
			// DetailsCardList doesn't generally need dynamic thumbs, so default enableIdle to false
			enableIdle = false;
		}
		
		private var _details:Details;

		/**
		 *  The main Details class.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function get details():Details
		{
			return _details;
		}

		/**
		 *  @private
		 */
		public function set details(value:Details):void
		{
			if (value == _details)
				return;
			_details = value;
			if (details && layout) {
				details.horizontalScrollPosition = layout.horizontalScrollPosition;
				if (layout is DetailsLayout) {
					(layout as DetailsLayout).details = _details;
				}
			}
			
			// Update existing detail cards with the new details value
			// (I don't think this would ever happen, but just in case)
			updatePropertyForAllCards("details", details);
		}
		
		private var horizontalWatcher:ChangeWatcher;
		private function horizontalScrollPositionChanged(value:Number):void
		{
			if (details) {
				details.horizontalScrollPosition = value;
			}
		}
		
		/**
		 *  Adds listeners and watchers that are automatically unwatched and unlistened.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		override public function addListenAndWatch():void
		{
			super.addListenAndWatch();
			addListener(this, MouseEvent.CLICK, onListClick);
		}

		/** Called when a skin has added a skin part for this instance.
		 */
		override protected function partAdded(partName:String, instance:Object):void
		{
			super.partAdded(partName, instance);
			switch (instance) {
				case dataGroup:
					if (layout)
						addWatcher(BindingUtils.bindSetter(horizontalScrollPositionChanged, layout, "horizontalScrollPosition"));
					if (layout is DetailsLayout) {
						(layout as DetailsLayout).details = details;
					}
					addListener(dataGroup, RendererExistenceEvent.RENDERER_ADD, rendererAddHandler);
					break;
				
				default:
					break;
			}
		}
		
		/**
		 *  @private
		 *  Called when an item has been added to this component.
		 */
		private function rendererAddHandler(event:RendererExistenceEvent):void
		{
			var card:DetailsCard = event.renderer as DetailsCard;
			if (!card)
				return;
			card.details = details;
		}

		protected function onListClick(event:MouseEvent):void
		{
			if ((event.target == this) || (event.target == skin) || (event.target == dataGroup)) {
				// Clear the selection if the user explicitly clicking on a non-interactive item
				collectionMgr.setSelectedItems([]);
			}
		}
		
		private function updatePropertyForAllCards(propName:String, propValue:*):void
		{
			// TODO If we're using a virtual layout, this may not set the prop for all renderers
			for (var index:int = 0; index < dataGroup.numElements; ++index) {
				var card:DetailsCard = dataGroup.getElementAt(index) as DetailsCard;
				if (!card)
					continue;
				card[propName] = propValue;
			}
		}
		
		/**
		 *  Called when the sort order is changed by clicking in the header of a column.
		 *  This method calls updateSortIndicators for each DetailsCard for this card list.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function updateSortIndicators():void
		{
			if (!dataGroup)
				return;
			for (var index:int = 0; index < dataGroup.numElements; ++index) {
				var card:DetailsCard = dataGroup.getElementAt(index) as DetailsCard;
				if (!card)
					continue;
				card.updateSortIndicators();
			}
		}
		
		/**
		 *  Returns focus to the details card list.
		 *  This is usually called after an action in the header has been processed (sort or resize for example).
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function returnFocusToDetailsCardList():void
		{
			focusManager.setFocus(this);
		}
		
		override protected function initializeSelectionFromHost():Boolean
		{
			var returnValue:Boolean = super.initializeSelectionFromHost();
			returnFocusToDetailsCardList();
			return returnValue;
		}
		
		override protected function keyDownHandler(event:KeyboardEvent):void
		{
			if ( (event.keyCode == Keyboard.LEFT) || (event.keyCode == Keyboard.RIGHT))
			{
				event.preventDefault();
				return;
			}
	 		
			if ( (event.keyCode == Keyboard.SPACE) && event.ctrlKey)
			{
				// The routine setSelectedIndicesWithChangeDispatch depends on a new Vector being passed in i.e. not a
				// modified selectedIndices.
				var newList:Vector.<int> = new Vector.<int>();
				
					// This key combo is supposed to toggle the row between being in and out of the selection set
				var indx:int = selectedIndices.indexOf( caretIndex);

				if ( indx == -1)
					newList.push( caretIndex);  // add to selection
				for (var i:int=0; i<selectedIndices.length; i++)
					if ( i != indx)				// don't add to newList i.e. this is a removal
						newList.push( selectedIndices[i]);
				
				setSelectedIndicesWithChangeDispatch( newList);
				return;
			}
			
			super.keyDownHandler( event);
		}
		
		override protected function onResize(event:ResizeEvent):void
		{
			// do nothing by design
		}
		
		/**
		 *  Gets the folder item at the drop location.
		 * 
		 *  @param dropLocation The location where the dropIndicator icon is to be displayed.
		 *  @return IAttachment of the folder item (or null)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		override public function getFolderItemAtDragLocation(dropLocation:DropLocation):IAttachment
		{
			var dropLocationAtt:IAttachment = super.getFolderItemAtDragLocation(dropLocation);
			
			if (!dropLocationAtt)
				return null;
			// If its folder, see where the drop point is:
			var elementBounds:Rectangle = layout.getElementBounds(dataProvider.getItemIndex(dropLocationAtt));
			if (!elementBounds)
				return null;
			var percentage:Number = Math.round(dropLocation.dropPoint.y - elementBounds.y)/elementBounds.height;
			if (percentage < 0.2 || percentage > 0.8)
				return null; // to allow dropping between folders.
			
			return dropLocationAtt;
		}
		
	}
}
