/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.supportClasses
{
	import com.adobe.portfolio.utils.LogHelper;
	
	import flash.events.IEventDispatcher;
	import flash.utils.Dictionary;
	
	import mx.utils.UIDUtil;
	

	/**
	 * The AutoListen class both adds and tracks event listeners and provides the ability to remove all listeners with a single call.
	 * 
  	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class AutoListen
	{
		
		private var listeners:Dictionary;
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function AutoListen()
		{
			listeners = new Dictionary(true);
		}

		/**
		 *  Adds an event listener to the specified event dispatcher and adds it to the list that <code>removeAllListeners</code> will reference when called.
		 * 
		 *  @param eventDispatcher The event dispatcher to which the listener is being added.
		 *  @param type The type of event.
		 *  @param listener The listener function that processes the event. This function must accept an event object as its only parameter and must return nothing.
		 *  @param useCapture Determines whether the listener works in the capture phase or the target and bubbling phases.
		 *  @param priority The priority level of the event listener.
		 * 
		 *  @see IEventDispatcher#addEventListener
		 * 
 		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function addListener(eventDispatcher:IEventDispatcher, type:String, listener:Function, 
									useCapture:Boolean = false, priority:int = 0):void {
			if (!eventDispatcher) {
				CONFIG::logging{
					LogHelper.logger.info("AutoListen.addListener:  null listener: " + type);
				}
				return;
			}
			eventDispatcher.addEventListener(type, listener, useCapture, priority, true);
			var key:String = getKey(eventDispatcher, type, listener);
			listeners[key] = {eventDispatcher:eventDispatcher, type:type, listener:listener};
		}
		
		private function getKey(eventDispatcher:IEventDispatcher, type:String, listener:Function):String {
			return "" + UIDUtil.getUID(eventDispatcher) + " type" + type + "listener" + UIDUtil.getUID(listener);
		}
		
		/**
		 *  Remove all listeners of the specified type from an eventDispatcher.  The eventDispatcher must 
		 *  have been previously remembered by calling <code>addListener</code>
		 * 
		 *  @param eventDispatcher The event dispatcher from which listeners will be removed.
		 *  @param type The type of event.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function removeListener(eventDispatcher:IEventDispatcher, type:String = null):void {
			if (!eventDispatcher)
				return;
			
			for  (var key:Object in listeners) {
				if (!key)
					continue;
				var o:Object = listeners[key];
				if (o.eventDispatcher == eventDispatcher && (!type || o.type == type)) {
					eventDispatcher.removeEventListener(o.type, o.listener);
					delete listeners[key];
				}
			}
		}
		
		/**
		 *  Remove all listeners that were added through addListener.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function removeAllListeners():void {
			for (var key:Object in listeners) {
				if (!key)
					continue;
				var o:Object = listeners[key];
				o.eventDispatcher.removeEventListener(o.type, o.listener);
			}
			listeners = new Dictionary();
		}
	}
}
