/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.supportClasses
{
	import com.adobe.portfolio.api.ICollection;
	import com.adobe.portfolio.components.Background;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.PropertyManager;
	import com.adobe.portfolio.managers.ResourceManager;
	import com.adobe.portfolio.managers.StringManager;
	import com.adobe.portfolio.properties.CanvasTemplate;
	import com.adobe.portfolio.properties.CommonProperties;
	import com.adobe.portfolio.properties.PropertyContext;
	import com.adobe.portfolio.properties.PropertyListValue;
	import com.adobe.portfolio.properties.PropertyValue;
	import com.adobe.portfolio.utils.LogHelper;
	
	import flash.display.DisplayObject;
	import flash.display.DisplayObjectContainer;
	import flash.display.Loader;
	import flash.display.LoaderInfo;
	import flash.events.Event;
	import flash.events.FocusEvent;
	import flash.events.IEventDispatcher;
	import flash.events.IOErrorEvent;
	import flash.events.MouseEvent;
	import flash.events.SecurityErrorEvent;
	import flash.geom.Point;
	import flash.net.URLLoader;
	import flash.utils.ByteArray;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.collections.ArrayCollection;
	import mx.core.ClassFactory;
	import mx.events.FlexEvent;
	import mx.managers.IFocusManagerComponent;
	import mx.utils.ColorUtil;
	import mx.utils.UIDUtil;
	
	import spark.components.Button;
	import spark.components.List;
	import spark.events.RendererExistenceEvent;
	import spark.layouts.BasicLayout;
	
	[SkinState("authoringNormal")]
	[SkinState("authoringOver")]
	[SkinState("authoringSelected")]
	[SkinState("authoringOverAndSelected")]
	/**
	 * The CanvasList class extends List to provide an area for a user to add, remove, and edit
	 * text and image items.
	 */ 
	public class CanvasList extends List implements IListenAndWatch
	{
		/**
		 * A string literal representing the element name used in the XML
		 * representation of a CanvasList.
		 */
		public static const XML_ELEMENT_NAME:String = "CanvasItems";

		/**
		 * A string literal representing the normal consumption state of this component's skin.
		 */
		public static const SKIN_STATE_NORMAL:String = "normal";

		/**
		 * A string literal representing the disabled consumption state of this component's skin.
		 */
		public static const SKIN_STATE_DISABLED:String = "disabled";
		
		/**
		 * A string literal representing the normal authoring state of this component's skin.
		 */
		public static const SKIN_STATE_AUTHORING_NORMAL:String = "authoringNormal";
		
		/**
		 * A string literal representing the over authoring state of this component's skin.
		 */
		public static const SKIN_STATE_AUTHORING_OVER:String = "authoringOver";
		
		/**
		 * A string literal representing the selected authoring state of this component's skin.
		 */
		public static const SKIN_STATE_AUTHORING_SELECTED:String = "authoringSelected";
		
		/**
		 * A string literal representing the overAndSelected authoring state of this component's skin.
		 */
		public static const SKIN_STATE_AUTHORING_OVER_AND_SELECTED:String = "authoringOverAndSelected";
				
		/**
		 * A string literal representing the name of the addImageButton skin part.
		 */
		public static const SKIN_PART_ADD_IMAGE_BUTTON:String = "addImageButton";
		
		/**
		 * A string literal representing the name of the addTextButton skin part.
		 */
		public static const SKIN_PART_ADD_TEXT_BUTTON:String = "addTextButton";
		
		/**
		 * A string literal representing the name of the removeButton skin part.
		 */
		public static const SKIN_PART_REMOVE_BUTTON:String = "removeButton";
		
		/**
		 * A string literal representing the name of the custom preset layout.
		 */
		public static const CUSTOM_PRESET_NAME:String = "custom";
		
		/**
		 * A string literal representing the name of the blank preset layout.
		 */
		public static const BLANK_PRESET_NAME:String = "blank";
		
		private var mouseIsOver:Boolean = false;
		
		[Bindable]
		private var canvasItems:ArrayCollection = new ArrayCollection();
		
		private var templateWasSet:Boolean = false;
		
		/**
		 * Default constructor
		 */
		public function CanvasList()
		{
			super();
			
			useVirtualLayout = false;
			
			// Set default item renderer; this can be overriden in mxml or AS
			itemRenderer = new ClassFactory(CanvasItem);
			
			// Set default layout; this can be overriden in mxml or AS
			layout = new BasicLayout();
			layout.clipAndEnableScrolling = false;
			
			dataProvider = canvasItems;
			
			addEventListener(FlexEvent.CREATION_COMPLETE,
				function (event:FlexEvent):void
				{
					removeEventListener(FlexEvent.CREATION_COMPLETE, arguments.callee);
					addListenAndWatch();
				}
			);
			
			hasFocusableChildren = true;
		}
		
		/** The navigator's CollectionManager instance. All host interaction should go through this
		 *  instance, rather than the INavigatorHost directly.
		 */
		protected const collectionManager:CollectionManager = CollectionManager.instance;
		
		
		[SkinPart(required="false")]
		/**
		 * addImageButton is the button control used to add an image item to this canvas list.
		 * It is an optional skin part, so not necessarily implemented for all skins.
		 */
		public var addImageButton:Button;
		
		[SkinPart(required="false")]
		/**
		 * addTextButton is the button control used to add a text item to this canvas list.
		 * It is an optional skin part, so not necessarily implemented for all skins.
		 */
		public var addTextButton:Button;
		
		[SkinPart(required="false")]
		/**
		 * removeButton is the button control used to remove selected items from this canvas list.
		 * It is an optional skin part, so not necessarily implemented for all skins.
		 */
		public var removeButton:Button;
		
		[Bindable]
		/**
		 * The background color of this canvas list.
		 */
		public var backgroundColor:uint = 0x00000000;
		
		[Bindable]
		/**
		 * The background opacity of this canvas list.
		 */
		public var backgroundOpacity:Number = 0;
		
		[Bindable]
		/**
		 * The background gradient type of this canvas list.
		 * 
		 * @see GradientType
		 */
		public var backgroundGradientType:String = "none";
		
		[Bindable]
		/**
		 * The secondary color to be used for radial and linear gradient types.
		 */
		public var backgroundGradientColor:uint = 0;
		
		private var _backgroundGradientIntensity:Number = 70;
		
		/**
		 * The gradient intensity used to determine the secondary color for radial and linear gradient types.
		 * 
		 * @see updateGradientColor
		 */
		protected function get backgroundGradientIntensity():Number
		{
			return _backgroundGradientIntensity;
		}
		
		
		private var _widthIfPresent:Number;
		[Bindable]
		/**
		 * widthIfPresent determines the width of this canvas list when the shouldDisplayCanvas is true;
		 * the width is zero otherwise.
		 */
		public function get widthIfPresent():Number
		{
			return _widthIfPresent;
		}
		/**
		 * @private
		 */
		public function set widthIfPresent(value:Number):void
		{
			_widthIfPresent = value;
			updateActualWidth()
		}
		
		private var _heightIfPresent:Number;
		[Bindable]
		/**
		 * heightIfPresent determines the height of this canvas list when the shouldDisplayCanvas is true;
		 * the height is zero otherwise.
		 */
		public function get heightIfPresent():Number
		{
			return _heightIfPresent;
		}
		public function set heightIfPresent(value:Number):void
		{
			_heightIfPresent = value;
			updateActualHeight()
		}
		
		/**
		 * Sets the width of this canvas list based on widthIfPresent and shouldDisplayCanvas.
		 */
		protected function updateActualWidth():void
		{
			if (isNaN(widthIfPresent)) {
				// Don't change the width if we're not using widthIfPresent
				return;
			}
			width = shouldDisplayCanvas() ? widthIfPresent : 0;
		}

		/**
		 * Sets the height of this canvas list based on heightIfPresent and shouldDisplayCanvas.
		 */
		protected function updateActualHeight():void
		{
			if (isNaN(heightIfPresent)) {
				// Don't change the height if we're not using heightIfPresent
				return;
			}
			height = shouldDisplayCanvas() ? heightIfPresent : 0;
		}
		
		private var _dirty:Boolean = false;
		/**
		 * True if this canvas has been changed and needs to be saved; false otherwise.
		 */
		public function get dirty():Boolean
		{
			return _dirty;
		}
		/**
		 * @private
		 */
		public function set dirty(value:Boolean):void
		{
			if (_dirty != value) {
				if (value && !templateWasSet) {
					return;
				}
				_dirty = value;
				if (_dirty) {
					collectionManager.setDirty();
				}
			}
		}
		
		private var _dataURL:String;
		/**
		 * The url location of the XML representing this canvas list, or null if it has
		 * no url representation.
		 */
		public function get dataURL():String
		{
			return _dataURL;
		}
		public function set dataURL(value:String):void
		{
			_dataURL = value;
		}
		
		private var _swfClass:String;
		[Bindable]
		/**
		 * The name of the Flash asset to be drawn over the background of this canvas list.
		 * 
		 * @see FlashSkinManager
		 */
		public function get swfClass():String
		{
			return _swfClass;
		}
		/**
		 * @private
		 */
		public function set swfClass(value:String):void
		{
			_swfClass = value;
		}
		
		/**
		 * A boolean indicating whether a property context should be auto-generated for
		 * this canvas list.
		 * 
		 * @see PropertyContext
		 */
		public var autoGenerateProperties:Boolean = false;
		
		/**
		 * The label string to be used when auto-generating a property context for this
		 * canvas list.
		 * 
		 * @see PropertyContext
		 */
		public var autoGeneratedLabel:String = "CANVAS_PROPERTIES";
		
		/**
		 * The key for a localized string to be used for the label string when auto-generating
		 * a property context for this canvas list.
		 * 
		 * @see StringManager
		 */
		public var autoGeneratedLabelKey:String = "canvasProperties";
		
		/**
		 * Returns true if this canvas list should be displayed; false otherwise.
		 * Canvas lists should generally be displayed when they contain valid items, or when
		 * in authoring mode.
		 */
		protected function shouldDisplayCanvas():Boolean
		{
			return (collectionManager.authoringMode || (canvasItems.length > 0));
		}
		
		private var autoGeneratedContext:PropertyContext;
		/**
		 * Returns an auto-generated property context that can be used to represent this canvas
		 * lists properties.
		 */
		protected function generatePropertyContext():PropertyContext
		{
			var context:PropertyContext = new PropertyContext();
			context.loadFromXML(
				<PropertyContext labelKey={autoGeneratedLabelKey} label={autoGeneratedLabel} uid={UIDUtil.createUID()} global="false">
					<PropertyColor name="backgroundColor" labelKey="canvasBackgroundColor" label="CANVAS_BACKGROUND_COLOR" value={backgroundColor}/>
					<PropertyNumber name="backgroundOpacity" labelKey="canvasBackgroundOpacity" label="CANVAS_BACKGROUND_OPACITY" minValue="0" value={backgroundOpacity} maxValue="100"/>
					<PropertyList name="gradientType" labelKey="backgroundGradientType" label="BACKGROUND_GRADIENT_TYPE" active={backgroundGradientType}>
						<PropertyValue name="none" labelKey="backgroundGradientNone" label="BACKGROUND_GRADIENT_NONE"/>
						<PropertyValue name="linear" labelKey="backgroundGradientLinear" label="BACKGROUND_GRADIENT_LINEAR"/>
						<PropertyValue name="radial" labelKey="backgroundGradientRadial" label="BACKGROUND_GRADIENT_RADIAL"/>
					</PropertyList>
					<PropertyNumber name="gradientIntensity" labelKey="backgroundGradientIntensity" label="BACKGROUND_GRADIENT_INTENSITY" minValue="-100" value={backgroundGradientIntensity} maxValue="100"/>
					<PropertyAction name={CommonProperties.PROPERTY_ADD_TEXT} labelKey="addText" label="ADD_TEXT_BUTTON"/>
					<PropertyAction name={CommonProperties.PROPERTY_ADD_IMAGE} labelKey="addImage" label="ADD_IMAGE_BUTTON"/>
				</PropertyContext>
			);
			return context;
		}
		
		/**
		 * Returns the property representing the list of preset layouts.
		 */
		protected function getTemplateProperty():PropertyListValue
		{
			// Subclasses should override if necessary
			return null;
		}
		
		// Watchers for properties; subclasses (like footer) can override addPropertyBindings
		// to watch alternate properties
		
		/**
		 * A ChangeWatcher used to detect changes to the preset layout list property.
		 * 
		 * @see #addPropertyBindings
		 */
		protected var templateWatcher:ChangeWatcher;
		
		/**
		 * A ChangeWatcher used to detect changes to the background color property.
		 * 
		 * @see #addPropertyBindings
		 */
		protected var backgroundColorWatcher:ChangeWatcher;
		
		/**
		 * A ChangeWatcher used to detect changes to the background opacity property.
		 * 
		 * @see #addPropertyBindings
		 */
		protected var backgroundOpacityWatcher:ChangeWatcher;
		
		/**
		 * A ChangeWatcher used to detect changes to the gradient type property.
		 * 
		 * @see #addPropertyBindings
		 */
		protected var backgroundGradientTypeWatcher:ChangeWatcher;
		
		/**
		 * A ChangeWatcher used to detect changes to the gradient intensity property.
		 * 
		 * @see #addPropertyBindings
		 */
		protected var backgroundGradientIntensityWatcher:ChangeWatcher;
		
		/**
		 * A ChangeWatcher used to detect when the add text property action is invoked by the user.
		 * 
		 * @see #addPropertyBindings
		 */
		protected var addTextWatcher:ChangeWatcher;
		
		/**
		 * A ChangeWatcher used to detect when the add image property action is invoked by the user.
		 * 
		 * @see #addPropertyBindings
		 */
		protected var addImageWatcher:ChangeWatcher;
		
		/**
		 * A ChangeWatcher used to detect when the add swf property action is invoked by the user.
		 * 
		 * @see #addPropertyBindings
		 */
		protected var addSwfWatcher:ChangeWatcher;
		
		private var authoringModeWatcher:ChangeWatcher;
		private var canvasXMLWatcher:ChangeWatcher;
		private var didAddListenAndWatch:Boolean = false;
		
		/**
		 * A function to add event listeners and change watchers required by this canvas list.
		 */
		public function addListenAndWatch():void
		{
			if (didAddListenAndWatch) {
				CONFIG::logging{
					LogHelper.logger.warn("Listeners are already setup");
				}
				return;
			}
			authoringModeWatcher = BindingUtils.bindSetter(setAuthoringMode, collectionManager, "authoringMode");
			if ((dataURL != null) && (dataURL != "")) {
				canvasXMLWatcher = BindingUtils.bindSetter(setCanvasXML, ResourceManager.instance.getResource(dataURL, false), "dataString");
			}
			addPropertyBindings();
			addEventListener(MouseEvent.ROLL_OVER, onMouseRoll, false, 0, true);
			addEventListener(FocusEvent.FOCUS_IN, onFocusInOrOut, false, 0, true);
			addEventListener(FocusEvent.FOCUS_OUT, onFocusInOrOut, false, 0, true);
			addEventListener(RendererExistenceEvent.RENDERER_ADD, onRendererChange, false, 0, true);
			addEventListener(RendererExistenceEvent.RENDERER_REMOVE, onRendererChange, false, 0, true);
			collectionManager.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent, false, 0, true);
			didAddListenAndWatch = true;
		}
		
		/**
		 * A function to clean up event listeners and change watchers required by this canvas list.
		 */
		public function removeListenAndWatch():void
		{
			if (!didAddListenAndWatch) {
				CONFIG::logging{
					LogHelper.logger.warn("Listeners were never setup");
				}
				return;
			}
			if (authoringModeWatcher != null) {
				authoringModeWatcher.unwatch();
			}
			if (canvasXMLWatcher != null) {
				canvasXMLWatcher.unwatch();
				canvasXMLWatcher = null;
			}
			removePropertyBindings();
			removeEventListener(MouseEvent.ROLL_OVER, onMouseRoll);
			removeEventListener(FocusEvent.FOCUS_IN, onFocusInOrOut);
			removeEventListener(FocusEvent.FOCUS_OUT, onFocusInOrOut);
			removeEventListener(RendererExistenceEvent.RENDERER_ADD, onRendererChange);
			removeEventListener(RendererExistenceEvent.RENDERER_REMOVE, onRendererChange);
			collectionManager.removeEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent);
			didAddListenAndWatch = false;
		}
		
		/**
		 * true if property bindings have been setup for this canvas list; false otherwise.
		 */
		protected var addedPropertyBindings:Boolean = false;
		
		/**
		 * A function to add event listeners and change watchers for binding portfolio properties to this
		 * canvas list.
		 */
		protected function addPropertyBindings():void
		{
			// Subclasses should override if necessary
			if (addedPropertyBindings) {
				return;
			}
			if (autoGenerateProperties) {
				if (autoGeneratedContext == null) {
					// Don't bother adding bindings if we haven't even setup the context yet
					return;
				}
				var property:PropertyValue = autoGeneratedContext.getProperty(CommonProperties.PROPERTY_PRESET_LIST);
				if (property != null) {
					templateWatcher = BindingUtils.bindSetter(setCanvasTemplate, property, "activeValue");
				} else {
					// There's no template property, so just pretend it was set so it's dirtied properly
					templateWasSet = true;
				}
				property = autoGeneratedContext.getProperty(CommonProperties.PROPERTY_BACKGROUND_COLOR);
				if (property != null) {
					backgroundColorWatcher = BindingUtils.bindSetter(setBackgroundColor, property, "numValue");
				}
				property = autoGeneratedContext.getProperty(CommonProperties.PROPERTY_BACKGROUND_OPACITY);
				if (property != null) {
					backgroundOpacityWatcher = BindingUtils.bindSetter(setBackgroundOpacity, property, "numValue");
				}
				property = autoGeneratedContext.getProperty(CommonProperties.PROPERTY_GRADIENT_TYPE);
				if (property != null) {
					backgroundGradientTypeWatcher = BindingUtils.bindSetter(setBackgroundGradientType, property, "activeValue");
				}
				property = autoGeneratedContext.getProperty(CommonProperties.PROPERTY_GRADIENT_INTENSITY);
				if (property != null) {
					backgroundGradientIntensityWatcher = BindingUtils.bindSetter(setBackgroundGradientIntensity, property, "numValue");
				}
				property = autoGeneratedContext.getProperty(CommonProperties.PROPERTY_ADD_TEXT);
				if (property != null) {
					addTextWatcher = BindingUtils.bindSetter(addTextAction, property, "value");
				}
				property = autoGeneratedContext.getProperty(CommonProperties.PROPERTY_ADD_IMAGE);
				if (property != null) {
					addImageWatcher = BindingUtils.bindSetter(addImageAction, property, "value");
				}
				property = autoGeneratedContext.getProperty(CommonProperties.PROPERTY_ADD_SWF);
				if (property != null) {
					addSwfWatcher = BindingUtils.bindSetter(addSwfAction, property, "value");
				}
			}
			addedPropertyBindings = true;
		}
		
		/**
		 * A function to clean up event listeners and change watchers for binding portfolio properties to this
		 * canvas list.
		 */
		protected function removePropertyBindings():void
		{
			if (templateWatcher != null) {
				templateWasSet = false;
				templateWatcher.unwatch();
			}
			if (backgroundColorWatcher != null) {
				backgroundColorWatcher.unwatch();
			}
			if (backgroundOpacityWatcher != null) {
				backgroundOpacityWatcher.unwatch();
			}
			if (backgroundGradientTypeWatcher != null) {
				backgroundGradientTypeWatcher.unwatch();
			}
			if (backgroundGradientIntensityWatcher != null) {
				backgroundGradientIntensityWatcher.unwatch();
			}
			if (addTextWatcher != null) {
				addTextWatcher.unwatch();
			}
			if (addImageWatcher != null) {
				addImageWatcher.unwatch();
			}
			if (addSwfWatcher != null) {
				addSwfWatcher.unwatch();
			}
			addedPropertyBindings = false;
		}
		
		/**
		 * Sets the preset layout property list to its custom entry.
		 * 
		 * @see #getTemplateProperty
		 * @see #CUSTOM_PRESET_NAME
		 */
		public function setCustomTemplate():Boolean
		{
			var templateList:PropertyListValue = getTemplateProperty();
			if (templateList == null) {
				return false;
			}
			if ((templateList.activeValue != null) && (templateList.activeValue.name == CUSTOM_PRESET_NAME)) {
				return true;
			}
			for each (var template:PropertyValue in templateList.subvalues) {
				if (template.name == CUSTOM_PRESET_NAME) {
					templateList.activeValue = template;
					return true;
				}
			}
			return false;
		}
		
		/**
		 * Sets the preset layout property list to its blank entry.
		 * 
		 * @see #getTemplateProperty
		 * @see #BLANK_PRESET_NAME
		 */
		public function setBlankTemplate():Boolean
		{
			var templateList:PropertyListValue = getTemplateProperty();
			if (templateList == null) {
				return false;
			}
			if ((templateList.activeValue != null) && (templateList.activeValue.name == BLANK_PRESET_NAME)) {
				return true;
			}
			for each (var template:PropertyValue in templateList.subvalues) {
				if (template.name == BLANK_PRESET_NAME) {
					templateList.activeValue = template;
					return true;
				}
			}
			return false;
		}
		
		/**
		 * Sets the list of items to match those from a preset list.
		 * 
		 * @see setCanvasTemplate
		 */
		public function applyPresetPositions(presetDataSet:Vector.<CanvasItemData>):void
		{
			if (!collectionManager.permissionCheck("Document", "Modify")) {
				return;
			}
			
			var newItems:Array = new Array();
			var textCandidates:Array = new Array();
			var imageCandidates:Array = new Array();
			
			// We collect the existing items in reverse order so that they can be more easily popped and pushed
			// from their arrays
			var existingItem:CanvasItemData;
			for (var existingIndex:int = canvasItems.length-1; existingIndex >= 0; --existingIndex) {
				existingItem = canvasItems[existingIndex] as CanvasItemData;
				if (existingItem == null) {
					continue;
				}
				if (existingItem.type == CanvasItemData.TEXT_TYPE) {
					textCandidates.push(existingItem);
				} else if (existingItem.type == CanvasItemData.IMAGE_TYPE) {
					imageCandidates.push(existingItem);
				} else {
					CONFIG::logging{
						LogHelper.logger.warn("Unrecognized existing CanvasItemData type: "+existingItem.type);
					}
				}
			}
			
			for (var i:uint = 0; i < presetDataSet.length; ++i) {
				var newItem:CanvasItemData = null;
				var currentPreset:CanvasItemData = presetDataSet[i];
				if (currentPreset == null) {
					continue;
				}
				var candidateArray:Array;
				if (currentPreset.type == CanvasItemData.TEXT_TYPE) {
					candidateArray = textCandidates;
				} else if (currentPreset.type == CanvasItemData.IMAGE_TYPE) {
					candidateArray = imageCandidates;
				} else {
					CONFIG::logging{
						LogHelper.logger.warn("Unrecognized preset CanvasItemData type: "+currentPreset.type);
					}
				}
				if ((candidateArray != null) && (candidateArray.length > 0)) {
					newItem = candidateArray.pop() as CanvasItemData;
					if (!newItem.setValuesFromData(currentPreset)) {
						candidateArray.push(newItem);
						newItem = null;
					} else {
						newItem.visible = true;
					}
				}
				if (newItem == null) {
					newItem = new CanvasItemData(this);
					newItem.type = currentPreset.type;
					newItem.visible = true;
					newItem.setValuesFromData(currentPreset);
				}
				newItems.push(newItem);
			}
			
			// We keep the unused candidates around (but not visible) so that their data doesn't get lost
			// when switching between different presets
			for each (existingItem in textCandidates) {
				existingItem.visible = false;
				newItems.push(existingItem);
			}
			for each (existingItem in imageCandidates) {
				existingItem.visible = false;
				newItems.push(existingItem);
			}
			
			canvasItems.source = newItems;
		}
		
		/**
		 * A function invoked when authoring mode changes.
		 */
		protected function setAuthoringMode(value:Boolean):void
		{
			if (!value) {
				selectedIndex = -1;
				// TODO disallow selection?
			}
			
			// Trigger a non-editing state
			invalidateSkinState();
			
			if (focusIn) {
				takeContext();
			}
			updateActualWidth();
			updateActualHeight();
		}
		
		/**
		 * Sets the list of items to match those from a preset layout.
		 * 
		 * @see CanvasTemplate
		 */
		protected function setCanvasTemplate(value:PropertyValue):void
		{
			if (value == null) {
				return;
			}
			var shouldBeCustom:Boolean = false;
			var presetDataSet:Vector.<CanvasItemData> = new Vector.<CanvasItemData>();
			
			switch (value.name) {
				case CUSTOM_PRESET_NAME:
					// Keep items as they are
					shouldBeCustom = true;
					break;
				default:
					if (value is CanvasTemplate) {
						presetDataSet = (value as CanvasTemplate).canvasItems;
					}
					break;
			}
			
			if (!shouldBeCustom) {
				// We only want to apply new values if we're switching to a non-custom preset
				applyPresetPositions(presetDataSet);
			}
			
			// Don't dirty ourselves if we're just initializing the template
			if (templateWasSet) {
				dirty = true;
			} else {
				templateWasSet = true;
			}
		}
		
		/**
		 * A function invoked when the background color property changes
		 */
		protected function setBackgroundColor(value:Number):void
		{
			backgroundColor = value;
			updateGradientColor();
		}
		
		/**
		 * A function invoked when the background opacity property changes
		 */
		protected function setBackgroundOpacity(value:Number):void
		{
			backgroundOpacity = value;
		}
		
		/**
		 * A function invoked when the background gradient type property changes
		 */
		protected function setBackgroundGradientType(value:PropertyValue):void
		{
			if (value == null) {
				return;
			}
			backgroundGradientType = value.name;
		}
		
		/**
		 * A function invoked when the background gradient intensity property changes
		 */
		protected function setBackgroundGradientIntensity(value:Number):void
		{
			_backgroundGradientIntensity = value;
			updateGradientColor();
		}
		
		/**
		 * Updates backgroundGradientColor according to the current state of the canvas list.
		 */
		protected function updateGradientColor():void
		{
			// Gradients require a second color, which is a delta intensity of the base color
			var brightnessChange:Number = -backgroundGradientIntensity;
			brightnessChange = Math.min(brightnessChange, 100);
			brightnessChange = Math.max(brightnessChange, -100);
			backgroundGradientColor = ColorUtil.adjustBrightness2(backgroundColor, brightnessChange);
		}
		
		/**
		 * Adds a text item to the canvas list.
		 * This function is suitable for binding to a property action value.
		 * 
		 * @see PropertyAction
		 */
		protected function addTextAction(value:uint):void
		{
			if (value != 0) {
				addTextItem();
			}
		}
		
		/**
		 * Adds an image item to the canvas list.
		 * This function is suitable for binding to a property action value.
		 * 
		 * @see PropertyAction
		 */
		protected function addImageAction(value:uint):void
		{
			if (value != 0) {
				addImageItem();
			}
		}
		
		/**
		 * Adds a swf item to the canvas list.
		 * This function is suitable for binding to a property action value.
		 * 
		 * @see PropertyAction
		 */
		protected function addSwfAction(value:uint):void
		{
			if (value != 0) {
				addSwfItem();
			}
		}
		
		/**
		 * Initializes this canvas list from an XML representation.
		 */
		protected function setCanvasXML(value:String):void
		{
			if ((value == null) || (value == "")) {
				return;
			}
			if (canvasXMLWatcher != null) {
				canvasXMLWatcher.unwatch();
				canvasXMLWatcher = null;
			}
			
			// Make sure the appropriate whitespace is maintiained between textflow elements
			var oldIgnoreWhitespace:Boolean = XML.ignoreWhitespace;
			var oldPrettyPrinting:Boolean = XML.prettyPrinting;
			XML.ignoreWhitespace = false;
			XML.prettyPrinting = false;
			
			var valueXML:XML = new XML(value);
			
			// Reset whitespace settings to their original values
			XML.ignoreWhitespace = oldIgnoreWhitespace;
			XML.prettyPrinting = oldPrettyPrinting;
			
			canvasItems.removeAll();
			var items:XMLList = valueXML.elements(CanvasItemData.XML_ELEMENT_NAME);
			if (items != null) {
				var subElements:XMLList = valueXML.elements();
				for each (var subElement:XML in subElements) {
					var canvasItemData:CanvasItemData = new CanvasItemData(this);
					if (!canvasItemData.loadFromXML(subElement)) {
						continue;
					} 
					canvasItems.addItem(canvasItemData);
				}
			}
			if (autoGenerateProperties) {
				var numberString:String = valueXML.@backgroundColor;
				var numberValue:Number = parseInt(numberString);
				if (!isNaN(numberValue)) {
					backgroundColor = numberValue;
				}
				numberString = valueXML.@backgroundOpacity;
				numberValue = parseFloat(numberString);
				if (!isNaN(numberValue)) {
					backgroundOpacity = numberValue;
				}
				
				backgroundGradientType = valueXML.@backgroundGradientType;
				
				numberString = valueXML.@backgroundGradientIntensity;
				numberValue = parseFloat(numberString);
				if (!isNaN(numberValue)) {
					_backgroundGradientIntensity = numberValue;
				}
				var templateName:String = valueXML.@template;
				if (templateName == CUSTOM_PRESET_NAME) {
					// This still falls under the intialization category, so don't let this dirty the document
					templateWasSet = false;
					
					setCustomTemplate();
				}
			}
			addPropertyBindings();
			invalidateSkinState();
		}
		
		/**
		 * Event listener for rollover events on this canvas list.
		 */
		protected function onMouseRoll(event:MouseEvent):void
		{
			mouseIsOver = (event.type == MouseEvent.ROLL_OVER);
			if (collectionManager.authoringMode) {
				// We don't care about rollover unless we're in authoring mode
				invalidateSkinState();
			}
		}
		
		private var focusIn:Boolean = false;
		private function onFocusInOrOut(event:FocusEvent):void
		{
			focusIn = (event.type == FocusEvent.FOCUS_IN);
			if ((event.target != this) && (event.target != skin)) {
				// This is an indirect event, so the only thing we might do is clear selection
				if (!focusIn && (focusManager != null)) {
					clearSelectionPendingFocus();
				}
				return;
			}
			if (focusIn) {
				if (collectionManager.authoringMode) {
					// We have focus, and we can edit, so set the current property context to canvas
					takeContext()
				}
			} else if ((focusManager.getFocus() == null) && (stage != null)) {
				// If nothing else actually got focus, then wait until something does
				stage.addEventListener(FocusEvent.FOCUS_IN, onStageFocusIn, false, 0, true);
			} else {
				loseContext();
			}
		}
		
		private function clearSelectionPendingFocus():void
		{
			var currentFocus:DisplayObject = (focusManager == null) ? null : focusManager.getFocus() as DisplayObject;
			var focusIsAChild:Boolean = false;
			while (currentFocus != null) {
				if ((currentFocus == this) || (currentFocus == skin)) {
					focusIsAChild = true;
					break;
				}
				if (currentFocus is DisplayObjectContainer) {
					currentFocus = currentFocus.parent;
				} else {
					break;
				}
			}
			if (!focusIsAChild) {
				selectedIndex = -1;
			}
		}
		
		private function onStageFocusIn(event:FocusEvent):void
		{
			var currentFocus:IFocusManagerComponent = focusManager.getFocus();
			if (currentFocus == null) {
				return;
			}
			if (event.currentTarget is IEventDispatcher) {
				(event.currentTarget as IEventDispatcher).removeEventListener(FocusEvent.FOCUS_IN, onStageFocusIn);
			}
			if ((currentFocus != this) && (currentFocus != skin)) {
				loseContext();
			}
		}
		
		/**
		 * Sets up the property context for this canvas list, auto-generating it
		 * if necessary.
		 */
		protected function setupContext():PropertyContext
		{
			// Subclasses should override if necessary
			
			if (autoGenerateProperties) {
				if (autoGeneratedContext == null) {
					autoGeneratedContext = generatePropertyContext();
				}
				addPropertyBindings();
				return autoGeneratedContext;
			}
			return null;
		}
		
		/**
		 * Sets the property context for this canvas list as the current property context.
		 * 
		 * @see PropertyContext
		 * @see PropertyManager#setCurrentContext
		 */
		public function takeContext():Boolean
		{
			// If the list is taking context, then clear the general selection
			selectedIndex = -1;

			var propertyContext:PropertyContext = setupContext();
			if (propertyContext != null) {
				PropertyManager.instance.commonProperties.setCurrentContext(propertyContext, this);
				return true;
			}
			return false;
		}
		
		/**
		 * Clears the property context for this canvas list as the current property context.
		 * 
		 * @see PropertyContext
		 * @see PropertyManager#setCurrentContext
		 */
		public function loseContext():Boolean
		{
			// If the list is losing context/focus, and a child didn't get it, then clear the general selection
			clearSelectionPendingFocus();
			
			if (PropertyManager.instance.commonProperties != null) {
				PropertyManager.instance.commonProperties.setCurrentContextByName(CommonProperties.CONTEXT_CLEAR, this);
			}
			return true;
		}
		
		private var dataToFocus:CanvasItemData;
		/**
		 * Event listener called when canvas item renderers are added or removed.
		 */
		protected function onRendererChange(event:RendererExistenceEvent):void
		{
			var item:CanvasItem = event.renderer as CanvasItem;
			if (item == null) {
				return
			}
			if (event.type == RendererExistenceEvent.RENDERER_ADD) {
				item.addEventListener(CanvasItem.EVENT_REMOVE, onRemoveItem);
				if ((dataToFocus != null) && (focusManager != null) && (item.canvasData == dataToFocus)) {
					selectedItem = dataToFocus;
					item.takeFocus();
				}
			} else {
				item.removeEventListener(CanvasItem.EVENT_REMOVE, onRemoveItem);
				
				// Stop any playing SWFs when its renderer is removed
				item.playSWF = false;
			}
		}
		
		/**
		 * Event listener invoked when a canvas item dispatches a remove event.
		 */
		protected function onRemoveItem(event:Event):void
		{
			var item:CanvasItem = event.target as CanvasItem;
			if ((item == null) || (item.canvasData == null)) {
				return;
			}
			if (canvasItems.contains(item.canvasData)) {
				canvasItems.removeItemAt(canvasItems.getItemIndex(item.canvasData));
				if (focusManager != null) {
					focusManager.setFocus(this);
				}
				if (canvasItems.length == 0) {
					setBlankTemplate();
				} else {
					setCustomTemplate();
				}
				dirty = true;
			}
		}
		
		/**
		 * Event listener invoked on CollectionManagerEvents.
		 */
		protected function onNavHostEvent(event:CollectionManagerEvent):void
		{
			if ((event.kind != CollectionManagerEvent.NEEDS_SAVE_KIND) || !dirty) {
				return;
			}
			if ((dataURL == null) || (dataURL == "")) {
				LogHelper.logger.error("CanvasList cannot be saved to an unspecified dataURL");
				return;
			}
			
			// Make sure the appropriate whitespace is maintiained between textflow elements
			var oldIgnoreWhitespace:Boolean = XML.ignoreWhitespace;
			var oldPrettyPrinting:Boolean = XML.prettyPrinting;
			XML.ignoreWhitespace = false;
			XML.prettyPrinting = false;
			
			var dirtyXML:XML = new XML("<"+XML_ELEMENT_NAME+"/>");
			if (autoGenerateProperties) {
				var templateProperty:PropertyListValue = getTemplateProperty();
				var templateValue:PropertyValue = (templateProperty != null) ? templateProperty.activeValue : null;
				if (templateValue != null) {
					dirtyXML.@template = templateValue.name;
				}
				dirtyXML.@backgroundColor = backgroundColor;
				dirtyXML.@backgroundOpacity = backgroundOpacity;
				dirtyXML.@backgroundGradientType = backgroundGradientType;
				dirtyXML.@backgroundGradientIntensity = backgroundGradientIntensity;
			}
			for each (var canvasData:CanvasItemData in canvasItems) {
				if ((canvasData == null) || !canvasData.visible) {
					continue;
				}
				var itemXML:XML = canvasData.generateXML();
				if (itemXML == null) {
					continue;
				}
				dirtyXML.appendChild(itemXML);
			}
			var dirtyBytes:ByteArray = new ByteArray();
			dirtyBytes.writeUTFBytes("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
			
			dirtyBytes.writeUTFBytes(dirtyXML.toXMLString());
			
			// Reset whitespace settings to their original values
			XML.ignoreWhitespace = oldIgnoreWhitespace;
			XML.prettyPrinting = oldPrettyPrinting;
			
			ResourceManager.instance.writeResource(dirtyBytes, dataURL);
			dirty = false;
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function getCurrentSkinState():String
		{
			if (collectionManager.authoringMode) {
				return mouseIsOver ? SKIN_STATE_AUTHORING_OVER : SKIN_STATE_AUTHORING_NORMAL;
			}
			return shouldDisplayCanvas() ? SKIN_STATE_NORMAL : SKIN_STATE_DISABLED;
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function partAdded(partName:String, instance:Object):void
		{
			//handlePartChange will call super.partAdded
			handlePartChange(partName, instance, true);
		}
		/** 
		 * @inheritDoc
		 */
		override protected function partRemoved(partName:String, instance:Object):void
		{
			//handlePartChange will call super.partRemoved
			handlePartChange(partName, instance, false);
		}
		
		/** Internal function for adding and removing event listeners for skin parts, depending on
		 *  whether the part is being added or removed for this instance.
		 */
		private function handlePartChange(partName:String, instance:Object, adding:Boolean):void
		{
			var listeners:Array = null;
			var types:Array = null;
			if ((instance != null) && (instance is IEventDispatcher)) {
				var activePart:IEventDispatcher = instance as IEventDispatcher;
				
				// Build up an array of event types and listeners that apply to this object
				switch (partName) {
					case SKIN_PART_ADD_IMAGE_BUTTON:
						types = [ MouseEvent.CLICK ];
						listeners = [ onAddImageButtonClick ];
						if (adding) {
							addImageButton.label = StringManager.instance.addImage;
						}
						break;
					case SKIN_PART_ADD_TEXT_BUTTON:
						types = [ MouseEvent.CLICK ];
						listeners = [ onAddTextButtonClick ];
						if (adding) {
							addTextButton.label = StringManager.instance.addText;
						}
						break;
					case SKIN_PART_REMOVE_BUTTON:
						types = [ MouseEvent.CLICK ];
						listeners = [ onRemoveButtonClick ];
						if (adding) {
							removeButton.label = StringManager.instance.remove;
						}
						break;
				}
				
				// Add or remove all the relevant event listeners
				var count:uint = (listeners == null) ? 0 : listeners.length;
				if (count > 0) {
					for (var i:uint = 0; i < count; ++i) {
						adding ? 	activePart.addEventListener(types[i], listeners[i]) :
							activePart.removeEventListener(types[i], listeners[i]);
					}
				}
			}
			adding ? super.partAdded(partName, instance) : super.partRemoved(partName, instance);
		}
		
		/**
		 * Adds an image item to the canvas list.
		 * This function is suitable for use as a button event listener.
		 */
		protected function onAddImageButtonClick(event:MouseEvent):void
		{
			addImageItem();
		}
		private var imageURLToValidate:String;
		/**
		 * Adds an image item to the canvas list.
		 * 
		 * @see CanvasItem
		 */
		protected function addImageItem():void
		{
			if (!collectionManager.permissionCheck("Document", "Modify")) {
				return;
			}
			
			var hostCollection:ICollection = (CollectionManager.instance.host != null) ? CollectionManager.instance.host.collection : null;
			var titleString:String = (hostCollection == null) ? "CANVAS_SELECT_IMAGE_TITLE" : hostCollection.getLocalizedString("canvasSelectImageTitle", "CANVAS_SELECT_IMAGE_TITLE");
			ResourceManager.instance.addResource(
				titleString,
				[ ResourceManager.instance.imageFilter ],
				"navigator/canvas_images/"+UIDUtil.createUID(),
				function (response:Object):void
				{
					var originalURL:String = response["originalURL"];
					if (!response["success"] as Boolean) {
						LogHelper.logger.error("addResource failed: "+originalURL);
						return;
					}
					imageURLToValidate = response["relativeURL"];
					ResourceManager.instance.getResource(imageURLToValidate, true, onImageBytesLoaded, onImageBytesLoaded, onImageBytesLoaded);
				}
			);
		}

		private function onImageBytesLoaded(event:Event):void
		{
			if (event == null) {
				LogHelper.logger.error("Null event loading imported image");
				addImageAbandon();
				return;
			}
			
			if ((event == null) || (event.type != Event.COMPLETE)) {
				LogHelper.logger.error("Error loading imported image: "+event.toString());
				addImageAbandon();
				return;
			}
			
			var loader:URLLoader = event.target as URLLoader;
			if (loader == null) {
				LogHelper.logger.error("Invalid loader for imported image: "+event.toString());
				addImageAbandon();
				return;
			}
			
			var imageLoader:Loader = new Loader();
			imageLoader.contentLoaderInfo.addEventListener(Event.COMPLETE, onImageLoaded);
			imageLoader.contentLoaderInfo.addEventListener(IOErrorEvent.IO_ERROR, onImageLoaded);
			imageLoader.contentLoaderInfo.addEventListener(SecurityErrorEvent.SECURITY_ERROR, onImageLoaded);
			
			imageLoader.loadBytes(loader.data);
		}
		private function onImageLoaded(event:Event):void
		{
			var loaderInfo:LoaderInfo = event.target as LoaderInfo;
			if ((loaderInfo == null) || (loaderInfo.loader == null)) {
				LogHelper.logger.error("Invalid loader for imported image: "+event.toString());
				addImageAbandon();
				return;
			}
			loaderInfo.removeEventListener(Event.COMPLETE, onImageLoaded);
			loaderInfo.removeEventListener(IOErrorEvent.IO_ERROR, onImageLoaded);
			loaderInfo.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, onImageLoaded);
			
			if ((event == null) || (event.type != Event.COMPLETE)) {
				LogHelper.logger.error("Error loading imported image: "+event.toString());
				loaderInfo.loader.unload();
				addImageAbandon();
				return;
			}

			// This prevents dealing with corrupted images
			if (loaderInfo.loader.width == 0 || loaderInfo.loader.height == 0) {
				LogHelper.logger.error("Zero width/height loading imported image: "+event.toString());
				loaderInfo.loader.unload();
				addImageAbandon();
				return;
			}
			
			// Only allow images (swfs masquerading as images are a security risk)
			switch (loaderInfo.contentType)
			{
				case Background.MIME_TYPE_JPG:
				case Background.MIME_TYPE_GIF:
				case Background.MIME_TYPE_PNG:
					loaderInfo.loader.unload();
					addImageFinalize();
					return;
				
				default:
					LogHelper.logger.error("Invalid mime type loading imported image: "+event.toString());
					loaderInfo.loader.unload();
					addImageAbandon();
					return;
			}
		}

		private function addImageFinalize():void
		{
			var canvasData:CanvasItemData = new CanvasItemData(this);
			canvasData.type = CanvasItemData.IMAGE_TYPE;
			canvasData.imageURL = imageURLToValidate;
			canvasData.width = 200;
			canvasData.left = findHorizontalFit(canvasData.width);
			canvasData.top = 10;
			canvasData.bottom = 10;
			dataToFocus = canvasData;
			canvasItems.addItem(canvasData);
			setCustomTemplate();
			dirty = true;
			imageURLToValidate = null;
		}
		
		private function addImageAbandon():void
		{
			if ((imageURLToValidate == null) || (imageURLToValidate == "")) {
				LogHelper.logger.warn("An unrecognized image import is being abandoned");
				return;
			}
			ResourceManager.instance.deleteResource(imageURLToValidate);
			imageURLToValidate = null;
		}

		private var swfURLToValidate:String;
		/**
		 * Adds an swf item to the canvas list.
		 * 
		 * @see CanvasItem
		 */
		protected function addSwfItem():void
		{
			if (!collectionManager.permissionCheck("Document", "Modify")) {
				return;
			}
			
			var hostCollection:ICollection = (CollectionManager.instance.host != null) ? CollectionManager.instance.host.collection : null;
			var titleString:String = (hostCollection == null) ? "CANVAS_SELECT_SWF_TITLE" : hostCollection.getLocalizedString("canvasSelectSwfTitle", "CANVAS_SELECT_SWF_TITLE");
			ResourceManager.instance.addResource(
				titleString,
				[ ResourceManager.instance.swfFilter ],
				"navigator/canvas_swfs/"+UIDUtil.createUID(),
				function (response:Object):void
				{
					var originalURL:String = response["originalURL"];
					if (!response["success"] as Boolean) {
						LogHelper.logger.error("addResource failed: "+originalURL);
						return;
					}
					swfURLToValidate = response["relativeURL"];
					ResourceManager.instance.getResource(swfURLToValidate, true, onSwfBytesLoaded, onSwfBytesLoaded, onSwfBytesLoaded);
				}
			);
		}
		
		private function onSwfBytesLoaded(event:Event):void
		{
			if (event == null) {
				LogHelper.logger.error("Null event loading imported swf");
				addSwfAbandon();
				return;
			}
			
			if ((event == null) || (event.type != Event.COMPLETE)) {
				LogHelper.logger.error("Error loading imported image: "+event.toString());
				addSwfAbandon();
				return;
			}
			
			var loader:URLLoader = event.target as URLLoader;
			if (loader == null) {
				LogHelper.logger.error("Invalid loader for imported image: "+event.toString());
				addSwfAbandon();
				return;
			}
			
			addSwfFinalize();
		}
		
		private function addSwfFinalize():void
		{
			var canvasData:CanvasItemData = new CanvasItemData(this);
			canvasData.type = CanvasItemData.SWF_TYPE;
			canvasData.swfURL = swfURLToValidate;
			canvasData.width = 200;
			canvasData.left = findHorizontalFit(canvasData.width);
			canvasData.top = 10;
			canvasData.bottom = 10;
			dataToFocus = canvasData;
			canvasItems.addItem(canvasData);
			setCustomTemplate();
			dirty = true;
			swfURLToValidate = null;
		}
		
		private function addSwfAbandon():void
		{
			if ((swfURLToValidate == null) || (swfURLToValidate == "")) {
				LogHelper.logger.warn("An unrecognized swf import is being abandoned");
				return;
			}
			ResourceManager.instance.deleteResource(swfURLToValidate);
			swfURLToValidate = null;
		}
		
		/**
		 * Adds a text item to the canvas list.
		 * This function is suitable for use as a button event listener.
		 */
		protected function onAddTextButtonClick(event:MouseEvent):void
		{
			addTextItem();
		}
		/**
		 * Adds a text item to the canvas list.
		 * 
		 * @see CanvasItem 
		 */
		protected function addTextItem():void
		{
			if (!collectionManager.permissionCheck("Document", "Modify")) {
				return;
			}
			
			var canvasData:CanvasItemData = new CanvasItemData(this);
			canvasData.type = CanvasItemData.TEXT_TYPE;
			canvasData.width = 200;
			canvasData.left = findHorizontalFit(canvasData.width);
			canvasData.top = 10;
			canvasData.bottom = 10;
			dataToFocus = canvasData;
			canvasItems.addItem(canvasData);
			setCustomTemplate();
			dirty = true;
		}
		
		/**
		 * Returns the left-most coordinate suitable for placing a canvas item with the
		 * given width amidst the current canvas items.
		 */
		private function findHorizontalFit(itemWidth:Number):Number
		{
			var returnValue:Number = 10;
			var pointRanges:Array = new Array();
			
			// First build a list of ranges for current items
			for each (var canvasData:CanvasItemData in canvasItems) {
				if ((canvasData == null) || !canvasData.visible) {
					continue;
				}
				
				var newPoint:Point = new Point();
				var didSetY:Boolean = false;
				
				if (!isNaN(canvasData.x)) {
					newPoint.x = canvasData.x;
				} else if (!isNaN(canvasData.left)) {
					newPoint.x = canvasData.left;
					if (!isNaN(canvasData.right)) {
						newPoint.y = width - canvasData.right;
						didSetY = true;
					}
				} else if (!isNaN(canvasData.horizontalCenter)) {
					newPoint.x = ((width - canvasData.width) / 2) + canvasData.horizontalCenter;
				} else if (!isNaN(canvasData.right)) {
					newPoint.y = width - canvasData.right;
					didSetY = true;
					newPoint.x = newPoint.y - canvasData.width;
				}
				if (!didSetY) {
					newPoint.y = newPoint.x + canvasData.width;
				}
				pointRanges.push(newPoint);
			}
			
			// Then sort the list of point ranges
			pointRanges = pointRanges.sort(function (a:Point, b:Point):int
			{
				var xDiff:Number = a.x - b.x;
				if (xDiff != 0) {
					return xDiff;
				}
				return a.y - b.y;
			});
			
			// Now find somewhere to fit
			var targetY:Number = returnValue + itemWidth;
			for each (var point:Point in pointRanges) {
				if (point.x > targetY) {
					return returnValue;
				}
				if (point.y > returnValue) {
					returnValue = point.y + 5;
					targetY = returnValue + itemWidth;
				}
			}
			
			if (targetY > width) {
				// We couldn't find a place to fit, so just overlap on the left
				returnValue = 10;
			}
			return returnValue;
		}
		
		/**
		 * Remoeves selected items from the canvas list.
		 * This function is suitable for use as a button event listener.
		 */
		protected function onRemoveButtonClick(event:MouseEvent):void
		{
			if (!collectionManager.permissionCheck("Document", "Modify")) {
				return;
			}
			
			var itemsToRemove:Vector.<Object> = selectedItems.slice();
			for each (var canvasData:CanvasItemData in itemsToRemove) {
				if (canvasItems.contains(canvasData)) {
					canvasItems.removeItemAt(canvasItems.getItemIndex(canvasData));
					dirty = true;
				}
			}
			if (canvasItems.length == 0) {
				setBlankTemplate();
			}
		}
	}
}
