/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.supportClasses
{
	import flash.events.EventDispatcher;
	import flash.events.IEventDispatcher;
	
	import mx.events.PropertyChangeEvent;
	
	/**
	 * A class containing standard color values used by PortfolioFramework
	 * components and their skins.
	 */
	public class ColorPalette extends EventDispatcher
	{
		/**
		 * Default constructor
		 */
		public function ColorPalette(target:IEventDispatcher = null)
		{
			super(target);
			reset();
		}
		
		/** Resets the color palette to its default values.
		 */
		public function reset():void
		{
			isHighContrast = false;
			background = 0x515151;
			cardBackground = 0xC4C5C5;
			cardBorder = 0x999999;
			primaryText = 0xFFFFFF;
			secondaryText = 0x222222;
		}
		
		/**
		 * Internal representation of the Boolean referenced by the isHighContrast getter and setter.
		 */
		private var _isHighContrast:Boolean = false;
		/**
		 * A boolean indicating whether or not this object represents a color palette in high contrast mode.
		 */
		[Bindable]
		public function get isHighContrast():Boolean
		{
			return _isHighContrast;
		}
		/**
		 * @private
		 */
		protected function set isHighContrast(value:Boolean):void
		{
			_isHighContrast = value;
		}
		
		/** Internal representation of the color referenced by the background getter and setter.
		 */
		private var _background:Number;
		/** The background color, as specified by this palette.
		 */
		[Bindable]
		public function get background():Number
		{
			return _background;
		}
		/**
		 * @private
		 */
		protected function set background(value:Number):void
		{
			_background = value;
		}
		
		/** Internal representation of the color referenced by the cardBackground getter and setter.
		 */
		private var _cardBackground:Number;
		/** The cardBackground color, as specified by this palette.
		 */
		[Bindable]
		public function get cardBackground():Number
		{
			return _cardBackground;
		}
		/**
		 * @private
		 */
		protected function set cardBackground(value:Number):void
		{
			if (value == _cardBackground) {
				return;
			}
			var oldValue:Number = _cardBackground;
			_cardBackground = value;

			// Dispatch an additional event for accentColor, since it shares the same underlying value
			dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "primaryColor", oldValue, value));
		}

		/**
		 * The primaryColor color, as specified by this palette. This property is identical to the
		 * <code>cardBackground</code> property, but is exposed with an alternate name to reflect usage.
		 * 
		 * @see #setColorPalette
		 */
		[Bindable (event="propertyChange")]
		public function get primaryColor():Number
		{
			return _cardBackground;
		}

		/** Internal representation of the color referenced by the cardBorder getter and setter.
		 */
		private var _cardBorder:Number;
		/** The cardBorder color, as specified by this palette.
		 */
		[Bindable]
		public function get cardBorder():Number
		{
			return _cardBorder;
		}
		/**
		 * @private
		 */
		protected function set cardBorder(value:Number):void
		{
			_cardBorder = value;
		}
		
		/** Internal representation of the color referenced by the primaryText getter and setter.
		 */
		private var _primaryText:Number;
		/** The primaryText color, as specified by this palette.
		 */
		[Bindable]
		public function get primaryText():Number
		{
			return _primaryText;
		}
		/**
		 * @private
		 */
		protected function set primaryText(value:Number):void
		{
			_primaryText = value;
		}
		
		/** Internal representation of the color referenced by the secondaryText getter and setter.
		 */
		private var _secondaryText:Number;
		/** The secondaryText color, as specified by this palette.
		 */
		[Bindable]
		public function get secondaryText():Number
		{
			return _secondaryText;
		}
		/**
		 * @private
		 */
		protected function set secondaryText(value:Number):void
		{
			if (value == _secondaryText) {
				return;
			}
			var oldValue:Number = _secondaryText;
			_secondaryText = value;
			
			// Dispatch an additional event for accentColor, since it shares the same underlying value
			dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "accentColor", oldValue, value));
		}
		
		/**
		 * The accentColor color, as specified by this palette. This property is identical to the
		 * <code>secondaryText</code> property, but is exposed with an alternate name to reflect usage.
		 * 
		 * @see #setColorPalette
		 */
		[Bindable (event="propertyChange")]
		public function get accentColor():Number
		{
			return _secondaryText;
		}		
		
		/**
		 * Sets this palette according to the raw color palette object specified by an INavigatorHost.
		 * 
		 * @see INavigatorHost#navigatorColorPalette
		 */
		public function setColorPalette(navHostPalette:Object):void
		{
			if (navHostPalette != null) {
				if (navHostPalette.hasOwnProperty("isHighContrast")) {
					isHighContrast = navHostPalette["isHighContrast"] as Boolean;
				}
				else {
					isHighContrast = false;
				}

				if (navHostPalette.propertyIsEnumerable("background")) {
					background = navHostPalette["background"] as Number;
				}
				if (navHostPalette.propertyIsEnumerable("cardBackground")) {
					cardBackground = navHostPalette["cardBackground"] as Number;
				}
				if (navHostPalette.propertyIsEnumerable("cardBorder")) {
					cardBorder = navHostPalette["cardBorder"] as Number;
				}
				if (navHostPalette.propertyIsEnumerable("primaryText")) {
					primaryText = navHostPalette["primaryText"] as Number;
				}
				if (navHostPalette.propertyIsEnumerable("secondaryText")) {
					secondaryText = navHostPalette["secondaryText"] as Number;
				}
			}
		}
		
		/**
		 * Returns a raw color palette object corresponding to this palette's color values
		 * suitable for updating the INavigatorHost.
		 */
		public function generateRawPalette():Object
		{
			var returnObj:Object = new Object();
			returnObj["background"] = background;
			returnObj["cardBackground"] = cardBackground;
			returnObj["cardBorder"] = cardBorder;
			returnObj["primaryText"] = primaryText;
			returnObj["secondaryText"] = secondaryText;
			return returnObj;
		}
	}
}