/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.supportClasses
{
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.PropertyManager;
	import com.adobe.portfolio.properties.CommonProperties;
	import com.adobe.portfolio.properties.PropertyContext;
	import com.adobe.portfolio.properties.PropertyListValue;
	import com.adobe.portfolio.properties.PropertyValue;
	import com.adobe.portfolio.utils.GlobalizationUtil;
	import com.adobe.portfolio.utils.LogHelper;
	
	import flash.events.FocusEvent;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.ui.Keyboard;
	
	import flashx.textLayout.elements.FlowElement;
	import flashx.textLayout.formats.LineBreak;
	import flashx.textLayout.formats.TextLayoutFormat;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.core.UIComponent;
	import mx.events.FlexEvent;
	import mx.managers.IFocusManager;
	import mx.managers.IFocusManagerComponent;
	
	import spark.components.Button;
	import spark.components.TextArea;
	import spark.components.TextSelectionHighlighting;
	import spark.events.TextOperationEvent;
	
	/**
	 * The normal state for multiline text.
	 */
	[SkinState("normalMultiline")]
	
	/**
	 * The disabled state for multiline text.
	 */
	[SkinState("disabledMultiline")]
	
	/**
	 * The editable state for multiline text.
	 */
	[SkinState("editableMultiline")]

	/**
	 * A text area that is capable of providing both single- and multi-line behavior, as well as
	 * a context for styling text.
	 */
	public class CustomizableTextArea extends TextArea implements IListenAndWatch
	{
		// String literals used to represent the states of this component's skins.
		
		/**
		 * A string literal representing the name of the normal state.
		 */
		static public const STATE_NORMAL:String = "normal";
		
		/**
		 * A string literal representing the name of the disabled state.
		 */
		static public const STATE_DISABLED:String = "disabled";

		/**
		 * A string literal representing the name of the editable state.
		 */
		static public const STATE_EDITABLE:String = "editable";
		
		/**
		 * A string literal representing the name of the normal multiline state.
		 */
		static public const STATE_NORMAL_MULTILINE:String = "normalMultiline";
		
		/**
		 * A string literal representing the name of the disabled multiline state.
		 */
		static public const STATE_DISABLED_MULTILINE:String = "disabledMultiline";
		
		/**
		 * A string literal representing the name of the editable multiline state.
		 */
		static public const STATE_EDITABLE_MULTILINE:String = "editableMultiline";
		
		/**
		 * The navigator's CollectionManager instance. All host interaction should go through this
		 *  instance, rather than the INavigatorHost directly.
		 */
		protected static const collectionManager:CollectionManager = CollectionManager.instance;
		
		private var _multiline:Boolean = false;
		/**
		 * A boolean indicating whether or not this component supports multiline text.
		 */
		[Bindable]
		public function get multiline():Boolean
		{
			return _multiline;
		}
		/**
		 * @private
		 */
		public function set multiline(value:Boolean):void
		{
			_multiline = value;
			handleMultiline();
		}

		private var _allowStyling:Boolean = true;
		/**
		 * A boolean indicating whether or not this component supports styled text.
		 */
		[Bindable]
		public function get allowStyling():Boolean
		{
			return _allowStyling;
		}
		public function set allowStyling(value:Boolean):void
		{
			_allowStyling = value;
		}
		
		/**
		 * The number to be used as the measuredWidth when measure() is called
		 * for this component. This can be used to prevent long text values from
		 * affecting the measuredWidth of the parent container.
		 * This value is ignored when set to NaN.
		 */
		public var forceMeasuredWidth:Number;
		
		/**
		 * @inheritDoc
		 */
		override protected function measure():void
		{
			super.measure();
			if (!isNaN(forceMeasuredWidth)) {
				measuredWidth = forceMeasuredWidth;
			}
		}
		
		/**
		 * Default constructor
		 */
		public function CustomizableTextArea()
		{
			super();
			selectionHighlighting = TextSelectionHighlighting.ALWAYS; // So we can have text stay selected while we use the panel
			addListenAndWatch();
		}

		/**
		 * Returns true if this component should be editable by the user.
		 */
		protected function shouldBeEditable():Boolean
		{
			return collectionManager.authoringMode && collectionManager.permissionCheck("Document", "Modify");
		}
		
		/**
		 * Updates the editable state of this component based on current state.
		 */
		protected function updateEditable():void
		{
			var newSkinIsEditable:Boolean = shouldBeEditable();
			// Make sure we initialize all this stuff on creationComplete
			if (!creationDidComplete || (newSkinIsEditable != skinIsEditable)) {
				skinIsEditable = newSkinIsEditable;
				editable = skinIsEditable;
				if (skinIsEditable && focusIn) {
					// We have focus, and we can edit, so set the current property context to textfield
					takeContext();
				}
				else if (!skinIsEditable) {
					loseContext();
				}
				selectable = skinIsEditable;
				invalidateSkinState();
			}
		}
		
		private function handleMultiline():void
		{
			if (textDisplay == null) {
				return;
			}
			if (multiline) {
				textDisplay.multiline = true;
				setStyle("lineBreak", LineBreak.TO_FIT);
				
				if (textDisplay is UIComponent) {
					(textDisplay as UIComponent).percentHeight = 100;
				}
			} else {
				textDisplay.multiline = false;
				setStyle("lineBreak", LineBreak.EXPLICIT);
				
				// TextInput should always be 1 line.
				heightInLines = 1;
			}
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function partAdded(partName:String, instance:Object):void
		{
			super.partAdded(partName, instance);
			
			if (instance == textDisplay) {
				handleMultiline();
			}
		}

		private var authoringModeWatcher:ChangeWatcher;

		/**
		 * A function to add event listeners and change watchers required by this canvas list.
		 */
		public function addListenAndWatch():void
		{
			// Don't forget to remove all these events if our part gets removed...
			
			// In the case of a NEEDS_SAVE_KIND event, we may need to dirty ourselves before other listeners
			// respond, so use a higher priority.
			const NAV_HOST_EVENT_PRIORITY:uint = 3;
			collectionManager.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent, false, NAV_HOST_EVENT_PRIORITY, true);
			
			authoringModeWatcher = BindingUtils.bindSetter(setAuthoringMode, collectionManager, "authoringMode");
			addEventListener(FlexEvent.CREATION_COMPLETE, onCreationComplete, false, 0, true);
			addEventListener(FocusEvent.FOCUS_IN, onFocusInOrOut, false, 0, true);
			addEventListener(FocusEvent.FOCUS_OUT, onFocusInOrOut, false, 0, true);
			addEventListener(KeyboardEvent.KEY_DOWN, onKeyDown, false, 0, true);
			addEventListener(MouseEvent.DOUBLE_CLICK, onDoubleClick, false, 0, true);
			addEventListener(TextOperationEvent.CHANGE, onTextChanged, false, 0, true);
			addEventListener(FlexEvent.SELECTION_CHANGE, onSelectionChanged, false, 0, true);
		}

		/**
		 * A function to clean up event listeners and change watchers required by this canvas list.
		 */
		public function removeListenAndWatch():void
		{
			collectionManager.removeEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent);
			if (authoringModeWatcher != null) {
				authoringModeWatcher.unwatch();
				authoringModeWatcher = null;
			}
			removeEventListener(FlexEvent.CREATION_COMPLETE, onCreationComplete);
			removeEventListener(FocusEvent.FOCUS_IN, onFocusInOrOut);
			removeEventListener(FocusEvent.FOCUS_OUT, onFocusInOrOut);
			removeEventListener(KeyboardEvent.KEY_DOWN, onKeyDown);
			removeEventListener(MouseEvent.DOUBLE_CLICK, onDoubleClick);
			removeEventListener(TextOperationEvent.CHANGE, onTextChanged);
			removeEventListener(FlexEvent.SELECTION_CHANGE, onSelectionChanged);
			if (fontWatcher != null) {
				fontWatcher.unwatch();
				fontWatcher = null;
			}
			if (sizeWatcher != null) {
				sizeWatcher.unwatch();
				sizeWatcher = null;
			}
			if (styleWatcher != null) {
				styleWatcher.unwatch();
				styleWatcher = null;
			}
			if (weightWatcher != null) {
				weightWatcher.unwatch();
				weightWatcher = null;
			}
			if (colorWatcher != null) {
				colorWatcher.unwatch();
				colorWatcher = null;
			}
			if (underlinedWatcher != null) {
				underlinedWatcher.unwatch();
				underlinedWatcher = null;
			}

		}
		
		private function onKeyDown(event:KeyboardEvent):void
		{
			if (event.isDefaultPrevented()) {
				return;
			}
			var shouldPreventDefault:Boolean = false;
			var shouldStopPropagation:Boolean = false;
			if (skinIsEditable) {
				// The user is typing in the field, so don't let the default navigator do its thing
				if (event.keyCode != Keyboard.TAB) {
					shouldStopPropagation = true;
				}
				switch (event.keyCode) {
					case Keyboard.ESCAPE:
						shouldPreventDefault = true;
						if (pendingChanges) {
							setTextFromData();
						}
						break;
					case Keyboard.ENTER:
						if (!multiline) {
							// Increment focus if this isn't multiline text
							shouldPreventDefault = true;
							setDataFromText();
							var nextFocus:IFocusManagerComponent = (focusManager == null) ? null : focusManager.getNextFocusManagerComponent();
							if (nextFocus != null) {
								focusManager.setFocus(nextFocus);
							}
						}
						break;
				}
			}
			if (shouldPreventDefault && event.cancelable) {
				event.preventDefault();
			}
			if (shouldStopPropagation && event.bubbles) {
				event.stopImmediatePropagation();
			}
		}
		
		private function onDoubleClick(event:MouseEvent):void
		{
			if (event.isDefaultPrevented()) {
				return;
			}
			// If user is clicking quickly on the up and down buttons
			// of the scroll bar, stop propagation. Otherwise the
			// double click is received at a higher level and either
			// the card is previewed or opened (depending on in layout
			// of files view).
			if (event.target is Button) {
				event.stopImmediatePropagation();
			}
		}
		
		private var skinIsEditable:Boolean = false;
		
		/**
		 * @inheritDoc
		 */
		override protected function getCurrentSkinState():String
		{
			if (!enabled) {
				return multiline ? STATE_DISABLED_MULTILINE : STATE_DISABLED;
			} else if (skinIsEditable) {
				return multiline ? STATE_EDITABLE_MULTILINE : STATE_EDITABLE;
			} else {
				return multiline ? STATE_NORMAL_MULTILINE : STATE_NORMAL;
			}
		}
		
		/**
		 * Event listener invoked on CollectionManagerEvents.
		 */
		protected function onNavHostEvent(event:CollectionManagerEvent):void
		{
			switch (event.kind) {
				case CollectionManagerEvent.PART_REMOVED_KIND:
					if (event.oldVal == this)
					{
						// This part is being removed, so clean up event listeners
						removeListenAndWatch();
					}
					break;
				case CollectionManagerEvent.NEEDS_SAVE_KIND:
					// If we're potentially about to save, then try to apply any pending changes
					setDataFromText();
					break;
			}
		}
		
		private var creationDidComplete:Boolean = false;

		/**
		 * Event listener invoked when this component is created..
		 */
		protected function onCreationComplete(event:FlexEvent):void
		{
			if (creationDidComplete) {
				LogHelper.logger.warn("onCreationComplete invoked twice");
				return;
			}
			updateEditable();
			creationDidComplete = true;
		}
		
		private function setAuthoringMode(value:Boolean):void
		{
			// If authoringMode is changing, then try to apply any pending changes
			setDataFromText();
			
			updateEditable();
		}
		
		private var focusIn:Boolean = false;
		private function onFocusInOrOut(event:FocusEvent):void
		{
			focusIn = (event.type == FocusEvent.FOCUS_IN);
			if (focusIn) {
				if (skinIsEditable) {
					// We have focus, and we can edit, so set the current property context to textfield
					takeContext();
				}
			} else {
				if (((focusManager == null) || (focusManager.getFocus() == null)) && (stage != null)) {
					// If nothing else actually got focus, then wait until something does
					stage.addEventListener(FocusEvent.FOCUS_IN, onStageFocusIn, false, 0, true);
				} else {
					// We lost focus, so clear the current property context
					loseContext();
				}
			}
		}
		
		private function onStageFocusIn(event:FocusEvent):void
		{
			var currentFocus:IFocusManagerComponent = (focusManager == null) ? null : focusManager.getFocus();
			if (currentFocus == null) {
				return;
			}
			if (event.currentTarget is IEventDispatcher) {
				(event.currentTarget as IEventDispatcher).removeEventListener(FocusEvent.FOCUS_IN, onStageFocusIn);
			}
			if ((currentFocus == this) || (currentFocus == skin)) {
				takeContext();
			} else {
				loseContext();
			}
		}
		
		private var flushingPropertyValues:Boolean;
		private function onFontChanged(fontFamily:String):void
		{
			if (flushingPropertyValues) {
				// We're setting this property ourselves, so just return
				return;
			}
			if ((fontFamily == null) || (fontFamily == "")) {
				LogHelper.logger.error("Invalid fontFamily");
				return;
			}
			var format:TextLayoutFormat = getFormatOfRange();
			if (format) {
				format.fontFamily = fontFamily;
				setFormatOfRange(format);
				pendingChanges = true;
			}
		}
		
		private function onSizeChanged(pointSize:Number):void
		{
			if (flushingPropertyValues) {
				// We're setting this property ourselves, so just return
				return;
			}
			if (isNaN(pointSize) || (pointSize <= 0)) {
				LogHelper.logger.error("Invalid pointSize");
				return;
			}
			var format:TextLayoutFormat = getFormatOfRange();
			if (format) {
				format.fontSize = pointSize;
				setFormatOfRange(format);
				pendingChanges = true;
			}
		}
		
		private function onStyleChanged(fontStyle:String):void
		{
			if (flushingPropertyValues) {
				// We're setting this property ourselves, so just return
				return;
			}
			if ((fontStyle == null) || (fontStyle == "")) {
				LogHelper.logger.error("Invalid fontStyle");
				return;
			}
			var format:TextLayoutFormat = getFormatOfRange();
			if (format) {
				format.fontStyle = fontStyle;
				setFormatOfRange(format);
				pendingChanges = true;
			}
		}
		
		private function onWeightChanged(fontWeight:String):void
		{
			if (flushingPropertyValues) {
				// We're setting this property ourselves, so just return
				return;
			}
			if ((fontWeight == null) || (fontWeight == "")) {
				//logger.error("Invalid fontWeight");
				return;
			}
			var format:TextLayoutFormat = getFormatOfRange();
			if (format) {
				format.fontWeight = fontWeight;
				setFormatOfRange(format);
				pendingChanges = true;
			}
		}
		private function onColorChanged(fontColor:int):void
		{
			if (flushingPropertyValues) {
				// We're setting this property ourselves, so just return
				return;
			}
			var format:TextLayoutFormat = getFormatOfRange();
			if (format) {
				format.color = fontColor;
				setFormatOfRange(format);
				pendingChanges = true;
			}
		}
		private function onUnderlinedChanged(fontUnderlined:Boolean):void
		{
			if (flushingPropertyValues) {
				// We're setting this property ourselves, so just return
				return;
			}
			var format:TextLayoutFormat = getFormatOfRange();
			if (format) {
				format.textDecoration = fontUnderlined ? flashx.textLayout.formats.TextDecoration.UNDERLINE : flashx.textLayout.formats.TextDecoration.NONE;
				setFormatOfRange(format);
				pendingChanges = true;
			}
		}
		
		private function addListValue(pl:PropertyListValue, v:String):void
		{
			// If it's already there, don't add it again
			if (pl.setValueFromString(v))
				return;
			
			// Make a copy of the current subvalues
			var newSubValues:Array = pl.subvalues.slice();
			var newActiveValue:PropertyValue = PropertyValue.createPropertyValue(<PropertyValue name={v} label={v}/>, pl);
			newSubValues.push(newActiveValue);
			if ((newSubValues != null) && (newActiveValue != null)) {
				pl.updateSubvalues(newSubValues, newActiveValue);
			}
		}
		
		private function onSelectionChanged(event:FlexEvent):void
		{
			if (!allowStyling) {
				// Don't update any properties if this isn't style-able text
				return;
			}
			var context:PropertyContext = PropertyManager.instance.currentContext;
			if ((context == null) || (context.name != CommonProperties.CONTEXT_TEXTFIELD)) {
				// Don't bother setting properties if we're not the current context
				return;
			}
			setPropertyValuesFromFormat();
		}
		
		private function setPropertyValuesFromFormat():void
		{
			var cp:CommonProperties = PropertyManager.instance.commonProperties;
			if ((cp == null) || (collectionManager.host == null)) {
				return;
			}
			
			// Make sure we don't try to update the text based on our own property changes
			var originalFlushingValue:Boolean = flushingPropertyValues;
			flushingPropertyValues = true;
			
			var format:TextLayoutFormat = getFormatOfRange();
			if (format) {
				if (format.hasOwnProperty("fontFamily") &&
					format.fontFamily &&
					(format.fontFamily != "") &&
					(cp.fontFamilyList != null))
				{
					if (!cp.fontFamilyList.setValueFromString(format.fontFamily))
					{
						addListValue(cp.fontFamilyList, format.fontFamily);
					}
				}
				if (format.hasOwnProperty("fontSize") &&
					(format.fontSize > 0) &&
					(cp.fontSize != null))
				{
					cp.fontSize.value = format.fontSize;
				}
				if (format.hasOwnProperty("fontStyle") &&
					format.fontStyle &&
					(format.fontStyle != "") &&
					(cp.fontStyleList != null))
				{				
					cp.fontStyleList.setValueFromString(format.fontStyle);
				}
				if (format.hasOwnProperty("fontWeight") &&
					format.fontWeight &&
					(format.fontWeight != "") &&
					(cp.fontWeightList != null))
				{				
					cp.fontWeightList.setValueFromString(format.fontWeight);
				}
				if (format.hasOwnProperty("color") && (cp.fontColor != null)) {
					cp.fontColor.value = format.color;
				}
				if (format.hasOwnProperty("textDecoration") &&
					format.textDecoration != "" &&
					(cp.fontUnderlined != null))
				{
					cp.fontUnderlined.value = (format.textDecoration == flashx.textLayout.formats.TextDecoration.UNDERLINE);
				}
			} else {
				LogHelper.logger.error("No format to retrieve text properties");
			}
			
			flushingPropertyValues = originalFlushingValue;
		}
		
		private var fontWatcher:ChangeWatcher;
		private var sizeWatcher:ChangeWatcher;
		private var styleWatcher:ChangeWatcher;
		private var weightWatcher:ChangeWatcher;
		private var colorWatcher:ChangeWatcher;
		private var underlinedWatcher:ChangeWatcher;
		
		/**
		 * Sets the property context for this text component as the current property context.
		 * 
		 * @see PropertyContext
		 * @see PropertyManager#setCurrentContext
		 */
		public function takeContext():void
		{
			if (!allowStyling) {
				// This instance doesn't support styling, so don't take context
				return;
			}
			if (fontWatcher != null) {
				// We've already taken context;
				return;
			}
			if (!CollectionManager.instance.apiIsSupported(CollectionManager.API_VERSION_10)) {
				// Necessary APIs are not supported, so don't bother
				return;
			}
			var commonProperties:CommonProperties = PropertyManager.instance.commonProperties;
			if (commonProperties != null) {
				var textContext:PropertyContext = commonProperties.getCurrentContextByName(CommonProperties.CONTEXT_TEXTFIELD);
				if (textContext == null) {
					return;
				}
				
				// Don't let our format values get affected by initializing the context
				var originalFlushingValue:Boolean = flushingPropertyValues;
				flushingPropertyValues = true;
				
				// First initialize the values based on this object
				setPropertyValuesFromFormat();
				
				commonProperties.setCurrentContext(textContext, this);
				
				fontWatcher = BindingUtils.bindSetter(onFontChanged, commonProperties, ["fontFamilyList", "activeValue", "name"]);
				sizeWatcher = BindingUtils.bindSetter(onSizeChanged, commonProperties, ["fontSize", "numValue"]);
				styleWatcher = BindingUtils.bindSetter(onStyleChanged, commonProperties, ["fontStyleList", "activeValue", "name"]);
				weightWatcher = BindingUtils.bindSetter(onWeightChanged, commonProperties, ["fontWeightList", "activeValue", "name"]);
				colorWatcher = BindingUtils.bindSetter(onColorChanged, commonProperties, ["fontColor", "numValue"]);
				underlinedWatcher = BindingUtils.bindSetter(onUnderlinedChanged, commonProperties, ["fontUnderlined", "booleanValue"]);
				
				flushingPropertyValues = originalFlushingValue;
			}
		}
		
		/**
		 * Clears the property context for this text component as the current property context.
		 * 
		 * @see PropertyContext
		 * @see PropertyManager#setCurrentContext
		 */
		public function loseContext():void
		{
			// If we're losing our context, try to apply any pending changes
			setDataFromText();
			
			// After editing text that scrolled to the right or down, display text from first character
			scrollToRange(0, 0);
			
			if (PropertyManager.instance.commonProperties != null) {
				PropertyManager.instance.commonProperties.setCurrentContextByName(CommonProperties.CONTEXT_CLEAR, this);
				if (fontWatcher != null) {
					fontWatcher.unwatch();
					fontWatcher = null;
				}
				if (sizeWatcher != null) {
					sizeWatcher.unwatch();
					sizeWatcher = null;
				}
				if (styleWatcher != null) {
					styleWatcher.unwatch();
					styleWatcher = null;
				}
				if (weightWatcher != null) {
					weightWatcher.unwatch();
					weightWatcher = null;
				}
				if (colorWatcher != null) {
					colorWatcher.unwatch();
					colorWatcher = null;
				}
				if (underlinedWatcher != null) {
					underlinedWatcher.unwatch();
					underlinedWatcher = null;
				}
			}
		}
		
		/**
		 * True if changes to the text have been made that have not yet been applied.
		 */
		protected var pendingChanges:Boolean = false;
		
		/**
		 * Event listener invoked when this components text is changed by the user.
		 */
		protected function onTextChanged(event:TextOperationEvent):void
		{
			pendingChanges = true;
			setTextDirection(text);
		}
		
		private var textDirection:String = null;
		
		/**
		 * A function that is invoked when the text direction is changed.
		 */
		protected function setTextDirection(textStr:String):void
		{
			// Get the text direction based on the text (or based on locale for empty strings)
			var newTextDirection:String = GlobalizationUtil.textDirection(collectionManager.host, textStr);
			// If the text direction changed, update the text layout direction
			if (newTextDirection != textDirection) {
				textDirection = newTextDirection;
				setStyle("direction", textDirection);
				// Setting the direction style is not enough to change the direction in a text area.
				// Note: No reason to get the existing text layout format first.
				// All fields of a new TextLayoutFormat are undefined initially
				// and only non-undefined fields (direction in this case) are changed,
				// all other fields are left alone.
				//				var textLayoutFormat:TextLayoutFormat = new TextLayoutFormat();
				//				textLayoutFormat.direction = textDirection;
				//				setFormatOfRange(textLayoutFormat);
				// And setFormatOfRange does not do the correct thing either.
				// This works - iterating over all of the flow elements and setting direction on each one.
				if (textFlow) {
					for (var index:int = 0; index < textFlow.numChildren; ++index) {
						var flowElement:FlowElement = textFlow.getChildAt(index);
						if (!flowElement)
							continue;
						flowElement.direction = textDirection;
					}
				}
			}
		}
		
		/**
		 * Applies pending text changes to the data model. Subclasses should
		 * override this function.
		 */
		protected function setDataFromText():void
		{
			if (pendingChanges) {
				dispatchEvent(new TextOperationEvent(TextOperationEvent.CHANGE));
				pendingChanges = false;
			}
		}
		
		/**
		 * Updates this component's text based on the data model.
		 * Subclasses should override this function.
		 */
		protected function setTextFromData():void
		{
			// TODO dispatch an event to reset the field?
		}
		
		/**
		 * Return true if a TextArea component currently has focus.
		 */
		public static function textAreaHasFocus(focusManager:IFocusManager):Boolean
		{
			var currentFocus:IFocusManagerComponent = (focusManager != null) ? focusManager.getFocus() : null;
			var textArea:TextArea = currentFocus as TextArea;
			return (textArea != null) && textArea.enabled && textArea.selectable;
		}
	}
}
