/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.supportClasses
{
	import com.adobe.portfolio.utils.GlobalizationUtil;
	
	import flash.display.InteractiveObject;
	import flash.events.FocusEvent;
	import flash.events.IEventDispatcher;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.globalization.DateTimeFormatter;
	import flash.globalization.DateTimeNameContext;
	import flash.globalization.DateTimeNameStyle;
	import flash.globalization.DateTimeStyle;
	import flash.globalization.LocaleID;
	import flash.globalization.NumberFormatter;
	import flash.ui.Keyboard;
	
	import mx.controls.DateField;
	import mx.events.CalendarLayoutChangeEvent;
	import mx.events.FlexEvent;
	
	import spark.components.Button;
	import spark.components.TextInput;
	import spark.components.supportClasses.SkinnableComponent;
	import spark.events.TextOperationEvent;
	import spark.skins.spark.TextInputSkin;
	
	/**
	 *  The DateTimeChooser class extends SkinnableComponent to implement a date and time chooser component. 
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */	
	public class DateTimeChooser extends SkinnableComponent
	{
		private static const MIN_HOUR_12:Number = 1;
		private static const MAX_HOUR_12:Number = 12;
		private static const MIN_HOUR_24:Number = 0;
		private static const MAX_HOUR_24:Number = 23;
		private static const MIN_MINUTE:Number = 0;
		private static const MAX_MINUTE:Number = 59;
		private static const MIN_SECOND:Number = 0;
		private static const MAX_SECOND:Number = 59;
		private static const MIN_MILLISECOND:Number = 0;

		private static const AM:String = "AM";
		private static const PM:String = "PM";
		
		/**
		 * String literals used to represent the skin parts exposed by DateTimeChooser
		 */
		private static const DATE_INPUT:String = "dateInput";
		private static const HOUR_INPUT:String = "hourInput";
		private static const MINUTE_INPUT:String = "minuteInput";
		private static const SECOND_INPUT:String = "secondInput";
		private static const MERIDIEM_INPUT:String = "meridiemInput";
		private static const UP_BUTTON:String = "upButton";
		private static const DOWN_BUTTON:String = "downButton";
		
		/**
		 *  The date chooser to select a new date.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="true")]
		public var dateInput:DateField;

		/**
		 *  The hours field.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="true")]
		public var hourInput:TextInput;
		
		/**
		 *  The minutes field.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="true")]
		public var minuteInput:TextInput;
		
		/**
		 *  The seconds field.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="false")]
		public var secondInput:TextInput;
		
		/**
		 *  The meridiem (am/pm) field.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="false")]
		public var meridiemInput:TextInput;
		
		/**
		 *  The up button (to increment the hour/minute/second/meridiem field values).
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="true")]
		public var upButton:Button;
		
		/**
		 *  The down button (to increment the hour/minute/second/meridiem field values).
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="true")]
		public var downButton:Button;
		
		private var dateOnly:Date = null;
		private var hour:Number = MAX_HOUR_12;
		private var minute:Number = MIN_MINUTE;
		private var second:Number = MIN_SECOND;
		private var millisecond:Number = MIN_MILLISECOND;
		private var meridiem:String = PM;
		
		private var lastInputWithFocus:TextInput = null;
		
		private var minHour:Number = MIN_HOUR_12;
		private var maxHour:Number = MAX_HOUR_12;
		private var amLocalized:String = AM;
		private var pmLocalized:String = PM;
		
		private var zeroDigit:String;
		private var validDigits:String;
		private var timeUsesMeridiemIndicators:Boolean = true;
		
		private var numFormatter:NumberFormatter;
		
		/**
		 *  Constructor. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function DateTimeChooser()
		{
			super();
			
			// Does the current locale use a 12 hour or 24 hour time system?
			var dateTimeFormatter:DateTimeFormatter = new DateTimeFormatter(LocaleID.DEFAULT, DateTimeStyle.NONE, DateTimeStyle.SHORT);
			var timePattern:String = dateTimeFormatter.getDateTimePattern();
			timeUsesMeridiemIndicators = (timePattern.search("a") != -1) || (timePattern.search("p") != -1);
			
			// These are used both for figuring out the am/pm indictor and the digit types
			var str:String;
			var date:Date = new Date();	// Arbitrary date - we only care about the hour
			
			if (timeUsesMeridiemIndicators) {
				minHour = MIN_HOUR_12;
				maxHour = MAX_HOUR_12;
				// Figure out the localized AM/PM indicators
				dateTimeFormatter.setDateTimePattern("ha");	// This means return hour in 12 hour mode 'plus' am/pm indicator
				date.hours = 3;
				str = dateTimeFormatter.format(date);
				if (str)
					amLocalized = str.substr(1);
				date.hours = 15;
				str = dateTimeFormatter.format(date);
				if (str)
					pmLocalized = str.substr(1);
			}
			else {
				minHour = MIN_HOUR_24;
				maxHour = MAX_HOUR_24;
			}

			// Assume "normal" digits unless time formatted by current locale uses something else (which is unusual)
			var timeLocale:String = "en-US";
			dateTimeFormatter.setDateTimePattern("HH");	// This means return hour in 24 hour mode
			date.hours = 3;
			str = dateTimeFormatter.format(date);
			if (str && str.substr(0, 1) != "0")
				timeLocale = LocaleID.DEFAULT;
			numFormatter = new NumberFormatter(timeLocale);
			numFormatter.fractionalDigits = 0;
			var zeroDigitCharCode:uint = numFormatter.digitsType;
			zeroDigit = String.fromCharCode(zeroDigitCharCode);
			validDigits = zeroDigit + "-" + String.fromCharCode(zeroDigitCharCode + 9);
		}
		
		/**
		 *  The selected date and time for the DateTimeChooser.
		 *  After creating DateTimeChooser set <code>selectedDate</code> if you want the date to be something other than current date.
		 *  When a FlexEvent.DATA_CHANGE is received from the DateTimeChooser, the new date is retrieved from <code>selectedDate</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get selectedDate():Date
		{
			if (!dateOnly)
				return null;
			var hour24:Number = hour;
			if (timeUsesMeridiemIndicators) {
				// This is required to convert 12 pm (noon) to 0 and the code below changes it back to 12 pm (noon)
				hour24 %= MAX_HOUR_12;
				if (meridiem == PM)
					hour24 += MAX_HOUR_12;
			}
			return new Date(dateOnly.fullYear, dateOnly.month, dateOnly.date, hour24, minute, second, millisecond);
		}

		/**
		 *  @private
		 */	
		public function set selectedDate(value:Date):void
		{
			if (value == null)
			{
				dateOnly = null;
				// Default the time to 12:00 noon
				hour = MAX_HOUR_12;
				minute = MIN_MINUTE;
				second = MIN_SECOND;
				millisecond = MIN_MILLISECOND;
				meridiem = PM;
			}
			else
			{
				dateOnly = new Date(value.getTime());
				if (timeUsesMeridiemIndicators) {
					hour = value.hours % MAX_HOUR_12;
					if (hour == 0)
						hour = MAX_HOUR_12;
					meridiem = value.hours >= MAX_HOUR_12 ? PM : AM;
				}
				else {
					hour = value.hours;
				}
				minute = value.minutes;
				second = value.seconds;
				millisecond = value.milliseconds;
			}
			takeFocus();
			updateDisplay();
		}
		
		private var valuesChanged:Boolean = false;
		override protected function commitProperties():void
		{
			if (valuesChanged) {
				dateInput.selectedDate = dateOnly;
				hourInput.text = formatNumber(hour);
				minuteInput.text = formatNumber(minute, 2);
				secondInput.text = formatNumber(second, 2);
				if (timeUsesMeridiemIndicators)
					meridiemInput.text = meridiem == AM ? amLocalized : pmLocalized;
				
				var hasDate:Boolean = dateOnly != null;
				var hourWasNotVisible:Boolean = !hourInput.visible;
				hourInput.enabled = hourInput.visible = hasDate; 
				minuteInput.enabled = minuteInput.visible = hasDate; 
				secondInput.enabled = secondInput.visible = hasDate; 
				meridiemInput.enabled = meridiemInput.visible = hasDate && timeUsesMeridiemIndicators;
				upButton.enabled = upButton.visible = hasDate;
				downButton.enabled = downButton.visible = hasDate;
				
				if (hasDate && hourWasNotVisible)
					takeFocus();

				valuesChanged = false;
			}
			super.commitProperties();
		}
		
		/** Called when a skin has added a skin part for this instance.
		 */
		override protected function partAdded(partName:String, instance:Object):void
		{
			handlePartChange(partName, instance, true);
		}
		
		/** Called when a skin has removed a skin part for this instance.
		 */
		override protected function partRemoved(partName:String, instance:Object):void
		{
			handlePartChange(partName, instance, false);
		}
		
		/** Internal function for adding and removing event listeners for skin parts, depending on
		 *  whether the part is being added or removed for this instance.
		 */
		private function handlePartChange(partName:String, instance:Object, adding:Boolean):void
		{
			var listeners:Array = null;
			var types:Array = null;
			var activePart:IEventDispatcher = instance as IEventDispatcher;
			if (activePart) {
				// Build up an array of event types and listeners that apply to this object
				switch (partName) {
					case DATE_INPUT:
						if (adding && dateInput) {
							// Localize the elements of the calendar
							var dateFormatter:DateTimeFormatter = new DateTimeFormatter(LocaleID.DEFAULT, DateTimeStyle.LONG, DateTimeStyle.LONG);
							dateInput.dayNames = dateFormatter.getWeekdayNames(DateTimeNameStyle.SHORT_ABBREVIATION, DateTimeNameContext.STANDALONE).toString().split(',');
							dateInput.monthNames = dateFormatter.getMonthNames(DateTimeNameStyle.FULL, DateTimeNameContext.STANDALONE).toString().split(',');
							dateInput.firstDayOfWeek = dateFormatter.getFirstWeekday();
							dateInput.labelFunction = formatDateFieldDate;
							dateInput.parseFunction = null;
						}
						types = [ CalendarLayoutChangeEvent.CHANGE, FocusEvent.FOCUS_OUT ];
						listeners = [ onDateChange, onFocusOut ];
						break;
					
					case HOUR_INPUT:
						if (adding && hourInput) {
							initTextInput(hourInput);
							takeFocus();
						}
						types = [ FocusEvent.FOCUS_IN, KeyboardEvent.KEY_DOWN, TextOperationEvent.CHANGE, FocusEvent.FOCUS_OUT ];
						listeners = [ onFocusIn, onKeyDown, onHourChange, onFocusOut ];
						break;
					
					case MINUTE_INPUT:
						if (adding && minuteInput) {
							initTextInput(minuteInput);
						}
						types = [ FocusEvent.FOCUS_IN, KeyboardEvent.KEY_DOWN, TextOperationEvent.CHANGE, FocusEvent.FOCUS_OUT ];
						listeners = [ onFocusIn, onKeyDown, onMinuteChange, onFocusOut ];
						break;
					
					case SECOND_INPUT:
						if (adding && secondInput) {
							initTextInput(secondInput);
						}
						types = [ FocusEvent.FOCUS_IN, KeyboardEvent.KEY_DOWN, TextOperationEvent.CHANGE, FocusEvent.FOCUS_OUT ];
						listeners = [ onFocusIn, onKeyDown, onSecondChange, onFocusOut ];
						break;
					
					case MERIDIEM_INPUT:
						if (adding && meridiemInput) {
							initTextInput(meridiemInput);
							meridiemInput.visible = timeUsesMeridiemIndicators;
						}
						types = [ FocusEvent.FOCUS_IN, KeyboardEvent.KEY_DOWN, FocusEvent.FOCUS_OUT ];
						listeners = [ onFocusIn, onKeyDown, onFocusOut ];
						break;
					
					case UP_BUTTON:
						types = [ MouseEvent.CLICK, FocusEvent.FOCUS_OUT ];
						listeners = [ onUpButtonClick, onFocusOut ];
						break;
					
					case DOWN_BUTTON:
						types = [ MouseEvent.CLICK, FocusEvent.FOCUS_OUT ];
						listeners = [ onDownButtonClick, onFocusOut ];
						break;
					
					default:
						break;
				}
				
				// Add or remove all the relevant event listeners
				var count:uint = listeners ? listeners.length : 0;
				if (count > 0) {
					for (var i:uint = 0; i < count; ++i) {
						adding ? activePart.addEventListener(types[i], listeners[i]) :
							activePart.removeEventListener(types[i], listeners[i]);
					}
				}
			}
			adding ? super.partAdded(partName, instance) : super.partRemoved(partName, instance);
		}
		
		private function initTextInput(input:TextInput):void
		{
			input.restrict = validDigits;
			var skin:TextInputSkin = input.skin as TextInputSkin;
			if (skin && skin.background)
				skin.background.alpha = 0.3;
		}
		
		private function takeFocus():void
		{
			if (!hourInput)
				return;
			// If the hour field exists, put the focus on it
			switchFocusToInput(hourInput);
		}

		private function onFocusIn(event:FocusEvent):void
		{
			lastInputWithFocus = event.currentTarget as TextInput;
		}
		
		private function onKeyDown(event:KeyboardEvent):void
		{
			var input:TextInput = event.currentTarget as TextInput;
			if (!input)
				return;

			switch (event.keyCode) {
				case Keyboard.UP:
				case Keyboard.NUMPAD_ADD:
					incrementInput(input);
					break;
				
				case Keyboard.DOWN:
				case Keyboard.NUMPAD_SUBTRACT:
					decrementInput(input);
					break;
				
				default:
					return;
			}
			event.preventDefault();
		}
		
		private function onUpButtonClick(event:MouseEvent):void
		{
			var input:TextInput = lastInputWithFocus;
			if (input == null)
				input = hourInput;
			incrementInput(input);				
			switchFocusToInput(input);
		}

		private function onDownButtonClick(event:MouseEvent):void
		{
			var input:TextInput = lastInputWithFocus;
			if (input == null)
				input = hourInput;
			decrementInput(input);				
			switchFocusToInput(input);
		}

		private function incrementInput(input:TextInput):void
		{
			if (input == hourInput)
				hour = increment(hour, minHour, maxHour);
			else if (input == minuteInput)
				minute = increment(minute, MIN_MINUTE, MAX_MINUTE);
			else if (input == secondInput)
				second = increment(second, MIN_SECOND, MAX_SECOND);
			else if (input == meridiemInput)
				meridiem = meridiem == AM ? PM : AM; 
			updateDisplay();
		}
		
		private function decrementInput(input:TextInput):void
		{
			if (input == hourInput)
				hour = decrement(hour, minHour, maxHour);
			else if (input == minuteInput)
				minute = decrement(minute, MIN_MINUTE, MAX_MINUTE);
			else if (input == secondInput)
				second = decrement(second, MIN_SECOND, MAX_SECOND);
			else if (input == meridiemInput)
				meridiem = meridiem == AM ? PM : AM; 
			updateDisplay();
		}
		
		private static function switchFocusToInput(input:TextInput):void
		{
			input.setFocus();
			input.selectAll();
		}
		
		private function onDateChange(event:CalendarLayoutChangeEvent):void
		{
			dateOnly = dateInput.selectedDate;
			updateDisplay();
		}
		
		private function onHourChange(event:TextOperationEvent):void
		{
			var text:String = hourInput.text;
			if (text == "")
				return;

			var number:Number = parseFloat(text);
			if (number >= minHour && number <= maxHour) 
				hour = number;
			else
				updateDisplay(); // revert to previous values
		}
		
		private function onMinuteChange(event:TextOperationEvent):void
		{
			var text:String = minuteInput.text;
			if (text == "")
				return;
			
			var number:Number = parseFloat(text);
			if (number >= MIN_MINUTE && number <= MAX_MINUTE) 
				minute = number;
			else
				updateDisplay(); // revert to previous values
		}
		
		private function onSecondChange(event:TextOperationEvent):void
		{
			var text:String = secondInput.text;
			if (text == "")
				return;
			
			var number:Number = parseFloat(text);
			if (number >= MIN_SECOND && number <= MAX_SECOND) 
				second = number;
			else
				updateDisplay(); // revert to previous values
		}
				
		private function updateDisplay():void
		{
			valuesChanged = true;
			invalidateProperties();
		}
		
		private function onFocusOut(event:FocusEvent):void
		{
			// If we're changing focus to another sub-component of our
			// component then we aren't really losing focus, so don't
			// propagate the focus out event
			if (focusIsInDateTimeChooser(event.relatedObject)) {
				event.stopImmediatePropagation();
				return;
			}
			dispatchEvent(new FlexEvent(FlexEvent.DATA_CHANGE));
		}
		
		private function focusIsInDateTimeChooser(focus:InteractiveObject):Boolean
		{
			if (!focus)
				return false;
			if (dateInput.contains(focus))
				return true;
			if (hourInput.contains(focus))
				return true;
			if (minuteInput.contains(focus))
				return true;
			if (secondInput.contains(focus))
				return true;
			if (meridiemInput.contains(focus))
				return true;
			if (upButton.contains(focus))
				return true;
			if (downButton.contains(focus))
				return true;
			return false;
		}
				
		// Provide the localized date string for the DateField input/display field
		private function formatDateFieldDate(currentDate:Date):String
		{
			if (currentDate == null)
				return "";
			return GlobalizationUtil.formatDateTime(currentDate, DateTimeStyle.SHORT, DateTimeStyle.NONE);
		}
		
		private function formatNumber(value:Number, requiredLength:int = 1):String
		{
			if (!numFormatter)
				return "";
			var result:String = numFormatter.formatNumber(value);
			var count:int = requiredLength - result.length; 
			if (count <= 0) return result;
			
			var zeros:String = "";
			while (count > 0)
			{
				zeros += zeroDigit;
				count--;
			}
			return zeros + result;
		}
		
		private static function increment(value:Number, min:Number, max:Number):Number
		{
			return value >= max ? min : value + 1;
		}
		
		private static function decrement(value:Number, min:Number, max:Number):Number
		{
			return value <= min ? max : value - 1;
		}
		
	}
}