/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.supportClasses
{
	import com.adobe.portfolio.api.FieldType;
	import com.adobe.portfolio.api.FieldValueChangedEvent;
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.api.IField;
	import com.adobe.portfolio.utils.FieldUtils;
	
	import flash.globalization.DateTimeStyle;
	
	import flashx.textLayout.elements.ParagraphElement;
	import flashx.textLayout.elements.SpanElement;
	import flashx.textLayout.formats.WhiteSpaceCollapse;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.events.PropertyChangeEvent;
	
	import spark.events.TextOperationEvent;
	import spark.utils.TextFlowUtil;
	
	/**
	 * A subclass of <code>CustomizableTextArea</code> that's specialized for displaying and editing
	 * attachment field names and values.
	 * 
	 * @see CustomizableTextArea
	 * @see IField
	 */
	[SkinState("editable")]
	public class FieldTextArea extends CustomizableTextArea
	{
		/**
		 * Default constructor
		 */
		public function FieldTextArea()
		{
			super();
			multiline = true;
			allowStyling = false;
		}
		
		private var _dateStyle:String = DateTimeStyle.SHORT;
		/**
		 * The style to use for the date when displaying date fields.
		 * 
		 * @see DateTimeStyle
		 */
		public function get dateStyle():String
		{
			return _dateStyle;
		}
		/**
		 * @private
		 */
		public function set dateStyle(value:String):void
		{
			if (value != _dateStyle) {
				_dateStyle = value;
				if (!pendingChanges) {
					setTextFromData();
				}
			}
		}
		
		private var _timeStyle:String = DateTimeStyle.MEDIUM;
		/**
		 * The style to use for the time when displaying date fields.
		 * 
		 * @see DateTimeStyle
		 */
		public function get timeStyle():String
		{
			return _timeStyle;
		}
		/**
		 * @private
		 */
		public function set timeStyle(value:String):void
		{
			if (value != _timeStyle) {
				_timeStyle = value;
				if (!pendingChanges) {
					setTextFromData();
				}
			}
		}
		
		private var _attachment:IAttachment;
		/**
		 * The data object whose field should be displayed. Only IAttachment objects are supported.
		 */
		[Bindable]
		public function get data():Object
		{
			return _attachment;
		}
		/**
		 * @private
		 */
		public function set data(value:Object):void
		{
			var newAttachment:IAttachment = value as IAttachment;
			if (_attachment != newAttachment) {
				if (_attachment != null) {
					_attachment.removeEventListener("fieldValueChanged", onFieldValueChanged);
					_attachment.removeEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onPropertyChanged);
				}
				_attachment = newAttachment;
				if (_attachment != null) {
					_attachment.addEventListener("fieldValueChanged", onFieldValueChanged);
					_attachment.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onPropertyChanged);
				}
				_fieldChanged = true;
				invalidateProperties();
			}
		}
		
		// Are we showing/editing the field's display name
		// rather than the field's value
		private var _editFieldDisplayName:Boolean = false;
		
		/**
		 * True if the display name of the field should be displayed, rather than the field value.
		 */
		public function get editFieldDisplayName():Boolean
		{
			return _editFieldDisplayName;
		}
		/**
		 * @private
		 */
		public function set editFieldDisplayName(value:Boolean):void
		{
			if (_editFieldDisplayName != value) {
				_editFieldDisplayName = value;
				_fieldChanged = true;
				invalidateProperties();
			}
		}
		
		private var _field:IField;
		/**
		 * The field whose value is to be displayed.
		 */
		public function get field():IField
		{
			return _field;
		}
		/**
		 * @private
		 */
		public function set field(value:IField):void
		{
			if (_field != value) {
				_field = value;
				_fieldChanged = true;
				invalidateProperties();
			}
		}
		
		private var _fieldName:String;
		/**
		 * The name of the field whose value is to be displayed.
		 */
		[Bindable]
		public function get fieldName():String
		{
			return _fieldName;
		}
		/**
		 * @private
		 */
		public function set fieldName(value:String):void
		{
			if (_fieldName != value) {
				_fieldName = value;
				_fieldNameChanged = true;
				invalidateProperties();
			}
		}
		
		private var _fieldChanged:Boolean;
		private var _fieldNameChanged:Boolean;
		private var _fieldsChangeWatcher:ChangeWatcher;
		
		/**
		 * @inheritDoc
		 */
		override protected function commitProperties():void
		{
			if (_fieldNameChanged) {
				var newField:IField = FieldUtils.getFieldByFieldName(collectionManager.fields, _fieldName);
				if ((newField == null) && (_fieldsChangeWatcher == null)) {
					// listen for fields array change
					_fieldsChangeWatcher = BindingUtils.bindSetter(fieldsChanged, collectionManager, "fields");
				}
				if (field != newField) {
					// Setting field will also result in _fieldChanged being set to true
					field = newField;
				}
				_fieldNameChanged = false;
			}
			if (_fieldChanged) {
				setTextFromData();
				_fieldChanged = false;
			}
			updateEditable();
			
			// Wait to call super.commitProperties until after updateEditable() has invalidated everything it needs to
			super.commitProperties();
		}
		
		private function fieldsChanged(value:Array):void
		{
			var newField:IField = FieldUtils.getFieldByFieldName(collectionManager.fields, _fieldName);
			if ((newField != null) && (newField != field)) {
				field = newField;
				if (_fieldsChangeWatcher != null) {
					_fieldsChangeWatcher.unwatch();
					_fieldsChangeWatcher = null;
				}
			}
		}
		
		private function updateToolTip():void
		{
			if (text != null) {
				toolTip = text;
			}
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function onTextChanged(event:TextOperationEvent):void
		{
			super.onTextChanged(event);
			updateToolTip();
		}
		
		/**
		 * Updates the displayed text based on the current field value.
		 * 
		 * @inheritDoc
		 */
		override protected function setTextFromData():void
		{
			if (field)
			{
				if (editFieldDisplayName) {
					text = field != null ? field.displayName : "";
				} else if (allowStyling) {
					var styled:XML = FieldUtils.getStyledTextFromField(_attachment, field);
					if (styled)
					{
						textFlow = TextFlowUtil.importFromXML(styled, WhiteSpaceCollapse.PRESERVE);
					} else {
						textFlow = null;
						text = "";
					}
				} else {
					text = FieldUtils.getTextFromField(_attachment, field, dateStyle, timeStyle);
				}
				pendingChanges = false;
			} else {
				textFlow = null;
				text = "";
				pendingChanges = false;
			}
			setTextDirection(text);
			
			updateToolTip();
		}
		
		/**
		 * Updates the field value based on the current text.
		 * 
		 * @inheritDoc
		 */ 
		override protected function setDataFromText():void
		{
			if (!pendingChanges)
				return;
			
			// Make sure the tooltip is in sync
			updateToolTip();

			// Note that the order of setting pendingChanges to false before calling
			// setFieldDisplayNameFromText, setFieldFromText and setFieldValue is intentional.
			// In at least one case (negative number input) the user is allowed to enter the
			// number in multiple formats, but the output needs to change to the required
			// format. If pendingChanges isn't changed to false until after the fieldUtils
			// calls then the change isn't detected/updated in onFieldValueChanged.
			
			if (editFieldDisplayName) {
				if (field != null && collectionManager.host != null) {
					pendingChanges = false;
					if (!FieldUtils.setFieldDisplayNameFromText(collectionManager.host.collection, field.name, text))
						setTextFromData();
				}
				return;
			}

			// Check for null attachment after checking for editing the field display name since
			// editing the field display name does not need an attachment (it is passed a field).
			if (_attachment == null)
				return;

			if (!allowStyling) {
				pendingChanges = false;
				FieldUtils.setFieldFromText(_attachment, field, text);
				return;
			}
			var styled:XML;
			if (textFlow) {
				// Fix bug 2707602 (which may be a side effect of deferred bug 2689558)
				// The IME is adding entries for its own use into the span elements
				// Remove them before writing out the styled data, otherwise the
				// text will always look like it is being edited in the IME.
				for (var peIndex:int = 0; peIndex < textFlow.numChildren; ++peIndex) {
					var paraElement:ParagraphElement = textFlow.getChildAt(peIndex) as ParagraphElement;
					if (!paraElement)
						continue;
					for (var seIndex:int = 0; seIndex < paraElement.numChildren; ++seIndex) {
						var spanElement:SpanElement = paraElement.getChildAt(seIndex) as SpanElement;
						if (!spanElement)
							continue;
						spanElement.userStyles = null;
					}
				}
				styled = TextFlowUtil.export(textFlow);
			}
			else
				styled = new XML( <TextFlow>{text}</TextFlow> );
			
			pendingChanges = false;
			_attachment.setFieldValue(field.name, styled);
		}
		
		/**
		 * @inheritDoc
		 */
		override public function removeListenAndWatch():void
		{
			super.removeListenAndWatch();
			if (_attachment != null) {
				_attachment.removeEventListener("fieldValueChanged", onFieldValueChanged);
				_attachment.removeEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onPropertyChanged);
			}
		}
		
		private function onFieldValueChanged(event:FieldValueChangedEvent):void
		{
			var fieldMatch:Boolean = (event.fieldName == _fieldName);
			if (!fieldMatch && (field != null) && (field.name == event.fieldName)) {
				fieldMatch = true;
			}
			if (fieldMatch) {
				if (pendingChanges) {
					// We've got pending changes, so just keep what we've got
				} else {
					// No pending changes, so no conflict. Just update our value
					setTextFromData();
				}
			}
		}
		
		// This is needed to detect when a file name is being restored to its
		// original value after the user has attempted to change the name to
		// one with an invalid character in it. I generalized it to include
		// the description field (the other attachment 'field' property that
		// users can change), although it should never be a problem because
		// the C++ side accepts all input for the description field.
		private function onPropertyChanged(event:PropertyChangeEvent):void
		{
			var propFieldName:String;
			switch(event.property) {
				case 'description':
					propFieldName = FieldType.DESCRIPTION;
					break;
				case 'fileName':
					propFieldName = FieldType.FILE_NAME;
					break;
				default:
					return;
			}
			if (propFieldName == (field ? field.name : _fieldName)) {
				if (pendingChanges) {
					// We've got pending changes, so just keep what we've got
				} else {
					// No pending changes, so no conflict. Just update our value
					setTextFromData();
				}
			}
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function shouldBeEditable():Boolean
		{
			if (super.shouldBeEditable()) {
				if (field != null) {
					// In the header, only user defined field names are editable
					if (editFieldDisplayName)
						return FieldUtils.isUserDefinedField(field);
					// User defined date fields are never directly editable
					if (field.type == FieldType.DATE)
						return false;
					return field.editable;
				}
			}
			return false;
		}
	}
}
