/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.supportClasses
{
	import com.adobe.portfolio.managers.FlashSkinManager;
	import com.adobe.portfolio.utils.LogHelper;
	
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.DisplayObject;
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.geom.Matrix;
	import flash.geom.Rectangle;
	
	import mx.controls.Image;
	import mx.events.ResizeEvent;
	import mx.logging.ILogger;
	import mx.logging.Log;
	
	import spark.components.SkinnableContainer;
	
	/**
	 * A FlashContainer is used to reference an object type in an external SWF file.
	 * This allows assets to be accessed dynamically rather than requiring them as a SWC
	 * at compile time. If the referenced object cannot be found, then the contents of
	 * the container are displayed instead.
	 */
	public class FlashContainer extends SkinnableContainer implements IFlashComponent
	{
		/** Logger used by all NavFramework components.
		 */
		private static const logger:ILogger = Log.getLogger("NavComponents");

		/**
		 * Default constructor.
		 */
		public function FlashContainer()
		{
			super();
			addEventListener(ResizeEvent.RESIZE, onResize, false, 0, true);
			FlashSkinManager.instance.addEventListener(Event.COMPLETE, onSkinsLoaded, false, 0, true);
			skinsAreLoaded = FlashSkinManager.instance.skinsAreLoaded();
		}
		
		/**
		 * Measures the contents of this Flex container. The SWF object is not
		 * considered when measuring.
		 * 
		 * @inheritDoc
		 */
		override protected function measure():void
		{
			super.measure();
			if (contentGroup != null) {
				measuredWidth = contentGroup.measuredWidth;
				measuredHeight = contentGroup.measuredHeight;
				measuredMinWidth = contentGroup.measuredMinWidth;
				measuredMinHeight = contentGroup.measuredMinHeight;
			}
		}
		
		private var _useFlashAsset:Boolean = false;
		/**
		 * True if the referenced SWF object should be displayed; false otherwise.
		 */
		[Bindable]
		public function get useFlashAsset():Boolean
		{
			return _useFlashAsset;
		}
		public function set useFlashAsset(value:Boolean):void
		{
			if (_useFlashAsset != value) {
				_useFlashAsset = value;
			}
		}
		
		/**
		 * An image skin part where the SWF object will be rendered.
		 */
		[SkinPart(required="true")]
		public var flashImage:Image = null;
		
		private var _swfClass:String = "";
		
		/**
		 * @inheritDoc
		 * 
		 * @see FlashSkinManager
		 */
		public function get swfClass():String
		{
			return _swfClass;
		}
		public function set swfClass(val:String):void
		{
			if (_swfClass != val) {
				skinObj = null;
				_swfClass = val;
				initSkinObj();
			}
		}
		
		/**
		 * The DisplayObject in the swf that should be rendered by this component.
		 */
		protected var skinObj:DisplayObject = null;

		/**
		 * An event listener to respond when a swf is loaded or unloaded.
		 */
		protected function onSkinsLoaded(evt:Event):void
		{
			skinsAreLoaded = FlashSkinManager.instance.skinsAreLoaded();
		}
		
		/**
		 * An event listener to respond when this element is resized. This is important when skinObj
		 * supports the 9-slice skinning model, since a simple scaling is not sufficient when redrawing.
		 */
		protected function onResize(evt:ResizeEvent):void
		{
			setSource();
		}
		
		private var _skinsAreLoaded:Boolean = false;
		
		/**
		 * An internal container sprite used to host skinObj, the visual asset from the swf.
		 *  This is necessary since a DisplayObject cannot be added directly to a Flex container.
		 */
		private var _sprite:Sprite = new Sprite();
		
		private var _skinBitmap:Bitmap = null;
		
		/**
		 * A Boolean indicating whether or not the skins contained in the swf have been loaded.
		 */
		protected function get skinsAreLoaded():Boolean
		{
			return _skinsAreLoaded;
		}
		protected function set skinsAreLoaded(val:Boolean):void
		{
			if (val != _skinsAreLoaded) {
				_skinsAreLoaded = val;
				initSkinObj();
			}
		}
		
		/**
		 * Initializes skinObj based on the current loaded swf and swfClass.
		 */
		protected function initSkinObj():void
		{
			skinObj = _skinsAreLoaded ? FlashSkinManager.instance.getSkin(swfClass) : null;
			setSource();
		}

		/**
		 * Sets the source of this component appropriately based on the availability of skinObj.
		 * If skinObj is non-null, then it is sized properly and used as the source. Otherwise the
		 * _backup Bitmap is used if it is non-null.
		 */
		protected function setSource():void
		{
			// Empty the current _sprite object, although it shouldn't be necessary
			if (_sprite.numChildren > 0) {
				CONFIG::logging{
					LogHelper.logger.warn("FlashContainer:setSource() - expected 0 children in _sprite");
				}
				while (_sprite.numChildren > 0) {
					_sprite.removeChildAt(0);
				}
			}
			if ((flashImage != null) && (skinObj != null)) {
				var bitmapWidth:int = (width > 0) ? width : 1;
				var bitmapHeight:int = (height > 0) ? height : 1;
				
				if ((_skinBitmap == null) || (_skinBitmap.bitmapData == null)) {
					_skinBitmap = new Bitmap(new BitmapData(bitmapWidth, bitmapHeight, true, 0x00000000));
				} else if ((_skinBitmap.bitmapData.width != width) || (_skinBitmap.bitmapData.height != height)) {
					var oldData:BitmapData = _skinBitmap.bitmapData
					_skinBitmap.bitmapData = new BitmapData(bitmapWidth, bitmapHeight, true, 0x00000000);
					oldData.dispose();
				} else {
					_skinBitmap.bitmapData.fillRect(new Rectangle(0, 0, bitmapWidth, bitmapHeight), 0x00000000);
				}
				if ((width > 0) && (height > 0)) {
					// Add skinObj to _sprite so that we can resize/reposition it
					_sprite.addChild(skinObj);
					
					if (skinObj.scale9Grid != null) {
						// Match the skinObj width and height to fill this entire component.
						skinObj.width = width;
						skinObj.height = height;
					} else {
						// This isn't a 9-slice asset, so maintain its aspect ratio
						var skinAspectRatio:Number = skinObj.width / skinObj.height;
						if ((width / skinAspectRatio) > height) {
							skinObj.width = height * skinAspectRatio;
							skinObj.height = height;
						} else {
							skinObj.width = width;
							skinObj.height = width / skinAspectRatio;
						}
					}
					
					// Offset skinObj so that it's bounds line up with the bounds of this component.
					var bounds:Rectangle = _sprite.getBounds(_sprite);
					var ratio:Number = width / bounds.width;
					var drawMatrix:Matrix = new Matrix(ratio, 0, 0, ratio, -bounds.left * ratio, -bounds.top * ratio);

					// Draw _sprite into the active bitmap; _sprite is not drawn directly
					_skinBitmap.bitmapData.draw(_sprite, drawMatrix);
					
					// Remove skinObj from _sprite so that it can be used by other instances
					_sprite.removeChildAt(0);
				}
				flashImage.source = _skinBitmap;
				useFlashAsset = true;
			} else {
				useFlashAsset = false;
			}
		}
	}
}