/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.supportClasses
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.layouts.FreeFormLayout;
	import com.adobe.portfolio.layouts.supportClasses.ICardProperties;
	
	import flash.events.EventDispatcher;
	import flash.geom.Point;
	
	import mx.logging.ILogger;
	import mx.logging.Log;
	
	import spark.components.supportClasses.GroupBase;
	
	/**
	 *  The FreeFormField class extends EventDispatcher to provide functionality specific to the Freeform navigator.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class FreeFormFields extends EventDispatcher implements ICardProperties
	{
		/**
		 *  Field name of the pdf file field to get/set the freeform card's x position.
		 *  The x position is relative to the center of the card.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const CARDS_X_POS:String = "adobe:FreeForm:x";
		
		/**
		 *  Field name of the pdf file field to get/set the freeform card's y position.
		 *  The y position is relative to the center of the card.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const CARDS_Y_POS:String = "adobe:FreeForm:y";
		
		/**
		 *  Field name of the pdf file field to get/set the freeform card's depth.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const CARDS_DEPTH:String = "adobe:FreeForm:depth";

		/**
		 *  Field name of the pdf file field to get/set the freeform card's rotation.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const CARDS_ROTATION:String = "adobe:FreeForm:angle";
		
		/**
		 *  The ILogger instance used by this component.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected static const logger:ILogger = Log.getLogger("FreeFormFields");
		
		private static const DEFAULT_CARDS_X_POS:Number = 0;
		private static const DEFAULT_CARDS_Y_POS:Number = 0;
		private static const DEFAULT_CARDS_DEPTH:Number = 0;
		private static const DEFAULT_CARDS_ROTATION:Number = 0;
		private static const BORDER_THICKNESS:Number = 10;
		
		// Note: These values can never change, because the x/y positions
		// saved in portfolios are saved relative (scaled) to these values.
		private static const VIRTUAL_CANVAS_WIDTH:Number = 2048;
		private static const VIRTUAL_CANVAS_HEIGHT:Number = 2048;
		
		/**
		 *  Constructor.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function FreeFormFields()
		{
			super();
		}

		
		private var _target:GroupBase = null;

		/**
		 *  The target component 
		 *  @see spark.components.supportClasses.GroupBase
		 * 
		 *  @return The target component that displays visual elements
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get target():GroupBase
		{
			return _target;
		}

		/**
		 *  @private
		 */ 
		public function set target(value:GroupBase):void
		{
			if (value == _target)
				return;
			_target = value;
		}
		
		private var _cardWidth:Number = FreeFormLayout.DEFAULT_CARD_WIDTH;

		/**
		 *  The card's current width 
		 * 
		 *  @return The width of the card
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get cardWidth():Number
		{
			return _cardWidth;
		}

		/**
		 *  @private
		 */ 
		public function set cardWidth(value:Number):void
		{
			_cardWidth = value;
		}

		private var _cardHeight:Number = FreeFormLayout.DEFAULT_CARD_HEIGHT;

		/**
		 *  The card's current height 
		 * 
		 *  @return The height of the card
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get cardHeight():Number
		{
			return _cardHeight;
		}

		/**
		 *  @private
		 */ 
		public function set cardHeight(value:Number):void
		{
			_cardHeight = value;
		}
		
		/**
		 *  The width and height of the target
		 * 
		 *  @return An ordered pair (width, height)
		 *  representing the width and height of the target in pixels
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function get targetSize():Point 
		{
			if (!target)
				return null;
			
			return new Point(target.width, target.height);
		}
		
		/**
		 *  Get the card layout location for the specified attachment.
		 * 
		 *  @param attachment The IAttachment for which you want to find
		 *  the layout location
		 *  @return The layout location of the center of the card
		 *  given as an ordered pair: (x,y)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function getLocation(attachment:IAttachment):Point 
		{
			var p:Point = new Point(
				attachment.getFieldValue(CARDS_X_POS),
				attachment.getFieldValue(CARDS_Y_POS));
			if (isNaN(p.length))
				return null;
			p.x *= (target.width / VIRTUAL_CANVAS_WIDTH);
			p.y *= (target.height / VIRTUAL_CANVAS_HEIGHT);
			
			return p;
		}
		
		/**
		 *  Indicates whether the location of
		 *  the card for the specified attachment has been set or not.
		 * 
		 *  @param attachment The IAttachment for which you want to know
		 *  if a location has been set for its card
		 *  @return <code>true</code> if the location of the card for the
		 *  specified IAttachment has been set; returns <code>false</code>
		 *  if it has not been set
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function isLocationSet(attachment:IAttachment):Boolean
		{
			if (!attachment)
				return false;
			
			return !isNaN(attachment.getFieldValue(FreeFormFields.CARDS_X_POS)) 
				&& !isNaN(attachment.getFieldValue(FreeFormFields.CARDS_Y_POS));
		}
		
		
		/**
		 *  Get the x value from the specified attachment's x position field.
		 *  If the field has not been set, the default 0 is returned.
		 * 
		 *  @param attachment The IAttachment for which you want to find
		 *  the x position of its card layout
		 *  @return The x position
		 * 	
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getX(attachment:IAttachment):Number
		{
			if (!target) {
				logger.error("getX - target is not set");
				return DEFAULT_CARDS_X_POS;
			}
			if (!attachment) {
				logger.error("getX - attachment is not set");
				return DEFAULT_CARDS_X_POS;
			}
			// Get the stored value
			var xPos:Number = attachment.getFieldValue(CARDS_X_POS);
			// If one existed, scale it to the canvas, else return a default
			var xScale:Number= target.width / VIRTUAL_CANVAS_WIDTH;
			return isNaN(xPos) ? DEFAULT_CARDS_X_POS : (xPos * xScale);
		}
		
		/**
		 *  Get the y value from the specified attachment's y position field.
		 *  If the field has not been set, the default 0 is returned.
		 * 
		 *  @param attachment The IAttachment for which you want to find
		 *  the y position of its card layout
		 *  @return The y position
		 * 	
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getY(attachment:IAttachment):Number
		{
			if (!target) {
				logger.error("getY - target is not set");
				return DEFAULT_CARDS_Y_POS;
			}
			if (!attachment) {
				logger.error("getY - attachment is not set");
				return DEFAULT_CARDS_Y_POS;
			}
			// Get the stored value
			var yPos:Number = attachment.getFieldValue(CARDS_Y_POS);
			// If one existed, scale it to the canvas, else return a default
			var yScale:Number= target.height / VIRTUAL_CANVAS_HEIGHT;
			return isNaN(yPos) ? DEFAULT_CARDS_Y_POS : (yPos * yScale);
		}
		
		/**
		 *  Get the rotation value from the specified attachment's rotation field.
		 *  If the field has not been set, the default 0 is returned.
		 * 
		 *  @param attachment The IAttachment for which you want to find
		 *  the rotation of its card layout
		 *  @return The rotation
		 * 	
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getRotation(attachment:IAttachment):Number
		{
			if (!attachment) {
				logger.error("getRotation - attachment is null");
				return DEFAULT_CARDS_ROTATION;
			}
			// Get the stored value
			var cardsRotation:Number = attachment.getFieldValue(CARDS_ROTATION);
			// If none existed, return a default
			return isNaN(cardsRotation) ? DEFAULT_CARDS_ROTATION : cardsRotation;
		}
		
		/**
		 *  Get the depth value from the specified attachment's depth field.
		 *  If the field has not been set, the default 0 is returned.
		 * 
		 *  @param attachment The IAttachment for which you want to find
		 *  the depth of its card layout
		 *  @return The depth
		 * 	
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getDepth(attachment:IAttachment):Number
		{
			if (!attachment) {
				logger.error("getDepth - attachment is null");
				return DEFAULT_CARDS_DEPTH;
			}
			// Get the stored value
			var cardsDepth:Number = attachment.getFieldValue(CARDS_DEPTH);
			// If none existed, return a default
			return isNaN(cardsDepth) ? DEFAULT_CARDS_DEPTH : cardsDepth;
		}
		
		/**
		 *  Set the x value for the specified attachment's x position field.
		 *  The new value is constrained to the target (greater or equal to zero and less than or equal to target width). 
		 * 
		 *  @param attachment The IAttachment for which you want to set
		 *  the x position of its card layout
		 *  @param value The x position
		 * 	
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function setX(attachment:IAttachment, value:Number):void
		{
			if (!target) {
				logger.error("setX - target is not set");
				return;
			}
			if (target.width == 0) {
				// Avoid divide by zero below
				logger.error("setX - target width is 0");
				return;
			}
			if (!attachment) {
				logger.error("setX - attachment is not set");
				return;
			}
			if (isNaN(value)) {
				logger.error("setX - value cannot be NaN");
				return;
			}
			// Ensure that the new position is on the canvas
			value = Math.max(value, 0);
			value = Math.min(value, target.width);
			// Scale the x pos to the virtual canvas
			var xScale:Number = target.width / VIRTUAL_CANVAS_WIDTH;
			attachment.setFieldValue(CARDS_X_POS, value / xScale);
		}
		
		/**
		 *  Set the y value for the specified attachment's y position field.
		 *  The new value is constrained to the target (greater or equal to zero and less than or equal to target height). 
		 * 
		 *  @param attachment The IAttachment for which you want to set
		 *  the y position of its card layout
		 *  @param value The y position
		 * 	
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function setY(attachment:IAttachment, value:Number):void
		{
			if (!target) {
				logger.error("setY - target is not set");
				return;
			}
			if (target.height == 0) {
				// Avoid divide by zero below
				logger.error("setY - target height is 0");
				return;
			}
			if (!attachment) {
				logger.error("setY - attachment is not set");
				return;
			}
			if (isNaN(value)) {
				logger.error("setY - value cannot be NaN");
				return;
			}
			// Ensure that the new position is on the canvas
			value = Math.max(value, 0);
			value = Math.min(value, target.height);
			// Scale the y pos to the virtual canvas
			var yScale:Number = target.height / VIRTUAL_CANVAS_HEIGHT;
			attachment.setFieldValue(CARDS_Y_POS, value / yScale);
		}
		
		/**
		 *  Sets the card layout location for the specified attachment.
		 * 
		 *  @param attachment The IAttachment for which you want to set
		 *  the layout location
		 *  @param value The layout location of the center of the card
		 *  given as an ordered pair: (x,y)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function setLocation(attachment:IAttachment, value:Point):void
		{
			setX(attachment, value.x);
			setY(attachment, value.y);
		}
		
		/**
		 *  Sets the rotation of the card layout for the specified attachment
		 * 
		 *  @param attachment The IAttachment for which you want to set
		 *  the rotation of its card layout
		 *  @param value The rotation of the card layout in degrees
		 *  of rotation from the horizontal
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function setRotation(attachment:IAttachment, value:Number):void
		{
			if (!attachment) {
				logger.error("setRotation - attachment is null");
				return;
			}
			if (isNaN(value)) {
				logger.error("setRotation - value cannot be NaN");
				return;
			}
			attachment.setFieldValue(CARDS_ROTATION, value);
		}
		
		/**
		 *  Set the depth value for the specified attachment's depth field.
		 * 
		 *  @param attachment The IAttachment for which you want to set
		 *  the depth of its card layout
		 *  @param value The depth
		 * 	
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function setDepth(attachment:IAttachment, value:Number):void
		{
			if (!attachment) {
				logger.error("setDepth - attachment is null");
				return;
			}
			if (isNaN(value)) {
				logger.error("setDepth - value cannot be NaN");
				return;
			}
			attachment.setFieldValue(CARDS_DEPTH, value);
		}
	}
}
