/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.components.supportClasses
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.managers.CollectionManager;
	
	import flash.utils.Dictionary;
	
	import spark.components.DataGroup;
	
	/**
	 * The NavContainerDataGroup class extends DataGroup to provide functionality specific to Portfolio navigators.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class NavContainerDataGroup extends DataGroup
	{
		/** 
		 *  The navigator's CollectionManager instance. All host interaction should go through this
		 *  instance, rather than the INavigatorHost directly.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public const collectionManager:CollectionManager = CollectionManager.instance;
		
		private var _restoreScrollPositionsOnFolderChange:Boolean;
		protected var currentFolder:IAttachment;
		protected var verticalScrollPositionDictionary:Dictionary = new Dictionary();
		
		/**
		 * Constructor.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function NavContainerDataGroup()
		{
			super();
		}
		
		[Bindable]
		public function get restoreScrollPositionsOnFolderChange():Boolean
		{
			return _restoreScrollPositionsOnFolderChange;
		}

		public function set restoreScrollPositionsOnFolderChange(value:Boolean):void
		{
			if (value)
				collectionManager.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onHostEvent);
			else
				collectionManager.removeEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onHostEvent);
			_restoreScrollPositionsOnFolderChange = value;
		}

		/**			 
		 * DataGroup's commitProperties resets the horizontal and vertical scroll positions to 0 when the dataprovider changes.
		 * It also resets the horizontal and vertical scroll positions to 0 when the dataProvider's IList is refreshed or reset.
		 * The dataProvider's IList (in our case, collection manager's attachments) is reset for a number of reasons, such as
		 * when the sort order changes, or the user edits and changes a field value. Since we don't want to scroll to the
		 * beginning of the list after actions like changing the sort order or editing a field, prevent commitProperties from
		 * resetting the horizontal and vertical scroll positions to 0.
		 * 
		 * 	@langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override protected function commitProperties():void
		{
			
			/*
			The code below was originally a workaround for the following bug: https://bugs.adobe.com/jira/browse/SDK-23175
			(which has been fixed) but does what we still want to do.
			*/
			
			// Remember the scroll positions
			var oldHorizontalScrollPosition:Number = horizontalScrollPosition;
			var oldVerticalScrollPosition:Number = verticalScrollPosition;
			
			super.commitProperties();
			
			// If scroll positions changed, restore them
			if (horizontalScrollPosition != oldHorizontalScrollPosition)
				horizontalScrollPosition = oldHorizontalScrollPosition;
			if (verticalScrollPosition != oldVerticalScrollPosition)
				verticalScrollPosition = oldVerticalScrollPosition;
		}
		
		/** An event listener to respond to events from the collectionManager.
		 *  In this case we are only interested in the "ITEMS_DID_CHANGE" event,
		 *  which occurs after a folder change. When this happens, we save the
		 *  vertical scroll position of the current folder (or root) and then set
		 *  up a delayed call to set/restore the vertical scroll position of
		 *  the new folder (or root).
		 */
		protected function onHostEvent(event:CollectionManagerEvent):void
		{
			// We only are concerned with "ITEMS_DID_CHANGE" events
			if (event.kind != CollectionManagerEvent.ITEMS_DID_CHANGE_KIND)
				return;
			
			// Save the previous folder and get the current folder
			var previousFolder:IAttachment = currentFolder;
			currentFolder = collectionManager.currentFolder;
			
			// Do nothing if they are the same (which happens the first time we are called)
			if (currentFolder == previousFolder)
				return;
			
			// Save the vertical scroll position of the previous folder (or root), if
			// we are traversing down the folder hierarchy (but not if going up)
			if (currentFolder && (currentFolder.parent == previousFolder))
				verticalScrollPositionDictionary[folderKey(previousFolder)] = verticalScrollPosition;
			
			var currentFolderKey:String = folderKey(currentFolder);
			
			// Get the last vertical scroll position of the new folder (or root)
			var newVerticalScrollPosition:Number = verticalScrollPositionDictionary[currentFolderKey];
			// If there was no previous vertical scroll position, set it to zero [fixes bug 2748121]
			if (isNaN(newVerticalScrollPosition)) {
				newVerticalScrollPosition = 0;
			}
			// If there was a previous vertical scroll position, use it and remove it from the dictionary
			// Doing this enforces the behavior that entering a folder always starts at the beginning,
			// but exiting a folder returns to the vertical scroll position saved above.
			else {
				delete verticalScrollPositionDictionary[currentFolderKey];
			}
			// If we're back to the root, remove all entries from the dictionary
			// The code above (deleting an entry at a time) works if "folder back" is used to traverse up folders,
			// but would leave around entries if "folder breadcrumbs" was used. This cleans up in those cases.
			if (currentFolder == null) {
				verticalScrollPositionDictionary = new Dictionary();
			}
			// Wait to actually set the vertical scroll position. Otherwise the target hasn't quite switched
			// over to using the new values and the index based on the scroll position will not be found.
			callLater(setVerticalScrollPosition, [newVerticalScrollPosition]);
		}
		
		protected function setVerticalScrollPosition(value:Number):void 
		{
		 	verticalScrollPosition = value;
		}
		
		private function folderKey(folder:IAttachment):String
		{
			if (folder == null)
				return null;
			var key:String = folder.fileName;
			for (var parent:IAttachment = folder.parent; parent != null; parent = parent.parent) {
				key = parent.fileName + "/" + key;
			}
			return key;
		}
	}
}