/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.layouts
{
	import com.adobe.portfolio.components.NavListDropIndicator;
	import com.adobe.portfolio.components.supportClasses.AutoListen;
	import com.adobe.portfolio.components.supportClasses.AutoWatch;
	import com.adobe.portfolio.events.LayoutEvent;
	import com.adobe.portfolio.layouts.supportClasses.CardLayoutHelper;
	import com.adobe.portfolio.layouts.supportClasses.ICardLayout;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.PropertyManager;
	import com.adobe.portfolio.properties.CommonProperties;
	
	import flash.display.DisplayObject;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	import mx.binding.utils.BindingUtils;
	import mx.core.IVisualElement;
	import mx.effects.Effect;
	import mx.events.ResizeEvent;
	import mx.logging.ILogger;
	import mx.logging.Log;
	
	import spark.components.supportClasses.GroupBase;
	import spark.layouts.TileLayout;
	import spark.layouts.supportClasses.DropLocation;
	
	/**
	 *  This event is dispatched after the layout has set its focus to a valid attachment.
	 *
	 *  @eventType com.adobe.portfolio.events.NavLayoutEvent.HANDLED_ATTACHMENT_FOCUS
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	[Event(name="HANDLED_ATTACHMENT_FOCUS", type="com.adobe.portfolio.events.LayoutEvent")]

	/**
	 * 
	 * The BasicTileLayout class extends TileLayout to implement a tile-based layout that 
	 * centers the group of tiles in the layout container.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class BasicTileLayout extends TileLayout implements ICardLayout
	{
		protected static const IBEAM_SIZE:Number = 2;
		protected static const IBEAM_PADDING:Number = 30;

		protected const collectionManager:CollectionManager = CollectionManager.instance;
		protected const logger:ILogger = Log.getLogger("BasicTileLayout");

		protected var autoListen:AutoListen;
		protected var autoWatch:AutoWatch;
		protected var fadeDelay:Number;
		
		private var _computedRowHeight:Number;
		private var _computedColumnWidth:Number;
		
		/**
		 *  @private
		 *  The x and y offset are simply the distance from the top left corner of the container to the top left corner of the layout.
		 *  This allows to center the tiles both vertically and horizontally in the container.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		private var xOffset:Number;
		private var yOffset:Number;
		
		private var _cardLayoutHelper:CardLayoutHelper;
		private var _thumbFitSize:Point;
		
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function BasicTileLayout()
		{
			super();
			xOffset = yOffset = 0;
			autoListen = new AutoListen();
			autoWatch = new AutoWatch();
			autoWatch.addWatcher(BindingUtils.bindSetter(cardSizeChanged, PropertyManager.instance, ["commonProperties","cardSize", "numValue"]));
			_cardLayoutHelper = new CardLayoutHelper(this);
			horizontalGap = 15;
			verticalGap = 15;
		}
		
		/**
		 *  Call this function when the card size has changed to
		 *  set the proper computedColumnWidth and computedRowHeight for the layout
		 * 
		 *  @param value the new width and height of the tile layout, in pixels
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function cardSizeChanged(value:Number):void {
			if (isNaN(value) || (value <= 0)) {
				// Let the card lay itself out using its default measured dimensions
				// TODO If this causes behavioral problems, then value can be set to
				//		BasicCard.DEFAULT_SMALL_DIMENSION as an alternative.
				computedColumnWidth = Number.NaN;
				computedRowHeight = Number.NaN;
				return;
			}
			if ((PropertyManager.instance.commonProperties != null) &&
				(PropertyManager.instance.commonProperties.portfolioCardSizeList != null) &&
				(PropertyManager.instance.commonProperties.portfolioCardSizeList.activeValue != null) &&
				(PropertyManager.instance.commonProperties.portfolioCardSizeList.activeValue.name == CommonProperties.PROPERTY_CARD_SIZE_SMALL))
			{
				// Let the card lay itself out using its default measured dimensions
				// TODO 	If this causes behavioral problems, then this block can be removed entirely.
				computedColumnWidth = Number.NaN;
				computedRowHeight = Number.NaN;
				return;
			}
			computedColumnWidth = computedRowHeight = value;
		}
		
		/**
		 *  Set the focus attachment
		 * 
		 *  @param attachment The IAttachment that is to be the new focus
		 *  @param oldFocus The IAttachment that is the current (old) focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function setFocusAttachment(attachment:Object, oldFocus:Object = null):void
		{
			dispatchEvent(new LayoutEvent(LayoutEvent.HANDLED_ATTACHMENT_FOCUS));
		}
		
		/**
		 *  Generate a focus change effect
		 * 
		 *  @param oldIndex The index value of the element that is the current (old) focus
		 *  @param newIndex The index value of the element that is to be the new focus
		 *  @param duration The duration, in milliseconds, the effect should last
		 *  @return The focus effect object
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function generateFocusEffect(oldIndex:int, newIndex:int, duration:Number):Effect
		{
			return null;
		}
		
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function set target(value:GroupBase):void 
		{
			autoListen.removeListener(target, ResizeEvent.RESIZE);
			super.target = value;
			autoListen.addListener(target, ResizeEvent.RESIZE, onResize );
		}
		
		
		/**
		 *  Set pending fade to true on a resize
		 * 
		 *  @param event The resize event to handle
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onResize(event:ResizeEvent):void
		{
			cardLayoutHelper.pendingFade = true;
		}

		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function elementAdded(index:int):void {
			super.elementAdded(index);
			cardLayoutHelper.pendingFade = true;
		}
		
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function updateDisplayList(unscaledWidth:Number, unscaledHeight:Number):void 
		{
			if (unscaledWidth > IBEAM_PADDING) {
				// Leave a little extra room so that the IBeam doesn't get completely cropped
				unscaledWidth -= IBEAM_PADDING;
			}
			super.updateDisplayList(unscaledWidth, unscaledHeight);

			justifyCards();
			
			if (isNaN(fadeDelay))
				fadeDelay = 2000;
			cardLayoutHelper.fadeInCards(fadeDelay);
			fadeDelay = 0;
			
			cardLayoutHelper.enableThumbnailGeneration();
		}
		
		/**
		 *  The number of columns that will fit in the layout
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function get justifiedColumnCount():int 
		{
			if (!target)
				return columnCount;
			
			return Math.min(columnCount, target.numElements);
		}
		
		
		protected function justifyCards():void
		{
			var columnCount:int = justifiedColumnCount;
			var columnSize:Number = columnCount * columnWidth;
			var xGapSize:Number = (columnCount - 1) * horizontalGap;
			xOffset = (target.width - columnSize - xGapSize) / 2;
			
			var rowSize:Number = rowCount * rowHeight;
			var yGapSize:Number = (rowCount - 1) * verticalGap;
			yOffset = Math.max(0, (target.height - rowSize - yGapSize) / 2);
			
			var numElements : int = target.numElements;
			for (var i:int = 0; i < numElements; i++ ) {
				var element:IVisualElement = target.getElementAt(i);
				if (element == null) {
					continue;
				}
				element.setLayoutBoundsPosition(element.x + xOffset, element.y + yOffset);
			}
		}
		
		
		/**
		 *  Find the list of cards that overlap the specified rectangle 
		 * 
		 *  @param rect A bounding region for which you want to find cards that intersect it.
		 *  @return Vector A vector of integers that represent the indexes of the cards in the layout that intersect
		 *  the specified rectangle.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function findCardsThatIntersect(rect:Rectangle):Vector.<int>
		{
			return cardLayoutHelper.findCardsThatIntersect(rect);
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  TileLayout.measure() is bizarre!  It calls the layout with explicitWidth & explicitHeight - instead of width & height
		 *  this causes it to re-layout the tiles with the wrong layout - but not showing the wrong layout.  This in turns screws 
		 *  up the drag and drop code.  A call to updateDisplayList causes the layout to be corrected. TileLayout has its issues!

		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function measure():void {
			super.measure();
			updateDisplayList(target.width, target.height);
		}
		
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override protected function calculateDropIndex(x:Number, y:Number):int {
			var dropIndex:int = super.calculateDropIndex(x-xOffset, y-yOffset);
//			logger.debug("dropIndex= " + dropIndex);
			return dropIndex;
		}
		
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function set dropIndicator(value:DisplayObject):void
		{
			super.dropIndicator = value;
			customizeDropIndicator(value);
		}
			
		/**
		 *  Customize the dropIndicator icon to look like an I-beam.
		 * 
		 *  @param dropIndicator The displayObject instance to use as the dropIndicator icon
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function customizeDropIndicator(dropIndicator:DisplayObject):void
		{
			if (!dropIndicator)
				return;

			var listDropIndicator:NavListDropIndicator = dropIndicator as NavListDropIndicator;
			listDropIndicator.getGraphicPosition = getDropIndicatorPosition;
			
			// If no one sets this, the default is the first element.
			if (!typicalLayoutElement)
				return;
			var elementHeight:Number = typicalLayoutElement.getLayoutBoundsHeight();
			listDropIndicator.setMaxHeight(elementHeight);
		}
		
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function showDropIndicator(dropLocation:DropLocation):void
		{
			super.showDropIndicator(dropLocation);
			
			var dropIndex:int = dropLocation.dropIndex;
			if (dropIndex < 0)
				return;
			
			var element:IVisualElement = target.getElementAt(dropIndex);
			if (!element)
			{
				dropIndicator.height = 0;
				return;
			}

			dropIndicator.height = element.height;
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  The bounds are calculated without offsetting the dropPoint, the bounds itself is offsetted.
		 *  Called by the showDropIndicator() method. 
		 * 
		 *  @param dropLocation The drop location where the dropIndicator will be drawn.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override protected function calculateDropIndicatorBounds(dropLocation:DropLocation):Rectangle
		{
			dropLocation.dropPoint.x -= xOffset;
			dropLocation.dropPoint.y -= yOffset;
			var indicatorBounds:Rectangle = super.calculateDropIndicatorBounds(dropLocation);
			dropLocation.dropPoint.x += xOffset;
			dropLocation.dropPoint.y += yOffset;
			
			indicatorBounds.x += xOffset;
			indicatorBounds.y += yOffset;
			
			return indicatorBounds;
		}
		
		
		/**
		 *  Given a location, width, and height for the dropIndicator icon,
		 *  return the drawing rectangle for the dropIndicator icon.
		 * 
		 *  @param w The width, in pixels of the drop indicator icon
		 *  @param h The height, in pixels of the drop indicator icon
		 *  @param x The horizontal position of the drop indicator, in pixels
		 *  @param y The vertical position of the drop indicator, in pixels
		 *  @param end If this is the last element
		 *  @return A Rectangle that is the bounding box of the dropIndicator icon as it will be drawn.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getDropIndicatorPosition(w:Number, h:Number, x:Number, y:Number, end:Boolean):Rectangle
		{
			// If the gap is 0 or negative, the layout would size us as 0 width/height,
			// we need some minimum to ensure drawing.
			var iBeamWidth:Number = Math.max(IBEAM_SIZE, w);
			var iBeamHeight:Number = Math.max(IBEAM_SIZE, h);
			
			// Make the shorter side IBEAM_SIZE pixels so that the drop indicator is always 
			// a IBEAM_SIZE pixel thick line regardless of vertical/horizontal orientation.
			if (iBeamWidth < iBeamHeight)
				iBeamWidth = IBEAM_SIZE;
			else
				iBeamHeight = IBEAM_SIZE;
			
			// Center the drawing within the bounds
			var draw:Rectangle = new Rectangle();
			var gap:Number = Math.round(horizontalGap/2);
			if (x <= xOffset)
				draw.x = gap < xOffset ? -gap : -xOffset/2;
			else if (x + xOffset + horizontalGap >= target.width)
				draw.x = horizontalGap +  (gap < xOffset ? gap : xOffset/2);
			else
				draw.x = gap;
			
			draw.y = IBEAM_SIZE;
			
			if (end)
			{
				draw.y = Math.round((h - iBeamHeight) / 2);
			//	draw.x = Math.round((w - iBeamWidth) / 2); 
			}
			
			draw.width = iBeamWidth;
			draw.height = iBeamHeight;
			
			return draw;
		}
		
		[Bindable]
		/**
		 *  The computed column width for each column in the tile layout
		 * 
		 *  @param value The column width in pixels
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get computedColumnWidth():Number
		{
			return _computedColumnWidth;
		}
		
		public function set computedColumnWidth(value:Number):void
		{
			_computedColumnWidth = value;
		}

		[Bindable]
		/**
		 *  The computed row height for each row in the tile layout
		 * 
		 *  @param value The row height in pixels
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get computedRowHeight():Number
		{
			return _computedRowHeight;
		}
		
		public function set computedRowHeight(value:Number):void
		{
			_computedRowHeight = value;
		}
		
		/**
		 *  The CardLayoutHelper property.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get cardLayoutHelper():CardLayoutHelper
		{
			return _cardLayoutHelper;
		}

		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function set verticalScrollPosition(value:Number):void
		{
			if (verticalScrollPosition == value)
				return;
			
			super.verticalScrollPosition = value;
			target.invalidateDisplayList();
		}

		/**
		 *  The fit size to use for thumbnail images in the layout
		 * 
		 *  @param value The size in pixels for thumbnails to fit in the tile layout
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get thumbFitSize():Point
		{
			return _thumbFitSize;
		}

		public function set thumbFitSize(value:Point):void
		{
			_thumbFitSize = value;
		}
	}
}
