/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.layouts
{
	import com.adobe.portfolio.api.FieldValueChangedEvent;
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.cards.BasicCard;
	import com.adobe.portfolio.cards.ISimpleCard;
	import com.adobe.portfolio.components.supportClasses.AutoListen;
	import com.adobe.portfolio.components.supportClasses.AutoWatch;
	import com.adobe.portfolio.components.supportClasses.FreeFormFields;
	import com.adobe.portfolio.components.supportClasses.IListenAndWatch;
	import com.adobe.portfolio.layouts.supportClasses.CardDistributor;
	import com.adobe.portfolio.layouts.supportClasses.CardLayoutHelper;
	import com.adobe.portfolio.layouts.supportClasses.ICardLayout;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.PropertyManager;
	import com.adobe.portfolio.managers.ThumbnailManager;
	import com.adobe.portfolio.properties.CommonProperties;
	import com.adobe.portfolio.properties.PropertyListValue;
	import com.adobe.portfolio.properties.PropertyNumberValue;
	
	import flash.geom.Matrix3D;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.geom.Vector3D;
	
	import mx.binding.utils.BindingUtils;
	import mx.core.IVisualElement;
	import mx.events.ResizeEvent;
	import mx.logging.ILogger;
	import mx.logging.Log;
	
	import spark.components.supportClasses.GroupBase;
	import spark.core.NavigationUnit;
	import spark.layouts.supportClasses.DropLocation;
	import spark.layouts.supportClasses.LayoutBase;
	
	/**
	 * 
	 *  The FreeFormLayout class extends LayoutBase to implement the layout of the Freeform navigator.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class FreeFormLayout extends LayoutBase implements ICardLayout, IListenAndWatch
	{
		/**
		 *  The default card width, which is based on the default small card dimension
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const DEFAULT_CARD_WIDTH:Number = BasicCard.DEFAULT_SMALL_DIMENSION;

		/**
		 *  The default card height, which is based on the default small card dimension
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const DEFAULT_CARD_HEIGHT:Number = BasicCard.DEFAULT_SMALL_DIMENSION;

		/**
		 *  The string constant used to specify automatic card size
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const AUTO_CARD_SIZE:String = "auto";

		/**
		 *  The collectionManager instance used by this navigator.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected const collectionManager:CollectionManager = CollectionManager.instance;
		/**
		 *  The ILogger instance used by this navigator.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected const logger:ILogger = Log.getLogger("FreeFormLayout");
		
		/**
		 *  The depth in the display stack of the topmost card.
		 *  The topmost card is 'above' all other cards, except for
		 *  a hover card, which always displays on top of all other cards.
		 *  So it's depth is greater than any other card in the stack, except a hover card.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		private static const TOPMOST_CARD_DEPTH:Number = int.MAX_VALUE - 1;

		/**
		 *  @private
		 * 
		 *  The depth in the display stack of the hover card.
		 *  A hover card always displays on top of all other cards, so it's
		 *  depth is the miximum possible value.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		private static const HOVERED_CARD_DEPTH:Number = int.MAX_VALUE;
		
		// The amount of adjustment applied to the angle of the selected card(s)
		// each time the user presses the ctrl-shift left or ctrl-shift right keys.
		private static const INCREMENTAL_ANGLE_ADJUSTMENT:Number = 5;

		private static const MAX_PERCENT_AREA_FOR_CARDS:Number = .8;
		
		private static const SLIGHT_ROTATION:Number = 5;

		/**
		 *  The card size.  The default is PROPERTY_CARD_SIZE_SMALL.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var cardSize:String = CommonProperties.PROPERTY_CARD_SIZE_SMALL;

		/**
		 *  The card width, in pixels.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var cardWidth:Number = DEFAULT_CARD_WIDTH;

		/**
		 *  The card height, in pixels.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var cardHeight:Number = DEFAULT_CARD_HEIGHT;

		/**
		 *  A Boolean flag that indicates whether the cards should be displayed
		 *  at an angle.
		 * 
		 *  @return <code>true</code> (the default) if the cards are displayed at
		 *  various angles; <code>false</code> if the cards are displayed straight up.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var showCardsAngled:Boolean = true;
		
		/**
		 *  A Boolean flag that indicates whether the cards are automatically sized or not
		 * 
		 *  @return <code>true</code> if the cards are automatically sized
		 *  <code>false</code> if they are not automatically sized.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var autoSizeCards:Boolean;

		/**
		 *  A Boolean flag that indicates whether the currently specified card size is
		 *  a valid size for the current target or not.
		 * 
		 *  @return <code>true</code> if the current card size is invalid
		 *  <code>false</code> if the current card size is valid.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var invalidCardSize:Boolean;
		
		/**
		 *  A Boolean flag that indicates whether the currently specified card states are
		 *  valid or not.
		 * 
		 *  @return <code>true</code> if the current card states are invalid
		 *  <code>false</code> if the current card states are valid.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var invalidCardStates:Boolean;
		
		/**
		 *  This is the card that the user is currently hovering over
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var hoveredCard:IAttachment;
		
		/**
		 *  The instance of the AutoListen class used by this navigator for event listeners
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var autoListen:AutoListen;
		
		/**
		 *  The instance of the AutoWatch class used by this navigator for change watchers
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var autoWatch:AutoWatch;
		
		/**
		 *  The card size for automatically sized cards
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var cardAutoSize:PropertyNumberValue;
		
		/**
		 *  The fade duration, in mimiseconds, for transition effects.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var fadeDelay:Number;
		
		/**
		 *  The instance of CardDistributor used by this navigator
		 *  for placing Cards pseudo-randomly in the layout.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var cardDistributor:CardDistributor;
		
		/**
		 *  A Boolean flag that indicates whether the currently specified card depths are
		 *  valid or not.
		 * 
		 *  @return <code>true</code> if one or more of the current card depths are invalid
		 *  <code>false</code> if the current card depths are valid.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var invalidCardDepths:Boolean;
		
		private var _cardLayoutHelper:CardLayoutHelper;
		private var _thumbFitSize:Point;
		private var _dragStartPoint:Point = new Point();
		private var _cardBoundsArray:Array;
		private var _isDragging:Boolean;
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function FreeFormLayout()
		{
			super();
			_freeFormFields = new FreeFormFields();
			_cardLayoutHelper = new CardLayoutHelper(this);
			autoListen = new AutoListen();
			autoWatch = new AutoWatch();
			addListenAndWatch();  // TODO remove once addListenAndWatch is called externally
			cardDistributor = new CardDistributor(freeFormFields);
			cardDistributor.cardAngle = SLIGHT_ROTATION;
			ThumbnailManager.instance.initialQueueIsFIFO = false;
		}
		
		/**
		 *  Add event listeners and change watchers to the navigator.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function addListenAndWatch():void
		{
			autoWatch.addWatcher(BindingUtils.bindSetter(cardSizeChanged, PropertyManager.instance, ["commonProperties", "cardSize", "value"]));
			autoWatch.addWatcher(BindingUtils.bindSetter(showCardsAngledChanged, PropertyManager.instance, ["commonProperties", "portfolioCardsAngled", "value"]));
		}
		
		/**
		 *  Remove all event listeners and change watchers from the navigator
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function removeListenAndWatch():void
		{
			autoListen.removeAllListeners();
			autoWatch.removeAllWatchers();
		}

		private var _freeFormFields:FreeFormFields = null;

		/**
		 *  These correspond to field names in the pdf file and are
		 *  used to get and set the x position, y position, depth and
		 *  rotation of the card when it is used in the FreeForm navigator.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get freeFormFields():FreeFormFields
		{
			return _freeFormFields;
		}
		
		/**
		 *  Set the target host for this navigator
		 * 
		 *  @param value The GroupBase instance of the new target
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function set target(value:GroupBase):void
		{
			autoListen.removeListener(target);
			super.target = value;
			freeFormFields.target = value;
			autoListen.addListener(target, ResizeEvent.RESIZE, resizeHandler);
		}
		
		/**
		 *  This is the event handler for target resize events.
		 * 
		 *  @param event A target resize event that needs to be handled.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function resizeHandler(event:ResizeEvent):void
		{
//			logger.debug("resizeHandler");
			invalidCardSize = true;
			invalidCardStates = true;
			if (target)
				target.invalidateSize();
		}
		
		/**
		 *  Set the layout properties: best card size, card states, card bounds, and card depths.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function commitProperties():void
		{
			findBestCardSize();
			setCardStates();
			setAllCardBounds();
			setDepthsOfCards();
		}
		
		/**
		 *  The array of card boundaries for the currently list of attachments.
		 * 
		 *  @param value An array of card bounds.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function set cardBoundsArray(value:Array):void
		{
			_cardBoundsArray = value;
		}
		/**
		 *  @private
		 */ 
		public function get cardBoundsArray():Array
		{
			return _cardBoundsArray;
		}

		/**
		 *  Update the display List
		 * 
		 *  @param width The new width of the target, in pixels, in the targets's coordinates.
		 *  @param height The new height of the target, in pixels, in the targets's coordinates.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function updateDisplayList(width:Number, height:Number):void
		{
			if (!target)
				return;
			
//			logger.debug("updateDisplayList: isDragging= " + isDragging);
			super.updateDisplayList(width, height);
			
			if (isDragging) {
				return;
			}
			
			commitProperties();
			
			var count:int = target.numElements;
			for (var index:int = 0; index < count; index++)
			{
				var card:BasicCard = getVisualElementAt(index) as BasicCard;
				if (!card)
					continue;
				
				card.enableThumbnailGeneration = true;
				card.visible = true;
				card.depth = getDepthForCard(card);
				setRotationForCard(card);
			}
			
			// Make sure that if the content spans partially over a pixel to the right/bottom,
			// the content size includes the whole pixel.
			target.setContentSize(width, height);

			fadeInCards();
			cardLayoutHelper.enableThumbnailGeneration();
		}
		
		/**
		 *  Run a fade-in animation effect for the cards.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function fadeInCards():void
		{
			var count:int = target.numElements;
			
			if (count < 1)
				return;
			
			if (isNaN(fadeDelay))
				fadeDelay = Math.min(5000, count * 750);
			cardLayoutHelper.fadeInCards(fadeDelay);
			fadeDelay = 0;
		}
		
		/**
		 *  Determine the best card size to use based on the number of cards and the
		 *	window size.  This method is called if auto-sizing was requested and either
		 *  the window was resized or the number of cards changed.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function findBestCardSize():void
		{
			// Nothing to do if not auto-sizing cards
			// or no change to window size or number of cards
			if (!autoSizeCards || !invalidCardSize)
				return;
			
			// Wait until we can actually figure this out
			if (!target || !target.width || !target.height || !target.numElements)
				return;
			
			if (!PropertyManager.instance.commonProperties)
				return;
			
			var list:PropertyListValue = PropertyManager.instance.commonProperties.portfolioCardSizeList;
			if (!list || !list.subvalues || !list.subvalues.length)
				return;
			
			var maxAreaForCards:Number = target.width * target.height * MAX_PERCENT_AREA_FOR_CARDS;
			
			// Find the largest card that fits (taking the number of cards into account)
			var smallestSize:PropertyNumberValue;
			var largestSize:PropertyNumberValue;
			var largestArea:Number = 0;
			for (var index:int = 0; index < list.subvalues.length; ++index) {
				var size:PropertyNumberValue = list.subvalues[index] as PropertyNumberValue;
				if (!size)
					continue;
				if (size.name == CommonProperties.PROPERTY_CARD_SIZE_SMALL)
					smallestSize = size;
				var cardsArea:Number = size.numValue * size.numValue * target.numElements;
				if (cardsArea > maxAreaForCards)
					continue;
				if (cardsArea > largestArea) {
					largestArea = cardsArea;
					largestSize = size;
				}
			}
			if (largestSize || smallestSize) {
				cardAutoSize = largestSize ? largestSize : smallestSize;
				setCardSize(cardAutoSize.numValue);
				invalidCardSize = false;
				invalidCardStates = true;
			}
		}
		
		/**
		 *  Let the navigator know that the card size has been changed.
		 *  This will resize the cards in the display.
		 * 
		 *  @param value The card size to use, in pixels, if the card size is not <code>AUTO_CARD_SIZE</code>
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function cardSizeChanged(value:Number):void
		{
			var commonProperties:CommonProperties = PropertyManager.instance.commonProperties;
			if (!commonProperties || !commonProperties.cardSize)
				return;
			
			cardSize = commonProperties.cardSize.name;
			
			autoSizeCards = (cardSize == AUTO_CARD_SIZE);
			if (autoSizeCards)
				invalidCardSize = true;
			
			if (!autoSizeCards) {
				setCardSize(value);
			}
			invalidateTarget();
			invalidCardStates = true;
		}
		
		/**
		 *  Set the card size explicitly.
		 * 
		 *  @param value The width and height, in pixels, to use for the card size.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function setCardSize(value:Number):void {
			if (isNaN(value) || (value <= 0)) {
				// Let the card lay itself out using its default measured dimensions
				// TODO If this causes behavioral problems, then value can be set to
				//		BasicCard.DEFAULT_SMALL_DIMENSION as an alternative.
				value = Number.NaN;
			}
			
			cardWidth = value;
			cardHeight = value;
			freeFormFields.cardWidth = value;
			freeFormFields.cardHeight = value;
		}
		
		/**
		 *  Let the navigator know that the showCardsAngled Boolean indicator has been changed.
		 *  This will set showCardsAngled to the value specified.
		 * 
		 *  @param value Set this to <code>true</code> if you want the cards to be displayed at
		 *  various angles (this is the default for free form layout; set this to <code>false</code>
		 *  if you want to cards to be displayed straight up.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function showCardsAngledChanged(value:Boolean):void
		{
			if (!PropertyManager.instance.commonProperties || !PropertyManager.instance.commonProperties.portfolioCardsAngled)
				return;
			showCardsAngled = value;
			invalidateTarget();
		}
		
		/**
		 *  Sets the attachment for the specified visual element as the hoveredCard
		 *  if hoveredOver is <code>true</code>.
		 * 
		 *  @param visualElement The visual element that is being hovered over
		 *  @param hoveredOver If <code>true</code> then <code>visualElement</code>'s related attachment 
		 *  will be the hovered card.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function hoveringOverElement(visualElement:IVisualElement, hoveredOver:Boolean):void
		{
//			logger.info("hoveringOverElement: " + hoveredOver);
			hoveredCard = null;
			var attachment:IAttachment = getAttachmentFromVisualElement(visualElement);
			hoveredCard = hoveredOver ? attachment : null;
			invalidateTarget();
		}
		
		/**
		 *  Resets the depth for each card in the vector of selectedItems.
		 * 
		 *  @param selectedItems A vector of IAttachements
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function selectionChanged(selectedItems:Vector.<Object>):void
		{
			if (!collectionManager.authoringMode)
				return;

//			logger.info("selectionChanged: " + selectedItems);
			
			var depth:int = TOPMOST_CARD_DEPTH;
			for each (var attachment:IAttachment in selectedItems) {
				freeFormFields.setDepth(attachment, depth--);
			}	
			invalidCardDepths = true;
			invalidateTarget();
		}
		
		/**
		 *  Adjusts the display angle for all the cards with an index 
		 *  in the indices vector by the specified navigationUnit
		 * 
		 *  @param indices A vector of card indices
		 *  @param navigationUnit The amount, in degrees, to adjust the angle of each card
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function adjustAngleOfCards(indices:Vector.<int>, navigationUnit:uint):void
		{
			// Nothing to do if not in edit mode
			if (!collectionManager.authoringMode)
				return;
			for each (var index:int in indices) {
				var attachment:IAttachment = getAttachmentFromVisualElement(getVisualElementAt(index));
				if (!attachment)
					continue;
				adjustCardAngle(attachment, navigationUnit);
			}
			invalidateTarget();
		}
		
		/**
		 *  Adjust the display angle of the card for the card displayed for the specified attachment
		 * 
		 *  @param attachment The attachment for which its card's display angle is to be changed.
		 *  @param navigationUnit A NavigationUnit enum value corresponding to a keyCode, such as UP, DOWN, LEFT, RIGHT, etc.
		 * 	
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function adjustCardAngle(attachment:IAttachment, navigationUnit:uint):void
		{
			var angle:Number = freeFormFields.getRotation(attachment);
			switch (navigationUnit) {
				case NavigationUnit.UP:
					angle = 0;
					break;
				
				case NavigationUnit.DOWN:
					angle = 180;
					break;
				
				case NavigationUnit.LEFT:
					angle -= INCREMENTAL_ANGLE_ADJUSTMENT;
					break;
				
				case NavigationUnit.RIGHT:
					angle += INCREMENTAL_ANGLE_ADJUSTMENT;
					break;
				
				default:
					return;
			}
			angle = Math.round(angle / INCREMENTAL_ANGLE_ADJUSTMENT) * INCREMENTAL_ANGLE_ADJUSTMENT; // make angle a multiple of INCREMENTAL_ANGLE_ADJUSTMENT
			freeFormFields.setRotation(attachment, angle);
		}
		
		/**
		 *  Handles adding a visual element: it gets the related IAttachment and sets the
		 *  appropriate event listeners.
		 * 
		 *  @param visualElement The visual element that has been added
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function elementWasAdded(visualElement:IVisualElement):void
		{
//			logger.info("elementWasAdded");
			invalidCardSize = true;
			cardLayoutHelper.pendingFade = true;

			var attachment:IAttachment = getAttachmentFromVisualElement(visualElement);
			if (attachment)
				attachment.addEventListener("fieldValueChanged", onFieldChange, false, 0, true);
		}
		
		/**
		 *  Handles removing a visual element: it gets the related IAttachment and removes the
		 *  appropriate event listeners.
		 * 
		 *  @param visualElement The visual element that has been removed
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function elementWasRemoved(visualElement:IVisualElement):void
		{
//			logger.info("elementWasRemoved");
			invalidCardSize = true;
			var attachment:IAttachment = getAttachmentFromVisualElement(visualElement);
			if (attachment)
				attachment.removeEventListener("fieldValueChanged", onFieldChange);
			if (attachment == hoveredCard)
				hoveredCard = null;
			if (collectionManager.authoringMode)
				invalidCardDepths = true;
			invalidateTarget();
		}
		
		/**
		 *  Handler for field value changed events
		 * 
		 *  @param e The <code>FieldValueChangedEvent</code> to handle
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onFieldChange(e:FieldValueChangedEvent):void
		{
			//	logger.debug("fieldName= " + e.fieldName);
			invalidateTarget();
		}
		
		/**
		 *  Find all cards that intersect the region between the anchor index and destination index.
		 *  This is used to implement shift-select to extend the selection.
		 * 
		 *  @param anchorIndex The index of the anchor point when doing shift-selection
		 *  @param destIndex The index of the destination or final point when doing shift-selection
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getExtendedSelection(anchorIndex:int, destIndex:int):Vector.<int>
		{
			var intersections:Vector.<int> = new Vector.<int>();
			if (anchorIndex < 0 || destIndex < 0)
				return intersections;
			if (!cardBoundsArray || anchorIndex >= cardBoundsArray.length || destIndex >= cardBoundsArray.length)
				return intersections;
			var anchorBounds:Rectangle = cardBoundsArray[anchorIndex];
			var destBounds:Rectangle = cardBoundsArray[destIndex];
			if (!anchorBounds || !destBounds)
				return intersections;
			
			var topMost:Rectangle = anchorBounds.top < destBounds.top ? anchorBounds : destBounds;
			var bottomMost:Rectangle = anchorBounds.bottom > destBounds.bottom ? anchorBounds : destBounds;
			var leftMost:Rectangle = anchorBounds.left < destBounds.left ? anchorBounds : destBounds;
			var rightMost:Rectangle = anchorBounds.right > destBounds.right ? anchorBounds : destBounds;
			
			// Do the anchor and destination rectangles intersect (horizontally)?
			if ((bottomMost.top >= topMost.top) && (bottomMost.top <= topMost.bottom)) {
				// If so, select all cards between them (horizontally)
				var testRect:Rectangle = new Rectangle(leftMost.x, topMost.y);
				testRect.right = rightMost.right;
				testRect.bottom = bottomMost.bottom;
				intersections = findCardsThatIntersect(testRect);
			}
			else {
				// If not, select all cards between them (both horizontally and vertically)
				var testRect1:Rectangle = topMost.clone();
				testRect1.right = target.width;
				var testRect2:Rectangle = new Rectangle(0, topMost.bottom);
				testRect2.right = target.width;
				testRect2.bottom = bottomMost.top;
				var testRect3:Rectangle = bottomMost.clone();
				testRect3.left = 0;
				intersections = findCardsThatIntersectRectangles([testRect1, testRect2, testRect3]);
			}
			
			return intersections;
		}
		
		/**
		 *  Find the list of cards that overlap the specified rectangle. 
		 * 
		 *  @param rect A bounding rectangle
		 *  @return A vector of integers that represent the indexes
		 *  of the cards in the layout that intersect the specified rectangle.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function findCardsThatIntersect(rect:Rectangle):Vector.<int>
		{
			return findCardsThatIntersectRectangles([rect]);
		}
		
		/**
		 *  Find the list of cards that overlap any of the rectangles in the specified array. 
		 * 
		 *  @param rects An array of bounding rectangle
		 *  @return A vector of integers that represent the indexes
		 *  of the cards in the layout that intersect the specified rectangles.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function findCardsThatIntersectRectangles(rects:Array):Vector.<int>
		{
			var intersections:Vector.<int> = new Vector.<int>();
			if (!cardBoundsArray || !rects || rects.length == 0)
				return intersections;
			
			for (var cIndex:int = 0; cIndex < cardBoundsArray.length; cIndex++) {
				var cardBounds:Rectangle = cardBoundsArray[cIndex] as Rectangle;
				// No bounds to check against
				if (!cardBounds)
					continue;
				for (var rIndex:int = 0; rIndex < rects.length; rIndex++) {
					var rect:Rectangle = rects[rIndex];
					if (rect.intersects(cardBounds)) {
						intersections.push(cIndex);
						break;
					}
				}
			}
			
			return intersections;
		}
		
		/**
		 *  Find the index value of the card that is the destination, based on the navigationUnit
		 *  supplied (UP, DOWN, LEFT, RIGHT, PAGE_UP, PAGE_DOWN, HOME).
		 * 
		 *  @param currentIndex The index of the currently selected card
		 *  @param navigationUnit A NavigationUnit enum value corresponding to a keyCode, such as UP, DOWN, LEFT, RIGHT, etc.
		 *  @param arrowKeysWrapFocus
		 *  @return The index value of the destination card.
		 *  Returns -1 if no destination index can be found.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function getNavigationDestinationIndex(currentIndex:int, navigationUnit:uint, arrowKeysWrapFocus:Boolean):int
		{
			var nearest:int = -1;
			
			if (!target)
				return nearest;
			
			var cardBounds:Rectangle = null;
			if (!cardBoundsArray)
				return nearest;
			
			if (currentIndex < 0 || currentIndex >= cardBoundsArray.length)
				return nearest;
			
			var currentCardBounds:Rectangle = cardBoundsArray[currentIndex] as Rectangle;
			if (!currentCardBounds)
				return nearest;
			
			// For many navigation changes, compare from the center of the currently selected card
			var compareFrom:Point = new Point(currentCardBounds.x + currentCardBounds.width / 2, currentCardBounds.y + currentCardBounds.height / 2);
			
			switch (navigationUnit) {
				case NavigationUnit.UP:
				case NavigationUnit.DOWN:
				case NavigationUnit.LEFT:
				case NavigationUnit.RIGHT:
					// Look for the card closest to the center of the currently selected card
					// compareFrom.x set above
					// compareFrom.y set above
					break;
				
				case NavigationUnit.PAGE_UP:
					// Look for the card farthest above the currently selected card
					// compareFrom.x set above
					compareFrom.y = 0;
					break;
				
				case NavigationUnit.PAGE_DOWN:
					// Look for the card farthest below the currently selected card
					// compareFrom.x set above
					compareFrom.y = target.height;
					break;
				
				case NavigationUnit.HOME:
					// Look for the card closest to the top left of the display
					compareFrom.x = 0;
					compareFrom.y = 0;
					break;
				
				case NavigationUnit.END:
					// Look for the card closest to the bottom right of the display
					compareFrom.x = target.width;
					compareFrom.y = target.height;
					break;
				
				default:
					return nearest;
			}
			
			var compareTo:Point = new Point();
			var nearestDelta:Number = Number.POSITIVE_INFINITY;
			
			for (var index:int = 0; index < cardBoundsArray.length; index++) {
				cardBounds = cardBoundsArray[index] as Rectangle;
				// No bounds to check against
				if (!cardBounds)
					continue;
				
				compareTo.x = cardBounds.x + (cardBounds.width / 2);
				compareTo.y = cardBounds.y + (cardBounds.height / 2);
				var deltaX:Number = compareTo.x - compareFrom.x;
				var deltaY:Number = compareTo.y - compareFrom.y;
				var absDeltaX:Number = Math.abs(deltaX);
				var absDeltaY:Number = Math.abs(deltaY);
				
				switch (navigationUnit) {
					case NavigationUnit.UP:
						// Skip the currently selected card
						if (index == currentIndex)
							continue;
						if (deltaY > 0 || absDeltaX > absDeltaY)
							continue;
						break;
					
					case NavigationUnit.DOWN:
						// Skip the currently selected card
						if (index == currentIndex)
							continue;
						if (deltaY < 0 || absDeltaX > absDeltaY)
							continue;
						break;
					
					case NavigationUnit.LEFT:
						// Skip the currently selected card
						if (index == currentIndex)
							continue;
						if (deltaX > 0 || absDeltaY > absDeltaX)
							continue;
						break;
					
					case NavigationUnit.RIGHT:
						// Skip the currently selected card
						if (index == currentIndex)
							continue;
						if (deltaX < 0 || absDeltaY > absDeltaX)
							continue;
						break;
					
					case NavigationUnit.PAGE_UP:
					case NavigationUnit.PAGE_DOWN:
					case NavigationUnit.HOME:
					case NavigationUnit.END:
					default:
						// Looking for closest to the compareFrom point, no restrictions
						break;
				}
				
				var delta:Number = Point.distance(compareFrom, compareTo);
				if (delta < nearestDelta) {
					nearestDelta = delta;
					nearest = index;
				}
			}
			
			return nearest;
		}

		/**
		 *  Set the card boundaries for all cards in the portfolio.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function setAllCardBounds():void
		{
//			logger.debug("setAllCardBounds");
			
			cardDistributor.ensureAllPositionsAreValid(cardLayoutHelper.getDataProvider());
			
			var numElements:int = target.numElements;
			var boundsArray:Array = new Array(numElements);
						
			for (var index:int = 0; index < numElements; index++) {
				var card:ISimpleCard = cardLayoutHelper.getCard(index);
				if (!card)
					continue;

				// Get the bounds of the thumbnail card
				boundsArray[index] = setCardBounds(card);
			}
			cardBoundsArray = boundsArray;
		}
		
		/**
		 *  Set the card boundary for the specified card
		 * 
		 *  @param card The card for which you want to find its bounding box
		 *  @return A rectangle the specifies the position and size of the card's bounding box.
		 *  This is the unrotated rectangle that contains the card.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function setCardBounds(card:IVisualElement):Rectangle
		{
			card.setLayoutBoundsSize(cardWidth, cardHeight, false);
			
			var bounds:Rectangle = new Rectangle(0, 0, card.width, card.height);
			
			var attachment:IAttachment = getAttachmentFromVisualElement(card);
			if (!attachment)
				return bounds;
			
			var x:Number = freeFormFields.getX(attachment);
			var y:Number = freeFormFields.getY(attachment);
			if (!isNaN(x) && !isNaN(y)) {
				bounds.x = x - bounds.width / 2;
				bounds.y = y - bounds.height / 2;
				card.setLayoutBoundsPosition(bounds.x, bounds.y, false);
			}
			
//			logger.debug("setCardBounds: card: " + card + " bounds: " + bounds);
			return bounds;
		}

		/**
		 *  Set the appropriate states for all cards used in the free form layout.  
		 *  These are the states used with <code>BasicCard</code>s.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function setCardStates():void
		{
			if (!invalidCardStates)
				return;
			
//			logger.debug("setCardStates");
			
			var numElements:int = target.numElements;
			for (var i:int = 0; i < numElements; i++) {
				var card:ISimpleCard = cardLayoutHelper.getCard(i);
				if (!card)
					continue;
				
				// Update the card state
				setStateForCard(card as BasicCard);
			}
			
			invalidCardStates = false;
		}
		
		/**
		 *  Set the appropriate states for the specified card.
		 *  This is normally called from setCardStates().
		 * 
		 *  @param card A <code>BasicCard</code> for which you want to set the valid states.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function setStateForCard(card:BasicCard):void
		{
			if (!card)
				return;
			
			switch (cardSize) {
				case CommonProperties.PROPERTY_CARD_SIZE_SMALL:
					card.currentState = BasicCard.STATE_SMALL;
					break;
				
				case CommonProperties.PROPERTY_CARD_SIZE_MEDIUM:
					card.currentState = BasicCard.STATE_MEDIUM;
					break;
				
				case CommonProperties.PROPERTY_CARD_SIZE_LARGE:
					// If the user has changed to the info state, don't change it back to large
					if (card.currentState != BasicCard.STATE_INFO) {
						card.currentState = BasicCard.STATE_LARGE;
					}
					break;
				
				case AUTO_CARD_SIZE:
					// If the user has changed to the info state, don't change it 
					if (card.currentState != BasicCard.STATE_INFO) {
						card.currentState = cardAutoSize.name;
					}
					break;
				
				default:
					card.currentState = BasicCard.STATE_SMALL;
					break;
			}
		}
		/**
		 *  Set the appropriate degree of rotation for the specified card
		 *  based on the "adobe:FreeForm:angle" field value in the PDF file.
		 * 
		 *  @param card A <code>BasicCard</code> for which you want to set the rotation.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function setRotationForCard(card:BasicCard):void
		{
			if (!card)
				return;
			
			var attachment:IAttachment = card.data as IAttachment;
			var angle:Number = (attachment && showCardsAngled) ? freeFormFields.getRotation(attachment) : 0;
			var matrix:Matrix3D = new Matrix3D();
			
//			logger.debug("card: " + card + " center = " + (card.x + card.width/2) + " , " + (card.y + card.height/2));
			var w2:Number = card.width/2;
			var h2:Number = card.height/2;
			matrix.appendTranslation(-w2,-h2,0);
			matrix.appendRotation( angle, Vector3D.Z_AXIS);
			matrix.appendTranslation(w2,h2,0);
			matrix.appendTranslation(card.x, card.y, 0);
			card.setLayoutMatrix3D( matrix, false);
		}

		/**
		 *  Get the depth in the visual stack of the specified card.  
		 *
 		 *  @param card A <code>BasicCard</code> for which you want to get its depth.
		 *  @return A number indicating where the card is in the visual stack. The card on top will have the highest depth number.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function getDepthForCard(card:BasicCard):Number
		{
			if (!card)
				return 0;
			
			var attachment:IAttachment = card.data as IAttachment;
			if (!attachment)
				return 0;
			else if (attachment == hoveredCard)
				return HOVERED_CARD_DEPTH;
			else
				return freeFormFields.getDepth(attachment);
		}
		
		/**
		 *  Set the card depths in the visual stack for all the cards in the portfolio.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function setDepthsOfCards():void
		{
			if (!invalidCardDepths)
				return;
			if (!target)
				return;
			
//			logger.debug("correctDepthsOfCards");
			var attachments:Array = getAttachments();

			// Sort the depths
			attachments.sort(compareAttachmentDepths);
			// Normalize the depths
			var currDepth:Number = Number.MIN_VALUE;
			var normalizedDepth:Number = -1;
			for each (var attachment:IAttachment in attachments) {
				if (!attachment)
					continue;
				var depth:Number = freeFormFields.getDepth(attachment);
				if (depth != currDepth) {
					currDepth = depth;
					normalizedDepth++;
				}
				if (depth != normalizedDepth)
					freeFormFields.setDepth(attachment, normalizedDepth);
			}
			invalidCardDepths = false;
		}
		
		/**
		 *  Get all the attachments in the portfolio.
		 * 
		 *  @return An array of attachments.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function getAttachments():Array
		{
			var attachments:Array = new Array();
			for (var index:int = 0; index < target.numElements; ++index) {
				var attachment:IAttachment = getAttachmentFromVisualElement(getVisualElementAt(index));
				if (!attachment)
					continue;
				attachments.push(attachment);
			}
			return attachments;
		}
		
		/**
		 *  Compare the depth in the visual stack of two attachments,
		 *  so that you can tell which is on top of the other.
		 * 
		 *  @param attach1 The first attachment to compare
		 *  @param attach2 The second attachment to compare
		 *  @return Returns <code>0</code> if the two attachments are at the same depth.
		 *  Returns <code>1</code> if the first attachment has a greater depth than (i.e., is on top of) the second attachment.
		 *  Returns <code>-1</code> if the first attachment has a lesser depth than (i.e., is below) the second attachment.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function compareAttachmentDepths(attach1:IAttachment, attach2:IAttachment):int
		{
			if (!attach1 && !attach2)
				return 0;
			if (!attach1)
				return -1;
			if (!attach2)
				return 1;
			var depth1:Number = freeFormFields.getDepth(attach1);
			var depth2:Number = freeFormFields.getDepth(attach2);
			if (depth1 == depth2)
				return 0;
			if (depth1 < depth2)
				return -1;
			return 1;
		}
		
		/**
		 *  Get the attachment that is associated with the specified visual element.
		 * 
		 *  @param visualElement The visual element for which you want the associated attachment
		 *  @return The attachment associated with the visual element
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function getAttachmentFromVisualElement(visualElement:IVisualElement):IAttachment
		{
			var card:BasicCard = visualElement as BasicCard;
			if (!card)
				return null;
			return card.data as IAttachment;
		}
		
		/**
		 *  Get the visual element that has the specified index value.
		 *  This is the compliment of getVisualElementIndex()
		 * 
		 *  @param index The index value for the visual element in which you are interested 
		 *  @return The visual element with the specified index value
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function getVisualElementAt(index:int):IVisualElement
		{
			// Validate that the index is in range
			if (!target || index < 0 || index >= target.numElements)
				return null;
			// Note that classes derived from list are virtual and need to use getVirtualElementAt
			var visualElement:IVisualElement = useVirtualLayout ? target.getVirtualElementAt(index) : target.getElementAt(index);
			if (!visualElement || !visualElement.includeInLayout)
				return null;
			return visualElement;
		}
		
		/**
		 *  Get the index of the specified visual element.
		 *  This is the compliment of getVisualElementAt()
		 * 
		 *  @param visualElement The visual element in which you are interested 
		 *  @return The index value for the visual element specified
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getVisualElementIndex(visualElement:IVisualElement):int
		{
			// Validate that the visualElement is not null
			if (!target || !visualElement)
				return -1;
			return target.getElementIndex(visualElement);
		}
		
		/**
		 *  Invaidates the display list for the target
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function invalidateTarget():void
		{
			if (target)
				target.invalidateDisplayList();
		}

		/**
		 *  The CardLayoutHelper instance used by this Navigator.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get cardLayoutHelper():CardLayoutHelper
		{
			return _cardLayoutHelper;
		}

		/**
		 *  The size, represented as a width and height ordered pair, to use for thumbnails
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get thumbFitSize():Point
		{
			return _thumbFitSize;
		}

		/**
		 *  @private
		 */
		public function set thumbFitSize(value:Point):void
		{
			_thumbFitSize = value;
		}
		
		/**
		 *  The starting position, represented as an (x,y) location, for a 
		 * 
		 *  @param value The (x,y) location, in the target's coordinate system of the starting
		 *  position for a drag, or shift-select operation.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function set dragStartPoint(value:Point):void
		{
			_dragStartPoint = value;
		}
		/**
		 *  @private
		 */
		public function get dragStartPoint():Point
		{
			return _dragStartPoint;
		}
		
		/**
		 *  Calculates the bounding rectangle defined by the dragStartPoint and the dropLocation.
		 *  This method returns <code>null</code> for drags initiated from the platform,
		 *  since we do not want to show the drag image in that situation.
		 * 
		 *  @param dropLocation The drop location for a drag and drop operation.
		 *  @return A rectangle defined by the dragStartPoint and the dropLocation.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override protected function calculateDropIndicatorBounds(dropLocation:DropLocation):Rectangle
		{
			// Return null for drags initiated from the platform as we do not want to show the drag image in that case 
			return (dropLocation.dragEvent.dragInitiator != null)? new Rectangle(dropLocation.dropPoint.x - _dragStartPoint.x
				, dropLocation.dropPoint.y - _dragStartPoint.y
				, dropIndicator.width, dropIndicator.height): null;
		}

		/**
		 *  A Boolean indicating whether a drag operation is occuring or not.
		 * 
		 *  @param value If <code>true</code> then a drag operation has started.
		 *  If <code>false</code> there is no current drag operation.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function set isDragging(value:Boolean):void
		{
			_isDragging = value;
			if (!value && target) {
				target.invalidateDisplayList();
				target.validateNow();
			}
		}		
		/**
		 *  @private
		 */	
		public function get isDragging():Boolean
		{
			return _isDragging;
		}
	}
}
