/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.layouts
{
	import com.adobe.portfolio.animations.AnimateFactory;
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.cards.ISimpleCard;
	import com.adobe.portfolio.components.Navigator;
	import com.adobe.portfolio.components.supportClasses.IListenAndWatch;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.events.LayoutEvent;
	import com.adobe.portfolio.layouts.supportClasses.CardLayoutHelper;
	import com.adobe.portfolio.layouts.supportClasses.ICardLayout;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.utils.LogHelper;
	
	import flash.events.IEventDispatcher;
	import flash.geom.Matrix3D;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.geom.Vector3D;
	
	import mx.core.IVisualElement;
	import mx.effects.Effect;
	import mx.effects.Parallel;
	import mx.events.EffectEvent;
	import mx.events.ResizeEvent;
	
	import spark.components.supportClasses.GroupBase;
	import spark.effects.Animate;
	import spark.layouts.supportClasses.LayoutBase;
	
	/**
	 *  Dispatched after the layout has set its focus to a valid attachment.
	 *
	 *  @eventType com.adobe.portfolio.events.NavLayoutEvent.HANDLED_ATTACHMENT_FOCUS
	 */
	[Event(name="HANDLED_ATTACHMENT_FOCUS", type="com.adobe.portfolio.events.LayoutEvent")]
	
	/**
	 * 
	 * The LinearLayout class extends LayoutBase to implement the layout of the Linear navigator.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class LinearLayout extends LayoutBase implements IListenAndWatch, ICardLayout
	{
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function LinearLayout()
		{
			super();
			addListenAndWatch();
			_cardLayoutHelper = new CardLayoutHelper(this);
		}
		
		/**
		 *  The IAttachment that currently has the focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var currentFocus:IAttachment;
		
		/**
		 *  The IAttachment that had the focus immediately
		 *  prior to the currentFocus.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var lastFocus:IAttachment;
		
		/**
		 *  The IAttachement that is to get the focus next.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var pendingFocus:IAttachment;
		
		/**
		 *  The effect to currently playing due to a focus change
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var focusEffect:Effect = null;
		
		/**
		 * When we're in virtual mode and we're not animating, we load a few extra cards to
		 * ensure smooth transitions later. However, when we first come out of an animation, the
		 * first update is slow, since we build new cards. This boolean is used to delay that
		 * until the SECOND update, so that the end of the animation doesn't look so choppy.
		 */
		private var skipNextPreload:Boolean;
		
		private var _cardLayoutHelper:CardLayoutHelper;
		private var _thumbFitSize:Point;

		private static const CARD_GAP:int = 50;
		private static const BACKGROUND_DISTANCE:Number = 700;
		private static const BACKGROUND_SCALE:Number = 0.5;
		private static const CARD_ROTATION:Number = 5;
		private const FOCUS_DURATION:Number = Navigator.FOCUS_DURATION;
		
		private var maxDimension:Number;
		private var horizontalFocusPosition:Number;
		
		/**
		 *  The collection manager instance used by this Navigator
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected const collectionManager:CollectionManager = CollectionManager.instance;

		/**
		 *  The maximum linear size of a square that can fit within the target
		 *  container.  This is used to calculate the layout position of element
		 *  that are in the layout.
		 * 
		 *  @param unscaledWidth Specifies the width of the target,
		 *  in pixels, in the target's coordinates, regardless of the value
		 *  of the scaleX property of the target.
		 *  @param unscaledHeight Specifies the width of the target,
		 *  in pixels, in the target's coordinates, regardless of the value
		 *  of the scaleY property of the target.
		 *  @return The minimum of the scaled width and scaled height of the target
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public static function getMaxDimension(unscaledWidth:Number, unscaledHeight:Number):Number
		{
			return Math.min(unscaledWidth * 0.5, unscaledHeight * 0.8);
		}
		
		/**
		 *  Returns the desired horizontal position within this layout for the active card.
		 * 
		 *  @param unscaledWidth Specifies the width of the target,
		 *  in pixels, in the target's coordinates, regardless of the value
		 *  of the scaleX property of the target.
		 *  @param unscaledHeight Specifies the width of the target,
		 *  in pixels, in the target's coordinates, regardless of the value
		 *  of the scaleY property of the target.
		 *  @return the desired horizontal position in pixels within this layout for the active item.
		 *  This will be offset a small amount to the left of the current layout.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public static function getHorizontalFocusPosition(unscaledWidth:Number, unscaledHeight:Number):Number
		{
			// Inset the active item a small amount from the left of the layout
			return CARD_GAP;
		}
		
		/**
		 *  Returns the horizontal position within this layout for element
		 *  with the given index value.
		 * 
		 *  @param index The index value of the element for which 
		 *  you want the horizontal scroll position
		 *  @param unscaledWidth Specifies the width of the target,
		 *  in pixels, in the target's coordinates, regardless of the value
		 *  of the scaleX property of the target.
		 *  @param unscaledHeight Specifies the width of the target,
		 *  in pixels, in the target's coordinates, regardless of the value
		 *  of the scaleY property of the target.
		 *  @return The horizontal position, in pixels from the left side of the
		 *  target, for the element with the given index position.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function getScrollPositionForIndex(index:int,
												  unscaledWidth:Number = 0,
												  unscaledHeight:Number = 0):Number
		{
			if (target == null) {
				return 0;
			}
			if (unscaledWidth == 0) {
				unscaledWidth = target.width;
			}
			if (unscaledHeight == 0) {
				unscaledHeight = target.height;
			}
			var maxDimensionPlusGap:Number = getMaxDimension(unscaledWidth, unscaledHeight) + CARD_GAP;
			return index * maxDimensionPlusGap;
		}
		
		/**
		 *  Add an event listener to the collection manager to 
		 *  listen for host events
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function addListenAndWatch():void
		{
			collectionManager.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent);
		}
		
		/**
		 *  Remove the current event listener for host events
		 *  from the collection manager
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function removeListenAndWatch():void
		{
			collectionManager.removeEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent);
		}
		
		/**
		 *  Sets the UI container instance for this layout and
		 *  handles setting and removing event listeners for the target.
		 * 
		 *  @param value The groupBase container instance to use for this layout
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function set target(value:GroupBase):void	
		{
			if (target != null) {
				target.removeEventListener(ResizeEvent.RESIZE, onTargetResize);
			}
			super.target = value;
			if (target != null) {
				target.addEventListener(ResizeEvent.RESIZE, onTargetResize);
			}
		}
		
		/**
		 *  Updates the layout scroll position when the target container
		 *  has been resized.
		 * 
		 *  @param event The resize event caused by the target container
		 *  being resized.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onTargetResize(event:ResizeEvent):void
		{
			if (target == null) {
				return;
			}
			var newHorizontalPosition:Number = 0;
			if (currentFocus != null) {
				newHorizontalPosition = getScrollPositionForIndex(cardLayoutHelper.getCurrentIndex(currentFocus));
			}
			if (horizontalScrollPosition != newHorizontalPosition) {
				horizontalScrollPosition = newHorizontalPosition;
			}
		}
		
		/**
		 *  Draws the layout and positions and sizes the target GroupBase's
		 *  card elements
		 * 
		 *  @param unscaledWidth Specifies the width of the target,
		 *  in pixels, in the target's coordinates, regardless of the value
		 *  of the scaleX property of the target.
		 *  @param unscaledHeight Specifies the width of the target,
		 *  in pixels, in the target's coordinates, regardless of the value
		 *  of the scaleY property of the target.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function updateDisplayList(unscaledWidth:Number, unscaledHeight:Number):void
		{
			super.updateDisplayList(unscaledWidth, unscaledHeight);
			
			if (target == null) {
				return;
			}
			
			if (currentFocus != collectionManager.focusAttachment) {
				if (cardLayoutHelper.getCurrentIndex(collectionManager.focusAttachment) >= 0) {
					CONFIG::logging{
						LogHelper.logger.warn("currentFocus didn't match collectionManager.focusAttachment");
					}
					pendingFocus = collectionManager.focusAttachment;
				}
			}
			
			var focusIndex:int = cardLayoutHelper.getCurrentIndex(currentFocus);
			var pendingIndex:int = cardLayoutHelper.getCurrentIndex(pendingFocus);
			
			maxDimension = getMaxDimension(unscaledWidth, unscaledHeight);
			horizontalFocusPosition = getHorizontalFocusPosition(unscaledWidth, unscaledHeight);
			thumbFitSize = new Point(maxDimension, maxDimension);
			var maxDimensionPlusGap:Number = maxDimension + CARD_GAP;
			var backgroundYOffset:Number = maxDimension / 3;
			var guessFocusWasWrong:Boolean = false;
			var horizontalFocusOffset:Number = horizontalScrollPosition + horizontalFocusPosition;
			
			
			var contentWidth:Number = horizontalFocusPosition + (maxDimensionPlusGap * (target.numElements - 1)) + unscaledWidth;
			target.setContentSize(contentWidth, unscaledHeight);
			
			var scrollRect:Rectangle = cardLayoutHelper.scrollRect(target);
			// need to offset by the parent's origin - not sure why this is the case
			scrollRect.x += target.parent.x;  
			scrollRect.y += target.parent.y;
			
//			logger.debug("LinearLayout: scrollRect= " + scrollRect);
			var animating:Boolean = (focusEffect != null) && focusEffect.isPlaying;
			if (animating) {
				skipNextPreload = true;
			} else if (focusIndex >= 0) {
				// We're not animating, so force our scroll position to what we expect
				var expectedScrollPosition:Number = getScrollPositionForIndex(focusIndex, unscaledWidth, unscaledHeight);
				if ((pendingFocus == null) && (horizontalScrollPosition != expectedScrollPosition)) {
					super.horizontalScrollPosition = expectedScrollPosition;
				}
			}
			var preloadElements:Boolean = useVirtualLayout && !animating && !skipNextPreload;
			
			var lastIndex:int = animating ? cardLayoutHelper.getCurrentIndex(lastFocus) : focusIndex;
			const virtualBuffer:uint = 3;
			for (var i:int = 0; i < target.numElements; ++i) {
				var shouldBeInBounds:Boolean = false;
				if (!preloadElements) {
					// Rely on code later to only hold onto the elements we've already got
				} else if (i == pendingIndex) {
					shouldBeInBounds = true;
				} else {
					var lowerBounds:int = focusIndex - 1;
					var upperBounds:int = focusIndex + virtualBuffer;
					if ((i >= lowerBounds) && (i <= upperBounds)) {
						// Only setup the next/previous cards if we're not animating
						shouldBeInBounds = true;
					}
				}
				
				var card:IVisualElement = (useVirtualLayout && shouldBeInBounds) ? 	target.getVirtualElementAt(i) :
																					target.getElementAt(i);
				if (card == null) {
					continue;
				}
				if (useVirtualLayout && !preloadElements) {
					// This is a leftover renderer, but don't get rid of one we might want next time around
					card = target.getVirtualElementAt(i);
				}
				card.setLayoutBoundsSize(maxDimension, maxDimension);
				var elementSize:Point = new Point(card.getLayoutBoundsWidth(), card.getLayoutBoundsHeight());
				
				var positionMatrix:Matrix3D = new Matrix3D();
				var horizontalCardPosition:Number = horizontalFocusPosition + (maxDimensionPlusGap * i);
				positionMatrix.position = new Vector3D(	horizontalCardPosition + ((maxDimension - elementSize.x) / 2),
														(unscaledHeight - elementSize.y) / 2, 0);
				card.depth = -i;
//				card.depth = contentWidth - horizontalCardPosition;
				
				// TODO I'd like to animate the cards directly as part of the focus effect, but that's
				//     causing layout problems.
				var ratio:Number = 1.0;
				if (horizontalCardPosition < horizontalFocusOffset) {
					if (horizontalCardPosition >= (horizontalFocusOffset - maxDimensionPlusGap)) {
						ratio = (horizontalFocusOffset - horizontalCardPosition) / maxDimensionPlusGap;
					}
					positionMatrix.appendTranslation(0, 0, -0.25 * BACKGROUND_DISTANCE * ratio);
					card.alpha = Math.max(0, 1.0 - ratio);
				} else if (horizontalCardPosition > horizontalFocusOffset) {
					if (true/*horizontalCardPosition <= (horizontalFocusOffset + maxDimensionPlusGap)*/) {
						ratio = (horizontalCardPosition - horizontalFocusOffset) / maxDimensionPlusGap;
						ratio = Math.pow(ratio, 0.25);
					}
					positionMatrix.appendTranslation(0, backgroundYOffset * ratio, BACKGROUND_DISTANCE * ratio);
					//positionMatrix.appendScale(BACKGROUND_SCALE * ratio, BACKGROUND_SCALE * ratio, 1.0);
					card.alpha = Math.max(0, 1.0 - (0.5 * ratio));
				} else {
					ratio = 0;
					card.alpha = 1.0;
				}
				positionMatrix.prependRotation(CARD_ROTATION * ratio, Vector3D.Y_AXIS);
				
				card.setLayoutMatrix3D(positionMatrix, true);
				
				if (card is ISimpleCard) {
					(card as ISimpleCard).enableThumbnailGeneration = useVirtualLayout ? true : cardLayoutHelper.elementIntersectsRect(card, scrollRect);
				}
			}
			if (!animating) {
				if (skipNextPreload) {
					skipNextPreload = false;
					// We'll need to draw again with skipNextPreload set to false
					target.callLater(target.invalidateDisplayList);
				}
				if (pendingFocus != null) {
					setFocusAttachment(pendingFocus, currentFocus);
				}
			}
		}
		
		/**
		 *  Find the horizontal distance needed to scroll to reach
		 *  the specified element. The element is identified by its index number.
		 * 
		 *  @param index The index value for the element to which you 
		 *  want to scroll
		 *  @return Returns the distance to the element as an ordered pair (delta_x, 0) 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function getScrollPositionDeltaToElement(index:int):Point
		{
			var elementR:Rectangle = getElementBounds(index);
			if (!elementR)
				return null;
			
			var scrollR:Rectangle = getScrollRect();
			if (!scrollR || !target.clipAndEnableScrolling)
				return null;
			
			// We only scroll horizontally
			return new Point(elementR.x - scrollR.x, 0);
		}
		
		/**
		 *  Set the target's focus to be an IAttachment
		 * 
		 *  @param newFocus The IAttachement element that is to be the new focus
		 *  @param oldFocus the IAttachment element is the the current focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function setFocusAttachment(newFocus:IAttachment, oldFocus:IAttachment):void
		{
			if (target != null)
				target.callLater(setFocusAttachmentInner, [newFocus, oldFocus]);
			else
				pendingFocus = newFocus;
		}
		
		// TODO It might be nice if CardList called this, rather than handling it all in the layout
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function setFocusAttachmentInner(newFocus:IAttachment, oldFocus:IAttachment):void
		{
			if ((newFocus == oldFocus) || (newFocus == currentFocus)) {
				return;
			}
			if ((target == null) || (focusEffect != null)) {
				pendingFocus = newFocus;
				return;
			}
			var newIndex:int = cardLayoutHelper.getCurrentIndex(newFocus);
			var oldIndex:int = cardLayoutHelper.getCurrentIndex(oldFocus);
			
			if (newIndex < 0) {
				// No item to transition to
				pendingFocus = newFocus;
				return;
			}
			pendingFocus = null;
					
			var count:int = target.numElements;
			if (newIndex >= count) {
				if (count < 0) {
					// We shouldn't get into this situation, but if we do, we want to avoid potential wraparound
					// (and might be that we're being attacked)
					CONFIG::logging{
						LogHelper.logger.warn("LinearLayout:setFocusAttachment - target.numElements is less than 0");
					}
					newIndex = -1;
				} else {
					newIndex = count-1;
				}
			}
			if (newIndex < 0) {
				currentFocus = null;
			} else {
				focusEffect = generateFocusEffect(oldIndex, newIndex, FOCUS_DURATION);
				lastFocus = currentFocus;
				if (focusEffect != null) {
					lastFocus = currentFocus;
					currentFocus = newFocus;
					focusEffect.addEventListener(EffectEvent.EFFECT_END, onEffectEnd);
					//target.autoLayout = false;
					focusEffect.play();
				} else {
					currentFocus = newFocus;
					horizontalScrollPosition = getScrollPositionForIndex(newIndex);
				}
			}
		}
		
		/**
		 *  Set the x coordinate of the origin of the viewport in the
		 *  target's coordinate system, where 0 corresponding to the left side
		 *  of the target
		 * 
		 *  @param value The x coordinate, in pixels of the viewport
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function set horizontalScrollPosition(value:Number):void
		{
			if (value != horizontalScrollPosition) {
				super.horizontalScrollPosition = value;
				if (target != null) {
					target.invalidateDisplayList();
					target.validateNow();
				}
			}
		}

		/**
		 *  Create a focus effect for the change of focus.
		 * 
		 *  @param oldIndex The index value of the card that currently 
		 *  has the focus
		 *  @param newIndex The index value of the card that is getting
		 *  the focus
		 *  @param duration The time, in milliseconds, for the focus effect to last
		 *  @return The focus effect to run
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function generateFocusEffect(oldIndex:int, newIndex:int, duration:Number):Effect
		{
			if (target == null) {
				return null;
			}
			var newCard:IVisualElement = target.getElementAt(newIndex);
			if (newCard == null) {
				// The card we want to focus on doesn't exist yet, so we can't generate the effect
				return null;
			}
			if (useVirtualLayout) {
				// If we're using a virtual layout, we should only animate the scroll if we
				// we have all the cards in between.
				if (oldIndex < 0) {
					return null;
				}
				var fromIndex:int = Math.min(oldIndex, newIndex);
				var toIndex:int = Math.max(oldIndex, newIndex);
				for (var i:int = fromIndex; i <= toIndex; ++i) {
					if (target.getElementAt(i) == null) {
						return null;
					}
				}
			}
			
			var newHorizontalPosition:Number = getScrollPositionForIndex(newIndex);
			if (newHorizontalPosition == horizontalScrollPosition) {
				return null;
			}
			
			var effect:Parallel = new Parallel(this);
			var effectChildren:Array = new Array();
			//var move:Move = null;
			//var easeIn:Power = new Power(.6, 6);
			//var easeOut:Power = new Power(.4, 6);

			var scrollEffect:Animate = AnimateFactory.createAnimate(this, duration, "horizontalScrollPosition", newHorizontalPosition);
			if (scrollEffect != null) {
				effectChildren.push(scrollEffect);
			}
			
			// TODO I'd like to animate the cards I'm interested in directly, but it's not
			// behaving well with the "real" layout in updateDisplayList
			/*
			for (var i:int = 0; i < target.numElements; ++i) {
				if ((i > oldIndex) && (i > newIndex)) {
					break;
				}
				if ((i < oldIndex) && (i < newIndex)) {
					continue;
				}
				var currentElement:IVisualElement = target.getElementAt(i);
				if (currentElement == null) {
					continue;
				}
				var moveEffect:Move3D = new Move3D(currentElement);
				if (i > oldIndex) {
					moveEffect.zFrom = BACKGROUND_DISTANCE;
				} else {
					moveEffect.zFrom = 0;
				}
				if (i > newIndex) {
					moveEffect.zTo = BACKGROUND_DISTANCE;
				} else {
					moveEffect.zTo = 0;
				}
				moveEffect.captureStartValues();
				if (moveEffect.zFrom == moveEffect.zTo) {
					continue;
				}
				effectChildren.push(moveEffect);
			}
			*/

			effect.children = effectChildren;
			return effect;
		}
		
		/**
		 *  Clean up at the end of an effect
		 * 
		 *  @param evt The effect event that was listening for the effect end
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onEffectEnd(evt:EffectEvent):void
		{
			(evt.target as IEventDispatcher).removeEventListener(EffectEvent.EFFECT_END, onEffectEnd);
			focusEffect = null;
			dispatchEvent(new LayoutEvent(LayoutEvent.HANDLED_ATTACHMENT_FOCUS));
			if (pendingFocus != null) {
				// we should already have a new focus, so change now
				setFocusAttachment(pendingFocus, currentFocus);
			} else if (target != null) {
				target.invalidateDisplayList();
				target.validateNow();
				//target.autoLayout = true;
			}
		}
		
		/**
		 *  Handle the specified navigator host event.
		 * 
		 *  @param evt The navigator host event to handle
		 *  primarily this will be either an ITEMS_DID_CHANGE
		 *  event or a FOCUS_DID_CHANGE_KIND event.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onNavHostEvent(evt:CollectionManagerEvent):void
		{
			switch (evt.kind) {
				case CollectionManagerEvent.ITEMS_DID_CHANGE_KIND:
					// Validate now, to make sure the visibility of items is correct
					if (target != null) {
						target.validateNow();
					}
					break;
				case CollectionManagerEvent.FOCUS_DID_CHANGE_KIND:
					setFocusAttachment(evt.newVal as IAttachment, evt.oldVal as IAttachment);
					break;
				default:
					break;
			}
			
		}

		/**
		 *  The instance of CardLayoutHelper used by this layout.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get cardLayoutHelper():CardLayoutHelper
		{
			return _cardLayoutHelper;
		}
		
		/**
		 *  Find the list of cards that overlap the specified marquee rectangle. 
		 * 
		 *  @param marqueeRect The bounding rectangle for the marquee
		 *  @return A vector of integers that represent the indexes
		 *  of the cards in the layout that intersect the specified marqueeRect.
		 * 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function findCardsThatIntersect(marqueeRect:Rectangle):Vector.<int>
		{
			return null; // only necessary for marquee support
		}

		/**
		 *  The largest size of a thumbnail image that will fit in the layout.
		 * 
		 *  @param value the size of the thumbnail given as an ordered pair
		 *  of pixel values: (width, height)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get thumbFitSize():Point
		{
			return _thumbFitSize;
		}
		
		/**
		 *  @private
		 */
		public function set thumbFitSize(value:Point):void
		{
			_thumbFitSize = value;
		}
	}
}
