/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.layouts
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.events.LayoutEvent;
	import com.adobe.portfolio.layouts.supportClasses.CardLayoutHelper;
	import com.adobe.portfolio.layouts.supportClasses.ICardLayout;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.ThumbnailManager;
	import com.adobe.portfolio.utils.LogHelper;
	
	import flash.events.IEventDispatcher;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	import mx.core.IVisualElement;
	import mx.effects.Effect;
	import mx.effects.Parallel;
	import mx.events.EffectEvent;
	
	import spark.components.supportClasses.ListBase;
	import spark.effects.Animate;
	import spark.effects.SetAction;
	import spark.effects.animation.Keyframe;
	import spark.effects.animation.MotionPath;
	import spark.effects.easing.Power;
	import spark.layouts.HorizontalLayout;
	import spark.layouts.VerticalAlign;
	import spark.layouts.supportClasses.DropLocation;

	
	/**
	 *  Dispatched after the layout has set its focus to a valid attachment.
	 *
	 *  @eventType com.adobe.portfolio.events.NavLayoutEvent.HANDLED_ATTACHMENT_FOCUS
	 */
	[Event(name="HANDLED_ATTACHMENT_FOCUS", type="com.adobe.portfolio.events.LayoutEvent")]
	
	/**
	 * The MiniNavLayout class extends HorizontalLayout to implement the layout of the MiniNav component.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class MiniNavLayout extends HorizontalLayout implements ICardLayout
	{
		
		/** The navigator's CollectionManager instance. All host interaction should go through this
		 *  instance, rather than the INavigatorHost directly.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected const collectionMgr:CollectionManager = CollectionManager.instance;
		
		/** The attachment which was last in focus. 
		 * 
		 *  @langversion 3.0
	 	 *  @playerversion Flash 10.1
	 	 *  @playerversion AIR 2.0
	 	 *  @productversion Portfolio 10.0
	 	 */
		protected static var lastFocus:IAttachment = null;
		
		/**
		 *  The time duration, in milliseconds, for a focus effect to last
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected static const FOCUS_DURATION:Number = 150;
		
		/**
		 *  The maxmum height for a thumbnail image in the MiniNavLayout
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected static const ITEM_MAX_HEIGHT:Number = 170;
		
		/**
		 *  The padding amount around images in the MiniNavLayout
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected static const PADDING:Number = 10;

		protected var focusAttachment:IAttachment = null;
		protected var pendingFocus:IAttachment = null;
		protected var pendingDuration:Number;
		protected var focusEffect:Effect;
		
		private var _cardLayoutHelper:CardLayoutHelper;
		private var _list:ListBase;
		private var _thumbFitSize:Point;

		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function MiniNavLayout()
		{
			super();
			columnWidth = 200;
			variableColumnWidth = true;
			gap = 20;
			paddingLeft = 10;
			paddingRight = 10;
			paddingTop = 15;
			paddingBottom = 15;
			verticalAlign = VerticalAlign.BOTTOM;
			useVirtualLayout = false;
			cardLayoutHelper = new CardLayoutHelper(this);
			collectionMgr.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onHostEvent);
			_thumbFitSize = new Point(ThumbnailManager.SMALL_THUMB_FIT_WIDTH, ThumbnailManager.SMALL_THUMB_FIT_HEIGHT);
		}
		
		/**
		 *  Updates the MiniNavLayouts display list.
		 * 
		 *  @param width The width of the display list in pixels
		 *  @param height The height of the display list in pixels
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function updateDisplayList(width:Number, height:Number):void
		{
			//Maxing out height for thumbnail card height, as this gets the default value of the basic card on a redraw.
			height = height > ITEM_MAX_HEIGHT ? ITEM_MAX_HEIGHT : height;
			
			var animating:Boolean = ((focusEffect != null) && focusEffect.isPlaying);
			var pendingFocusIndex:int = getPendingFocusIndex();
			
			if (!animating && (pendingFocusIndex >= 0)) {
				var pendingCard:IVisualElement = target.getElementAt(pendingFocusIndex);
				if (pendingCard != null) {
					// We wanted to transition to this card, but it didn't exist yet; do it now
					target.callLater(setFocusAttachmentInner, [pendingFocus, focusAttachment]);
				}
			}

			// Let our parent class layout elements using the default size
			super.updateDisplayList(width, height);
			
			if (!animating) {
				ensureFocusCentered(lastFocus);
				// This is a relatively expensive call, so don't make it when we're animating
				cardLayoutHelper.enableThumbnailGeneration();
			}
		}
			
		/**
		 *  Get the index value of the pending focus item from the collection manager
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function getPendingFocusIndex():int
		{
			if (focusAttachment != collectionMgr.focusAttachment) {
				if (getCurrentIndex(collectionMgr.focusAttachment) >= 0) {
					CONFIG::logging{
						LogHelper.logger.warn("currentFocus didn't match collectionManager.focusAttachment in mini Nav");
					}
					pendingFocus = collectionMgr.focusAttachment;
				}
			}
			
			return getCurrentIndex(pendingFocus);
		}

		
		/**
		 *  Make sure the display is centered on the last focus item.
		 *  If needed, a pan animation is created to center the display.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function ensureFocusCentered(focus:IAttachment):void
		{
			//To ensure we are always centered on the last focus item.
			var lastFocusIndex:int = getCurrentIndex(focus);
			var delta:Point = getScrollPositionDeltaToElement(lastFocusIndex);
			if (delta != null)
			{
				if (delta.x == 0)
					return;
				createPanAnimation(delta.x).play();
			}
		}
		
		
		/**
		 *  Find the x and y distances needed to scroll to reach the 
		 *  specified element.  The element is identified by its index number
		 * 
		 *  @param index The index value of the element for which you want to know
		 *  the distance to scroll.
		 *  @return Returns the distance to the element as an ordered pair (delta_x, delta_y)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function getScrollPositionDeltaToElement(index:int):Point
		{
			var elementR:Rectangle = getElementBounds(index);
			if (!elementR)
				return null;
			
			var scrollR:Rectangle = getScrollRect();
			if (!scrollR || !target.clipAndEnableScrolling)
				return null;
			
			return new Point( (elementR.x + (elementR.width/2)) - (scrollR.x + (scrollR.width/2)),
				(elementR.y + (elementR.height)) - (scrollR.y + (scrollR.height)));
			
		}
		
		
		/**
		 *  Set the focus to a new element in the Portfolio
		 * 
		 *  @param newFocus The IAttachment that is to be the new focus
		 *  @param oldFocus The IAttachment that is the current (old) focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function setFocusAttachment(newFocus:IAttachment, oldFocus:IAttachment = null):void
		{
			if (target != null) {
				target.callLater(setFocusAttachmentInner, [newFocus, oldFocus]);
				lastFocus = newFocus;
			} else {
				pendingFocus = newFocus;
			}
		}
		
		
		/**
		 * @private
		 * Internal function used to set a new attachment to focus.
		 * 
		 *  @param newFocus The IAttachment that is to be the new focus
		 *  @param oldFocus The IAttachment that is the current (old) focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function setFocusAttachmentInner(newFocus:IAttachment, oldFocus:IAttachment):void
		{
			if ((newFocus == oldFocus) || (newFocus == focusAttachment)) {
				return;
			}
			if (focusEffect) {
				// we're already changing focus, so keep attachment as pending
				// TODO pendingFocusAttachment = attachment;
				pendingFocus = newFocus;
				return;
			}
				
			pendingFocus = null;
			
			if (!newFocus) {
				focusAttachment = null;
				return;
			} 
			
			var newIndex:int = getCurrentIndex(newFocus);
			if (newIndex < 0) {
				// We're probably in a folder, so just return
				return;
			}
			
			var oldIndex:int = getCurrentIndex(oldFocus);
			focusEffect = generateFocusEffect(oldIndex, newIndex, FOCUS_DURATION);
			if (focusEffect) {
				focusAttachment = newFocus;
				focusEffect.addEventListener(EffectEvent.EFFECT_END, onEffectEnd);
				target.callLater(focusEffect.play);
			} else {
				//dispatchEvent(new NavLayoutEvent(NavLayoutEvent.HANDLED_ATTACHMENT_FOCUS));
				pendingFocus = newFocus;
			}
		}
		
		/**
		 *  Create a focus change effect
		 * 
		 *  @param oldIndex The index value of the element that is the current (old) focus
		 *  @param newIndex The index value of the element that is to be the new focus
		 *  @param duration The duration, in milliseconds, the effect should last
		 *  @return The focus effect object
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function generateFocusEffect(oldIndex:int, newIndex:int, duration:Number):Effect
		{
			if (newIndex < 0)
				return null;
		
			if (oldIndex >= 0) {
				// just use the standard panning effect
				return generatePanEffect(newIndex, duration);
			}

			// if there's no previous index to pan from, forget the animation and just set the position
			var delta:Point = getScrollPositionDeltaToElement(newIndex);
			if ((delta != null) && ((delta.x != 0))) {
				var effect:Parallel = new Parallel(this);
				effect.children = new Array();
				var setAction:SetAction = new SetAction(this);
				setAction.property = "horizontalScrollPosition";
				setAction.value = horizontalScrollPosition + delta.x;
				setAction.duration = 0;
				effect.children.push(setAction);
				return effect;
			}
			return null;
		}
		
		/**
		 *  Generate the appropriate pan effect to center on the sepcified element 
		 *  in the element list.
		 * 
		 *  @param index The index value of the element to which to pan
		 *  @param dur The duration, in milliseconds, the effect should last
		 *  @return A parallel animation effect object
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function generatePanEffect(index:int, dur:Number):Parallel
		{
			var delta:Point = getScrollPositionDeltaToElement(index);
			if ((delta == null) || (delta.x == 0))
				return null;

			var effect:Parallel = new Parallel(this);
			effect.children = new Array();
			effect.children.push(createPanAnimation(delta.x, dur));
			
			return effect;
		}
		
		/**
		 *  Create a pan animation effect
		 * 
		 *  @param delta The distance to pan; this may be either positive or negative.
		 *  @param duration The time duration, in milliseconds of the pan animation effect.
		 *  @param applyEaser If <code>true</code> apply easing beahvior for this effect.
		 *  This will convert the elapsed fraction of the animation into an eased fraction, 
		 *  which is then used to calculate the value at that eased elapsed fraction.
		 *  @return The animation effect object
		 * 
		 *  @See spark.effects.Animate.easer 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function createPanAnimation(delta:Number, duration:Number = FOCUS_DURATION, applyEaser:Boolean = true):Animate
		{
			var animation:Animate = new Animate(this);
			var motionPaths:Vector.<MotionPath> = new Vector.<MotionPath>();
			var path:MotionPath = new MotionPath("horizontalScrollPosition");
			var keyframes:Vector.<Keyframe> = new Vector.<Keyframe>();
			keyframes.push(new Keyframe(0, horizontalScrollPosition));
			keyframes.push(new Keyframe(duration, horizontalScrollPosition + delta));
			path.keyframes = keyframes;
			motionPaths.push(path);
			if (applyEaser)
				animation.easer = new Power(.3, 4);
			animation.motionPaths = motionPaths;
			
			return animation;
		}
		
		/**
		 *  Cleans up after an effect has finished playing.
		 * 
		 *  @param evt The NavLayout event effect that is to be handled
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onEffectEnd(evt:EffectEvent):void
		{
			(evt.target as IEventDispatcher).removeEventListener(EffectEvent.EFFECT_END, onEffectEnd);
			focusEffect = null;
			dispatchEvent(new LayoutEvent(LayoutEvent.HANDLED_ATTACHMENT_FOCUS));
			if (pendingFocus != null) {
				// we should already have a new focus, so change now
				target.callLater(setFocusAttachmentInner, [pendingFocus, focusAttachment]);
				//setFocusAttachment(pendingFocusAttachment, prevAttachment);
			}
			cardLayoutHelper.enableThumbnailGeneration();
		}
		
		/**
		 *  Generate the scrolling effect needed when dragging an item beyond the visible elements 
		 *  in the MiniNavLayout so that the hidden items come to view.
		 * 
		 *  @dropIndex The index value for where the item is to be dropped in the 
		 *  list of elements. If this is the last or first item visible, scroll to show the hidden elements.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function generateScrollEffect(dropIndex:int):void
		{
			// Don't go past and beginning.
			if ((dropIndex == target.numElements - 1 || dropIndex == 0 ) && fractionOfElementInView(dropIndex) == 1)
				return;
			
			var scrollPastWidth:Number = 0;
			
			if (dropIndex == lastIndexInView)
			{
				var element1:IVisualElement = target.getElementAt(dropIndex);
				scrollPastWidth = element1 != null ? element1.width + PADDING : 0;
			}
			else if (dropIndex == firstIndexInView || dropIndex == firstIndexInView + 1 && firstIndexInView != 0)
			{
				var element2:IVisualElement = target.getElementAt(dropIndex < target.numElements ? dropIndex : firstIndexInView);
				scrollPastWidth =  element2 != null ? -element2.width + PADDING : 0;
			}
			if (scrollPastWidth != 0)
				createPanAnimation(scrollPastWidth, 100, false).play();
		}
		
		
		/**
		 * Listener that responds to events from the Collection Manager.
		 * 
		 *  @param evt The Collection Manager event to be handled
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onHostEvent(evt:CollectionManagerEvent):void
		{
			switch (evt.kind) {
				case CollectionManagerEvent.FOCUS_DID_CHANGE_KIND:
					setFocusAttachment(evt.newVal as IAttachment, evt.oldVal as IAttachment);
					break;
			}
		}
		
		/**
		 *  Sizes, positions and parents the drop indicator based on the specified drop location. 
		 *  Use the calculateDropLocation() method to obtain the DropLocation object. 
		 *  Display the visual drop indicator for drag and drop operations.
		 * 
		 *  @param dropLocation Specifies the location to dispaly the drop indicator.
		 *  Drop location is obtained through the <code>computeDropLocation()</code> method.
		 * 
		 *  @See spark.layouts.supportClasses.LayoutBase.showDropIndicator
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function showDropIndicator(dropLocation:DropLocation):void
		{
			// Scroll the list if the drag location is on the edge of the view and there are more items available
			generateScrollEffect(dropLocation.dropIndex);
			
			// TODO: LayoutBase has its own scroll processing. Check if that does not interfere with the above
			// scroll processing. It would be even better if this could use that instead of a custom one.
			super.showDropIndicator(dropLocation);
			
			if (dropLocation.dropIndex < 0)
				return;
			
			var element:IVisualElement = target.getElementAt(dropLocation.dropIndex);
			if (!element)
			{
				dropIndicator.height = 0;
				return;
			}
			
			// Make the dropIndicator the height of the element next to which it is being drawn. 
			dropIndicator.height = element.height;
			var p:Point = new Point(element.x, element.y);
			dropIndicator.y =  element.parent.localToGlobal(p).y;
		}
		
		/**
		 *  @private
		 */
		private function getCurrentIndex(attachment:IAttachment):int
		{
			return cardLayoutHelper.getCurrentIndex(attachment);
		}
		
		/**
		 *  The width and height for small thumbnail images to fit in the MiniNavLayout.
		 *  This is given as an ordered pair of numbers: (width, height)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get thumbFitSize():Point
		{
			return _thumbFitSize;
		}
		/**
		 * @private
		 */
		public function set thumbFitSize(value:Point):void
		{
			// use the default
		}

		/**
		 *  The current card layout helper
		 * 
		 *  @return a CardLayoutHelper component object
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get cardLayoutHelper():CardLayoutHelper
		{
			return _cardLayoutHelper;
		}
		/**
		 *  @private
		 */
		public function set cardLayoutHelper(value:CardLayoutHelper):void
		{
			_cardLayoutHelper = value;
		}
		
		/**
		 *  Find the list of cards that overlap the specified rectangle
		 * 
		 *  @param marqueeRect The bounding rectangle for the marquee
		 *  @return A vector of integers that represent the indexes
		 *  of the cards in the layout that intersect the specified marqueeRect.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function findCardsThatIntersect(marqueeRect:Rectangle):Vector.<int>
		{
			return null; // only necessary for marquee support
		}
	}
}
