/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.layouts
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.api.INavigatorHost;
	import com.adobe.portfolio.cards.ICardProvider;
	import com.adobe.portfolio.cards.ISimpleCard;
	import com.adobe.portfolio.components.Navigator;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.events.LayoutEvent;
	import com.adobe.portfolio.layouts.supportClasses.CardLayoutHelper;
	import com.adobe.portfolio.layouts.supportClasses.ICardLayout;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.utils.LogHelper;
	
	import flash.display.DisplayObject;
	import flash.display.DisplayObjectContainer;
	import flash.events.Event;
	import flash.events.IEventDispatcher;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	import mx.collections.IList;
	import mx.core.IVisualElement;
	import mx.effects.Effect;
	import mx.effects.Parallel;
	import mx.events.EffectEvent;
	import mx.managers.IFocusManager;
	import mx.managers.IFocusManagerComponent;
	
	import spark.effects.Move;
	import spark.effects.easing.Power;
	import spark.layouts.supportClasses.LayoutBase;
	
	/**
	 *  Dispatched after the layout has set its focus to a valid attachment.
	 *
	 *  @eventType com.adobe.portfolio.events.NavLayoutEvent.HANDLED_ATTACHMENT_FOCUS
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	[Event(name="HANDLED_ATTACHMENT_FOCUS", type="com.adobe.portfolio.events.LayoutEvent")]
	
	/**
	 * 
	 * The OneUpLayout class is a LayoutBase that implements the layout of the BasicClickThrough navigator.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class OneUpLayout extends LayoutBase implements ICardLayout, ICardProvider
	{
		/**
		 *  The instance of the CollectionManager class used by this Navigator
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected const collectionManager:CollectionManager = CollectionManager.instance;
		
		/**
		 *  The time, in mimiseconds for focus effects to run
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected const FOCUS_DURATION:Number = Navigator.FOCUS_DURATION;
		
		/**
		 *  The IAttachment instance that currently has the focus.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var currentFocus:IAttachment;
		
		/**
		 *  The IAttachment instance that is will next get the focus (pending)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var pendingFocus:IAttachment;
		
		/**
		 *  The IAttachment instance that last had the focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var lastFocus:IAttachment;
		
		/**
		 *  The time, in mimiseconds for pending effects to run
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var pendingDuration:Number = 0;
		
		/**
		 *  The effect to currently playing due to a focus change
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected var focusEffect:Effect = null;
		
		private var host:INavigatorHost;
		private var _cardLayoutHelper:CardLayoutHelper;
		private var _thumbFitSize:Point;
		private var _cardPosition:Number;
		private var _pendingCardPosition:Number;
		private var _maxCardPosition:Number;
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function OneUpLayout()
		{
			super();
			useVirtualLayout = true;
			collectionManager.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onHostEvent, false, 0, true);
			_cardLayoutHelper = new CardLayoutHelper(this);
		}
		
		/**
		 *  Measures the target's default size based on its content.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function measure():void
		{
			super.measure();
			
			// Set default size of target
			target.measuredWidth = 500;
			target.measuredHeight = 400;
		}
		
		/**
		 *  Handles the bookeeping when an element is added
		 * 
		 *  @param index The index value of the element that has been added
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function elementAdded(index:int):void
		{
			super.elementAdded(index);
			updateMaxCardPosition();
		}
		
		/**
		 *  Handles the bookeeping when an element is removed
		 * 
		 *  @param index The index value of the element that has been removed
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function elementRemoved(index:int):void
		{
			super.elementRemoved(index);
			updateMaxCardPosition();
		}
		
		/**
		 *  Updates the position of the maximum card (last card in the list)
		 *  This is called when elements have been added or removed from the list of cards
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function updateMaxCardPosition():void
		{
			maxCardPosition = cardPositionFromIndex(target.numElements - 1);  
		}
		
		/**
		 *  update the display list
		 * 
		 *  @param width The new width of the target, in pixels, in the targets's coordinates.
		 *  @param height The new height of the target, in pixels, in the targets's coordinates.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function updateDisplayList(width:Number, height:Number):void
		{
			super.updateDisplayList(width, height);

			var animating:Boolean = ((focusEffect != null) && focusEffect.isPlaying);
			var count:int = target.numElements;
			var maxX:Number = 0;
			var maxY:Number = 0;
			
			if (currentFocus != collectionManager.focusAttachment) {
				if (getCurrentIndex(collectionManager.focusAttachment) >= 0) {
					CONFIG::logging{
						LogHelper.logger.warn("currentFocus didn't match collectionManager.focusAttachment");
					}
					pendingFocus = collectionManager.focusAttachment;
				}
			}
			var focusIndex:int = getCurrentIndex(currentFocus);
			var pendingIndex:int = getCurrentIndex(pendingFocus);
			var lastIndex:int = getCurrentIndex(lastFocus);
			if (focusIndex >= count) {
				if (count < 0) {
					// We shouldn't get into this situation, but if we do, we want to avoid potential wraparound
					// (and might be that we're being attacked)
					CONFIG::logging{
						LogHelper.logger.warn("OneUpLayout:updateDisplayList - target.numElements is less than 0");
					}
					focusIndex = -1;
				} else {
					focusIndex = count-1;
				}
			}
			for (var i:int = 0; i < count; ++i)
			{
				var shouldBeInBounds:Boolean = false;
				if ((i == focusIndex) || (i == pendingIndex) || (i == lastIndex))
				{
					shouldBeInBounds = true;
				} else if (!animating && (pendingIndex < 0) && ((i == focusIndex+1) || (i == focusIndex-1)))
				{
					// Only setup the next/previous cards if we're not animating, and we're not about to
					shouldBeInBounds = true;
				}
				var layoutElement:IVisualElement = null;
				if (useVirtualLayout) {
					if (!shouldBeInBounds) {
						// This renderer isn't important enough to create, but we should update its
						// properties if it already exists.
						layoutElement = target.getElementAt(i);
					} else if (animating) {
						// We're animating, so just hold onto the renderers that have already been created
						layoutElement = target.getElementAt(i);
						if (layoutElement != null) {
							// If getElementAt succeeded, then call getVirtualElementAt
							// to make sure this renderer isn't cleaned up yet
							layoutElement = target.getVirtualElementAt(i);
						}
					} else {
						// We're not animating, and this renderer is important enough to be created now
						layoutElement = target.getVirtualElementAt(i);
					}
				} else {
					layoutElement = target.getElementAt(i);
				}
				if ((layoutElement == null) || !layoutElement.includeInLayout)
					continue;
				
				// If we're not animating, then go ahead and set the card's visiblity, size, and position
				var shouldPositionCard:Boolean = !animating;
				if (!shouldPositionCard && (layoutElement.getLayoutBoundsY() == 0)) {
					// This case has caused major strangeness. The layout seems to be reset
					// under some circumstances, and I'm not sure how to detect it. This block of
					// code will break the currently-running animation, but at least it gets the cards
					// back into an ok state.
					LogHelper.logger.error("NavLayoutBasic: elements were reset, and we're animating");
					shouldPositionCard = true;
				}
				if (shouldPositionCard) {
					var card:ISimpleCard = layoutElement as ISimpleCard;
					var visible:Boolean = (i == focusIndex);
					card.visible = visible;
					card.enableThumbnailGeneration = Math.abs(i-focusIndex) <= 1;
					
					if (shouldBeInBounds) {
						var fitSize:Number = 0.8 * Math.min(width,height);
						card.setLayoutBoundsSize(fitSize, fitSize);
						layoutElement.setLayoutBoundsPosition(	(width - card.getLayoutBoundsWidth()) / 2,
																(height - card.getLayoutBoundsHeight()) / 2);
					}
				}
				if (!animating && (i == pendingIndex)) {
					// We wanted to transition to this card, but it didn't exist yet; do it now
					target.callLater(setFAInner, [pendingFocus, currentFocus]);
				}
			}
		}
		
		/**
		 *  Set the target's focus to be the specified attachment
		 * 
		 *  @param attachment The IAttachement element that is to be the new focus
		 *  @param oldFocus The previous FocusAttachment
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function setFocusAttachment(attachment:IAttachment, oldFocus:IAttachment = null):void
		{
			if (target != null) {
				target.callLater(setFAInner, [attachment, oldFocus]);
			} else {
				pendingFocus = attachment;
			}
		}
		
		/**
		 *  This is a helper function called by setFocusAttachment,
		 *  updateDisplayList and other methods to set the focus
		 *  to the specified IAttachement instance, when that action needs to be
		 *  defered.
		 * 
		 *  @param newFocus The IAttachment instance that is getting the focus
		 *  @param oldFocus The IAttachment instance that just had the focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function setFAInner(newFocus:IAttachment, oldFocus:IAttachment):void
		{
			if ((newFocus == oldFocus) || (newFocus == currentFocus)) {
				return;
			}
			if (focusEffect != null) {
				// we're already changing focus, so keep attachment as pending
				// TODO pendingFocusAttachment = attachment;
				pendingFocus = newFocus;
				return;
			}
			
			if (!target.focusManager)
				return;
			
			pendingFocus = null;
			
			var newIndex:int = getCurrentIndex(newFocus);
			if (newIndex < 0) {
				return;
			}
			
			var oldIndex:int = getCurrentIndex(oldFocus);
			setFocus(newIndex, oldIndex);
			
			var count:int = target.numElements;
			if (newIndex >= count) {
				if (count < 0) {
					// We shouldn't get into this situation, but if we do, we want to avoid potential wraparound
					// (and might be that we're being attacked)
					CONFIG::logging{
						LogHelper.logger.warn("OneUpLayout:setFAInner - target.numElements is less than 0");
					}
					newIndex = -1;
				} else {
					newIndex = count-1;
				}
			}
			if (newIndex < 0) {
				currentFocus = null;
			} else {
				// TODO prevAttachment = attachment;
				focusEffect = generateFocusEffect(oldIndex, newIndex, FOCUS_DURATION);
				if (focusEffect != null) {
					lastFocus = oldFocus;
					currentFocus = newFocus;
					focusEffect.addEventListener(EffectEvent.EFFECT_END, onEffectEnd);
					focusEffect.play();
				} else {
					//dispatchEvent(new NavLayoutEvent(NavLayoutEvent.HANDLED_ATTACHMENT_FOCUS));
					pendingFocus = newFocus;
					target.invalidateDisplayList();
				}
			}
		}
		
		/**
		 *  Handled the change of focus from IAttachement instance that currently has the focus
		 *  to a new IAttachment instance with the given index value.
		 * 
		 *  @param newFocus The index value of the IAttachment instance that is getting the focus
		 *  @param oldFocus The index value of the IAttachment instance that just had the focus
		 * 
		 *  @param newIndex
		 *  @param oldIndex
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function setFocus(newIndex:int, oldIndex:int):void {
			if ((oldIndex < 0) || (oldIndex >= target.numElements))
				return;
			
			var oldElement:DisplayObjectContainer = target.getElementAt(oldIndex) as DisplayObjectContainer;
			if (oldElement == null)
				return;
			
			var focusManager:IFocusManager = target.focusManager;
			var currentFlexFocus:DisplayObject = focusManager.getFocus() as DisplayObject;
			if (currentFlexFocus == null) 
				return;

			if (!oldElement.contains(currentFlexFocus)) 
				return;

			// We're about to fade out the element that has the current Flex focus, so move it elsewhere first
			var newElement:IFocusManagerComponent;
			if ((newIndex >= 0) && (newIndex < target.numElements)) {
				newElement = target.getElementAt(newIndex) as IFocusManagerComponent;
				if (newElement != null) { 
					focusManager.setFocus(newElement);
				}
			}
			if (!newElement && collectionManager.defaultFocusItem != null) {
				focusManager.setFocus(collectionManager.defaultFocusItem);
			}
		}
		
		/**
		 *  Handler for effect end events
		 * 
		 *  @param evt The effect end event to be handled
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onEffectEnd(evt:EffectEvent):void
		{
			(evt.target as IEventDispatcher).removeEventListener(EffectEvent.EFFECT_END, onEffectEnd);
			focusEffect = null;
			dispatchEvent(new LayoutEvent(LayoutEvent.HANDLED_ATTACHMENT_FOCUS));
			
			// Null out lastFocus, since updateDisplayList no longer needs to make sure
			// its renderer sticks around
			lastFocus = null;
			
			// Let updateDisplayList handle pendingFocus as needed
			target.invalidateDisplayList();
		}
		
		/**
		 *  Create a focus change effect
		 * 
		 *  @param oldIndex The index value of the element that is the current (old) focus
		 *  @param newIndex The index value of the element that is to be the new focus
		 *  @param duration The duration, in milliseconds, the effect should last
		 *  @return The focus effect object
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function generateFocusEffect(oldIndex:int, newIndex:int, duration:Number):Effect
		{
			var oldCard:IVisualElement = (oldIndex >= 0) ? target.getElementAt(oldIndex) : null;
			var newCard:IVisualElement = target.getElementAt(newIndex);
			if (newCard == null) {
				// The card we want to focus on doesn't exist yet, so we can't generate the effect
				return null;
			}
			
			var effect:Parallel = new Parallel(this);
			var effectChildren:Array = new Array();
			var move:Move = null;
			var easeIn:Power = new Power(.6, 6);
			var easeOut:Power = new Power(.4, 6);
			
			var xDelta:Number = (target.width );
			if (newIndex > oldIndex) {
				xDelta = -xDelta;
			}
			
			if (oldCard != null) {
				
				move = new Move(oldCard);
				move.duration = duration;
				move.xBy = xDelta;
				move.easer = easeIn;
				effectChildren.push(move);
			}
			if (newCard != null) {
				newCard.visible = true;

				var lx:Number = newCard.getLayoutBoundsX();
				var ly:Number = newCard.getLayoutBoundsY();
				move = new Move(newCard);
				move.duration = duration;
				move.easer = easeOut;
				move.xTo = (target.width - newCard.width) / 2;
				move.xFrom = move.xTo - xDelta;
				effectChildren.push(move);
			}
			effect.children = effectChildren;
			return effect;
		}
		
		/**
		 *  Handler for Navigator host events passed to the Navigator
		 * 
		 *  @param evt The host event to handle
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onHostEvent(evt:CollectionManagerEvent):void
		{
			var newHost:INavigatorHost = evt.newVal as INavigatorHost;

			switch (evt.kind) {
				case CollectionManagerEvent.ITEMS_DID_CHANGE_KIND:
					// Validate now, to make sure the visibility of items is correct
					if (host && target != null) {
						target.validateNow();
					}
					break;
				case CollectionManagerEvent.FOCUS_DID_CHANGE_KIND:
					if (host)
						setFocusAttachment(evt.newVal as IAttachment, evt.oldVal as IAttachment);
					break;
				case CollectionManagerEvent.HOST_CHANGED_KIND:
					if (newHost) {
						host = newHost;
					} else if (host) {
						collectionManager.removeEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onHostEvent);
					} 
					break;
				default:
					break;
			}

		}
		
		/**
		 *  @private
		 *
		 */
		private function getCurrentIndex(attachment:Object, dataProvider:IList = null):int
		{
			if (attachment == null) {
				return -1;
			}
			if (dataProvider == null) {
				if ((target != null) && target.hasOwnProperty("dataProvider")) {
					dataProvider = target["dataProvider"];
				}
			}
			if (dataProvider == null) {
				return -1;
			}
			
			return dataProvider.getItemIndex(attachment);
		}
		
		/**
		 *  Create an animation effect for a gap in the list.
		 * 
		 *  @param splitIndex The position in the card index where the gap is to appear
		 *  @param duration The time, in mimiseconds, for the animation effect to last
		 *  @return The animation effect
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function generateGapEffect(splitIndex:int, duration:Number):Effect
		{
			return null;
		}

		/**
		 *  The instance of CardLayoutHelper used by this Navigator
		 * 
		 *  @return An instance of the CardLayoutHelper class
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get cardLayoutHelper():CardLayoutHelper
		{
			return _cardLayoutHelper;
		}
		
		/**
		 *  Find the list of cards that overlap the specified marquee rectangle. 
		 * 
		 *  @param marqueeRect A bounding rectangle
		 *  @return A vector of integers that represent the indexes
		 *  of the cards in the layout that intersect the specified marquee rectangle.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function findCardsThatIntersect(marqueeRect:Rectangle):Vector.<int>
		{
			return null; // only necessary for marquee support
		}

		/**
		 *  The size, represented as a width and height ordered pair, to use for thumbnails
		 * 
		 *  @param value The image size to use for thumbnails
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get thumbFitSize():Point
		{
			return _thumbFitSize;
		}
		/**
		 *  @private
		 * 
		 */
		public function set thumbFitSize(value:Point):void
		{
			_thumbFitSize = value;
		}

		/**
		 *  Returns the card position, in pixels, given
		 *  the cards index value.
		 * 
		 *  @param index The index value for the card
		 *  @return the position of the card, given in pixels
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function cardPositionFromIndex(elementIndex:int):Number
		{
			return elementIndex * 100;
		}
		
		/**
		 *  Returns the index value for the card at the given horizontal position
		 *  the cards index value.
		 * 
		 *  @param cardPosition The horizontal position of the card, in pixels
		 *  @return The index value of the card at the given position
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function cardIndexFromPosition(cardPosition:Number):int
		{
			return cardPosition / 100;
		}
		
		/**
		 *  The card position for the card that is pending as the current card.
		 * 
		 *  @param value The card position value for the pending card
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function set pendingCardPosition(value:Number):void 
		{
			if (_pendingCardPosition == value)
				return;
			
			_pendingCardPosition = value;
			
			if (isNaN(value))
				return;
			
			collectionManager.setSelectedItems([getCard(value).data as IAttachment]);
		}
		/**
		 *  @private
		 *
		 */
		public function get pendingCardPosition():Number
		{
			return _pendingCardPosition;
		}
		
		/**
		 *  Returns the card that is at the specified position
		 * 
		 *  @param cardPosition The position, in pixels, of the card
		 *  @return the instance of ISimpleCard located at the specified position
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function getCard(cardPosition:Number):ISimpleCard
		{
			return cardLayoutHelper.getLayoutElement(cardIndexFromPosition(cardPosition)) as ISimpleCard;
		}
		
		[Bindable (name="cardPositionChanged")]
		/**
		 *  The horizontal position of the card, in pixels.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get cardPosition():Number
		{
			return _cardPosition;
		}
		/**
		 *  @private
		 *
		 */
		public function set cardPosition(value:Number):void
		{
			if (_cardPosition == value)
				return;
			
			if (value < 0)
				return;
			
			_cardPosition = value;
			cardLayoutHelper.invalidateTarget();
			
			if (isNaN(pendingCardPosition)) {
				dispatchEvent(new Event("cardPositionChanged"));
			}
		}
		
		[Bindable (name="maxCardPositionChanged")]
		/**
		 *  The maximum horizontal position that a card can have.
		 *  This is dependent on the target width.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get maxCardPosition():Number
		{
			return _maxCardPosition;
		}
		/**
		 *  @private
		 *
		 */
		public function set maxCardPosition(value:Number):void
		{
			if (_maxCardPosition == value)
				return;
			
			_maxCardPosition = value;
			dispatchEvent(new Event("maxCardPositionChanged"));
		}
	}
}
