/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.layouts
{
	import com.adobe.portfolio.animations.AnimateFactory;
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.cards.ICardProvider;
	import com.adobe.portfolio.cards.ISimpleCard;
	import com.adobe.portfolio.components.supportClasses.AutoListen;
	import com.adobe.portfolio.components.supportClasses.IListenAndWatch;
	import com.adobe.portfolio.events.LayoutEvent;
	import com.adobe.portfolio.events.ThumbnailManagerEvent;
	import com.adobe.portfolio.layouts.supportClasses.CardLayoutHelper;
	import com.adobe.portfolio.layouts.supportClasses.ICardLayout;
	import com.adobe.portfolio.lists.CardList;
	import com.adobe.portfolio.managers.ThumbnailManager;
	import com.adobe.portfolio.utils.MathUtil;
	
	import flash.events.Event;
	import flash.events.TimerEvent;
	import flash.geom.Matrix3D;
	import flash.geom.PerspectiveProjection;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.geom.Vector3D;
	import flash.utils.Timer;
	
	import mx.core.ILayoutElement;
	import mx.core.IUIComponent;
	import mx.core.IVisualElement;
	import mx.core.UIComponent;
	import mx.events.EffectEvent;
	import mx.events.ResizeEvent;
	import mx.logging.ILogger;
	import mx.logging.Log;
	
	import spark.components.supportClasses.GroupBase;
	import spark.effects.Animate;
	import spark.effects.easing.EaseInOutBase;
	import spark.filters.BlurFilter;
	import spark.filters.ColorMatrixFilter;
	import spark.layouts.supportClasses.LayoutBase;
	
	/**
	 *  Dispatched after the layout has set its focus to a valid attachment.
	 *  @eventType com.adobe.portfolio.events.NavLayoutEvent.HANDLED_ATTACHMENT_FOCUS
	 */
	[Event(name="HANDLED_ATTACHMENT_FOCUS", type="com.adobe.portfolio.events.LayoutEvent")]
	
	/**
	 * The WaveLayout class extends LayoutBase to implement the layout of the Wave navigator.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class WaveLayout extends LayoutBase implements IListenAndWatch, ICardLayout, ICardProvider
	{
		private const logger:ILogger = Log.getLogger("WaveLayout");
		
		protected const DEG2RAD:Number = 2 * Math.PI / 360;
		protected var autoListen:AutoListen;
		protected var perspectiveProjection:PerspectiveProjection;
		protected var animate:Animate;
		protected var totalWidth:Number;
		protected var layoutWidth:Number;
		protected var focusCardInvalid:Boolean;
		protected var sizeInvalid:Boolean;
		protected var cardSizesInvalid:Boolean;
		protected var scrollAgain:Boolean;
		protected var cardPositionTimer:Timer;
		
		private var _itemDistance:Number; // percent
		private var _itemWidth:Number;
		private var _itemHeight:Number;
		private var _fieldOfView:Number; // degrees
		private var _yAngle:Number;
		private var _focusIndex:Number;
		private var _zDistance:Number;
		private var _zDistancePower:Number;
		private var _zAngle:Number;
		private var _sineAmp:Number;
		private var _sinePhase:Number; // degrees
		private var _easer:EaseInOutBase;
		private var _list:CardList;
		private var _thumbFitSize:Point;
		private var _cardLayoutHelper:CardLayoutHelper;
		private var _cardPosition:Number;
		private var _maxCardPosition:Number;
		private var _pendingCardPosition:Number;
		private var _focusAttachment:IAttachment;
		
		private static var blurFilters:Array;
		private static var colorFilters:Array;
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function WaveLayout()
		{
			super();
			initBlurFilters();
			initColorFilters();
			autoListen = new AutoListen();
			_fieldOfView = 25; 
			_zDistancePower = 1;
			_sineAmp = 1;
			_sinePhase = 30;
			_focusIndex = 0;
			_cardPosition = 0;
			_cardLayoutHelper = new CardLayoutHelper(this);
			addListenAndWatch();
			focusCardInvalid = true;
		}
		
		/**
		 *  Add event listeners and change watchers to the navigator.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function addListenAndWatch():void
		{
			autoListen.addListener(ThumbnailManager.instance, ThumbnailManagerEvent.ASPECT_RATIO_CHANGED, onAspectRatioChanged);
		}
		
		/**
		 *  Remove all event listeners and change watchers from the navigator
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function removeListenAndWatch():void
		{
			autoListen.removeAllListeners();
		}

		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function set target(value:GroupBase):void 
		{
			autoListen.removeListener(target);
			super.target = value;
			autoListen.addListener(target, ResizeEvent.RESIZE, onResize);
		}
		
		protected function onAspectRatioChanged(event:ThumbnailManagerEvent):void
		{
			cardSizesInvalid = true;
		}
		
		protected function onResize(event:ResizeEvent):void
		{
			invalidateSize();
		}
		
		protected function invalidateSize():void
		{
			sizeInvalid = true;
			cardSizesInvalid = true;
			if (target)
				target.invalidateSize();
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function measure():void
		{
			super.measure();
			commitProperties();
		}
		
		protected function commitProperties():void
		{
			if (cardSizesInvalid) {
				resetCardSizes();
			}
			if (sizeInvalid) {
				resetSize();
			}
			if (focusCardInvalid) {
				scroll();
				focusCardInvalid = false;
			}
		}
		
		/**
		 *  Resets the target's content size
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function resetSize():void {
			
			if (!target)
				return;
			
			var numElements:int = target.numElements;
			layoutWidth = target.width;
			totalWidth = layoutWidth + (numElements - 1) * cardDistance;
			updateMaxCardPosition();
			target.setContentSize(totalWidth, target.height);
			resetPerspective();
			scroll(0);
			sizeInvalid = false;
		}

		protected function resetPerspective():void 
		{
			if (!target)
				return;
			
			if (!perspectiveProjection) {
				perspectiveProjection = new PerspectiveProjection();
			}
			perspectiveProjection.fieldOfView = fieldOfView;
			perspectiveProjection.projectionCenter = getProjectionCenter();
			target.transform.perspectiveProjection = perspectiveProjection;
			invalidateTarget();
		}
		
		protected function getProjectionCenter():Point 
		{
			return target.localToGlobal(
				new Point (target.x + target.width/2, 
						   target.y + target.height/2));
		}

		protected function initBlurFilters():void 
		{
			if (!blurFilters) {
				blurFilters = [];
				for (var i:int=0; i<4; i++) {
					var blur:Number = Math.pow(2, i);
					// i.e. 1,2,4,8
					blurFilters[i] = new BlurFilter(blur, blur); 
				}
			}
		}
		
		protected function initColorFilters():void 
		{
			if (!colorFilters) {
				colorFilters = [];
				for (var i:int=0; i<256; i++) {
					var v:Number = 1 - i/255;
					// i.e. 1, 254/255, 253/255, etc.
					colorFilters[i] = new ColorMatrixFilter(
						[v,0,0,0,0,
							0,v,0,0,0,
							0,0,v,0,0,
							0,0,0,1,0]);
				}
			}
		}
		
		/**
		 *  Scroll the cards to the current focusIndex
		 * 
		 *  @param duration The duration in miliseconds for the scrolling.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function scroll(duration:Number = NaN):void
		{
			if (focusIndex == -1)
				return;
			
			var newCardPosition:Number = cardPositionFromIndex(focusIndex);
			
			if (newCardPosition == cardPosition)
				return;
			
			if (isNaN(duration))
				duration = 750 * Math.abs(newCardPosition - cardPosition) / cardDistance;
			if (duration > 5000)
				duration = 5000;
			else if (duration < 0 || isNaN(duration))
				duration = 0;
			
			enableCards(false);

			if (animate) {
				scrollAgain = true;
				return;
			}
			
			ThumbnailManager.instance.pauseQueue = true;
			animate = AnimateFactory.createAnimate(this, duration, "cardPosition", newCardPosition);
			if (easer)
				animate.easer = easer;
			animate.addEventListener(EffectEvent.EFFECT_END, onEndScroll);
			animate.play();
		}
		
		protected function enableCards(enabled:Boolean):void
		{
			if (!target)
				return;
			
			var numElements:int = target.numElements;
			for (var i:int = 0; i < numElements; i++) {
				
				var card:IUIComponent = target.getElementAt(i) as IUIComponent;
				if (card == null) {
					continue;
				}
				card.enabled = enabled;
			}
		}
		
		/**
		 *  Scroll the wave layout display to the card with the specified index.
		 * 
		 *  @param cardIndex The index of the card that is to be made the focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function scrollTo(cardIndex:int):void 
		{
			if (!list || !list.dataProvider)
				return;
			
			focusAttachment = list.dataProvider.getItemAt(cardIndex) as IAttachment;
			scroll();
		}
		
		protected function centerCard(newCardPosition:Number = NaN):void 
		{
			if (isNaN(cardDistance))
				return;
			
			if (isNaN(newCardPosition))
				newCardPosition = cardPosition;
			
			scrollTo(Math.round(newCardPosition / cardDistance));
		}
		
		protected function get cardDistance():Number 
		{
			return layoutWidth * itemDistance / 100.0;
		}
		
		/**
		 *  nudgeCard is called when the user drags a card either left or right.  In response 
		 *  the function scrolls one card to the left or right.
		 * 
		 *  @param delta The amount of the drag move.  This must be at least 20 pixels to trigger a scroll.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function nudgeCard(delta:Number):void 
		{
			var absDelta:Number = Math.abs(delta);
			
			if (absDelta > 20 && absDelta < cardDistance) {
				var newIndex:int = focusIndex + (delta > 0 ? 1 : -1);
				newIndex = MathUtil.clipToRange(newIndex, 0, target.numElements);
				scrollTo(newIndex);
			} else {
				centerCard();
			}
		}
		
		protected function onEndScroll(e:EffectEvent):void 
		{
			var numElements:int = target.numElements;
			for (var i:int = 0; i < numElements; i++) {
				var card:ISimpleCard = cardLayoutHelper.getLayoutElement(i) as ISimpleCard;
				if (card)
					card.doubleClickEnabled = (i == focusIndex);
			}
			animate.removeEventListener(EffectEvent.EFFECT_END, onEndScroll);
			animate = null;
			notifyCardPosition();
			enableCards(true);
			ThumbnailManager.instance.pauseQueue = false;
			
			if (scrollAgain) {
				scrollAgain = false;
				scroll(500);
			}
		}

		protected function notifyCardPosition():void
		{
			// delay before dispatching new position to lessen card slider jitter
			if (!cardPositionTimer) {
				cardPositionTimer = new Timer(300);
				cardPositionTimer.addEventListener(TimerEvent.TIMER,
					function (e:Event):void {
						cardPositionTimer.stop();
						dispatchEvent(new Event("cardPositionChanged"));
					}
				);
			}
			cardPositionTimer.reset();
			cardPositionTimer.start();
		}

		/**
		 *  @inheritDoc
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function elementAdded(index:int):void {
			super.elementAdded(index);
			cardLayoutHelper.pendingFade = true;
			cardSizesInvalid = true;
			sizeInvalid = true;
			focusCardInvalid = true;
		}
		
		/**
		 *  @inheritDoc
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function elementRemoved(index:int):void {
			super.elementRemoved(index);
			cardSizesInvalid = true;
			sizeInvalid = true;
			focusCardInvalid = true;
		}
		
		protected function updateMaxCardPosition():void
		{
			maxCardPosition = cardPositionFromIndex(target.numElements - 1);  
		}
		
		/**
		 *  Set the target's focus to be the specified attachment
		 * 
		 *  @param attachment The IAttachement element that is to be the new focus
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function setFocusAttachment(attachment:Object):void
		{
			dispatchEvent(new LayoutEvent(LayoutEvent.HANDLED_ATTACHMENT_FOCUS));
		}
		
		protected function resetCardSizes():void 
		{
			if (!cardSizesInvalid)
				return;
			
			if (!target)
				return;
			
			var width:Number = target.width;
			var height:Number = target.height;
		
			var fitSize:Number = 0.8 * Math.min(width,height);
			fitSize = Math.min(fitSize, 600);
			var numElements:int = target.numElements;
			if (numElements <= 0)
				return;
			
			cardSizesInvalid = false;
			for (var i:int = 0; i < numElements; i++) {
				
				var card:IVisualElement = target.getElementAt(i);
				if (card == null) {
					continue;
				}
				if (isNaN(itemWidth) || isNaN(itemHeight)) {
					// Let the card's setLayoutBoundsSize logic handle any aspect ratio adjustments
					card.setLayoutBoundsSize(fitSize, fitSize, false); 
				} else {
					card.setLayoutBoundsSize(itemWidth, itemHeight, false);
				}
			}
		}

		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		override public function updateDisplayList (width:Number, height:Number):void
		{
//			logger.debug("updateDisplayList: cardPosition= " + cardPosition);
			super.updateDisplayList(width, height);

			var numElements:int = target.numElements;
			
			for (var i:int = 0; i < numElements; i++) {

				var card:IVisualElement = target.getElementAt(i);
				if (card == null) {
					continue;
				}

				var matrix:Matrix3D = new Matrix3D();
				var position:Number = cardDistance * i - cardPosition;
				var indexedPosition:Number = position / cardDistance;
				var roundedIndexedPosition:Number = Math.round(indexedPosition);
				
				if (Math.abs(indexedPosition - roundedIndexedPosition) < 0.01)
					indexedPosition = roundedIndexedPosition;
				
//				logger.debug("i: " + i + " position: " + position + " indexedPosition: " + indexedPosition);
				
				setAlpha(card, indexedPosition);
				setFilters(card, indexedPosition);
				
				setZDistance(card, matrix, indexedPosition);
				rotateZ(card, matrix, indexedPosition);
				rotateY(card, matrix, indexedPosition);
				centerElement(card, matrix, width, height); 
				
				setXPosition(matrix, position);
				setYPosition(matrix, indexedPosition, card.height);
	
				card.setLayoutMatrix3D(matrix, false);
			}
			
			if (animate == null) {
				cardLayoutHelper.enableThumbnailGeneration();
				fadeInCards();
			}
		}
		
		/**
		 *  Runs the fadeInCards animation effect.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function fadeInCards():void
		{
			var count:int = target.numElements;
			if (count <= 0)
				return;
			cardLayoutHelper.fadeInCards(0, Math.min(100, 1000/count));
		}
		
		protected function setAlpha(card:IVisualElement, indexedPosition:Number):void 
		{
			var absIndexedPosition:Number = Math.abs(indexedPosition);
//			card.visible = absIndexedPosition < 2;

			if (cardLayoutHelper.isPendingOrFading())
				return;
			
//			card.alpha = (absIndexedPosition < 1) ? (1 - (0.25 * absIndexedPosition)):0.75; // slows the animation
			card.alpha = (absIndexedPosition < 0.25) ? 1 : 0.75;
		}
		
		protected function setFilters(layoutElement:ILayoutElement, indexedPosition:Number):void
		{
			var uiElement:UIComponent = layoutElement as UIComponent;
			var absIndexedPosition:Number = Math.abs(indexedPosition);
			
			var blurIndex:uint = Math.min(3, absIndexedPosition > 0.25 ? 1 : 0) as uint;
			var newBlurFilter:BlurFilter = blurFilters[blurIndex];
			var curBlurFilter:BlurFilter = uiElement.filters ? uiElement.filters[0] : null;
			
			var colorIndex:uint = Math.min(128, absIndexedPosition*32);
			var newColorFilter:ColorMatrixFilter = colorFilters[colorIndex];
			var curColorFilter:ColorMatrixFilter = uiElement.filters ? uiElement.filters[1] : null;
			
			if (!curBlurFilter || curBlurFilter.blurX != newBlurFilter.blurX
					|| !curColorFilter || curColorFilter.matrix[0] != newColorFilter.matrix[0]) {
				uiElement.filters = [newBlurFilter, newColorFilter];
			}
		}
		
		protected function setXPosition(matrix:Matrix3D, position:Number):void 
		{
			matrix.appendTranslation(position, 0, 0);
		}
		
		protected function setYPosition(matrix:Matrix3D, indexedPosition:Number, scale:Number):void 
		{
			if (indexedPosition == 0)
				return;
						
			var angle:Number = indexedPosition * sinePhase;
			var yDelta:Number = sineAmp * scale * Math.sin(DEG2RAD * angle);
			matrix.appendTranslation(0, yDelta, 0);
		}
		
		protected function rotateY(card:IVisualElement, matrix:Matrix3D, indexedPosition:Number):void 
		{
			var absIndexedPosition:Number = Math.abs(indexedPosition);

			if (indexedPosition == 0)
				return;
			
			var angle: Number = indexedPosition < 0 ? -yAngle : yAngle;
			if (absIndexedPosition < 1)
				angle *= absIndexedPosition;
			
			matrix.prependTranslation(card.width/2,card.height/2,0);
			matrix.prependRotation(angle, Vector3D.Y_AXIS);
			matrix.prependTranslation(-card.width/2,-card.height/2,0);
		}

		protected function rotateZ(card:IVisualElement, matrix:Matrix3D, indexedPosition:Number):void 
		{
			if (indexedPosition == 0)
				return;
			
			var angle:Number = indexedPosition * zAngle;
//			angle = clipToRange(angle, -60, 60);
			matrix.appendTranslation(-card.width/2,-card.height/2,0);
			matrix.appendRotation(angle, Vector3D.Z_AXIS);
			matrix.appendTranslation(card.width/2,card.height/2,0);
		}
		
		protected function setZDistance(card:IVisualElement, matrix:Matrix3D, indexedPosition:Number):void 
		{
			var distance:Number = Math.abs(indexedPosition) * zDistance;
			distance = Math.pow(distance, zDistancePower);
			card.depth = -distance;
			matrix.appendTranslation(0, 0, distance);
		}
		
		private static function centerElement(card:IVisualElement, matrix:Matrix3D, width:Number, height:Number):void 
		{
			matrix.appendTranslation(width / 2 - (card.width / 2),  
									  height / 2 - (card.height / 2), 0); 
		}
		
		/**
		 *  Returns the horizontal card position, in pixels, given
		 *  the cards index value.
		 * 
		 *  @param index The index value for the card
		 *  @return the horizontal position of the card, given in pixels
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function cardPositionFromIndex(index:int):Number
		{
			return index * ((totalWidth - layoutWidth) / (target.numElements - 1));
		}
		
		/**
		 *  Returns the index value for the card at the given horizontal position
		 *  the cards index value.
		 * 
		 *  @param cardPosition The horizontal position of the card, in pixels
		 *  @return The index value of the card at the given position
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function cardIndexFromPosition(cardPosition:Number):int
		{
			var index:int = Math.round(cardPosition / cardDistance);
			if (index < 0)
				index = 0;
			if (index >= target.numElements)
				index = target.numElements - 1;
			return index;
		}
		
		/**
		 *  The pendingCard position
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get pendingCardPosition():Number
		{
			return _pendingCardPosition;
		}
		

		public function set pendingCardPosition(value:Number):void 
		{
			if (_pendingCardPosition == value)
				return;
			
			_pendingCardPosition = value;
			
			if (isNaN(value))
				return;
			
			list.setSelectedIndexWithChangeDispatch(cardIndexFromPosition(pendingCardPosition));
		}
		
		/**
		 *  Get the card that has the specified card position
		 * 
		 *  @param cardPosition The horizontal position of the card, in pixels.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function getCard(cardPosition:Number):ISimpleCard
		{
			return cardLayoutHelper.getLayoutElement(cardIndexFromPosition(cardPosition)) as ISimpleCard;
		}
		
		[Bindable (name="cardPositionChanged")]
		/**
		 *  The horizontal position of the card, in pixels.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get cardPosition():Number
		{
			return _cardPosition;
		}
		
		public function set cardPosition(value:Number):void
		{
			if (_cardPosition == value)
				return;
			
			value = MathUtil.clipToRange(value, 0, maxCardPosition);
			
			_cardPosition = value;
			invalidateTarget();
			
			if (!animate) {
				dispatchEvent(new Event("cardPositionChanged"));
			}
		}
		
		[Bindable (name="maxCardPositionChanged")]
		/**
		 *  The maximum horizontal position that a card can have.
		 *  This is dependent on the target width.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get maxCardPosition():Number
		{
			return _maxCardPosition;
		}
		
		public function set maxCardPosition(value:Number):void
		{
			value = Math.floor(value);
			if (_maxCardPosition == value)
				return;
			
			_maxCardPosition = value;
			dispatchEvent(new Event("maxCardPositionChanged"));
		}
		
		/**
		 *  Invalidate the current targer
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function invalidateTarget():void
		{
			cardLayoutHelper.invalidateTarget();
		}

		/**
		 *  The index of the focus (centered) item
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function get focusIndex():Number
		{
			if (!list || !list.dataProvider)
				return -1;
			
			return list.dataProvider.getItemIndex(focusAttachment);
		}
				
		[Bindable]
		/**
		 *  The pre-transform distance between each visual item in the wave layout in pixels
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get itemDistance():Number
		{
			return _itemDistance;
		}

		public function set itemDistance(value:Number):void
		{
			_itemDistance = value;
			invalidateSize();
		}
		
		[Bindable]
		/**
		 *  Controls the y-axis rotation of cards other than the focus card
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get yAngle():Number
		{
			return _yAngle;
		}

		public function set yAngle(value:Number):void
		{
			_yAngle = value;
			invalidateTarget();
		}
		
		[Bindable]
		/**
		 *  Controls the z-distance of cards other than the focus card
		 * 
		 *  @see flash.geom.PerspectiveProjection
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get zDistance():Number
		{
			return _zDistance;
		}

		public function set zDistance(value:Number):void
		{
			_zDistance = value;
			invalidateTarget();
		}

		[Bindable]
		/**
		 *  The y-axis rotation of cards other than the focus card
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get zAngle():Number
		{
			return _zAngle;
		}
		
		public function set zAngle(value:Number):void
		{
			_zAngle = value;
			invalidateTarget();
		}

		[Bindable]
		/**
		 *  The amplitude of the sine wave used to place cards in the layout, given as a 
		 *  fraction of the target height (nominally, sineAmp is in the range 0 to 1). 
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get sineAmp():Number
		{
			return _sineAmp;
		}

		public function set sineAmp(value:Number):void
		{
			_sineAmp = value;
			invalidateTarget();
		}

		[Bindable]
		/**
		 *  The starting phase angle for the wave function
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get sinePhase():Number
		{
			return _sinePhase;
		}

		public function set sinePhase(value:Number):void
		{
			_sinePhase = value;
			invalidateTarget();
		}

		[Bindable]
		/**
		 *  An explicit width for cards in the layout.  Setting it to NaN provides automatic sizing.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get itemWidth():Number
		{
			return _itemWidth;
		}
		
		public function set itemWidth(value:Number):void
		{
			_itemWidth = value;
		}

		[Bindable]
		/**
		 *  An explicit height for cards in the layout.  Setting it to NaN provides automatic sizing.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get itemHeight():Number
		{
			return _itemHeight;
		}

		public function set itemHeight(value:Number):void
		{
			_itemHeight = value;
		}

		[Bindable]
		/**
		 *  The easer used for the card animation effect
		 * 
		 *  @see spark.effects.easing
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get easer():EaseInOutBase
		{
			return _easer;
		}
		/**
		 *  @private
		 */
		public function set easer(value:EaseInOutBase):void
		{
			_easer = value;
		}

		[Bindable]
		/**
		 *  The z-distance of cards is determined (in part) by zDistance^^zDistancePower
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get zDistancePower():Number
		{
			return _zDistancePower;
		}

		public function set zDistancePower(value:Number):void
		{
			_zDistancePower = value;
			invalidateTarget();
		}

		[Bindable]
		/**
		 *  The angle, specified in degrees greater than 0 and less than 180, for the field of view in three dimensions.
		 * 
		 *  @see flash.geom.PerspectiveProjection
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get fieldOfView():Number
		{
			return _fieldOfView;
		}

		public function set fieldOfView(value:Number):void
		{
			_fieldOfView = value;
			resetPerspective();
		}

		[Bindable]
		/**
		 *  The list of attachments being displayed (as cards)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get list():CardList
		{
			return _list;
		}

		public function set list(value:CardList):void
		{
			_list = value;
		}

		/**
		 *  The instance of the CardLayoutHelper used by this layout
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get cardLayoutHelper():CardLayoutHelper
		{
			return _cardLayoutHelper;
		}

		/**
		 *  Find the list of cards that overlap the specified marquee rectangle. 
		 * 
		 *  @param marqueeRect A bounding rectangle
		 *  @return A vector of integers that represent the indexes
		 *  of the cards in the layout that intersect the specified marquee rectangle.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function findCardsThatIntersect(marqueeRect:Rectangle):Vector.<int>
		{
			return null; // only necessary for marquee support
		}

		/**
		 *  The size to use for preview thumbnails
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get thumbFitSize():Point
		{
			return _thumbFitSize;
		}

		public function set thumbFitSize(value:Point):void
		{
			_thumbFitSize = value;
		}

		[Bindable]
		public function get focusAttachment():IAttachment
		{
			return _focusAttachment;
		}

		public function set focusAttachment(value:IAttachment):void
		{
			logger.debug("set focusAttachment: " + value);
			_focusAttachment = value;
			scroll();
		}

	}
}
