/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.layouts.supportClasses
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.components.supportClasses.AutoListen;
	import com.adobe.portfolio.managers.PropertyManager;
	import com.adobe.portfolio.properties.CommonProperties;
	
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.utils.Dictionary;
	
	import mx.collections.IList;
	import mx.events.CollectionEvent;
	import mx.events.CollectionEventKind;
	import mx.logging.ILogger;
	import mx.logging.Log;

	/**
	 * 
	 * The CardDistributor class implements an algorithm for placing Cards pseudo-randomly in the layout.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class CardDistributor
	{
		private var _cardAngle:Number;
		
		protected var cardsChangedMap:Dictionary = new Dictionary(true);
		protected var autoListen:AutoListen;
		protected var autoPositionCardSize:Number;
		protected var autoPositionCardSizeDiameter:Number;
		protected var gridPoints:Array;
		protected var cardProperties:ICardProperties;

		private static const logger:ILogger = Log.getLogger("CardDistributor");
			
		// Cards were added or removed
		private var cardsChanged:Boolean = true;

		
		/**
		 *  Constructor 
		 * 
		 *  @param cardProperties An interface that allows for setting/getting card location and angle properties.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function CardDistributor(cardProperties:ICardProperties)
		{
			this.cardProperties = cardProperties;
			autoListen = new AutoListen();
			initGrid(9);
		}
		
		protected function onCollectionChange(event:CollectionEvent):void
		{
			switch (event.kind) {
				case CollectionEventKind.ADD:
				case CollectionEventKind.REMOVE:
				case CollectionEventKind.RESET:
					cardsChangedMap[event.currentTarget] = true;
					break;
			}
		}
		
		protected function initAutoPositionCardSize():Boolean {
			if (!isNaN(autoPositionCardSize))
				return true;
			
			autoPositionCardSize = getCardSize(CommonProperties.PROPERTY_CARD_SIZE_MEDIUM);
			if (isNaN(autoPositionCardSize))
				return false;
			autoPositionCardSizeDiameter = autoPositionCardSize * Math.SQRT2;
			
			return true;
		}
		
		/**
		 *  Verifies that all card positions are valid.
		 *  This method should be called whenever there is a change in the
		 *  collection or the target size.
		 * 
		 *  @param collection A collection (list) of attachments.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function ensureAllPositionsAreValid(collection:IList):void
		{
			if ((collection == null) || !cardProperties.targetSize) {
				return;
			}
			if (cardsChangedMap[collection] == null) {
				autoListen.addListener(collection, CollectionEvent.COLLECTION_CHANGE, onCollectionChange);
				cardsChangedMap[collection] = true;
			} else if (!(cardsChangedMap[collection] as Boolean)) {
				return;
			}
			
			if (!initAutoPositionCardSize())
				return;
			
			
			// Give the unassigned positions some random values
			createRandomPositions(collection);
			
			// Done
			cardsChangedMap[collection] = false;
		}
		
		/**
		 *  This method goes through all the cards in the unsetPositions array it is passed,
		 *  gets their position, then adds them to the setPositions array and removes 
		 *  them from the unsetPositions array.
		 * 
		 *  @param collection A collection of attachments
		 *  @param setPositions an Array that is filled with the attachments that have their position set
		 *  @param unsetPositions an Array that is filled with the attachments that have NOT had their position set
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function getCardPositions(collection:IList, setPositions:Array, unsetPositions:Array):void
		{
			if (!collection || collection.length <= 0)
				return;
			
			if (!setPositions || !unsetPositions) {
				CONFIG::logging{
					logger.warn("getCardPositions() expected non-null arguments");
				}
				return;
			}
			
			setPositions.splice(0, setPositions.length);
			unsetPositions.splice(0, unsetPositions.length);
			
			// See if any of the attachments has unset x/y position fields
			var collectionLength:int = collection.length;
			for (var index:int = 0; index < collectionLength; ++index) {
				var attachment:IAttachment = collection.getItemAt(index) as IAttachment;
				if (!attachment)
					continue;
				if (cardProperties.isLocationSet(attachment)) {
					setPositions.push(attachment);
				} else {
					unsetPositions.push(attachment);
				}
			}
		}
		
		/**
		 *  @private
		 * 
		 */
		private function createRandomPositions(collection:IList):void
		{
			// This algorithm places cards based on a Poisson-disk distribution and is based in part
			// on concepts presented here:  http://www.cs.virginia.edu/~gfx/pubs/antimony/
			//
			var targetSize:Point = cardProperties.targetSize;
			if (!targetSize)
				return;
			
			var diameter:Number = autoPositionCardSizeDiameter;
			var border:Number =  getCardSize(CommonProperties.PROPERTY_CARD_SIZE_MEDIUM) * 1.1 / 2;
			var sampleRect:Rectangle = new Rectangle(border, border, targetSize.x - 2 * border, targetSize.y - 2 * border);
			
			var setPositions:Array = new Array();
			var unsetPositions:Array = new Array();
			getCardPositions(collection, setPositions, unsetPositions);
			var seedPositions:Array = setPositions.slice();
			
			var p:Point;
			for each (var attachment:IAttachment in unsetPositions) {
				if (!attachment)
					continue;
				
				while (true) {
					p = getPoissonPoint(diameter, sampleRect, attachment, seedPositions, setPositions);
					if (p)
						break;
					// In the case where it's not possible to satisfy the constraint, the diameter is reduced 
					// via a geometric sequence (dividing by 1.2) until the constraint can be satisfied.
					diameter /= 1.2;
				}
				
				cardProperties.setLocation(attachment, p);
				cardProperties.setRotation(attachment, (2 * Math.random() - 1) * cardAngle);
			}
			sortXCoordinateByIndexOrder(unsetPositions);
		}
		
		/**
		 *  @private
		 * 
		 */
		private function sortXCoordinateByIndexOrder(unsetPositions:Array):void
		{
			var points:Array = [];
			for each (var attachment:IAttachment in unsetPositions) {
				if (!attachment)
					continue;
				points.push(cardProperties.getLocation(attachment));
			}
			var newSort:Array = points.sort(
				function (a:Point, b:Point):Number {
					return a.length - b.length;
				}
			);
			for each (attachment in unsetPositions) {
				if (!attachment)
					continue;
				var p:Point = newSort.shift();
				cardProperties.setLocation(attachment, p);
				// don't worry about the swapping the rotation
			}
		}
		
		/**
		 *  @private
		 * 
		 *  The getPoissonPoint() function attempts to generate sample points with a Poisson-disk distribution.  It starts with an initial seed point (currently
		 *  at 0.5, 0.5, then aggregrates subsequent sample points near the existing sample points (contained in setPositions).
		 *  
		 *  @param diameter the constraint diameter - sample points should ideally be at least diameter distance away from other sample points (i.e. a Poisson-disk distribution)
		 *  @param sampleRect returned points must lie within this rectangle
		 *  @param attachment the attachment that needs a location
		 *  @param seedPositions an Array containing attachments that are candidates for seeds (other attachments may fit nearby)
		 *  @param setPositions an Array containing attachments that have already been set
		 * 
		 *  @return a Point that (if possible) satisfies a Poisson disk distribution at the specified diameter. 
		 * 
		 *  The seedPositions and setPositions arrays are updated by getPoissonPoint() and are ready to be used for
		 *  the next call to getPoissonPoint(). 
		 */		
		private function getPoissonPoint(diameter:Number, sampleRect:Rectangle, 
										 attachment:IAttachment,
										 seedPositions:Array,
										 setPositions:Array):Point 
		{
			var setPositionsLen:int = setPositions.length;
			
//			logger.debug("getPoissonPoint: attachment= " + attachment.fileName);
//			logger.debug("getPoissonPoint: diameter= " + diameter);
			
			if (setPositionsLen == 0) {
				seedPositions.push(attachment);
				setPositions.push(attachment);
				return getScaledPoint(new Point(0.5, 0.5), sampleRect);
			}
			
			diameter = Math.floor(diameter);
			var diameter2:Number = diameter * 2;
			
			// loop through the attachments that already have their positions set
			while (seedPositions.length > 0) {
				var subSampleRect:Rectangle = null;
				var attachmentSeed:IAttachment = seedPositions.shift();
				if (attachmentSeed == attachment)
					continue;
				
//			 	logger.debug("getPoissonPoint: attachmentSeed= " + attachmentSeed.fileName);
				var center:Point = cardProperties.getLocation(attachmentSeed);
				if (center) {
					// logger.debug("getPoissonPoint: attachmentSeed center= " + center);
					
					// initialize a sample rect centered about the existing attachment point having a size of diameter * 2
					//	any available sample points near attachmentSeed must lie within this rect
					subSampleRect = new Rectangle(center.x, center.y, diameter2, diameter2);
					subSampleRect.offset(-diameter, -diameter);
				}
				
				var gridLen:int = gridPoints.length;
				// choose a random starting location within the grid
				var start:int = (gridLen-1) * Math.random();
				
				// visit all of the points in the grid
				for (var j:int = 0; j < gridLen; ++j) {
					var jj:int = (j + start) % gridLen;
					var p:Point = getGridPoint(subSampleRect, jj);
					
					// Eliminate gridPoints that are < diameter away or > 2 * diameters away
					var pDistance:Number = p.subtract(center).length;
					if (pDistance < diameter || pDistance > diameter2)
						continue;
					
					if (!sampleRect.containsPoint(p))
						continue;
					
					var checkLen:int = Math.min(100, setPositionsLen); // check no more than 100 attachments
					for (var k:int = 0; k < checkLen; k++) {// compare to existing attachments
						var a:IAttachment = setPositions[k];
						var q:Point = cardProperties.getLocation(a);
						if (!q)
							continue;
						
						var distance:Number = Math.ceil(q.subtract(p).length);
						if (distance < diameter) { // too close
//							logger.debug("getPoissonPoint: too close: a= " + a.fileName + " p= " + p + " q= " + q + " distance= " + distance);
							p = null;
							break;
						} else { // far enough
//							logger.debug("getPoissonPoint: far enough: a= " + a.fileName + " p= " + p + " q= " + q + " distance= " + distance);
						}
					}
					if (p) { // found a point that satisfies the constraint
//						logger.debug("getPoissonPoint: found point: attachment= " + attachment.fileName + "p= " + p);
						if (attachmentSeed)
							seedPositions.unshift(attachmentSeed);
						seedPositions.push(attachment);
						setPositions.push(attachment);
						return p;
					}
				}
			}
			// couldn't satisfy constraint
//			logger.debug("Poisson over-constrained: " + attachment.fileName + " diameter= " + diameter);

			// reinitialize the seedPositions dictionary
			arrayCopy(seedPositions, setPositions);
			
			return null;
		}
		
		/**
		 *  @private
		 * 
		 *  Replace the contents of the destination array with a copy of the contents of the source 
		 *  @param dst The destination array
		 *  @param src The source array
		 * 
		 */		
		private static function arrayCopy(dst:Array, src:Array):void 
		{
			dst.splice(0, dst.length);
			for (var i:int = 0; i < src.length; ++i) {
				dst.push(src[i]);
			}
		}

		/**
		 *  @private
		 * 
		 */
		private function getGridPoint(sampleRect:Rectangle, index:int = -1):Point 
		{
			if (index == -1)
				index = Math.floor(Math.random()*(gridPoints.length-1));
			var p:Point = gridPoints[index];
			return getScaledPoint(p, sampleRect);
		}
		
		/**
		 *  @private
		 * 
		 */
		private function getScaledPoint(p:Point, sampleRect:Rectangle):Point {
			
			if (!p || !sampleRect)
				return null;
			
			// jitter by up to 10%
			p.x += Math.random() * 0.1 - 0.05;
			p.y += Math.random() * 0.1 - 0.05;
			
			return new Point(
				sampleRect.x + (sampleRect.width * p.x),
				sampleRect.y + (sampleRect.height * p.y));
		}
		
		/**
		 *  @private
		 * 
		 */
		private function initGrid(gridSize:Number):void {
			if (gridPoints)
				return;
			gridPoints = [];
			gridSize = gridSize - 1;
			for (var i:int = 0; i <= gridSize; i++) {
				for (var j:int = 0; j <= gridSize; j++) {
					gridPoints.push(new Point(j/gridSize, i/gridSize));
				}
			}
		}
		
		/**
		 *  Returns the size, in pixels of the specified card size name.
		 * 
		 *  @param sizeName The name of the card size: one of: PROPERTY_CARD_SIZE_SMALL, 
		 *  PROPERTY_CARD_SIZE_MEDIUM, or PROPERTY_CARD_SIZE_LARGE
		 * 
		 *  @see com.adobe.portfolio.properties.CommonProperties#PROPERTY_CARD_SIZE_SMALL
		 *  @see com.adobe.portfolio.properties.CommonProperties#PROPERTY_CARD_SIZE_MEDIUM
		 *  @see com.adobe.portfolio.properties.CommonProperties#PROPERTY_CARD_SIZE_LARGE
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected static function getCardSize(sizeName:String):Number 
		{
			var commonProperties:CommonProperties = PropertyManager.instance.commonProperties;

			return commonProperties ? commonProperties.getCardSize(sizeName) : NaN;
		}
		
		/**
		 *  The maximum angle at which new cards are rotated (if applicable)
		 * 
		 *  @param value The angle, expressed in degrees
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function set cardAngle(value:Number):void
		{
			_cardAngle = value;
		}
		/**
		 *  @private
		 */
		public function get cardAngle():Number
		{
			return _cardAngle;
		}

	}
}
