/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.layouts.supportClasses
{
	import com.adobe.portfolio.animations.SequentialFade;
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.cards.ICardProvider;
	import com.adobe.portfolio.cards.ISimpleCard;
	import com.adobe.portfolio.components.supportClasses.AutoListen;
	import com.adobe.portfolio.lists.CardList;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.ThumbnailManager;
	import com.adobe.portfolio.utils.MathUtil;
	
	import flash.display.Bitmap;
	import flash.events.EventDispatcher;
	import flash.events.TimerEvent;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.utils.Dictionary;
	import flash.utils.Timer;
	
	import mx.collections.IList;
	import mx.core.IVisualElement;
	import mx.logging.ILogger;
	import mx.logging.Log;
	
	import spark.components.DataGroup;
	import spark.components.supportClasses.GroupBase;
	
	
	/**
	 * 
	 * The CardLayoutHelper class is a set of utility functions to help Card-based layouts.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class CardLayoutHelper extends EventDispatcher
	{
		protected static const logger:ILogger = Log.getLogger("CardLayoutHelper");
		protected static const collectionMgr:CollectionManager = CollectionManager.instance;
		
		protected var cardLayout:ICardLayout;
		protected var cardProvider:ICardProvider;
		protected var idleItems:Array;
		protected var idleTimer:Timer;
		protected var thumbnailMgr:ThumbnailManager;
		protected var autoListen:AutoListen;
		protected var idleDelay:int = 2000;
		protected var attachmentProperties:Dictionary;
		protected var sequentialFade:SequentialFade;
		protected var enableDynamicThumbnailsTimer:Timer;
		
		private var _pendingFade:Boolean;
		private var _enableIdle:Boolean;

		/**
		 *  Constructor.
		 * 
		 *  @param An ICardLayout to help
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function CardLayoutHelper(cardLayout:ICardLayout)
		{
			this.cardLayout = cardLayout;
			cardProvider = cardLayout as ICardProvider;
			attachmentProperties = new Dictionary();
			sequentialFade = new SequentialFade();
			sequentialFade.onElementDone = onElementDone;
			thumbnailMgr = ThumbnailManager.instance;
			idleTimer = new Timer(idleDelay);
			autoListen = new AutoListen();
			autoListen.addListener(idleTimer, TimerEvent.TIMER, onIdleTime);
			idleItems = new Array();
			_pendingFade = false;
			_enableIdle = true;
		}
		
		/**
		 *  Iterates the cards in a layout and enables thumbnail generation.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function enableThumbnailGeneration():void 
		{
			if (!target)
				return;
			
			var numElements:int = target.numElements;
			if (numElements == 0)
				return;
			
			var scrollRect:Rectangle = scrollRect(target);
			for (var i:int = 0; i < numElements; ++i ) {
				var card:ISimpleCard = getCard(i);
				if (!card)
					continue;
				if (card.enableThumbnailGeneration)
					continue;
				var elementIsVisible:Boolean =  elementIntersectsRect(card, scrollRect);
//				CONFIG::logging {
//					logger.debug(cardLayout + " intersects: " + i + ": " + elementIsVisible + ": " + card.fileName );
//				}			
				card.enableThumbnailGeneration = elementIsVisible;
			}
			enableDynamicThumbnails();
		}
		
		protected function enableDynamicThumbnails():void 
		{
			if (thumbnailMgr.enableDynamicThumbnails || enableDynamicThumbnailsTimer)
				return;
			
			enableDynamicThumbnailsTimer = new Timer(1000);
			enableDynamicThumbnailsTimer.addEventListener(TimerEvent.TIMER,
				function (e:TimerEvent):void {
					enableDynamicThumbnailsTimer.removeEventListener(TimerEvent.TIMER, arguments.callee);
					thumbnailMgr.enableDynamicThumbnails = true;
					enableDynamicThumbnailsTimer = null;
				}
			);
			enableDynamicThumbnailsTimer.start();
		}
		
		/**
		 *  Returns a scroll rect for the group.
		 * 
		 *  @param group The group
		 *  @return A Rectangle corresponding to the scrollRect
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function scrollRect(group:GroupBase):Rectangle 
		{
			var scrollRect:Rectangle = group.scrollRect;
			if (!scrollRect || scrollRect.width == 0 || scrollRect.height == 0) {
				scrollRect = new Rectangle(0,0, group.width, group.height);
			}
			return scrollRect;			
		}
		
		/**
		 *  Returns true if an element intersects the specified rectangle
		 * 
		 *  @param element The element
		 *  @param rect The rectangle
		 *  @return True if the element intersects the specified rectangle
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function elementIntersectsRect(element:IVisualElement, rect:Rectangle):Boolean
		{
			var elementRect:Rectangle = getLayoutRect(element);
//			CONFIG::logging {
//				var card:ISimpleCard = element as ISimpleCard;
//				if (card)
//					logger.debug(cardLayout + " elementRect: " + elementRect + " : " + card.fileName );
//			}
			return (elementRect != null) && elementRect.intersects(rect);
		}
		
		/**
		 *  Returns the bounding box of a layout element
		 * 
		 *  @param element The element
		 *  @return The bounding box of the element 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getLayoutRect(element:IVisualElement):Rectangle
		{
			if (element == null) {
				return null;
			}
				
			return MathUtil.projectedBounds(new Rectangle(0, 0, element.width, element.height), 
					element.getLayoutMatrix3D(), 
					target.transform.perspectiveProjection);
		}
		
		/**
		 *  Returns a vector of indices of cards that intersect a given rectangle
		 * 
		 *  @param rect The rectangle
		 *  @return A vector of indices
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function findCardsThatIntersect(rect:Rectangle):Vector.<int>
		{
			var intersectingCards: Vector.<int> = new Vector.<int>();
			var target:GroupBase = cardLayout.target;
			var numElements:int = target.numElements;
			for (var i:int = 0; i < numElements; i++ ) {
				var element:IVisualElement = getLayoutElement(i);
				if (elementIntersectsRect(element, rect)) {
					intersectingCards.push(i);
				}
			}
			
			return intersectingCards;
		}
		
		/**
		 *  Fades in cards in a layout
		 * 
		 *  @param delay The delay in msec before starting the fade
		 *  @param duration The duration of the fade for each card in msec
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function fadeInCards(delay:Number, duration:Number = 50):void 
		{
			if (sequentialFade.fading) {
				return;
			}
			if (!pendingFade && !cardLayout.useVirtualLayout) {
				// We only return here when cardLayout.useVirtualLayout is false, since we know
				// all our renderes existed and were faded in.
				return;
			}
			
			sequentialFade.fadeInElements(target, duration, delay, 
				function ():void {
					fadeInCards(0);
				}
			);
			pendingFade = false;
		}
		
		protected function onElementDone(target:Object):void 
		{
			saveAlpha(target as ISimpleCard);
		}
		
		/**
		 *  Responds to a layout resize by resetting the ThumbnailManager thumb size and idling thumbs around the focus card 
		 * 
		 *  @param dataProvider The dataProvider of the layout (a list of attachments)
		 *  @param width The thumb fit width
		 *  @param height The thumb fit height
		 *  @param delta Specifies how many cards from the focus card to idle (i.e. 1 idles focusCard-1, focusCard, focusCard+1, etc.)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function resizeCachedThumbs(dataProvider:IList, width:Number, height:Number, delta:int):void 
		{
			if (!dataProvider)
				return;
			
			var thumbSize:Point = new Point(width, height);
			if (thumbFitSize && thumbSize.x == thumbFitSize.x && thumbSize.y == thumbFitSize.y) 
				return;
			
			var idleItems:Boolean = thumbnailMgr.cacheThumbFitSize.x < thumbSize.x || thumbnailMgr.cacheThumbFitSize.y < thumbSize.y;
			if (idleItems)
				thumbnailMgr.cacheThumbFitSize = thumbSize;
			
			thumbFitSize = thumbnailMgr.cacheThumbFitSize;
			
			if (idleItems) {
				idleCardsAroundFocus(dataProvider, collectionMgr.focusAttachment, delta);
			}
		}
		
		/**
		 *  Lazily rebuilds thumbnails in a region around a specified focus after a resize of the layout
		 * 
		 *  @param dataProvider The dataProvider of the layout (a list of attachments)
		 *  @param delta Specifies how many cards from the focusAttachment to idle (i.e. 1 idles focusAttachment-1, focusAttachment, focusAttachment+1, etc.)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function idleCardsAroundFocus(dataProvider:IList, focusAttachment:IAttachment, delta:int):void
		{
			if (!enableIdle)
				return;
			if (!dataProvider)
				return;
			
			var focusIndex:int = dataProvider.getItemIndex(focusAttachment);
			if (focusIndex == -1) {
				target.callLater(idleCardsAroundFocus, [dataProvider, focusAttachment, delta]);
				return;
			}
			
			resetIdleItems();
			
			for (var i:int=focusIndex-delta; i <= focusIndex+delta; i++) {
				if (i < 0)
					continue;
				if (i >= dataProvider.length)
					continue;
				handleItemIdle(dataProvider.getItemAt(i) as IAttachment);
			}
		}
		
		protected function handleItemIdle(attachment:IAttachment):void 
		{
			if (!enableIdle)
				return;
			
			if (!attachment)
				return;
			
			idleTimer.reset();
//			CONFIG::logging {
//				logger.debug("handleItemIdle: " + cardLayout + attachment.fileName + " thumbFitSize= " + thumbFitSize);
//			}
			idleItems.push(attachment);
			startIdleTimer(idleDelay);
		}
		
		private function startIdleTimer(delay:int):void {
			if (!enableIdle)
				return;
			idleTimer.delay = delay;
			idleTimer.start();
		}
		
		protected function resetIdleItems():void 
		{
			idleTimer.reset();
			idleItems = new Array();
		}
		
		protected function onIdleTime(e:TimerEvent):void
		{
			idleTimer.reset();
			if (idleItems.length == 0)
				return;
			
			var attachment:IAttachment = idleItems.shift();
			if (!attachment)
				return;
			
			if (!thumbFitSize) {
				thumbFitSize = thumbnailMgr.cacheThumbFitSize;
			}
			var fitSize:Point = thumbFitSize;
			
			if (!fitSize) {
				CONFIG::logging {
					logger.warn("expected valid fitSize");
				}
				return;
			}
			
			if (thumbnailMgr.isThumbnailCached(attachment, fitSize.x, fitSize.y, 1)) {
				startIdleTimer(10);
				return;
			}
			
			preloadThumb(attachment, fitSize);
		}
		
		protected function preloadThumb(attachment:IAttachment, fitSize:Point):void 
		{
			CONFIG::logging {
				logger.debug("preloadThumb: " + cardLayout + attachment.fileName + " fitSize= " + fitSize);
			}
			const downSampleOK:Boolean = false;
			const upSampleOK:Boolean = true;
			const queuePriority:int = ThumbnailManager.IDLE_THUMB_PRIORITY;
			
			thumbnailMgr.requestThumbnail("preloadThumb_" + attachment.fileName, attachment, fitSize.x, fitSize.y, 1, 
				function (id:*, width:Number, height:Number, pageNumber:uint, thumbnail:Bitmap):void {
					CONFIG::logging {
						logger.debug("idle loaded: " + cardLayout + attachment.fileName + " fitSize= " + fitSize);
					}
					var card:ISimpleCard = CardList.cardFromAttachment(cardLayout.target as DataGroup, attachment);
					if (card) {
						card.enableThumbnailGeneration = true;
						card.invalidatePreview(false, true);
					}
					if (idleItems.length > 0) {
						startIdleTimer(idleDelay);
					}
				},
				downSampleOK, upSampleOK, queuePriority
			);
		}

		
		/**
		 *  True if a fade (of cards) is pending
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get pendingFade():Boolean
		{
			return _pendingFade;
		}

		public function set pendingFade(value:Boolean):void
		{
			_pendingFade = value;
		}
		
		/**
		 *  True if a fade (of cards) is pending or the fade is active
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function isPendingOrFading():Boolean
		{
			return pendingFade || sequentialFade.fading;
		}

		protected function get thumbFitSize():Point
		{
			return cardLayout.thumbFitSize;
		}
		
		protected function set thumbFitSize(value:Point):void
		{
			if (!cardLayout)
				return;
			
			cardLayout.thumbFitSize = value;
		}

		/**
		 *  True if thumbnail idle is enabled
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get enableIdle():Boolean
		{
			return _enableIdle;
		}

		public function set enableIdle(value:Boolean):void
		{
			_enableIdle = value;
		}
		
		protected function saveAlpha(card:ISimpleCard):void 
		{
			if (!card)
				return;
			
			attachmentProperties[card.data] = {alpha:card.alpha};
			// TODO need to delete dictionary entries when cards are deleted
		}
		
		/**
		 *  Returns the alpha value of a card
		 * 
		 *  @param card The card
		 *  @return The alpha value
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getAlpha(card:ISimpleCard):Number 
		{
			if (!card)
				return 1;
			
			var properties:Object = attachmentProperties[card.data];
			return properties ? properties.alpha : card.alpha;
		}
		
		protected function get target():GroupBase
		{
			return cardLayout.target;
		}
		
		/**
		 *  Invalidate the displaylist of the target of the associated layout
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function invalidateTarget():void
		{
			if (!target)
				return;
			
			target.invalidateDisplayList();
		}
		
		/**
		 *  Returns the element in the target at the specified index
		 * 
		 *  @param index The index
		 *  @return The element (of the layout)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getLayoutElement(index:int):IVisualElement
		{
			// Note: This should never call getVirtualElementAt.
			// getVirtualElementAt should only be called from updateDisplayList
			return target.getElementAt(index);
		}
		
		/**
		 *  Returns the card in the target at the specified index
		 * 
		 *  @param index The index
		 *  @return The card (of the layout)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getCard(index:int):ISimpleCard
		{
			return getLayoutElement(index) as ISimpleCard;
		}
		
		/**
		 *  Returns the layout target's dataProvider
		 * 
		 *  @return The dataProvider of the layout target
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getDataProvider():IList
		{
			if ((target != null) && target.hasOwnProperty("dataProvider")) {
				return target["dataProvider"] as IList;
			}
			return null;
		}
		
		/**
		 *  Returns the index in the specified dataProvider of the given attachement
		 * 
		 *  @param The attachment
		 *  @param The dataProvider (if null it uses the dataProvider associated with the layout)
		 *  @return The index of the attachment in the dataProvider.  Returns -1 if not found.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getCurrentIndex(attachment:IAttachment, dataProvider:IList = null):int
		{
			if (attachment == null) {
				return -1;
			}
			if (dataProvider == null) {
				dataProvider = getDataProvider();
			}
			if (dataProvider == null) {
				return -1;
			}
			
			return dataProvider.getItemIndex(attachment);
		}
		
		/**
		 *  Returns true if the dataProvider is "current".
		 * 
		 *  @return True if the dataProvider is "current".
		 * 
		 *  @see com.adobe.portfolio.managers.CollectionManager#itemsAreCurrent
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function dataProviderIsCurrent():Boolean
		{
			var dataProvider:IList = getDataProvider();
			return dataProvider ? collectionMgr.itemsAreCurrent(dataProvider) : true;
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function toString():String
		{
			return "[CardLayoutHelper]";
		}
	}
}