/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.lists
{
	import com.adobe.portfolio.cards.CardFactory;
	import com.adobe.portfolio.cards.IBasicCard;
	import com.adobe.portfolio.cards.ISimpleCard;
	import com.adobe.portfolio.components.MarqueeSelection;
	import com.adobe.portfolio.components.NavPreview;
	import com.adobe.portfolio.components.supportClasses.IListenAndWatch;
	import com.adobe.portfolio.events.LayoutEvent;
	import com.adobe.portfolio.events.MarqueeSelectionEvent;
	import com.adobe.portfolio.managers.PropertyManager;
	
	import flash.events.IEventDispatcher;
	import flash.events.MouseEvent;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	
	import mx.binding.utils.BindingUtils;
	import mx.logging.ILogger;
	import mx.logging.Log;
	import mx.states.State;
	
	import spark.events.RendererExistenceEvent;
	
	
	[SkinState("intro")]
	[SkinState("normal")]
	[SkinState("disabled")]

	/**
	 * The CardContainer class extends CardList to provide marquee support and other utility functions 
	 * useful for Navigators that display cards of different sizes.
	 * 
  	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class CardContainer extends CardList implements IListenAndWatch
	{
		/**
		 *  The "intro" state of both the CardContainer and its skin.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const STATE_INTRO:String = "intro";

		/**
		 *  The "normal" state of both the CardContainer and its skin.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const STATE_NORMAL:String = "normal";

		/**
		 *  The "disabled" state of both the CardContainer and its skin.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const STATE_DISABLED:String = "disabled";
		
		CONFIG::logging {
			private static const logger:ILogger = Log.getLogger("CardContainer");
		}
		
		/**
		 *  A MarqueeSelection (with frame and shaded center) used to highlight the area of a marquee select.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		[SkinPart(required="false")]
		public var marqueeSelection:MarqueeSelection;

		protected var cardSize:String;
		
		private var _navPreview:NavPreview;
		private var marqueeMove:Boolean = false;
		private var baseSelection:Vector.<int>;
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function CardContainer()
		{
			initStates();

			// Set default item renderer; this can be overriden in mxml or AS
			const alpha:Number = 0;
			itemRenderer = new CardFactory(CardFactory.BASIC_CARD, cardSize, null, null, alpha);
			useDragDropReordering = true;
			useVirtualLayout = false; // until we get card flashing sorted out...
		}
		
		// override as necessary
		protected function initStates():void {
			// Setup the states for this instance. They should match the skin states.
			states = [
				new State({name: STATE_INTRO}), 
				new State({name: STATE_NORMAL}), 
				new State({name: STATE_DISABLED})
			];
			currentState = STATE_INTRO;
		}
		
		/**
		 *  Adds listeners and watchers that are automatically unwatched and unlistened.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		override public function addListenAndWatch():void {
			super.addListenAndWatch();
			addWatcher(BindingUtils.bindSetter(cardSizeChanged, PropertyManager.instance, 
				["commonProperties", "cardSize", "name"]));
			
			addWatcher(BindingUtils.bindSetter(onNavPreviewCurrentState, this, ["navPreview", "currentState"]));
		}
		
		override protected function partAdded(partName:String, instance:Object):void
		{
			super.partAdded(partName, instance);
			switch (instance) {
				case dataGroup:
					addListener(dataGroup.layout, LayoutEvent.HANDLED_ATTACHMENT_FOCUS, onHandledFocus);
					CONFIG::logging {
						addListener(dataGroup, RendererExistenceEvent.RENDERER_ADD, onRendererAdd);
					}
					break;
				
				case marqueeSelection:
					addListener(marqueeSelection, MarqueeSelectionEvent.BEGIN, onMarqueeSelectionBegin);
					addListener(marqueeSelection, MarqueeSelectionEvent.MOVE, onMarqueeSelectionMove);
					addListener(marqueeSelection, MarqueeSelectionEvent.END, onMarqueeSelectionEnd);
					addListener(marqueeSelection, MarqueeSelectionEvent.CANCELLED, onMarqueeSelectionCancelled);
					break;
				
				default:
					break;
			}
		}
		
		override protected function partRemoved(partName:String, instance:Object):void
		{
			super.partRemoved(partName, instance);
			autoListen.removeListener(instance as IEventDispatcher);
		}
		
		/** onHandledFocus is used to ensure that the layout has successfully handled
		 *  the layout of its initial focused attachment before transitioning in the remaining pieces
		 *  of UI.
		 */
		protected function onHandledFocus(event:LayoutEvent):void
		{
			(event.target as flash.events.IEventDispatcher).removeEventListener(LayoutEvent.HANDLED_ATTACHMENT_FOCUS, onHandledFocus);
			if (currentState != STATE_NORMAL) {
				currentState = STATE_NORMAL;
			}
		}
		
		CONFIG::logging {
			private function onRendererAdd(event:RendererExistenceEvent):void
			{
				logger.debug("renderer added: " + (event.renderer as ISimpleCard).fileName);
			}
		}
		
		protected function onMarqueeSelectionBegin(event:MarqueeSelectionEvent):void
		{
			marqueeMove = false;
		}
		
		protected function onMarqueeSelectionMove(event:MarqueeSelectionEvent):void
		{
			if (!cardLayout)
				return;

			if (!marqueeMove) {
				// If shift or ctrl selected
				if (shiftOrCtrlDownButNotBoth(event))
					// Get base selection that will be modified
					baseSelection = selectedIndices.slice();
				else {
					// Clear base selection and current selection
					baseSelection = null;
					selectedItem = null;
				}
				marqueeMove = true;
			}
			
			// Get the items within the marquee selection boundary
			var pos:Point = dataGroup.globalToLocal(new Point(event.x, event.y));
			var marqueeRect:Rectangle = new Rectangle(pos.x, pos.y, event.width, event.height);
			var indices:Vector.<int> = cardLayout.findCardsThatIntersect(marqueeRect);
			// Grow and/or shrink selection if shift/ctrl modifiers active
			if (shiftOrCtrlDownButNotBoth(event))
				indices = growOrShrinkSelection(event, indices);
			// Update the multi selection
			updateMultiSelection(indices);
		}
		
		protected function growOrShrinkSelection(event:MarqueeSelectionEvent, indices:Vector.<int>):Vector.<int>
		{
			// If no new selection (i.e. selection is beginning outside of item(s) to be marqueed), don't make any changes
			if (!indices || !indices.length)
				return baseSelection;
			// If no existing selection, leave new selection unchanged
			if (!selectedIndices)
				return indices;
			var newSelection:Vector.<int> = baseSelection.slice();
			for each (var item:int in indices) {
				var indexInOld:int = newSelection.indexOf(item);
				// If item was not previously selected
				if (indexInOld < 0)
					// Add it
					newSelection.push(item);
				// If item was previously selected and ctrl modifier set
				else if (event.ctrlKey)
					// Remove it
					newSelection.splice(indexInOld, 1);
			}
			return newSelection;
		}
		
		protected function shiftOrCtrlDownButNotBoth(event:MarqueeSelectionEvent):Boolean
		{
			return (event.shiftKey || event.ctrlKey) && !(event.shiftKey && event.ctrlKey);
		}
		
		// override as needed
		protected function onMarqueeSelectionEnd(event:MarqueeSelectionEvent):void
		{
		}
		
		protected function onMarqueeSelectionCancelled(event:MarqueeSelectionEvent):void
		{
			collectionMgr.setSelectedItems([]);
		}
		
		protected function updateMultiSelection(indices:Vector.<int>):void
		{
			if (!indices)
				return;
			setSelectedIndicesWithChangeDispatch(indices);
		}
		
		/** Returns the current state of this instance, since its skins should have a matching
		 *  set of states.
		 */
		override protected function getCurrentSkinState():String
		{
			return currentState;
		}

		/**
		 *  Sets the state of the skin as well as the state of this instance.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		override public function setCurrentState(stateName:String, playTransition:Boolean = true):void
		{
			if (stateName == currentState) {
				return;
			}
			
			super.setCurrentState(stateName, playTransition);
			invalidateSkinState();
		}
		
		protected function modifiersActive(event:MouseEvent, shiftKeyOK:Boolean = false, ctrlKeyOK:Boolean = false):Boolean
		{
			if (event.altKey)
				return true;
			if (event.shiftKey && !shiftKeyOK)
				return true;
			if (event.ctrlKey && !ctrlKeyOK)
				return true;
			return false;
		}
		
		private function cardSizeChanged(value:String):void
		{
			if (!value)
				return;
			
			if (cardSize == value)
				return;
			
			cardSize = value;
			resetCardSizes();
		}
		
		protected function resetCardSizes():void {
			if (!dataGroup || !dataProvider)
				return;
			
			for (var i:int=0; i < dataGroup.numElements; ++i) {
				var card:IBasicCard = (useVirtualLayout ? dataGroup.getVirtualElementAt(i) : dataGroup.getElementAt(i)) as IBasicCard;
				if (!card)
					continue;
				if (card.currentState == cardSize)
					return;
				card.setCurrentState(cardSize);
			}
		}
		
		[Bindable]
		/**
		 * 
		 * @return The NavPreview associated with this container.
		 * 
	  	 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get navPreview():NavPreview
		{
			return _navPreview;
		}

		/**
		 *  @private
		 */
		public function set navPreview(value:NavPreview):void
		{
			_navPreview = value;
		}

		protected function onNavPreviewCurrentState(value:String):void
		{
			if (marqueeSelection)
				marqueeSelection.enabled = (value == NavPreview.STATE_DISABLED);
		}
		
	}
}
