/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.lists
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.api.INavigatorHost;
	import com.adobe.portfolio.cards.ISimpleCard;
	import com.adobe.portfolio.cards.SimpleCard;
	import com.adobe.portfolio.components.NavListDropIndicator;
	import com.adobe.portfolio.components.supportClasses.AutoListen;
	import com.adobe.portfolio.components.supportClasses.AutoWatch;
	import com.adobe.portfolio.components.supportClasses.IListenAndWatch;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.layouts.supportClasses.CardLayoutHelper;
	import com.adobe.portfolio.layouts.supportClasses.ICardLayout;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.ThumbnailManager;
	import com.adobe.portfolio.previews.IPreview;
	import com.adobe.portfolio.utils.JobQueue;
	import com.adobe.portfolio.utils.LogHelper;
	import com.adobe.portfolio.utils.SortUtil;
	
	import flash.display.DisplayObject;
	import flash.events.Event;
	import flash.events.IEventDispatcher;
	import flash.events.KeyboardEvent;
	import flash.events.TimerEvent;
	import flash.geom.Point;
	import flash.ui.Keyboard;
	import flash.utils.Timer;
	
	import mx.binding.utils.ChangeWatcher;
	import mx.collections.ArrayCollection;
	import mx.collections.IList;
	import mx.core.DragSource;
	import mx.core.IVisualElement;
	import mx.core.mx_internal;
	import mx.events.CollectionEvent;
	import mx.events.CollectionEventKind;
	import mx.events.DragEvent;
	import mx.events.FlexEvent;
	import mx.events.ResizeEvent;
	import mx.managers.DragManager;
	import mx.managers.IFocusManagerComponent;
	
	import spark.components.DataGroup;
	import spark.components.List;
	import spark.components.TextArea;
	import spark.events.IndexChangeEvent;
	import spark.layouts.supportClasses.DropLocation;
	
	/**
	 * The CardList class extends List to provide support for Navigators that display lists of cards.
	 * It also handles drag and drop gestures and allows for reordering of items. 
	 * 
  	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class CardList extends List implements IListenAndWatch
	{
		/** 
		 *  The navigator's CollectionManager instance. All host interaction should go through this
		 *  instance, rather than the INavigatorHost directly.
		 *  
		 *  @langversion 3.0
	 	 *  @playerversion Flash 10.1
	 	 *  @playerversion AIR 2.0
	 	 *  @productversion Portfolio 10.0
		 */
		public const collectionMgr:CollectionManager = CollectionManager.instance;
		
		/**
		 * Format that defines the set of Acrobat drag and drop attachments. 
		 *  
		 *  @langversion 3.0
	 	 *  @playerversion Flash 10.1
	 	 *  @playerversion AIR 2.0
	 	 *  @productversion Portfolio 10.0
		 */
		public static const DRAG_DROP_ATTACHMENTS:String = "attachments";
		
		/**
		 * Format that defines the default set of List drag and drop items. 
		 *  
		 *  @langversion 3.0
	 	 *  @playerversion Flash 10.1
	 	 *  @playerversion AIR 2.0
	 	 *  @productversion Portfolio 10.0
		 */
		public static const DRAG_DROP_ITEMS:String = "itemsByIndex";
		
		private static var disabledDragAndDropCounter:Number = 0;
		/**
		 * Functions used to disable and enable drag and drop. If disabledDragAndDropCounter > 0,
		 * a drag will not be started. Drags can be disabled by other components like ScrollPad.
		 *   
		 *  @langversion 3.0
	 	 *  @playerversion Flash 10.1
	 	 *  @playerversion AIR 2.0
	 	 *  @productversion Portfolio 10.0
		 */
		public static function disableDragAndDrop():void
		{
			disabledDragAndDropCounter++;
		}
		public static function enableDragAndDrop():void
		{
			if (disabledDragAndDropCounter > 0) {
				--disabledDragAndDropCounter;
			}
		}
		
		/**
		 * True if drag and drop has been temporarily disabled for card lists.
		 * 
		 * @see disableDragAndDrop
		 * @see enableDragAndDrop
		 */
		protected static function get isDragAndDropDisabled():Boolean
		{
			return (disabledDragAndDropCounter > 0);
		}
		
		/**
		 * @private
		 * A timer to check if hover on a folder(during a drag) is long enough to open it.
		 *  
		 *  @langversion 3.0
	 	 *  @playerversion Flash 10.1
	 	 *  @playerversion AIR 2.0
	 	 *  @productversion Portfolio 10.0
		 */
		private static var openFolderTimer:Timer = new Timer(1000, 1);
		
		
		/** 
		 *  A boolean to indicate that a drag operation is in progress
		 *  
		 *  @langversion 3.0
	 	 *  @playerversion Flash 10.1
	 	 *  @playerversion AIR 2.0
	 	 *  @productversion Portfolio 10.0
		 */
		public var dragStarted:Boolean = false;
		
		/** 
		 * To allow for reordering of items using simple drag and drop gestures. 
		 * If this is set to false, drag and drop will still be allowed, but reordering will not occur.
		 *
		 *  @langversion 3.0
	 	 *  @playerversion Flash 10.1
	 	 *  @playerversion AIR 2.0
	 	 *  @productversion Portfolio 10.0
		 */
		public var useDragDropReordering:Boolean = false;
		
		/** 
		 * To enable dragDrop on folders during a drag operation. 
		 * Some layouts like miniNav allow for drag/drop on folders, whereas others like Linear don't.
		 *
		 *  @langversion 3.0
	 	 *  @playerversion Flash 10.1
	 	 *  @playerversion AIR 2.0
	 	 *  @productversion Portfolio 10.0
		 */
		public var useDragDropOnFolders:Boolean = false;
		
		protected var autoListen:AutoListen;
		protected var autoWatch:AutoWatch;
		protected var cardLayout:ICardLayout;
		protected var cardLayoutHelper:CardLayoutHelper;
		protected var focusCardDelta:int;

		private var handleCurrentFolderTimer:Timer = new Timer(200, 1);
		
		/**
		 * @private
		 * Used by drag/drop. When hovered over a folder during drag, the folder 
		 * to change to after the openFolderTimer expires, is cached.
		 */
		private var toChangeFolder:IAttachment;
		
		/**
		 * @private 
		 * New selection indices after reordering
		 */
		private var newSelectionAfterReordering:Vector.<int>;
		
		/**
		 * @private
		 * <code>selectionWasInitializedFromHost</code> gets set to true when we've successfully updated our own
		 * selection from the host selection. This is to ensure that we don't accidentally push our
		 * uninitialized selection to the host prematurely.
		 */
		private var selectionWasInitializedFromHost:Boolean = false;
		
		/**
		 * Constructor
		 * 
		 *  @langversion 3.0
	 	 *  @playerversion Flash 10.1
	 	 *  @playerversion AIR 2.0
	 	 *  @productversion Portfolio 10.0
		 */
		public function CardList()
		{
			super();
			
			// IAttachments don't have a 'label' propery, so default to 'fileName'
			labelField = "fileName";
			
			allowMultipleSelection = true;
			dragEnabled = true; 
			dragMoveEnabled = true;
			dropEnabled = true;
			addListenAndWatch();
			focusCardDelta = 1;
		}
		
		/**
		 *  Adds listeners and watchers that are automatically unwatched and unlistened.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function addListenAndWatch():void
		{
			removeListenAndWatch();
			autoListen = new AutoListen();
			autoWatch = new AutoWatch();
			addListener(collectionMgr, CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent);
			addListener(this, IndexChangeEvent.CHANGE, onIndexChangeEvent);
			addListener(this, FlexEvent.CREATION_COMPLETE, onCreationComplete);
			addEventListener(flash.events.Event.SELECT_ALL, selectEverything);
		}
		
		/**
		 * @private
		 * Look at <code>CardLayoutHelper</code> to see how this is used.
		 */
		private var _enableIdle:Boolean = true;
		public function get enableIdle():Boolean
		{
			return _enableIdle;
		}
		public function set enableIdle(value:Boolean):void
		{
			if (value == _enableIdle) {
				return;
			}
			_enableIdle = value
			if (cardLayoutHelper != null) {
				cardLayoutHelper.enableIdle = _enableIdle;
			}
		}
		
		/**
		 *  Function to cleanup the event listeners required by miniNav list. This should be
		 *  called when the component is no longer needed, to prevent memory leaks.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function removeListenAndWatch():void
		{
			if (autoListen)
				autoListen.removeAllListeners();
			autoListen = null;
			if (autoWatch)
				autoWatch.removeAllWatchers();
			autoWatch = null;
		}
		
		protected function addListener(listener:IEventDispatcher, type:String, handler:Function):void {
			autoListen.addListener(listener, type, handler);
		}
		
		protected function removeListener(listener:IEventDispatcher, type:String = null):void {
			autoListen.removeListener(listener, type);
		}
		
		protected function addWatcher(watcher:ChangeWatcher):void {
			autoWatch.addWatcher(watcher);
		}

		protected function onCreationComplete(event:FlexEvent):void
		{
			cardLayout = layout as ICardLayout;
			if (cardLayout) {
				cardLayoutHelper = cardLayout.cardLayoutHelper;
				if (cardLayoutHelper != null) {
					cardLayoutHelper.enableIdle = enableIdle;
				}
			} else {
				CONFIG::logging{
					LogHelper.logger.warn("CardList.onCreationComplete: expected an layout of type ICardLayout");
				}
			}
			addListener(this, ResizeEvent.RESIZE, onResize);
		}

		/**
		 * @inheritDoc
		 */
		override protected function partAdded(partName:String, instance:Object):void {
			super.partAdded(partName, instance);
			switch (instance) {
				case dataGroup:
					dispatchEvent(new Event("dataGroupChanged"));
					break;
				case scroller:
					addListener(scroller.horizontalScrollBar, FlexEvent.CHANGE_START, onScroll); 
					addListener(scroller.horizontalScrollBar, FlexEvent.CHANGE_END, onScroll); 
					addListener(scroller.verticalScrollBar, FlexEvent.CHANGE_START, onScroll); 
					addListener(scroller.verticalScrollBar, FlexEvent.CHANGE_END, onScroll); 
					break;
			}
		}
		
		/**
		 * Listener to handle scroll events of the scroller on cardLists.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onScroll(event:Event):void {
			switch (event.type) {
				case FlexEvent.CHANGE_START:
				case FlexEvent.CHANGE_END:
					var pause:Boolean = event.type == FlexEvent.CHANGE_START;
					ThumbnailManager.instance.pauseQueue = pause;
					var rootItemsJobQueue:JobQueue = collectionMgr.rootItemsJobQueue;
					if (rootItemsJobQueue)
						rootItemsJobQueue.pause = pause;
					break;
				
				default:
					break;
			}
		}
		
		/**
		 * @inheritDoc
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override protected function partRemoved(partName:String, instance:Object):void {
			super.partRemoved(partName, instance);
			switch (instance) {
				case dataGroup:
					dispatchEvent(new Event("dataGroupChanged"));
					break;
			}
		}
		
		/**
		 * Returns true if the dataProvider for this list matches the CollectionMgr's currentItems.
		 * Things like focus and selection should only be propagated to the CollectionMgr if the
		 * dataProvider is current.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function dataProviderIsCurrent():Boolean
		{
			return collectionMgr.itemsAreCurrent(dataProvider);
		}
	
		/**
		 *  Listener to handle events from the Collection Manager.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onNavHostEvent(event:CollectionManagerEvent):void
		{
			switch (event.kind) {
				case CollectionManagerEvent.HOST_CHANGED_KIND:
					var newHost:INavigatorHost = event.newVal as INavigatorHost;
					if (!newHost && event.oldVal) {
						removeListenAndWatch();
						removeEventListener(event.kind, arguments.callee);
					} 
					break;
				
				case CollectionManagerEvent.PART_REMOVED_KIND:
					if (event.oldVal == this) {
						// This part is being removed, so clean up event listeners
						removeListenAndWatch();
					}
					break;
				
				case CollectionManagerEvent.FOCUS_DID_CHANGE_KIND:
					if (visible && enabled)
						handleFocusChange(event);
					break;
				
				case CollectionManagerEvent.SELECTION_DID_CHANGE_KIND:
					if (visible && enabled)
						initializeSelectionFromHost();
					break;
				
				default:
					break;
			}
		}
		
		/**
		 * Triggers generation of thumbnails around the new focus card.
		 *
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function handleFocusChange(event:CollectionManagerEvent):void
		{
			if (!dataProviderIsCurrent()) {
				// If our dataProvider isn't current, don't bother responding to focus events
				return;
			}
			var oldAttachment:IAttachment = event.oldVal as IAttachment;
			var card:ISimpleCard = cardFromAttachment(dataGroup, oldAttachment);
			if (card) {
				var preview:IPreview = card.previewComponent;
				if (preview)
					preview.stopPlayback();
			}
			if (cardLayoutHelper != null) {
				cardLayoutHelper.idleCardsAroundFocus(dataProvider, event.newVal as IAttachment, focusCardDelta);
			}
		}
		
		private function selectEverything(event:Event):void
		{
			if (event.isDefaultPrevented()) {
				// Something already handled this event and called preventDefault, so ignore
				return;
			}
			
			// If current item is edit control then ignore this command
			var currentFocus:IFocusManagerComponent = (focusManager != null) ? focusManager.getFocus() : null;
			var textArea:TextArea = currentFocus as TextArea;
			if ((textArea != null) && textArea.enabled && textArea.selectable)
				return;

			// If current item is not edit control and parent is card list then do select all
			var itemParent:Object = event.target;
			while (itemParent) 
			{
				if (itemParent == this)
				{
					// Use CollectionManager.setSelectedItems rather than setSelectedIndicesWithChangeDispatch
					// to avoid scrolling the list (due to a caret/index change) on a select all.
					var newArray:Array = new Array();
					for (var j:int = 0; j < dataProvider.length; ++j) {
						newArray.push(dataProvider[j]);
					}
					
					collectionMgr.setSelectedItems(newArray);
					return;
				}
				itemParent = itemParent.parent;
			}
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function keyDownHandler(event:KeyboardEvent):void
		{
			if ( (event.keyCode == Keyboard.A) && event.ctrlKey)
			{
				if (event.shiftKey) {
					// deselect everything
					setSelectedIndicesWithChangeDispatch( new Vector.<int>());
				} else {
					// select everything
					selectEverything(event);
				}
				event.preventDefault();
				return;
			}
			
			super.keyDownHandler( event);
		}		
		
		protected function initializeSelectionFromHost():Boolean
		{
			if (dataProvider == null) {
				return false;
			}
			var newSelection:Vector.<Object> = new Vector.<Object>();
			var currentProvider:Boolean = dataProviderIsCurrent();
			if (currentProvider) {
				selectionWasInitializedFromHost = true;
				var dataLength:int = dataGroup ? dataGroup.numElements : 0;
				if (dataLength > 0) {
					var hostSelection:ArrayCollection = collectionMgr.selectedItems;
					if (hostSelection != null) {
						for each (var hostSelectedItem:Object in hostSelection) {
							newSelection.push(hostSelectedItem);
						}
					}
				}
			}
			
			// Avoid recursion. The nav host event could have been triggered from IndexChange event below. 
			if ((!requireSelection || newSelection.length > 0) && mx_internal::_proposedSelectedIndex < 0)
			{
				// It seems like checking against selectedItems would make more sense, but that causes
				// problems when there are mismatches between selection and the dataProvider. Using
				// selectedIndices is much safer.
				if ((selectedIndices != null) && (newSelection.length == selectedIndices.length)) {
					// Don't bother updating the current selection if it's already up-to-date
					var didChangeSelection:Boolean = false;
					for (var i:uint = 0; i < newSelection.length; ++i) {
						var existingSelectedIndex:int = selectedIndices[i];
						if ((existingSelectedIndex < 0) || (existingSelectedIndex >= dataProvider.length)) {
							didChangeSelection = true;
							break;
						}
						if (newSelection[i] != dataProvider.getItemAt(existingSelectedIndex)) {
							didChangeSelection = true;
							break;
						}
					}
					if (!didChangeSelection) {
						return false;
					}
				}
				selectedItems = newSelection;
				return true;
			} else if (currentProvider && (selectedItem != null)) {
				// The host is trying to deselect all items, but we need to keep one selected
				collectionMgr.setSelectedItems([selectedItem]);
			}
			return false;
		}
		
		/**
		 * Sets the dataProvider for this list.
		 * @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		override public function set dataProvider(value:IList):void
		{
			removeListener(dataProvider, CollectionEvent.COLLECTION_CHANGE);
			super.dataProvider = value;
			if (!value)
				return;
			addListener(dataProvider, CollectionEvent.COLLECTION_CHANGE, onCollectionChange);
			if (!selectionWasInitializedFromHost) {
				// Try to inialize selection from the host, since we haven't already
				initializeSelectionFromHost();
			}
			resizeCachedThumbs();
		}
		
		/**
		 * Listener to handle collection change events.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onCollectionChange(e:CollectionEvent):void {
			if (!enabled || !cardLayoutHelper)
				return;
			switch (e.kind) {
				case CollectionEventKind.ADD:
					initializeSelectionFromHost();
					break;
				case CollectionEventKind.REMOVE:
				case CollectionEventKind.REPLACE:
					initializeSelectionFromHost();
					break;
				case CollectionEventKind.REFRESH:
				case CollectionEventKind.RESET:
					// send out a caret change event since it's a whole new world (potentially)
					var event:IndexChangeEvent = new IndexChangeEvent(IndexChangeEvent.CARET_CHANGE);
					dispatchEvent(event);
					break;
			}
			if (!selectionWasInitializedFromHost) {
				// Try to inialize selection from the host, since we haven't already
				initializeSelectionFromHost();
			}
		}
		
		/**
		 * Force selection change.
		 * @param value The index to change selection to.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function setSelectedIndexWithChangeDispatch(value:int):void
		{
			mx_internal::setSelectedIndex(value, true);
		}
		
		/**
		 * Force selection change.
		 * @param value A vector of indices to change selection to.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function setSelectedIndicesWithChangeDispatch(value:Vector.<int>):void 
		{
			mx_internal::setSelectedIndices(value, true);
		}
		
		/**
		 *  This listener is called when the selection changes based on user interaction, or
		 *  when <code>setSelectedIndexWithChangeDispatch</code> or <code>setSelectedIndicesWithChangeDispatch</code> is used
		 *  to set the selection.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		protected function onIndexChangeEvent(event:IndexChangeEvent):void
		{
			if (!dataProviderIsCurrent()) {
				// If our dataProvider isn't current, don't bother changing the selection,
				return;
			}
			
			// This listener is only triggered on explicit list changes, so force the host
			// to update
			handleSelectionChanged();
			
			// If we've selected an item, make it the focused item
			if(event.newIndex >= 0)
			{
				var attachment:IAttachment = dataProvider.getItemAt(event.newIndex) as IAttachment;
				// To avoid recursion
				if (collectionMgr.focusAttachment != attachment)
					setFocusOnIndexChange(attachment);
			}
		}
		
		/**
		 * Updates the host's selection based on the current list selection.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function handleSelectionChanged():void
		{
			if (selectionWasInitializedFromHost && dataProviderIsCurrent()) {
				var newSelection:Array = new Array();
				for each (var selectedItem:Object in selectedItems) {
					if (selectedItem is IAttachment) {
						newSelection.push(selectedItem);
					}
				}
				collectionMgr.setSelectedItems(newSelection);
			}
		}
		protected function setFocusOnIndexChange(attachment:IAttachment):void
		{
			if (!attachment)
				return;

			collectionMgr.focusAttachment = attachment;
		}
		
		protected function onResize(event:ResizeEvent):void
		{
			resizeCachedThumbs();
		}
		
		// override as necessary
		protected function resizeCachedThumbs():void 
		{
			if (!cardLayoutHelper)
				return;
			
			cardLayoutHelper.resizeCachedThumbs(dataProvider, parent.width, parent.height, focusCardDelta);
		}
		
		/**
		 * @inheritDoc
		 */
		override public function updateRenderer(renderer:IVisualElement, itemIndex:int, data:Object):void
		{
			var newCard:ISimpleCard = renderer as ISimpleCard;
			var newCardData:Object = (newCard == null) ? null : newCard.data;
			super.updateRenderer(renderer, itemIndex, data);
			
			var oldCard:ISimpleCard = dataGroup.getElementAt(itemIndex) as ISimpleCard;
			if (!newCard || !oldCard)
				return;
			
			if (cardLayoutHelper)
				newCard.alpha = cardLayoutHelper.getAlpha(newCard);
			
			// Give the card a chance to update the new renderer
			if (newCard != oldCard)
				newCard.updateRenderer(oldCard);
			
			// The data of a card changed, so make sure the layout can account for it
			if ((newCardData != null) && (data != newCardData) && (dataGroup != null)) {
				dataGroup.invalidateDisplayList();
			}
		}
		
		/**
		 * 
		 * Drag methods
		 * 
		 */
		
		/**
		 * Takes the data set by the list for ordered items and adds it to the attachments
		 * data format. Acrobat uses that for external drag-drops 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function addDragData(dragSource:DragSource):void
		{
			// Add list specific items such as ordered items
			super.addDragData(dragSource);
			
			// Add attachments data (just a copy of the default drag/drop items)
			dragSource.addData(dragSource.dataForFormat(DRAG_DROP_ITEMS) as Vector.<Object>, DRAG_DROP_ATTACHMENTS);
		}
		
		/** 
		 * Does not start drag if some other component has disabled it. Also distinguishes text selection from a drag.
		 * @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override protected function dragStartHandler(event:DragEvent):void
		{
			if (isDragAndDropDisabled) {
				// Don't intiate a drag if something else (like ScrollPad) has disabled it
				return;
			}
			
			// Don't actually start dragging if user is selecting text
			var textArea:TextArea = (focusManager == null) ? null : focusManager.getFocus() as TextArea;
			if ((textArea != null) && textArea.enabled && textArea.selectable) {
				return;
			}
			
			dragStarted = true;
			super.dragStartHandler(event);
		}
		
		/** 
		 * Calculate the dropLocation of the given dragEvent. This is the location where the mouse is gestured to drop. 
		 * @return The drop location for this event, or null if the drop operation is not available. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function calculateDropLocation(event:DragEvent):DropLocation
		{
			// Verify data format
			if (!enabled || !event.dragSource.hasFormat(DRAG_DROP_ATTACHMENTS))
				return null;
			
			// Calculate the drop location
			return layout.calculateDropLocation(event);
		}
		
		/**
		 * Creates a custom dropIndicator for the cardList.
		 * @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function createDropIndicator():DisplayObject
		{
			// Do we have a drop indicator already?
			if (layout.dropIndicator)
				return layout.dropIndicator;
			
			var dropIndicatorInstance:DisplayObject = dropIndicator ? 
				createDynamicPartInstance("dropIndicator") as DisplayObject : new NavListDropIndicator();

			if (dropIndicatorInstance is IVisualElement)
				(dropIndicatorInstance as IVisualElement).owner = this;
			
			// Set it in the layout
			layout.dropIndicator = dropIndicatorInstance;
			return dropIndicatorInstance;
		}
		
		/** 
		 * @private
		 * Handles drags on folder items. It hides the drop indicator, caches the folder to change to and starts the timer.
		 * This should not be called with non-folder attachments 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		private function handleDragOnFolder(folderAtt:IAttachment):void
		{
			if (!folderAtt)
				return;
			
			layout.hideDropIndicator();
			if (openFolderTimer.running)
				return;
			
			toChangeFolder = folderAtt;
			collectionMgr.collectionDropManager.targetPath = toChangeFolder;
			openFolderTimer.addEventListener(TimerEvent.TIMER_COMPLETE, onOpenFolderTimeComplete);
			openFolderTimer.start();
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function dragEnterHandler(event:DragEvent):void
		{
			if (event.isDefaultPrevented())
				return;
			
			var dropLocation:DropLocation = calculateDropLocation(event);
			if (enabled && collectionDropEnabled(event))
			{		
				dragStarted = true;

				// Create the dropIndicator instance. The layout will take care of
				// parenting, sizing, positioning and validating the dropIndicator.
				createDropIndicator();
			
				if (event.type == DragEvent.DRAG_ENTER)
					DragManager.acceptDragDrop(this);
				
				// Notify manager we can drop. Platform drags always provide copy feedback
				if (event.dragInitiator)
					// TODO: Support copy. Needs changes on the framework side
					DragManager.showFeedback(DragManager.MOVE);
				else
					// Platform drag
					DragManager.showFeedback(DragManager.COPY);
				
				// If we need list to show the drop indicator, create a drop indicator first. See miniNavList for example.
				if (dropLocation && (useDragDropReordering || useDragDropOnFolders))
				{
					// Get the item at drag location
					var folderItem:IAttachment = getFolderItemAtDragLocation(dropLocation);
					if (!folderItem)
					{
						// Show the drop indicator if the mouse location is not on folder
						layout.showDropIndicator(dropLocation);
						resetOpenFolderTimer();
					}
					else
						handleDragOnFolder(folderItem);
				}
			}
			else
			{
				// Using platform to show the drop indicator. Hence this has no effect if a dropIndicator has not been created. 
				// If we need list to show the drop indicator, create a drop indicator first. See miniNavList for example.
				// Hide if previously showing
				layout.hideDropIndicator();
				
				DragManager.showFeedback(DragManager.NONE);
			}
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function dragOverHandler(event:DragEvent):void
		{
			dragEnterHandler(event);
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function dragExitHandler(event:DragEvent):void
		{
			if (event.isDefaultPrevented())
				return;
			
			// Hide the drop indicator
			layout.hideDropIndicator();
			destroyDropIndicator();
			dragStarted = false;
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function dragCompleteHandler(event:DragEvent):void
		{
			if (event.isDefaultPrevented())
				return;

			// Hide the drop indicator
			layout.hideDropIndicator();
			destroyDropIndicator();
			dragStarted = false;			
			
			toChangeFolder = null;
		}
		
		protected function updateSelection():void
		{
			if (!newSelectionAfterReordering || newSelectionAfterReordering.length <= 0 || newSelectionAfterReordering == selectedIndices)
				return;
			
			selectedIndices = newSelectionAfterReordering;
			selectedIndex = newSelectionAfterReordering.length > 0 ? newSelectionAfterReordering[newSelectionAfterReordering.length - 1] : selectedIndex;
			newSelectionAfterReordering = null;
		}
		
		/** 
		 * 
		 * Drop methods
		 * 
		 */
		
		/**
		 * Helper function that determines whether drag drop is supported
		 *
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function collectionDropEnabled(event: DragEvent): Boolean
		{
			// Document Permission calls are not checked here as they are hard requirements enforced 
			// by Acrobat platform drop handler. Platform drop handler will not call the dragenter,
			// dragover, dragexit, and dragdrop even in the following conditions
			// - EF/Create and EF/Import is not available for copy operations
			// - EF/Modify is not available for move operations
			// The checks done here are navigator specific requirements
			return ((collectionMgr.authoringMode) && (event.dragSource.hasFormat(DRAG_DROP_ATTACHMENTS)));
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function dragDropHandler(event:DragEvent):void
		{
			if (event.isDefaultPrevented())
				return;
			
			// Hide the drop indicator
			layout.hideDropIndicator();
			destroyDropIndicator();
			dragStarted = false;
			
			resetOpenFolderTimer();
		
			if ((useDragDropReordering) && collectionDropEnabled(event))
				handleDragDropReordering(event);
		}
		
		/** 
		 * Handles the reordering of items after a dragDrop. However, it does not modify the dataProvider, but invokes
		 * <code>makeNewOrderedList</code> to create an ordered list of items after the drop. Drop on a folder is handled separately by
		 * <code>handleDropOnFolder</code>. It also handles adding of items as a result of an external drag/drop (dragInitiator is null in this case). 
		 * However,the new items are already added to the dataProvider in this case as well, this function simply puts them in the correct order.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function handleDragDropReordering(event:DragEvent):void
		{
			if (!event.dragSource)
				return;
			
			var dragItems:Vector.<Object> = event.dragSource.dataForFormat(DRAG_DROP_ATTACHMENTS) as Vector.<Object>;
			var att:IAttachment;
			if (event.dragInitiator == null)
			{
				// Files have already been added (not at the ibeam position), remove them for now. Add them back later.
				for each (att in dragItems)
				{
					var itemIndex:int = dataProvider.getItemIndex(att);
					if (itemIndex >= 0)
						dataProvider.removeItemAt(itemIndex);
				}
			}
			
			// Get the dropLocation
			var dropLocation:DropLocation = calculateDropLocation(event);
			if (!dropLocation)
				return;
			
			CONFIG::logging{
				LogHelper.logger.debug("DropIndex: " + dropLocation.dropIndex);
			}
			
			var copyItems:Boolean = event.action == DragManager.COPY || event.ctrlKey;			
			// Make sure the manager has the appropriate action
			DragManager.showFeedback(copyItems ? DragManager.COPY : DragManager.MOVE);
			var folderItem:IAttachment = getFolderItemAtDragLocation(dropLocation);
			if (folderItem)
			{
				// We should drop on (inside) the folder
				collectionMgr.newOrderedList = handleDropOnFolder(folderItem, event);
				reorderAttachments(folderItem.children);
				return;
			}
						
			if(dragItems == null || dragItems.length == 0)
			{
				CONFIG::logging{
					LogHelper.logger.warn("Expected attachments, but found none during drop");
				}
				return;
			}
			
			collectionMgr.newOrderedList = makeNewOrderedList(dropLocation.dropIndex, getDroppingItems(dragItems));
			if (event.dragInitiator == null)
			{
				// this is a drag from platform, adding back the files that were removed earlier.
				for each (att in dragItems)
				dataProvider.addItem(att);
			}
			
			reorderAttachments(dataProvider);
		}
		
		private function resetOpenFolderTimer():void 
		{
			if (!openFolderTimer.running)
				return;
			
			toChangeFolder = null;
			collectionMgr.collectionDropManager.targetPath = collectionMgr.currentFolder;
			openFolderTimer.removeEventListener(TimerEvent.TIMER_COMPLETE, onOpenFolderTimeComplete);
			openFolderTimer.reset();
		}
		
		private function getDroppingItems(dragItems:Vector.<Object>):ArrayCollection
		{
			var droppingItems:ArrayCollection = new ArrayCollection();
			var registered:Boolean = false;
			
			for (var i:int = 0; i < dragItems.length; ++i)
			{
				var att:IAttachment = dragItems[i] as IAttachment;				
				var attIndex:int = dataProvider.getItemIndex(att);
				var newFolder:IAttachment = collectionMgr.currentFolder;
				if (attIndex < 0 && att.parent != newFolder && att != newFolder)
				{
					// Parent has changed. This will do the needed remove and add.
					att.parent = newFolder;
					if (!registered)
					{
						var children:IList = newFolder ? newFolder.children : collectionMgr.rootItems;
						if (children) {
							addListener(children, CollectionEvent.COLLECTION_CHANGE, onChildrenCollectionChange)
							registered = true;
						}
					}
				}
				// Cache the items to be dropped at the dropIndex
				droppingItems.addItem(att);				
			}
			return droppingItems;
		}

		/** 
		 * Finds the folder at a given dropLocation
		 * @param dropLocation The drop location at which the folder item needs to be determined.
		 * @return The folder item at the drop location. Returns null if there is no folder at the given location
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getFolderItemAtDragLocation(dropLocation:DropLocation):IAttachment
		{
			// Get the card at the current location. For some reason droppoint is corrupted here
			var objects:Array = this.stage.getObjectsUnderPoint(new Point(dropLocation.dragEvent.stageX, dropLocation.dragEvent.stageY));
			for (var index:int = 0; index < objects.length; ++index) 
			{
				if ( !objects[index].hasOwnProperty("parent") )
					continue;
				
				var itemParent:ISimpleCard = objects[index].parent as ISimpleCard 
				if (itemParent && itemParent.owner == this)
				{
					var dropLocationAtt:IAttachment = itemParent.data as IAttachment;
					if ((dropLocationAtt) && (dropLocationAtt.isFolder))
					{
						// Make sure the folder is not part of the drag
						var dragItems:Vector.<Object> = dropLocation.dragEvent.dragSource.dataForFormat(DRAG_DROP_ATTACHMENTS) as Vector.<Object>;
						// Note that dragItems are not sent for drags initiated from the platform. The check
						// is not necessary in those cases too
						if ((dragItems) && (dragItems.indexOf(dropLocationAtt) != -1))
								return null;
						return dropLocationAtt;
					}
				}
			}
			
			return null;
		}
		
		/**
		 * Called when the <code>openFolderTimer</code> completes. The current folder is now changed to the cached
		 * <code>toChangeFolder</code> during an ongoing drag.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onOpenFolderTimeComplete(evt:TimerEvent) : void
		{
			if (evt.target)
				evt.target.removeEventListener(TimerEvent.TIMER_COMPLETE, onOpenFolderTimeComplete);
			if (!toChangeFolder || !dragStarted)
				return;
			CONFIG::logging{
				LogHelper.logger.debug("onOpenFolderTimeComplete");
			}
			collectionMgr.collectionDropManager.targetPath = toChangeFolder;
			collectionMgr.currentFolder = toChangeFolder;
		}
		
		/**
		 * This function is used when attachments are moved across folders. Once the parent of all attachments
		 * point to the new location, and the attachment has been added to the new collection, the attachments are
		 * sent for reordering.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onChildrenCollectionChange(event:CollectionEvent):void
		{
			var reordered:Boolean = false;
			if (event.kind == CollectionEventKind.ADD)
			{
				var source:IList = event.currentTarget as IList;
				reordered = reorderAttachments(source);
			}
			if (reordered)
			{
				if (event.target is IEventDispatcher)
					(event.target as IEventDispatcher).removeEventListener(CollectionEvent.COLLECTION_CHANGE, onChildrenCollectionChange);
			}
		}
		
		/** 
		 * Handles drop on a folder. Changes the parent to point to the new location.
		 * @param folder The folder to which the items are being moved to or added to.
		 * @param event The drag event associated with this drop.
		 * @return The ordered list of items in the given folder
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function handleDropOnFolder(folder:IAttachment, event:DragEvent):IList
		{
			var dragSource:DragSource = event.dragSource;
			var dragItems:Vector.<Object> = dragSource.dataForFormat(DRAG_DROP_ATTACHMENTS) as Vector.<Object>;
			if(dragItems == null || dragItems.length == 0)
			{
				CONFIG::logging{
					LogHelper.logger.warn("Expected attachments, but found none during drop");
				}
				return null;
			}
			
			var dragItemsCollection:ArrayCollection = new ArrayCollection;
			// Make a list of the items in the folder.
			var newList:ArrayCollection = new ArrayCollection;
			var finalList:ArrayCollection = new ArrayCollection;
			CONFIG::logging{
				if (folder.children.length == 0)
					LogHelper.logger.warn("Folder is empty, might need first access");
			}
			
			for (var i:int = 0; i < dragItems.length; ++i)
			{
				dragItemsCollection.addItem(dragItems[i] as IAttachment);
			}
				
			for each (var child:IAttachment in folder.children)
			{
				//If this is not a move, but an add as a result of an external drag, 
				// attachments have already been added by the viewer. Don't add again.
				if (dragItemsCollection.contains(child))
					continue;
				newList.addItem(child);
			}
			
			newList.sort = SortUtil.createSortFieldObject(collectionMgr.sort);
			newList.refresh();
			
			for each (var att:IAttachment in newList)
			{
				finalList.addItem(att);
			}
			
			for each (att in dragItemsCollection)
			{
				// This will take care of add and delete
				if (event.dragInitiator != null)
					att.parent = folder;
				// Add the new item to the end
				finalList.addItem(att);
			}
			
			// Reorder command must be sent only after the item has been added to the new list. 
			addListener(folder.children, CollectionEvent.COLLECTION_CHANGE, onChildrenCollectionChange);
			return finalList;
		}
		
		/** 
		 * Creates an ordered list of items in the dataProvider after the drop.
		 * @param dropIndex The destination index where the items are being dropped to.
		 * @param droppingItems The items being droppped.
		 * @return An ordered list of the items in the data Provider. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function makeNewOrderedList(dropIndex:int, droppingItems:ArrayCollection):IList
		{
			var newOrderedList:ArrayCollection = new ArrayCollection;
			
			newSelectionAfterReordering= new Vector.<int>();
			var countingIndex:int = 0;
			
			if (droppingItems == null || droppingItems.length <=0 )
				return null;
			
			var att:IAttachment;
			
			// Push the items before the moving items
			for (var index:int = 0; index < dropIndex && index < dataProvider.length; index++)
			{
				if ( droppingItems.contains(dataProvider.getItemAt(index)) )
				{
					// Don't push moving items
					continue;
				}
				
				newOrderedList.addItem(dataProvider.getItemAt(index));
				++countingIndex;
			}
			
			// Push the moving items
			for(var k:int = 0; k < droppingItems.length;  k++)
			{
				newOrderedList.addItem(droppingItems.getItemAt(k));
				newSelectionAfterReordering.push(countingIndex + k);
			}
			
			// Push the remaining items before the moving items
			for(index = dropIndex; index < dataProvider.length; index++)
			{
				if ( droppingItems.contains(dataProvider.getItemAt(index)) )
				{
					// Don't push moving items
					continue;
				}
				newOrderedList.addItem(dataProvider.getItemAt(index));
			}
			
			return newOrderedList;
		}
		
		/** 
		 * Sends the reorder command once all the attachments are found in the given list. If items were moved between
		 * lists, this function may have to be called repetitively until it returns true to ensure all items
		 * have been successfully moved and the reorder command has been sent.
		 * @param source An ordered list of items in the dataProvider.
		 * @return True if all items are found in the given list and the reorder command was sent successfully.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function reorderAttachments(source:IList):Boolean
		{
			var newOrderedList:IList = collectionMgr.newOrderedList;
			// Should wait until the attachment has actually been deleted/readded.
			if (newOrderedList == null || newOrderedList.length <= 0)
				return false;
			
			for (var index:int = 0; index < newOrderedList.length; ++index)
			{
				var att:IAttachment = newOrderedList.getItemAt(index) as IAttachment;
				// First check if we are at the same level
				if ( source.getItemIndex(att) >= 0)
					continue;
				
				// Attachments may not be the same, if we are at a different level, compare filename.
				var newAtt:IAttachment = getAttachmentByName(att.fileName, source);
				if (!newAtt)
				{
					// data not in yet, wait for collection_change event before calling reordering.
					return false;
				}
				else
				{
					// remake the list, the attachments have gone stale.
					newOrderedList.removeItemAt(index);
					newOrderedList.addItemAt(newAtt, index)
				}
			}
			//Send the command over the bridge.
			collectionMgr.host.collection.reorderFiles(newOrderedList.toArray());
			
			//Clear the collectionMgr's list
			collectionMgr.newOrderedList = null;
			
			updateSelection();
			return true;
		}
		
		/** 
		 * @private
		 * Helper function. Given a name, returns the corresponding attachment
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		private function getAttachmentByName(name:String, source:IList):IAttachment
		{
			for each (var att:IAttachment in source)
			{
				if (att.fileName == name)
					return att;
			}
			return null;
		}
		
		/**
		 * Finds the card corresponding to the attachment in focus.
		 * @return The Card corresponding to the attachment in focus.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function findCardWithAttachmentFocus():SimpleCard
		{
			var focus:IAttachment = collectionMgr.focusAttachment;
			// If no attachment has focus, then neither does any card
			if (!focus)
				return null;
			
			if (!dataGroup)
				return null;
			for (var index:int = 0; index < dataGroup.numElements; ++index) {
				var card:SimpleCard = dataGroup.getElementAt(index) as SimpleCard;
				if (!card)
					continue;
				var cardAttachment:IAttachment = card.data as IAttachment;
				if (!cardAttachment)
					continue;
				if (cardAttachment == focus)
					return card;
			}
			
			return null;
		}
		
		// bindableDataGroup's only use is to eliminate annoying warnings when 
		//	dataGroup is referenced in MXML code
		[Bindable (name="dataGroupChanged")]
		public function get bindableDataGroup():DataGroup {
			return dataGroup;
		}
		
		private function set bindableDataGroup(value:DataGroup):void {
		}
		
		/**
		 * Finds the card associated with the given attachment
		 *  @param _dataGroup The datagroup to which the attachment belongs. 
		 *  @param attachment The attachment whose card is needed.
		 *  @return The card corresponding to the given attachment.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function cardFromAttachment(_dataGroup:DataGroup, attachment:Object):ISimpleCard
		{
			if (!_dataGroup)
				return null;
			
			if (!_dataGroup.dataProvider)
				return null;
			
			var newIndex:int = _dataGroup.dataProvider.getItemIndex(attachment);
			var card:ISimpleCard = _dataGroup.getElementAt(newIndex) as ISimpleCard;
			
			if (card && (card.data != attachment)) {
				var length:int = _dataGroup.numElements;
				for (var i:int = 0; i<length; ++i) {
					card = _dataGroup.getElementAt(i) as ISimpleCard;
					if (card && card.data == attachment)
						return card;
				}
				return null;  
			}
			
			return card;
		}
		
	}
}
