/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.lists
{
	import com.adobe.portfolio.cards.BasicCard;
	import com.adobe.portfolio.cards.CardFactory;
	import com.adobe.portfolio.cards.ISimpleCard;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.layouts.LinearLayout;
	import com.adobe.portfolio.managers.CollectionManager;
	
	import flash.events.IEventDispatcher;
	import flash.events.MouseEvent;
	import flash.events.TimerEvent;
	import flash.geom.PerspectiveProjection;
	import flash.geom.Point;
	import flash.utils.Timer;
	
	import mx.collections.IList;
	import mx.core.IButton;
	import mx.effects.Effect;
	import mx.events.CollectionEvent;
	import mx.events.EffectEvent;
	import mx.events.FlexEvent;
	import mx.events.ResizeEvent;
	
	import spark.effects.Fade;

	/**
	 * The LinearList class extends CardList to provide functionality specific to the Linear navigator.
	 * 
  	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class LinearList extends CardList
	{
		// String literals used to represent the skin parts exposed by LinearList to its skins.
		
		/**
		 * A string literal representing the name of the previous button skin part.
		 */
		static public const PART_PREV_BUTTON:String = "prevButton";
		
		/**
		 * A string literal representing the name of the next button skin part.
		 */
		static public const PART_NEXT_BUTTON:String = "nextButton";
		
		/**
		 * A string literal representing the name of the guide card skin part.
		 */
		static public const PART_GUIDE_CARD:String = "guideCard";
		
		/**
		 * The projection applied to the dataGroup so that 3D effects display properly.
		 * 
		 * @see #centerProjection
		 */
		protected var projection:PerspectiveProjection;
		
		public function LinearList()
		{
			super();
			
			// Set default item renderer; this can be overriden in mxml or AS
			itemRenderer = new CardFactory(CardFactory.BASIC_CARD, BasicCard.STATE_MEDIUM, CardFactory.DROP_SHADOW_BOTTOM);
			
			// Set default layout; this can be overriden in mxml or AS
			layout = new LinearLayout();
			useDragDropReordering = false;
			
			// We want the user to be able to tab to the next/previous buttons, and the guide card
			hasFocusableChildren = true;
			focusEnabled = false;
		}
		
		/**
		 * A button for navigating to the previous attachment, optionally provided by this instance's skin.
		 */
		[SkinPart(required="false")]
		public var prevButton:IButton;
		
		/**
		 * A button for navigating to the next attachment, optionally provided by this instance's skin.
		 */
		[SkinPart(required="false")]
		public var nextButton:IButton;
		
		/**
		 * A card used to display descriptive information about the current attachment.
		 */
		[SkinPart(required="false")]
		public var guideCard:ISimpleCard;
		
		/**
		 * @inheritDoc
		 */
		override protected function onResize(event:ResizeEvent):void
		{
			super.onResize(event);
			centerProjection();
			invalidateSize();
		}
		
		/**
		 * An empty implementation to prevent selected items from appear selected.
		 * 
		 * @inheritDoc
		 */
		override protected function itemSelected(index:int, selected:Boolean):void
		{
			// do nothing by design
		}
		
		/**
		 * Updates projection so that it aligns properly with the current list dimensions.
		 */
		protected function centerProjection():void
		{
			if (dataGroup == null) {
				return;
			}
			if ((stage == null) || (transform == null)) {
				// Try again later
				callLater(centerProjection);
				return;
			}
			
			if (!projection)
				projection = new PerspectiveProjection();
			projection.fieldOfView = 55;
			// Keep the projectionCenter to the right of the component, to keep the "back" cards on
			// on the right from being obscured by the "front" cards on the left.
			projection.projectionCenter = localToGlobal(new Point(width * 0.5, height * 0.5));
			
			dataGroup.transform.perspectiveProjection = projection;
			
			invalidateDisplayList();
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function onNavHostEvent(event:CollectionManagerEvent):void
		{
			if ((event.kind == CollectionManagerEvent.FOCUS_DID_CHANGE_KIND) && dataProviderIsCurrent()) {
				if (dataGroup != null) {
					dataGroup.invalidateDisplayList();
				}
				updateGuideCard();
			}
			super.onNavHostEvent(event);
		}
		
		private function updateGuideCard():void
		{
			if ((guideCard != null) &&
				(guideCard.data != collectionMgr.focusAttachment) &&
				dataProviderIsCurrent())
			{
				var guideEffect:Effect;
				if (guideCard.alpha > 0) {
					guideEffect = generateGuideCardEffect(true);
					if (guideEffect != null) {
						guideEffect.addEventListener(EffectEvent.EFFECT_END, onGuideOut);
						guideEffect.play();
					}
				} else {
					guideEffect = generateGuideCardEffect(false);
					if (guideEffect != null) {
						guideCard.data = collectionMgr.focusAttachment;
						guideEffect.play();
					}
				}
				if (guideEffect == null) {
					guideCard.data = collectionMgr.focusAttachment;
					guideCard.alpha = 1.0;
				}
			}
		}
	
		private function onGuideOut(event:EffectEvent):void
		{
			if (event.currentTarget is IEventDispatcher) {
				(event.currentTarget as IEventDispatcher).removeEventListener(event.type, onGuideOut);
			}
			guideCard.data = collectionMgr.focusAttachment;
			var guideIn:Effect = generateGuideCardEffect(false);
			if (guideIn != null) {
				guideIn.play();
			} else {
				guideCard.alpha = 1.0;
			}
		}
		private function generateGuideCardEffect(fadeOut:Boolean):Effect
		{
			var fade:Fade = new Fade(guideCard);
			fade.alphaFrom = fadeOut ? 1.0 : 0.0;
			fade.alphaTo = fadeOut ? 0.0 : 1.0;
			fade.duration = 375;
			return fade;
		}
		
		/**
		 * @inheritDoc
		 */
		override public function addListenAndWatch():void
		{
			super.addListenAndWatch();
			
			// If guideCard was already created, then we may have missed a focusDidChange event,
			// so make sure it's been setup properly
			updateGuideCard();
		}

		/**
		 * @inheritDoc
		 */
		override protected function partAdded(partName:String, instance:Object):void
		{
			handlePartChange(partName, instance, true);
			if (partName == "dataGroup") {
				centerProjection();
			}
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function partRemoved(partName:String, instance:Object):void
		{
			handlePartChange(partName, instance, false);
		}
		/** Internal function for adding and removing event listeners for skin parts, depending on
		 *  whether the part is being added or removed for this instance.
		 */
		private function handlePartChange(partName:String, instance:Object, adding:Boolean):void
		{
			var listeners:Array = null;
			var types:Array = null;
			if ((instance != null) && (instance is IEventDispatcher)) {
				var activePart:IEventDispatcher = instance as IEventDispatcher;
				
				// Build up an array of event types and listeners that apply to this object
				switch (partName) {
					case PART_NEXT_BUTTON:
						types = [ MouseEvent.CLICK ];
						listeners = [ collectionMgr.onGotoNextAttachment ];
						break;
					case PART_PREV_BUTTON:
						types = [ MouseEvent.CLICK ];
						listeners = [ collectionMgr.onGotoPreviousAttachment ];
						break;
					case PART_GUIDE_CARD:
						if (adding) {
							updateGuideCard();
						}
						break;
				}
				
				// Add or remove all the relevant event listeners
				var count:uint = (listeners == null) ? 0 : listeners.length;
				if (count > 0) {
					for (var i:uint = 0; i < count; ++i) {
						adding ? 	activePart.addEventListener(types[i], listeners[i]) :
							activePart.removeEventListener(types[i], listeners[i]);
					}
				}
			}
			adding ? super.partAdded(partName, instance) : super.partRemoved(partName, instance);
		}
		
		/**
		 * @inheritDoc
		 */
		override public function set dataProvider(value:IList):void
		{
			super.dataProvider = value;
			updateGuideCard();
		}
		
		/**
		 * @inheritDoc
		 */
		override protected function onCollectionChange(e:CollectionEvent):void
		{
			updateGuideCard();
		}
	}
}