/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

////////////////////////////////////////////////////////////////////////////////
//
//  ADOBE SYSTEMS INCORPORATED
//  Copyright 2010 Adobe Systems Incorporated
//  All Rights Reserved.
//
//  NOTICE: Adobe permits you to use, modify, and distribute this file
//  in accordance with the terms of the license agreement accompanying it.
//
////////////////////////////////////////////////////////////////////////////////

package com.adobe.portfolio.managers
{
	import com.adobe.portfolio.api.IAttachment;
	
	import mx.core.EventPriority;
	import mx.core.UIComponent;
	import mx.core.mx_internal;
	import mx.events.DragEvent;
	import mx.managers.DragManager;

	/**
	 *  The AttachmentDropManager class provides support for dropping IAttachments to the UI Component classes.  
	 * 
	 *  Any UI Component that registers with this class gets the platform 
	 *  attachment drop support. This class is based on the Spark List drag drop code.
	 *
	 *  @mxml <p>The <code>&lt;s:AttachmentDragManager&gt;</code> has the following tag attributes:</p>
	 *
	 *  <pre>
	 *  &lt;s:AttachmentDragManager
	 *    <strong>Properties</strong>
	 *    dropTarget="null"
	 *    targetPath="null"
	 *  /&gt;
	 *  </pre>
	 *
	 *  @see spark.skins.spark.List
	 *
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class AttachmentDropManager
	{
		//--------------------------------------------------------------------------
		//
		//  Class constants
		//
		//--------------------------------------------------------------------------

		/**
		 *  The <code>AttachmentDragManager.DRAG_DROP_ATTACHMENTS</code> constant defines the value of the 
		 *  <code>format</code> property for the Attachment drag data.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		static public const DRAG_DROP_ATTACHMENTS:String = "attachments";

		/**
		 *  The <code>AttachmentDragManager.DRAG_DROP_ITEMS</code> constant defines the value of the 
		 *  <code>format</code> property for the default set of List drag and drop items.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		static public const DRAG_DROP_ITEMS:String = "itemsByIndex";

		//--------------------------------------------------------------------------
		//
		//  Properties
		//
		//--------------------------------------------------------------------------
		
		//----------------------------------
		//  dropTarget
		//----------------------------------

		/**
		 *  @private
		 * 	Storage for the target UI Component for the drop.
		 */
		private var _dropTarget:UIComponent = null;
		
		[Inspectable(defaultValue="null")]
		
		/**
		 *  This contains the UI Component registered by the caller as the drop target.
		 *
		 *  <p>If you set this property, the UI component that is set is enabled as 
		 *  a drop target for attachments. If this is set to null or to a different 
		 *  UI component, the previous UI component is unregistered as a drop target. </p>
		 *
		 *  @default null
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get dropTarget():UIComponent
		{
			return _dropTarget;
		}
		
		/**
		 *  @private
		 */
		public function set dropTarget(value:UIComponent):void
		{
			if (value == _dropTarget)
				return;
			
			// Remove event handlers from the old drop target and apply it to the new one
			if (_dropTarget)
			{
				dropTarget.removeEventListener(DragEvent.DRAG_ENTER, dragEnterHandler, false);
				dropTarget.removeEventListener(DragEvent.DRAG_EXIT, dragExitHandler, false);
				dropTarget.removeEventListener(DragEvent.DRAG_OVER, dragOverHandler, false);
				dropTarget.removeEventListener(DragEvent.DRAG_DROP, dragDropHandler, false);
			}
			_dropTarget = value;
			if (_dropTarget)
			{
				dropTarget.addEventListener(DragEvent.DRAG_ENTER, dragEnterHandler, false, EventPriority.DEFAULT_HANDLER);
				dropTarget.addEventListener(DragEvent.DRAG_EXIT, dragExitHandler, false, EventPriority.DEFAULT_HANDLER);
				dropTarget.addEventListener(DragEvent.DRAG_OVER, dragOverHandler, false, EventPriority.DEFAULT_HANDLER);
				dropTarget.addEventListener(DragEvent.DRAG_DROP, dragDropHandler, false, EventPriority.DEFAULT_HANDLER);
			}
		}

		//----------------------------------
		//  targetPath
		//----------------------------------
		
		/**
		 *  @private
		 * 	Storage for the drop folder.
		 */
		private var _targetPath:IAttachment = null;
		
		[Inspectable(defaultValue="null")]
		
		/**
		 *  This contains the target folder to which the attachment will be added.
		 * 
		 *  <p>If this value is null then the attachment is added to the root folder.</p>
		 *
		 *  @default null
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get targetPath():IAttachment
		{
			return _targetPath;
		}
		
		/**
		 *  @private
		 */
		public function set targetPath(value:IAttachment):void
		{
			if (value == _targetPath)
				return;

			_targetPath = value;
		}

		/** 
		 *  @private
		 *  The navigator's CollectionManager instance. All host interaction should go through this
		 *  instance, rather than the INavigatorHost directly.
		 */
		private const collectionManager:CollectionManager = CollectionManager.instance;
		
		//--------------------------------------------------------------------------
		//
		//  Constructor
		//
		//--------------------------------------------------------------------------
		
		/**
		 *  Constructor.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function AttachmentDropManager()
		{
		}
		
		//--------------------------------------------------------------------------
		//
		//  Methods
		//
		//--------------------------------------------------------------------------
		
		//--------------------------------------------------------------------------
		//
		//  Drop methods
		//
		//--------------------------------------------------------------------------
		
		/**
		 *  Handles <code>DragEvent.DRAG_ENTER</code> events.  This method
		 *  determines if the DragSource object contains valid elements and uses
		 *  the <code>DragManager.showDropFeedback()</code> method to set up the 
		 *  UI feedback.
		 *
		 *  @param event The DragEvent object.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function dragEnterHandler(event: DragEvent): void
		{
			if (event.isDefaultPrevented())
				return;
			
			if (collectionDropEnabled(event))
			{
				if (event.type == DragEvent.DRAG_ENTER)
					DragManager.acceptDragDrop(dropTarget);
				
				dropTarget.mx_internal::drawFocusAnyway = true;
				dropTarget.drawFocus(true);
				
				// Notify manager we can drop. Platform drags always provide copy feedback
				if (event.dragInitiator)
					// TODO: Support copy. Needs changes on the framework side
					DragManager.showFeedback(DragManager.MOVE);
				else
					// Platform drag
					DragManager.showFeedback(DragManager.COPY);
			}
			else
			{
				// Hide focus
				dropTarget.drawFocus(false);
				dropTarget.mx_internal::drawFocusAnyway = false;
				
				DragManager.showFeedback(DragManager.NONE);
			}
		}

		/**
		 *  Handles <code>DragEvent.DRAG_OVER</code> events. This method
		 *  determines if the DragSource object contains valid elements and uses
		 *  the <code>showDropFeedback()</code> method to set up the UI feedback. 
		 *
		 *  @param event The DragEvent object.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function dragOverHandler(event: DragEvent): void
		{
			dragEnterHandler(event);
		}
		
		/**
		 *  Handles <code>DragEvent.DRAG_EXIT</code> events. This method hides
		 *  the UI feedback by calling the <code>hideDropFeedback()</code> method.
		 *
		 *  @param event The DragEvent object.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function dragExitHandler(event:DragEvent):void
		{
			if (event.isDefaultPrevented())
				return;
			
			// Hide focus
			dropTarget.drawFocus(false);
			dropTarget.mx_internal::drawFocusAnyway = false;
		}

		/**
		 *  Handles <code>DragEvent.DRAG_DROP events</code>. 
		 *  
		 *  @param event The DragEvent object.
		 *
		 *  @see mx.utils.ObjectUtil
		 *  @see mx.utils.UIDUtil
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function dragDropHandler(event:DragEvent):void
		{
			if (event.isDefaultPrevented())
				return;
			
			// Hide focus
			dropTarget.drawFocus(false);
			dropTarget.mx_internal::drawFocusAnyway = false;
			
			// Set the target path for the drop. Note that the target path
			// for the components need not be the current folder stored inside
			// collection manager
			// Collection drop manager cannot be null as collectionDropEnabled
			// would be false in that case
			collectionManager.collectionDropManager.targetPath = targetPath;
			
		}
		
		/**
		 *  @private
		 *  Helper function that determines whether drop is enabled.
		 *  
		 *  @param event The DragEvent object.
		 * 
	     *  @return It returns <code>true</code> when all the following are true:
		 *  1) collectionDropManager is not null,
		 *  2) drop target is enabled, 
		 *  3) host is in Authoring mode, and
		 *  4) drag data contains attachments.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function collectionDropEnabled(event: DragEvent): Boolean
		{
			// Document Permission calls are not checked here as they are hard requirements enforced 
			// by Acrobat platform drop handler. Platform drop handler will not call the dragenter,
			// dragover, dragexit, and dragdrop even in the following conditions
			// - EF/Create and EF/Import is not available for copy operations
			// - EF/Modify is not available for move operations
			// The checks done here are navigator specific requirements
			return ((dropTarget.enabled) && (collectionManager.collectionDropManager) 
				&& (collectionManager.authoringMode) && (event.dragSource.hasFormat(DRAG_DROP_ATTACHMENTS)));
		}
	}
}