/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.managers
{
	import bridge.logging.BridgeLogger;
	
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.api.ICollectionDropManager;
	import com.adobe.portfolio.api.IField;
	import com.adobe.portfolio.api.INavigatorHost;
	import com.adobe.portfolio.api.RightClickEvent;
	import com.adobe.portfolio.components.supportClasses.ColorPalette;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.events.JobQueueEvent;
	import com.adobe.portfolio.properties.CommonProperties;
	import com.adobe.portfolio.utils.AttachmentCollection;
	import com.adobe.portfolio.utils.JobQueue;
	import com.adobe.portfolio.utils.LogHelper;
	import com.adobe.portfolio.utils.SortUtil;
	
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.TimerEvent;
	import flash.system.ApplicationDomain;
	import flash.system.Capabilities;
	import flash.text.Font;
	import flash.utils.Timer;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.collections.ArrayCollection;
	import mx.collections.IList;
	import mx.collections.ListCollectionView;
	import mx.collections.Sort;
	import mx.collections.SortField;
	import mx.events.CollectionEvent;
	import mx.events.CollectionEventKind;
	import mx.events.PropertyChangeEvent;
	import mx.managers.IFocusManagerComponent;
	
	import spark.components.Application;
	
	/**
	 *  Dispatched after various updates to the navigator host.
	 *
	 *  @eventType com.adobe.portfolio.events.NavHostEvent.NAV_HOST
	 * 
	 *  @see CollectionManagerEvent
	 */
	[Event(name="navHost", type="com.adobe.portfolio.events.CollectionManagerEvent")]
	
	/**
	 *  Dispatched when a right click event is received from the host.
	 *
	 *  @eventType com.adobe.portfolio.api.RightClickEvent.RIGHT_CLICK
	 */
	[Event(name="rightClick", type="com.adobe.portfolio.api.RightClickEvent")]
	
	/**
	 * CollectionManager acts as a layer between navigator components and an INavigatorHost
	 * for managing the collection of files in the Portfolio.  All host interactions 
	 * should go through a CollectionManager instance, rather than through the INavigatorHost directly.
	 * 
	 *  @see  flash.events.EventDispatcher
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *	
	 */
	public class CollectionManager extends EventDispatcher
	{
		/**
		 *  An array representing the Portfolio API v9.5.
		 *
		 * 	@see #apiIsSupported
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const API_VERSION_9_5:Array = [9, 5, 0, 0];

		/**
		 *  An array representing the Portfolio API v10.0.
		 *
		 *  @see #apiIsSupported
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const API_VERSION_10:Array = [10, 0, 0, 0];
		
		private static var _instance:CollectionManager = null;
		private static var _token:Object;
		
		private var savedRootItems:Array;
		private var suspendFocusAndSelectionEventsCounter:uint;
		private var suspendedOldFocus:IAttachment;
		private var _enabled:Boolean;
		private var _isAdobeHost:Boolean = false;
		private var _isNavPreviewCardPreview:Boolean;
		private var _rootItemsJobQueue:JobQueue;
		
		/**
		 *  A cache of the newly ordered list. This needs to be sent over the bridge for reordering.
		 *  Since we can reorder between folders and hence can be in different lists alltogether, 
		 *  this has to be cached centrally.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public var newOrderedList:IList;
		
		/**
		 * A static instance that can be shared by Flex components.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function get instance():CollectionManager
		{
			if (_instance == null) {
				_token = new Object();
				_instance = new CollectionManager(_token);
			} 
			return _instance;
		}
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function CollectionManager(token:Object)
		{
			super();
			if (_token != token)
				throw new Error("can't construct this object");
			_enabled = false;
			_currentItems.addEventListener(CollectionEvent.COLLECTION_CHANGE, onCurrentItemsChange);
		}
		
		// This string must exactly match the one in
		// BootStrapLoader/src/com/adobe/AdobeHost.as.
		private static const AdobeCopyrightString:String = 
			"©2010 Adobe Systems Incorporated.  All rights reserved.";

		/**
		 *  Determines if the provided Application is an Adobe-supported host.
		 * 
		 *  @review
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function queryAdobeHost(app:Application):Boolean
		{
			try
			{
				// If the host is Acrobat or Reader it is by definition an Adobe host.
				// I don't think there is any way to spoof this, since it is a 
				// Player class.
				if (Capabilities.isEmbeddedInAcrobat)
					return true;
				
				// Standalone hosting environments qualify as an Adobe host if they
				// have a Copyright class that returns the expected text.
				// The class will only exist if the navigator's loader forces it
				// into the ApplicationDomain. By design a navigator project will have
				// no compile-time access to the class to prevent accidental inclusion
				// in the navigator SWF itself.
				var c:Class = ApplicationDomain.currentDomain.getDefinition("com.adobe::AdobeHost") as Class;
				var text:String = c['text'];
				if (text == AdobeCopyrightString)
					return true;
			}
			catch (err:Error)
			{
				// Most likely the class doesn't exist.
			}

			return false;
		}

		/**
		 *  Returns the INavigatorHost's font
		 * 
		 * 	@see INavigatorHost#navigatorFont
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function get applicationFont():Font
		{
			if (_host == null) {
				return null;
			}
			return _host.navigatorFont;
		}
		
		/**
		 *  Dispatches a <code>CollectionManagerEvent</code> indicating that a skin part has been removed.
		 *  This allows skin parts to listen for events from the collection manager so that cleanup
		 *  (removing event listeners) can be performed.
		 * 
		 * 	@see CollectionManagerEvent#PART_REMOVED_KIND
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function dispatchPartRemoved(oldPart:Object):void
		{
			dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.PART_REMOVED_KIND, oldPart));
		}
		
		/**
		 *  Dispatches a <code>CollectionManagerEvent</code> indicating that a navigator preview should be displayed.
		 * 
		 * 	@see CollectionManagerEvent#NAVIGATOR_PREVIEW_KIND
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function dispatchNavigatorPreview(attachment:IAttachment):void
		{
			dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.NAVIGATOR_PREVIEW_KIND, null, attachment, false, true));
		}
		
		/**
		 * Invokes the appropriate preview behavior for both file and folder attachments.
		 */
		public function handlePreview(attachment:IAttachment):void 
		{
			if (attachment == null)
				return;	
			
			if (attachment.isFolder) {
				currentFolder = attachment;
				return;
			} 
			
			dispatchNavigatorPreview(attachment);
		}
		
		/**
		 *  A UI component that's suitable for receiving focus by default.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public var defaultFocusItem:IFocusManagerComponent;
		
		/** Calling this function tells the navigator to ignore focus changes.
		 *  This is useful when a series of events need to occur before the actual focus change.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function suspendFocusAndSelectionEvents():void
		{
			if (suspendFocusAndSelectionEventsCounter == 0) {
				suspendedOldFocus = focusAttachment;
			}
			++suspendFocusAndSelectionEventsCounter;
		}
		
		/**
		 *  This function turns off the previously called suspension
		 *  of Focus and Selection Events. 
		 *  There may be multiple calls to suspendFocusAndSelectionEvents(),
		 *  so the navigator maintains a count of the number of such calls.
		 *  For each call to suspendFocusAndSelectionEvents() you must have a
		 *  matching call to unsuspendFocusAndSelectionEvents() before the focus
		 *  change will be handled.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *	
		 */				
		public function unsuspendFocusAndSelectionEvents():void
		{
			if (suspendFocusAndSelectionEventsCounter == 0) {
				LogHelper.logger.error("Mismatched unsuspendFocusAndSelectionEvents call");
			} else {
				--suspendFocusAndSelectionEventsCounter;
				if (suspendFocusAndSelectionEventsCounter == 0) {
					handleNewFocus(suspendedOldFocus);
					dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.SELECTION_DID_CHANGE_KIND));
					suspendedOldFocus = null;
				}
			}
		}
		/**
		 *  Returns true if Focus and Selection Events are currently suspended;
		 *  false otherwise.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *	
		 */				
		public function focusAndSelectionEventsSuspended():Boolean
		{
			return (suspendFocusAndSelectionEventsCounter > 0);
		}

		private var _focusIndex:int = -1;

		[Bindable]
		/**
		 *  Indicates the index of the current focus attachment. 
		 *  focusIndex is used to keep track of what IAttachment
		 *  is currently in focus and what its next and previous attachments are.
		 *  focusIndex can be set explicitly by <code>set focusIndex()</code>, or indirectly
		 *  by <code>handleNewFocus()</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function get focusIndex():int
		{
			return _focusIndex;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function set focusIndex(value:int):void
		{
			_focusIndex = value;
		}

		private var _focusAttachment:IAttachment = null;

		[Bindable]
		/**
		 * Indicates the current IAttachment considered to have focus. This may be null.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get focusAttachment():IAttachment
		{
			return _focusAttachment;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function set focusAttachment(val:IAttachment):void
		{
			var oldVal:IAttachment = _focusAttachment;
			_focusAttachment = val;
			if (!focusAndSelectionEventsSuspended()) {
				focusIndex = -1;
				handleNewFocus(oldVal);
			}
		}
		
		private var _nextAttachment:IAttachment = null;

		[Bindable]
		/**
		 * Indicates the attachment following focusAttachment in the current sort order. This may be null.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get nextAttachment():IAttachment
		{
			return _nextAttachment;
		}

		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function set nextAttachment(value:IAttachment):void
		{
			_nextAttachment = value;
			if (_host != null) {
				_host.setNextPrevious(nextAttachment, prevAttachment);
			}
		}
		

		private var _prevAttachment:IAttachment = null;
	
		[Bindable]
		/**
		 * Indicates the attachment following focusAttachment in the current sort order. This may be null.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get prevAttachment():IAttachment
		{
			return _prevAttachment;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */		
		public function set prevAttachment(value:IAttachment):void
		{
			_prevAttachment = value;
			if (_host != null) {
				_host.setNextPrevious(nextAttachment, prevAttachment);
			}
		}

		/**
		 * An event listener that gives focus to the previous attachment, if it exists.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function onGotoPreviousAttachment(evt:Event):void
		{
			if (prevAttachment != null) {
				setSelectedItems([prevAttachment]);
			}
		}
		
		/**
		 * An event listener that gives focus to the next attachment, if it exists.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function onGotoNextAttachment(evt:Event):void
		{
			if (nextAttachment != null) {
				setSelectedItems([nextAttachment]);
			}
		}

		private var _lastFolder:IAttachment = null;
		
		[Bindable]
		/**
		 *  Indicates the current folder whose contents are being viewed,
		 *  or null if the root is currently being viewed.
		 *  The items of the folder that is returned correspond
		 *  to the currentItems property of this folder.  When a new
		 *  current folder is set, the SelectedItems array is set to [], since
		 *  the current folder has been changed, so there are no selected items yet.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get currentFolder():IAttachment
		{
			// TODO Once an ActionScript preview/debugging solution is in place that implements
			// an INavigatorHost, we should remove all (_host == null) functionality.
			return (_host != null) ? _host.currentFolder : null;
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function set currentFolder(val:IAttachment):void
		{
			if (_host == null || _host.currentFolder == val)
				return;
			
			// Clear the selection when changing folders
			setSelectedItems([]);
			
			_host.currentFolder = val;
		}
		
		// There are two levels of indirection for these items to allow the lists
		// to be hidden from clients until enabled is set to true
		private var _currentItems:AttachmentCollection = new AttachmentCollection();		
		private var _rootItems:AttachmentCollection = new AttachmentCollection();
		private var __rootItems:AttachmentCollection;
		private var __currentItems:AttachmentCollection;
		
		[Bindable]
		/**
		 * Indicates the items currently being viewed. 
		 * The items in this collection are contained in the object
		 * specified by <code>currentFolder</code>.
		 * 
		 * This property will be null until the instance has been enabled.
		 * 
		 * @see #enabled
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get currentItems():AttachmentCollection
		{
			return __currentItems;
		}
		/**
		 *  @private 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */		
		private function set currentItems(value:AttachmentCollection):void
		{
			__currentItems = value;
		}
		
		/**
		 * Returns true if the dataProvider for a list matches the CollectionMgr's currentItems.
		 * Things like focus and selection should only be propagated to the CollectionMgr if the
		 * dataProvider is current.
		 */
		public function itemsAreCurrent(dataProvider:IList):Boolean
		{
			// TODO:  This method doesn't always return the right answer.. we
			// should figure out how to eliminate using it.
			//
			if (dataProvider == null)
				return false;
			
			if (dataProvider == currentItems) {
				return true;
			}
			if (currentItems == null) {
				return false;
			}
			
			// sorted collections sometimes don't find items when they should.  use the backing list if available.
			var dataProviderCollectionView:ListCollectionView = dataProvider as ListCollectionView;
			if (dataProviderCollectionView && dataProviderCollectionView.sort)
				dataProvider = dataProviderCollectionView.list;
			
			// It's possible that currentItems has changed but the dataProvider hasn't been updated yet
			// Check for the focusAttachment just in case
			if ((focusAttachment != null) && (dataProvider.getItemIndex(focusAttachment) >= 0)) {
				return true;
			}

			if (dataProvider.length != currentItems.length) {
				if (selectedItems && (selectedItems.length > 0) && (dataProvider.getItemIndex(selectedItems[0]) >= 0))
					return true;
				return false;
			}
			
			// If the first items are the same, we'll assume our is current
			if ((dataProvider.length > 0) &&
				(currentItems.length > 0) &&
				(dataProvider.getItemAt(0) == currentItems.getItemAt(0)))
			{
				return true;
			}
			return false;
		}
		
		[Bindable]
		/**
		 * The root items in a collection.
		 * 
		 * This property will be null until the instance has been enabled.
		 * 
		 * @see #enabled
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get rootItems():AttachmentCollection
		{
			return __rootItems;
		}

		/**
		 *  @private 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */				
		private function set rootItems(value:AttachmentCollection):void {
			__rootItems = value;
			if (queueRootItems) {
				requeueList(rootItemsJobQueue, __rootItems);
			} else {
				// Make sure the queue events still get handled properly
				dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.ROOT_ITEMS_QUEUE_BEGIN_KIND));
				dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.ROOT_ITEMS_QUEUE_END_KIND));
			}
		}
		
		/**
		 * If true, then attachments will be added to the rootItems collection incrementally,
		 * rather than all at once. This can improve startup performance for navigators whose
		 * startup performance is dependent on the number of root items.
		 */
		public var queueRootItems:Boolean = true;
		
		/**
		 *  Trickles attachments into the root collection in the initialization phase of the Navigator.  
		 * 
		 *  <p>This gives Navigators the chance to update the UI and provide for scrolling, etc. while the attachments
		 *  are being added.</p>
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */				
		protected function requeueList(jobUtil:JobQueue, items:IList, timerInterval:Number = 20):void 
		{
			if (!items) {
				if (rootItemsJobQueue && rootItemsJobQueue.hasJobs()) {
					rootItemsJobQueue.killAllJobs();
				}
				return;
			}
			
			dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.ROOT_ITEMS_QUEUE_BEGIN_KIND));
			if (items.length == 0)
				return;
			
			if (!savedRootItems)
				savedRootItems = items.toArray();
			
			items.removeAll();
			if (!rootItemsJobQueue) {
				_rootItemsJobQueue= new JobQueue("rootItemsQueue", true, true, timerInterval);
				rootItemsJobQueue.addEventListener(JobQueueEvent.QUEUE_IS_EMPTY,
					function (e:Event):void {
						if (items.length == savedRootItems.length) {
							savedRootItems = null;
							dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.ROOT_ITEMS_QUEUE_END_KIND));
							rootItemsJobQueue.killAllJobs();
							_rootItemsJobQueue = null;
						} else {
							rootItemsJobQueue.startJobTimer();
						}
					}
				);
			}
			for (var i:int=0; i < savedRootItems.length; ++i) {
				rootItemsJobQueue.pushJob(i, i,
					function (key:Object, job:Object):void {
						items.addItem(savedRootItems[job]);
					}
				);
			}
		}
		
		
		/**
		 *  Takes the IList of items passed to it and make these 
		 *  the list of currentItems.  It broadcasts appropriate events
		 *  so that clients can respond to the change in current items.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function setCurrentItems(value:IList):void
		{
			dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.ITEMS_WILL_CHANGE_KIND, _currentItems, value));
			
			// We use the contents of value to create a collection that we can manipulate independently
			_currentItems.hostItems = value;
			
			dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.ITEMS_DID_CHANGE_KIND, null, _currentItems));
		}
		
		/**
		 *  A listener that adjusts the current focus when currentItems changes.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onCurrentItemsChange(evt:CollectionEvent):void
		{
			var shouldHandleFocus:Boolean = true;
			
			switch(evt.kind)
			{
				case CollectionEventKind.REFRESH:
					// Called after a sort, focus may need to be handled.
					break;
				
				case CollectionEventKind.MOVE:
				case CollectionEventKind.UPDATE:
					//do nothing
					// TODO Not refreshing items. This is not needed after a preview update. Actions triggering an update should not require a refresh.
					shouldHandleFocus = false;
					break;
				
				case CollectionEventKind.ADD:
					// For first time add, set the sort field.
					if(sort == null || sort.length == 0)
					{
						if (_host) 
						{	
							sort = _host.collection.sort;								
						}
					}
					if (!handleCurrentFolderSuspended &&
						(evt.items != null) &&
						(evt.items.length > 0))
					{
						// If the host adds multiple files one at a time, only the last one will
						// wind up selected. We'd need to use some timing heuristics if we want to
						// select them all in that case.
						// Fix bug 2722975: Platform Preview scrolls through previewing several PDFs when A8 Packages are opened.
						// Limit the update of the focus and selection, after adding an item, to authoring mode.
						if (authoringMode) {
							focusAttachment = evt.items[0];
							setSelectedItems(evt.items);
						}
					}
					break;
				
				case CollectionEventKind.REMOVE:
					// Focus may need to be handled, so leave shouldHandleFocus set to true.
					var relevantAttachment:IAttachment = focusAttachment;
					if ((relevantAttachment == null) &&
						(selectedItems != null) &&
						(selectedItems.length > 0))
					{
						relevantAttachment = selectedItems.getItemAt(0) as IAttachment;
					}
					if (relevantAttachment != null) {
						for each (var removedItem:Object in evt.items) {
							if (relevantAttachment == removedItem) {
								if ((evt.location >= 0) && (_currentItems.length > 0)) {
									if (evt.location < _currentItems.length) {
										focusAttachment = _currentItems[evt.location];
									} else {
										focusAttachment = _currentItems[_currentItems.length - 1];
									}
									setSelectedItems([focusAttachment]);
								} else {
									// TODO What should we set the focus to?
									focusAttachment = null;
									setSelectedItems([]);
								}
								break;
							}
						}
					}
					break;
				
				default:
					break;	
			}
			
			if (shouldHandleFocus) {
				if (handleCurrentFolderSuspended) {
					// Reset our timer. It will unsuspend focus and selection events after a short amount
					// of time, unless we reset the timer by calling it again first.
					resetCurrentFolderTimer();
				} 
				else {
					handleNewFocus(suspendedOldFocus);
				}
			}
		}
		
		/**
		 *  Returns the index of <code>attachment</code> in <code>currentItems</code>,
		 *  or -1 if currentItems does not contain <code>attachment</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function getCurrentIndex(attachment:IAttachment):int
		{
			if (attachment != null) {
				return _currentItems.getItemIndex(attachment);
			} else {
				return -1;
			}
		}
		
		/**
		 *  Returns the file name of the collection,
		 *  or the empty string if the host or the collection is null.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get fileName():String
		{
			if ((_host != null) && (_host.collection != null)) {
				return _host.collection.fileName;
			} else {
				return "";
			}
		}
		
		/** 
		 * Marks the host collection as "dirty", which enables Save in the application UI.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function setDirty():void
		{
			if ((_host != null) && (_host.collection != null)) {
				_host.collection.setDirty();
			}
		}
		
		
		/**
		 * Asynchronously returns the state of the host collection dirty flag.
		 * 
		 * 	@see ICollection#getDirty
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function getDirty(response:Function):void
		{
			if ((_host != null) && (_host.collection != null)) {
				_host.collection.getDirty(response);
			}
			
		}
		
		/**
		 *  Returns <code>true</code> if the given category/permission combination is allowed in the host,
		 *  or <code>false</code> otherwise.
		 * 
		 *  @see IPermissions#check
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function permissionCheck(category:String, permission:String):Boolean
		{
			if ((_host != null) &&
				(_host.collection != null) &&
				(_host.collection.permissions != null))
			{
				return _host.collection.permissions.check(category, permission);
			} else {
				return false;
			}
		}
		
		[Bindable]
		/**
		 * The current color palette as specified by the viewing application.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public var colorPalette:ColorPalette = new ColorPalette();
		
		private var _fixedNavigatorColorPalette:Object;
		
		/**
		 * Indicates the fixed navigator color palette, if there is one.
		 * Some navigators (such as Details) want to always have the same look independent
		 * of the host color palette changed, unless a high contrast color palette is used.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get fixedNavigatorColorPalette():Object
		{
			return _fixedNavigatorColorPalette;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function set fixedNavigatorColorPalette(value:Object):void
		{
			_fixedNavigatorColorPalette = value;
			setColorPalette(_host ? _host.navigatorColorPalette : null);
		}
		
		/**
		 * Conditionally sets the Navigator color palette.
		 * If the fixed Navigator palette is set, use it,
		 * unless the host color palette is using high contrast mode.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function setColorPalette(navHostPalette:Object):void
		{
			var highContrast:Boolean = false;
			if (_host && _host.navigatorColorPalette &&
				_host.navigatorColorPalette.hasOwnProperty("isHighContrast"))
				highContrast = _host.navigatorColorPalette["isHighContrast"] as Boolean;
			if (_fixedNavigatorColorPalette && !highContrast)
				colorPalette.setColorPalette(_fixedNavigatorColorPalette);
			else
				colorPalette.setColorPalette(navHostPalette);
		}
		
		/**
		 *  Updates the color palette used by the viewing application
		 *  to be the palette passed to updateColorPalette().
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function updateColorPalette(paletteObject:Object):void
		{
			if (_host == null) {
				return;
			}
			if (!permissionCheck("Document", "Modify")) {
				CONFIG::logging{
					LogHelper.logger.warn("Insufficient permissions to change the navigatorColorPalette");
				}
				return;
			}
			var newPalette:Object = (paletteObject is ColorPalette) ? (paletteObject as ColorPalette).generateRawPalette() : paletteObject; 
			if (_host.navigatorColorPalette == null) {
				_host.navigatorColorPalette = newPalette;
				return;
			}
			if (_host.navigatorColorPalette.hasOwnProperty("isHighContrast") &&
				_host.navigatorColorPalette["isHighContrast"])
			{
				CONFIG::logging{
					LogHelper.logger.warn("Unable to update the high-contrast color palette");
				}
				return;
			}
			
			newPalette = new Object();
			var newColor:String;
			
			// Build an object with the current colors...
			for (newColor in _host.navigatorColorPalette) {
				newPalette[newColor] = _host.navigatorColorPalette[newColor];
			}
			
			// ...then add the updated colors
			for (newColor in paletteObject) {
				if (newColor == "isHighContrast") {
					// Don't include the high-constrast value
					continue;
				}
				newPalette[newColor] = paletteObject[newColor];
			}
			_host.navigatorColorPalette = newPalette;
		}
		
		/** Internal representation of the fields array referenced by the fields getter and setter.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private var _fields:Array = [];
		
		[Bindable]
		/**
		 *  The array of fields displayed in the Files view of the Portfolio
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get fields():Array
		{
			return _fields;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set fields(val:Array):void
		{
			_fields = val;
		}
		
		/** Internal representation of the sort array referenced by the sort getter and setter.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
	 	 */
		private var _sort:Array = [];
		
		[Bindable]
		/**
		 *  An ISort array the indicates the sort order to use.
		 * 
		 * 	@see ISort
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function get sort():Array
		{
			return _sort;
		}
		
		/** 
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function set sort(val:Array):void
		{
			if (!SortUtil.compareSortArrays(val, _sort))
			{
				_sort = val;
				var newSort:Sort = SortUtil.createSortFieldObject(val);
				_currentItems.sort = newSort;
				_currentItems.refresh();
				
				_rootItems.sort = newSort;
				_rootItems.refresh();
				
				// If authoring mode and host present, change host's sort value
				if (_host && authoringMode && permissionCheck("Document", "Modify"))
				{
					_host.collection.sort = SortUtil.createSortArray(_currentItems.sort);
				}
				
				// Lists tend to track selection by index, so broadcast to make sure the
				// index is right.
				dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.SELECTION_DID_CHANGE_KIND));
			}
		}

		/** Internal representation of the Boolean referenced by the authoringMode getter and setter.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private var _authoringMode:Boolean = false;
		
		[Bindable]
		/**
		 * A Boolean indicating whether or not the host is in authoring mode.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get authoringMode():Boolean
		{
			return _authoringMode;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		private function set authoringMode(value:Boolean):void
		{
			if (value && !isAdobeHost) {
				LogHelper.logger.error("We shouldn't be setting authoring mode to true in non-Adobe environments");
				value = false;
			}
			if (value != _authoringMode) {
				_authoringMode = value;
				if ((_authoringMode) && (_applicationFontList == null))
					_applicationFontList = _host.applicationFontList;
			}
		}
		/**
		 * Indicates the <code>ICollectionDropManager</code> exposed by the host.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get collectionDropManager():ICollectionDropManager
		{
			return (_host != null) ? _host.collectionDropManager : null;
		}
		
		/**
		 *  Sets the host to be used by this class in the context of the given application.
		 * 	Applications that are not Adobe-supported may have limited functionality.
		 * 
		 * 	@see #host
		 * 	@see #queryAdobeHost
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function setHostFromApplication(val:INavigatorHost, app:Application):Boolean
		{
			_isAdobeHost = queryAdobeHost(app);
			if (!isAdobeHost) {
				authoringMode = false;
			}
			host = val;
			return isAdobeHost;
		}
		
		private var _host:INavigatorHost = null;
		
		[Bindable]
		/**
		 *  Indicates the current INavigatorHost used to communicate with the viewing application.
		 *  When the navigator host changes, the old host's event listeners are removed and
		 *  the following new event listerners are added to the new navigator host:
		 * 	"currentFolderChanged", "navigatorColorPaletteChanged",
		 *  RightClickEvent.RIGHT_CLICK, "needsSave", "permissionsChanged",
		 *  PropertyChangeEvent.PROPERTY_CHANGE, PropertyChangeEvent.PROPERTY_CHANGE,
		 *  "applicationFontListChanged".  Also, events of type CollectionManagerEvent.HOST_CHANGED_KIND
		 *  and CollectionManagerEvent.PERMISSIONS_CHANGED_KIND are dispatched so that clients can properly 
		 *  handle these events.
		 * 
		 * 	@see #setHostFromApplication
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */ 
		public function get host():INavigatorHost
		{
			return _host;
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function set host(val:INavigatorHost):void
		{
			var oldHost:INavigatorHost = _host;
			if (oldHost != null) {
				if (selectionWatcher != null) {
					selectionWatcher.unwatch();
					selectionWatcher = null;
				}
				if (embeddedFontWatcher != null) {
					embeddedFontWatcher.unwatch();
					embeddedFontWatcher = null;
				}
				oldHost.removeEventListener("currentFolderChanged", onCurrentFolderChanged);
				oldHost.removeEventListener("navigatorColorPaletteChanged", onHostPropertyChanged);
				oldHost.removeEventListener(RightClickEvent.RIGHT_CLICK, onRightClick);
				oldHost.collection.removeEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onCollectionPropertyChanged);
				oldHost.collection.removeEventListener("needsSave", onCollectionPropagateEvent);
				oldHost.collection.removeEventListener("permissionsChanged", onCollectionPropagateEvent);
				if (apiIsSupported(API_VERSION_10)) {
					oldHost.removeEventListener("authoringModeChanged", onHostPropertyChanged);
				}
				_rootItems.hostItems = null;
			}
			_host = val;
			if (_host != null) {
				selectionWatcher = BindingUtils.bindSetter(onSelectionChanged, _host, "selection");
				_host.addEventListener("currentFolderChanged", onCurrentFolderChanged);
				_host.addEventListener("navigatorColorPaletteChanged", onHostPropertyChanged);
				_host.addEventListener(RightClickEvent.RIGHT_CLICK, onRightClick);
				_host.collection.addEventListener("needsSave", onCollectionPropagateEvent);
				_host.collection.addEventListener("permissionsChanged", onCollectionPropagateEvent);
				_host.collection.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onCollectionPropertyChanged);
				embeddedFontWatcher = BindingUtils.bindSetter(setEmbeddedFontList, _host.collection, "embeddedFontList");
				
				// We only support NavFramework authoring for versions >= 10
				if (apiIsSupported(API_VERSION_10) && isAdobeHost) {
					_host.addEventListener("authoringModeChanged", onHostPropertyChanged);
					authoringMode = _host.authoringMode;
				} else {
					authoringMode = false;
				}
				// Wait for updates on application font list
				_host.addEventListener("applicationFontListChanged", onHostPropertyChanged);
				
				dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.HOST_CHANGED_KIND, oldHost, _host));
					
				setColorPalette(_host.navigatorColorPalette);
				fields = host.collection.fields;
				if ((host.collection.sort == null) || (host.collection.sort.length == 0)) {
					if ((fields != null) && (fields.length > 0)) {
						var firstField:IField = fields[0];
						if (firstField != null) {
							var newSort:Sort = new Sort();
							newSort.fields = [new SortField(firstField.name, true)];
							sort = SortUtil.createSortArray(newSort);
						}
					}
				} else {
					sort = host.collection.sort;
				}
				handleCurrentFolder(_host.currentFolder);
				_rootItems.hostItems = _host.collection.items;
			} else {
				onSelectionChanged(null);
				authoringMode = false;
				embeddedFontList = null;
				dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.HOST_CHANGED_KIND, oldHost, _host));
				
				colorPalette.reset();
				handleCurrentFolder(null);
			}
			
			// Make sure clients handle any potential permission changes
			dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.PERMISSIONS_CHANGED_KIND));
		}
		
		/**
		 * The internal representation of the ArrayCollection referenced by the selectedItems getter.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private var _selectedItems:ArrayCollection = new ArrayCollection();
		
		/**
		 *  Indicates the currently selected items in the collection.
		 * 
		 * 	@see #setSelectedItems
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get selectedItems():ArrayCollection
		{
			return _selectedItems;
		}
		
		/**
		 *  Sets the selected items to an array of IAttachment objects.
		 * 
		 * 	@see #selectedItems
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function setSelectedItems(val:Array):void
		{
			if (_host != null) {
				_host.selection = val;
			} else {
				_selectedItems = new ArrayCollection();
			}
		}
		
		/** 
		 *  A watcher that defines utility methods to let you 
		 *  define event handlers to handle changes from the host.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private var selectionWatcher:ChangeWatcher;
		/** 
		 *  Listener to handle selection changes from the host.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onSelectionChanged(value:Array):void
		{
			if (_host != null) {
				if (((value == null) || (value.length == 0)) &&
					((_selectedItems == null) || (_selectedItems.length == 0)))
				{
					// Selection was already empty
					return;
				} else if (value.length == _selectedItems.length) {
					// Check to see if the selection has really changed
					var didChange:Boolean = false;
					for (var index:int = 0; index < _selectedItems.length; ++index) {
						if (value[index] != _selectedItems.getItemAt(index)) {
							didChange = true;
							break;
						}
					}
					if (!didChange) {
						return;
					}
				}
				_selectedItems = new ArrayCollection(value);
				if (_selectedItems.length == 1) {
					// We'll get this event when Acrobat goes to the next/previous attachment
					// from platform preview. We need to respond to it so that next/previous are
					// updated properly.
					focusAttachment = _selectedItems.getItemAt(0) as IAttachment; 
				}
			} else {
				_selectedItems.removeAll();
			}
			if (!focusAndSelectionEventsSuspended()) {
				dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.SELECTION_DID_CHANGE_KIND));
			}
		}

		
		/** Listener to handle folder changes from the host.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onCurrentFolderChanged(evt:Event):void
		{
			handleCurrentFolder(currentFolder);
		}

		private var handleCurrentFolderSuspended:Boolean = false;
		private var handleCurrentFolderTimer:Timer;
		
		/**
		 *  A Boolean switch to determine whether to unsuspend focus/selection
		 *  as soon as items are available, or to wait until the items are ready.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private const useCurrentFolderTimer:Boolean = true;
		
		/**
		 *  A function that sets up <code>handleCurrentFolderTimer</code> as needed,
		 *  and resets the timer to unsuspend focus events after a fixed amount of time,
		 *  unless resetCurrentFolderTimer() called again before the fixed amount of time
		 *  elapses.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function resetCurrentFolderTimer():void
		{
			if (handleCurrentFolderTimer == null) {
				var timerComplete:Function = function(event:TimerEvent):void
				{
					if (handleCurrentFolderSuspended) {
						var shouldUnsuspend:Boolean = true;
						if (_currentItems.length > 0) {
							if ((selectedItems != null) && (selectedItems.length > 0)) {
								focusAttachment = selectedItems.getItemAt(0) as IAttachment;
							} else {
								focusAttachment = _currentItems.getItemAt(0) as IAttachment;
								
								// This is one of two code paths that will cause the first
								// item in a list to be selected by default.
								// TODO 	This may overwrite the current user selection on a navigator switch if
								//		we haven't that selection yet.
								setSelectedItems([focusAttachment]);
							}
						} else {
							// We still don't have any current items, so keep waiting
							shouldUnsuspend = false;
						}
						if (shouldUnsuspend) {
							unsuspendFocusAndSelectionEvents();
							handleCurrentFolderSuspended = false;
						}
					} else {
						CONFIG::logging{
							LogHelper.logger.warn("handleCurrentFolderTimer shouldn't have been running if handleCurrentFolderSuspended was false");
						}
					}
				};
				if (!useCurrentFolderTimer) {
					// The folder timer is disabled, so just call its handler and return
					timerComplete(null);
					return;
				}
				handleCurrentFolderTimer = new Timer(300, 1);
				handleCurrentFolderTimer.addEventListener(TimerEvent.TIMER_COMPLETE, timerComplete);
			} else {
				handleCurrentFolderTimer.reset();
			}
			handleCurrentFolderTimer.start();
		}
		
		/** 
		 *  This function adjusts the state of the CollectionManager, given a new current folder.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function handleCurrentFolder(newFolder:IAttachment):void
		{
			// First remove the current focus and selection, so fade outs and such can be handled
			focusAttachment = null;
			
			if ((handleCurrentFolderTimer != null) && handleCurrentFolderTimer.running) {
				handleCurrentFolderTimer.stop()
			}
			if (handleCurrentFolderSuspended) {
				unsuspendFocusAndSelectionEvents();
				handleCurrentFolderSuspended = false;
			}
			
			if (_host == null) {
				setSelectedItems(null);
				return;
			}
			
			// Don't dispatch any more focus/selection events until we're all ready
			suspendFocusAndSelectionEvents();
			
			// We can be smart about which attachment receives focus on a folder change.
			// It's not always the first item.
			var newFocus:IAttachment = null;
			
			// If we're moving to an ancestor of our last folder, then the child of newFolder
			// leading to lastFolder should get focus.
			for (var level:IAttachment = _lastFolder; level != null; level = level.parent) {
				if (level.parent == newFolder) {
					newFocus = level;
					break;
				}
			}
			
			if ((newFolder ==_lastFolder) && (newFolder != null)) {
				// folder hasn't actually changed
			} else {
				_lastFolder = newFolder;
				if (newFolder != null) {
					setCurrentItems(newFolder.children);
				} else if ((_host != null) && (_host.collection != null)) {
					setCurrentItems(_host.collection.items);
				} else {
					setCurrentItems(new ArrayCollection());
				}
			}
			
			var shouldSetSelection:Boolean = (newFocus != null);
			if ((newFocus == null) && (selectedItems != null) && (selectedItems.length > 0)) {
				// We already have a selection, so use it to pick a focus attachment
				newFocus = selectedItems.getItemAt(0) as IAttachment;
				shouldSetSelection = false;
			}
			if ((newFocus == null) && (_currentItems != null) && (_currentItems.length > 0)) {
				newFocus = _currentItems.getItemAt(0) as IAttachment;

				// Our new set of current items already exists, so use the first one as focus
				// This is one of two code paths that will cause the first item in a list to
				// be selected by default.
				// TODO 	This may overwrite the current user selection on a navigator switch if
				//		we haven't that selection yet.
				shouldSetSelection = true;
			}
			if (newFocus != null) {
				// Only set the new focus after current items has been updated.
				focusAttachment = newFocus;
				if (shouldSetSelection) {
					setSelectedItems([newFocus]);
				}
				unsuspendFocusAndSelectionEvents();
			} else {
				// Wait for items to be added before adjusting focus;
				// this will happen in onCurrentItemsChange
				handleCurrentFolderSuspended = true;
				
				if (newFolder != null) {
					// Our new folder might not actually contain any items, so kick off the timer
					// right away so we don't wait around forever
					resetCurrentFolderTimer();
				}
			}
		}
		
		/**
		 *  Listener to handle simple property changes from the host.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onHostPropertyChanged(event:Event):void
		{
			if (_host == null) {
				return;
			}
			switch (event.type) {
				case "navigatorColorPaletteChanged":
					setColorPalette(_host.navigatorColorPalette);
					break;
				case "authoringModeChanged":
					// We only support NavFramework authoring for versions >= 10
					authoringMode = (isAdobeHost && apiIsSupported(API_VERSION_10)) ? _host.authoringMode : false;
					break;
				case "applicationFontListChanged":
					applicationFontList = _host.applicationFontList;
					break;
			}
		}
		
		/**
		 * Listener to propagate right-click events from the host.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onRightClick(event:RightClickEvent):void
		{
			// TODO This should probably be a NavFramework event, but for now we'll just reuse the RightClickEvent
			dispatchEvent(event.clone());
		}

		/** 
		 * This function adjusts the state of the CollectionManager, given a new focused attachment.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function handleNewFocus(oldVal:IAttachment):void
		{
			if (focusAndSelectionEventsSuspended()) {
				// Don't bother with this setup until we're unsuspeded
				return;
			}
			var newFocusIndex:int = ((focusAttachment == null) || (_currentItems.length == 0)) ? -1 : _currentItems.getItemIndex(focusAttachment);
			if (newFocusIndex < 0) {
				nextAttachment = null;
				prevAttachment = null;
			} else {
				if (newFocusIndex == 0) {
					prevAttachment = null;
				} else {
					prevAttachment = _currentItems.getItemAt(newFocusIndex-1) as IAttachment;
				}
				if ((newFocusIndex+1) >= _currentItems.length) {
					nextAttachment = null;
				} else {
					nextAttachment = _currentItems.getItemAt(newFocusIndex+1) as IAttachment;
				}
			}
			if (_host != null) {
				// Set the host's next/prev to a non-folder attachment. The only place using this is platform preview.
				var hostNextAtt:IAttachment = nextAttachment;
				var hostPrevAtt:IAttachment = prevAttachment;
				var i:int = newFocusIndex + 2; // Start at next + 1
				while (hostNextAtt && hostNextAtt.isFolder && i < _currentItems.length)
				{
					hostNextAtt = _currentItems.getItemAt(i) as IAttachment;
					i++;
				}
				if (hostNextAtt && hostNextAtt.isFolder) // If the last item is also a folder, set next to null.
					hostNextAtt = null;
				i = newFocusIndex - 2; // Start at prev - 1
				while (hostPrevAtt && hostPrevAtt.isFolder && i >= 0)
				{
					hostPrevAtt = _currentItems.getItemAt(i) as IAttachment;
					i--;
				}
				if (hostPrevAtt && hostPrevAtt.isFolder) // If the first item is also a folder, set prev to null
					hostPrevAtt = null;
				
				_host.setNextPrevious(hostNextAtt, hostPrevAtt);
			}
			if (newFocusIndex != focusIndex)
			{
				focusIndex = newFocusIndex;
				dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.FOCUS_DID_CHANGE_KIND, oldVal, focusAttachment));
			}
				
		}
		
		/**
		 * A listener that receives events from host.collection and propagates them using NavHostEvents.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onCollectionPropagateEvent(event:Event):void
		{
			switch (event.type) {
				case "needsSave":
					dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.NEEDS_SAVE_KIND));
					var commonProperties:CommonProperties = PropertyManager.instance.commonProperties;
					if ((commonProperties != null) &&
						(commonProperties.portfolioFontName != null) &&
						(commonProperties.portfolioFontName.name != "applicationFont"))
					{
						FontManager.instance.addFont(commonProperties.portfolioFontName.name);
					}
					FontManager.instance.registerFontsWithHost();
					break;
				case "permissionsChanged":
					dispatchEvent(new CollectionManagerEvent(CollectionManagerEvent.PERMISSIONS_CHANGED_KIND));
					break;
				default:
					CONFIG::logging{
						LogHelper.logger.warn("Unexpected collection event: "+event.type);
					}
					break;
			}
		}
		
		/**
		 * A listener to handle changes to properties on host.collection.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function onCollectionPropertyChanged(evt:PropertyChangeEvent):void
		{
			switch(evt.property)
			{
				case "sort":
					// Resort
					sort = evt.newValue as Array;
					break;
				case "fields":
					if(evt.newValue != evt.oldValue)
						fields = evt.newValue as Array;
					break;
				
				case "decorateFields":
					// Do something
					break;
				
				case  "permissions":
					// Do something
					break;
				
				default:break;
			}
			return;
		}

		/** 
		 *  <code>true</code> if the collectionManager instance is enabled,
		 *  <code>false</code> otherwise. rootItems and currentItems will
		 * 	be null until the instance is enabled.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get enabled():Boolean
		{
			return _enabled;
		}

		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set enabled(value:Boolean):void
		{
			if (_enabled == value)
				return;
			
			CONFIG::logging{
				LogHelper.logger.debug("CollectionManager.enabled: value= " + value);
			}
			
			_enabled = value;
			
			currentItems = _enabled ? _currentItems : null;
			rootItems = _enabled ? _rootItems : null;
		}
		
		/**
		 *  Deletes the set of attachments in <code>attachmentArray</code>
		 *  from the Portfolio
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function deleteAttachments(attachmentArray:Array, response:Function = null):void
		{
			if ((_host == null) || (_host.collection == null) || (attachmentArray == null)) {
				return;
			}
			_host.collection.deleteItems(attachmentArray, response);
		}
		
		/**
		 *  Code in the framework should never call BridgeLogger.logger.logPerformance directly.
		 *  Instead they should call this helper method (which verifies that the logPerformance call exists before calling it).
		 *  Furthermore, the number of logPeformance calls made from the framework should be kept to a minimum and we may
		 *  remove all of them before releasing the product (since it is difficult to justify a public call that is only used internally).
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		CONFIG::performance {
			public function logPerformance(id:String, desc:String=null, tag:int=BridgeLogger.NEW_PERFORMANCE_TAG, placement:int=0, extras:String=null):void
			{
				if (BridgeLogger.logger != null &&
					BridgeLogger.logger.hasOwnProperty("logPerformance") &&
					BridgeLogger.logger.logPerformance != null) {
					BridgeLogger.logger.logPerformance(id, desc, tag, placement, extras);
				}
			}
		}
		
		/**
		 *  This function returns <code>true</code> if the host supports
		 *  this version of the Portfolio API; otherwise it returns <code>false</code>.
		 * 
		 * 	@see #API_VERSION_9_5
		 * 	@see #API_VERSION_10
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function apiIsSupported(requestedVersion:Array):Boolean
		{
			if ((requestedVersion == null) || (_host == null) || (_host.apiVersion == null)) {
				return false;
			}
			var apiVersion:Array = _host.apiVersion;
			var compareLength:int = Math.max(apiVersion.length, requestedVersion.length);;
			for (var i:int = 0; i < compareLength; ++i) {
				var currentRequestedNumber:Number = (i >= requestedVersion.length) ? 0 : requestedVersion[i];
				var currentAPINumber:Number = (i >= apiVersion.length) ? 0 : apiVersion[i];
				if (isNaN(currentRequestedNumber) || (currentRequestedNumber > currentAPINumber)) {
					return false;
				} else if (currentRequestedNumber < currentAPINumber) {
					// If we encounter a version number greater than what we're interested
					// in, return true without bothering with the more-minor version numbers
					return true;
				}
			}
			return true;
		}

		// This is being kept as class property to avoid bindings on this to trigger gets on the host applicationFontList property.
		// The host applicationFontList is being requested only when the authoring mode is set to true for the first time
		private var _applicationFontList:Array;
		
		[Bindable]
		/**
		 *  The application's list of supported fonts
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get applicationFontList():Array
		{
			return _applicationFontList;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function set applicationFontList(value:Array):void
		{
			if (value != _applicationFontList) {
				_applicationFontList = value;
			}
		}

		private var embeddedFontWatcher:ChangeWatcher;
		private function setEmbeddedFontList(value:Array):void
		{
			embeddedFontList = value;
		}
		
		private var _embeddedFontList:Array;
		[Bindable]
		/**
		 *  This function returns the array of fonts that are embedded
		 *  in the portfolio.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get embeddedFontList():Array
		{
			return _embeddedFontList;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function set embeddedFontList(value:Array):void
		{
			_embeddedFontList = value;
		}
		
		//  default font overrides
		private var defaultFontTable:Array = new Array(
			{ fontName:"Adobe \u660e\u9ad4 Std L", language:"zh_TW" }, // AdobeMingStd-Light 
			{ fontName:"Adobe \u5b8b\u4f53 Std L", language:"zh_CN" }, // AdobeSongStd-Light
			{ fontName:"\u5c0f\u585a\u660e\u671d Pr6N R", language:"ja" }, // KozMinPr6N-Regular
			{ fontName:"Adobe \uba85\uc870 Std M", language:"ko" }, // AdobeMyungjoStd-Medium
			{ fontName:"Adobe Hebrew", language:"he" }, // AdobeHebrew-Regular
			{ fontName:"Adobe Arabic", language:"ar" }, // AdobeArabic-Regular
			{ fontName:"Adobe Thai", language:"th" } // AdobeThai-Regular
			);

		// This is the font used as the "default", which includes some language-dependent
		// overrides on the application font
		/**
		 *  This function returns the name of the application's default font.
		 *  This includes some language-dependent overrides on the application font.
		 *  The usual default is "Myriad Pro".  If there is no default font, the
		 *  string "applicationFont" is returned.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function getDefaultFontName(includeDefaultFont:Boolean):String
		{
			if ((_host == null) ||
				(_host.locale == null) ||
				(_host.locale.length < 2))
			{
				return "applicationFont";
			}
			
			// See if we have one of the overrides
			var locale:String = _host.locale;
			for each (var entry:Object in defaultFontTable)
			{
				if (locale.substr(0, entry.language.length) == entry.language)
				{
					return entry.fontName;
				}
			}
			// Didn't find one, maybe use Myriad if it's supported in this locale
			if (isMyriadSupported())
				return "Myriad Pro";

			if (includeDefaultFont || applicationFont.fontName == null)
				return "applicationFont"
				
			return applicationFont.fontName;
		}
		
		/**
		 *  Returns the prefix for the locale language, for example "en" or "fr"
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function get localeLanguagePrefix():String
		{
			if ((_host == null) ||
				(_host.locale == null) ||
				(_host.locale.length < 2))
			{
				return null;
			}
			return _host.locale.substr(0, 2).toLowerCase();
		}
		
		/**
		 *  This function returns <code>true</code> if "Myriad Pro" font is
		 *  supported, <code>false</code> otherwise.  "Myriad Pro" is the usual default font.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		public function isMyriadSupported():Boolean
		{
			var localePrefix:String = localeLanguagePrefix;
			if ((localePrefix == "en") ||	// English
				(localePrefix == "fr") ||	// French
				(localePrefix == "de") ||	// German
				
				(localePrefix == "sq") ||	// Albanian
				(localePrefix == "be") ||	// Belarusian
				(localePrefix == "bg") ||	// Bulgarian
				(localePrefix == "ca") ||	// Catalan
				(localePrefix == "hr") ||	// Croatian
				(localePrefix == "cs") ||	// Czech 
				(localePrefix == "da") ||	// Danish
				(localePrefix == "nl") ||	// Dutch
				(localePrefix == "et") ||	// Estonian
				(localePrefix == "fi") ||	// Finish
				(localePrefix == "el") ||	// Greek
				(localePrefix == "hu") ||	// Hungarian
				(localePrefix == "is") ||	// Icelandic
				(localePrefix == "it") ||	// Italian
				(localePrefix == "lv") ||	// Latvian
				(localePrefix == "lt") ||	// Lithuanian
				(localePrefix == "mk") ||	// Macedonian
				(localePrefix == "nb") ||	// Norwegian
				(localePrefix == "pl") ||	// Polish
				(localePrefix == "pt") ||	// Portugese
				(localePrefix == "ro") ||	// Romanian
				(localePrefix == "ru") ||	// Russian
				(localePrefix == "sr") ||	// Serbian
				(localePrefix == "sk") ||	// Slovakian
				(localePrefix == "sl") ||	// Slovenian
				(localePrefix == "es") ||	// Spanish
				(localePrefix == "sv") ||	// Swedish
				(localePrefix == "tr") ||	// Turkish
				(localePrefix == "vi"))		// Vietnamese
			{
				return true;
			}
			return false;
		}

		private function get isAdobeHost():Boolean
		{
			return _isAdobeHost;
		}
		
		[Bindable]
		/**
		 *  Indicates whether a card is being previewing in the NavPreview component
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get isNavPreviewCardPreview():Boolean
		{
			return _isNavPreviewCardPreview;
		}
		
		public function set isNavPreviewCardPreview(value:Boolean):void
		{
			_isNavPreviewCardPreview = value;
		}

		/**
		 *  The job queue that trickles attachments into the root collection in the initialization phase of the Navigator.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get rootItemsJobQueue():JobQueue
		{
			return _rootItemsJobQueue;
		}

	}
}
