/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.managers
{
	import com.adobe.portfolio.api.CommandType;
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.api.ICommand;
	import com.adobe.portfolio.api.INavigatorHost;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.utils.FileUtil;
	import com.adobe.portfolio.utils.LogHelper;
	
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.IEventDispatcher;
	
	import mx.collections.ArrayCollection;
	
	
	/**
	 *  The CommandManager is the central control for navigator commands that directly
	 *  affect the attachments in the Portfolio, such as previewing, editing, and extracting
	 *  attachments. It has a cache to keep track of what command need to run and what commands
	 *  could not be executed because they require a request from the host.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */		
	public class CommandManager extends EventDispatcher
	{
		/** Constructor
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function CommandManager(target:IEventDispatcher=null)
		{
			super(target);
			collectionManager.addEventListener(CollectionManagerEvent.HOST_CHANGED_KIND, onHostEvent);
			setupHost(null, collectionManager.host);
		}
		
		/** Internal representation of the CommandManager referenced by the instance getter.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		private static var _instance:CommandManager = null;
		
		/**
		 * A CommandManager static instance that can be shared by Flex components.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function get instance():CommandManager
		{
			if (_instance == null) {
				_instance = new CommandManager();
			}
			return _instance;
		}
		
		/** 
		 *  The navigator's CollectionManager instance. 
		 *  All host interaction should go through this instance,
		 *  rather than the INavigatorHost directly.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected const collectionManager:CollectionManager = CollectionManager.instance;
		
		/** 
		 *  Sets up a listener to detect and handle a change of collectionManager's underlying INavigatorHost.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onHostEvent(event:CollectionManagerEvent):void
		{
			if (event.kind == CollectionManagerEvent.HOST_CHANGED_KIND) {
				setupHost(event.oldVal as INavigatorHost, event.newVal as INavigatorHost);
			}
		}
		
		/** 
		 *  Performs the shutdown and setup required when the collectionManager's underlying host changes.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function setupHost(oldHost:INavigatorHost, newHost:INavigatorHost):void
		{
			if (oldHost != null) {
				oldHost.removeEventListener("commandsChanged", onCommandsChanged);
			}
			if (newHost != null) {
				newHost.addEventListener("commandsChanged", onCommandsChanged);
				commandCache = newHost.commands;
				
				// If platform preview comes back, then we should cache the PREVIEW command here, rather than EDIT_NATIVE
				if (getCachedCommand(CommandType.EDIT_NATIVE) == null) {
					// Request the Open File command if we don't already have it,
					// since we're very likely to need it.
					newHost.requestCommands([CommandType.EDIT_NATIVE]);
				}
			} else {
				commandCache = null;
			}
		}
		
		/** 
		 *  An internal cache of host's current commands to be executed.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var commandCache:Object = null;
		
		/** 
		 *  An internal map to track commands which could not be executed because they required
		 *  a request from the host.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var pendingCommandMap:Object = new Object();
		
		/** 
		 *  Adds a command to pendingCommandMap, along with its operand and arguments.
		 *  The pendingCommandMap tracks commands which could not be executed because 
		 *  they required a request from the host.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function addPendingCommand(commandType:String, operand:*, arguments:*):void
		{
			var pendingArray:Array = pendingCommandMap[commandType];
			if (pendingArray == null) {
				pendingArray = new Array();
				pendingCommandMap[commandType] = pendingArray;
			}
			pendingArray.push({ "operand": operand, "arguments": arguments });
		}
		
		/** 
		 *  Listens for command changes from the host, and executes any pending commands that
		 *  have become available.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onCommandsChanged(event:Event):void
		{
			commandCache = collectionManager.host.commands;
			for (var commandType:String in pendingCommandMap) {
				var cachedCommand:ICommand = getCachedCommand(commandType);
				if ((cachedCommand != null) && cachedCommand.enabled) {
					var pendingArray:Array = pendingCommandMap[commandType];
					pendingCommandMap[commandType] = null;
					var pendingCount:uint = (pendingArray == null) ? 0 : pendingArray.length;
					for (var i:uint = 0; i < pendingCount; ++i) {
						var pendingObj:Object = pendingArray[i];
						if (pendingObj == null) {
							continue;
						}
						var pendingOperand:* = pendingObj["operand"];
						var pendingArguments:* = pendingObj["arguments"];
						pendingArray[i] = null;
						cachedCommand.execute(pendingOperand, pendingArguments);
					}
				} 
			}
		}
		
		/** 
		 *  Returns the ICommand of type commandType from the local commandCache, or null
		 *  if there is no such ICommand in commandCache.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function getCachedCommand(commandType:String):ICommand
		{
			if (commandCache != null) {
				return commandCache[commandType];
			} else {
				return null;
			}
		}
		
		/** 
		 *  A callback that is triggered when a command request has been fulfilled.
		 *  This doesn't currently do anything useful, but I suspect will be
		 *  needed to make CommandManager more robust, and is useful for debugging.
		 *  
		 *  @private
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		private function onInvokeRequestResult(result:Object):void
		{
			if (result == null) {
				LogHelper.logger.error("Command request callback with null result");
			} else if (!result.propertyIsEnumerable("success") || !(result["success"] as Boolean)) {
				LogHelper.logger.error("Command request callback with unsuccessful result");
			}
			// No need to actually invoke the command, since it will be triggered by a call to onCommandsChanged
		}
		
		/** 
		 *  Executes the ICommand of type commandType from the commandCache, 
		 *  or requests the ICommand if it is not current present in commandCache.
		 *  Returns true if the command was succesfully executed or requested;
		 *  returns false if the command is not enabled.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function invokeCommand(commandType:String, operand:* = null, arguments:Array = null):Boolean
		{
			var command:ICommand = getCachedCommand(commandType);
			if (command == null) {
				if (collectionManager.host != null) {
					addPendingCommand(commandType, operand, arguments);
					collectionManager.host.requestCommands([commandType], onInvokeRequestResult);
					// TODO return some sort of token to listen for the result
					return true;
				}
			} else if (command.enabled) {
				command.execute(operand, arguments);
				return true;
			}
			return false;
		}
		
		/** 
		 *  Determines and then executes the appropriate command for previewing the indicated attachment.
		 * 
		 * 	@see CommandType#PREVIEW
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function previewAttachment(attachment:IAttachment):void
		{
			if (!attachment || attachment.isFolder)
				return;
			
			collectionManager.setSelectedItems([attachment]);
			invokeCommand(CommandType.PREVIEW, attachment);
		}
		
		/** 
		 *  Determines and then executes the appropriate command for editing the indicated attachment
		 *  in its native application. If previewAppropriateTypes is true then it checks to see
		 *  if the attachment is an HTML file or SWF file, in which case it will use the preview command
		 *  instead.
		 * 
		 * 	@see CommandType#PREVIEW
		 * 	@see CommandType#EDIT_NATIVE
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function editAttachment(attachment:IAttachment, previewAppropriateTypes:Boolean = true):void
		{
			if (attachment != null) {
				collectionManager.setSelectedItems([attachment]);
				if (!previewAppropriateTypes) {
					invokeCommand(CommandType.EDIT_NATIVE, attachment);
					return;
				}
				
				// Some types are better handled by the preview command, so use that when appropriate
				var extension:String = FileUtil.getExtension(attachment.fileName);
				extension = (extension == null) ? "" : extension.toLowerCase();
				if ((extension == "html") || (extension == "swf")) {
					invokeCommand(CommandType.PREVIEW, attachment);
				} else {
					invokeCommand(CommandType.EDIT_NATIVE, attachment);
				}
			}
		}
		
		/** 
		 *  Determines and then executes the appropriate command
		 *  for extracting the given attachment from the Portfolio.
		 * 
		 * 	@see CommandType#EXTRACT_FILES
		 *   
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function extractAttachment(attachment:IAttachment):void
		{
			if (attachment != null) {
				collectionManager.setSelectedItems([attachment]);
				invokeCommand(CommandType.EXTRACT_FILES, attachment);
			}
		}
		
		/** 
		 * Determines and then executes the appropriate command for deleting the given attachment.
		 *   
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function deleteAttachment(attachment:IAttachment):void
		{
			if (attachment != null) {
				// It's easier for lists to recover from deletions when the deleted item
				// is selected, so go ahead and select as we do when invoking actual commands.
				collectionManager.setSelectedItems([attachment]);
				collectionManager.deleteAttachments([attachment]);
			}
		}
		
		/** 
		 * Determines and then executes the appropriate command for deleting the set of selected attachments.
		 *   
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function deleteSelection():void
		{
			var selectedItems:ArrayCollection = collectionManager.selectedItems;
			if ((selectedItems != null) && (selectedItems.length > 0)) {
				collectionManager.deleteAttachments(selectedItems.toArray());
			}
		}
		
		/** Executes the appropriate command for popping the context menu.
		 * 
		 * 	@see CommandType#CONTEXT_MENU
		 *   
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function showContextMenu(x:Number, y:Number):void
		{
			invokeCommand(CommandType.CONTEXT_MENU, null, [x, y]);
		}
	}
}
