/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.managers
{
	import com.adobe.portfolio.api.ICollection;
	import com.adobe.portfolio.api.INavigatorHost;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.properties.CommonProperties;
	import com.adobe.portfolio.utils.LogHelper;
	import com.adobe.portfolio.utils.NavURLLoader;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;

	
	/**
	 *  The FontManager handles adding and registering fonts and keeping track of
	 *  whether fonts are to be embedded when the Portfolio is saved. 
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */	
	public class FontManager
	{
		private static var _instance:FontManager = null;
		
		/** The navigator's CollectionManager instance. All host interaction should go through this
		 *  instance, rather than the INavigatorHost directly.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected const collectionManager:CollectionManager = CollectionManager.instance;

		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */			
		private var fontMap:Object;
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		private var fontResources:Object;
		
		/**
		 * A FontManager static instance that can be shared by Flex components.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public static function get instance():FontManager
		{
			if (_instance == null) {
				_instance = new FontManager();
			}
			return _instance;
		}

		/**
		 *  Default constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function FontManager()
		{
			collectionManager.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent, false, 0, true);
		}

		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		private var embedOnSaveWatcher:ChangeWatcher;

		/**
		 *  Listener to handle Navigator host events.
		 *  The only such event of interest is a change in hosts
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onNavHostEvent(event:CollectionManagerEvent):void
		{
			if ((event == null) || (event.kind != CollectionManagerEvent.HOST_CHANGED_KIND)) {
				// We only care about host changes
				return;
			}
			if (embedOnSaveWatcher != null) {
				embedOnSaveWatcher.unwatch();
				embedOnSaveWatcher = null;
			}
			var navHost:INavigatorHost = event.newVal as INavigatorHost;
			if ((navHost != null) && (navHost.collection != null)) {
				embedOnSaveWatcher = BindingUtils.bindSetter(setEmbedOnSave, navHost.collection, "embedFontsOnSave");
			}
		}

		private function setEmbedOnSave(value:Boolean):void
		{
			embedFontsOnSave = value;
		}

		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		private var _embedFontsOnSave:Boolean;
		
		[Bindable]
		/**
		 *  Indicates whether if the host collection 
		 *  will embed the fonts it uses when the portfolio is saved.
		 *  This property can be used as the source for data binding.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get embedFontsOnSave():Boolean
		{
			return _embedFontsOnSave;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set embedFontsOnSave(value:Boolean):void
		{
			_embedFontsOnSave = value;
			var hostCollection:ICollection = (collectionManager.host == null) ? null : collectionManager.host.collection;
			if (hostCollection != null) {
				hostCollection.embedFontsOnSave = value;
			}
		}
		
		/**
		 *  Scans any available font resources for font references to add to the local store, and
		 *  registers all fonts in the local store with the host for embedding.
		 *  Font resources are created by the various addfont methods: <code>addFont()</code>,
		 *  <code>addFontsFromTLF()</code>, and <code>addFontResource()</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function registerFontsWithHost():void
		{
			var hostCollection:ICollection = (collectionManager.host == null) ? null : collectionManager.host.collection;
			if (hostCollection == null) {
				CONFIG::logging{
					LogHelper.logger.warn("Host collection is null");
				}
				return;
			}
			if (!collectionManager.apiIsSupported(CollectionManager.API_VERSION_10)) {
				CONFIG::logging{
					LogHelper.logger.warn("Insufficient API for registering fonts");
				}
				return;
			}
			if (!collectionManager.permissionCheck("Document", "Modify")) {
				CONFIG::logging{
					LogHelper.logger.warn("Insufficient permissions for registering fonts");
				}
				return;
			}
			var commonProperties:CommonProperties = PropertyManager.instance.commonProperties;
			if ((commonProperties != null) &&
				(commonProperties.portfolioEmbedFonts != null) &&
				!commonProperties.portfolioEmbedFonts.booleanValue)
			{
				// Don't bother registering fonts if we're not embedding
				return;
			}
			if (fontResources != null) {
				for each (var resourceLoader:NavURLLoader in fontResources) {
					if (resourceLoader == null) {
						continue;
					}
					if (resourceLoader.bytesTotal > 0) {
						CONFIG::logging{
							LogHelper.logger.warn("Incomplete loader: "+resourceLoader.url);
						}
						continue;
					}
					addFontsFromTLF(new XML(resourceLoader.dataString));
				}
			}
			if (fontMap == null) {
				// No fonts to embed
				return;
			}
			
			var fontArray:Array = new Array();
			for (var fontName:String in fontMap) {
				var mapObj:Object = fontMap[fontName];
				if (mapObj == null) {
					continue;
				}
				if (mapObj.normal)
					fontArray.push({
						fontName: fontName,
						fontWeight: "normal",
						fontPosture: "normal"
					});
				if (mapObj.bold)
					fontArray.push({
						fontName: fontName,
						fontWeight: "bold",
						fontPosture: "normal"
					});
				if (mapObj.italic)
					fontArray.push({
						fontName: fontName,
						fontWeight: "normal",
						fontPosture: "italic"
					});
				if (mapObj.boldItalic)
					fontArray.push({
						fontName: fontName,
						fontWeight: "bold",
						fontPosture: "italic"
					});
			}
			hostCollection.registerFonts(fontArray);
		}
		
		/**
		 * Adds a particular font to the local store so that it can be registered with
		 * subsequent calls to <code>registerFontsWithHost</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function addFont(fontName:String,
								includeNormal:Boolean = true,
								includeBold:Boolean = true,
								includeItalic:Boolean = true,
								includeBoldItalic:Boolean = true):void
		{
			// This method ccould be extended to take an id object. If the fontObj maintained
			// a Dictionary of id objects, then it could serve as a type of refcounting that
			// could be used to maintain an ongoing store of fonts that are actually being referenced,
			// rather than scanning content on a save.
			
			if ((fontName == null) || (fontName == "")) {
				LogHelper.logger.error("Unsupported empty font name");
				return;
			}
			if (fontMap == null) {
				fontMap = new Object();
			}
			if (!fontMap.propertyIsEnumerable(fontName)) {
				fontMap[fontName] = { 	normal: includeNormal,
										bold: includeBold,
										italic: includeItalic,
										boldItalic: includeBoldItalic};
				return;
			}
			var fontObj:Object = fontMap[fontName];
			fontObj.normal = includeNormal || fontObj.normal;
			fontObj.bold = includeBold || fontObj.bold;
			fontObj.italic = includeItalic || fontObj.italic;
			fontObj.boldItalic = includeBoldItalic || fontObj.boldItalic;
		}
		
		/**
		 *  Adds all fonts used within the given TLF (Text Layout Framework) XML block
		 *  to the local store so that they can be registered with subsequent calls 
		 *  to <code>registerFontsWithHost</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function addFontsFromTLF(tlfXML:XML):void
		{
			if (tlfXML == null) {
				LogHelper.logger.error("Null tlfXML");
				return;
			}
			var spanName:QName = new QName("http://ns.adobe.com/textLayout/2008", "span");
			var spans:XMLList = tlfXML.descendants(spanName);
			for each (var span:XML in spans) {
				var fontFamily:String = span.@fontFamily;
				if ((fontFamily == null) || (fontFamily == "")) {
					continue;
				}
				var includeBold:Boolean = ("bold" == span.@fontWeight);
				var includeItalic:Boolean = ("italic" == span.@fontStyle);
				addFont(fontFamily,
						!includeBold && !includeItalic,
						includeBold && !includeItalic,
						includeItalic && !includeBold,
						includeBold && includeItalic);
			}
		}
		
		/**
		 * Adds a reference to the specified resourceURL so that its contents can be scanned
		 * for font references and added to the local store with subsequent calls
		 * to <code>registerFontsWithHost</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function addFontResource(resourceURL:String):void
		{
			if ((resourceURL == null) || (resourceURL == "")) {
				LogHelper.logger.error("Unsupported empty resource URL");
				return;
			}
			if (fontResources == null) {
				fontResources = new Object();
			}
			if (fontResources.propertyIsEnumerable(resourceURL)) {
				// resourceURL is already present
				return;
			}
			fontResources[resourceURL] = ResourceManager.instance.getResource(resourceURL, false);
		}
		
		/**
		 * Removes a reference to resourceURL so that its contents will no longer be scanned
		 * for font references on subsequent calls to <code>registerFontsWithHost</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function removeFontResource(resourceURL:String):void
		{
			if ((resourceURL == null) || (resourceURL == "")) {
				LogHelper.logger.error("Unsupported empty resource URL");
				return;
			}
			if (fontResources == null) {
				// no fontResources present
			}
			if (!fontResources.propertyIsEnumerable(resourceURL)) {
				// resourceURL isn't present
				return;
			}
			delete fontResources[resourceURL];
		}
	}
}
