/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.managers
{
	import com.adobe.portfolio.api.INavigatorHost;
	import com.adobe.portfolio.api.PropertyValueChangedEvent;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.events.PropertyManagerEvent;
	import com.adobe.portfolio.properties.CommonProperties;
	import com.adobe.portfolio.properties.PropertyContext;
	import com.adobe.portfolio.properties.PropertyValue;
	import com.adobe.portfolio.utils.LogHelper;
	import com.adobe.portfolio.utils.NavURLLoader;
	
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.IEventDispatcher;
	import flash.events.IOErrorEvent;
	import flash.events.SecurityErrorEvent;
	import flash.net.URLLoader;
	import flash.net.URLRequest;
	import flash.utils.ByteArray;
	import flash.utils.Dictionary;
	
	import mx.binding.utils.BindingUtils;
	import mx.collections.ArrayCollection;
	
	[Event (name="navPropertyContextLoaded", type="com.adobe.portfolio.events.PropertyManagerEvent")]
	[Event (name="navPropertyUpdatedContexts", type="com.adobe.portfolio.events.PropertyManagerEvent")]
	[Event (name="complete")]
	
	/** PropertyManager loads nav property XML files and propagates their contents to the navigator host.
	 *  PropertyManager does not contain logic for handling property behavior, but does contain a reference to
	 *  a CommonProperties object for dealing with such behavior.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class PropertyManager extends EventDispatcher
	{
		private static var _instance:PropertyManager = null;
		

		[Bindable]
		/** A CommonProperties instance that can be accessed by navigator components for setting and 
		 *  responding to property changes, and is suitable for binding.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public var commonProperties:CommonProperties;
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private var pendingPropertiesUpdate:Boolean = false;
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private var pendingCurrentProperties:PropertyContext;
		
		/** Returns the navigator property with the specified name (propertyName)
		 *  from the given context (contextName), or null if no such property has been loaded.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function getNavProperty(contextName:String, propertyName:String):PropertyValue
		{
			for each (var context:PropertyContext in _globalContexts) {
				if ((context != null) && (context.name == contextName)) {
					return context.getProperty(propertyName);
				}
			}
			return null;
		}
		
		/**
		 *  A PropertyManager static instance that can be shared by Flex components.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public static function get instance():PropertyManager
		{
			if (_instance == null) {
				_instance = new PropertyManager();
			}
			return _instance;
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private var contextMap:Dictionary = new Dictionary();
		
		private var _knownContexts:ArrayCollection = new ArrayCollection();
		
		[Bindable]
		/**
		 *  Returns an ArrayCollection that contains the currently
		 *  loaded property contexts.  A property context provides 
		 *  labels and descriptions for UI properties
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get knownContexts():ArrayCollection
		{
			return _knownContexts;
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function set knownContexts(newContexts:ArrayCollection):void
		{
			if (newContexts != _knownContexts) {
				_knownContexts = newContexts;
			}
		}
		
		private var _globalContexts:ArrayCollection = new ArrayCollection();
		[Bindable]
		/**
		 *  Returns an ArrayCollection of global loaded property contexts. A
		 * 	global property context contains properties that are always
		 * 	appropriate to display to the user.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get globalContexts():ArrayCollection
		{
			return _globalContexts;
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function set globalContexts(newContexts:ArrayCollection):void
		{
			if (newContexts != _globalContexts) {
				_globalContexts = newContexts;
				// To avoid duplicate calls, this does NOT automatically set the host contexts.
			}
		}
		
		/**
		 *  Adds the specified propertyContext to the knownContents ArrayCollection.
		 * 	The new list of contexts will not be communicated to the host until
		 * 	updateHostContexts is called.
		 * 
		 * 	@see #updateHostContexts
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected function addContext(newContext:PropertyContext):void
		{
			if (knownContexts == null) {
				knownContexts = new ArrayCollection();
			}
			knownContexts.addItem(newContext);
			if (!newContext.global) {
				return;
			}
			
			// Use the public getter to create _globalContexts as needed
			var contexts:ArrayCollection = globalContexts;
			if (contexts != null) {
				contexts.addItem(newContext);
				// TODO set host contexts
			} else {
				LogHelper.logger.error("globalContexts was null");
			}
		}
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function PropertyManager(target:IEventDispatcher=null)
		{
			super(target);
			CollectionManager.instance.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onHostChanged);
			BindingUtils.bindSetter(setAuthoringMode, CollectionManager.instance, "authoringMode");
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function onHostChanged(event:CollectionManagerEvent):void
		{
			if (event.kind == CollectionManagerEvent.HOST_CHANGED_KIND) {
				if ((event.oldVal != null) && (event.oldVal is INavigatorHost)) {
					(event.oldVal as INavigatorHost).removeEventListener(PropertyValueChangedEvent.PROPERTY_VALUE_CHANGED, onPropertyValueChanged);
				}
				if ((event.newVal != null) &&
					(event.newVal is INavigatorHost) &&
					CollectionManager.instance.apiIsSupported(CollectionManager.API_VERSION_10))
				{
					if (globalContexts.length > 0) {
						updateHostContexts();
					}
					if ((pendingCurrentProperties != null) && CollectionManager.instance.authoringMode) {
						(event.newVal as INavigatorHost).currentPropertyContext = pendingCurrentProperties;
						pendingCurrentProperties = null
					}
					(event.newVal as INavigatorHost).addEventListener(PropertyValueChangedEvent.PROPERTY_VALUE_CHANGED, onPropertyValueChanged);
				}
			} else if (event.kind == CollectionManagerEvent.NEEDS_SAVE_KIND) {
				// We're getting ready to save, so flush out our dirty property contexts
				if (CollectionManager.instance.host != null) {
					var knownContext:PropertyContext;
					var dirtyURL:String;
					var dirtyXML:XML;
					var dirtyURLs:ArrayCollection = new ArrayCollection();
					for each (knownContext in knownContexts) {
						if (knownContext.dirty && (knownContext.url != "") && !dirtyURLs.contains(knownContext.url)) {
							dirtyURLs.addItem(knownContext.url);
						}
					}
					if ((dirtyURLs.length > 0) && CollectionManager.instance.permissionCheck("Document", "Modify")) {
						// TODO We should probably make a call to CollectionManager.instance.permissionCheck before continuing
						var dirtyXMLs:Object = new Object();
						for each (dirtyURL in dirtyURLs) {
							dirtyXMLs[dirtyURL] = new XML("<NavProperties/>");
						}
						for each (knownContext in knownContexts) {
							if (dirtyURLs.contains(knownContext.url)) {
								dirtyXML = dirtyXMLs[knownContext.url];
								if (dirtyXML != null) {
									var contextXML:XML = knownContext.generateXML();
									if (contextXML != null) {
										dirtyXML.appendChild(contextXML);
									} else {
										LogHelper.logger.error("Couldn't generate XML for PropertyContext "+knownContext.name);
									}
								}
							}
						}
						for each (dirtyURL in dirtyURLs) {
							dirtyXML = dirtyXMLs[dirtyURL];
							var dirtyBytes:ByteArray = new ByteArray();
							dirtyBytes.writeUTFBytes("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
							dirtyBytes.writeUTFBytes(dirtyXML.toXMLString());
							ResourceManager.instance.writeResource(dirtyBytes, dirtyURL);
						}
					}
				}
			} else if (event.kind == CollectionManagerEvent.PERMISSIONS_CHANGED_KIND) {
				// TODO Update properties' enabled states based on CollectionManager.instance.permissionCheck
			}
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function setAuthoringMode(value:Boolean):void
		{
			if (value) {
				if (pendingPropertiesUpdate) {
					updateHostContexts();
				}
				if ((pendingCurrentProperties != null) &&
					(CollectionManager.instance.host != null) &&
					CollectionManager.instance.apiIsSupported(CollectionManager.API_VERSION_10))
				{
					CollectionManager.instance.host.currentPropertyContext = pendingCurrentProperties;
					pendingCurrentProperties = null
				}
			}
		}
		
		/**
		 *  Checks to see if we are in authoring mode and environment that supports this API.
		 *  If so, the host's propertyContexts is set to the globalContexts.
		 *  Otherwise pendingPropertiesUpdate is set to <code>true</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function updateHostContexts():void
		{
			if (CollectionManager.instance.authoringMode &&
				(CollectionManager.instance.host != null) &&
				CollectionManager.instance.apiIsSupported(CollectionManager.API_VERSION_10))
			{
				CollectionManager.instance.host.propertyContexts = globalContexts.toArray();
				pendingPropertiesUpdate = false;
			} else {
				pendingPropertiesUpdate = true;
			}
			dispatchEvent(new PropertyManagerEvent(PropertyManagerEvent.UPDATED_CONTEXTS));
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function onPropertyValueChanged(event:PropertyValueChangedEvent):void
		{
			var context:PropertyContext = contextMap[event.contextName];
			if (context != null) {
				context.updatePropertyFromHost(event.propertyName, event.propertyValue);
			}
		}
		
		private var _currentContext:PropertyContext;
		
		/**
		 *  The current property context of the host
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get currentContext():PropertyContext
		{
			return _currentContext;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function set currentContext(newContext:PropertyContext):void
		{
			if (newContext != _currentContext) {
				// TODO set host context
				_currentContext = newContext;
				
				// Add to the contextMap if necessary
				if ((_currentContext != null) && (contextMap[_currentContext.name] == null)) {
					contextMap[_currentContext.name] = _currentContext;
				}
				
				if (CollectionManager.instance.authoringMode &&
					(CollectionManager.instance.host != null) &&
					CollectionManager.instance.apiIsSupported(CollectionManager.API_VERSION_10))
				{
					CollectionManager.instance.host.currentPropertyContext = newContext;
				} else {
					pendingCurrentProperties = newContext;
				}
			}
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private var propertyLoaders:ArrayCollection;
		
		/**
		 *  Load the properties specified in the XML file specified
		 *  by propertyURL and propagates the contents to the Navigator host.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function loadPropertyXML(propertyURL:String):void
		{
			//var propertyLoader:URLLoader = new URLLoader();
			var propertyLoader:URLLoader = new NavURLLoader();
			if (propertyLoaders == null) {
				propertyLoaders = new ArrayCollection();
			}
			propertyLoaders.addItem(propertyLoader);
			
			var loadComplete:Function;
			var loadError:Function;
			
			loadComplete = function (event:Event):void {
				var propertyLoader:URLLoader = event.target as URLLoader;
				if (propertyLoader == null) {
					return;
				}
				propertyLoader.removeEventListener(Event.COMPLETE, loadComplete);
				propertyLoader.removeEventListener(IOErrorEvent.IO_ERROR, loadError);
				propertyLoader.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, loadError);
				
				var propertyXML:XML = new XML(propertyLoader.data);
				var contextXMLList:XMLList = propertyXML.PropertyContext;
				var numContexts:int = contextXMLList.length();
				for (var i:int = 0; i < numContexts; ++i) {
					var newContext:PropertyContext = new PropertyContext(propertyURL);
					if (newContext.loadFromXML(contextXMLList[i])) {
						var propEvent:PropertyManagerEvent = new PropertyManagerEvent(PropertyManagerEvent.CONTEXT_LOADED);
						propEvent.context = newContext;
						dispatchEvent(propEvent);
					}
					contextMap[newContext.name] = newContext;
					addContext(newContext);
				}
				dispatchEvent(event);
			};
			
			loadError = function (event:Event):void {
				var propertyLoader:URLLoader = event.target as URLLoader;
				if (propertyLoader == null) {
					return;
				}
				propertyLoader.removeEventListener(Event.COMPLETE, loadComplete);
				propertyLoader.removeEventListener(IOErrorEvent.IO_ERROR, loadError);
				propertyLoader.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, loadError);
				
				LogHelper.logger.error("Error loading property xml: "+propertyURL);
				dispatchEvent(event);
			};
				
			propertyLoader.addEventListener(Event.COMPLETE, loadComplete);
			propertyLoader.addEventListener(IOErrorEvent.IO_ERROR, loadError);
			propertyLoader.addEventListener(SecurityErrorEvent.SECURITY_ERROR, loadError);

			ResourceManager.instance.referenceResource(propertyURL);
			propertyLoader.load(new URLRequest(propertyURL));
		}
	}
}
