/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.managers
{
	import com.adobe.portfolio.api.ICollection;
	import com.adobe.portfolio.components.NavAlert;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.utils.LogHelper;
	import com.adobe.portfolio.utils.NavURLLoader;
	
	import flash.display.DisplayObject;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.IEventDispatcher;
	import flash.events.IOErrorEvent;
	import flash.events.SecurityErrorEvent;
	import flash.net.FileFilter;
	import flash.net.URLLoaderDataFormat;
	import flash.net.URLRequest;
	import flash.utils.ByteArray;
	
	import mx.events.PropertyChangeEvent;
	
	/**
	 *  The ResourceManager is a singleton class that facilitates the use of resource data associated with
	 *  the navigator.  The resourceManager handles adding, loading, deleting, and 
	 *  modification of resource data.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */	

	public class ResourceManager extends EventDispatcher
	{
		private static var _instance:ResourceManager = null;
		
		/**
		 * 	The navigator's CollectionManager instance. All host interaction should go through this
		 *  instance, rather than the INavigatorHost directly.
		 */
		protected const collectionManager:CollectionManager = CollectionManager.instance;
		
		private var _enabled:Boolean;
		
		/**
		 *  A ResourceManager static instance that can be shared by Flex components.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public static function get instance():ResourceManager
		{
			if (_instance == null) {
				_instance = new ResourceManager();
			}
			return _instance;
		}
		
		private var _loadingCount:uint = 0;
		
		[Bindable (name="propertyChange")]
		/**
		 *  Returns a count of the number of resource loads that are currently in the process
		 *  of being loaded.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get loadingCount():uint
		{
			return _loadingCount;
		}
		
		/**
		 *  Increments the resource <code>loadingCount</code> by 1.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function incrementLoadingCount():void
		{
			if (loadingCount == uint.MAX_VALUE) {
				LogHelper.logger.error("incrementLoadingCount exceeded uint.MAX_VALUE");
				return;
			}
			++_loadingCount;
			dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "loadingCount", _loadingCount-1, _loadingCount));
		}
		
		/**
		 *  Decrements the resource <code>loadingCount</code> by 1.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function decrementLoadingCount():void
		{
			if (_loadingCount == 0) {
				LogHelper.logger.error("decrementLoadingCount at zero count");
				return;
			}
			--_loadingCount;
			dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "loadingCount", _loadingCount+1, _loadingCount));
		}
		
		/**
		 * A Boolean indicating whether or not resources should be auto-purged on a needsSave event.
		 *
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public var purgeOnNeedsSave:Boolean = false;
		
		private var loaders:Object = new Object();
		private var references:Object = new Object();
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function ResourceManager(target:IEventDispatcher=null)
		{
			super(target);
			collectionManager.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent);
		}
		
		private var _imageFilter:FileFilter;
		
		/**
		 *  Returns the filter for image files, if there is one.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get imageFilter():FileFilter
		{
			if (_imageFilter != null) {
				return _imageFilter;
			}
			var hostCollection:ICollection = (CollectionManager.instance.host != null) ? CollectionManager.instance.host.collection : null;
			var filterName:String = (hostCollection == null) ? "IMAGES_FILTER_NAME" : hostCollection.getLocalizedString("imagesFilterName", "IMAGES_FILTER_NAME");
			_imageFilter = new FileFilter(filterName, "*.jpg;*.jpeg;*.gif;*.png");
			return _imageFilter;
		}
		
		private var _swfFilter:FileFilter;
		/**
		 *  Returns the filter for SWF files, if there is one.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get swfFilter():FileFilter
		{
			if (_swfFilter != null) {
				return _swfFilter;
			}
			var hostCollection:ICollection = (CollectionManager.instance.host != null) ? CollectionManager.instance.host.collection : null;
			var filterName:String = (hostCollection == null) ? "SWF_FILTER_NAME" : hostCollection.getLocalizedString("swfFilterName", "SWF_FILTER_NAME");
			_swfFilter = new FileFilter(filterName, "*.swf");
			return _swfFilter;
		}
		
		
		/**
		 * Adds custom resources needed by the navigators. It brings up a file browsing dialog prompting the user 
		 * to select the needed file.  Types that are not blacklisted will be added to the resources.
		 * 
		 * <p>This operation is only allowed when the permission "Document"/"Modify" has been granted.</p>
		 * 
		 * <p>It accepts an optional callback function as an argument.  
		 * When a callback function is given, it is invoked to indicate the result of the operation.  
		 * The callback function receives a single argument, an <code>Object</code>, and has the following format:
		 * 		<pre>    function callback(response:Object):void</pre>
		 * The <code>response</code> parameter contains properties that describe the response.
		 * These include:</p>
		 * <table class="innertable">
		 *		<tr><th>Property</th><th>Type</th><th>Value</th></tr>
		 *		<tr>
		 *			<td><code>'success'</code></td>
		 *			<td><code>Boolean</code></td>
		 *			<td><code>true</code> if the operation succeeded; <code>false</code> otherwise</td>
		 *		</tr>
		 *		<tr>
		 *			<td><code>'relativeURL'</code></td>
		 *			<td><code>String</code></td>
		 *			<td>The path that identifes the resource.
		 * 				This is the full relative path to the resource, and may differ from <code>originalURL</code> 
		 * 				if the caller passes an incomplete path, which is then completed by addResource().
		 * 				Use this url to later read the resource.
		 * 			</td>
		 *		</tr>
		 *		<tr>
		 *			<td><code>'originalURL'</code></td>
		 *			<td><code>String</code></td>
		 *			<td>The path under which this resource was requested.
		 * 				It will match <code>relativeURL</code> except in the case where 
		 * 				<code>addResource()</code> has completed an incomplete URL.
		 * 			</td>
		 *		</tr>
		 *		<tr>
		 *			<td><code>'error'</code></td>
		 *			<td><code>String</code></td>
		 *			<td>A non-human-readable String describing the error.  
		 * 				It is non-<code>null</code> only if <code>success</code> is <code>false</code>.
		 * 			</td>
		 *		</tr>
		 * </table>
		 * 
		 * @param browseDialogTitle
		 * 			The title of the File browse dialog. 
		 * 			This is prefixed by a string to indicate to the user that 
		 * 			this is coming from the navigator.
		 * @param browseDialogFilters
		 * 			An <code>Array</code> of <code>FileFilter</code> objects and will be used in the File browse dialog.
		 * @param relativeURL
		 * 			The path where the resource will be added. Must begin with "navigator/".
		 * 			It must be a relative path, "/" separated, but with no leading "/".
		 * 			If the URL does not have a file extension (that is, it does not contain a '.'), 
		 * 			then the file extension of the added file (if any) will be appended.
		 * 			If the URL ends with "/", then the entire file name of the selected file
		 * 			will be appended.
		 * @param response An optional notification function, as described above.
		 *
		 * @throws Error
		 * 		An <code>Error</code> is thrown if the 
		 * 		permission "Document"/"Modify" has not been granted.
		 *
		 *  @see #writeResource()
		 *  @see #deleteResource()
		 * 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */ 
		public function addResource(browseDialogTitle:String,
									browseDialogFilters:Array,
									relativeURL:String,
									response:Function = null):void
		{
			if ((relativeURL == null) ||
				(collectionManager.host == null) ||
				(collectionManager.host.collection == null))
			{
				// TODO Delay call or invoke response?
				LogHelper.logger.error("addResource called with a null url, host, or collection");
				return;
			}
			collectionManager.host.collection.addResource(browseDialogTitle, browseDialogFilters, relativeURL, response);
		}
		
		/**
		 *  Writes the contents of the specified buffer to the location indicated by the relativeURL.
		 * 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function writeResource(buffer:ByteArray, relativeURL:String, response:Function = null):void
		{
			if (relativeURL == null) {
				LogHelper.logger.error("writeResource called with a null url");
				return;
			}
			var loader:NavURLLoader = loaders[relativeURL] as NavURLLoader;
			if (loader != null) {
				// We've already got a cached loader, so have it write/update itself
				loader.write(buffer, response);
				return;
			}
			
			// We don't have a cached loader, so just pass along the write to the ICollection
			if ((buffer == null) ||
				(collectionManager.host == null) ||
				(collectionManager.host.collection == null))
			{
				// TODO Delay call or invoke response?
				LogHelper.logger.error("writeResource called with a null buffer, host, or collection");
				return;
			}
			collectionManager.host.collection.writeResource(buffer, relativeURL, response);
		}
		
		/**
		 *  Deletes the resource specified by relativeURL. 
		 *  This operation is only allowed when the permission "Document"/"Modify" has been granted.
		 * 
		 *  <p>Accepts an optional callback function as an argument. When the callback function is non-null, 
		 *  it is invoked to indicate the result of the operation. The callback function receives a single
		 *  argument, an Object, and has the following form: </p>
		 *  <pre>    function callback(response:Object):void</pre>
		 *  <p>The response parameter contains properties that describe the response. These include:</p>
		 *  <table class="innertable">
		 *   <tr>
		 *     <th>Property</th>
		 *     <th>Type</th>
		 *     <th>Value</th>
		 *   </tr>
		 *   <tr>
		 *     <td>'success'</td>
		 *     <td>Boolean</td>
		 *     <td><code>true</code> if the deletion succeeded, <code>false</code>otherwise</td>
		 *   </tr>    
		 *   <tr>
		 *     <td>'error'</td>
		 *     <td>String</td>
		 *     <td>A String describing the error. It is null if success is <code>true</code></td>
		 *   </tr>
		 *   <tr>
		 *     <td>'relativeURL'</td>
		 *     <td>String</td>
		 *     <td>The value of the <code>relativeURL></code> parameter in the function call.</td>
		 *   </tr>
		 * </table>
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function deleteResource(relativeURL:String, response:Function = null):void
		{
			if ((relativeURL == null) ||
				(collectionManager.host == null) ||
				(collectionManager.host.collection == null))
			{
				// TODO Delay call or invoke response?
				LogHelper.logger.error("deleteResource called with a null url, host, or collection");
				return;
			}
			collectionManager.host.collection.deleteResource(relativeURL, response);
			
			// This resource has been deleted, so remove it from our cache
			loaders[relativeURL] = null;
		}
		
		/**
		 * Returns a <code>NavURLLoader</code> suitable for loading the given resource URL,
		 * and providing a binding to its resulting data. The loader begins loading immediately.
		 * If event listeners are provided by the caller, then they will be invoked on the appropriate
		 * events, and should not be removed by the caller.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function getResource(relativeURL:String,
									binaryFormat:Boolean,
									completeListener:Function = null,
									ioErrorListener:Function = null,
									securityErrorListener:Function = null):NavURLLoader
		{
			var loader:NavURLLoader = loaders[relativeURL] as NavURLLoader;
			if (loader != null) {
				;// fixes warning
				CONFIG::logging{
					if (binaryFormat && (loader.dataFormat != URLLoaderDataFormat.BINARY)) {
						LogHelper.logger.warn("Mismatched format in getLoader");
					}
				}
			} else {
				loader = new NavURLLoader(relativeURL, binaryFormat);
				loaders[relativeURL] = loader;
				referenceResource(relativeURL);
			}
			
			var defaultCompleteListener:Function;
			var defaultIOErrorListener:Function;
			var defaultSecurityErrorListener:Function;
			
			var removeEventListeners:Function = function():void
			{
				loader.removeEventListener(Event.COMPLETE, defaultCompleteListener);
				loader.removeEventListener(IOErrorEvent.IO_ERROR, defaultIOErrorListener);
				loader.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, defaultSecurityErrorListener);
			}
			
			defaultCompleteListener = function (event:Event):void
			{
				removeEventListeners();
				if (completeListener != null) {
					completeListener(event);
				}
			}
			defaultIOErrorListener = function (event:IOErrorEvent):void
			{
				removeEventListeners();
				if (ioErrorListener != null) {
					ioErrorListener(event);
				}
			}
				
			defaultSecurityErrorListener = function (event:SecurityErrorEvent):void
			{
				removeEventListeners();
				if (securityErrorListener != null) {
					securityErrorListener(event);
				}
			}
				
			
			loader.addEventListener(Event.COMPLETE, defaultCompleteListener);
			loader.addEventListener(IOErrorEvent.IO_ERROR, defaultIOErrorListener);
			loader.addEventListener(SecurityErrorEvent.SECURITY_ERROR, defaultSecurityErrorListener);
			loader.load(new URLRequest(relativeURL));
			return loader;
		}
		
		/**
		 *  Deletes resources from the collectiont that have been referenced,but whose refcounts have gone
		 *  to zero. If purgeUnreferenced is set to true, then resources that have not been referenced at all
		 *  will also be deleted.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function purgeResources(purgeUnreferenced:Boolean = false):void
		{
			var relativeURL:String;
			var deletedResources:Array = new Array();
			for (relativeURL in references) {
				if (references[relativeURL] as uint == 0) {
					deletedResources.push(relativeURL);
					deleteResource(relativeURL);
				}
			}
			if (purgeUnreferenced) {
				// TODO enumerate resources and delete those that aren't referenced.
			}
			for each (relativeURL in deletedResources) {
				delete references[relativeURL];
			}
		}
		
		/**
		 *  Increments by one the reference count for the specified relativeURL
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function referenceResource(relativeURL:String):void
		{
			if ((relativeURL == null) || (relativeURL == "")) {
				return;
			}
			var refCount:uint = references[relativeURL];
			references[relativeURL] = refCount + 1;
		}
		
		/**
		 *  Decrements by one the reference count for the specified relativeURL.
		 *  If the tryToDelete flag is set to <code>true</code> and the reference
		 *  count, after being decremented, is 0, the resource is also deleted.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function dereferenceResource(relativeURL:String, tryToDelete:Boolean = false):void
		{
			if ((relativeURL == null) || (relativeURL == "")) {
				return;
			}
			var refCount:uint = references[relativeURL];
			if (refCount < 1) {
				CONFIG::logging{
					LogHelper.logger.warn("unmatched dereference for resource: "+relativeURL);
				}
				return;
			}
			references[relativeURL] = (refCount - 1);
			if (tryToDelete && (refCount == 1)) {
				deleteResource(relativeURL);
			}
		}
		
		/**
		 *  Set the reference count (initially to 0) for the relativeURL passed,
		 *  if the reference count has not already been set. This is functionally
		 * 	equivalent to incrementing and decrementing the refcount.
		 * 
		 * 	A resource with a reference count of zero is more likely to be purged
		 * 	by a call to purgeResources than a resource with no refcount.
		 * 
		 * 	@see #purgeResources
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function touchResource(relativeURL:String):void
		{
			if ((relativeURL == null) || (relativeURL == "")) {
				return;
			}
			if (references[relativeURL] == null) {
				references[relativeURL] = 0;
			}
		}
		
		/**
		 *  An event listener that handles host changed events for resources.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function onNavHostEvent(event:CollectionManagerEvent):void
		{
			if (event.kind == CollectionManagerEvent.HOST_CHANGED_KIND) {
				// TODO handle pending events, and maybe reset references?
			} else if (event.kind == CollectionManagerEvent.NEEDS_SAVE_KIND) {
				// TODO I'm not sure how to account for the dirty flag here.
				//	If the doc is being closed, then I only really want to purge if the doc is already dirty.
				//	On the other hand, if the doc is being saved, then I want to purge regardless.
				//	For now I always purge but never set dirty, so for case #1 we won't prompt to save, but
				//	for case #2 we'll successfully purge.
				if (purgeOnNeedsSave) {
					purgeResources();
				}
			}
		}
		
		private var resourceAlert:NavAlert;
		private var resourceAlertBaseMessage:String;
		private var resourceEntries:Vector.<String>;
		
		/**
		 * Displays an alert as a popup with the given parent indicating the the given
		 * resource entry cannot be found. Returns true if the popup is displayed;
		 * false otherwise.
		 */
		public function showMissingResourcePopup(entry:String, parent:DisplayObject):Boolean
		{
			if ((resourceAlertBaseMessage == null) || (resourceAlertBaseMessage == "")) {
				resourceAlertBaseMessage = StringManager.instance.getLocalizedString("resourceError");
				if (resourceAlertBaseMessage == null) {
					return false;
				}
				while (resourceAlertBaseMessage.indexOf("%newline%") >= 0) {
					resourceAlertBaseMessage = resourceAlertBaseMessage.replace("%newline%", "\n");
				}
			}
			if (resourceEntries == null) {
				resourceEntries = new Vector.<String>;
			}
			resourceEntries.push(entry);
			var entryString:String = resourceEntries[0];
			for (var entryIndex:uint = 1; entryIndex < resourceEntries.length; ++entryIndex) {
				entryString += "\n";
				entryString += resourceEntries[entryIndex];
			}
			var message:String = resourceAlertBaseMessage.replace("%files%", entryString);
			
			if (resourceAlert != null) {
				// Adjust current alert to include this resource
				resourceAlert.message = message;
			} else if (parent != null) {
				resourceAlert = NavAlert.showPopup(parent, message);
				if (resourceAlert != null) {
					resourceAlert.addEventListener(Event.REMOVED_FROM_STAGE, onResourceAlertClose);
				}
			} else {
				return false;
			}
			return true;
		}
		
		private function onResourceAlertClose(event:Event):void
		{
			if (resourceAlert == null) {
				return;
			}
			while (resourceEntries.length > 0) {
				resourceEntries.pop();
			}
			resourceAlert.removeEventListener(Event.REMOVED_FROM_STAGE, onResourceAlertClose);
			resourceAlert = null;
		}
		

	}
}
