/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.managers
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.events.JobQueueEvent;
	import com.adobe.portfolio.events.ThumbnailManagerEvent;
	import com.adobe.portfolio.previews.Preview;
	import com.adobe.portfolio.previews.PreviewFactory;
	import com.adobe.portfolio.utils.FileUtil;
	import com.adobe.portfolio.utils.JobQueue;
	import com.adobe.portfolio.utils.OrderedDictionary;
	
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.TimerEvent;
	import flash.geom.Point;
	import flash.utils.Dictionary;
	import flash.utils.Timer;
	
	import mx.events.PropertyChangeEventKind;
	import mx.logging.ILogger;
	import mx.logging.Log;
	import mx.utils.UIDUtil;
	
	/**
	 *  Dispatched when the aspectRatio of an attachment changes.
	 *
	 *  @eventType com.adobe.portfolio.events.ThumbnailManagerEvent
	 */
	[Event(name="aspectRatioChanged", type="com.adobe.portfolio.events.ThumbnailManagerEvent")]

	/**
	 * The ThumbnailManager caches and manages thumbnail images of attachments.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */	
	public class ThumbnailManager extends EventDispatcher
	{
		CONFIG::logging {
			protected static const BYPASS_ACROBAT:Boolean = false; 
		}
		
		
		/**
		 *  The fit width of static cached thumbnails
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const SMALL_THUMB_FIT_WIDTH:int = 128;
		
		/**
		 *  The fit height of static cached thumbnails
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const SMALL_THUMB_FIT_HEIGHT:int = 128;
		
		/**
		 *  The size of "small" previews supplied by Acrobat.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const SMALL_PREVIEW_THUMBSIZE:int = 512;
		
		/**
		 *  The priority that allows any thumbnail job to be executed.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const ANY_THUMB_PRIORITY:int = int.MIN_VALUE;

		/**
		 *  The nominal priority of idle thumb generation
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const IDLE_THUMB_PRIORITY:int = -1;
		
		/**
		 *  The nominal priority of dynamic thumb generation
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const DYNAMIC_THUMB_PRIORITY:int = 0;
		
		/**
		 *  The nominal priority of static thumb requests
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const STATIC_THUMB_PRIORITY:int = 1;

		/**
		 *  The nominal priority of preview thumb requests
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const PREVIEW_THUMB_PRIORITY:int = 10;
		
		protected static var smallPreviewDict:Dictionary;
		
		private static const logger:ILogger = Log.getLogger("ThumbnailMgr");
		
		private static var _token:Object;
		private static var _instance:ThumbnailManager;
		
//		/**
//		 * A timer used as a heuristic to determine when dynamic thumbnails should be enabled.
//		 */
//		private var enableDynamicThumbnailsTimer:Timer;
//		private function onReadyToEnableDynamicThumbnails(event:TimerEvent):void
//		{
//			enableDynamicThumbnailsTimer.removeEventListener(TimerEvent.TIMER_COMPLETE, onReadyToEnableDynamicThumbnails);
//			enableDynamicThumbnailsTimer = null;
//			enableDynamicThumbnails = true;
//		}
		
		protected var STATIC_CACHE_MAX_MEM_CACHE_SIZE:int = 200;
		protected var DYNAMIC_CACHE_MAX_MEM_CACHE_SIZE:int = 50;
		
		protected var staticThumbCache:OrderedDictionary;
		protected var dynamicThumbCache:OrderedDictionary;
		protected var pendingRequests:Dictionary;
		protected var pendingStaticRequests:Dictionary;
		protected var timeoutTimer:Timer;

		/**
		 *  The thumbnail job queue active before dynamic thumbs are enabled.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var initialJobQueue:JobQueue;
		
		/**
		 *  The thumbnail job queue active after dynamic thumbs are enabled.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected var jobQueue:JobQueue;

		[Embed("../../../../assets/FolderImage_simple.png")]
		protected var folderImage:Class;
		[Embed("../../../../assets/AX_PortSWF_Xl_N.png")]
		protected var swfImage:Class;
		[Embed("../../../../assets/AX_PortMov_Xl_N.png")]
		protected var videoImage:Class;
		[Embed("../../../../assets/AX_PortAudio_Xl_N.png")]
		protected var audioImage:Class;
		
		private var _initialQueueIsFIFO:Boolean = true;
		private var _enableDynamicThumbnails:Boolean = false;
		private var _minThumbQueuePriority:int = ANY_THUMB_PRIORITY;
		private var _pauseQueue:Boolean;
		
		private var _cacheThumbFitSize:Point;
		private var aspectRatioMap:Dictionary = new Dictionary();
		private var aspectRatioMediaMap:Dictionary = new Dictionary();
		private var collectionMgr:CollectionManager;
		
		
		/**
		 *  Returns the thumbnailManager instance; it creates one if is does not exist already.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function get instance():ThumbnailManager 
		{
			if (!_instance) {
				_instance = new ThumbnailManager(_token = new Object());
			}
			return _instance;
		}
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */	
		public function ThumbnailManager(token:Object) 
		{
			if (_token != token)
				throw new Error("can't construct this object");
			_cacheThumbFitSize = new Point(512, 512);
			staticThumbCache = new OrderedDictionary(); // don't use weak keys 
			dynamicThumbCache = new OrderedDictionary(); // don't use weak keys 
			pendingRequests = new Dictionary();
			pendingStaticRequests = new Dictionary();
			collectionMgr = CollectionManager.instance;
			collectionMgr.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onRootItemQueueChange);
			initialJobQueue = new JobQueue("initialThumbnailQueue", initialQueueIsFIFO, false, 5);
			initialJobQueue.addEventListener(JobQueueEvent.QUEUE_IS_EMPTY,
				function (e:Event):void {
					initialJobQueue = null;
					CONFIG::logging {
						logger.debug("starting primary job queue");
					}
					jobQueue.startJobTimer();
				}
			);
			initialJobQueue.minExecutionPriority = STATIC_THUMB_PRIORITY;
			jobQueue = new JobQueue("thumbnailQueue", false, false, 25);
			jobQueue.minExecutionPriority = STATIC_THUMB_PRIORITY;
		}
		
		/**
		 *  Returns the aspect ratio for the specified attachment.
		 *  If object is null then it returns NaN if allowNaN is <code>true</code>,
		 *  or 1 if allowNaN is <code>false</code>
		 * 
		 * @param attachment the attachment for which you are requesting the aspect ratio
		 * @param mediaValue true if the attachment is media content (audio/video)
		 * @param allowNaN if <code>true</code> then the defaultApectRatio is allowed to be NaN
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getAspectRatio(attachment:IAttachment, mediaValue:Boolean = false, allowNaN:Boolean = false):Number
		{
			var defaultAspectRatio:Number = allowNaN ? Number.NaN : 1;
			
			if (attachment == null) {
				return defaultAspectRatio;
			}
			var currentAspectRatio:Number = mediaValue ? aspectRatioMediaMap[attachment] : aspectRatioMap[attachment] as Number;
			if (isNaN(currentAspectRatio) || (currentAspectRatio <= 0)) {
				return defaultAspectRatio;
			}
			return currentAspectRatio;
		}
		
		/**
		 *  Sets the aspect ratio for the specified attachment.
		 * 
		 *  @param attachment The attachment
		 *  @param aspectRatio The aspect ratio
		 *  @param mediaValue True if the attachment is media content (audio/video)
		 *  @param forceUpdate True to force an update of the aspectRatio
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function setAspectRatio(attachment:IAttachment, aspectRatio:Number, mediaValue:Boolean = false, forceUpdate:Boolean = false):void
		{
			if (attachment == null) {
				return;
			}
			
			// TODO recode to eliminate the dup calls to innerSetAspectRatio
			var currentAspectRatio:Number = mediaValue ? aspectRatioMediaMap[attachment] : aspectRatioMap[attachment] as Number;
			if (forceUpdate) {
				if (currentAspectRatio != aspectRatio) {
					innerSetAspectRatio(attachment, aspectRatio, currentAspectRatio, mediaValue);
				}
			} else if (isNaN(currentAspectRatio) || (currentAspectRatio <= 0)) {
				innerSetAspectRatio(attachment, aspectRatio, currentAspectRatio, mediaValue);
			}
		}
		
		protected function innerSetAspectRatio(attachment:IAttachment, aspectRatio:Number, currentAspectRatio:Number, mediaValue:Boolean):void 
		{
			if (mediaValue) {
				aspectRatioMediaMap[attachment] = aspectRatio;
			} else {
				aspectRatioMap[attachment] = aspectRatio;
			}
			dispatchEvent(new ThumbnailManagerEvent(ThumbnailManagerEvent.ASPECT_RATIO_CHANGED,
				false,
				false,
				PropertyChangeEventKind.UPDATE,
				mediaValue ? "mediaAspectRatio" : "aspectRatio",
				currentAspectRatio,
				aspectRatio,
				attachment));
		}
		
		/**
		 *  Requests a thumbnail to be generated that will fit into the specified width and height. 
		 * 
		 *  @param id An id that the client can use for tracking purposes. If a request with the same id 
		 * 				is already queued, then the new request will supercede the old request.
		 *  
		 *  @param attachment Specifies the attachment (an IAttachment) for which the thumbnail is generated.
		 *  @param width  The desired width of the thumbnail, in pixels. It must be gt; 0 and &lt;= 2048.
		 *  @param height The desired  height of the thumbnail, in pixels.  It must be &gt; 0 and &lt;= 2048.
		 *  @param pageNumber The desired page number of the attachment. It must be &gt;= 1.
		 *  @param callback A callback function that will be called when the thumbnail has been generated.  
		 *  <p>It has the following signature and parameters:</p>
		 * 
		 *  <p><code>callback(id:*, width:Number, height:Number, thumbnail:Bitmap)</code></p>
		 * 
		 *  <p><b>callback parameters</b>:</p>
		 * 
		 *  <table class="innertable">
		 *    <tr>
		 *      <th>Parameter:Type</th>
		 * 	    <th>Value</th>
		 *    </tr>
		 *    <tr>
		 *      <td>id:String</td>
		 * 	    <td>the value of the id param passed to requestThumbnail</td>
		 *    </tr>		
		 *    <tr>
		 *      <td>width:Number</td>
		 *  	<td>the value of the width parameter passed to requestThumbnail</td>
		 *    </tr>
		 *    <tr>
		 *      <td>height:Number</td>
		 * 	   <td>the value of the height param to requestThumbnail</td>
		 *    </tr>	
		 *    <tr>
		 *      <td>pageNumber:uint</td>
		 *   	<td>the value of the pageNumber param to requestThumbnail</td>
		 *    </tr>
		 *    <tr>
		 *      <td>thumbnail:Bitmap</td>
		 *      <td>the generated thumbnail</td>
		 *    </tr>		
  	 	 *  </table>
		 * 	
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function requestThumbnail(id:*, attachment:IAttachment, width:Number, height:Number, 
										 pageNumber:uint, callback:Function, 
										 downSampleOK:Boolean = true, upSampleOK:Boolean = true,
										 queuePriority:int = 0):void 
		{
			if (!attachment)
				return;
			
			CONFIG::logging {
				logEntry("entry: ", id, attachment, width, height, pageNumber, queuePriority);
			}
			
//			if (!enableDynamicThumbnails) {
//				// If we haven't yet enabled dynamic thumbnails, go ahead and do so once we've gone the alloted
//				// time without a request
//				if (enableDynamicThumbnailsTimer == null) {
//					enableDynamicThumbnailsTimer = new Timer(1000, 1);
//					enableDynamicThumbnailsTimer.addEventListener(TimerEvent.TIMER_COMPLETE, onReadyToEnableDynamicThumbnails);
//					enableDynamicThumbnailsTimer.start();
//				} else {
//					enableDynamicThumbnailsTimer.reset();
//					enableDynamicThumbnailsTimer.start();
//				}
//			}
			
			width = Math.floor(width);
			height = Math.floor(height);
			
			if (downSampleOK) {
				width = validateThumbSize(width, downSampleOK);
				height = validateThumbSize(height, downSampleOK);
			}
			
			var cacheObject:CacheObject = new CacheObject(id, attachment, width, height, pageNumber, downSampleOK, upSampleOK, queuePriority);
			var pending:Boolean = isPending(cacheObject.thumbKey, pendingRequests);
			if (pending) {
				// If we've already got pending requests, listen for them as well in case
				// we get a better thumbnail down the road
				addToPending(cacheObject.thumbKey, callback, pendingRequests);
			}
			
			var thumbnail:Bitmap = getThumbFromDynamicCache(cacheObject.thumbKey);
			if (thumbnail) {
				cacheObject.bitmapData = thumbnail.bitmapData;
				CONFIG::logging {
					logInfo("found cachedThumb: ", cacheObject);
				}
				callback(id, width, height, pageNumber, thumbnail);
				return;
			}
			
			thumbnail = getPredefinedThumbnail(cacheObject.attachment);
			if (thumbnail) {
				// Assume we support dynamic thumbs for types that use predefined thumbnails
				cacheThumb(staticThumbCache, callback, cacheObject, thumbnail);
				return;
			}
			
			thumbnail = getDownsampledThumb(cacheObject, downSampleOK);
			if (thumbnail) {
				cacheObject.bitmapData = thumbnail.bitmapData;
				CONFIG::logging {
					logInfo("got downsampledThumb: ", cacheObject);
				}
				cacheThumb(dynamicThumbCache, callback, cacheObject, thumbnail);
				return;
			}
			
			CONFIG::logging {
				if (BYPASS_ACROBAT) {
					var colors:Array = [0xffaaaa, 0xaaffaa, 0xaaaaff];
					var _thumbnail:Bitmap = new Bitmap(new BitmapData(width, height, false, colors[int(Math.random() * 3)]));
					cacheThumb(dynamicThumbCache, callback, cacheObject, _thumbnail);
					return;
				}
			}

			if (getStaticCachedThumb(callback, cacheObject))
				return;
			
			if (pageNumber <= 1) {
				// Go ahead and request the static thumbnail for any size, so we can display it before we get the real one.
				getStaticCachedThumb(callback, cacheObject, true, false);
			}
			
			if (pending) {
				startTimeOutTimer();
				return;
			} else {
				// We're going to make some sort of request, so add to pending regardless
				addToPending(cacheObject.thumbKey, callback, pendingRequests);
			}
			
			performRequest(cacheObject, width, height, pageNumber, callback);
		}
		
		private function get currentJobQueue():JobQueue
		{
			return initialJobQueue ? initialJobQueue : jobQueue;
		}
		
		private function performRequest(cacheObject:CacheObject, width:Number, height:Number, pageNumber:uint, callback:Function):void
		{
			var attachment:IAttachment = cacheObject.attachment;
			if (attachment == null) {
				return;
			}
			
			CONFIG::logging {
				logger.info("pushJob: " + cacheObject);
			}

			var uid:String = UIDUtil.getUID(cacheObject.id);
			currentJobQueue.pushJob(cacheObject.attachment.fileName + "_p" + cacheObject.pageNumber + "_" + uid, cacheObject,
				function (jobKey:Object, cacheObject:CacheObject):void {
					// If requestPageThumbnail isn't supported with the current API version, then we'll just rely on
					// the built-in fallback functionality to give us something, and we'll cache that.
					// We must convert between our 1-based index and requestPageThumbnail's 0-based index.
					CONFIG::logging {
						logInfo("calling requestPageThumbnail: ", cacheObject);
					}
					attachment.requestPageThumbnail(cacheObject.id, pageNumber-1, width, height,
						function (id:*, pageNumber:uint, width:uint, height:uint, 
								  thumbnail:Bitmap):void 
						{
							pageNumber += 1;
							if (thumbnail)
								cacheObject.bitmapData = thumbnail.bitmapData;
							CONFIG::logging {
								if (thumbnail == null) {
									logger.warn("thumbnail was null for " + attachment.fileName);
								} else if (thumbnail.bitmapData == null) {
									logger.warn("thumbnail.bitmapData was null for " + attachment.fileName);
								} else {
									logInfo("got new thumb: ", cacheObject);
								}
							}
							if (thumbnail == null) {
								getStaticCachedThumb(callback, cacheObject, true);
							} else {
								cacheThumb(dynamicThumbCache, callback, cacheObject, thumbnail);
							}
							startJobs();
						} 
					);
//					startJobs();
					startTimeOutTimer();
				}, cacheObject.queuePriority
			);
			startJobs();
		}
		
		protected function startTimeOutTimer():void
		{
			if (!timeoutTimer) {
				timeoutTimer = new Timer(2000);
				timeoutTimer.addEventListener(TimerEvent.TIMER, onTimeout, false, 0, true);
			} else {
				timeoutTimer.reset();
			}
			timeoutTimer.start();
		}
		
		protected function onTimeout(e:Event):void {
			CONFIG::logging {
				logger.debug("onTimeout: currentJobQueue.hasJobs()= " + currentJobQueue.hasJobs());
			}			
			if (currentJobQueue.hasJobs()) {
				startJobs();
			} else {
				timeoutTimer.stop();
				timeoutTimer = null;
			}
		}
		
		/**
		 *  Returns the predefined thumbnail images for the specified attachment,
		 *  such as folders, SWF files, video, and audio files.  If the there is no
		 *  predefined image for the attachment type, it returns null.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function getPredefinedThumbnail(attachment:IAttachment):Bitmap
		{
			if (attachment.isFolder) {
				return new folderImage();
			}
			
			// Check the mimeType/extension to see if we should be using one of our baked assets
			var mimeType:String = attachment.mimeType;
			var extension:String = FileUtil.getExtension(attachment.fileName);
			if ((mimeType == "application/x-shockwave-flash") ||
				((extension != null) && (extension.toLowerCase() == "swf")))
			{
				return new swfImage();
			}
			var previewState:String = PreviewFactory.instance.previewStateFromMimeType(mimeType, extension);
			if (previewState == Preview.VIDEO) {
				return new videoImage();
			} else if (previewState == Preview.AUDIO) {
				return new audioImage();
			}
			
			return null;
		}
		
		
		/**
		 *  Returns <code>true</code> if the specified attachment has a dynamic thumb cached for the give width, height and pageNumber
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function isThumbnailCached(attachment:IAttachment, width:Number, height:Number, pageNumber:uint):Boolean 
		{
			var thumbKey:String = CacheObject.createThumbKey(attachment,width,height,pageNumber);
			var cacheObject:CacheObject = dynamicThumbCache.getValue(thumbKey) as CacheObject;
			
			return cacheObject != null;
		}
		
		
		CONFIG::logging {
			protected function logEntry(prefix:String, id:*, attachment:IAttachment, width:Number, height:Number, pageNumber:uint, priority:int):void 
			{
				logger.info (prefix + id + " : " + attachment.fileName + " w,h= " + width + " , " + height + " pn= " + pageNumber
					+ " priority= " + priority);
			}
		
			protected function logInfo(prefix:String, cacheObject:CacheObject):void 
			{
				CONFIG::logging {
					logger.info (prefix + cacheObject);
				}
			}
		}
		
		
		protected function isPending(thumbKey:Object, requests:Dictionary):Boolean {
			return requests[thumbKey] is Array;
		}
		
		
		protected function addToPending(thumbKey:Object, callback:Function, requests:Dictionary):void {
			var pendingArray:Array = requests[thumbKey];
			if (pendingArray)
				pendingArray.push(callback);
			else {
				requests[thumbKey] = [callback];
			}
		}
		
		
		protected function callPendingCallbacks(thumbKey:Object, cacheObject:CacheObject, requests:Dictionary, clearPendingRequests:Boolean = true):void {
			var pendingArray:Array = requests[thumbKey];
			if (!pendingArray)
				return; // a folder
			
			for (var i:int = 1; i < pendingArray.length; ++i) {
				var _callback:Function = pendingArray[i] as Function;
				if (_callback == null)
					continue;
				_callback(cacheObject.id, cacheObject.width, cacheObject.height, cacheObject.pageNumber, wrapBitmapData(cacheObject.bitmapData));
			}
			if (clearPendingRequests) {
				delete requests[thumbKey];
			}
		}
		
		
		protected function cacheThumb(thumbCache:OrderedDictionary, callback:Function, cacheObject:CacheObject, thumbnail:Bitmap, clearPendingRequests:Boolean = true):void 
		{
			if (!thumbnail)
				return;
			
			var isStaticCache:Boolean = thumbCache == staticThumbCache;
			var maxCacheSize:int = isStaticCache ? STATIC_CACHE_MAX_MEM_CACHE_SIZE : DYNAMIC_CACHE_MAX_MEM_CACHE_SIZE;
			
			if (thumbCache.length >= maxCacheSize) {
				var cacheObj:CacheObject = thumbCache.shift();
				CONFIG::logging {
					logInfo("thumb uncached: ", cacheObj);
				}
			}
			cacheObject.bitmapData = thumbnail.bitmapData;
			CONFIG::logging {
				logInfo("thumb cached: ", cacheObject);
			}
			var thumbKey:Object = isStaticCache ? cacheObject.attachment : cacheObject.thumbKey;
			thumbCache.push(thumbKey, cacheObject);

			callback(cacheObject.id, cacheObject.width, cacheObject.height, cacheObject.pageNumber, wrapBitmapData(thumbnail.bitmapData));
			callPendingCallbacks(cacheObject.thumbKey, cacheObject, pendingRequests, clearPendingRequests);
			callPendingCallbacks(cacheObject.attachment, cacheObject, pendingStaticRequests, true);
		}
		
		
		protected function getThumbFromDynamicCache(thumbKey:String):Bitmap
		{
			var cacheObject:CacheObject =  dynamicThumbCache.getValue(thumbKey) as CacheObject;
			var bitmapData:BitmapData = cacheObject ? cacheObject.bitmapData : null;
			return wrapBitmapData(bitmapData);
		}

		protected function getThumbFromStaticCache(attachment:IAttachment):Bitmap
		{
			var cacheObject:CacheObject =  staticThumbCache.getValue(attachment) as CacheObject;
			var bitmapData:BitmapData = cacheObject ? cacheObject.bitmapData : null;
			return wrapBitmapData(bitmapData);
		}

		
		protected function getDownsampledThumb(c:CacheObject, downSampleOK:Boolean):Bitmap
		{
			var attachment:* = c.attachment;
			var width:Number = c.width;
			var height:Number = c.height;
			var pageNumber:uint = c.pageNumber;
			var length:uint = dynamicThumbCache.length;
			for (var i:int = 0; i < length; ++i) {
				var cacheObject:CacheObject = dynamicThumbCache.getValueByIndex(i) as CacheObject;
				if (cacheObject.attachment != attachment)
					continue;
				if (cacheObject.pageNumber != pageNumber)
					continue;
				
				if (cacheObject.width >= width || cacheObject.height >= height) {
					if (!downSampleOK) {
						if (cacheObject.width != width && cacheObject.height != height)
							break;
					}
					if ((cacheObject.bitmapData == null) ||
						((cacheObject.bitmapData.width < width) && (cacheObject.bitmapData.height < height)))
					{
						// The cached bitmap is actually smaller than what we asked for in both dimensions - keep looking
						continue;
					}
					var wRatio:Number = cacheObject.width / width;
					var hRatio:Number = cacheObject.height / height;
					if (wRatio > 4 || hRatio > 4)
						continue; // downsample ratio is too big - keep looking
					cacheObject.downsampled = true;
					
					CONFIG::logging {
						logInfo("found thumb for downsampling", cacheObject);
					}
					
					return wrapBitmapData(cacheObject.bitmapData);
				}
			}
			return null;
		}
		
		
		protected function wrapBitmapData(bitmapData:BitmapData):Bitmap
		{
			if (!bitmapData)
				return null;
			
			var thumbNail:Bitmap = new Bitmap(bitmapData);
			if (thumbNail)
				thumbNail.smoothing = true;
			
			return thumbNail;
		}
		
		protected function getStaticCachedThumb(callback:Function, cacheObject:CacheObject, anySizeOK:Boolean = false, clearPendingRequests:Boolean = true):Boolean
		{
			if (!anySizeOK && (cacheObject.width > SMALL_THUMB_FIT_WIDTH || cacheObject.height > SMALL_THUMB_FIT_HEIGHT))
				return false;
			
			var thumbnail:Bitmap = getThumbFromStaticCache(cacheObject.attachment);
			if (thumbnail) {
				CONFIG::logging {
					cacheObject.bitmapData = thumbnail.bitmapData;
					logInfo("got static cached thumb", cacheObject);
				}
				callback(cacheObject.id, cacheObject.width, cacheObject.height, cacheObject.pageNumber, wrapBitmapData(thumbnail.bitmapData));
				return true;
			}
			
			var pending:Boolean = isPending(cacheObject.attachment, pendingStaticRequests);
			addToPending(cacheObject.attachment, callback, pendingStaticRequests);
			if (pending) {
				return true;
			}

			CONFIG::logging {
				logger.info("pushStaticJob: " + cacheObject);
			}
			var hasJobs:Boolean = currentJobQueue.hasJobs();
			currentJobQueue.pushJob(cacheObject.attachment, cacheObject,
				function (jobKey:Object, job:CacheObject):void {
					CONFIG::logging {
						logInfo("calling requestCachedThumbnail", job);
					}
					job.attachment.requestCachedThumbnail(job.id, 
						function (id:*, bitmap:Bitmap):void {
							if (bitmap && bitmap.bitmapData) {
								CONFIG::logging {
									logInfo("got static thumb from host", job);
								}
								cacheThumb(staticThumbCache, callback, job, wrapBitmapData(bitmap.bitmapData), clearPendingRequests);
							} else if (!anySizeOK) {
								// If we couldn't get a static cached thumbnail, go ahead and request a new one
								// (but not if anySizeOK is true, to avoid an infinite loop)
								performRequest(job, job.width, job.height, 1, callback);
							}
							startJobs();
						}
					);
//					startJobs();
					startTimeOutTimer();
				}, STATIC_THUMB_PRIORITY
			);
			if (!hasJobs)
				startJobs();

			return true;
		}
		
		
		/**
		 *  The preferred fit size for preview thumbs
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get cacheThumbFitSize():Point
		{
			return _cacheThumbFitSize.clone();
		}
		
		
		public function set cacheThumbFitSize(value:Point):void
		{
			var x:Number = validateThumbSize(value.x);
			var y:Number = validateThumbSize(value.y);
			
			if (_cacheThumbFitSize.x == x && _cacheThumbFitSize.y == y)
				return;
			
			_cacheThumbFitSize = new Point(x,y);
		}
		
		
		protected function validateThumbSize(value:Number, downSampleOK:Boolean = true):Number
		{
			// this causes fidelity problems
//			if (downSampleOK)
//				value = Math.ceil(value/ 8) * 8;
			if (value < 128 )
				value = 128;
			if (value > 2048)
				value = 2048;
			
			return value;
		}
		
		protected function startJobs():void
		{
			currentJobQueue.startJobTimer();
		}

		[Bindable]
		/**
		 * True to enable processing of dynamic thumbs (they're still queue'ed regardless of this value).
		 */
		public function get enableDynamicThumbnails():Boolean
		{
			return _enableDynamicThumbnails;
		}

		public function set enableDynamicThumbnails(value:Boolean):void
		{
			CONFIG::logging {
				logger.info("enableDynamicThumbnails= " + value);
			}
			_enableDynamicThumbnails = value;
			if (value) {
				if (initialJobQueue && !initialJobQueue.hasJobs()) {
					initialJobQueue = null;
				}
				
				if (initialJobQueue)
					initialJobQueue.minExecutionPriority = ANY_THUMB_PRIORITY;
				if (jobQueue)
					jobQueue.minExecutionPriority = ANY_THUMB_PRIORITY;
				startJobs();
			}
		}

		/**
		 *  Type of initial queue (FIFO/FILO)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get initialQueueIsFIFO():Boolean
		{
			return _initialQueueIsFIFO;
		}

		public function set initialQueueIsFIFO(value:Boolean):void
		{
			_initialQueueIsFIFO = value;
			if (initialJobQueue) {
				initialJobQueue.fifo = value;
			}
		}
		
		protected function onRootItemQueueChange(e:CollectionManagerEvent):void
		{
			if (e.kind == CollectionManagerEvent.ROOT_ITEMS_QUEUE_BEGIN_KIND)
				enableDynamicThumbnails = false;
			else if (e.kind == CollectionManagerEvent.ROOT_ITEMS_QUEUE_END_KIND)
				enableDynamicThumbnails = true;
		}
		
		
		/**
		 *  The minimum priority required to execute a queued thumb request.  Requests below this priority
		 *  will remain in the queue.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get minThumbQueuePriority():int
		{
			return _minThumbQueuePriority;
		}

		public function set minThumbQueuePriority(value:int):void
		{
			if (_minThumbQueuePriority == value)
				return;
			
			_minThumbQueuePriority = value;
			
			if (initialJobQueue)
				initialJobQueue.minExecutionPriority = value;
			if (jobQueue)
				jobQueue.minExecutionPriority = value;
		}

		/**
		 *  True if Acrobat generates small previews for the specified mimetype.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function isSmallPreview(mimeType:String):Boolean
		{
			if (!smallPreviewDict) {
				smallPreviewDict = new Dictionary();
				smallPreviewDict["application/msword"] = true;
				smallPreviewDict["application/vnd.openxmlformats-officedocument.wordprocessingml.document"] = true;
				smallPreviewDict["application/vnd.ms-excel"] = true;
				smallPreviewDict["application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"] = true;
				smallPreviewDict["application/vnd.ms-powerpoint"] = true;
				smallPreviewDict["application/vnd.openxmlformats-officedocument.presentationml.presentation"] = true;
				smallPreviewDict["text/plain"] = true;
				smallPreviewDict["text/html"] = true;
			}
			return smallPreviewDict[mimeType] != null;
		}

		
		/**
		 *  Pauses or starts the thumbnail queue.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get pauseQueue():Boolean
		{
			return _pauseQueue;
		}

		public function set pauseQueue(value:Boolean):void
		{
			if (_pauseQueue == value)
				return;

			_pauseQueue = value;
			
			if (initialJobQueue)
				initialJobQueue.pause = value;
			if (jobQueue)
				jobQueue.pause = value;
		}

		
	}
}

import com.adobe.portfolio.api.IAttachment;

import flash.display.BitmapData;
import flash.utils.Dictionary;

import mx.logging.ILogger;
import mx.logging.Log;

/** 
 *  @private
 *  The following class, CacheObject, is effectively private,
 *  since it is the second class defined in this file.
 *  Do not document the properties and methods
 */
class CacheObject {
	
	protected static const logger:ILogger = Log.getLogger("ThumbnailMgr");
	
	public var id:*;
	public var attachment:IAttachment;
	public var width:Number;
	public var height:Number;
	public var pageNumber:uint;
	public var bitmapData:BitmapData;
	public var downsampled:Boolean;
	public var thumbKey:String;
	public var downSampleOK:Boolean;
	public var upSampleOK:Boolean;
	public var queuePriority:int;
	
	public function CacheObject(id:*, attachment:IAttachment, width:Number, height:Number, pageNumber:uint, 
								downSampleOK:Boolean, upSampleOK:Boolean, queuePriority:int,
								thumbnailData:BitmapData = null, downsampled:Boolean = false)
	{
		this.id = id;
		this.attachment = attachment;
		this.width = width;
		this.height = height;
		this.pageNumber = pageNumber;
		this.downSampleOK = downSampleOK;
		this.upSampleOK = upSampleOK;
		this.queuePriority = queuePriority;
		this.bitmapData = thumbnailData;
		this.downsampled = downsampled;
		
		thumbKey = createThumbKey(attachment,width,height,pageNumber);
	}
	

	public static function createThumbKey(attachment:IAttachment, width:Number, height:Number, pageNumber:uint):String
	{
		return attachment.path + " w=" + width + " h=" + height + " p=" + pageNumber;
	}

	public function toString():String
	{
		return "[cache: " + id + ": " + attachment.fileName 
			+ " w,h= " + width + " , " + height + " pn= " + pageNumber + " priority= " + queuePriority
			+ (bitmapData ? (" bitmap w,h= " + bitmapData.width + " , " + bitmapData.height) : " <no data>")
			+ "] ";
	}
}
