/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.previews
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.components.NavSkinnableComponent;
	import com.adobe.portfolio.components.ScrollPad;
	import com.adobe.portfolio.events.PreviewEvent;
	import com.adobe.portfolio.managers.ThumbnailManager;
	
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.events.Event;
	import flash.events.MouseEvent;
	import flash.geom.Point;
	import flash.utils.Dictionary;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.collections.ArrayList;
	import mx.controls.Image;
	import mx.logging.ILogger;
	import mx.logging.Log;
	import mx.utils.UIDUtil;
	
	[SkinState("thumbnail")]
	[SkinState("pagingThumbnail")]
	[SkinState("audio")]
	[SkinState("video")]
	[SkinState("thumbnailHover")]
	[SkinState("pagingThumbnailHover")]
	[SkinState("audioHover")]
	[SkinState("videoHover")]

	/**
	 *  Dispatched when the user requests to show info for the previewed content.
	 * 
	 *  @see com.adobe.portfolio.events.PreviewEvent#HANDLE_FILE_INFO
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	[Event(name="handleFileInfo", type="com.adobe.portfolio.events.PreviewEvent")]
	
	/**
	 *  Dispatched when the user requests to extract the previewed content.
	 * 
	 *  @see com.adobe.portfolio.events.PreviewEvent#HANDLE_FILE_EXTRACT
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	[Event(name="handleFileExtract", type="com.adobe.portfolio.events.PreviewEvent")]
	
	/**
	 * 	The Preview class extends NavSkinnableComponent to preview single and multipage documents,
	 * 	as well as audio and video content.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class Preview extends NavSkinnableComponent implements IPreview
	{
		/**
		 * 	The "thumbnail" state of the Preview and its skin. Used to preview most file types.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const THUMBNAIL:String = "thumbnail";

		/**
		 * 	The "pagingThumbnail" state of the Preview and its skin.  Used to preview multipage file types.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const PAGING_THUMBNAIL:String = "pagingThumbnail";

		/**
		 * 	The "audio" state of the Preview and its skin.  Used to preview audio file types.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const AUDIO:String = "audio";

		/**
		 * 	The "video" state of the Preview and its skin.  Used to preview video file types.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const VIDEO:String = "video";
		

		private static const THUMBNAIL_HOVER:String = "thumbnailHover";
		private static const PAGING_THUMBNAIL_HOVER:String = "pagingThumbnailHover";
		private static const AUDIO_HOVER:String = "audioHover";
		private static const VIDEO_HOVER:String = "videoHover";

		private static var toHover:Object;
		private static var fromHover:Object;
		private static var hoverStates:Dictionary;
		private static var normalStates:Dictionary;
		private static var thumbnailStates:Dictionary;
		
		protected static var thumbnailMgr:ThumbnailManager = ThumbnailManager.instance;
		CONFIG::logging {
			protected static var logger:ILogger = Log.getLogger("Preview");
		}
		
		[SkinPart(required="false")]
		/**
		 * 	An optional skin part that's used to display a single thumbnail preview.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public var thumbnail:Image;

		[SkinPart(required="false")]
		/**
		 * 	An optional skin part that's used to display a multi-page preview.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public var scrollPad:ScrollPad;
		
		[SkinPart(required="false")]
		/**
		 * 	An optional skin part that's used to display an audio preview.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public var audio:AudioPreview;

		[SkinPart(required="false")]
		/**
		 * 	An optional skin part that's used to display a video preview.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public var video:VideoPreview;
		
		[SkinPart(required="false")]
		/**
		 * 	An optional skin part that's used to display an icon.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public var icon:Image;
		
		protected var thumbsData:Object;
		protected var modelWatcher:ChangeWatcher;
		protected var lastAttachmentRequest:IAttachment;
		protected var deferredRequest:Object;
		
		private var _model:IPreviewModel;
		private var _smallPreview:Boolean;
		private var _aspectRatio:Number;
		private var _thumbnailPriority:int;
		private var _pageNavigationEnabled:Boolean;

		/**
		 * 	Constructor.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function Preview() {
			_aspectRatio = 1;
			_smallPreview = true;
			doubleClickEnabled = true;
			_thumbnailPriority = ThumbnailManager.DYNAMIC_THUMB_PRIORITY;
			_pageNavigationEnabled = false;
		}
		
		override protected function initStates():void {
			if (hoverStates)
				return;
			
			normalStates = new Dictionary();
			normalStates[THUMBNAIL] = true;
			normalStates[PAGING_THUMBNAIL] = true;
			normalStates[AUDIO] = true;
			normalStates[VIDEO] = true;
			
			hoverStates = new Dictionary();
			hoverStates[THUMBNAIL_HOVER] = true;
			hoverStates[PAGING_THUMBNAIL_HOVER] = true;
			hoverStates[AUDIO_HOVER] = true;
			hoverStates[VIDEO_HOVER] = true;
			
			thumbnailStates = new Dictionary();
			thumbnailStates[THUMBNAIL] = true;
			thumbnailStates[THUMBNAIL_HOVER] = true;
			thumbnailStates[PAGING_THUMBNAIL] = true;
			thumbnailStates[PAGING_THUMBNAIL_HOVER] = true;
			
			toHover = new Object();
			toHover[THUMBNAIL] = THUMBNAIL_HOVER;
			toHover[PAGING_THUMBNAIL] = PAGING_THUMBNAIL_HOVER;
			toHover[AUDIO] = AUDIO_HOVER;
			toHover[VIDEO] = VIDEO_HOVER;
			
			fromHover = new Object();
			fromHover[THUMBNAIL_HOVER] = THUMBNAIL;
			fromHover[PAGING_THUMBNAIL_HOVER] = PAGING_THUMBNAIL;
			fromHover[AUDIO_HOVER] = AUDIO;
			fromHover[VIDEO_HOVER] = VIDEO;
		}
		
		override protected function attachSkin():void {
			super.attachSkin();
			addListener(this, MouseEvent.ROLL_OVER, handleRollOver); 
			addListener(this, MouseEvent.ROLL_OUT, handleRollOut);
		}
		
		override protected function partAdded(partName:String, instance:Object):void {
			switch (instance) {
				
				case scrollPad:
					addModelWatcher();
					addWatcher(BindingUtils.bindProperty(scrollPad, "showButtons", this, "smallPreview"));
					addListener(scrollPad, Event.COMPLETE, onScrollComplete);
					break;
				
				case video:
					addWatcher(BindingUtils.bindProperty(video, "visible", parent.parent, "visible"));
					addWatcher(BindingUtils.bindSetter(videoAspectChanged, video, "aspectRatio"));
					break;
				
				case audio:
					addWatcher(BindingUtils.bindProperty(audio, "visible", parent.parent, "visible"));
					break;
				
				case icon:
					if (lastAttachmentRequest != null) {
						lastAttachmentRequest.addEventListener("iconChanged", onIconChange);
						setIcon(lastAttachmentRequest.icon);
					}
					break;

				default:
					break;
			}
		}
		
		override protected function partRemoved(partName:String, instance:Object):void {
			switch (instance) {
				case icon:
					if (lastAttachmentRequest != null) {
						lastAttachmentRequest.removeEventListener("iconChanged", onIconChange);
					}
					break;
			}
			super.partRemoved(partName, instance);
		}
		
		private function addModelWatcher():void {
			if (modelWatcher) {
				modelWatcher.unwatch();
				modelWatcher = null;
			}
			if (!previewMultiPageModel)
				return;
			modelWatcher = BindingUtils.bindProperty(scrollPad, "model", this, "model");
			addWatcher(modelWatcher);
		}
		
		/**
		 *  @private
		 */			
		private function setIcon(value:Bitmap):void
		{
			if (icon == null) {
				return;
			}
			if ((value == null) || (value.bitmapData == null)) {
				icon.source = null;
				return;
			}
			icon.source = new Bitmap(value.bitmapData);
		}

		private function onIconChange(event:Event):void
		{
			var attachment:IAttachment = event.target as IAttachment;
			if (attachment != null) {
				setIcon(attachment.icon);
			}
		}
		
		private function handleRollOver(e:Event):void {
			
			if (skinState == null) {
				removeEventListener(e.type, arguments.callee);
				return;
			}
			var state:String = skinState;
			if (hoverStates[state])
				return;
			
			if (video && video.playing)
				return;
			
			skinState = toHover[state];
			if (skinState == null)
				throw new Error("couldn't find toHover state for: " + state);				
		}
		
		private function handleRollOut(e:Event):void {
			if (skinState == null) {
				removeEventListener(e.type, arguments.callee);
				return;
			}
			var state:String = skinState;
			if (normalStates[state])
				return;
			
			if (video && video.playing)
				return;
			
			skinState = fromHover[state];
			if (skinState == null)
				throw new Error("couldn't find fromHover state for: " + state);

		}
		
		private function get previewMultiPageModel():IPreviewMultiPageModel {
			return model as IPreviewMultiPageModel;
		}
			
		private function videoAspectChanged(videoAspectRatio:Number):void {
			if (videoAspectRatio == 1)
				return;
			
			var w:Number = getPreferredBoundsWidth();
			var h:Number = getPreferredBoundsHeight();
			aspectRatio = w / h;
		}

		[Bindable(name="modelChanged")]
		/**
		 * 	The current preview model.
		 * 
	 	 *  @see com.adobe.portfolio.events.PreviewEvent#MODEL_CHANGED
		 *  @see IPreviewModel
		 *  @see IPreviewSinglePageModel
		 *  @see IPreviewMultiPageModel
		 *  @see IPreviewMediaModel
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get model():IPreviewModel {
			return _model;
		}
		
		public function set model(value:IPreviewModel):void
		{
			if (_model == value)
				return;
			_model = value;
			
			if (modelWatcher)
				modelWatcher.unwatch();
			if (previewMultiPageModel && scrollPad) {
				addModelWatcher();
			} else {
				modelWatcher = null;
			}

			dispatchEvent(new PreviewEvent(PreviewEvent.MODEL_CHANGED));
		}
		
		override protected function getCurrentSkinState():String
		{
			return skinState;
		}
		
		/**
		 * 	Return true if the preview is a single or multipage preview (not an audio or video).
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function isThumbnailState():Boolean 
		{
			var result:* = thumbnailStates[getCurrentSkinState()];
			return result ? true : false;
		}

		[Bindable(name="aspectRatioChanged")]
		/**
		 * 	The aspect ratio of the preview
		 * 
	 	 *  @see com.adobe.portfolio.events.PreviewEvent#ASPECT_RATIO_CHANGED
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get aspectRatio():Number
		{
			return _aspectRatio;
		}

		public function set aspectRatio(value:Number):void
		{
			if (_aspectRatio == value)
				return;
			_aspectRatio = value;
			dispatchEvent(new PreviewEvent(PreviewEvent.ASPECT_RATIO_CHANGED));
		}

		[Bindable]
		/**
		 * 	Whether the preview is a "small" preview that should not have next/prev buttons in the multipage case.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get smallPreview():Boolean
		{
			return _smallPreview;
		}

		public function set smallPreview(value:Boolean):void
		{
			_smallPreview = value;
		}
		
		/**
		 * 	Override to propagate the doubleClickEnabled property to sub-components of Preview.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		override public function set doubleClickEnabled(value:Boolean):void
		{
			if (audio)
				audio.doubleClickEnabled = value;
			if (video)
				video.doubleClickEnabled = value;
			if (scrollPad)
				scrollPad.doubleClickEnabled = value;
			super.doubleClickEnabled = value;	
		}
		
		override protected function measure():void
		{
			// If we're a media format, use the measured dimensions of the media player
			if (model is PreviewMediaModel) {
				if (video != null) {
					measuredWidth = video.measuredWidth;
					measuredHeight = video.measuredHeight;
					return;
				}
				if (audio != null) {
					measuredWidth = audio.measuredWidth;
					measuredHeight = audio.measuredHeight;
					return;
				}
			}
			
			// Otherwise always default to the small thumb size
			measuredWidth = ThumbnailManager.SMALL_THUMB_FIT_WIDTH;
			measuredHeight = ThumbnailManager.SMALL_THUMB_FIT_HEIGHT;
		}
		
		/**
		 * 	Stop the playback of media content in this preview.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function stopPlayback():void {
			if (audio)
				audio.pause();
			if (video)
				video.pause();
		}

		/**
		 * 	Generate a thumbnail.
		 * 
		 * 	@param id An arbitrary id that is used to prune multiple duplicate requests.  
		 *  @param attachment The attachment used to generate the thumbnail.
		 *  @param previewSize The desired size of the thumbnail in pixels.
		 *  @param downSampleOK A boolean that indicates whether it's ok to return a downsampled thumb
		 *  @param onThumbComplete An optional callback function that is called when the thumbnail generation is complete.  It takes the form
		 * 	<p><code>
		 * 	function onThumbComplete(pageNumber:uint, retWidth:Number, retHeight:Number):void
		 *  </code></p>
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function generateThumbnail(id:*,
										  attachment:IAttachment, 
										  previewSize:Point = null,
										  downSampleOK:Boolean = true,
										  onThumbComplete:Function = null):void 
		{
			if (!model) {
//				logger.warn("generateThumbnail: early exit: model == null" + " fileName= " + attachment.fileName);
				return;
			}
			if (!isThumbnailState()) {
//				logger.warn("generateThumbnail: early exit: isThumbnailState == false" + " fileName= " + attachment.fileName);
				return;
			}
			
			if (previewSize == null) {
				if (width > 0 && height > 0)
					previewSize = new Point (width, height);
				else
					previewSize = new Point(128, 128);
			}
			
			if (previewMultiPageModel) {
				requestMultiPageThumbnails(id, attachment, previewSize, downSampleOK, onThumbComplete);
			} else {
				requestThumbnail(id, attachment, previewSize, 1, downSampleOK, thumbnailPriority, onThumbComplete);
			}
		}
		
		protected function requestThumbnail(id:*, attachment:IAttachment, previewSize:Point, pageNumber:uint, downSampleOK:Boolean, 
											priority:int, onThumbComplete:Function = null):void 
		{
			if (!previewSize || previewSize.x == 0 || previewSize.y == 0)
				return;

			if (pageNumber < 1 || !attachment)
				return;
			
			if (previewMultiPageModel && pageNumber > previewMultiPageModel.pageCount)
				return;
			
			CONFIG::logging {
				log(attachment, "attachment: " + attachment + " previewSize=" + previewSize + "pageNumber= " + pageNumber, "requestThumbnail entry:");
			}
			
			if (lastAttachmentRequest != attachment) {
				if ((icon != null) && (lastAttachmentRequest != null)) {
					lastAttachmentRequest.removeEventListener("iconChanged", onIconChange);
				}
				lastAttachmentRequest = attachment;
				if ((icon != null) && (lastAttachmentRequest != null)) {
					lastAttachmentRequest.addEventListener("iconChanged", onIconChange);
					setIcon(lastAttachmentRequest.icon);
				}
			}
			
			thumbnailMgr.requestThumbnail(id, attachment, previewSize.x, previewSize.y, pageNumber,
				function (id:*, _width:Number, _height:Number, pageNumber:uint, thumbnail:Bitmap):void 
				{
					if (!model)
						return;
					
					if (!thumbnail || !thumbnail.bitmapData)
						return;
					
					if (lastAttachmentRequest != attachment) {
						// This must be an old request
						return;
					}
					
					// We only display the first page, so ignore thumbnails generated for other pages
					if (!previewMultiPageModel && pageNumber > 1) {
						CONFIG::logging {
							log(attachment, "singe-page preview called back with non-first page: " + pageNumber + " w x h=" + thumbnail.bitmapData.width + "," + thumbnail.bitmapData.height, "requestThumbnail");
						}
						return;
					}
					
					var retWidth:Number = thumbnail.bitmapData.width;
					var retHeight:Number = thumbnail.bitmapData.height;
					var modelThumbnail:Bitmap = getModelThumbnail(pageNumber);
					
					if (modelThumbnail && modelThumbnail.bitmapData) {
						CONFIG::logging {
							log(attachment, " modelThumbnail w x h=" + modelThumbnail.bitmapData.width + "," + modelThumbnail.bitmapData.height, "requestThumbnail");
						}
						if (!shouldUseNewBitmap(modelThumbnail.bitmapData, thumbnail.bitmapData)) {
							return;
						}
					} else {
						CONFIG::logging {
							log(attachment, (model as Object).toString(), "requestThumbnail");
							log(attachment, modelThumbnail ? "modelThumbnail.bitmapData==null" : "modelThumbnail==null", "requestThumbnail");
						}
					}
					
					setModelThumbnail(thumbnail, pageNumber);
					
					CONFIG::logging {
						log(attachment, "page thumb returned: page: " + pageNumber + " w x h=" + thumbnail.bitmapData.width + "," + thumbnail.bitmapData.height, "requestThumbnail");
					}
					if (onThumbComplete != null)
						onThumbComplete(pageNumber, retWidth, retHeight);
				},
				downSampleOK, true, priority
			);
		}
		
		protected function ensureMultipageThumbnailsReady(bottomPage:int, topPage:int):void
		{
			if (!previewMultiPageModel || !previewMultiPageModel.thumbnails)
				return;
			
			var thumbnails:ArrayList = previewMultiPageModel.thumbnails;
			var topPageIndex:int = topPage - 1;
			if (topPageIndex == previewMultiPageModel.thumbnails.length) {
				thumbnails.addItemAt(new Bitmap(), topPageIndex);
			} else if (topPageIndex > thumbnails.length) {
				var newLength:int = Math.min(topPageIndex+100, previewMultiPageModel.pageCount);
				thumbnails = previewMultiPageModel.thumbnails = new ArrayList(new Array(newLength));
			}			
			
			for (var i:int = bottomPage - 1; i <= topPageIndex; i++) {
				if (thumbnails.getItemAt(i) == null)
					thumbnails.setItemAt(new Bitmap(), i);
			}
		}
		
		protected function getModelThumbnail(pageNumber:int):Bitmap
		{
			if (!previewMultiPageModel)
				return model.thumbnail;
			
			var index:int = pageNumber - 1;
			if (index < 0)
				return null;
			if (index >= previewMultiPageModel.thumbnails.length)
				return null;
			
			return previewMultiPageModel.thumbnails.getItemAt(index) as Bitmap;
		}
		
		protected function setModelThumbnail(thumbnail:Bitmap, pageNumber:int):void
		{
			if (!previewMultiPageModel) {
				model.thumbnail = thumbnail;
				return;
			}
			var index:int = pageNumber - 1;
			if (index < 0)
				return;
			
			var thumbnails:ArrayList = previewMultiPageModel.thumbnails;
			if (index < thumbnails.length) {
				thumbnails.setItemAt(thumbnail, index);
				if (scrollPad)
					scrollPad.invalidateScrolling();
			}
		}
		
		/**
		 * Returns true if newBitmap dimensions are a sufficiently better fit than oldBitmap
		 * given the current dimensions of the preview; false otherwise.
		 */
		protected function shouldUseNewBitmap(oldBitmap:BitmapData, newBitmap:BitmapData):Boolean
		{
			if (newBitmap == null) {
				return false;
			}
			if (oldBitmap == null) {
				return true;
			}
			
			var oldWidth:Number = oldBitmap.width;
			var oldHeight:Number = oldBitmap.height;
			var newWidth:Number = newBitmap.width;
			var newHeight:Number = newBitmap.height;
			
			if (oldWidth == newWidth && oldHeight == newHeight)
				return false;
			
			var compareWidths:Boolean =	(oldWidth / oldHeight) >= (width / height);
			var currentDiff:Number = compareWidths ? oldWidth - width : oldHeight - height;
			if (currentDiff == 0) {
				// We've already got a perfect fit, so keep the old thumbnail
				return false;
			}
			var newDiff:Number = compareWidths ? newWidth - width : newHeight - height;
			
			// Upsampling is worse than downsampling so weight the diffs with that in mind
			const upsamplingWeight:Number = 6.0;  // changed the weight from 3 to 6 to allow a larger difference - bh.  TODO: revisit this code and use ratios instead?
			if (currentDiff < 0) {
				currentDiff = -currentDiff * upsamplingWeight
			}
			if (newDiff < 0) {
				newDiff = -newDiff * upsamplingWeight
			}
			
			if (newDiff >= currentDiff) {
				// The new thumbnail's width isn't a better fit
				return false;
			}
			return true;
		}
		
		protected function requestMultiPageThumbnails(id:*,
													  attachment:IAttachment, 
													  previewSize:Point = null,
													  downSampleOK:Boolean = true,
													  onThumbComplete:Function = null):void 
		{
			if (!enabled) {
				deferredRequest = {id:id, attachment:attachment, previewSize:previewSize, downSampleOK:downSampleOK, onThumbComplete:onThumbComplete};
				return;
			}
			
			// we keep thumbnails around for the current page plus 2 thumbs before and after
			var pageNumber:int = previewMultiPageModel.pageNumber;
			
			var bottomPage:int = pageNumber - 2;
			if (bottomPage < 1)
				bottomPage = 1;
			
			var topPage:int = pageNumber + 2;
			if (previewMultiPageModel.thumbnails == null)
				topPage = bottomPage;
			else if (topPage > previewMultiPageModel.pageCount)
				topPage = previewMultiPageModel.pageCount;
			
			ensureMultipageThumbnailsReady(bottomPage, topPage); // this call may change previewMultiPageModel.thumbnails
			
			var thumbnailsSource:Array = previewMultiPageModel.thumbnails.source;
			
			// special case first page for performance
			if (pageNumber == 1)
				topPage = 2;
			
			var nextPage:int = pageNumber + 1;
			if (thumbnailsSource == null)
				nextPage = bottomPage;
			else if (nextPage > thumbnailsSource.length)
				nextPage = thumbnailsSource.length;

			if (previewSize.x == 0 || previewSize.y == 0)
				return;
			
			// request current page at higher priority
			requestThumbnail(id, attachment, previewSize, pageNumber, downSampleOK, thumbnailPriority+1, onThumbComplete);
			
			thumbsData = {pageNumber:pageNumber, nextPage:nextPage, bottomPage:bottomPage, topPage:topPage, id:id, 
				attachment:attachment, previewSize:previewSize, downSampleOK:downSampleOK, 
				thumbnailPriority:thumbnailPriority, onThumbComplete:onThumbComplete};
			
			// clear out thumbs outside of the current range (they'll stay in the thumbnail cache, however)
			// TODO: use a smaller list (for just the range) and adjust ScrollPad accordingly - that way this code
			// can go away (along w/ its slower performance for docs w/ many pages)
			if (thumbnailsSource != null) {
				for (var i:int = 1; i < bottomPage; ++i)
					thumbnailsSource[i-1] = null;
				for (var j:int = topPage+1; j <= thumbnailsSource.length; ++j)
					thumbnailsSource[j-1] = null;
			}
		}
		
		protected function onScrollComplete(e:Event):void
		{
			if (!thumbsData)
				return;
			
			// request thumbs for the current range
//			logger.debug("getting NeighboringThumbs: pageNumber= " + thumbsData.pageNumber);
			
			// request next page first at higher priority
			requestThumbnail(thumbsData.id, thumbsData.attachment, 
				thumbsData.previewSize, thumbsData.nextPage, thumbsData.downSampleOK, thumbsData.thumbnailPriority+1, thumbsData.onThumbComplete);
			
			for (var page:int = thumbsData.bottomPage; page <= thumbsData.topPage; page++) {
				if (page == thumbsData.pageNumber)
					continue;
				if (page == thumbsData.nextPage)
					continue;
				requestThumbnail(thumbsData.id, thumbsData.attachment, 
					thumbsData.previewSize, page, thumbsData.downSampleOK, thumbsData.thumbnailPriority, thumbsData.onThumbComplete);
			}
		}

		CONFIG::logging {
			private function log (attachment:IAttachment, msg:String, prefix:String):void {
				var uid:String = UIDUtil.getUID(skin ? skin : this);
				uid = attachment ? attachment.fileName + ":" + uid.substr(uid.length - 40) : "NO_ATTACHMENT";
				logger.info(prefix + ": " + uid + ": " + msg)
			}
		}

		/**
		 * 
		 * Whether this preview should use platform preview (as opposed to navigator preview).
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get usePlatformPreview():Boolean
		{
			if (!thumbnail)
				return false;
			
			if (model is PreviewMediaModel)
				return false;
			
			if (previewMultiPageModel != null) {
				// Use the same check used by the skin to display the missing file artwork
				return (previewMultiPageModel.pageCount == 0);
			}
			return (thumbnail.content == null);
		}

		/**
		 * 
		 * The priority for generating this thumbnail.  
		 * It can a priority defined by the ThumbnailManager, but is not limited to those values. 
		 * The priority is used by the ThumbnailManager to control a priority queue for thumbnail generation.
		 * 
		 * @see com.adobe.portfolio.managers.ThumbnailManager#IDLE_THUMB_PRIORITY
		 * @see com.adobe.portfolio.managers.ThumbnailManager#DYNAMIC_THUMB_PRIORITY
		 * @see com.adobe.portfolio.managers.ThumbnailManager#STATIC_THUMB_PRIORITY
		 * @see com.adobe.portfolio.managers.ThumbnailManager#PREVIEW_THUMB_PRIORITY
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		public function get thumbnailPriority():int
		{
			return _thumbnailPriority;
		}

		public function set thumbnailPriority(value:int):void
		{
			_thumbnailPriority = value;
		}
		
		override public function set enabled(value:Boolean):void
		{
			if (enabled == value)
				return;
			
			var processDeferredRequest:Boolean = !enabled && value && deferredRequest;
			
			super.enabled = value;
			
			if (processDeferredRequest) {
				requestMultiPageThumbnails(deferredRequest.id, deferredRequest.attachment, deferredRequest.previewSize, deferredRequest.downSampleOK, deferredRequest.onThumbComplete);
				deferredRequest = null;
			}
		}
		
		[Bindable]
		/**
		 *  Whether page navigation is enabled (via keyboard, mousewheel)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.1
		 */	
		public function get pageNavigationEnabled():Boolean
		{
			return _pageNavigationEnabled;
		}
		
		public function set pageNavigationEnabled(value:Boolean):void
		{
			_pageNavigationEnabled = value;
		}
		
	}
}
