/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.previews
{
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.utils.FileUtil;
	
	import flash.utils.Dictionary;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.core.ClassFactory;
	import mx.logging.ILogger;
	import mx.logging.Log;
	
	/**
	 * 
	 * The PreviewFactory class is a factory for constructing IPreviewModels.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class PreviewFactory
	{
		CONFIG::logging {
			protected static var logger:ILogger = Log.getLogger("Preview");
		}
		private static var _token:Object;
		private static var _instance:PreviewFactory;
		
		private var _modelFactories:Dictionary;
		private var _skinStateDict:Dictionary;
		
		
		/**
		 * Returns the singleton instance.
		 * 
		 * @return A PreviewFactory
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function get instance():PreviewFactory 
		{
			if (!_instance) {
				_instance = new PreviewFactory(_token = new Object());
			}
			return _instance;
		}
		
		/**
		 * Constructor.  Consider this private - it will throw an error if called directly.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function PreviewFactory(token:Object) 
		{
			if (_token != token)
				throw new Error("can't construct this object");
		}
		
		/**
		 * Initializes a Preview and creates a preview model if necessary.
		 *  
		 * @param preview The Preview component to initialize.
		 * @param previewState The type of preview. Valid values: Preview.THUMBNAIL, Preview.PAGING_THUMBNAIL, Preview.AUDIO, Preview.VIDEO. 
		 * @param attachment  The associated attachment.
		 * 
		 *  @see Preview
		 *  @see Preview#THUMBNAIL
		 *  @see Preview#PAGING_THUMBNAIL
		 *  @see Preview#AUDIO
		 *  @see Preview#VIDEO
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function initPreview(preview:Preview, previewState:String, attachment:IAttachment):void
		{
			if (!preview || !attachment || !attachment.fileName)
				return;
			
			if (!preview.model) {
				preview.model = createPreviewModel(previewState, attachment);
			}
			preview.skinState = previewState;
			preview.aspectRatio = 1;
		}
		
		/**
		 * Creates a preview model for use by the Preview component.
		 *  
		 * @param previewState The type of preview. Valid values: Preview.THUMBNAIL, Preview.PAGING_THUMBNAIL, Preview.AUDIO, Preview.VIDEO. 
		 * @param attachment  The associated attachment
		 * 
		 * @return A previewModel corresponding to the previewState.
		 * 
		 *  @see Preview#THUMBNAIL
		 *  @see Preview#PAGING_THUMBNAIL
		 *  @see Preview#AUDIO
		 *  @see Preview#VIDEO
		 *  @see IPreviewModel
		 *  @see IPreviewSinglePageModel
		 *  @see IPreviewMultiPageModel
		 *  @see IPreviewMediaModel
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function createPreviewModel(previewState:String, attachment:IAttachment):IPreviewModel
		{
			var modelFactory:ClassFactory = modelFactoryFromSkinState(previewState);
			if (!modelFactory)
				throw new Error("PreviewFactory.modelFactory: invalid previewState: " + previewState);
		
			modelFactory.properties = {mimeType: attachment.mimeType};
			var previewModel:IPreviewModel = modelFactory.newInstance();
			
			if (previewState == Preview.AUDIO || previewState == Preview.VIDEO) {
				var cw:ChangeWatcher = BindingUtils.bindProperty(previewModel, "source", attachment, "dataUrl");
				// TODO : this CW needs to be stored (probably in the model) and unwatched when the model is no longer used
				   // probably best to move this into PreviewMediaModel, but make the signature such that attachment is opaque
				
				// ensure that the mimeType is set for m4a files
				// todo: ensure that the mimeType is always set for all file types - the preview factory 
				// and the way it's called can be refactored a bit to make this cleaner
				if (previewModel.mimeType == null && previewState == Preview.AUDIO 
						&& FileUtil.getExtension(attachment.fileName).toLowerCase() == "m4a") {
					var model:PreviewModel = previewModel as PreviewModel;
					if (model)
						model.mimeType = "audio/mp4";
				}
			}
			
			return previewModel;
		}
		
		private function modelFactoryFromSkinState(skinState:String):ClassFactory 
		{
			if (!_modelFactories) {
				_modelFactories = new Dictionary();
				_modelFactories[Preview.THUMBNAIL] = new ClassFactory(PreviewSinglePageModel);
				_modelFactories[Preview.PAGING_THUMBNAIL] = new ClassFactory(PreviewMultiPageModel);
				_modelFactories[Preview.AUDIO] = new ClassFactory(PreviewMediaModel);
				_modelFactories[Preview.VIDEO] = new ClassFactory(PreviewMediaModel);
			}
			return _modelFactories[skinState];
		}
		
		/**
		 * Returns a preview state from a file/attachment mimeType and/or extension.  It uses the extension only if the mimeType is unknown or not supplied.
		 *  
		 * @param mimeType The file/attachment mimeType.
		 * @param extension  The file/attachment extension.
		 * 
		 * @return A previewState.  One of: Preview.THUMBNAIL, Preview.PAGING_THUMBNAIL, Preview.AUDIO, Preview.VIDEO. 
		 * 
		 *  @see Preview#THUMBNAIL
		 *  @see Preview#PAGING_THUMBNAIL
		 *  @see Preview#AUDIO
		 *  @see Preview#VIDEO
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function previewStateFromMimeType(mimeType:String, extension:String):String 
		{
			if (!_skinStateDict) {
				_skinStateDict = new Dictionary();
				
				_skinStateDict["image/gif"] = Preview.THUMBNAIL;
				_skinStateDict["image/jpeg"] = Preview.THUMBNAIL;
				_skinStateDict["image/png"] = Preview.THUMBNAIL;
				_skinStateDict["application/x-shockwave-flash"] = Preview.THUMBNAIL;
				_skinStateDict["text/html"] = Preview.THUMBNAIL;
				
				_skinStateDict["application/pdf"] = Preview.PAGING_THUMBNAIL;
				_skinStateDict["application/msword"] = Preview.PAGING_THUMBNAIL;
				_skinStateDict["application/vnd.openxmlformats-officedocument.wordprocessingml.document"] = Preview.PAGING_THUMBNAIL;
				_skinStateDict["application/vnd.ms-excel"] = Preview.PAGING_THUMBNAIL;
				_skinStateDict["application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"] = Preview.PAGING_THUMBNAIL;
				_skinStateDict["application/vnd.ms-powerpoint"] = Preview.PAGING_THUMBNAIL;
				_skinStateDict["application/vnd.openxmlformats-officedocument.presentationml.presentation"] = Preview.PAGING_THUMBNAIL;
				
				_skinStateDict["audio/mpeg"] = Preview.AUDIO;
				_skinStateDict["audio/x-mpeg"] = Preview.AUDIO;
				_skinStateDict["audio/mp3"] = Preview.AUDIO;
				_skinStateDict["audio/x-mp3"] = Preview.AUDIO;
				_skinStateDict["audio/mpeg3"] = Preview.AUDIO;
				_skinStateDict["audio/x-mpeg3"] = Preview.AUDIO;
				_skinStateDict["audio/mpg"] = Preview.AUDIO;
				_skinStateDict["audio/x-mpg"] = Preview.AUDIO;
				_skinStateDict["audio/x-mpegaudio"] = Preview.AUDIO;
				_skinStateDict["audio/mp4"] = Preview.AUDIO;
				_skinStateDict["audio/mp4a-latm"] = Preview.AUDIO;

				_skinStateDict["video/x-flv"] = Preview.VIDEO;
				_skinStateDict["video/mp4"] = Preview.VIDEO;

				_skinStateDict["video/x-m4v"] = Preview.VIDEO;
				_skinStateDict["video/quicktime"] = Preview.VIDEO;
				_skinStateDict["video/3gpp"] = Preview.VIDEO;
				_skinStateDict["video/3gpp2"] = Preview.VIDEO;
			}
			var skinState:String = _skinStateDict[mimeType];
			
			if (skinState == null) {
				CONFIG::logging{
					if (mimeType != null)
						logger.warn("unknown mime type: " + mimeType);
				}
				skinState = previewStateFromExtension(extension);
				if (skinState == null)
					skinState = Preview.THUMBNAIL;
			}
				
			return skinState;
		}
		
		/// REMOVE this block when mimeType doesn't always return null
		protected var skinStateExtensionDict:Dictionary;
		
		protected function previewStateFromExtension(extension:String):String {
			if (!skinStateExtensionDict) {
				skinStateExtensionDict = new Dictionary();
				skinStateExtensionDict["pdf"] = Preview.PAGING_THUMBNAIL;
				skinStateExtensionDict["doc"] = Preview.PAGING_THUMBNAIL;
				skinStateExtensionDict["docx"] = Preview.PAGING_THUMBNAIL;
				skinStateExtensionDict["xls"] = Preview.PAGING_THUMBNAIL;
				skinStateExtensionDict["xlsx"] = Preview.PAGING_THUMBNAIL;
				skinStateExtensionDict["ppt"] = Preview.PAGING_THUMBNAIL;
				skinStateExtensionDict["pptx"] = Preview.PAGING_THUMBNAIL;
				skinStateExtensionDict["flv"] = Preview.VIDEO;
				skinStateExtensionDict["f4v"] = Preview.VIDEO;
				skinStateExtensionDict["mov"] = Preview.VIDEO;
				skinStateExtensionDict["m4v"] = Preview.VIDEO;
				skinStateExtensionDict["mp4"] = Preview.VIDEO;
				skinStateExtensionDict["3gp"] = Preview.VIDEO;
				skinStateExtensionDict["3g2"] = Preview.VIDEO;
				skinStateExtensionDict["mp3"] = Preview.AUDIO;
				skinStateExtensionDict["m4a"] = Preview.AUDIO;
				skinStateExtensionDict["jpg"] = Preview.THUMBNAIL;
			}
			var skinState:String = skinStateExtensionDict[extension ? extension.toLowerCase() : null];
			
			if (skinState == null)
				skinState = Preview.THUMBNAIL;
			
			return skinState;
		}

	}
}
