/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.previews
{
	import com.adobe.portfolio.components.supportClasses.AutoWatch;
	import com.adobe.portfolio.events.PreviewEvent;
	import com.adobe.portfolio.utils.DoubleClickHandler;
	
	import flash.display.DisplayObject;
	import flash.events.Event;
	import flash.events.KeyboardEvent;
	import flash.events.MouseEvent;
	import flash.ui.Keyboard;
	
	import mx.binding.utils.BindingUtils;
	import mx.core.mx_internal;
	import mx.events.VideoEvent;
	import mx.logging.ILogger;
	import mx.logging.Log;
	import mx.states.State;
	
	import org.osmf.events.MediaPlayerStateChangeEvent;
	import org.osmf.media.MediaPlayerState;
	
	import spark.components.Button;
	import spark.components.Group;
	import spark.components.VideoPlayer;
	import spark.components.mediaClasses.ScrubBar;

	/**
	 * 
	 * The VideoPreview class extends VideoPlayer to play video content with more controls.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class VideoPreview extends VideoPlayer
	{
		CONFIG::logging {
			protected static const logger:ILogger = Log.getLogger("VideoPreview");
		}
		protected static const NEAR_ZERO:Number = 1/24;

		[SkinPart(required="false")]
		
		/**
		 *  An optional skin part to display the hover play button.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public var hoverPlay:Button;
		
		[SkinPart(required="false")]
		/**
		 *  An optional skin part to display a slideable control bar
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public var controlBar:Group;

		[SkinPart(required="false")]
		/**
		 *  An optional skin part to display the minimalScrubBar play button.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public var minimalScrubBar:ScrubBar;
		
		[SkinPart(required="true")]
		/**
		 *  A required skin part to handle double-click events.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public var mouseButton:Button;

		/**
		 *  The "stopped" state for VideoPreview and its skin
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public static var STOPPED:String = "stopped";

		/**
		 *  The "paused" state for VideoPreview and its skin
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public static var PAUSED:String = "paused";

		/**
		 *  The "playing" state for VideoPreview and its skin
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public static var PLAYING:String = "playing";

		/**
		 *  The "stoppedHover" state for VideoPreview and its skin
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public static var STOPPED_HOVER:String = "stoppedHover";

		/**
		 *  The "pausedHover" state for VideoPreview and its skin
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public static var PAUSED_HOVER:String = "pausedHover";

		/**
		 *  The "playingHover" state for VideoPreview and its skin
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public static var PLAYING_HOVER:String = "playingHover";
		
		private static var toHover:Object;
		
		protected var autoWatch:AutoWatch;
		protected var doubleClickHandler:DoubleClickHandler;
		protected var controlBarChanging:Boolean;
		protected var rollOver:Boolean;
		
		private var _skinState:String;
		private var _aspectRatio:Number;
		private var _model:IPreviewModel;
		
		/**
		 *  Constructor
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10
		 *  @playerversion AIR 1.5
		 *  @productversion Flex 4
		 */
		public function VideoPreview()
		{
			super();
			initStates();
			_aspectRatio = 1;
			autoWatch = new AutoWatch();
			autoWatch.addWatcher(BindingUtils.bindProperty(this, "source", this, ["model", "source"]));
		}
		
		
		protected function get skinState():String
		{
			return _skinState;
		}
		
		protected function set skinState(value:String):void
		{
			if (_skinState == value)
				return;
			
			_skinState = value;
			invalidateSkinState();
		}

		protected function initStates():void 
		{
			states = [
				new State({name: PAUSED_HOVER}), 
				new State({name: PLAYING_HOVER}),
				new State({name: STOPPED_HOVER}),
				new State({name: STOPPED})
			];
			
			if (toHover)
				return;
						
			toHover = new Object();
			toHover[STOPPED] = STOPPED_HOVER;
			toHover[PAUSED] = PAUSED_HOVER;
			toHover[PLAYING] = PLAYING_HOVER;
		}
		
		override protected function attachSkin():void
		{
			super.attachSkin();
			addEventListener("mediaPlayerStateChange", onMediaStateChange);
			addEventListener("currentTimeChange", playHeadChangedHandler);
			addEventListener("totalTimeChanged", totalTimeChangedHandler);
			addEventListener(VideoEvent.PLAYHEAD_UPDATE, playHeadUpdateHandler);
			addEventListener(MouseEvent.ROLL_OVER, rollOverHandler);
			addEventListener(MouseEvent.ROLL_OUT, rollOutHandler);
			systemManager.getSandboxRoot().addEventListener(KeyboardEvent.KEY_DOWN, keyboardHandler);
		}

		override protected function detachSkin():void
		{
			super.detachSkin();
			removeEventListener("mediaPlayerStateChange", onMediaStateChange);
			removeEventListener(MouseEvent.ROLL_OVER, rollOverHandler);
			removeEventListener(MouseEvent.ROLL_OUT, rollOutHandler);
			removeEventListener("currentTimeChange", playHeadChangedHandler);
			removeEventListener("durationChanged", totalTimeChangedHandler);
			removeEventListener(VideoEvent.PLAYHEAD_UPDATE, playHeadUpdateHandler);
			systemManager.getSandboxRoot().removeEventListener(KeyboardEvent.KEY_DOWN, keyboardHandler);
		}

		
		override protected function partAdded(partName:String, instance:Object):void 
		{
			super.partAdded(partName, instance);
			switch (instance) {
				case minimalScrubBar:
					updateScrubBar();
					break;
				
				case mouseButton:
					doubleClickHandler = new DoubleClickHandler(mouseButton, mouseClickHandler);
					break;
				
				case controlBar:
					controlBar.addEventListener(MouseEvent.MOUSE_DOWN, controlBar_mouseDownHandler);
					controlBar.addEventListener(MouseEvent.MOUSE_UP, controlBar_mouseUpHandler);
					break;
				
				default:
					break;
			}
		}
		
		override protected function partRemoved(partName:String, instance:Object):void 
		{
			super.partAdded(partName, instance);
			
			switch (instance) {
				case minimalScrubBar:
					updateScrubBar();
					break;
				
				case mouseButton:
					if (doubleClickHandler)
						doubleClickHandler.removeAllListeners();
					break;
				
				case controlBar:
					controlBar.removeEventListener(MouseEvent.MOUSE_DOWN, controlBar_mouseDownHandler);
					controlBar.removeEventListener(MouseEvent.MOUSE_UP, controlBar_mouseUpHandler);
					break;
				
				default:
					break;
			}
		}
		
		protected function onMediaStateChange(e:MediaPlayerStateChangeEvent):void 
		{
//			logger.debug(this + " video: " + e.state);
			
			if (!enabled)
				return;
			
			switch (e.state) {
				case MediaPlayerState.PLAYBACK_ERROR:
					setInvalidVideo();
					break;
				
				case MediaPlayerState.PAUSED:
					
					setDimensions();
					break;
			}
		}
		
		protected function setInvalidVideo():void
		{
			skinState = MediaPlayerState.PLAYBACK_ERROR;
			enabled = false;
		}
		
		protected function setDimensions():void
		{
			if (aspectRatio != 1)
				return;
			
			if (!videoDisplay.videoObject)
				return;
			
			var w:Number = videoDisplay.videoObject.videoWidth;
			var h:Number = videoDisplay.videoObject.videoHeight;
			
			if (w == 0 || h == 0)
				return;
			
			aspectRatio = w / h;
			videoDisplay.width = width;
			videoDisplay.height = width / aspectRatio;
		}
		
		/**
		 *  @private
		 */
		protected function playHeadChangedHandler(event:Event):void
		{
			updateScrubBar();
		}
		
		/**
		 *  @private
		 */
		protected function totalTimeChangedHandler(event:Event):void
		{
			updateScrubBar();
		}
		
		/**
		 *  @private
		 */
		protected function playHeadUpdateHandler(event:VideoEvent):void
		{
			updateScrubBar();
		}
		
		protected function mouseClickHandler(event:MouseEvent):void 
		{
			if (playing) {
				pause();
			} else {
				play();
			}
		}
		
		protected function keyboardHandler(event:KeyboardEvent):void 
		{
			if (event.isDefaultPrevented())
				return;
			if (event.keyCode != Keyboard.SPACE)
				return;
			for (var displayObject:DisplayObject = this; displayObject != null; displayObject = displayObject.parent) {
				if (!displayObject.visible || (displayObject.alpha == 0)) {
					return;
				}
			}
			if (playing) {
				pause();
			} else {
				play();
			}
 		}
		
		protected function controlBar_mouseDownHandler(e:Event):void {
			controlBarChanging = true;
			e.stopImmediatePropagation();
		}

		protected function controlBar_mouseUpHandler(e:Event):void {
			controlBarChanging = false;
			invalidateSkinState();
		}

		protected function rollOverHandler(e:Event):void {
			rollOver = true;
			invalidateSkinState();
		}

		protected function rollOutHandler(e:Event):void {
			rollOver = false;
			invalidateSkinState();
		}
		
		/**
		 *  @private
		 */
		protected function updateScrubBar():void
		{
			if (!videoDisplay || !minimalScrubBar)
				return;
			
			minimalScrubBar.minimum = 0;
			minimalScrubBar.maximum = isNaN(duration) ? 0 : duration;
			minimalScrubBar.value = isNaN(currentTime) ? 0 : currentTime;
			minimalScrubBar.loadedRangeEnd = videoDisplay.mx_internal::videoPlayer.bytesLoaded/videoDisplay.mx_internal::videoPlayer.bytesTotal * duration;
		}

		override protected function getCurrentSkinState():String
		{
			if (!initialized)
				return null;

			var state:String = super.getCurrentSkinState();
//			CONFIG::logging {
//				logger.debug(this + " super skinState= " + state);
//			}
			
			if (skinState == MediaPlayerState.PLAYBACK_ERROR) {
				state =  MediaPlayerState.PLAYBACK_ERROR;
			} else if (currentTime <= NEAR_ZERO) { // convert to stopped states if appropriate
				state = rollOver ? STOPPED_HOVER : STOPPED;
			} else if (controlBarChanging) { 
					state =  (state == STOPPED) ? STOPPED_HOVER : PAUSED_HOVER;
			} else if (rollOver) {
				if (toHover[state]) {
					state = toHover[state];
				}
			} else if (skinState) {
				state = skinState;
			}
			
//			CONFIG::logging {
//				logger.debug(this + " getCurrentSkinState: state= " + state);
//			}
			
			return state;
		}
		
		[Bindable(name="aspectRatioChanged")]
		/**
		 * 
		 *  The aspect ratio of the preview.
		 * 
		 *  @see com.adobe.portfolio.events.PreviewEvent#ASPECT_RATIO_CHANGED
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get aspectRatio():Number
		{
			return _aspectRatio;
		}
		
		public function set aspectRatio(value:Number):void
		{
			if (_aspectRatio == value)
				return;
			_aspectRatio = value;
			dispatchEvent(new PreviewEvent(PreviewEvent.ASPECT_RATIO_CHANGED));
		}
		
		[Bindable]
		/**
		 * 
		 *  The current preview model.
		 * 
		 *  @see IPreviewModel
		 *  @see com.adobe.portfolio.events.PreviewEvent#MODEL_CHANGED
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get model():IPreviewModel
		{
			return _model;
		}
		
		public function set model(value:IPreviewModel):void
		{
			_model = value;
		}
		
		override public function set source(value:Object):void {
			var mediaModel:IPreviewMediaModel = _model as IPreviewMediaModel;
			if (mediaModel && mediaModel.source != null) {
				enabled = true;
				skinState = null;
				aspectRatio = 1;
				super.source = mediaModel.source;
			}
		}
	}
}
