/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.properties
{
	import com.adobe.portfolio.api.ICollection;
	import com.adobe.portfolio.api.PropertyType;
	import com.adobe.portfolio.components.Background;
	import com.adobe.portfolio.events.PropertyManagerEvent;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.FlashSkinManager;
	import com.adobe.portfolio.managers.FontManager;
	import com.adobe.portfolio.managers.PropertyManager;
	import com.adobe.portfolio.managers.ResourceManager;
	import com.adobe.portfolio.utils.LogHelper;
	import com.adobe.portfolio.utils.NavURLLoader;
	
	import flash.display.DisplayObject;
	import flash.display.GradientType;
	import flash.events.Event;
	import flash.events.EventDispatcher;
	import flash.events.IOErrorEvent;
	import flash.events.SecurityErrorEvent;
	import flash.events.TimerEvent;
	import flash.net.URLLoader;
	import flash.net.URLRequest;
	import flash.system.ApplicationDomain;
	import flash.system.LoaderContext;
	import flash.utils.Timer;
	
	import mx.binding.utils.BindingUtils;
	import mx.controls.SWFLoader;
	import mx.core.UIComponent;
	import mx.events.ModuleEvent;
	import mx.events.PropertyChangeEvent;
	import mx.modules.IModuleInfo;
	import mx.modules.ModuleManager;
	import mx.styles.IStyleManager2;
	import mx.styles.IStyleModule;
	import mx.utils.UIDUtil;

	/** 
	 *  A class containing properties shared among PortfolioFramework navigators. This class can be extended
	 *  to handle navigator-specific properties.
	 *  
	 *  <p>The set of property classes provides a general mechanism that can be used by navigators
	 *  to expose their internal properties for viewing and editing by a host application. The nature
	 *  and behavior of these properties is considered "private" to the navigator, but exposing 
	 *  information about them allows a viewing application to provide a consistent UI 
	 *  for manipulating them.</p>
	 * 
	 *  <p>The Property object contains all the information required by a host application to present
	 *  an interface to the user for configuring the property's value.</p>
	 * 
	 *  <p>The representation and behavior of navigator properties are entirely controlled by
	 *  the Navigator itself. The host (Acrobat) has no knowledge of how properties are stored,
	 *  or what effect they have on the Navigator. The host simply allows the navigator to specify
	 *  its properties (along with type and context information) through a collection of property interfaces.
	 *  A host need never implement any of these interfaces; it will only use them to get, set, and
	 *  enumerate properties through the implementations provided by the Navigators.</p>
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */	
	public class CommonProperties extends EventDispatcher
	{
		/** 
		 *  Property manager shared by navigator components.
		 * 
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected static const propertyManager:PropertyManager = PropertyManager.instance;
		
		/**
		 *  The collection manager instance used by this Navigator
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		protected static const collectionManager:CollectionManager = CollectionManager.instance;
		
		public static const IMPORTED_THEME_NAME:String = "user_theme";
		
		// Functions that should be overridden by subclasses
		
		/** 
		 *  Creates an array of URIs to be loaded by the PropertyManager. This method can be overriden
		 *  to handle additional property files.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function createXMLList():Array
		{
			return ["navigator/properties.xml"];
		}
		
		/** 
		 *  Loads known properties from a property context. This method can be overriden
		 *  to handle additional properties and contexts.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function loadContext(context:PropertyContext):void
		{
			if (context == null) {
				LogHelper.logger.error("loadContext() - context was null");
				return;
			}
			var list:PropertyValue;
			var resourceValue:PropertyValue;
			switch (context.name) {
				case CONTEXT_THEMES:
					skinList = context.getProperty(PROPERTY_SKINS) as PropertyListValue;
					if (skinList != null) {
						// Touch the skin modules so that unused ones are removed on a purge
						skinList.referenceSubvalueResources(0);
						
						// Reference an imported skin so it doesn't get purged
						for each (resourceValue in skinList.subvalues) {
							if ((resourceValue.type == PropertyType.URL) &&
								(resourceValue.name == IMPORTED_THEME_NAME))
							{
								ResourceManager.instance.referenceResource(resourceValue.value as String);
								break;
							}
						}
						skinList.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onValueChange);
						skin = (skinList.activeValue as SkinValue);
					}
					
					var importThemeAction:PropertyActionValue = context.getProperty(PROPERTY_IMPORT_THEME) as PropertyActionValue;
					if (importThemeAction != null) {
						BindingUtils.bindSetter(setImportThemeAction, importThemeAction, "value");
					}
					break;
				
				case CONTEXT_BACKGROUND:
					// Image source
					backgroundImageSourceList = context.getProperty(PROPERTY_IMAGE_SOURCE) as PropertyListValue;
					if (backgroundImageSourceList != null) {
						// Touch the images so that unused ones are removed on a purge
						backgroundImageSourceList.referenceSubvalueResources(0);

						// Reference an imported background image so it doesn't get purged
						for each (resourceValue in backgroundImageSourceList.subvalues) {
							if ((resourceValue.type == PropertyType.URL) &&
								(resourceValue.name == Background.IMPORTED_IMAGE_NAME))
							{
								ResourceManager.instance.referenceResource(resourceValue.value as String);
								break;
							}
						}
					}
					// Image scale
					backgroundImageScaleList = context.getProperty(PROPERTY_IMAGE_SCALE) as PropertyListValue;
					if (backgroundImageScaleList != null) {
						backgroundImageScaleList.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onValueChange);
						backgroundImageScale = backgroundImageScaleList.activeValue;
					}
					// Image position
					backgroundImagePositionList = context.getProperty(PROPERTY_IMAGE_POSITION) as PropertyListValue;
					if (backgroundImagePositionList != null) {
						backgroundImagePositionList.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onValueChange);
						backgroundImagePosition = backgroundImagePositionList.activeValue;
					}
					// Image opacity
					backgroundImageOpacity = context.getProperty(PROPERTY_IMAGE_OPACITY) as PropertyNumberValue;
					// Image blur
					backgroundImageBlur = context.getProperty(PROPERTY_IMAGE_BLUR) as PropertyNumberValue;
					// Gradient type
					list = context.getProperty(PROPERTY_GRADIENT_TYPE) as PropertyValue;
					if ((list != null) && (list.type == PropertyType.LIST)) {
						list.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onValueChange);
						backgroundGradientType = (list as PropertyListValue).activeValue;
					}
					// Gradient intensity
					backgroundGradientIntensity = context.getProperty(PROPERTY_GRADIENT_INTENSITY) as PropertyNumberValue;
					// Gradient opacity
					backgroundGradientOpacity = context.getProperty(PROPERTY_GRADIENT_OPACITY) as PropertyNumberValue;
					break;
				
				case CONTEXT_TEXTFIELD:
					if (textFieldContext != null) {
						LogHelper.logger.error("Duplicate textFieldContext");
					}
					textFieldContext = context;
					
					fontFamilyList = context.getProperty(PROPERTY_TEXTFIELD_FONTFAMILY) as PropertyListValue;
					fontSize = context.getProperty(PROPERTY_TEXTFIELD_FONTSIZE) as PropertyNumberValue;
					fontStyleList = context.getProperty(PROPERTY_TEXTFIELD_FONTSTYLE) as PropertyListValue;
					fontWeightList = context.getProperty(PROPERTY_TEXTFIELD_FONTWEIGHT) as PropertyListValue;
					embedFonts = context.getProperty(PROPERTY_EMBED_FONTS) as PropertyBooleanValue;
					if (embedFonts != null) {
						BindingUtils.bindSetter(setEmbedFonts, embedFonts, "booleanValue");
					}
					
					fontColor = context.getProperty(PROPERTY_TEXTFIELD_FONTCOLOR) as PropertyColorValue;
					fontUnderlined = context.getProperty(PROPERTY_TEXTFIELD_FONTUNDERLINED) as PropertyBooleanValue;
					break;
				case CONTEXT_PORTFOLIO:
					if (portfolioContext != null) {
						LogHelper.logger.error("Duplicate portfolioContext");
					}
					portfolioContext = context;
					
					list = context.getProperty(PROPERTY_TEXTFIELD_FONTFAMILY) as PropertyValue;
					if ((list != null) && (list.type == PropertyType.LIST)) {
						list.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onValueChange);
						portfolioFontName = (list as PropertyListValue).activeValue;
					}
					portfolioEmbedFonts = context.getProperty(PROPERTY_EMBED_FONTS) as PropertyBooleanValue;
					if (portfolioEmbedFonts != null) {
						BindingUtils.bindSetter(setEmbedFonts, portfolioEmbedFonts, "booleanValue");
					}
					
					list = context.getProperty(PROPERTY_CARD_SIZE) as PropertyValue;
					if ((list != null) && (list.type == PropertyType.LIST)) {
						list.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onValueChange);
						portfolioCardSizeList = list as PropertyListValue;
						cardSize = (portfolioCardSizeList.activeValue) as PropertyNumberValue;
					}
					portfolioCardsAngled = context.getProperty(PROPERTY_CARDS_ANGLED) as PropertyBooleanValue;
					portfolioCardsAutoSize = context.getProperty(PROPERTY_CARDS_AUTOSIZE) as PropertyBooleanValue;
					break;
				case CONTEXT_HEADER:
					if (headerContext != null) {
						LogHelper.logger.error("Duplicate headerContext");
					}
					headerContext = context;
					
					FontManager.instance.addFontResource("navigator/header.xml");
					FontManager.instance.registerFontsWithHost();
					headerTemplateList = context.getProperty(PROPERTY_PRESET_LIST) as PropertyListValue;
					headerBackgroundColor = context.getProperty(PROPERTY_BACKGROUND_COLOR) as PropertyColorValue;
					headerBackgroundOpacity = context.getProperty(PROPERTY_BACKGROUND_OPACITY) as PropertyNumberValue;
					headerBackgroundGradientType = context.getProperty(PROPERTY_GRADIENT_TYPE) as PropertyListValue;
					if (headerBackgroundGradientType != null) {
						headerBackgroundGradientType.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onValueChange);
					}
					headerBackgroundGradientIntensity = context.getProperty(PROPERTY_GRADIENT_INTENSITY) as PropertyNumberValue;
					headerAddTextAction = context.getProperty(PROPERTY_ADD_TEXT) as PropertyActionValue;
					headerAddImageAction = context.getProperty(PROPERTY_ADD_IMAGE) as PropertyActionValue;
					break;
				case CONTEXT_INITIALIZE:
					var initialized:PropertyBooleanValue = context.getProperty(PROPERTY_INITIALIZED) as PropertyBooleanValue;
					if ((initialized != null) && !initialized.booleanValue) {
						// Set initialized to true; this lets us distinguish between initial creation and consumption
						initialized.booleanValue = true;
					}
					break;
			}
		}

		// Known context names
		/**
		 *  The string to specify the context name for Navigator themes (skins)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const CONTEXT_THEMES:String = "com.adobe.navigator.themes";
		
		/**
		 *  The string to specify the context name for Navigator backgrounds 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const CONTEXT_BACKGROUND:String = "com.adobe.navigator.background";
		
		/**
		 *  The string to specify the context name for Navigator text fields 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const CONTEXT_TEXTFIELD:String = "com.adobe.navigator.textField";
		
		/**
		 *  The string to specify the context name for Navigator Portfolios
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const CONTEXT_PORTFOLIO:String = "com.adobe.navigator.portfolio";
		
		/**
		 *  The string to specify the context name for Navigator headers
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const CONTEXT_HEADER:String = "com.adobe.navigator.header";
		
		/**
		 *  The string to specify the context name for Navigator initialization
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const CONTEXT_INITIALIZE:String = "com.adobe.navigator.initalize";
		
		/**
		 *  The string used to clear the current context
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const CONTEXT_CLEAR:String = "";

		// Known properties available to clients
		/**
		 *  The string to specify the skins property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_SKINS:String = "skins";
		
		/**
		 *  The string to specify the import theme property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMPORT_THEME:String = "importTheme";
		
		/**
		 *  The string to specify the image source property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_SOURCE:String = "imageSource";
		
		/**
		 *  The string to specify the image scale property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_SCALE:String = "imageScale";
		
		/**
		 *  The string to specify the image position property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_POSITION:String = "imagePosition";
		
		/**
		 *  The string to specify the image opacity property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_OPACITY:String = "imageOpacity";
		
		/**
		 *  The string to specify the image blur property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_BLUR:String = "imageBlur";
		
		/**
		 *  The string to specify the gradient type property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_GRADIENT_TYPE:String = "gradientType";
		
		/**
		 *  The string to specify the gradient intensity property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_GRADIENT_INTENSITY:String = "gradientIntensity";
		
		/**
		 *  The string to specify the gradient opacity property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_GRADIENT_OPACITY:String = "gradientOpacity";
		
		/**
		 *  The string to specify the card size property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_CARD_SIZE:String = "cardSize";
		
		/**
		 *  The string to specify the cards angled property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_CARDS_ANGLED:String = "cardsAngled";
		
		/**
		 *  The string to specify the cards autosized property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_CARDS_AUTOSIZE:String = "cardsAutoSize";
		
		/**
		 *  The string to specify the text field font family property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_TEXTFIELD_FONTFAMILY:String = "fontFamily";
		
		/**
		 *  The string to specify the text field font size property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_TEXTFIELD_FONTSIZE:String = "fontSize";
		
		/**
		 *  The string to specify the text field font style property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_TEXTFIELD_FONTSTYLE:String = "fontStyle";
		
		/**
		 *  The string to specify the text field font weight property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_TEXTFIELD_FONTWEIGHT:String = "fontWeight";
		
		/**
		 *  The string to specify the text field font color property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_TEXTFIELD_FONTCOLOR:String = "fontColor";
		
		/**
		 *  The string to specify the text field font underlined property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_TEXTFIELD_FONTUNDERLINED:String = "fontUnderlined";
		
		/**
		 *  The string to specify the embed fonts property available to clients 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_EMBED_FONTS:String = "embedFonts";

		/**
		 *  The string to specify the preset list property
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_PRESET_LIST:String = "presetList";
		
		/**
		 *  The string to specify the background color property
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_BACKGROUND_COLOR:String = "backgroundColor";
		
		/**
		 *  The string to specify the background opacity property 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_BACKGROUND_OPACITY:String = "backgroundOpacity";
		
		/**
		 *  The string to specify the add text property
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_ADD_TEXT:String = "addText";
		
		/**
		 *  The string to specify the add image property
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_ADD_IMAGE:String = "addImage";
		
		/**
		 *  The string to specify the add swf property
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_ADD_SWF:String = "addSwf";
		
		// Choices in the image source property
		/**
		 *  Indicates that the image source is a placeholder only 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_FILE_PLACEHOLDER:String = "nothing";
		
		/**
		 *  The "select" choice for the image source property 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_FILE_SELECT:String = "select";
		
		/**
		 *  Removes the image source
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_FILE_REMOVE:String = "remove";
		
		// Choices in the image scale property
		/**
		 *  Displays image at its full resolution (not scaled).
		 *  If the image is larger than the window it is clipped.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_SCALE_ACTUAL_SIZE:String = "actualSize";
		
		/**
		 *  Images are tiled (repeated) until the entire containing window is filled with the original image.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_SCALE_ACTUAL_SIZE_TILED:String = "actualSizeTiled";
		
		/**
		 *  Displays image scaled to fill the containing window without clipping on the top, bottom or sides.
		 *  If the containing window is not the same aspect ratio as the image,
		 *  the image fills the window in one dimension and borders may be visible in the other dimension.
		 *  Resizing the containing window causes the image to rescale to the new size.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_SCALE_SHOW_ALL:String = "showAll";
		
		/**
		 *  Displays image scaled to fill the containing window, but clipping may occur on the top and bottom or sides.
		 *  If the containing window is not the same aspect ratio as the image,
		 *  the image fills the window in one dimension and is clipped in the other dimension.
		 *  Resizing the containing window causes the image to rescale to the new size.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_SCALE_NO_BORDER:String = "noBorder";
		
		// Choices in the image position property
		/**
		 *  Displays image so its top edge is aligned with the top of the window
		 *  and the image's left edge is aligned with the left of the window.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_POSITION_TOP_LEFT:String = "topLeft";
		
		/**
		 *  Displays image so its top edge is aligned with the top of the window
		 *  and the image's horizontal center is aligned with the horizontal center of the window.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_POSITION_TOP_CENTER:String = "topCenter";
		
		/**
		 *  Displays image so its top edge is aligned with the top of the window
		 *  and the image's right edge is aligned with the right of the window.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_POSITION_TOP_RIGHT:String = "topRight";
		
		/**
		 *  Displays image so its vertical center is aligned with the vertical center of the window
		 *  and the image's left edge is aligned with the left of the window.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_POSITION_MIDDLE_LEFT:String = "middleLeft";
		
		/**
		 *  Displays image relative to the center of the window.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_POSITION_MIDDLE_CENTER:String = "center";
		
		/**
		 *  Displays image so its vertical center is aligned with the vertical center of the window
		 *  and the image's right edge is aligned with the right of the window.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_POSITION_MIDDLE_RIGHT:String = "middleRight";
		
		/**
		 *  Displays image so its bottom edge is aligned with the bottom of the window
		 *  and the image's left edge is aligned with the left of the window.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_POSITION_BOTTOM_LEFT:String = "bottomLeft";
		
		/**
		 *  Displays image so its bottom edge is aligned with the bottom of the window
		 *  and the image's horizontal center is aligned with the horizontal center of the window.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_POSITION_BOTTOM_CENTER:String = "bottomCenter";
		
		/**
		 *  Displays image so its bottom edge is aligned with the bottom of the window
		 *  and the image's right edge is aligned with the right of the window.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_IMAGE_POSITION_BOTTOM_RIGHT:String = "bottomRight";
		
		
		// Names of the different card sizes
		/**
		 *  The smallest of the card image sizes
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_CARD_SIZE_SMALL:String = "small";
		
		/**
		 *  The medium images size for a card
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_CARD_SIZE_MEDIUM:String = "medium";
		
		/**
		 *  The largest of the card size images
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_CARD_SIZE_LARGE:String = "large";
		

		// Location of the background image
		/**
		 *  The URL for the source of the background image 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const BACKGROUND_IMAGE_SOURCE_URL:String = "navigator/backgroundImage.ext";

		// Font info
		/**
		 *  Indicates the property's text field should be a normal weight font.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_TEXTFIELD_NORMAL:String = "normal";
		
		/**
		 *  Indicates the property's text field should be a bold weight font. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_TEXTFIELD_BOLD:String = "bold";
		
		/**
		 *  Indicates the property's text field should have an italic style.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_TEXTFIELD_ITALIC:String = "italic";
		
		/**
		 *  Indicates the property has been initialized.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const PROPERTY_INITIALIZED:String = "initialized";

		/**
		 *  The portfolio property context
		 * 
		 * 	@see #CONTEXT_PORTFOLIO
		 *
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var portfolioContext:PropertyContext;
		
		/**
		 *  The text field property context. This context is auto-generated when
		 * 	requested by a call to getCurrentContextByName.
		 * 
		 * 	@see #CONTEXT_TEXTFIELD
		 * 	@see #getCurrentContextByName
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var textFieldContext:PropertyContext;
		
		/**
		 *  The header property context
		 * 
		 * 	@see #CONTEXT_HEADER
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var headerContext:PropertyContext;
		
		/**
		 *  The unique identifier of the current context.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		private var currentContextID:Object;
		
		/**
		 *  Returns the current property context for the type of context specified by contextName.
		 *  If no contextName is specified the portfolioContext is returned.
		 * 
		 *  @param contextName Identifies the context of interest.  Possible values are:
		 *  <ul>
		 *     <li>CONTEXT_CLEAR -- this will return the default current context</li>
		 *     <li>CONTEXT_TEXTFIELD -- returns the text properties</li>
		 *     <li>CONTEXT_HEADER -- returns the header properties</li>
		 *  </ul>
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function getCurrentContextByName(contextName:String):PropertyContext
		{
			if (contextName == CONTEXT_CLEAR) {
				contextName = getDefaultCurrentContextName();
			}
			if (contextName == CONTEXT_TEXTFIELD) {
				if (textFieldContext == null) {
					// Auto-generate the context for text fields
					textFieldContext = new PropertyContext();
					textFieldContext.loadFromXML(
						<PropertyContext labelKey="textfieldProperties" label="TEXTFIELD_PROPERTIES" uid={CONTEXT_TEXTFIELD} global="false">
							<FontList name="fontFamily" includeDefaultFont="false" labelKey="textfieldFontFamily" label="TEXTFIELD_FONT_FAMILY"/>
							<PropertyNumber name="fontSize" labelKey="textfieldFontSize" label="TEXTFIELD_FONT_SIZE" minValue="0" value="12" maxValue="128"/>
							<PropertyList name="fontStyle" labelKey="textfieldFontStyle" label="TEXTFIELD_FONT_STYLE" active="normal">
								<PropertyValue name="normal" labelKey="textfieldFontNormal" label="TEXTFIELD_FONT_NORMAL"/>
								<PropertyValue name="italic" labelKey="textfieldFontItalic" label="TEXTFIELD_FONT_ITALIC"/>
							</PropertyList>
							<PropertyList name="fontWeight" labelKey="textfieldFontWeight" label="TEXTFIELD_FONT_WEIGHT" active="normal">
								<PropertyValue name="normal" labelKey="textfieldFontNormal" label="TEXTFIELD_FONT_NORMAL"/>
								<PropertyValue name="bold" labelKey="textfieldFontBold" label="TEXTFIELD_FONT_BOLD"/>
							</PropertyList>
						    <PropertyColor name="fontColor" labelKey="textfieldFontColor" label="TEXTFIELD_FONT_COLOR" active="normal" value="0x000000" />
						    <PropertyBoolean name="fontUnderlined" labelKey="textfieldFontUnderlined" label="TEXTFIELD_FONT_UNDERLINED" active="normal" value="false" />
							<PropertyBoolean name="embedFonts" labelKey="shouldEmbedFonts" label="SHOULD_EMBED_FONTS" descriptionKey="shouldEmbedFontsDesc" description="SHOULD_EMBED_FONTS_DESC" value={FontManager.instance.embedFontsOnSave}/>
						</PropertyContext>
					);
					loadContext(textFieldContext);
				}
				return textFieldContext;
			} else if (contextName == CONTEXT_HEADER) {
				if (headerContext != null) {
					return headerContext;
				}
			}
			
			// portfolioContext is our go-to default, so return it if we haven't found anything else
			return portfolioContext;
		}
		
		/**
		 *  Sets the current property context to the one with the given name.
		 *  <code>CONTEXT_CLEAR</code> may be used when the current context is no longer
		 *  applicable to the caller; however, the context is only cleared if the id provided
		 *  is the same id used in the previous call to <code>setCurrentContext</code>.
		 * 
		 *  @param contextName The name of the new propertyContext to use
		 *  @param id The identifier of the current propertyContext you wish to change
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function setCurrentContextByName(contextName:String, id:Object):void
		{
			setCurrentContext(getCurrentContextByName(contextName), id, (contextName == CONTEXT_CLEAR));
		}
		
		/**
		 *  Set the current propertyContext
		 * 
		 *  @param propertyContext The name of the new propertyContext to set.
		 *  @param id The identifier of the current propertyContext you wish to change
		 *  @param clearing A boolean indicating whether the current context should be cleared or not.
		 *  The default is <code>false</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function setCurrentContext(propertyContext:PropertyContext, id:Object, clearing:Boolean = false):void
		{
			if (clearing) {
				if ((id != currentContextID) && (currentContextID != null)) {
					// Someone has already changed the current context, so don't clear it.
					return;
				} else {
					currentContextID = null;
				}
			} else {
				currentContextID = id;
			}
			propertyManager.currentContext = propertyContext;
		}
		
		/**
		 *  Returns the name of the default non-global property context.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function getDefaultCurrentContextName():String
		{
			return CONTEXT_PORTFOLIO;
		}
		
		/**
		 *  The property list of themes (skins) available to this Navigator.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var skinList:PropertyListValue;

		[Bindable]
		/**
		 *  The image source list for the Navigator background.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var backgroundImageSourceList:PropertyListValue;
		
		[Bindable]
		/**
		 *  The image scale list for the Navigator background.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var backgroundImageScaleList:PropertyListValue;
		
		[Bindable]
		/**
		 *  The image position list for the Navigator background.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var backgroundImagePositionList:PropertyListValue;
		
		[Bindable]
		/**
		 *  The image scale value for the Navigator background.
		 * 
		 * 	@see Background#imageScale
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var backgroundImageScale:PropertyValue;
		
		[Bindable]
		/** 
		 *  The image position value for the Navigatro background.
		 * 
		 * 	@see Background#imagePosition 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var backgroundImagePosition:PropertyValue;
		

		[Bindable]
		/** 
		 *  The image opacity for the Navigator background.
		 * 
		 * 	@see Background#imageOpacity
		 *
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var backgroundImageOpacity:PropertyNumberValue;
		
		[Bindable]
		/**
		 *  The image blur for the navigator background.
		 * 
		 * 	@see Background#imageBlur
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var backgroundImageBlur:PropertyNumberValue;
		
		[Bindable]
		/**
		 *  The gradient type for the navigator background.
		 * 
		 * 	@see Background#gradientType
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var backgroundGradientType:PropertyValue;
		
		[Bindable]
		/**
		 *  The gradient intensity for the Navigator background.
		 * 
		 * 	@see Background#gradientIntensity
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var backgroundGradientIntensity:PropertyNumberValue;
		
		[Bindable]
		/**
		 *  The gradient opacity for the Navigator background.
		 * 
		 * 	@see Background:gradientOpacity
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var backgroundGradientOpacity:PropertyNumberValue;
		

		[Bindable]
		/**
		 *  The name of the font family used by the current text field selection.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var fontFamilyList:PropertyListValue;

		[Bindable]
		/**
		 *  The font size used by the current text field selection.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var fontSize:PropertyNumberValue;

		[Bindable]
		/**
		 *  The font style used by the current text field selection.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var fontStyleList:PropertyListValue;

		[Bindable]
		/**
		 *  The font weight used by the current text field selection.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var fontWeightList:PropertyListValue;
		
		[Bindable]
		/**
		 *  The font color used by the current text field selection.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var fontColor:PropertyColorValue;
		
		[Bindable]
		/**
		 *  Indicates whether the font used by the current text field selection is underlined or not.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var fontUnderlined:PropertyBooleanValue;
		
		[Bindable]
		/**
		 *  Checkbox indicating whether or not fonts should be embedded,
		 * 	as specified by the text field context.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var embedFonts:PropertyBooleanValue;

		[Bindable]
		/**
		 *  Checkbox indicating whether or not fonts should be embedded,
		 * 	as specified by the portfolio context.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var portfolioEmbedFonts:PropertyBooleanValue;

		[Bindable]
		/**
		 *  The card size for cards, as specified by the portfolio context.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var cardSize:PropertyNumberValue;
		
		[Bindable]
		/**
		 *  The card size list for card, as specified by the portfolio context.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var portfolioCardSizeList:PropertyListValue;
		
		[Bindable]
		/**
		 *  Indicates whether cards should automatically change size when the window size changes
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var portfolioCardsAutoSize:PropertyBooleanValue;
		
		[Bindable]
		/**
		*  Indicates whether cards should be shown at random angles or not.
		* 
		*  @langversion 3.0
		*  @playerversion Flash 10.1
		*  @playerversion AIR 2.0
		*  @productversion Portfolio 10.0
		*
		*/	
		public var portfolioCardsAngled:PropertyBooleanValue;
		
		[Bindable]
		/**
		 *  The name of the Portfolio font
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var portfolioFontName:PropertyValue;
		
		[Bindable]
		/**
		 *  The font size of the Portfolio font
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var portfolioFontSize:PropertyNumberValue;
		
		[Bindable]
		/**
		 *  The font style of the Portfolio font
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var portfolioFontStyle:PropertyValue;
		
		[Bindable]
		/**
		 *  The selected template for the Navigator header.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var headerTemplateList:PropertyListValue;
		
		[Bindable]
		/**
		 *  The background color for the Navigator header component.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var headerBackgroundColor:PropertyColorValue;
		
		[Bindable]
		/**
		 *  The background opacity for the Navigator header component.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var headerBackgroundOpacity:PropertyNumberValue;
		
		[Bindable]
		/**
		 *  The background gradient type for the Navigator header component.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var headerBackgroundGradientType:PropertyListValue;
		
		[Bindable]
		/**
		 *  The background gradient intensity for the Navigator header component.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var headerBackgroundGradientIntensity:PropertyNumberValue;
		
		[Bindable]
		/**
		 *  The action control used to add a text item to the Navigator header.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var headerAddTextAction:PropertyActionValue;
		
		[Bindable]
		/**
		 *  The action control used to add an image to the Navigator header.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public var headerAddImageAction:PropertyActionValue;
		
		private var _skin:SkinValue;
		[Bindable]
		/**
		 *  The current theme (skin), as specified by the activeValue of skinList.
		 * 
		 * 	@see #skinList
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get skin():SkinValue
		{
			return _skin;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set skin(value:SkinValue):void
		{
			if (value != _skin) {
				if (_navigator == null) {
					_skin = value;
				} else {
					unloadSkin((value == null) || (value.skinType != "flex"));
					_skin = value;
					loadSkin();
				}
			}
		}
		
		/*
		InternalAdobePatentID="B1283E1"
		InternalAdobePatentID="B1283E2"
		*/
		/**
		 *  Loads the current skin, as specified by the activeValue of skinList.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function loadSkin():void 
		{
			if (!_skin)
				return;
			
			switch (_skin.skinType) {
				case "flex":
					// load module
					loadSkinModule(_skin.url);
					break;
				case "flash":
					FlashSkinManager.instance.loadFlashSkins(_skin.url);
					// TODO For now, loadStyleDeclarations triggers collectionManager.enabled to get set to true,
					//     so force it for flash skins as well.
					collectionManager.enabled = true;
					break;
				default:
					// TODO For now, loadStyleDeclarations triggers collectionManager.enabled to get set to true,
					//     so force it for default skins as well.
					collectionManager.enabled = true;
					break;
			}
			ResourceManager.instance.referenceResource(_skin.url);
		}
		
		/**
		 *  Unloads the current skin.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function unloadSkin(update:Boolean):void
		{
			if (!_skin)
				return;
			
			collectionManager.enabled = false;
			try {
				_navigator.styleManager.unloadStyleDeclarations(_skin.url);
				switch (_skin.skinType) {
					case "flex":
						// unload module
						unloadCurrentSkinModule(update);
						break;
					case "flash":
						FlashSkinManager.instance.unloadFlashSkins(_skin.url);
						break;
				}
				ResourceManager.instance.dereferenceResource(_skin.url);
			} catch (e:Error) {
				var g:Boolean = true;
				// trying to unload styles that have already been unloaded throws errors
			}
		}
		
		private var currentSkinModule:IModuleInfo;
		private var currentSkinStyleModule:IStyleModule;
		
		/**
		 * @private
		 * 
		 *  A variant of styleManager.loadStyleDeclarations that allows us to load
		 *  skin modules from a ByteArray rather than a url.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		private function loadSkinModule(moduleURL:String):Boolean
		{
			var module:IModuleInfo = ModuleManager.getModule(moduleURL);
			var thisStyleManager:IStyleManager2 = _navigator.styleManager;
			
			var readyHandler:Function;
			
			// TODO For some reason, sometimes module is not dispatching a ready event, even
			// when it becomes ready. This timer is intended to catch this case, although oddly
			// enough I can't reproduce it when the timer is started.
			var readyTimerHandler:Function = function(event:TimerEvent):void
			{
				if (module.ready) {
					CONFIG::logging{
						LogHelper.logger.warn("timer says we're ready");
					}
					readyHandler(null);
				}
			}
			var readyTimer:Timer = new Timer(500);
			readyTimer.addEventListener(TimerEvent.TIMER, readyTimerHandler);
			readyTimer.start();
			
			readyHandler = function(moduleEvent:ModuleEvent):void
			{
				readyTimer.stop();
				readyTimer.removeEventListener(TimerEvent.TIMER, readyTimerHandler);
				readyTimer = null;
				
				currentSkinModule.removeEventListener(ModuleEvent.READY, readyHandler);
				currentSkinModule.removeEventListener(ModuleEvent.ERROR, errorHandler);
				
				currentSkinStyleModule = IStyleModule(currentSkinModule.factory.create());
				
				// Register the style module to use this style manager.
				currentSkinModule.factory.registerImplementation("mx.styles::IStyleManager2", thisStyleManager);
				currentSkinStyleModule.setStyleDeclarations(thisStyleManager);
				thisStyleManager.styleDeclarationsChanged();
			};
			
			var setupHandler:Function = function (moduleEvent:ModuleEvent):void
			{
				currentSkinModule = moduleEvent.module;
				currentSkinModule.removeEventListener(ModuleEvent.SETUP, setupHandler);
				if (currentSkinModule.ready) {
					readyHandler(moduleEvent);
				} else {
					currentSkinModule.addEventListener(ModuleEvent.READY, readyHandler);
				}
			};
			
			
			var errorHandler:Function = function(moduleEvent:ModuleEvent):void
			{
				readyTimer.stop();
				readyTimer.removeEventListener(TimerEvent.TIMER, readyTimerHandler);
				readyTimer = null;
				
				module.removeEventListener(ModuleEvent.READY, readyHandler);
				module.removeEventListener(ModuleEvent.ERROR, errorHandler);
				LogHelper.logger.error("Error loading module: "+moduleURL);
				currentSkinModule = null;
				currentSkinStyleModule = null;
				onSkinModuleLoadError(moduleURL);
			};
			
			module.addEventListener(ModuleEvent.SETUP, setupHandler, false, 0, true);
			module.addEventListener(ModuleEvent.ERROR, errorHandler, false, 0, true);
			
			var bytesLoaderHandler:Function = function(event:Event):void
			{
				var thisLoader:URLLoader = event.target as URLLoader;
				if (thisLoader == null) {
					LogHelper.logger.error("thisLoader shouldn't have been null");
				}
				thisLoader.removeEventListener(Event.COMPLETE, bytesLoaderHandler);
				thisLoader.removeEventListener(IOErrorEvent.IO_ERROR, bytesLoaderHandler);
				thisLoader.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, bytesLoaderHandler);
				
				switch (event.type) {
					case Event.COMPLETE:
						module.load(new ApplicationDomain(ApplicationDomain.currentDomain),
							null/* securityDomain */,
							thisLoader.data);
						break;
					case IOErrorEvent.IO_ERROR:
						LogHelper.logger.error("IOError loading module: "+moduleURL);
						onSkinModuleLoadError(moduleURL);
						break;
					case SecurityErrorEvent.SECURITY_ERROR:
						LogHelper.logger.error("SecurityError loading module: "+moduleURL);
						onSkinModuleLoadError(moduleURL);
						break;
				}
				
			}
			var bytesLoader:NavURLLoader = new NavURLLoader(null, true);
			bytesLoader.addEventListener(Event.COMPLETE, bytesLoaderHandler);
			bytesLoader.addEventListener(IOErrorEvent.IO_ERROR, bytesLoaderHandler);
			bytesLoader.addEventListener(SecurityErrorEvent.SECURITY_ERROR, bytesLoaderHandler);
			
			bytesLoader.load(new URLRequest(moduleURL));
			
			return true;
		}
		
		/**
		 * Displays an error message to the user that the given skin resource could not be found.
		 * 
		 * @see ResourceManager#showMissingResourcePopup()
		 */
		protected function onSkinModuleLoadError(moduleURL:String):void
		{
			// If we've failed to load a skin module, go ahead and enable collectionManager
			// with the default skin
			collectionManager.enabled = true;
			
			if (_skin == null) {
				return;
			}
			ResourceManager.instance.showMissingResourcePopup(moduleURL, _navigator);
		}
		
		/**
		 *  @private
		 * 
		 *  A variant of styleManager.unloadStyleDeclarations that compliments
		 *  loadSkinModule.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		private function unloadCurrentSkinModule(update:Boolean):void
		{
			if (currentSkinStyleModule != null) {
				currentSkinStyleModule.unload();
				currentSkinStyleModule = null;
			}
			
			if (currentSkinModule != null) {
				currentSkinModule.unload();
				currentSkinModule = null;
			}
			
			if (update)
				_navigator.styleManager.styleDeclarationsChanged();
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function setImportThemeAction(value:uint):void
		{
			if (value == 0) {
				// Initializing the action; don't trigger anything
				return;
			}
			if (_navigator == null) {
				importTheme();
			} else {
				_navigator.callLater(importTheme);
			}
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private var themeURLToValidate:String;
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function importTheme():void
		{
			if (!collectionManager.permissionCheck("Document", "Modify")) {
				LogHelper.logger.error("Insufficient permissions to import a theme");
				return;
			}
			if (skinList == null) {
				LogHelper.logger.error("No skin list to update");
				return;
			}
			if ((themeURLToValidate != null) && (themeURLToValidate != "")) {
				LogHelper.logger.error("We're already validating a theme, so don't load another yet");
				return;
			}
			var hostCollection:ICollection = (CollectionManager.instance.host != null) ? CollectionManager.instance.host.collection : null;
			// TODO Add a importThemeSelectFile string, rather than using the button
			var titleString:String = (hostCollection == null) ? "IMPORT_THEME_SELECT_FILE" : hostCollection.getLocalizedString("importThemeButton", "IMPORT_THEME_SELECT_FILE");
			
			ResourceManager.instance.addResource(titleString, [ResourceManager.instance.swfFilter], "navigator/customthemes/"+UIDUtil.createUID()+".swf",
				function (result:Object):void
				{
					if (!result.propertyIsEnumerable("success") || !result.success) {
						LogHelper.logger.error("Theme import did not succeed");
						return;
					}
					var relativeURL:String = result.propertyIsEnumerable("relativeURL") ? result.relativeURL : "";
					if (relativeURL == "") {
						LogHelper.logger.error("Invalid URL");
						return;
					}
					
					const validateSWF:Boolean = true;
					if (validateSWF) {
						// Load the URL first to confirm that the SWF is a "valid" theme
						themeURLToValidate = relativeURL;
						ResourceManager.instance.getResource(themeURLToValidate, true, onThemeBytesLoaded, onThemeBytesLoaded, onThemeBytesLoaded);
					} else {
						// TODO For now we'll assume this is a designer skin set, but we could support Flex skins
						// in the future if we sniffed the bytes before creating a new property.
						importThemeFinalize(relativeURL, false);
					}
				});
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function onThemeBytesLoaded(event:Event):void
		{
			if (event == null) {
				LogHelper.logger.error("Null event loading imported theme");
				importThemeAbandon();
				return;
			}
			
			if ((event == null) || (event.type != Event.COMPLETE)) {
				LogHelper.logger.error("Error loading imported theme: "+event.toString());
				importThemeAbandon();
				return;
			}
			
			var loader:URLLoader = event.target as URLLoader;
			if (loader == null) {
				LogHelper.logger.error("Invalid loader for imported theme: "+event.toString());
				importThemeAbandon();
				return;
			}
			

			// Now that we've got the bytes, try to construct a SWFLoader to validate the theme
			var swfLoader:SWFLoader = new SWFLoader();
			swfLoader.loaderContext = new LoaderContext();
			swfLoader.loaderContext.allowCodeImport = true;
			swfLoader.addEventListener(Event.COMPLETE, onThemeSWFLoaded);
			swfLoader.addEventListener(IOErrorEvent.IO_ERROR, onThemeSWFLoaded);
			swfLoader.addEventListener(SecurityErrorEvent.SECURITY_ERROR, onThemeSWFLoaded);
			swfLoader.load(loader.data);
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function onThemeSWFLoaded(event:Event):void
		{
			if (event == null) {
				LogHelper.logger.error("Null event");
				importThemeAbandon();
				return;
			}
			var swfLoader:SWFLoader = event.target as SWFLoader;
			if (swfLoader == null) {
				LogHelper.logger.error("Null SWFLoader");
				importThemeAbandon();
				return;
			}
			
			swfLoader.removeEventListener(Event.COMPLETE, onThemeSWFLoaded);
			swfLoader.removeEventListener(IOErrorEvent.IO_ERROR, onThemeSWFLoaded);
			swfLoader.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, onThemeSWFLoaded);
			if (event.type != Event.COMPLETE) {
				LogHelper.logger.error("Error loading swf from imported bytes: "+event.toString());
				importThemeAbandon();
				return;
			}
			
			// Validate loaded swf
			var theme:DisplayObject = swfLoader.content;
			if (theme == null) {
				LogHelper.logger.error("Error loading swf from imported bytes: "+event.toString());
				importThemeAbandon();
				return;
			}
			
			var foundAsset:Boolean = false;
			// Make sure we have at least one recognized asset
			for each (var assetName:String in FlashSkinManager.knownAssetNames) {
				if (theme.hasOwnProperty(assetName)) {
					foundAsset = true;
					break;
				}
			}
			if (foundAsset) {
				importThemeFinalize(themeURLToValidate, false);
			} else {
				LogHelper.logger.error("Invalid theme swf: "+themeURLToValidate);
				importThemeAbandon();
			}
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function importThemeFinalize(relativeURL:String, flexTheme:Boolean = false):void
		{
			if (skinList == null) {
				LogHelper.logger.error("No skin list to update");
				importThemeAbandon();
				return;
			}
			// We're successfully importing a theme, so clear out the validation value (if necessary)
			themeURLToValidate = null;
			
			// TODO All of our themes should force explicit values for all background properties and colors.
			// For now these are hard-coded, but we might consider slamming in the values as they currently are.
			var newSkin:PropertyBinaryValue = PropertyValue.createPropertyValue(
				<SkinBinary name={IMPORTED_THEME_NAME} labelKey="customTheme" label="CUSTOM_THEME" url={relativeURL} type={flexTheme ? "flex" : "flash"}>
					<SelectEffect type="propertyValue" context="com.adobe.navigator.background" property="gradientType" value="linear"/>
					<SelectEffect type="propertyValue" context="com.adobe.navigator.background" property="gradientIntensity" value="70"/>
					<SelectEffect type="propertyValue" context="com.adobe.navigator.background" property="imageScale" value="noBorder"/>
					<SelectEffect type="propertyValue" context="com.adobe.navigator.background" property="imageOpacity" value="100"/>
					<SelectEffect type="propertyValue" context="com.adobe.navigator.background" property="imagePosition" value="center"/>				
					<SelectEffect type="propertyValue" context="com.adobe.navigator.background" property="imageSource" value="remove"/>
					<SelectEffect type="propertyValue" context="com.adobe.navigator.background" property="imageBlur" value="0"/>				
					<SelectEffect type="colorValue" property="background" value="0xC4C5C5"/>
					<SelectEffect type="colorValue" property="cardBackground" value="0x0078A6"/>
					<SelectEffect type="colorValue" property="cardBorder" value="0x999999"/>
					<SelectEffect type="colorValue" property="primaryText" value="0xFFFFFF"/>
					<SelectEffect type="colorValue" property="secondaryText" value="0xE9C171"/>
				</SkinBinary>,
				skinList) as PropertyBinaryValue;
			ResourceManager.instance.referenceResource(relativeURL);
			
			var newSkins:Array = new Array();
			var oldImportedURL:String = null;
			for each (var skinProperty:PropertyBinaryValue in skinList.subvalues) {
				if (skinProperty == null) {
					continue;
				}
				if (skinProperty.name == IMPORTED_THEME_NAME) {
					oldImportedURL = skinProperty.url;
					// We're going to replace this property, so dereference it
					ResourceManager.instance.dereferenceResource(oldImportedURL);
					newSkins.push(newSkin);
					continue;
				}
				newSkins.push(skinProperty);
			}
			if (oldImportedURL == null) {
				newSkins.push(newSkin);
			}
			skinList.updateSubvalues(newSkins, newSkin);

			// Delete the old URL
			if (oldImportedURL != null) {
				ResourceManager.instance.deleteResource(oldImportedURL);
			}
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function importThemeAbandon():void
		{
			if ((themeURLToValidate == null) || (themeURLToValidate == "")) {
				LogHelper.logger.warn("An unrecognized theme import is being abandoned");
				return;
			}
			ResourceManager.instance.deleteResource(themeURLToValidate);
			themeURLToValidate = null;
		}
		
		// Miscellaneous plumbing that should generally be ignored
		
		/**
		 *  The internal representation of the navigator hosting this object.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		private var _navigator:UIComponent;
		
		/**
		 *  Sets the navigator hosting this object.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set navigator(value:UIComponent):void
		{
			if (value == _navigator) {
				return;
			}
			_navigator = value;
			if (_navigator != null) {
				propertyManager.addEventListener(PropertyManagerEvent.CONTEXT_LOADED, onContextLoaded);
				propertyManager.addEventListener(Event.COMPLETE, onLoadXMLCompleteOrError);
				propertyManager.addEventListener(IOErrorEvent.IO_ERROR, onLoadXMLCompleteOrError);
				propertyManager.addEventListener(SecurityErrorEvent.SECURITY_ERROR, onLoadXMLCompleteOrError);
				initializeXMLQueue();
				
				// TODO Perhaps this assignment should wait until all the contexts are loaded?
				propertyManager.commonProperties = this;
				
				// TODO Do I need to check which contexts have already been loaded, or can
				// we assume we're doing all the loading?
				if (propertyManager.knownContexts != null) {
					for each (var context:PropertyContext in propertyManager.knownContexts) {
						if (context == null) {
							// Will "for each" ever do this to us?
							continue;
						}
						loadContext(context);
					}
				}
				handleNextXML();
			}
		}
		
		private var nextXML:String;
		private var xmlQueue:Array;
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function initializeXMLQueue():void
		{
			xmlQueue = createXMLList();
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function loadNextXML():Boolean
		{
			if (xmlQueue == null) {
				LogHelper.logger.error("loadNextXML() - xmlQueue was null");
				return false;
			}
			while (xmlQueue.length > 0) {
				nextXML = xmlQueue.shift();
				if (nextXML != "") {
					propertyManager.loadPropertyXML(nextXML);
					return true;
				}
			}
			xmlQueue = null;
			return false;
		}
		
		/** 
		 *  An event listener than is invoked when the PropertyManager
		 *  has finished loading a property context.
		 * 
		 *  @param event The property context loaded event
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onContextLoaded(event:PropertyManagerEvent):void
		{
			loadContext(event.context);
		}
		
		/**
		 *  An event listener than is invoked when a load XML command has
		 *  completed or has generated an error.
		 * 
		 *  @param event The load XML complete event or an error event generated when
		 *  trying to load the XML
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onLoadXMLCompleteOrError(event:Event):void
		{
			// TODO make sure this event corresponds to nextXML?
			handleNextXML();
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function handleNextXML():void
		{
			if (!loadNextXML()) {
				// We must be done loading all our properties
				propertyManager.removeEventListener(PropertyManagerEvent.CONTEXT_LOADED, onContextLoaded);
				propertyManager.removeEventListener(Event.COMPLETE, onLoadXMLCompleteOrError);
				propertyManager.removeEventListener(IOErrorEvent.IO_ERROR, onLoadXMLCompleteOrError);
				propertyManager.removeEventListener(SecurityErrorEvent.SECURITY_ERROR, onLoadXMLCompleteOrError);
				
				propertyManager.updateHostContexts();
				setCurrentContextByName(getDefaultCurrentContextName(), null);
			}
		
		}
		
		/**
		 *  A common event listener for handling changed properties.
		 *  This event listener will be called twice - once when the activeValue is changed
		 *  and again when the activeValue's value is changed. This listener ignores
		 *  the second event (when the activeValue's value is changed).
		 * 
		 *  @param event A property change event.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function onValueChange(event:PropertyChangeEvent):void
		{
			// This gets called twice - once when the activeValue is changed
			// and again when the activeValue's value is changed. Filter
			// out the unwanted second time we get called.
			if (event.property != "activeValue")
				return;

			if (event.target == skinList) {
				skin = event.newValue as SkinValue;
			}
			
			// Background properties
			if (event.oldValue == backgroundImageScale)
				backgroundImageScale = event.newValue as PropertyValue;
			if (event.oldValue == backgroundImagePosition)
				backgroundImagePosition = event.newValue as PropertyValue;
			if (event.oldValue == backgroundGradientType) {
				backgroundGradientType = event.newValue as PropertyValue;
				if ((backgroundGradientIntensity != null) && (backgroundGradientType != null)) {
					backgroundGradientIntensity.enabled = (backgroundGradientType.name != "none");
				}
			}
			
			var sizeValue:PropertyNumberValue;
			
			// portfolio properties
			if (event.oldValue == cardSize)
				cardSize = event.newValue as PropertyNumberValue;
			if (event.oldValue == portfolioFontName)
				portfolioFontName = event.newValue as PropertyValue;
			if (event.oldValue == portfolioFontSize)
				portfolioFontSize = event.newValue as PropertyNumberValue;
			if (event.oldValue == portfolioFontStyle)
				portfolioFontStyle = event.newValue as PropertyValue;
			
			// header properties
			if ((event.target == headerBackgroundGradientType) &&
				(headerBackgroundGradientType.activeValue != null) &&
				(headerBackgroundGradientIntensity != null))
			{
				var gradientType:String = headerBackgroundGradientType.activeValue.name;
				headerBackgroundGradientIntensity.enabled = ((gradientType == GradientType.LINEAR) || (gradientType == GradientType.RADIAL));
			}
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */
		private function setEmbedFonts(value:Boolean):void
		{
			FontManager.instance.embedFontsOnSave = value;
			
			// We've got two different properties for the same param,
			// so make sure they stay in sync
			if ((embedFonts != null) && (embedFonts.booleanValue != value)) {
				embedFonts.booleanValue = value;
			}
			if ((portfolioEmbedFonts != null) && (portfolioEmbedFonts.booleanValue != value)) {
				portfolioEmbedFonts.booleanValue = value;
			}
		}
		
		/**
		 *  Returns the size, in pixels, of cards with the specified size name
		 * 
		 *  @param sizeName The name of the card size: "small", "medium", or "large"
		 *  (PROPERTY_CARD_SIZE_SMALL, PROPERTY_CARD_SIZE_MEDIUM, PROPERTY_CARD_SIZE_LARGE)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function getCardSize(sizeName:String):Number 
		{
			var list:PropertyListValue = portfolioCardSizeList;
			if (!list || !list.subvalues || !list.subvalues.length)
				return NaN;
			
			for (var index:int = 0; index < list.subvalues.length; ++index) {
				var size:PropertyNumberValue = list.subvalues[index] as PropertyNumberValue;
				if (!size)
					continue;
				if (size.name == sizeName)
					return size.numValue;
			}
			return NaN;
		}
	}
}
