/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.properties
{
	import com.adobe.portfolio.api.ICollection;
	import com.adobe.portfolio.managers.CollectionManager;
	
	import flash.events.IEventDispatcher;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.events.PropertyChangeEvent;
	
	/**
	 *  The class of properties for font lists.  This is an extension of
	 *  the PropertyListValue class.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */	
	public class FontList extends PropertyListValue
	{
		/**
		 *  The XML element name for FontLists.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const XML_ELEMENT_NAME:String = "FontList";
		
		private var includeDefaultFont:Boolean = true;
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function FontList(target:IEventDispatcher=null)
		{
			super(target);
		}
		
		private var listenForEmbedded:Boolean;
		private var listenForApplication:Boolean;
		
		/**
		 * A listener for handling changes to the application or embedded font arrays.
		 */
		protected function onCollectionPropertyChange(event:PropertyChangeEvent):void	
		{
			if (event == null) {
				return;
			}
			if (event.property == "embeddedFontList") {
				if (!listenForEmbedded) {
					// We don't care about embedded fonts yet, so just return
					return;
				}
			} else if (event.property == "applicationFontList") {
				if (!listenForApplication) {
					// We don't care about application fonts yet, so just return
					return;
				}
			} else {
				// This isn't a font list change, so just return
				return;
			}
			
			// Load any newly-available fonts, but maintain our current selection
			loadFonts((activeValue == null) ? "" : activeValue.name, listenForApplication, listenForEmbedded);
		}
		
		/**
		 *  Encode bold/italic flags in the font attributes
		 * 
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		private function setFontStyleFlag(theFont:XML, fontEntry:Object):void
		{
			if (fontEntry.fontPosture == "italic")
			{
				if (fontEntry.fontWeight == "normal")
					theFont.@hasItalic = true;
				else
					theFont.@hasBoldItalic = true;
			}
			else
			{
				if (fontEntry.fontWeight == "bold")
					theFont.@hasBold = true;
				else
					theFont.@hasRegular = true;
			}
		}

		/**
		 *  Loads the list of available fonts as subvalues.
		 * 
		 * 	@param activeName The name of the desired active item once fonts have been loaded
		 * 	@param loadApplicationFonts true if all system fonts exposed by the application should be loaded; false otherwise
		 * 	@param loadEmbeddedFonts true if all embedded fonts in the portfolio should be loaded; false otherwise
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function loadFonts(activeName:String = "", loadApplicationFonts:Boolean = true, loadEmbeddedFonts:Boolean = true):void
		{
			var newActiveValue:PropertyValue = null;
			var newSubvalues:Array = new Array();
			
			// This is the name of the font we want to fallback to if the activeName isn't valid
			var backupName:String = "Myriad Pro";
			
			var fontProperty:PropertyValue;
			if (includeDefaultFont) {
				var defaultLabel:String = "TEXTFIELD_FONT_DEFAULT";
				fontProperty = new PropertyValue();
				if (CollectionManager.instance && (CollectionManager.instance.host != null)) {
					var hostCollection:ICollection = (CollectionManager.instance.host != null) ? CollectionManager.instance.host.collection : null;
					if (hostCollection != null) {
						if (CollectionManager.instance.applicationFont != null) {
							defaultLabel = hostCollection.getLocalizedString("textfieldFontDefaultWithName", "");
							if ((defaultLabel != null) && (defaultLabel != "")) {
								defaultLabel = defaultLabel.replace("%font%", CollectionManager.instance.applicationFont.fontName);
							} else {
								defaultLabel = hostCollection.getLocalizedString("textfieldFontDefault", "TEXTFIELD_FONT_DEFAULT");
							}
						} else {
							defaultLabel = hostCollection.getLocalizedString("textfieldFontDefault" ,"TEXTFIELD_FONT_DEFAULT");
						}
					}
				}
				fontProperty.loadFromXML(<PropertyValue name="applicationFont" label={defaultLabel}/>);
				newSubvalues.push(fontProperty);
			}
			
			if (CollectionManager.instance && (loadEmbeddedFonts || loadApplicationFonts))
			{
				if (!listenForEmbedded && !listenForApplication) {
					// Start listening for font changes
					CollectionManager.instance.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onCollectionPropertyChange, false, 0, true);
				}
				backupName = CollectionManager.instance.getDefaultFontName(includeDefaultFont);
				// Collect all the fontnames and sort them so we can only push one font entry
				// but keep track of which styles the font has
				
				// TODO This assumes that embeddedFontList is available when asked for
				// If we need to support lazy population of these lists, then they need to be made Bindable,
				// and we need to bind to them here.
				var embeddedFonts:Array = new Array();
				if (loadEmbeddedFonts) {
					if (CollectionManager.instance.embeddedFontList != null) {
						for each (fontEntry in CollectionManager.instance.embeddedFontList)
						{
							embeddedFonts.push(fontEntry);
						}
					}
					listenForEmbedded = true;
				}
				if (loadApplicationFonts) {
					if (CollectionManager.instance.applicationFontList != null) {
						// Add the system fonts as well
						for each (var fontEntry:Object in CollectionManager.instance.applicationFontList)
						{
							if (fontEntry.fontLookup == "embedded_cff")
							{
								embeddedFonts.push(fontEntry);
							}
						}
					}
					listenForApplication = true;
				}
				
				var previousName:String = null;
				embeddedFonts.sortOn("fontName");
				var theFont:XML = <PropertyValue />;
				for each (fontEntry in embeddedFonts)
				{
					if (fontEntry.fontName != previousName)
					{
						if (previousName != null)
						{
							// Don't take it if there's no regular face, it makes Flash sad
							if (theFont.@hasRegular == "true")
							{
								fontProperty = new PropertyValue();
								fontProperty.loadFromXML(theFont);
								newSubvalues.push(fontProperty);
							}
						}
						previousName = fontEntry.fontName;
						theFont.@name = fontEntry.fontName;
						theFont.@label = fontEntry.fontName;
						theFont.@hasRegular = false;
						theFont.@hasBold = false;
						theFont.@hasItalic = false;
						theFont.@hasBoldItalic = false;
						setFontStyleFlag(theFont, fontEntry);
						theFont.@family = fontEntry.psName.split("-")[0];
					}
					else
					{
						setFontStyleFlag(theFont, fontEntry);
					}
				}
				
				if (previousName != null && theFont.@hasRegular == "true")
				{
					fontProperty = new PropertyValue();
					fontProperty.loadFromXML(theFont);
					newSubvalues.push(fontProperty);
				}
			}
			
			var backupValue:PropertyValue;
			for each (fontProperty in newSubvalues) {
				if (fontProperty.name == activeName) {
					newActiveValue = fontProperty;
				} else if (fontProperty.name == backupName) {
					backupValue = fontProperty;
				}
			}
			if (newActiveValue == null) {
				if (activeName != "") {
					// We might not have loaded our desired font yet, so just create a dummy entry for now
					var fakeFont:XML = <PropertyValue />;
					if (activeName != null && activeName != "")
					{
						fakeFont.@name = activeName;
						fakeFont.@label = activeName;
					}
					else
					{
						fakeFont.@name = backupName;
						fakeFont.@label = backupName;
					}
					fakeFont.@hasRegular = true;
					fakeFont.@hasBold = true;
					fakeFont.@hasItalic = true;
					fakeFont.@hasBoldItalic = true;
					var fakeValue:PropertyValue = new PropertyValue();
					fakeValue.loadFromXML(fakeFont);
					newSubvalues.push(fakeValue);
					newActiveValue = fakeValue;
				}
				else {
					// No specific entry was requested, so use our backup value (or the first in the list if we don't have one)
					if (backupValue != null) {
						newActiveValue = backupValue;
					} else if (newSubvalues.length > 0) {
						newActiveValue = newSubvalues[0];
					}
				}
			}
			
			// Don't dirty the document when we update the font list
			var wasInitializing:Boolean = initializing;
			initializing = true;
			
			updateSubvalues(newSubvalues, newActiveValue);
			
			initializing = wasInitializing;
		}
		
		/**
		 * 	Since font list items are not stored in the portfolio itself, its
		 * 	subvalues are not loaded or saved to XML.
		 * 
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function get loadAndSaveValues():Boolean
		{
			return false;
		}
		
		/**
		 *  Load the specified XML object FontList.
		 *  If in authoring mode, also load all the system fonts.
		 *  If possible, the localized label and description will be used.
		 * 
		 *  @return <code>true</code> if the font list was successfully loaded,
		 *  <code>false</code> otherwise.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function loadFromXML(valueXML:XML):Boolean
		{
			if (super.loadFromXML(valueXML)) {
				if (String(valueXML.@includeDefaultFont).toLowerCase() == "false") {
					includeDefaultFont = false;
				}
				
				// Only load all the system fonts if we're in authoring mode
				var activeName:String = valueXML.@active;
				if ((CollectionManager.instance.authoringMode) 
					&& (CollectionManager.instance.applicationFontList)
					&& (CollectionManager.instance.applicationFontList.length != 0)) {
					loadFonts(activeName, true);
				} else {
					loadFonts(activeName, false);
					_authoringWatcher = BindingUtils.bindSetter(loadFontsWhenReady, CollectionManager.instance, "authoringMode");
				}
				return true;
			}
			return false;
		}

		private var _authoringWatcher:ChangeWatcher = null;
		private function loadFontsWhenReady(value: Object): void
		{
			// All variables not initialized
			if (!CollectionManager.instance.authoringMode) 
				return;
			
			// Remove watchers
			if (_authoringWatcher != null) {
				_authoringWatcher.unwatch();
				_authoringWatcher = null;
			}

			if ((CollectionManager.instance.applicationFontList != null) &&
				(CollectionManager.instance.applicationFontList.length > 0))
			{
				// Application fonts are already available, so go ahead and load them
				loadFonts((activeValue == null) ? "" : activeValue.name, true);
			} else {
				// Application fonts aren't available, so let onCollectionPropertyChange
				// process them when they are.
				listenForApplication = true;
			}
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function getXMLElementName():String
		{
			return XML_ELEMENT_NAME;
		}
		
		/**
		 *  Returns <code>false</code>. The list of fonts is generated programatically, 
		 *  so we don't output them to XML
		 * 
		 * 	@inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function get shouldOutputSubvalues():Boolean
		{
			// The list of fonts is generated programatically, so don't output them to XML
			return false;
		}		
	}
}
