/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.properties
{
	import com.adobe.portfolio.api.IPropertyContext;
	import com.adobe.portfolio.events.PropertyManagerEvent;
	import com.adobe.portfolio.managers.CollectionManager;
	
	import flash.events.EventDispatcher;
	import flash.events.IEventDispatcher;
	import flash.xml.XMLNode;
	
	import mx.events.PropertyChangeEvent;
	
	/**
	 *  This class provides support for groups of properties displayed in the user interface. 
	 * 
	 *  <p>The PropertyContext object represents a grouping of properties that are relevant to each other.
	 *  Property contexts can be global, which indicates that the group of properties they contain
	 *  are always suitable for presentation in the UI, or transient, in which case the properties
	 *  may or may not be relevant depending on the user's actions.</p>
	 * 
	 *  <p>A PropertyContext object contains information for displaying a group of properties in the UI. 
	 *  The getProperty function allows a navigator host to trigger the creation of Property objects, 
	 *  while still allowing the Navigator on the other side of the bridge to actually create them.
	 *  The navigator host can create its needed properties, configure them, and then set their values.</p>
	 * 
	 *  <p>Note that the full set of properties available at any point is the union of the global properties
	 *  (those perperties whose global property is <code>true</code>) and the properties relevant to the
	 *  current context.</p>
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */	
	public class PropertyContext extends EventDispatcher implements IPropertyContext
	{
		/**
		 *  Constructor
		 * 
		 *  @param url	The URL of the property context.  The default is ""
		 *  @param target The event dispatcher to be passed to the host The default is null
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function PropertyContext(url:String = "", target:IEventDispatcher=null)
		{
			super(target);
			_url = url;
		}
		
		private var _properties:Array = new Array();
		/**
		 *  The array of properties
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get properties():Array
		{
			return _properties;
		}
		
		/**
		 *  Returns the PropertyValue of the property with the given name,
		 *  <code>null</code> otherwise.
		 * 
		 *  @param name the name of the specified property
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function getProperty(name:String):PropertyValue
		{
			for each (var propertyObj:Object in _properties) {
				var property:PropertyValue = propertyObj as PropertyValue;
				if (property == null) {
					// This array can also contain nested contexts
					continue;
				}
				if (property.name == name) {
					return property;
				}
			}
			return null;
		}
		
		private var _url:String = "";
		/**
		 *  The URL of the PropertyContext
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get url():String
		{
			return _url;
		}
		
		private var _label:String = "";
		/**
		 *  The localized label for the PropertyContext
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get label():String
		{
			return _label;
		}
		
		/**
		 *  The string used as a key to request a localized label string.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var labelKey:String = "";
		
		/**
		*  The default label string to be used when the localized label is unavailable.
		* 
		*  @langversion 3.0
		*  @playerversion Flash 10.1
		*  @playerversion AIR 2.0
		*  @productversion Portfolio 10.0
		*
		*/	
		protected var labelDefault:String = "";
		
		private var _description:String = "";
		/**
		 *  The localized description of the PropertyContext
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get description():String
		{
			return _description;
		}
		
		/**
		 *  The string used as a key to request a localized description string.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var descriptionKey:String = "";
		
		/**
		 *  The default description string to be used when the localized description is unavailable.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var descriptionDefault:String = "";
		
		private var _global:Boolean = false;
		/**
		 *  Returns a value of <code>true</code> if the controls for this context
		 *  are always suitable for display in the user interface. Returns 
		 *  <code>false</code> if the controls for this context should only be 
		 *  displayed when it is set as the current context. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get global():Boolean
		{
			return _global;
		}
		
		private var _name:String = "";
		/**
		 *  Returns a unique string identifier for this PropertyContext.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get name():String
		{
			return _name;
		}
		
		private var _dirty:Boolean = false;
		/**
		 *  <code>true</code> if this PropertyContext has been
		 *  modified and needs to be saved; <code>false</code> otherwise.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get dirty():Boolean
		{
			return _dirty;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set dirty(value:Boolean):void
		{
			if (_dirty != value) {
				_dirty = value;
				if (_dirty && (url != "")) {
					CollectionManager.instance.setDirty();
				}
			}
		}

		/**
		 *  Load the PropertyContext from an XML object and
		 *  localize the values.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function loadFromXML(contextXML:XML, forceGlobal:Boolean = false):Boolean
		{
			if (forceGlobal) {
				_global = true;
			} else {
				var xmlBooleanValue:String = contextXML.@global;
				_global = (xmlBooleanValue.toLowerCase() == "true");
			}
			_name = contextXML.@uid;

			// Try to get a localized label
			labelKey = contextXML.@labelKey;
			if ((labelKey != null) &&
				(labelKey != "") &&
				(CollectionManager.instance.host != null) &&
				(CollectionManager.instance.host.collection != null))
			{
				labelDefault = contextXML.@label;
				_label = CollectionManager.instance.host.collection.getLocalizedString(labelKey, labelDefault);
			} else {
				labelDefault = "";
				_label = contextXML.@label;
			}
			
			// Try to get a localized description
			descriptionKey = contextXML.@descriptionKey;
			if ((descriptionKey != null) &&
				(descriptionKey != "") &&
				(CollectionManager.instance.host != null) &&
				(CollectionManager.instance.host.collection != null))
			{
				descriptionDefault = contextXML.@description;
				_description = CollectionManager.instance.host.collection.getLocalizedString(descriptionKey, descriptionDefault);
			} else {
				descriptionDefault = "";
				_description = contextXML.@description;
			}
			
			var values:XMLList = contextXML.elements();
			for each (var valueNode:XML in values) {
				var localName:String = valueNode.localName();
				if (localName == "PropertyContext") {
					var newSubcontext:PropertyContext = new PropertyContext(url);
					newSubcontext.loadFromXML(valueNode);
					_properties.push(newSubcontext);
				} else {
					var newValue:PropertyValue = PropertyValue.createPropertyValue(valueNode, this);
					if (newValue != null) {
						_properties.push(newValue);
						newValue.addEventListener(PropertyChangeEvent.PROPERTY_CHANGE, onPropertyUpdate);
					}
				}
			}
			return true;
		}
		
		/**
		 *  Create a set of XML elements from the PropertyContext.
		 * 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function generateXML():XML
		{
			var newXML:XML = <PropertyContext/>;
			newXML.@uid = _name;
			newXML.@global = _global;
			
			if ((labelKey != null) && (labelKey != "")) {
				newXML.@label = labelDefault;
				newXML.@labelKey = labelKey;	
			} else {
				newXML.@label = _label;
			}
			if ((descriptionKey != null) && (descriptionKey != "")) {
				newXML.@description = descriptionDefault;
				newXML.@descriptionKey = descriptionKey;	
			} else {
				newXML.@description = _description;
			}
			
			for each (var propertyObj:Object in _properties) {
				var currentProperty:PropertyValue = propertyObj as PropertyValue;
				if (currentProperty != null) {
					var propertyXML:XML = currentProperty.generateXML();
					if (propertyXML != null) {
						newXML.appendChild(propertyXML);
					}
					continue;
				}
				var subcontext:PropertyContext = propertyObj as PropertyContext;
				if (subcontext != null) {
					var subcontextXML:XML = subcontext.generateXML();
					if (subcontextXML != null) {
						newXML.appendChild(subcontextXML);
					}
					continue;
				}
			}
			dirty = false;
			return newXML;
		}
		
		private var updatesFromHost:Object = new Object();
		/**
		 *  Set a new value for a property that is supplied by the host
		 * 
		 *  @param propertyName The name of the property that is to be updated.
		 *  @param newValue the new value to be set for the property
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function updatePropertyFromHost(propertyName:String, newValue:*):void
		{
			var property:PropertyValue = getProperty(propertyName) as PropertyValue;
			if (property != null) {
				updatesFromHost[propertyName] = true;
				property.value = newValue;
				updatesFromHost[propertyName] = false;
			}
		}
		
		/**
		 *  Watcher to handle property change events
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		private function onPropertyUpdate(event:PropertyChangeEvent):void
		{
			if (CollectionManager.instance.host == null) {
				return;
			}
			var propertyValue:PropertyValue = event.target as PropertyValue;
			if (propertyValue == null) {
				return;
			}
			
			// Only certain property changes should propagate across the bridge.
			// If properties are added to IProperty, then they should be added here as well
			var propertyString:String = event.property as String;
			if ((propertyString != "value") &&
				(propertyString != "label") &&
				(propertyString != "description") &&
				(propertyString != "visible") &&
				(propertyString != "enabled") &&
				(propertyString != "hasMinMax") &&
				(propertyString != "minValue") &&
				(propertyString != "maxValue") &&
				(propertyString != "subvalues") &&
				(propertyString != "subvaluesLocked") &&
				(propertyString != "subvaluesType"))
			{
				return;
			}
			
			if (!updatesFromHost[propertyValue.name]) {
				CollectionManager.instance.host.updateProperty(this, propertyValue);
			}
		}
	}
}
