/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.properties
{
	import com.adobe.portfolio.api.PropertyType;
	import com.adobe.portfolio.managers.ResourceManager;
	import com.adobe.portfolio.utils.LogHelper;
	
	import flash.events.IEventDispatcher;
	
	/**
	 *  This interface extends PropertyValue to provide support
	 *  for lists of PropertyValues displayed in the user interface.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */	
	public class PropertyListValue extends PropertyValue
	{
		/**
		 *  The XML element name for PropertyLists
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const XML_ELEMENT_NAME:String = "PropertyList";

		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function PropertyListValue(target:IEventDispatcher=null)
		{
			super(target);
		}
		
		private var _subvaluesLocked:Boolean = true;
		/**
		 *  Returns <code>true</code> if elements of subvalues for this property cannot be added or removed.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function get subvaluesLocked():Boolean
		{
			return _subvaluesLocked;
		}
		
		/**
		 *  @inheritDoc
		 * 
		 * 	@see PropertyType#LIST
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function get type():String
		{
			return PropertyType.LIST;
		}
		
		/**
		 *  Identifies the data type of the IProperty elements found in the subvalues array.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function get subvaluesType():String
		{
			return PropertyType.UNDEFINED;
		}
		
		[Bindable]
		/**
		 *  The currently active value.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get activeValue():PropertyValue
		{
			return value as PropertyValue;
		}
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function set activeValue(value:PropertyValue):void
		{
			super.value = value;
			
			// Don't play the activeValue's effects if our own effects are disabled
			if ((dontPlayEffectsCounter == 0) && (value != null)) {
				value.playSelectEffects();
			}
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function set value(newValue:Object):void
		{
			// Force all value assignments to trigger an activeValue PropertyChangeEvent as well
			activeValue = newValue as PropertyValue;
		}
		
		/**
		 *  The array of possible values for a PropertyList
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var _subvalues:Array = new Array();
		/**
		 *  The array of possible values for a PropertyList
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function get subvalues():Array
		{
			return _subvalues;
		}
		
		/**
		 * Returns <code>false</code> if loading and saving values to and from XML should be skipped.
		 * This may be desirable if the list values are based on the runtime state of the system.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function get loadAndSaveValues():Boolean
		{
			return true;
		}
		
		/**
		 *  Update the subvalues and active value of the property list to the new list and preperty value
		 * 
		 *  @param newSubvalues The new list of property values to use
		 *  @param newActiveValue The value to use as the new active value
		 * 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function updateSubvalues(newSubvalues:Array, newActiveValue:PropertyValue = null):void
		{
			if (newSubvalues == null) {
				LogHelper.logger.error("updateSubvalues cannot be called with null newSubvalues");
				return;
			}
			_subvalues = newSubvalues;
			
			// TODO Do we need a new way to dispatch an event when subvalues change? Or maybe just overload value changed?
			if (newActiveValue != null) {
				activeValue = newActiveValue;
			}
		}
		
		/**
		 * References, dereferences, or touches all URL subvalues.
		 * 
		 * @param referenceType If greater than zero, referenceResource is called on all URL subvalues;
		 * 						if less than zero, dereferenceResource is called; if equal to zero,
		 * 						touchResource is called.
		 * 
		 * @see com.adobe.portfolio.managers.ResourceManager#referenceResource
		 * @see com.adobe.portfolio.managers.ResourceManager#dereferenceResource
		 * @see com.adobe.portfolio.managers.ResourceManager#touchResource
		 * @see com.adobe.portfolio.managers.ResourceManager#purgeResources
		 */
		public function referenceSubvalueResources(referenceType:int):void
		{
			if ((subvalues == null) || (subvalues.length == 0)) {
				return;
			}
			var resourceManager:ResourceManager = ResourceManager.instance;
			var referenceFunction:Function;
			if (referenceType > 0) {
				referenceFunction = resourceManager.referenceResource;
			} else if (referenceType < 0) {
				referenceFunction = resourceManager.dereferenceResource;	
			} else {
				referenceFunction = resourceManager.touchResource;
			}
			for each (var subproperty:PropertyValue in subvalues) {
				if (subproperty.type == PropertyType.URL) {
					referenceFunction.call(resourceManager, subproperty.value as String);
				}
			}
		}
		
		/**
         *  Loads the list of property values from the specified XML value. 
		 * 
		 *  @param valueXML The XML object to load that contains the list of property values as its elements. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function loadFromXML(valueXML:XML):Boolean
		{
			if (super.loadFromXML(valueXML)) {
				++dontPlayEffectsCounter;
				if (loadAndSaveValues) {
					var newSubvalues:Array = new Array();
					var newActiveValue:PropertyValue;
					var activeName:String = valueXML.@active;
					var subElements:XMLList = valueXML.elements();
					for each (var subElement:XML in subElements) {
						var subvalue:PropertyValue = PropertyValue.createPropertyValue(subElement, this);
						if (subvalue != null) {
							newSubvalues.push(subvalue);
							if (subvalue.name == activeName) {
								newActiveValue = subvalue;
							}
						}
					}
					updateSubvalues(newSubvalues, newActiveValue);
				}
				if (String(valueXML.@fixed).toLowerCase() == "false") {
					_subvaluesLocked = false;
				}
				--dontPlayEffectsCounter;
				return true;
			} else {
				return false;
			}
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function getXMLElementName():String
		{
			return XML_ELEMENT_NAME;
		}
		
		/**
		 *  Returns <code>true</code>. The default (unless overridden) is to output subvalues to XML.
		 * 
		 * 	@inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function get shouldOutputSubvalues():Boolean
		{
			return true;
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function setXMLElements(propertyXML:XML):void
		{
			if (propertyXML == null) {
				return;
			}
			super.setXMLElements(propertyXML);
			
			propertyXML.@active = (activeValue == null) ? "" : activeValue.name;
			propertyXML.@fixed = _subvaluesLocked ? "true" : "false";
			
			if (shouldOutputSubvalues) {
				for each (var subvalue:PropertyValue in _subvalues) {
					var subvalueXML:XML = subvalue.generateXML();
					if (subvalueXML != null) {
						propertyXML.appendChild(subvalueXML);
					}
				}
			}
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function setValueFromString(valueString:String):Boolean
		{
			if (valueString == null) {
				return false;
			}
			for each (var subvalue:PropertyValue in _subvalues) {
				if (subvalue.name == valueString) {
					activeValue = subvalue;
					return true;
				}
			}
			return false;
		}
	}
}
