/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.properties
{
	import com.adobe.portfolio.api.PropertyType;
	
	import flash.events.IEventDispatcher;
	
	/**
	 *  This class handles property values that are numeric.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */	
	public class PropertyNumberValue extends PropertyValue
	{
		/**
		 *  The XML element name for numeric property values
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const XML_ELEMENT_NAME:String = "PropertyNumber";
		
		/**
		 * The numeric type of this property. If "int" then the value is treated as
		 * an int; if "uint" then the value is treated as a uint; otherwise the
		 * value is treated as a Number.
		 */
		protected var numType:String = "";
		
		private var _minValue:Number = Number.NEGATIVE_INFINITY;
		private var _maxValue:Number = Number.POSITIVE_INFINITY;
		
		/**
		 *  @inheritDoc
		 * 
		 * 	@see PropertyType#INTEGER
		 * 	@see PropertyType#NUMBER
		 * 	@see PropertyType#UNSIGNED_INTEGER
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function get type():String
		{
			switch (numType) {
				case "int":
					return PropertyType.INTEGER;
				case "uint":
					return PropertyType.UNSIGNED_INTEGER;
				default:
					return PropertyType.NUMBER;
			}
		}
		
		[Bindable]
		/**
		 *  The value as a number. Setting this value will trigger a numValue PropertyChangeEvent as well.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get numValue():Number
		{
			return value as Number;
		}
		/**
		 *  @private
		 */	
		public function set numValue(newValue:Number):void
		{
			if (newValue > _maxValue) {
				super.value = _maxValue;
			} else if (newValue < _minValue) {
				super.value = _minValue;
			} else {
				super.value = newValue;
			}
		}
		
		/**
		 *  @private
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function set value(newValue:Object):void
		{
			// Force all value assignments to trigger a numValue PropertyChangeEvent as well
			if (numValue == (newValue as Number)) {
				return;
			}
			if (isNaN(numValue) && isNaN(newValue as Number)) {
				return;
			}
			numValue = newValue as Number;
		}
		
		/**
		 *  Returns <code>true</code> if the propertyNumberValue is an int;
		 *  returns <code>false</code> otherwise.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get isInt():Boolean
		{
			return (numType == "int");
		}
		
		/**
		 *  Returns <code>true</code> if the propertyNumberValue is an unsigned integer;
		 *  returns <code>false</code> otherwise.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function get isUint():Boolean
		{
			return (numType == "uint");
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function get hasMinMax():Boolean
		{
			return ((_minValue > Number.NEGATIVE_INFINITY) || (_maxValue < Number.POSITIVE_INFINITY));
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function get minValue():Number
		{
			return _minValue;
		}
		
		public function set minValue(newValue:Number):void
		{
			if (_minValue != newValue) {
				var oldValue:Object = _minValue;
				_minValue = newValue;
				if (!initializing) {
					// If we're not initializing this property, dirty it when its value changes
					dirty = true;
				}
			}
		}

		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function get maxValue():Number
		{
			return _maxValue;
		}
		
		public function set maxValue(newValue:Number):void
		{
			if (_maxValue != newValue) {
				var oldValue:Object = _maxValue;
				_maxValue = newValue;
				if (!initializing) {
					// If we're not initializing this property, dirty it when its value changes
					dirty = true;
				}
				//				dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "maxValue", oldValue, _maxValue));
			}
		}

		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function PropertyNumberValue(target:IEventDispatcher=null)
		{
			super(target);
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function loadFromXML(valueXML:XML):Boolean
		{
			if (super.loadFromXML(valueXML)) {
				var xmlNumType:String = valueXML.@type;
				if (xmlNumType != null && xmlNumType.length > 0)
					numType = xmlNumType;
				var xmlNumValue:String = valueXML.@value;
				if (xmlNumValue != null && xmlNumValue.length > 0) {
					if ((numType == "uint") || (numType == "int")) {
						numValue = parseInt(xmlNumValue);
					} else {
						numValue = parseFloat(xmlNumValue);
					}
				}
				var xmlMinValue:String = valueXML.@minValue;
				if (xmlMinValue != null && xmlMinValue.length > 0)
					_minValue = parseFloat(xmlMinValue);
				var xmlMaxValue:String = valueXML.@maxValue;
				if (xmlMaxValue != null && xmlMaxValue.length > 0)
					_maxValue = parseFloat(xmlMaxValue);
				return true;
			} else {
				return false;
			}
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function getXMLElementName():String
		{
			return XML_ELEMENT_NAME;
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function setXMLElements(propertyXML:XML):void
		{
			if (propertyXML == null) {
				return;
			}
			super.setXMLElements(propertyXML);
			propertyXML.@type = numType;
			switch (numType) {
				case "int":
					propertyXML.@value = numValue as int;
					break;
				case "uint":
					propertyXML.@value = numValue as uint;
					break;
				default:
					propertyXML.@value = numValue;
			}
			if (_minValue > Number.NEGATIVE_INFINITY) {
				propertyXML.@minValue = _minValue;
			}
			if (_maxValue < Number.POSITIVE_INFINITY) {
				propertyXML.@maxValue = _maxValue;
			}
		}
		
		/**
		 *  Set the property number value equal to the specified string.
		 *  The value parsed will depend on the numType of the propertyNumberValue.
		 *  If the type is an integer or unsigned integer, the value will be parsed as an integer,
		 *  otherwise it will be parse as a floating point number.
		 * 
		 *  @param valueString The numeric value as a string.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function setValueFromString(valueString:String):Boolean
		{
			if (valueString == null) {
				return false;
			}
			if ((numType == "uint") || (numType == "int")) {
				numValue = parseInt(valueString);
			} else {
				numValue = parseFloat(valueString);
			}
			return true;
		}
	}
}