package com.adobe.portfolio.properties
{
	import com.adobe.portfolio.api.PropertyType;
	import com.adobe.portfolio.managers.PropertyManager;
	import com.adobe.portfolio.managers.ResourceManager;
	import com.adobe.portfolio.utils.LogHelper;
	
	import flash.events.IEventDispatcher;
	
	import mx.binding.utils.BindingUtils;
	import mx.binding.utils.ChangeWatcher;
	import mx.collections.ArrayCollection;

	/**
	 *  This class handles property values that are theme filtered lists.  Theme filtered lists
	 *  are property lists that filter out subvalues whose names match those of themes
	 *  that are not currently active.
	 * 
	 * 	@see CommonProperties#skinList
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */	
	public class ThemeFilteredList extends PropertyListValue
	{
		/**
		 *  The XML element name for theme filtered lists
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public static const XML_ELEMENT_NAME:String = "ThemeFilteredList";
		
		/**
		 *  Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public function ThemeFilteredList(target:IEventDispatcher=null)
		{
			super(target);
		}
		
		/**
		 *  Update the subvalues and active value of the theme filtered list to the new filter list and active value
		 * 
		 *  @param newSubvalues The new list of theme values to use for filtering
		 *  @param newActiveValue The value to use as the new active value
		 * 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		public override function updateSubvalues(newSubvalues:Array, newActiveValue:PropertyValue = null):void
		{
			if ((newSubvalues == null) || !buildThemedSubvalues(newSubvalues)) {
				// We couldn't build our list of themes subvalues, so just incorporate all values
				super.updateSubvalues(newSubvalues, newActiveValue);
				return;
			}
			var adjustedActiveValue:PropertyValue = updateFilteredValues(newSubvalues, newActiveValue);
			super.updateSubvalues(filteredSubvalues, adjustedActiveValue);
		}
		
		/**
		 *  The complete (unfiltered) list of subvalues
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var completeSubvalues:Array;
		
		/**
		 *  The filtered list of subvalues; subvalues whose name matches a non-active theme
		 * 	have been removed.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var filteredSubvalues:Array;
		
		/**
		 *  An array of all theme names
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var themeNames:ArrayCollection;
		
		/**
		 *  An array of subvalues whose name matches a theme name
		 * 
		 *  @review
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var themedSubvalues:ArrayCollection;
		
		/**
		 *  The name of the current active theme
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected var currentThemeName:String = "";
		
		private var currentThemeWatcher:ChangeWatcher;
		private function setActiveTheme(value:PropertyValue):void
		{
			currentThemeName = (value == null) ? "" : value.name;
			updateSubvalues(subvalues, activeValue);
		}
		
		/**
		 * @inheritDoc
		 */
		override public function referenceSubvalueResources(referenceType:int):void
		{
			if ((completeSubvalues == null) || (completeSubvalues.length == 0)) {
				return;
			}
			var resourceManager:ResourceManager = ResourceManager.instance;
			var referenceFunction:Function;
			if (referenceType > 0) {
				referenceFunction = resourceManager.referenceResource;
			} else if (referenceType < 0) {
				referenceFunction = resourceManager.dereferenceResource;	
			} else {
				referenceFunction = resourceManager.touchResource;
			}
			for each (var subproperty:PropertyValue in completeSubvalues) {
				if (subproperty.type == PropertyType.URL) {
					referenceFunction.call(resourceManager, subproperty.value as String);
				}
			}
		}
		
		/**
		 *  Sets the theme list to use for filtering and updates the current theme watcher.
		 * 
		 *  @param value The list of themes on which to filter
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function setThemeList(value:PropertyListValue):void
		{
			if (currentThemeWatcher != null) {
				currentThemeWatcher.unwatch();
				currentThemeWatcher = null;
			}
			
			if (value == null) {
				themeNames = null;
				currentThemeName = "";
				return;
			}
			
			themeNames = new ArrayCollection();
			for each (var themeProperty:PropertyValue in value.subvalues) {
				themeNames.addItem(themeProperty.name);
			}
			currentThemeWatcher = BindingUtils.bindSetter(setActiveTheme, value, "activeValue");
		}
		
		/**
		 *  Create the list of themed subvalues from the valueArray passed.
		 * 
		 *  @param valueArray The new list of value to 
		 *  @param forceRebuild If forceRebuild is <code>false</code> (the default), the 
		 *  themedSubvalues array with not be rebuilt if it already exists. If forceRebuild is <code>true</code>
		 *  the current themedSubvalues array will be replaced with the new list.
		 *  @return <code>true</code> if the list of themed values could be built, otherwise <code>false</code>.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function buildThemedSubvalues(valueArray:Array, forceRebuild:Boolean = false):Boolean
		{
			if (!forceRebuild && (themedSubvalues != null)) {
				return true;
			}
			if ((themeNames == null) ||
				(themeNames.length <= 0) ||
				(valueArray == null) ||
				(valueArray.length == 0))
			{
				themedSubvalues = null;
				return false;
			}
			themedSubvalues = new ArrayCollection();
			for each (var subvalue:PropertyValue in valueArray) {
				if (themeNames.contains(subvalue.name)) {
					themedSubvalues.addItem(subvalue);
				}
			}
			return true;
		}
		
		/**
		 *  Update the list of filtered subvalues based on the full list of subvalues and the
		 * 	currently active theme.
		 * 
		 *  @param newSubvalues The new list of theme values to use for filtering
		 *  @param newActiveValue The value to use as the new active value
		 *  @return Returns the currentTheme if that is part of the newSubvalues list.  Otherwise it returns
		 *  the newActiveValue.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		protected function updateFilteredValues(newSubvalues:Array, newActiveValue:PropertyValue):PropertyValue
		{
			if ((newSubvalues == null) || (themedSubvalues == null) || (currentThemeName == ""))
			{
				return null;
			}
			var pendingActiveValue:PropertyValue = (newActiveValue == null) ? activeValue : newActiveValue;
			var matchingSubvalue:PropertyValue;
			var didRemoveActiveValue:Boolean = false;
			var desiredCurrentIndex:int = -1;
			var filteredValues:ArrayCollection = new ArrayCollection();
			var completeValues:ArrayCollection = new ArrayCollection();
			for (var i:int; i < newSubvalues.length; ++i) {
				var subvalue:PropertyValue = newSubvalues[i];
				if (subvalue == null) {
					continue;
				}
				if (!themeNames.contains(subvalue.name)) {
					// Not a theme-specific value, so keep it
					filteredValues.addItem(subvalue);
					completeValues.addItem(subvalue);
					continue;
				}
				if (desiredCurrentIndex < 0) {
					desiredCurrentIndex = i;
				}
				if (currentThemeName == "") {
					// We don't have a current them, so keep everything
					filteredValues.addItem(subvalue);
					continue;
				}
				if (subvalue.name == currentThemeName) {
					// Matches current theme, so keep it
					filteredValues.addItem(subvalue);
					matchingSubvalue = subvalue;
					continue;
				}
				// Doesn't match the current theme, so remove it, but hold onto the index
				// so that we can insert the matching subvalue if necessary
				if (subvalue == pendingActiveValue) {
					// We're about to remove the active value, so replace it later
					didRemoveActiveValue = true;
				}
			}
			if ((currentThemeName != "") && (matchingSubvalue == null)) {
				// We didn't find a property for our current theme, so try to find one
				for each (var themedSubvalue:PropertyValue in themedSubvalues) {
					if (themedSubvalue.name == currentThemeName) {
						matchingSubvalue = themedSubvalue;
						break;
					}
				}
				if (matchingSubvalue == null) {
					LogHelper.logger.error("Couldn't find a subvalue to match: "+matchingSubvalue);
				} else if (desiredCurrentIndex < 0) {
					// We didn't find a theme-specific value, so just add our match to the end
					filteredValues.addItem(matchingSubvalue);
				} else {
					filteredValues.addItemAt(matchingSubvalue, desiredCurrentIndex);
				}
			}
			
			if (desiredCurrentIndex < 0) {
				desiredCurrentIndex = completeValues.length;
			}
			for (var j:int = themedSubvalues.length; j > 0; --j) {
				completeValues.addItemAt(themedSubvalues[j-1], desiredCurrentIndex);
			}
			
			filteredSubvalues = filteredValues.toArray();
			completeSubvalues = completeValues.toArray();
			
			return didRemoveActiveValue ? matchingSubvalue : pendingActiveValue;
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function getXMLElementName():String
		{
			return XML_ELEMENT_NAME;
		}
		
		private var didBindToThemes:Boolean = false;
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function loadFromXML(valueXML:XML):Boolean
		{
			if (!didBindToThemes) {
				BindingUtils.bindSetter(setThemeList, PropertyManager.instance, ["commonProperties", "skinList"], false, true);
				didBindToThemes = true;
			}
			return super.loadFromXML(valueXML);
		}
		
		/**
		 *  @inheritDoc
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override protected function setXMLElements(propertyXML:XML):void
		{
			if (propertyXML == null) {
				return;
			}
			
			var oldSubvalues:Array = _subvalues;
			
			// Repopulate subvalues with our complete set before saving
			_subvalues = completeSubvalues;
			super.setXMLElements(propertyXML);
			
			_subvalues = oldSubvalues;
		}
		
		/**
		 *  Set the theme filtered list's active subvalue (the value that is being filtered out)
		 *  equal to the specified string.
		 * 
		 *  @param valueString The desired value to filter out, expressed as a string.
		 *  @return Returns <code>true</code> if the specified value is being filterd out, 
		 *  otherwise it returns <code>false</code>
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 *
		 */	
		override public function setValueFromString(valueString:String):Boolean
		{
			if (super.setValueFromString(valueString)) {
				return true;
			}
			// Check to see if the desired value is being filtered out
			if (themedSubvalues != null) {
				for each (var subvalue:PropertyValue in themedSubvalues) {
					if (subvalue.name == valueString) {
						activeValue = subvalue;
						return true;
					}
				}
			}
			return false;
		}
	}
}