/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.utils
{
	import com.adobe.portfolio.api.FieldType;
	import com.adobe.portfolio.api.FieldValueChangedEvent;
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.components.supportClasses.AttachmentFieldSort;
	
	import mx.collections.ArrayCollection;
	import mx.collections.ArrayList;
	import mx.collections.IList;
	import mx.collections.ListCollectionView;
	import mx.collections.SortField;
	import mx.events.CollectionEvent;
	import mx.events.CollectionEventKind;
	
	/**
	 * The AttachmentCollection class extends ListCollectionView to support a collection of IAttachment objects. 
	 * 
	 * Its source list can be changed dynamically so that the same object can effectively be redirected
	 * to different collections with minimal overhead for clients.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class AttachmentCollection extends ListCollectionView
	{
		/**
		 * Constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */ 
		public function AttachmentCollection(source:IList=null)
		{
			if (!source)
				source = new ArrayList();
			super(source);
			
			// Default to ascending sort by file name (if no other sort specified by the host)
			sort = new AttachmentFieldSort();
			sort.fields = [new SortField(FieldType.FILE_NAME, true)];
			
			addEventListener(CollectionEvent.COLLECTION_CHANGE, onOurItemsChanged, false, 0, true);
		}
		/**
		 * @private
		 * The host items of this collection.
		 */
		private var _hostItems:IList;
		public function set hostItems(value:IList):void
		{
			if (_hostItems != null) {
				_hostItems.removeEventListener(CollectionEvent.COLLECTION_CHANGE, onHostItemsChanged);
			}
			
			_hostItems = value;
			if (_hostItems != null) {
				_hostItems.addEventListener(CollectionEvent.COLLECTION_CHANGE, onHostItemsChanged);
			}
			
			var attachment:IAttachment;
			
			// Remove existing event listeners...
			for each (attachment in list)
			{
				attachment.removeEventListener("fieldValueChanged", onFieldValueChanged);
			}
			
			list = new ArrayCollection((_hostItems == null) ? null : _hostItems.toArray());
			
			// ...and add new event listeners
			for each (attachment in list)
			{
				attachment.addEventListener("fieldValueChanged", onFieldValueChanged);
			}
		}

		/**
		 * Listener to handle collection change events on the attachment collection.
		 * Adds field listeners to the new attachments.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onOurItemsChanged(evt:CollectionEvent):void
		{
			var attachment:IAttachment;
			switch(evt.kind)
			{
				case CollectionEventKind.ADD:
					for each( attachment in evt.items)
					{
						attachment.addEventListener("fieldValueChanged", onFieldValueChanged);
					}
//					refresh();  // removing this call because it kills performance drastically
					break;
				case CollectionEventKind.REMOVE:
					for each( attachment in evt.items)
					{
						attachment.removeEventListener("fieldValueChanged", onFieldValueChanged);
					}
//					refresh();	// removing this call because it kills performance drastically
					break;
			}
		}

		/**
		 * Listener to handle collection change events on the hostItems, and propogate them to this collection.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onHostItemsChanged(evt:CollectionEvent):void
		{
			if(evt && evt.target) {
				var attachment:IAttachment;
				switch (evt.kind) {
					case CollectionEventKind.MOVE:
					case CollectionEventKind.UPDATE:
						// do nothing
						break;
					case CollectionEventKind.REFRESH:
						// do nothing
						break;
					case CollectionEventKind.ADD:
						for each (attachment in evt.items) {
							addItem(attachment);
						}
						break;
					case CollectionEventKind.REMOVE:
						for each (attachment in evt.items) {
							if (!contains(attachment)) {
								continue;
							}
							var itemIndex:int = getItemIndex(attachment);
							if (itemIndex >= 0) {
								removeItemAt(itemIndex);
							}
						}
						break;
					case CollectionEventKind.RESET:
						hostItems = (evt.target as IList);
						break;
				}
			}
		}
		
		/**
		 * Listener to handle a change in the field value of an attachment. 
		 * Uses the internal sort function to resort the collection based on the new field value.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		protected function onFieldValueChanged(evt:FieldValueChangedEvent):void
		{
			var attachmentSort:AttachmentFieldSort = sort as AttachmentFieldSort;
			if (!attachmentSort || !attachmentSort.fields || (attachmentSort.fields.length <= 0))
				return;
			for each ( var sortField:SortField in attachmentSort.fields)
			{
				if (sortField.name == evt.fieldName)
				{
					refresh();
					break;
				}
			}
		}
	}
}