/*************************************************************************
 *
 * ADOBE SYSTEMS INCORPORATED
 *  Copyright 2011 Adobe Systems Incorporated
 *  All Rights Reserved.
 *
 * NOTICE:  Adobe permits you to use, modify, and distribute this 
 * file in accordance with the terms of the Adobe license agreement 
 * accompanying it.  If you have received this file from a source 
 * other than Adobe, then your use, modification, or distribution 
 * of it requires the prior written permission of Adobe.
 **************************************************************************/

package com.adobe.portfolio.utils
{
	import flash.display.Stage;
	import flash.events.MouseEvent;
	import flash.events.TimerEvent;
	import flash.utils.Timer;
	
	import mx.core.UIComponent;

	/**
	 * The ClickHandler class is a utility class that allows components to dispatch click events only
	 * when certain time or drag limits are not exceeded.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class ClickHandler
	{
		protected var component:UIComponent;
		protected var singleClickHandler:Function;
		protected var startTimer:Timer;
		protected var timeLimit:uint;
		protected var translationLimit:uint;
		
		protected var disallowClick:Boolean;
		
		protected var listeningForUp:Boolean;
		protected var listeningForMove:Boolean;
		protected var stageX:Number;
		protected var stageY:Number;
		
		/**
		 *  Constructor.
		 * 
		 * 	@param component The host component (listen to its events)
		 * 	@param singleClickHandler The function to call if a single click is detected.  The function takes the form: singleClickHandler(e:MouseEvent):void.
		 * 	@param translationLimit The number of units the user must drag, either horizontally or vertically, before a click event is canceled.
		 * 	@param timeLimit The number of milliseconds the user must delay before a click event is canceled.
		 * 	@param priority The event priority to use for the mouse event listeners on component.
		 */
		// Note: We'd like to use EventPriority.DEFAULT as a default for that last param, but it breaks PyFlexBuild
		public function ClickHandler(component:UIComponent, singleClickHandler:Function, translationLimit:uint = 5, timeLimit:uint = 0, priority:int = -50):void
		{
			this.timeLimit = timeLimit;
			if (timeLimit > 0) {
				startTimer = new Timer(timeLimit, 1);
				startTimer.addEventListener(TimerEvent.TIMER, timeoutHandler, false, 0, true);
			}
			this.translationLimit = translationLimit;
			this.component = component;
			this.singleClickHandler = singleClickHandler;
			if ((component != null) && (singleClickHandler != null)) {
				component.addEventListener(MouseEvent.CLICK, mouseClickHandler, false, priority, true);
				component.addEventListener(MouseEvent.MOUSE_DOWN, mouseDownHandler, false, 0, true);
			}
		}
		
		/**
		 *  Removes all event listeners.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function removeAllListeners():void
		{
			if ((component != null) && (singleClickHandler != null)) {
				component.removeEventListener(MouseEvent.CLICK, mouseClickHandler);
				component.removeEventListener(MouseEvent.MOUSE_DOWN, mouseDownHandler);
			}
			if (startTimer != null) {
				startTimer.removeEventListener(TimerEvent.TIMER, timeoutHandler);
			}
			if (listeningForUp) {
				listeningForUp = false;
				component.stage.removeEventListener(MouseEvent.MOUSE_UP, mouseUpHandler, false);
			}
			if (listeningForMove) {
				listeningForMove = false;
				component.stage.removeEventListener(MouseEvent.MOUSE_MOVE, mouseMoveHandler, false);
			}
		}
		
		protected function mouseClickHandler(event:MouseEvent):void 
		{
			if (disallowClick) {
				return;
			}
			singleClickHandler(event);
		}
		
		protected function mouseDownHandler(event:MouseEvent):void 
		{
			disallowClick = false;
			stageX = event.stageX;
			stageY = event.stageY;
			if (translationLimit > 0) {
				component.stage.addEventListener(MouseEvent.MOUSE_MOVE, mouseMoveHandler, false, 0, true);
				listeningForMove = true;
				component.stage.addEventListener(MouseEvent.MOUSE_UP, mouseUpHandler, false, 0, true);
				listeningForUp = true;
			}
			if (startTimer != null) {
				startTimer.reset();
				startTimer.start();
			}
		}
		
		protected function mouseUpHandler(event:MouseEvent):void 
		{
			listeningForUp = false;
			component.stage.removeEventListener(MouseEvent.MOUSE_UP, mouseUpHandler, false);
			if (listeningForMove) {
				listeningForMove = false;
				component.stage.removeEventListener(MouseEvent.MOUSE_MOVE, mouseMoveHandler, false);
			}
		}

		protected function mouseMoveHandler(event:MouseEvent):void 
		{
			if ((Math.abs(event.stageX - stageX) > translationLimit) ||
				(Math.abs(event.stageY - stageY) > translationLimit))
			{
				disallowClick = true;
				listeningForMove = false;
				component.stage.removeEventListener(MouseEvent.MOUSE_MOVE, mouseMoveHandler, false);
			}
		}

		protected function timeoutHandler(e:TimerEvent):void {
			disallowClick = true;
		}
	}
}