/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.utils
{
	import com.adobe.portfolio.api.FieldType;
	import com.adobe.portfolio.api.IAttachment;
	import com.adobe.portfolio.api.ICollection;
	import com.adobe.portfolio.api.IField;
	import com.adobe.portfolio.managers.CollectionManager;
	
	import mx.logging.ILogger;

	/**
	 * The FieldUtils class provides a set of utility functions for setting and getting attachment field values.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class FieldUtils
	{
		/**
		 * The string literal used for the field name of the field, "Display Name". 
		 * This is different from the string that is used in the UI to display this field.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const DISPLAY_NAME_FIELD_NAME:String = "adobe:DisplayName";
		/**
		 * The string literal used for the field name of the field, "Tags". 
		 * This is different from the string that is used in the UI to display this field.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static const TAGS_FIELD_NAME:String = "adobe:Tags";

		/**
		 * Empty constructor
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function FieldUtils()
		{
		}
		
		/**
		 * setFieldValue sets the value of the given feld and attachment.
		 * @param item The attachment whose field value is to be set
		 * @param field The field whose value is to be set.
		 * @param value The value to be set.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function setFieldValue(item:IAttachment, field:IField, value:*):void {
			try {
				if (item == null || field == null)
					return;
				var propertyName:String = getAttachmentPropertyName(field);
				if (propertyName)
					item[propertyName] = value;
				else
					item.setFieldValue(field.name, value);
			} catch (e:Object) {
				LogHelper.logger.error("FieldUtils:setFieldValue"+e.toString());
			}
		}
		
		/**
		 * Get the field value for the given attachment, and the field or its property. If the field is specified, the
		 * property string literal is ignored. If the field is null, the property is used to obtain the field value.
		 * @param item The attachment whose field or property value is to be retrieved.
		 * @param field The field whose value is to be retrieved.
		 * @param prop The string literal representing the field property.
		 * @return The field value for the given attachment and field or property.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function getFieldOrPropertyValue(item:IAttachment, field:IField, prop:String):* {
			try {
				if (item == null)
					return undefined;
				if (field != null)
					return getFieldValue(item, field);
				else
					return item[prop]
			} catch (e:Object) {
				return undefined;
			}
		}
		
		/**
		 * Get the field value for the given attachment and field.
		 * @param item The attachment whose field value is to be retrieved.
		 * @param field The field whose value is to be retrieved.
		 * @return The field value for the given attachment and field.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function getFieldValue(item:IAttachment, field:IField):* {
			try {
				if (item == null || field == null)
					return undefined;
				var propertyName:String = getAttachmentPropertyName(field);
				if (propertyName)
					return item[propertyName];
				else {
					var result:* = item.getFieldValue(field.name);
					if ((field.name == DISPLAY_NAME_FIELD_NAME) && (result == undefined))
						result = item.fileName;
					return result;
				}
			} catch (e:Object) {
				LogHelper.logger.error("FieldUtils:getFieldValue"+e.toString());
				return undefined;
			}
		}
		
		/** Get the field, given its field name.
		 * @param fields The array of fields that contains the field.
		 * @param fieldName The name of the field that is to be retrieved,
		 * @param allowFallback When true, fallback to filename or to one of the built-in field types.
		 * @return The field with the given name.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function getFieldByFieldName(fields:Array, fieldName:String, allowFallback:Boolean = true):IField {
			var fallbackToFileName:Boolean = allowFallback && (fieldName == DISPLAY_NAME_FIELD_NAME);
			var fallbackToType:Boolean = false;
			if (allowFallback && (	(fieldName == FieldType.COMPRESSED_SIZE) ||
									(fieldName == FieldType.CREATION_DATE) ||
									(fieldName == FieldType.DESCRIPTION) ||
									(fieldName == FieldType.FILE_NAME) ||
									(fieldName == FieldType.MOD_DATE) ||
									(fieldName == FieldType.SIZE)))
			{
				// If the name matches one of our recognized built-in types, then check for that as well
				fallbackToType = true;
			}
			var fallbackField:IField = null;
			for each (var field:IField in fields) {
				if (field.name == fieldName)
					return field;
				else if (fallbackToFileName && (field.type == FieldType.FILE_NAME))
					fallbackField = field;
				else if (fallbackToType && (field.type == fieldName))
					fallbackField = field;
			}
			return fallbackField;			
		}
		
		/**
		 * Get the display name of the field, given its field name.
		 * @param fields The array of fields that contains the field.
		 * @param fieldName The name of the field whose display name is to be retrieved,
		 * @param includeColon When true, use the localized string to determine where the colon should be.
		 * @return The display name of the given field
		 * Example: Field with the field name "adobe:Tags" has the display name "Tags"
		 * 
 		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function getDisplayNameByFieldName(fields:Array, fieldName:String, includeColon:Boolean = false):String
		{
			return getDisplayNameByField(getFieldByFieldName(fields, fieldName), includeColon);
		}
		
		/**
		 * Get the display name of the field.
		 * @param field The field whose display name is to be retrieved,
		 * @param includeColon When true, use the localized string to determine where the colon should be.
		 * @return The display name of the given field
		 * Example: Field with the field name "adobe:Tags" has the display name "Tags"
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function getDisplayNameByField(field:IField, includeColon:Boolean = false):String
		{
			var fieldName:String = (field != null) ? field.displayName : "";
			if (includeColon) {
				// Use a localized string to determine where the colon (or whatever symbol) should go
				var hostCollection:ICollection = (CollectionManager.instance.host != null) ? CollectionManager.instance.host.collection : null;
				var colonString:String = (hostCollection == null) ? "%field%:" : hostCollection.getLocalizedString("fieldWithColon", "%field%:");
				fieldName = colonString.replace("%field%", fieldName);
			}
			return fieldName;
		}
		
		/**
		 * Get the attachment's property name, given the field.
		 * @param field The field whose property name needs to be retrieved.
		 * @return The property name for the given field
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function getAttachmentPropertyName(field:IField):String {
			if (field)
			{
				// now, check for type to get correct behavior per Rob
				switch (field.type) {
					case FieldType.FILE_NAME: return "fileName";
					case FieldType.DESCRIPTION: return "description";
					case FieldType.SIZE: return "size";
					case FieldType.COMPRESSED_SIZE: return "compressedSize";
					case FieldType.CREATION_DATE: return "creationDate";
					case FieldType.MOD_DATE: return "modDate";
					case FieldType.STRING:
					case FieldType.DATE:
					case FieldType.NUMBER:
						break;
					default:
						// technically an error condition, but treat as default
						LogHelper.logger.error("ERROR: FieldUtils.getAttachmentPropertyName has field with bogus FieldType");
						break;
				}
			}
			return null;
		}
		
		/** Determines if a given field belongs to one of the default fields. You must use property notation
		 * for the default fields instead of using the getFieldValue/setFieldValue notation.
		 * This also includes icon since that is also a property that is put in grids.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function isPropertyOfAttachment(field:IField):Boolean {
			return getAttachmentPropertyName(field) != null;
		}	
		
		/** Determines if a given field is a user defined field. This is just the opposite of isPropertyOfAttachment.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function isUserDefinedField(field:IField):Boolean
		{
			return getAttachmentPropertyName(field) == null;
		}
		
		/** Get field value as text from the given field
		 * @param attachment The attachment whose field value needs to be retrieved.
		 * @param field The field whose value needs to be retrieved.
		 * @param dateStyle Specifies the style to use when formatting dates. The default is DateTimeStyle.SHORT.
		 * @param timeStyle Specifies the style to use when formatting times. The default is DateTimeStyle.MEDIUM.
		 * @return the field value as text formatted according to the given styles.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function getTextFromField(attachment:IAttachment, field:IField, dateStyle:String = "", timeStyle:String = ""):String
		{
			if ((attachment == null) || (field == null))
				return "";

			var fieldValue:* = attachment.getFieldValue(field.name);
			
			switch (field.type) {
				case FieldType.STRING:
					if (field.name == DISPLAY_NAME_FIELD_NAME) {
						if ((fieldValue == null) || ((fieldValue as String) == "")) {
							// DisplayName is a special case; if it's blank, use the FileName value instead
							fieldValue = attachment.fileName;
						}
					}
					return fieldValue;
				case FieldType.FILE_NAME:
				case FieldType.DESCRIPTION:
					return fieldValue;
					
				case FieldType.DATE:
				case FieldType.MOD_DATE:
				case FieldType.CREATION_DATE:
					var dateValue:Date = fieldValue as Date;
					if (!dateValue)
						break;
					// This uses the system locale to format the date and time
					return GlobalizationUtil.formatDateTime(dateValue, dateStyle, timeStyle);
					
				case FieldType.NUMBER:
					var numberValue:Number = fieldValue as Number;
					if (fieldValue == undefined || isNaN(numberValue))
						break;
					// This uses the system locale to format the number
					return GlobalizationUtil.formatNumber(numberValue);
					
				case FieldType.SIZE:
				case FieldType.COMPRESSED_SIZE:
					var sizeValue:Number = fieldValue as Number;
					if (fieldValue == undefined || isNaN(sizeValue))
						break;
					return GlobalizationUtil.formatSize(sizeValue);
					
				default:
					CONFIG::logging{
						LogHelper.logger.warn("Unknown field type, "+field.type+", for field: "+field);
					}
					break;
			}
			return "";
		}
		
		/**
		 * @private
		 */
		// Wrap a non-styled field in a default style dict
		private static function wrapInStyle(val:*):XML
		{
			var unstyled:String;
			if (val is Date)
				unstyled = (val as Date).toLocaleString();
			else
				unstyled = String(val);
			
			var styled:XML = 
				<TextFlow><p><span>{unstyled}</span></p></TextFlow> ;
			return styled;
		}
			
		
		/** Get the styled version of the field value.
		 * @param attachment The attachment whose field value needs to be retrieved.
		 * @param field The field whose value needs to be retrieved as stylized text.
		 * @return The field value as stylized text.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function getStyledTextFromField(attachment:IAttachment, field:IField):XML
		{
			var fieldText:XML;
			if ((attachment == null) || (field == null)) {
				return fieldText;
			}
			switch (field.type) {
				case FieldType.STRING:
				case FieldType.DESCRIPTION:
					fieldText = attachment.getFieldStyledText(field.name);
					break;
				default:
					var unstyled:String = getTextFromField(attachment, field);
					fieldText = wrapInStyle(unstyled);
					break;
			}
			return fieldText;
		}

		/**
		 * Sets the given text as the field value.
		 * @param attachment The attachment whose field value needs to be set.
		 * @param field The field whose value needs to be set
		 * @param fieldText The text to be set. 
		 * @return True if the field value is successfully set.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function setFieldFromText(attachment:IAttachment, field:IField, fieldText:String):Boolean
		{
			var success:Boolean = false;
			if ((attachment == null) || (field == null)) {
				return success;
			}
			switch (field.type) {
				case FieldType.FILE_NAME:
					// TODO Check for valid characters?
					attachment.fileName = fieldText;
					success = true;
					break;
				case FieldType.DESCRIPTION:
					attachment.description = fieldText;
					success = true;
					break;
				case FieldType.STRING:
					attachment.setFieldValue(field.name, fieldText);
					success = true;
					break;
				case FieldType.DATE:
				case FieldType.MOD_DATE:
				case FieldType.CREATION_DATE:
					CONFIG::logging{
						LogHelper.logger.warn("We don't parse date values yet");
					}
					var parsedDate:Date = new Date(fieldText);
					if (parsedDate != null) {
						// TODO Check for validity?
						attachment.setFieldValue(field.name, parsedDate);
						success = true;
					}
					break;
				case FieldType.NUMBER:
					// This uses the system locale to parse the number
					attachment.setFieldValue(field.name, GlobalizationUtil.parseNumber(fieldText));
					success = true;
					break;
				case FieldType.SIZE:
				case FieldType.COMPRESSED_SIZE:
					attachment.setFieldValue(field.name, parseInt(fieldText));
					success = true;
					break;
				default:
					CONFIG::logging{
						LogHelper.logger.warn("Unknown field type: "+field.type);
					}
					break;
			}
			return success;
		}
		
		/** Give a new display name to a user defined field.
		 * @param collection The collection containing the field
		 * @param fieldName The name of the field whose display name needs to be set.
		 * @param displayName The new display name.
		 * @return True if the display name is successfully set. 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function setFieldDisplayNameFromText(collection:ICollection, fieldName:String, displayName:String):Boolean
		{
			// Can't access current fields (or they don't exist)
			if (collection == null || collection.fields == null)
				return false;
			// Invalid field name
			if (fieldName == null)
				return false;
			// Invalid display name
			if (displayName == null || displayName.length == 0)
				return false;
			// Change must be made in a copy of the fields
			var fieldsToModify:Array = collection.fields.slice();
			// Can't change it if the field does not exist
			var fieldToModify:IField = FieldUtils.getFieldByFieldName(fieldsToModify, fieldName);
			if (fieldToModify == null)
				return false;
			// Can't change it if the field is not a user defined field
			if (!isUserDefinedField(fieldToModify))
				return false;
			fieldToModify.displayName = displayName;
			collection.fields = fieldsToModify;
			return true;
		}
	}
}
