/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.utils
{
	import com.adobe.portfolio.api.INavigatorHost;
	
	import flash.globalization.DateTimeFormatter;
	import flash.globalization.DateTimeStyle;
	import flash.globalization.LocaleID;
	import flash.globalization.NationalDigitsType;
	import flash.globalization.NumberFormatter;
	
	import flashx.textLayout.formats.Direction;
	
	/**
	 *  The GlobalizationUtil class is a set of globalization utility functions.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class GlobalizationUtil
	{
		private static const SIZE_EXTS:Array = ["bytes", "KB", "MB" ,"GB","TB","PB","EB","ZB","YB"];
		private static const ONES_EXT:String = "byte";
		private static const BASE:Number = 1000;
		private static const PRECISION_DIGITS:int = 4;
		
		// Different ways to format negative numbers
		// These aren't defined in the globalization class, so I'm defining them here (for better code readability below)
		private static const NEGATIVE_NUMBER_FORMAT_PARENTHESES:uint = 0;
		private static const NEGATIVE_NUMBER_FORMAT_LEADING:uint = 1;
		private static const NEGATIVE_NUMBER_FORMAT_LEADING_WITH_SPACE:uint = 2;
		private static const NEGATIVE_NUMBER_FORMAT_TRAILING:uint = 3;
		private static const NEGATIVE_NUMBER_FORMAT_TRAILING_WITH_SPACE:uint = 4;
		
		// The table below is a condensed version of the following RTL Unicode scripts:
		// Hebrew    590 - 5FF
		// Arabic    600 - 6FF
		// Syriac    700 - 74F
		// Arabic    750 - 77F
		// Thâna     780 - 7BF
		// Nko       7C0 - 7FF
		// Samaritan 800 - 83F
		// Mandaic   840 - 85F
		// Resv RTL  860 - 89F
		// Arabic    8A0 - 8FF
		// Hebrew Presentation Forms           FB1D - FB4F
		// Arabic Presentation Forms-A         FB50 - FDFF
		// Arabic Presentation Forms-B         FE70 - FEFF
		// Alphabetic and syllabic RTL scripts 10800 - 10FFF
		// Reserved RTL scripts                1E800 - 1EFFF
		// Note: The entries in this list must be ordered (from low to high value)
		private static const RTL_CHAR_CODE_RANGES:Array = [
			{low: 0x00590,	high: 0x008FF},	// Hebrew, Arabic, Syriac, Arabic, Thâna, NKo, Samaritan, Mandiac, Reserved, Arabic
			{low: 0x0FB1D,	high: 0x0FDFF}, // Hebrew Presentation Forms, Arabic Presentation Forms-A
			{low: 0x0FE70,	high: 0x0FEFF}, // Arabic Presentation Forms-B
			// Skip these because we don't parse multi-byte Unicode characters correctly
			// and it would be very rare to encounter any of these characters. I have
			// been told that even the presentation characters (above) are rare, too,
			// but there is no harm in including them.
//			{low: 0x10800,	high: 0x10FFF}, // Alphabetic and syllabic RTL scripts
//			{low: 0x1E800,	high: 0x1EFFF}  // Reserved RTL scripts
		];
		
		/*
		 * Note about the default arguments in the static functions below:
		 * There is a bug in the mxml/actionscript command line compiler which
		 * treats constant default values (such as LocaleID.DEFAULT) in static
		 * functions as invalid syntax. The error returned is the following:
		 * Error: Parameter initializer unknown or is not a compile-time constant.
		 * The bug is filed here: http://bugs.adobe.com/jira/browse/ASC-4065.
		 * To get around that error, the default arguments, which are all Strings,
		 * have been changed to the empty string and the actual default argument
		 * is substituted in the body of the method.
		 */

		/**
		 *  Returns a formatted date and time string using the provided date and time formatting styles, according to the conventions of the specified locale.
 		 *  @see flash.globalization.DateTimeFormatter 
		 *  @see flash.globalization.DateTimeStyle
		 *  @see flash.globalization.LocaleID
 		 *
		 *  @param date Date to be formatted.
		 *  @param dateStyle Specifies the style to use when formatting dates. The default is DateTimeStyle.SHORT.
		 *  @param timeStyle Specifies the style to use when formatting times. The default is DateTimeStyle.MEDIUM.
		 *  @param locale Specifies the locale to use when formatting dates/times. The default is LocaleID.DEFAULT.		 *
		 *  @return Formatted date/time string.
 		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function formatDateTime(date:Date, dateStyle:String = "", timeStyle:String = "", locale:String = ""):String
		{
			if (dateStyle == "")
				dateStyle = DateTimeStyle.SHORT;
			if (timeStyle == "")
				timeStyle = DateTimeStyle.MEDIUM;
			if (locale == "")
				locale = LocaleID.DEFAULT;
			var dtFormatter:DateTimeFormatter = new DateTimeFormatter(locale, dateStyle, timeStyle);
			return dtFormatter.format(date);
		}
		
		/**
		 *  Returns a formatted number string according to the conventions of the specified locale.
		 *  Trailing zeros are supressed.
		 *  @see flash.globalization.NumberFormatter 
		 *  @see flash.globalization.LocaleID
		 *
		 *  @param number Number to be formatted.
		 *  @param locale Specifies the locale to use when formatting the number. The default is LocaleID.DEFAULT.
		 *  @return Formatted number string.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function formatNumber(number:Number, locale:String = ""):String
		{
			if (locale == "")
				locale = LocaleID.DEFAULT;
			var numFormatter:NumberFormatter = new NumberFormatter(locale);
			// Don't display trailing zeros on whole numbers (such as "3.00")
			numFormatter.trailingZeros = false;
			return numFormatter.formatNumber(number);
		}
		
		/**
		 *  Returns a formatted size string according to the conventions of the specified locale.
		 *  The number is normalized to the closest formatted unit size.
		 *  <p>Examples of formatted strings are: 1.445 KB, 12.50 MB, and 200 bytes</p>
		 *  @see flash.globalization.NumberFormatter 
		 *  @see flash.globalization.LocaleID
		 *
		 *  @param size Size to be formatted.
		 *  @param locale Specifies the locale to use when formatting the number. The default is LocaleID.DEFAULT.
		 *  @return Formatted size string.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function formatSize(size:Number, locale:String = ""):String
		{
			if (isNaN(size))
				return "";
			
			if (locale == "")
				locale = LocaleID.DEFAULT;

			var num:Number = size;
			var extIndex:uint = 0;
			while (num > BASE) {
				num /= BASE;
				++extIndex;
			} 
			
			var numStr:String = (extIndex == 0) ? num.toString() : num.toPrecision(PRECISION_DIGITS);
			var rest:String = (extIndex == 0 && num == 1) ? ONES_EXT : SIZE_EXTS[extIndex];
			
			var numFormatter:NumberFormatter = new NumberFormatter(locale);
			
			// France uses Mo instead of MB
			if (numFormatter.actualLocaleIDName.toLowerCase() == "fr-fr")
				rest = rest.replace("B", "o");
			
			// Fix bug #2619544
			// Since we had to format this string ourselves, because NumberFormatter
			// doesn't have a formatting option similar to Number.toPrecision, we
			// have to deal with substituting national digits ourselves, too.
			numStr = useNationalDigitsIn(numStr, numFormatter.digitsType);
			
			return numStr.replace(".", numFormatter.decimalSeparator) + " " + rest;
		}
		
		/**
		 *  Returns a number derived from the passed in string according to the conventions of the specified locale.
		 *  It is assumed that the passed in string had been formatted using the same locale passed to this method.
		 *  @see flash.globalization.NumberFormatter 
		 *  @see flash.globalization.LocaleID
		 *  
		 *  @param numberStr Formatted string representation of a number.
		 *  @param locale Specifies the locale to use when parsing the number. The default is LocaleID.DEFAULT.
		 *  @return Number.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function parseNumber(numberStr:String, locale:String = ""):Number
		{
			if (locale == "")
				locale = LocaleID.DEFAULT;
			var numFormatter:NumberFormatter = new NumberFormatter(locale);
			var result:Number = numFormatter.parseNumber(numberStr);
			if (!isNaN(result))
				return result;
			// The same property is used to define how negative numbers are formatted and parsed.
			// No matter what the output format is for negative numbers, allow user to input
			// the negative number with a leading negative sign.
			numFormatter.negativeNumberFormat = NEGATIVE_NUMBER_FORMAT_LEADING;
			return numFormatter.parseNumber(numberStr);
		}
		
		/**
		 *  Returns a text direction (left to right or right to left) based on the locale of the host.
		 *  @see com.adobe.portfolio.api.INavigatorHost 
		 *  @see flash.globalization.LocaleID
		 *  @see flashx.textLayout.formats.Direction
		 *  
		 *  @param host The <code>INavigator</code> instance that hosts the PDF portfolio navigator.
		 *  @return Text direction.
		 *  
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function localeDirection(host:INavigatorHost):String
		{
			// If we can't query the locale of the host, assume it is left to right
			if (!host)
				return Direction.LTR;
			var locale:LocaleID;
			// Note: The format of host.locale is 'll_CC', whereas LocaleID expects 'll-CC'.
			// Fortunately LocaleID is smart enough to detect and accept variants like this.
			locale = new LocaleID(host.locale);
			return locale.isRightToLeft() ? Direction.RTL : Direction.LTR;
		}
		
		/**
		 *  Returns a text direction (left to right or right to left) based on a text string.
		 *  The text direction is based on the first character in the string.
		 *  If the string is empty, then the direction of the locale is used.
		 *  @see localeDirection
		 *  @see com.adobe.portfolio.api.INavigatorHost 
		 *  @see flashx.textLayout.formats.Direction
		 *  
		 *  @param host The <code>INavigator</code> instance that hosts the PDF portfolio navigator.
		 *  @param str The text string that should be displayed either left to right or right to left.
		 *  @return Text direction.
		 *
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function textDirection(host:INavigatorHost, str:String):String
		{
			if (!str || str.length == 0)
				return localeDirection(host);
			var unicodeChar:Number = str.charCodeAt();
			for (var index:uint = 0; index < RTL_CHAR_CODE_RANGES.length; ++index) {
				// If less than the beginning of the RTL range, it's a LTR character
				if (unicodeChar < RTL_CHAR_CODE_RANGES[index].low)
					return Direction.LTR;
				// If less than or equal to the end of the RTL range, it's a RTL character
				if (unicodeChar <= RTL_CHAR_CODE_RANGES[index].high)
					return Direction.RTL;
			}
			// Outside of the table, it's a LTR character
			return Direction.LTR;
		}
		
		// This method substitutes all European digits ("0" - "9") for the national digit types passed in.
		private static function useNationalDigitsIn(str:String, digitsType:Number):String
		{
			// If the requested digits type is EUROPEAN ("0" - 9") then done
			if (digitsType == NationalDigitsType.EUROPEAN)
				return str;
			var result:String = "";
			for (var index:int = 0; index < str.length; ++index) {
				var charCode:Number = str.charCodeAt(index);
				if ((charCode >= NationalDigitsType.EUROPEAN) && (charCode <= NationalDigitsType.EUROPEAN + 9))
					charCode = charCode - NationalDigitsType.EUROPEAN + digitsType;
				result += String.fromCharCode(charCode);
			}
			return result;
		}
	}
}