/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.utils
{
	import com.adobe.portfolio.events.JobQueueEvent;
	
	import flash.events.EventDispatcher;
	import flash.events.TimerEvent;
	import flash.utils.Dictionary;
	import flash.utils.Timer;
	
	import mx.collections.ArrayCollection;
	import mx.collections.Sort;
	import mx.logging.ILogger;
	import mx.logging.Log;
	
	/**
	 *  Dispatched when the   queue empties.
	 *
	 *  @eventType com.adobe.portfolio.events.JobQueueEvent
	 *  
	 */
	[Event(name="queueIsEmpty", type="com.adobe.portfolio.events.JobQueueEvent")]
	
 	/**
	 *  The JobQueue class implements a priority job queue that runs jobs sequentially.
	 * 
	 *  <p>JobQueue implements a queue that is particularly useful for deferred computation of list display elements, 
	 *  and can also be used for tasks that need to be done serially.</p>
	 *  <p>JobQueue can be either FILO (first in, last out) or FIFO (first in, first out).</p>
	 *	
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class JobQueue extends EventDispatcher
	{
		private static const logger:ILogger = Log.getLogger("JobQueue");
		
		private var jobDict:Dictionary;
		private var jobStack:ArrayCollection;
		private var jobTimer:Timer;
		private var timerInterval:int;
		private var autoStart:Boolean;
		private var jobNumber:int;
		private var _queueName:String;
		private var _fifo:Boolean; 
		private var id:String;
		private var _pause:Boolean;
		private var _minExecutionPriority:int; 
		
		/**
		 *  Constructor.
		 * 
		 *  @param queueName Optional descriptive name of the queue
		 *  @param fifo Optional boolean to control fifo behavior (true for fifo, false for filo)
		 *  @param autoStart Optional boolean to control job start behavior - 
		 * 		set to true to have JobQueue automatically start the next job in the queue.  
		 * 		Using autoStart == true is generally not a good idea for asynchronous jobs.
		 *  @param timerInterval Time in msec before the next job is begun (after completion of a previous job).
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function JobQueue(queueName:String = "", fifo:Boolean = false, autoStart:Boolean = true, timerInterval:int = 50)
		{
			this.fifo = fifo;
			this.timerInterval = timerInterval;
			this.autoStart = autoStart;
			_queueName= queueName;
			this.id = queueName + "_" + new Date().valueOf();
			jobNumber = 0;
			_minExecutionPriority = int.MIN_VALUE;
		}

		/**
		 *  Push a job onto the queue.
		 * 
		 *  @param jobKey A unique key for the job. 
		 *  @param job An object that generally contains the parameters of the job
		 *  @param callBack A function that is called to "execute" the job.  It takes the form:
		 * 					callback(jobKey:Object, job:Object):void
		 *  @param jobPriority The priority of the job.  The fifo property dictates the order of execution for
		 * 						all jobs with the same priority.  Jobs with a higher priority get executed before jobs
		 * 						with a lower priority.
		 * 
		 *  If a job is pushed onto the queue with a key that matches an existing job in the queue, then
		 *  the new job takes it's place, and the old job is discarded.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function pushJob(jobKey:Object, job:Object, callBack:Function, jobPriority:int = 0):void 
		{
			if (jobKey == null)
				throw new Error("JobQueue.pushJob(): invalid jobKey");
			
			if (!jobDict)
				jobDict = new Dictionary();
			if (!jobStack) {
				jobStack = new ArrayCollection();
				jobStack.sort = jobSort();
			}
			
			var myJobNumber:int;
			if (jobDict[jobKey])  { // remove existing duplicate request 
				CONFIG::logging{
					logger.debug("dup request: " + jobKey);
				}
				myJobNumber = jobDict[jobKey].jobNumber;
				removeJob(jobKey);
			} else {
				myJobNumber = ++jobNumber;
			}
			
			jobDict[jobKey] = {job:job, callBack:callBack, jobPriority:jobPriority, jobNumber:myJobNumber};
			jobStack.addItem(jobKey);
			jobStack.refresh();
			
			if (autoStart)
				startJobTimer();
		}
		
		protected function jobSort():Sort
		{
			var sort:Sort = new Sort();
			sort.compareFunction = 
				function (a:Object, b:Object, fields:Array = null):int {
					var priorityDelta:int = jobDict[b].jobPriority - jobDict[a].jobPriority; // reverse priority (high is first)
					if (priorityDelta != 0)
						return (priorityDelta > 0) ? 1 : -1;
					
					var jobNumberDelta:int = jobDict[a].jobNumber - jobDict[b].jobNumber;
					if (!fifo)
						jobNumberDelta = -jobNumberDelta;
					
					if (jobNumberDelta == 0)
						return 0;
					else
						return (jobNumberDelta > 0) ? 1 : -1;
				};
			return sort;
		}
		
		/**
		 *  Remove a job from the queue.
		 * 
		 *  @param jobKey The key for the job. 
		 *  @return True if the job was found and removed
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function removeJob(jobKey:Object):Boolean {
			
			if (!jobDict)
				return false;
				
			var isRunning:Boolean = jobTimer && jobTimer.running;
			if (isRunning)
				jobTimer.stop();
			var index:int = jobStack.getItemIndex(jobKey);
			if (index != -1)
				jobStack.removeItemAt(index);
			else 
				throw new Error("dict & stack are unbalanced");
			delete jobDict[jobKey];
			if (isRunning)
				jobTimer.start();
				
			return index != -1;
		}
		
		/**
		 *  Whether the queue has pending jobs.
		 * 
		 *  @return True if the queue has pending jobs.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function hasJobs():Boolean
		{
			return jobStack && jobStack.length > 0;
		}
		
		/**
		 *  Kills all pending jobs.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function killAllJobs():void {
			if (!jobTimer)
				return;
			
			jobTimer.stop();
			jobTimer = null;
			jobDict = null;
			jobStack = null;
		}
		
		/**
		 *  Start the queue to process the next job.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function startJobTimer():void {
			if (!jobTimer) {
				jobTimer = new Timer(timerInterval);
				jobTimer.addEventListener(TimerEvent.TIMER, 
					function (event:TimerEvent):void {
						jobTimer.stop();
						var jobKey:Object = getJobKey();
						if (jobKey != null) {
							var job:Object = jobDict[jobKey];
							removeJob(jobKey);
							CONFIG::logging{
								logger.debug(queueName + " starting " + jobKey);
							}
							job.callBack(jobKey, job.job);
							if (autoStart  && !_pause)
								jobTimer.start();
						} else if (jobStack == null || jobStack.length == 0) {
							CONFIG::logging{
								logger.debug(queueName + ": completed jobs");
							}
							jobTimer = null;
							dispatchEvent(new JobQueueEvent(JobQueueEvent.QUEUE_IS_EMPTY));
						}
					}
				);
			}
			if (!jobTimer.running && !_pause)
				jobTimer.start();
		}
		
		/**
		 *  Return the key of the next job to be executed.
		 * 
		 *  @return The key of the next job to be executed.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getJobKey():Object {
			if (!jobStack)
				return null;
			if (jobStack.length == 0)
				return null;
			
			var jobKey:Object;
			for (var i:int = 0; i < jobStack.length; ++i) {
				jobKey = jobStack.getItemAt(i);
				var job:Object = jobDict[jobKey];
				if (job.jobPriority < minExecutionPriority) {
					jobKey = null;
				} else {
					break;
				}
			}
			
			return jobKey;
		}
		
		/**
		 *  Get the job associated with the specified job key.
		 * 
		 *  @return The job associated with the specified job key.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getJob(jobKey:Object):Object {
			return jobDict[jobKey].job;
		}
		
		/**
		 *  Pause (or start) the job queue.
		 * 
		 *  @param val True to pause the queue, false to start it.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function set pause(val:Boolean):void 
		{
			if (_pause == val)
				return;
			
			_pause = val;
			CONFIG::logging{
				logger.debug("pause= " + _pause);
			}
			if (_pause) {
				if (jobTimer)
					jobTimer.stop();
			} else {
				startJobTimer();
			}
		}
		
		/**
		 *  The queue name.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get queueName():String {
			return _queueName;
		}

		/**
		 *  The queue ordering:  fifo = true -> fifo queuing;  fifo = false -> filo queuing
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get fifo():Boolean
		{
			return _fifo;
		}

		public function set fifo(value:Boolean):void
		{
			_fifo = value;
		}

		/**
		 *  The minimum execution priority.  Jobs must have at least this priority to be executed
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get minExecutionPriority():int
		{
			return _minExecutionPriority;
		}

		public function set minExecutionPriority(value:int):void
		{
			_minExecutionPriority = value;
		}

		
		// debug code
		//		private function validate():void {
		//			var dictSize:int = dictSize(jobDict);
		//			var stackSize:int = jobStack.length;
		//			if (dictSize != stackSize) {
		//				var stop:Boolean = true;
		//			}
		//		}
		//		
		//		private function dictSize(dict:Dictionary):int {
		//			var size:int = 0;
		//			for each (var key:Object in dict ) {
		//				size++;
		//			}
		//			return size;
		//		}
	}
}