/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.utils
{
	import flash.geom.Matrix3D;
	import flash.geom.PerspectiveProjection;
	import flash.geom.Point;
	import flash.geom.Rectangle;
	import flash.geom.Utils3D;
	import flash.geom.Vector3D;

	/**
	 * The MathUtil class is a set of math utility functions.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 *
	 */
	public class MathUtil
	{
		/**
		 *  Returns the dimensions that will fit within the specified maximum width and height.  The dimensions will
		 *  have the same aspect ratio as the input dimensions (the size parameter)
		 * 
		 *  @param size a Point that contains the input dimensions
		 *  @param maxWidth the maximum width of the output
		 *  @param maxHeight the maximum height of the output
		 * 
		 *  @return a Point that contains the largest dimensions that satisfy the constraint
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function fitSize(size:Point, maxWidth:Number, maxHeight:Number):Point {
			
			if (!size)
				return null;
			
			var scale:Number =  Math.min(
				maxWidth / size.x,
				maxHeight / size.y);
			
			return new Point(size.x * scale, size.y * scale);
		}
		
		/**
		 *  Clips an input value to the specified range.
		 * 
		 *  @param value the input value to be clipped
		 *  @param min the minimum value 
		 *  @param max the maximum value
		 * 
		 *  @return a value clipped to the range [min, max] (inclusive)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function clipToRange(value:Number, min:Number, max:Number):Number {
			if (value < min)
				value = min;
			else if (value > max)
				value = max;
			
			return value;
		}
		
		/**
		 *  Returns the projected bounds of a rectangle given a matrix and a perspective projecion.
		 * 
		 *  @param bounds The rectangle
		 *  @param matrix The matrix
		 *  @param matrix The projection
		 *  @return The projected bounds of the specified rectangle
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function projectedBounds(bounds:Rectangle,
											   matrix:Matrix3D,
											   projection:PerspectiveProjection):Rectangle
		{
			// Setup the matrix
			if (projection) {
				var centerX:Number = projection.projectionCenter.x;
				var centerY:Number = projection.projectionCenter.y;
				matrix.appendTranslation(-centerX, -centerY, projection.focalLength);
				matrix.append(projection.toMatrix3D());
			}
			
			// Project the corner points
			var pt1:Vector3D = new Vector3D(bounds.left, bounds.top, 0);
			var pt2:Vector3D = new Vector3D(bounds.right, bounds.top, 0)
			var pt3:Vector3D = new Vector3D(bounds.left, bounds.bottom, 0);
			var pt4:Vector3D = new Vector3D(bounds.right, bounds.bottom, 0);
			pt1 = Utils3D.projectVector(matrix, pt1);
			pt2 = Utils3D.projectVector(matrix, pt2);
			pt3 = Utils3D.projectVector(matrix, pt3);
			pt4 = Utils3D.projectVector(matrix, pt4);
			
			// Find the bounding box in 2D
			var maxX:Number = Math.max(Math.max(pt1.x, pt2.x), Math.max(pt3.x, pt4.x));
			var minX:Number = Math.min(Math.min(pt1.x, pt2.x), Math.min(pt3.x, pt4.x));
			var maxY:Number = Math.max(Math.max(pt1.y, pt2.y), Math.max(pt3.y, pt4.y));
			var minY:Number = Math.min(Math.min(pt1.y, pt2.y), Math.min(pt3.y, pt4.y));
			
			// Set the bounds
			bounds.x = minX;
			bounds.y = minY;
			bounds.width = maxX - minX;
			bounds.height = maxY - minY;
			
			// Add back the projection center
			if (projection) {
				bounds.x += centerX;
				bounds.y += centerY;
			}
			
			return bounds;
		}
		
	}
}