/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.utils
{
	import com.adobe.portfolio.api.ICollection;
	import com.adobe.portfolio.events.CollectionManagerEvent;
	import com.adobe.portfolio.managers.CollectionManager;
	import com.adobe.portfolio.managers.ResourceManager;
	
	import flash.events.Event;
	import flash.events.IOErrorEvent;
	import flash.events.SecurityErrorEvent;
	import flash.net.URLLoader;
	import flash.net.URLLoaderDataFormat;
	import flash.net.URLRequest;
	import flash.utils.ByteArray;
	
	import mx.events.PropertyChangeEvent;
	
	/**
	 * The NavURLLoader class extends URLLoader to provide URL loading for Navigators.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class NavURLLoader extends URLLoader
	{
		private var cachedRequest:URLRequest;

		private var _url:String;
		/**
		 * The URL string to be loaded. Setting this value will immediately call load.
		 */
		public function get url():String
		{
			return _url;
		}
		/**
		 * @private
		 */
		public function set url(value:String):void
		{
			if (value != _url) {
				_url = value;
				if ((_url != null) && (_url != "")) {
					load(new URLRequest(_url));
				}
			}
		}
		
		private var _dataBytes:ByteArray;
		/**
		 * A ByteArray representing the bytes loaded from the URL. This value is only
		 * updated for binary requests.
		 * 
		 * @see URLLoader#dataFormat
		 */
		[Bindable (name="propertyChange")]
		public function get dataBytes():ByteArray
		{
			return _dataBytes;
		}
		
		private var _dataString:String;
		/**
		 * A String representing the text loaded from the URL. This value is only updated
		 * for text requests.
		 * 
		 * @see URLLoader#dataFormat
		 */
		[Bindable (name="propertyChange")]
		public function get dataString():String
		{
			return _dataString;
		}
		
		/**
		 * Default constructor
		 */
		public function NavURLLoader(urlString:String=null, binaryFormat:Boolean = false)
		{
			super();
			if (binaryFormat) {
				dataFormat = URLLoaderDataFormat.BINARY;
			}
			if (urlString != null) {
				url = urlString; 
			}
		}
		
		private function onNavHostEvent(event:CollectionManagerEvent):void
		{
			if (event.kind != CollectionManagerEvent.HOST_CHANGED_KIND) {
				return;
			}
			if (event.newVal != null) {
				CollectionManager.instance.removeEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent);
				load(cachedRequest);
				cachedRequest = null;
			}
		}
		
		
		private var _pendingCount:Boolean = false;
		/**
		 * @inheritDoc
		 */
		override public function load(request:URLRequest):void
		{
			if (CollectionManager.instance.host == null) {
				if (cachedRequest == null) {
					CollectionManager.instance.addEventListener(CollectionManagerEvent.COLLECTION_CHANGE, onNavHostEvent);
				}
				cachedRequest = request;
				return;
			}
			// TODO Should _url be updated?
			CONFIG::logging{
			if ((request.url != _url) && (_url != null) && (_url != ""))
				LogHelper.logger.warn("request doesn't match URL property");
			}
			
			if (CollectionManager.instance.host.collection != null) {
				request.url = CollectionManager.instance.host.collection.resolvePath(request.url);
			}
			
			if (!_pendingCount) {
				ResourceManager.instance.incrementLoadingCount();
				_pendingCount = true;
			}
			super.load(request);
		}
		
		/**
		 * Writes the given bytes into the resource file referenced by this loader. <code>dataBytes</code> and
		 * <code>dataString</code> are updated immediately with the new data. The <code>response</code> function
		 * will be called as described by the <code>ICollection</code> intefaces, and load is also triggered so
		 * that any event listeners present will be called appropriately.
		 * 
		 * @see ICollection#writeResource
		 */
		public function write(buffer:ByteArray, response:Function = null):void
		{
			if ((buffer == null) || (url == null)) {
				LogHelper.logger.error("writeResource called with a null buffer or url");
				return;
			}
			var hostCollection:ICollection = (CollectionManager.instance.host == null) ? null : CollectionManager.instance.host.collection;
			if (hostCollection == null) {
				LogHelper.logger.error("writeResource called with a null host or collection");
				return;
			}
			
			if (dataFormat == URLLoaderDataFormat.BINARY) {
				var oldBytes:ByteArray = _dataBytes;
				_dataBytes = buffer;
				super.dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "dataBytes", oldBytes, _dataBytes));
			} else if (dataFormat == URLLoaderDataFormat.TEXT) {
				var oldString:String = _dataString;
				_dataString = buffer.toString();
				super.dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "dataString", oldString, _dataString));
			} else {
				LogHelper.logger.error("Unexpected dataFormat type: "+dataFormat);
			}
			hostCollection.writeResource(buffer, url, function (result:Object):void {
				// Call load to stay in sync with the resource bytes
				load(new URLRequest(url));
				if (response != null) {
					response(result);
				}
			});
		}
		
		/**
		 * @inheritDoc
		 */
		override public function dispatchEvent(event:Event):Boolean
		{
			if (event != null) {
				if (event.type == Event.COMPLETE) {
					if (dataFormat == URLLoaderDataFormat.BINARY) {
						var oldBytes:ByteArray = _dataBytes;
						_dataBytes = data as ByteArray;
						super.dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "dataBytes", oldBytes, _dataBytes));
					} else if (dataFormat == URLLoaderDataFormat.TEXT) {
						var oldString:String = _dataString;
						_dataString = data as String;
						super.dispatchEvent(PropertyChangeEvent.createUpdateEvent(this, "dataString", oldString, _dataString));
					} else {
						LogHelper.logger.error("Unexpected dataFormat type: "+dataFormat);
					}
					if (_pendingCount) {
						ResourceManager.instance.decrementLoadingCount();
						_pendingCount = false;
					}
				} else if ((event.type == SecurityErrorEvent.SECURITY_ERROR) ||
						   (event.type == IOErrorEvent.IO_ERROR))
				{
					if (_pendingCount) {
						ResourceManager.instance.decrementLoadingCount();
						_pendingCount = false;
					}
				}
			}
			return super.dispatchEvent(event);
		}
	}
}
