/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.utils
{
	import flash.utils.Dictionary;
	
	import mx.collections.ArrayCollection;
	
	/**
	 * The OrderedDictionary class implements an ordered dictionary.  
	 * Elements can be pushed/popped to/from the end of the list, 
	 * as while as shifted/unshifted to/from the beginning of the list.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class OrderedDictionary 
	{
		private var dictionary:Dictionary;
		private var order:ArrayCollection;
		private var weakKeys:Boolean;
		
		/**
		 *  Constructor
		 * 
		 *  @param weakKeys whether to use weak keys for the OrderedDictionary entries
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function OrderedDictionary(weakKeys:Boolean=false)
		{
			this.weakKeys = weakKeys;
			clear();
		}
		
		/**
		 *  Clears the dictionary of all entries.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function clear():void {
			dictionary = new Dictionary(weakKeys);
			order = new ArrayCollection();
		}
		
		/**
		 *  Returns true if the dictionary has a key that matches the specified key/value pair.
		 * 
		 *  @param key The key
		 *  @param value The value
		 *  @return true if the dictionary contains the key and matches the value.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function isKeyValueMatch(key:Object, value:*):Boolean
		{
			return dictionary[key] == value;
		}
		
		/**
		 *  Returns true if the dictionary has the specified key
		 * 
		 *  @param key The key
		 *  @return true if the dictionary contains the key 
		 * 
		 *  The value of the key must be non-null)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function keyExists(key:Object):Boolean
		{
			return dictionary[key] != null;
		}
		
		/**
		 *  Pushes the key/value pair onto the end of the dictionary
		 * 
		 *  @param key The key
		 *  @param value The value (if null, then the key is used as a value)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function push(key:Object, value:* = null):void 
		{
			if (!value)
				value = key;
			if (isKeyValueMatch(key, value))
				return;
			if (keyExists(key)) {
				remove(key);
			}
			dictionary[key] = value;
			order.addItem(key);
		}
		
		
		/**
		 *  Pops the last key/value pair from the end of the dictionary and returns its value
		 * 
		 *  @return The value
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function pop():* 
		{
			var key:Object = order.removeItemAt(order.length-1);
			if (!key)
				return null;
			var value:Object = dictionary[key];
			delete dictionary[key];
			
			return value;
		}
		
		/**
		 *  Shifts the first key/value pair from the beginning of the dictionary and returns its value
		 * 
		 *  @return The value
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function shift():* 
		{
			if (order.length == 0)
				return;
			
			var key:Object = order.removeItemAt(0);
			if (!key)
				return null;
			var value:Object = dictionary[key];
			delete dictionary[key];
			
			return value;
		}
		
		/**
		 *  Unshifts (inserts) the key/value pair at the beginning of the dictionary.
		 * 
		 *  @param key The key
		 *  @param value The value (if null, then the key is used as a value)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function unshift(key:Object, value:* = null):void
		{
			if (!value)
				value = key;
			if (isKeyValueMatch(key, value))
				return;
			if (keyExists(key)) {
				remove(key);
			}
			dictionary[key] = value;
			order.addItemAt(key, 0);
		}
		
		/**
		 *  Gets the value of the specified key.
		 * 
		 *  @param key The key
		 *  @return The value (null if the key doesn't exist)
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getValue(key:Object):*
		{
			var index:int = order.getItemIndex(key);
			if (index == -1)
				return null;
			
			return dictionary[key];
		}
		
		/**
		 *  Removes a key/value from the dictionary
		 * 
		 *  @param key The key
		 *  @return True if the key exists and the key/value was removed.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function remove(key:Object):Boolean
		{
			var index:int = order.getItemIndex(key);
			if (index == -1)
				return false;
			
			order.removeItemAt(index);
			delete dictionary[key];
			
			return true;
		}
		
		/**
		 *  The length of the dictionary
		 * 
		 *  @return The length of the dictionary.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get length():uint 
		{
			return order.length;
		}
		
		/**
		 *  Get a value by index
		 * 
		 *  @param The index of the desired key/value pair
		 *  @return The value
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function getValueByIndex(index:int):*
		{
			return dictionary[order.getItemAt(index)];
		}
		
		/**
		 *  Gets the keys in the order stored.
		 * 
		 *  @return An array of the keys in the order 
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public function get orderedKeys():Array 
		{
			return order.toArray();
		}
		
	}
}