/*************************************************************************
*
* ADOBE SYSTEMS INCORPORATED
*  Copyright 2010 Adobe Systems Incorporated
*  All Rights Reserved.
*
* NOTICE:  Adobe permits you to use, modify, and distribute this 
* file in accordance with the terms of the Adobe license agreement 
* accompanying it.  If you have received this file from a source 
* other than Adobe, then your use, modification, or distribution 
* of it requires the prior written permission of Adobe.
**************************************************************************/

package com.adobe.portfolio.utils
{
	import com.adobe.portfolio.api.FieldType;
	import com.adobe.portfolio.api.ISort;
	import com.adobe.portfolio.components.supportClasses.AttachmentFieldSort;
	
	import mx.collections.Sort;
	import mx.collections.SortField;
	
	/**
	 * The SortUtil class provides a set of utility functions that can be used by Sort objects.
	 * 
	 *  @langversion 3.0
	 *  @playerversion Flash 10.1
	 *  @playerversion AIR 2.0
	 *  @productversion Portfolio 10.0
	 */
	public class SortUtil
	{
		/**
		 * compareSortArrays checks if the two given sort arrays are equal. 
		 * The equality is determined by comparing the length of the arrays and each sort object in the array.
		 * @return True if the arrays are the equal.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function compareSortArrays(a:Array, b:Array):Boolean 
		{
			if (a == b)
				return true;
			if (a == null || b == null)
				return false;
			
			if (a.length == b.length) {
				for (var i:Number = 0; i < a.length; ++i) {
					if (!compareSort(a[i], b[i]))
						return false;
				}
				return true;
			}
			return false;
		}
		
		/** compareSort checks if the two sorts are equal. The equality is determined by comparing the name and the order of the sort.
		 * @return True if the two sort objects are equal
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function compareSort(a:ISort, b:ISort) : Boolean
		{
			if (a.name != b.name)
				return false;
			if (a.ascending != b.ascending)
				return false;
			return true;
		}
		
		/** Creates a Sort object from a given array of sort fields. 
		 * @param collectionSortFields An array of sort fields. Mostly the fields used to sort the given collection.
		 * @return A mx.collections.Sort object.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function createSortFieldObject(collectionSortFields:Array):Sort
		{
			var result:Sort = new AttachmentFieldSort;
			
			var sortFields:Array = new Array;
			if ((collectionSortFields != null) && (collectionSortFields.length > 0)) { 
				for each(var sort:ISort in collectionSortFields)
				{
					if(sort == null)
						continue;
					var sf:SortField = new SortField(sort.name, true, !sort.ascending);
					sortFields.push(sf);
				}
			} else {
				sortFields.push(new SortField(FieldType.FILE_NAME, true, false));
			}
			result.fields = sortFields;
			return result;
		}
		
		/** Creates an array of Sort fields given a Sort Object. This is the opposite of createSortFieldObject.
		 * @param sort The Sort or ISort object to be converted into an array.
		 * @return An array of sort fields. The fields used to sort a given collection.
		 * 
		 *  @langversion 3.0
		 *  @playerversion Flash 10.1
		 *  @playerversion AIR 2.0
		 *  @productversion Portfolio 10.0
		 */
		public static function createSortArray(sort:Object):Array
		{
			var result:Array = new Array();
			// make our fields...
			if (sort && sort.hasOwnProperty("fields") && sort.fields)
			{
				for each (var sorter:SortField in sort.fields)
				{
					if(sorter == null)
						continue;
					result.push(new SupportSort(sorter.name, !sorter.descending));
				}
			}
			return result;
		}
	}
}

import com.adobe.portfolio.api.ISort;
/** An internal class used to store the name and order of a sort field.
 * 
 *  @langversion 3.0
 *  @playerversion Flash 10.1
 *  @playerversion AIR 2.0
 *  @productversion Portfolio 10.0
 */
class SupportSort implements ISort
{
	private var _name:String = "";
	private var _ascending:Boolean = false;
	
	public function SupportSort(n:String, a:Boolean)
	{
		_name = n;
		_ascending = a;
	}
	
	// ISort method
	public function get name():String
	{
		return _name;
	}
	
	// ISort method
	public function get ascending():Boolean
	{
		return _ascending;
	}
	
}